/*
 * imalib
 * 
 * Copyright 2017 Aberystwyth University
 * 
 * This library and program are free software: you can redistribute them
 * and/or modify them under the terms of the GNU General Public License as 
 * published by the Free Software Foundation, either version 3 of the License,
 * or (at your option) any later version.
 * 
 * The program and library are distributed in the hope that they will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*!
 * \file imalibDocumentation.h
 * \brief A set of random documentation.
 * \author Fred Labrosse <ffl@aber.ac.uk>
 * \copyright 2017 Aberystwyth University
 */

/*!
 * \mainpage imalib
 *
 * imalib is a library developed to perform image processing useful for
 * some of the research done in the Intelligent Robotics research group of the
 * Department of Computer Science, Aberystwyth University.  It is therefore not
 * a general image processing library.
 * 
 * The library is licensed under GPLv3.  See COPYING for the license.
 *
 * You may find the following links useful:
 * - \subpage autotools "Some help on setting up Makefiles and autoconf scripts"
 * - \subpage imalibBibliography "Bibliography"
 */

/*!
 * \page autotools Help on setting up Makefiles and autoconf scripts
 *
 * imalib installs the script "imalib-config" in the prefix/bin directory.
 * The script can be used in Makefiles to tell the compiler what flags are
 * needed for imalib (with the --cflags option) or the linker what libraries
 * are needed (with the option --libs).  Other options will make the script do
 * other things.  Running the script without options will list the possible
 * options.
 *
 * The --cflags option gives a number of defined terms (-D option to compiler)
 * that were used during the compilation of imalib.  These are mentioned in the
 * documentation of the relevant classes.  These defines can be used in your own
 * code if you need to cope with non-availability of these features.
 *
 * The script can also be used in an m4 script to create an autoconf function
 * to make the autoconf scrit automatically detect imalib and handle it
 * properly.  An example m4 script is bellow:
 * \verbatim
dnl Check for imalib
dnl
dnl Usage:
dnl IMALIB_CHECK([ACTION-IF-FOUND [, ACTION-IF-NOT-FOUND]])
dnl
dnl IMALIB_INCLUDES and IMALIB_LIBS will contain the includes and libs
dnl needed.  IMALIB_FLAGS will contain flags with which imalib was compiled
dnl that influence what was compiled in imalib (such as picasso support).
dnl
dnl Author: Fred Labrosse <ffl@aber.ac.uk>.
dnl
AC_DEFUN([IMALIB_CHECK], [
   AC_SUBST(IMALIB_FLAGS)
   AC_SUBST(IMALIB_INCLUDES)
   AC_SUBST(IMALIB_LIBS)
   AC_MSG_CHECKING([for imalib])
   if imalib-config --version >/dev/null 2>/dev/null; then
      IMALIB_INCLUDES="`imalib-config --cflags`"
      IMALIB_LIBS="`imalib-config --libs`"
      imalib_avail=yes
      IMALIB_VERSION="`imalib-config --version`"
      AC_MSG_RESULT([found version $IMALIB_VERSION])
      AC_MSG_CHECKING([if imalib compiled with Picasso support])
      if imalib-config --with-picasso >/dev/null 2>/dev/null; then
         IMALIB_INCLUDES="$IMALIB_INCLUDES -DWITH_PICASSO"
         IMALIB_FLAGS="WITH_PICASSO=1"
         AC_MSG_RESULT([yes])
      else
         AC_MSG_RESULT([no])
      fi
   else
      imalib_avail=no
      AC_MSG_RESULT([no])
   fi

   if test "$imalib_avail" = yes; then
      ifelse($1, , :, $1)
   else
      ifelse($2, , :, $2)
   fi
])\endverbatim
 *
 * The m4 script must be handled by aclocal (by running "aclocal -I m4", if m4
 * is the name of the directory containing m4 scripts) and automake by having
 * the line
 * \verbatim
ACLOCAL_AMFLAGS = -I m4 \endverbatim
 * in the file Makefile.am at the top of the hierarchy of your software.
 *
 * Alternativelly, it should be possible (but not as nice ;-) to have the
 * function IMALIB_CHECK directly defined in your file configure.ac.
 *
 * You then call the function IMALIB_CHECK in your configure.ac file:
 * \verbatim
IMALIB_CHECK(CPPFLAGS="$CPPFLAGS $IMALIB_INCLUDES"
  LIBS="$LIBS $IMALIB_LIBS",
  AC_MSG_ERROR([imalib is needed]))\endverbatim
 * This will automagically modify your variables $CPPFLAGS and $LIBS if imalib
 * is found or output an error message if not found.
 */

/*!
\page imalibBibliography Bibliography

<dl>
  <dt> [\b BL06] \anchor BL06 </dt>
  <dd>
  Dave Binding and Fred Labrosse, "Visual local navigation
  using warped panoramic images".  \em Proceedings \em of \em Towards \em
  Autonomous \em Robotic \em Systems, pages 19-26,
  2006.
  </dd>

  <dt> [\b CLD13] \anchor CLD13 </dt>
  <dd>
  Juan Cao, Frédéric Labrosse, and Hannah Dee, "An evaluation of image-based
  robot orientation estimation".  To appear in \em Proceedings \em of \em
  Towards \em Autonomous \em Robotic \em Systems, 2013.
  </dd>

  <dt> [\b Fai98] \anchor Fai98 </dt>
  <dd>
  Mark D. Fairchild, \em Color \em Appearance \em
  Models, Addison-Wesley, 1998.
  </dd>

  <dt> [\b Hec94] \anchor Hec94 </dt>
  <dd>
  Paul S. Heckbert, "Bilinear Coons Patch Image
  Warping".  In \em Graphics \em Gems \em IV, pages 438-446, 1994.
  </dd>

  <dt> [\b KCB09] \anchor KCB09 </dt>
  <dd>
  Ioannis Katramados, Steve Crumpler, and Toby P. Breckon, "Real-Time
  Traversable Surface Detection by Colour Space Fusion and Temporal Analysis".
  In \em Proceedings \em of \em the \em International \em Conference \em on
  \em Computer \em Vision \em Systems, pages 265-274, Lecture Notes in Computer
  Science, volume 5815, 2009.
  </dd>

  <dt> [\b Lab04] \anchor Lab04 </dt>
  <dd>
  Fred Labrosse, "Visual compass".  In \em Proceedings
  \em of \em Towards \em Autonomous \em Robotic \em Systems, pages 85-92,
  2004.
  </dd>

  <dt> [\b Lab06] \anchor Lab06 </dt>
  <dd>
  Fred Labrosse, "Appearance-based heading
  estimation: the visual compass", Technical Report UWA-DCS-06-048,
  Computer Science Department, University of Wales, Aberystwyth, 2006.
  </dd>

  <dt> [\b Lab07] \anchor Lab07 </dt>
  <dd>
  Fred Labrosse, "Short and long-range visual navigation using warped panoramic
  images". \em Robotics \em and \em Autonomous \em Systems, 55(9),
  pages 675-684, 2007.
  </dd>

  <dt> [\b Mey98] \anchor Mey98 </dt>
  <dd>
  Scott Meyers. Effective C++, Second Edition. Addison-Wesley, 1998.
  </dd>
  
  
  <dt> [\b ML04] \anchor ML04 </dt>
  <dd>
  Tristan Mitchell and Fred Labrosse, "Visual
  homing: a purely appearance-based approach".  In \em Proceedings \em of \em
  Towards \em Autonomous \em Robotic \em Systems, pages 101-108, 2004.
  </dd>
</dl>
*/
