/*
 * imalib
 * 
 * Copyright 2017 Aberystwyth University
 * 
 * This library and program are free software: you can redistribute them
 * and/or modify them under the terms of the GNU General Public License as 
 * published by the Free Software Foundation, either version 3 of the License,
 * or (at your option) any later version.
 * 
 * The program and library are distributed in the hope that they will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*!
 * \file dc1394ImageGrabber.cpp
 * \brief Image grabber classe for firewire (dc1394) frame grabber. 
 * Implementation.
 *
 * \author Fred Labrosse
 * \copyright 2017 Aberystwyth University
 */

#include "imalib/imageRGB.h"
#include "imalib/dc1394ImageGrabber.h"

// #include <math.h>
// #include <stdio.h>
// #include <unistd.h>
#include <sys/times.h>

#include <fstream>
#include <iostream>

#include <boost/program_options.hpp>
namespace po = boost::program_options;

namespace imalib
{

/*!
 * \brief Constructor.
 *
 * \param[in] upsideDown Set to true if the camera is upside-down.
 * \param[in] flip Set to true to horizontally flip the images.
 * \param[in] cameraNumber The number of the camera on the firewire system.
 * \param[in] posX
 * \param[in] posY The position of the viewport of the grabbed part.
 * \param[in] width
 * \param[in] height The width and height of the grabbed area and of the images.
 */
dc1394ImageGrabber::dc1394ImageGrabber(bool upsideDown,
                                       bool flip,
                                       unsigned short cameraNumber,
                                       unsigned short posX,
                                       unsigned short posY,
                                       unsigned short width,
                                       unsigned short height)
: context(NULL),
camera(NULL),
captureRunning(false),
// grabbedImage(NULL)
grabbedFrame(NULL)
{
   init(upsideDown,
        flip,
        cameraNumber,
        posX, posY,
        width, height);
}

/*!
 * \brief Constructor.
 * \param[in] fileName Name of the configuration file.
 */
dc1394ImageGrabber::dc1394ImageGrabber(const std::string& fileName)
: context(NULL),
camera(NULL),
captureRunning(false),
// grabbedImage(NULL)
grabbedFrame(NULL)
{
   // Let's extract the options from the configuration file.
   po::options_description optionsDesc("picassoImageGrabber options");
   unsigned short camNum = 0;
   int poX = 0;
   int poY = 0;
   int wid = IMAGE_DEFAULT_INPUTWIDTH;
   int hei = IMAGE_DEFAULT_INPUTHEIGHT;
   bool upDown = false;
   bool fl = false;
   optionsDesc.add_options()
      ("cameraNumber", po::value<unsigned short>(&camNum)->default_value(camNum), "")
      ("posX", po::value<int>(&poX)->default_value(poX), "")
      ("posY", po::value<int>(&poY)->default_value(poY), "")
      ("width",
       po::value<int>(&wid)->default_value(wid), "")
      ("height",
       po::value<int>(&hei)->default_value(hei), "")
      ("upsideDown", po::value<bool>(&upDown)->default_value(upDown), "")
      ("flip", po::value<bool>(&fl)->default_value(fl), "")
      ;

   po::variables_map vm;
   std::ifstream configFile(fileName.c_str());
   if (configFile.is_open())
   {
      po::store(po::parse_config_file(configFile, optionsDesc), vm);
      po::notify(vm);
      configFile.close();
   }
   else
   {
      std::cerr << "picassoImageGrabber::picassoImageGrabber():\n"
         << "WARNING: could not open configuration file " << fileName
         << "!\n"
         << "\tUsing defaults.\n";
   }

   init(upDown,
        fl,
        camNum,
        poX, poY,
        wid, hei);
}


/*!
 * \brief Destructor.
 */
dc1394ImageGrabber::~dc1394ImageGrabber()
{
//    std::cout << "Stopping the camera...";
//    std::cout.flush();
   if (camera)
   {
      if (captureRunning)
         dc1394_capture_stop(camera);
      dc1394_video_set_transmission(camera, DC1394_OFF);
      dc1394_camera_free(camera);
   }
   if (context)
      dc1394_free(context);

//    delete [] grabbedImage;
   free(grabbedFrame->image); // allocated with alloc in libdc1394.
   delete grabbedFrame;

//    std::cout << " done.\n";
}

/*!
 * \brief Display on the standard output some information about the
 * source of the images.
 *
 * \param[in,out] aStream The stream to which the info is added.
 */
void dc1394ImageGrabber::printInfoAboutSource(std::ostringstream* aStream)
{
   *aStream << "----------------------------------------------\n"
      << "Image source:\n"
      << "\tVendor: " << camera->vendor << std::endl
      << "\tModel: " << camera->model << std::endl
//       << "\tGUID: " << camera->guid << std::endl
      << "\tUnits: " << widthUnit << "x" << heightUnit
      << " at (" << xUnit << "," << yUnit << ")\n"
      << "\tMax size: " << widthMax << "x" << heightMax << std::endl;

      // Could more here.  See dc1394error_t
      // dc1394_camera_print_info(dc1394camera_t *camera, FILE* fd) in
      // libdc1394's control.c

   printBasicInfoAboutSource(aStream);
   *aStream << "----------------------------------------------\n";
}

/*!
 * \brief Get an image from the frame grabber.
 *
 * The image might have been pre-captured (with captureImage()), but
 * if not will be captured here.
 *
 * The images are resized here.
 *
 * \param[out] theImage The image in which to save the grabbed image, if
 * not NULL.
 * \param[out] rawImage The image in which to save the raw image (before any
 * transformation is applied), if not NULL.
 */
void dc1394ImageGrabber::getImage(imageRGB* theImage, imageRGB* rawImage)
{
   dc1394video_frame_t* frame = NULL;
   bool enqueue = false;
   dc1394error_t err;
   dc1394bool_t corruptFrame = DC1394_FALSE;

   // Get the data.
   if (connected)
   {
      err = dc1394_video_set_transmission(camera, DC1394_ON);
      if (err != DC1394_SUCCESS)
      {
         std::cerr << "Error while switching on the transmission:\n"
            << dc1394_error_get_string(err) << std::endl;
         connected = false;
      }
      else
      {
         err = dc1394_capture_dequeue(camera, DC1394_CAPTURE_POLICY_WAIT, &frame);
         if (err != DC1394_SUCCESS)
         {
            std::cerr << "Error while getting an image:\n"
               << dc1394_error_get_string(err) << std::endl;
            connected = false;
         }
         else
         {
            enqueue = true;

            corruptFrame = dc1394_capture_is_frame_corrupt(camera, frame);
            if (corruptFrame == DC1394_TRUE)
            {
               std::cerr << "Corrupt frame: ignoring.\n";
            }
            else
            {
               assert(connected && (frame->size[0] == imgMap->sourceWidth())
                      && (frame->size[1] == imgMap->sourceHeight()));

               err = dc1394_convert_frames(frame, grabbedFrame);
//                err = dc1394_bayer_decoding_8bit(frame->image, grabbedImage,
//                                                 frame->size[0], frame->size[1],
//                                                 DC1394_COLOR_FILTER_GRBG,
//                                                 DC1394_BAYER_METHOD_SIMPLE);
               if (err != DC1394_SUCCESS)
               {
                  std::cerr << "Error while converting an image to RGB8:\n"
                     << dc1394_error_get_string(err) << std::endl;
                  connected = false;
               }
            }
         }
      }
   }

   // Play with the data.
   if (connected && (corruptFrame == DC1394_FALSE))
   {
      unsigned long index;
      if (theImage)
      {
         theImage->setSize(imgMap->targetWidth(),
                           imgMap->targetHeight());
         std::vector<imageMapping::coordsPair>::iterator iter;
         for (iter = imgMap->map.begin();
              iter < imgMap->map.end();
              ++iter)
         {
            index = (iter->sY * imgMap->sourceWidth() + iter->sX) * 3;
            theImage->setPixel(iter->tX, iter->tY,
                               &grabbedFrame->image[index]);
//                                &grabbedImage[index]);
         }
      }
      if (rawImage)
      {
         rawImage->setSize(imgMap->sourceWidth(),
                           imgMap->sourceHeight());
         unsigned long width3 = imgMap->sourceWidth() * 3;
         for (unsigned short indY = 0;
              indY < imgMap->sourceHeight();
              ++indY)
         {
            index = indY * width3;
            for (unsigned short indX = 0;
                 indX < imgMap->sourceWidth();
                 ++indX)
            {
               rawImage->setPixel(indX, indY,
                                  &grabbedFrame->image[index]);
//                                   &grabbedImage[index]);
               index += 3;
            }
         }
      }
   }

   // release buffer
   if (enqueue)
   {
      err = dc1394_capture_enqueue(camera, frame);
      if (err != DC1394_SUCCESS)
      {
         std::cerr << "Error while enqueueing a frame:\n"
            << dc1394_error_get_string(err) << std::endl;
         connected = false;
      }
   }

   if (connected)
   {
      err = dc1394_video_set_transmission(camera, DC1394_OFF);
      if (err != DC1394_SUCCESS)
      {
         std::cerr << "Error while switching off the transmission:\n"
         << dc1394_error_get_string(err) << std::endl;
         connected = false;
      }
   }
}

/*!
 * \brief Set the size of the image to be captured.
 * \param[in] imgWidth
 * \param[in] imgHeight The width and height of the images to grab.
 * \return true if successful, false otherwise.
 */
bool dc1394ImageGrabber::setImageSize(unsigned short imgWidth,
                                      unsigned short imgHeight)
{
   imageWidth = imgWidth;
   imageHeight = imgHeight;

//    if (grabbedImage)
//       delete [] grabbedImage;
//    grabbedImage = new unsigned char[imageWidth * imageHeight * 3];

   return(useInputRectangle());
}

/*!
 * \brief Specify the input rectangle (viewport on the grabbed image).
 * \param[in] x
 * \param[in] y Coordinates of the top-left corner of the rectangle.
 * \param[in] w
 * \param[in] h Size of the rectangle.
 * \return true is successful, false otherwise.
 * \sa useInputRectangle().
 */
bool dc1394ImageGrabber::setInputRectangle(unsigned short x,
                                           unsigned short y,
                                           unsigned short w,
                                           unsigned short h)
{
   inputX = x;
   inputY = y;
   imageWidth = w;
   imageHeight = h;

   return(useInputRectangle());
}

/*!
 * \brief Use the input rectangle currently set.
 * \return true if successful, false otherwise.
 * \sa setInputRectangle().
 */
bool dc1394ImageGrabber::useInputRectangle()
{
   bool success = true;
   bool changedImageSize = false;

//    std::cout << "Trying " << inputWidth << "x" << inputHeight
//       << " at (" << inputX << "," << inputY << ")\n";

   if (imageWidth < widthUnit)
      imageWidth = widthUnit;
   else if (imageWidth > widthMax)
      imageWidth = widthMax;
   else
      imageWidth = roundNearestMultiple(imageWidth, widthUnit);

   if (imageHeight < heightUnit)
      imageHeight = heightUnit;
   else if (imageHeight > heightMax)
      imageHeight = heightMax;
   else
      imageHeight = roundNearestMultiple(imageHeight, heightUnit);

   unsigned short imageWidthOld = imageWidth;
   unsigned short imageHeightOld = imageHeight;

   inputX = roundNearestMultiple(inputX, xUnit);
   inputY = roundNearestMultiple(inputY, yUnit);

   dc1394error_t err;
   if (captureRunning)
   {
      err = dc1394_capture_stop(camera);
      if (err != 0)
      {
         std::cerr << "Error while stopping the capture:\n"
            << dc1394_error_get_string(err) << std::endl;
         success = false;
      }
      else
      {
         captureRunning = false;
      }
   }

//    dc1394switch_t onOff;
//    if (success)
//    {
//       err = dc1394_video_get_transmission(camera, &onOff);
//       if (err != 0)
//       {
//          std::cerr << "Error while getting the transmission status:\n"
//             << dc1394_error_get_string(err) << std::endl;
//          success = false;
//       }
//       else
//       {
//          if (onOff == DC1394_ON)
//          {
//             err = dc1394_video_set_transmission(camera, DC1394_OFF);
//             if (err != 0)
//             {
//                std::cerr << "Error while switching off the transmission:\n"
//                   << dc1394_error_get_string(err) << std::endl;
//                success = false;
//             }
//          }
//       }
//    }

   if (success)
   {
      // Both image and grabbed sizes have to be the same.
      imageGrabber::setImageSize(imageWidth, imageHeight);
//       if (grabbedImage)
//          delete [] grabbedImage;
//       grabbedImage = new unsigned char[imageWidth * imageHeight * 3];

      changedImageSize = true;

      err = dc1394_format7_set_roi(camera, DC1394_VIDEO_MODE_FORMAT7_0,
                                   DC1394_COLOR_CODING_YUV422,
//                                    DC1394_COLOR_CODING_RAW8,
                                   DC1394_USE_MAX_AVAIL, // use max packet size
                                   inputX, inputY, // left, top
                                   imageWidth, imageHeight);  // width, height
      if (err != 0)
      {
         std::cerr << "Error while setting the parameters of the camera:\n"
            << dc1394_error_get_string(err) << std::endl;
         success = false;
      }
   }

   if (success)
   {
      err = dc1394_capture_setup(camera, 4, DC1394_CAPTURE_FLAGS_DEFAULT);
      if (err != 0)
      {
         std::cerr << "Error while setting up the capture:\n"
            << dc1394_error_get_string(err) << std::endl;
         success = false;
      }
      else
      {
         captureRunning = true;
      }
   }

   if (success)
   {
//       err = dc1394_video_set_transmission(camera, DC1394_ON);
//       if (err != DC1394_SUCCESS)
//       {
//          std::cerr << "Error while switching on the transmission:\n"
//            << dc1394_error_get_string(err) << std::endl;
//          success = false;
//       }
//       else
//       {
         std::cout << "Capture rectangle: " << imageWidth << "x" << imageHeight
            << " at (" << inputX << "," << inputY << ")\n";
//       }
   }

   if (!success && changedImageSize)
   {
      imageGrabber::setImageSize(imageWidthOld, imageHeightOld);
//       if (grabbedImage)
//          delete [] grabbedImage;
//       grabbedImage = new unsigned char[imageWidthOld * imageHeightOld * 3];
   }

   return(success);
}

void dc1394ImageGrabber::moveRight()
{
   if (inputX >= xUnit)
   {
      inputX -= xUnit;
      useInputRectangle();
   }
}

void dc1394ImageGrabber::moveLeft()
{
   inputX += xUnit;
   useInputRectangle();
}

void dc1394ImageGrabber::moveUp()
{
   inputY += yUnit;
   useInputRectangle();
}

void dc1394ImageGrabber::moveDown()
{
   if (inputY >= yUnit)
   {
      inputY -= yUnit;
      useInputRectangle();
   }
}

void dc1394ImageGrabber::increaseWidth()
{
   if (imageWidth >= 2 * widthUnit)
   {
      imageWidth -= widthUnit;
      useInputRectangle();
   }
}

void dc1394ImageGrabber::decreaseWidth()
{
   imageWidth += widthUnit;
   useInputRectangle();
}

void dc1394ImageGrabber::increaseHeight()
{
   if (imageHeight >= 2 * heightUnit)
   {
      imageHeight -= heightUnit;
      useInputRectangle();
   }
}

void dc1394ImageGrabber::decreaseHeight()
{
   imageHeight += heightUnit;
   useInputRectangle();
}

/*!
 * \brief Initialise the grabber.
 * \param[in] upsideDown Set to true if the camera is upside-down.
 * \param[in] flip Set to true to horizontally flip the images.
 * \param[in] cameraNumber The number of the camera on the firewire system.
 * \param[in] posX
 * \param[in] posY The position of the viewport of the grabbed part.
 * \param[in] width
 * \param[in] height The width and height of the grabbed area anf of the
 * images.
 */
void dc1394ImageGrabber::init(bool upsideDown,
                              bool flip,
                              unsigned short cameraNumber,
                              unsigned short posX,
                              unsigned short posY,
                              unsigned short width,
                              unsigned short height)
{
   bool cont = true;
   connected = false;

   setParameters(width, height, upsideDown, flip);

   grabbedFrame = new dc1394video_frame_t;
   grabbedFrame->image = NULL;
   grabbedFrame->allocated_image_bytes = 0;
   grabbedFrame->color_coding = DC1394_COLOR_CODING_RGB8;

   context = dc1394_new();
   if (!context)
   {
      cont = false;
      std::cerr << "Error while initialising the dc1394 stack.\n";
   }

   dc1394camera_list_t* cameraList = NULL;
   dc1394error_t err;
   if (cont)
   {
      err = dc1394_camera_enumerate(context, &cameraList);
      if (err != DC1394_SUCCESS)
      {
         std::cerr << "Error while enumarating cameras:\n"
            << dc1394_error_get_string(err) << std::endl;
         cont = false;
      }
      else if (cameraList->num == 0)
      {
         std::cerr << "No camera found.\n";
         cont = false;
      }
   }
   if (cont)
   {
      if (cameraList->num < cameraNumber)
      {
         std::cerr << "The specified camera index (" << cameraNumber
            << ") does not exist.\n";
         cont = false;
      }
   }

   if (cont)
   {
      camera = dc1394_camera_new(context, cameraList->ids[cameraNumber].guid);
      if (!camera)
      {
         std::cerr << "Failed to initialise camera number " << cameraNumber
            << " with guid " << cameraList->ids[cameraNumber].guid << std::endl;
         cont = false;
      }

      dc1394_camera_free_list(cameraList);
      cameraList = NULL; // Just to make sure.
   }

   //
   // Setting parameters (format 7 so that we can change the viewport).
   //
   if (cont)
   {
      err = dc1394_video_set_mode(camera, DC1394_VIDEO_MODE_FORMAT7_0);
      if (err != DC1394_SUCCESS)
      {
         std::cerr << "Error while setting the mode (mode 7) of the camera:\n"
         << dc1394_error_get_string(err) << std::endl;
         cont = false;
      }
   }

   // We first make sure the capture is not already running, in case the
   // constructor was not called last time.
//    err = dc1394_capture_stop(camera);
//    if (err != 0)
//    {
//       std::cerr << "Error while stopping the capture:\n"
//          << dc1394_error_get_string(err) << std::endl;
//       cont = false;
//    }

   err = dc1394_video_set_transmission(camera, DC1394_OFF);
   if (err != 0)
   {
      std::cerr << "Error while switching off the transmission:\n"
         << dc1394_error_get_string(err) << std::endl;
      cont = false;
   }

   // Then we set some parameters.
   if (cont)
   {
      err = dc1394_video_set_iso_speed(camera, DC1394_ISO_SPEED_400);
      if (err != DC1394_SUCCESS)
      {
         std::cerr << "Error while setting the ISO speed of the camera:\n"
            << dc1394_error_get_string(err) << std::endl;
         cont = false;
      }
   }

   // Image sizes.
   if (cont)
   {
      err = dc1394_format7_get_unit_size(camera, DC1394_VIDEO_MODE_FORMAT7_0,
                                         &widthUnit, &heightUnit);
      if (err != DC1394_SUCCESS)
      {
         std::cerr << "Error while getting the unit size of the images:\n"
            << dc1394_error_get_string(err) << std::endl;
         cont = false;
      }
      err = dc1394_format7_get_max_image_size(camera, DC1394_VIDEO_MODE_FORMAT7_0,
                                              &widthMax, &heightMax);
      if (err != DC1394_SUCCESS)
      {
         std::cerr << "Error while getting the max size of the images:\n"
            << dc1394_error_get_string(err) << std::endl;
         cont = false;
      }
      err = dc1394_format7_get_unit_position(camera, DC1394_VIDEO_MODE_FORMAT7_0,
                                             &xUnit, &yUnit);
      if (err != DC1394_SUCCESS)
      {
         std::cerr << "Error while getting the unit position of the images:\n"
            << dc1394_error_get_string(err) << std::endl;
         cont = false;
      }
   }

   if (cont)
   {
      cont = setInputRectangle(posX, posY, width, height);
      // This also initialises the frame grabber.
   }

   // Let's get some control values.
   if (cont)
   {
      err = dc1394_feature_get_boundaries(camera,
                                          DC1394_FEATURE_GAIN,
                                          &minGain, &maxGain);
      if (err != DC1394_SUCCESS)
      {
         std::cerr << "Error while getting the gain boundaries:\n"
            << dc1394_error_get_string(err) << std::endl;
         cont = false;
      }
   }

   if (cont)
   {
      err = dc1394_feature_get_boundaries(camera,
                                          DC1394_FEATURE_SHUTTER,
                                          &minShutter, &maxShutter);
      if (err != DC1394_SUCCESS)
      {
         std::cerr << "Error while getting the shutter boundaries:\n"
            << dc1394_error_get_string(err) << std::endl;
         cont = false;
      }
   }

   if (cont)
   {
      err = dc1394_feature_get_boundaries(camera,
                                          DC1394_FEATURE_EXPOSURE,
                                          &minExposure, &maxExposure);
      if (err != DC1394_SUCCESS)
      {
         std::cerr << "Error while getting the exposure boundaries:\n"
            << dc1394_error_get_string(err) << std::endl;
         cont = false;
      }
   }

   if (!cont)
   {
      connected = false;
      // We may have to stop a few things too.
   }
   else
   {
      connected = true;
   }
}

/*!
 * \brief Set the gain.
 *
 * \param[in] value The new gain value.  A negative value means automatic
 * gain control.
 */
void dc1394ImageGrabber::setGain(long value)
{
   assert(camera);
   dc1394error_t err;
   if (value >= 0)
   {
      if (value > maxGain)
         value = maxGain;
      if (value < minGain)
         value = minGain;

      err = dc1394_feature_set_mode(camera, DC1394_FEATURE_GAIN,
                                    DC1394_FEATURE_MODE_MANUAL);
      if (err != DC1394_SUCCESS)
      {
         std::cerr
            << "Error while setting the manual gain control of the camera:\n"
            << dc1394_error_get_string(err) << std::endl;
      }
      err = dc1394_feature_set_value(camera,
                                     DC1394_FEATURE_GAIN, value);
      if (err != DC1394_SUCCESS)
      {
         std::cerr << "Error while setting the gain of the camera:\n"
            << dc1394_error_get_string(err) << std::endl;
      }
   }
   else
   {
      err = dc1394_feature_set_mode(camera, DC1394_FEATURE_GAIN,
                                    DC1394_FEATURE_MODE_AUTO);
      if (err != DC1394_SUCCESS)
      {
         std::cerr
            << "Error while setting the automatic gain control of the camera:\n"
            << dc1394_error_get_string(err) << std::endl;
      }
   }
}

/*!
 * \brief Get the current gain value.
 * \return The value.
 */
long dc1394ImageGrabber::getGain()
{
   uint32_t value;
   dc1394error_t err = dc1394_feature_get_value(camera,
                                                DC1394_FEATURE_GAIN, &value);
   if (err != DC1394_SUCCESS)
   {
      std::cerr
         << "Error while getting the gain value:\n"
         << dc1394_error_get_string(err) << std::endl;
   }

   return(static_cast<long>(value));
}

/*!
 * \brief Set the exposure.
 *
 * \param[in] value The new exposure value.  A negative value means automatic
 * exposure control.
 */
void dc1394ImageGrabber::setExposure(long value)
{
   assert(camera);
   dc1394error_t err;
   if (value >= 0)
   {
      if (value > maxExposure)
         value = maxExposure;
      if (value < minExposure)
         value = minExposure;

      err = dc1394_feature_set_mode(camera, DC1394_FEATURE_EXPOSURE,
                                    DC1394_FEATURE_MODE_MANUAL);
      if (err != DC1394_SUCCESS)
      {
         std::cerr
            << "Error while setting the manual exposure control of the camera:\n"
            << dc1394_error_get_string(err) << std::endl;
      }
      err = dc1394_feature_set_value(camera,
                                     DC1394_FEATURE_EXPOSURE, value);
      if (err != DC1394_SUCCESS)
      {
         std::cerr << "Error while setting the exposure of the camera:\n"
            << dc1394_error_get_string(err) << std::endl;
      }
   }
   else
   {
      err = dc1394_feature_set_mode(camera, DC1394_FEATURE_EXPOSURE,
                                    DC1394_FEATURE_MODE_AUTO);
      if (err != DC1394_SUCCESS)
      {
         std::cerr
            << "Error while setting the automatic exposure control of the camera:\n"
            << dc1394_error_get_string(err) << std::endl;
      }
   }
}

/*!
 * \brief Get the current exposure value.
 * \return The value.
 */
long dc1394ImageGrabber::getExposure()
{
   uint32_t value;
   dc1394error_t err = dc1394_feature_get_value(camera,
                                                DC1394_FEATURE_EXPOSURE,
                                                &value);
   if (err != DC1394_SUCCESS)
   {
      std::cerr
         << "Error while getting the exposure value:\n"
         << dc1394_error_get_string(err) << std::endl;
   }

   return(static_cast<long>(value));
}

/*!
 * \brief Set the shutter.
 *
 * \param[in] value The new shutter value.  A negative value means automatic
 * shutter control.
 */
void dc1394ImageGrabber::setShutter(long value)
{
   assert(camera);
   dc1394error_t err;
   if (value >= 0)
   {
      if (value > maxShutter)
         value = maxShutter;
      if (value < minShutter)
         value = minShutter;

      err = dc1394_feature_set_mode(camera, DC1394_FEATURE_SHUTTER,
                                    DC1394_FEATURE_MODE_MANUAL);
      if (err != DC1394_SUCCESS)
      {
         std::cerr
            << "Error while setting the manual shutter control of the camera:\n"
            << dc1394_error_get_string(err) << std::endl;
      }
      err = dc1394_feature_set_value(camera,
                                     DC1394_FEATURE_SHUTTER, value);
      if (err != DC1394_SUCCESS)
      {
         std::cerr << "Error while setting the shutter of the camera:\n"
            << dc1394_error_get_string(err) << std::endl;
      }
   }
   else
   {
      err = dc1394_feature_set_mode(camera, DC1394_FEATURE_SHUTTER,
                                    DC1394_FEATURE_MODE_AUTO);
      if (err != DC1394_SUCCESS)
      {
         std::cerr
            << "Error while setting the automatic shutter control of the camera:\n"
            << dc1394_error_get_string(err) << std::endl;
      }
   }
}

/*!
 * \brief Get the current shutter value.
 * \return The value.
 */
long dc1394ImageGrabber::getShutter()
{
   uint32_t value;
   dc1394error_t err = dc1394_feature_get_value(camera,
                                                DC1394_FEATURE_SHUTTER,
                                                &value);
   if (err != DC1394_SUCCESS)
   {
      std::cerr
         << "Error while getting the shutter value:\n"
         << dc1394_error_get_string(err) << std::endl;
   }

   return(static_cast<long>(value));
}

} // namespace imalib.
