/*
 * imalib
 * 
 * Copyright 2017 Aberystwyth University
 * 
 * This library and program are free software: you can redistribute them
 * and/or modify them under the terms of the GNU General Public License as 
 * published by the Free Software Foundation, either version 3 of the License,
 * or (at your option) any later version.
 * 
 * The program and library are distributed in the hope that they will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*!
 * \file fileGrabber.h
 * \brief A class create a "frame grabber" that grabs images from files.
 * (Implementation.)
 * 
 * \author Fred Labrosse
 * \copyright 2017 Aberystwyth University
 */

#include "imalib/imageRGB.h"
#include "imalib/fileGrabber.h"
#include "imalib/imageMapping.h"

#include <iostream>

#include <boost/program_options.hpp>
namespace po = boost::program_options;

namespace imalib
{
/*!
 * \brief Constructor.
 * \param[in] names The names of the files.
 * \param[in] configName Name of the configuration file.
 */
fileGrabber::fileGrabber(const std::vector<std::string>& names,
                         const std::string& configName)
   : imageGrabber(1, 1, false, false),    // Non zero sizes to avoid the 0 size warning.
   imageIndex(0)
{
   if (configName != "")
   {
      // Let's extract the options from the configuration file.  But we
      // only want upsideDown.
      po::options_description optionsDesc("fileGrabber options");
      optionsDesc.add_options()
      ("grabberNumber", "")
      ("devicePort", "")
      ("videoDevice", "")
      ("videoPort", "")
      ("posX", "")
      ("posY", "")
      ("sourceWidth", "")
      ("sourceHeight", "")
      ("imageWidth", "")
      ("imageHeight", "")
      ("upsideDown", po::value<bool>(&_upsideDown)->default_value(false), "")
      ("flip", po::value<bool>(&_flip)->default_value(false), "")
      ;

      po::variables_map vm;
      std::ifstream configFile(configName.c_str());
      if (configFile.is_open())
      {
         po::store(po::parse_config_file(configFile, optionsDesc), vm);
         po::notify(vm);
         configFile.close();
      }
      else
      {
         std::cerr << "fileGrabber::fileGrabber():\n"
            << "WARNING: could not open configuration file " << configName
            << "!\n"
            << "\tUsing defaults.\n";
      }
   }
   else
   {
      _upsideDown = false;
   }

   fileNames = names;
   static imageRGB grabbedImage;
   if (grabbedImage.read(fileNames[imageIndex]))
   {
     imageWidth = grabbedImage.nbColumns();
     imageHeight = grabbedImage.nbRows();
     doUnwrap(_unwrap,
              _outerRadius, _innerRadius,
              _centreX, _centreY,
              _angularRes,
              _unwrappingPower);
     // The doUnwrap above also updates the mapping (therefore uses the
     // _upsideDown setting).

     resetImage(0);
   }
   else
   {
     std::cerr
       << "fileGrabber:: Could not read the first image.  Bailing out.\n";
     connected = false;
   }
}

fileGrabber::~fileGrabber()
{
}

/*!
 * \brief Display on the standard output some information about the source of
 * the images.
 *
 * \param[in,out] aStream The stream to which the info is added.
 */
void fileGrabber::printInfoAboutSource(std::ostringstream* aStream)
{
   *aStream << "----------------------------------------------\n"
      << "Image source:\n"
      << "\tfiles (" << fileNames[0] << "...)\n";
   printBasicInfoAboutSource(aStream);
   *aStream << "----------------------------------------------\n";
}


/*!
 * \brief Get an image from the next file.
 *
 * The image is resized here.
 *
 * \param theImage (OUT) The image in which to save the grabbed image,
 * if not NULL.
 * \param[out] rawImage The image in which to save the raw image (before any
 * transformation is applied), if not NULL.
 */
void fileGrabber::getImage(imageRGB* theImage, imageRGB* rawImage)
{
   if (connected)            // To make sure we still have an image to read.
   {
      static imageRGB tmpImg;
      static imageRGB* grabbedImage = NULL;
      if (rawImage != NULL)
      {
         grabbedImage = rawImage;
      }
      else
      {
         grabbedImage = new imageRGB;
      }
//          std:: cout << fileNames[imageIndex] << std::endl;
      grabbedImage->read(fileNames[imageIndex++]);
      if ((grabbedImage->nbColumns() != imageWidth)
          || (grabbedImage->nbRows() != imageHeight))
      {
         imageWidth = grabbedImage->nbColumns();
         imageHeight = grabbedImage->nbRows();
         doUnwrap(_unwrap,
                  _outerRadius, _innerRadius,
                  _centreX, _centreY,
                  _angularRes,
                  _unwrappingPower);
      }

      tmpImg.setSize(imgMap->targetWidth(),
                     imgMap->targetHeight());
      if (theImage)
      {
         theImage->setSize(imgMap->targetWidth(),
                           imgMap->targetHeight());

         // Apply the mapping (which might be a noop).
         imageRGB* imgPtr;
         if (_blur)
         {
            imgPtr = &tmpImg;
         }
         else
         {
            imgPtr = theImage;
         }

         std::vector<imageMapping::coordsPair>::iterator iter;
         for (iter = imgMap->map.begin();
              iter < imgMap->map.end();
              ++iter)
         {
            imgPtr->setPixel(iter->tX, iter->tY,
                             grabbedImage->getActualPixel(iter->sX,
                                iter->sY));
         }

         // Possibly blur the image.
         if (_blur)
         {
            tmpImg.scaleBlur(_blurWidth, _blurHeight, theImage);
         }
      }

      if (imageIndex >= fileNames.size())
         connected = false;
   }
}

/*!
 * \brief Reset the next image to be grabbed to be the one at the
 * given index in the list of name.
 *
 * If the given index is out of range, then the grabber becomes
 * disconnected.
 *
 * \param index (IN) The index of the next image to be grabbed.
 * \return true if we are connected, false otherwise.
 */
bool fileGrabber::resetImage(unsigned long index)
{
   imageIndex = index;
   if (imageIndex >= fileNames.size())
      connected = false;
   else
      connected = true;

   return(connected);
}

} // namespace imalib

//  LocalWords:  namespace imalib param imageRGB
