/*
 * imalib
 * 
 * Copyright 2017 Aberystwyth University
 * 
 * This library and program are free software: you can redistribute them
 * and/or modify them under the terms of the GNU General Public License as 
 * published by the Free Software Foundation, either version 3 of the License,
 * or (at your option) any later version.
 * 
 * The program and library are distributed in the hope that they will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*!
 * \file grabPairs.cpp
 * \brief A program to grab pairs of images and save them to disk.
 * \author Fred Labrosse <ffl@aber.ac.uk>
 * \copyright 2017 Aberystwyth University
 */

#include "imalib/imageRGB.h"

#include "imalib/imageGrabber.h"
#ifdef WITH_PICASSO
#  include "imalib/picassoImageGrabber.h"
#endif // WITH_PICASSO
#ifdef WITH_PICASSO_V4L
#  include "imalib/picassoImageGrabber.h"
#endif // WITH_PICASSO_V4L
#ifdef WITH_V4L
#   include "imalib/v4lImageGrabber.h"
#endif // WITH_V4L

#include <stdio.h>
#include <stdlib.h>     /* getenv(), etc. */
#include <termios.h>
#include <unistd.h>
#include <sys/time.h>

#include <string>
#include <iostream>

#include <boost/program_options.hpp>
namespace po = boost::program_options;

struct termios origTermios;

double firstTime;
double currentTime;
double lastTime;
unsigned long nbFrames;

int videoDevice;
int videoPort_r;
int videoPort_l;
unsigned short posX, posY, sourceWidth, sourceHeight;
unsigned short imageWidth;
unsigned short imageHeight;
bool statistics;
bool unWrap;
short innerRadius = -1;
short angularRes = 1; // Sign will be the direction.
short direction = 1;
double power = 1.0;
bool blur;
unsigned short blurSize;
bool upsideDown;
bool flip;

bool save;                      // If true, then save the images.
bool saveOne = false;           // If true, save one image.
unsigned short savedCounter_r = 0;
unsigned short savedCounter_l = 0;

enum possibleGrabberTypes
{
   #ifdef WITH_V4L
   v4l = 0, // For a v4l image grabber.
   #endif // WITH_V4L
   #ifdef WITH_PICASSO
   picasso, // For a picasso image grabber.
   #endif // WITH_PICASSO
   #ifdef WITH_PICASSO_V4L
   picassoV4L, // For a v4l picasso image grabber.
   #endif // WITH_PICASSO_V4L
   fake // For a fake frame grabber.
} grabberType;


imalib::imageRGB theImage_l;
imalib::imageGrabber* theGrabber_l;
imalib::imageRGB theImage_r;
imalib::imageGrabber* theGrabber_r;

/*
\brief Retrieves a timestamp in seconds using gettimeofday() for portable
timing
\return The current time since the  Epoch  (see time(2)).
*/
inline double time()
{
   timeval tv;
   gettimeofday(&tv, 0);

   return(static_cast<double>(tv.tv_sec)
         + static_cast<double>(tv.tv_usec) / 1000000.0);
}

void commandLine(int argc, char** argv)
{
   std::ostringstream helpKeys;
   helpKeys
   << "While running,'q' will quit, 's' will save one pair and 'S' will\n"
   << "continuously save.";

   po::options_description generalOptionsDesc("General options");
   generalOptionsDesc.add_options()
   ("help", "Prints this help message and exits")
   ("blurSize,b", po::value<unsigned short>(&blurSize)->default_value(1),
    "To blur the images (blurSize is the size of the square neighbourhood)");

   po::options_description cameraOptionsDesc("Camera options");
   cameraOptionsDesc.add_options()
#ifdef WITH_V4L
   ("v4l", "to grab using a v4l image grabber (default)")
#endif // WITH_V4L
#ifdef WITH_PICASSO
   ("picasso", "to grab using a picasso image grabber")
#endif // WITH_PICASSO
#ifdef WITH_PICASSO_V4L
   ("picassoV4L", "to grab using a v4l picasso image grabber")
#endif // WITH_PICASSO_V4L
   ("fake", "to grab using a fake image grabber"
#ifndef WITH_V4L
    " (default)"
#endif // WITH_V4L
   )
   ("statistics,s", "to output some statistics while running")
   ("videoDevice,d", po::value<int>(&videoDevice)->default_value(0),
    "Number of the video device")
   ("videoPortRight", po::value<int>(&videoPort_r)->default_value(0),
    "Number of the port on the frame grabber for the \"right\" image")
   ("videoPortLeft", po::value<int>(&videoPort_l)->default_value(0),
    "Number of the port on the frame grabber for the \"left\" image")
   ("posX,x", po::value<unsigned short>(&posX)->default_value(0),
    "X position of the source viewport")
   ("posY,y", po::value<unsigned short>(&posY)->default_value(0),
    "Y position of the source viewport")
   ("sourceWidth,w",
    po::value<unsigned short>(&sourceWidth)->
       default_value(imalib::IMAGE_DEFAULT_INPUTWIDTH),
    "Width of the viewport")
   ("sourceHeight,h",
    po::value<unsigned short>(&sourceHeight)->
       default_value(imalib::IMAGE_DEFAULT_INPUTHEIGHT),
    "Height of the viewport")
   ("imageWidth,W",
    po::value<unsigned short>(&imageWidth)->
       default_value(imalib::IMAGE_DEFAULT_INPUTWIDTH),
    "Width of the image")
   ("imageHeight,H",
    po::value<unsigned short>(&imageHeight)->
       default_value(imalib::IMAGE_DEFAULT_INPUTHEIGHT),
    "Height of the image")
    ("upsideDown,U", "To use if the camera is upsidedown")
    ("flip,f", "To horizontally flip the images")
    ;

//    po::options_description
//       unwrappingOptionsDesc("Unwrapping options (any will unwrap)");
//    unwrappingOptionsDesc.add_options()
//    ("innerRadius,r", po::value<short>(&innerRadius),
//     "The inner radius of the doughnut used for unwrapping")
//    ("power,u", po::value<double>(&power),
//     "To unwrap the images (omni-directional to panoramic):"
//     " this is the power of the polynomial function along a radial axis"
//     " used to scale differently the inside and the outside"
//     " (the higher the value, the more the outside is important)")
//    ("pixelsPerDegree,R",
//     po::value<short>(&angularRes),
//     "Number of pixels per degree of angle when unwrapping")
//    ("direction,D", po::value<short>(&direction),
//     "Direction of the unwrapping from left to right in the images:"
//     " 1 counter-clockwise, -1 clockwise");

   po::options_description  optionsDesc("");
   optionsDesc.add(generalOptionsDesc);
   optionsDesc.add(cameraOptionsDesc);
//    optionsDesc.add(unwrappingOptionsDesc);

   po::variables_map vm;
   try
   {
      po::store(po::parse_command_line(argc, argv, optionsDesc), vm);
      po::notify(vm);
   }
   catch (const std::exception& e)
   {
      std::cerr << "Error: " << e.what() << std::endl << std::endl
      << optionsDesc << std::endl;
      exit(0);
   }

   if (vm.count("help"))
   {
      std::cout << optionsDesc << std::endl
         << helpKeys.str() << std::endl;
      exit(1);
   }

   if (vm.count("statistics"))
   {
      statistics = true;
   }
   else
   {
      statistics = false;
   }

   innerRadius = 50;
   power = 1.0;
   angularRes = 1;
   direction = -1;
   if(vm.count("innerRadius")
      || vm.count("power")
      || vm.count("pixelsPerDegree")
      || vm.count("direction"))
   {
      unWrap = true;
      if (direction < 0)
         direction = -1;
      else
         direction = 1;
   }

#ifdef WITH_V4L
   grabberType = v4l;
#else // WITH_V4L
   grabberType = fake;
#endif // WITH_V4L
#ifdef WITH_V4L
   if (vm.count("v4l"))
   {
      grabberType = v4l;
   }
#endif // WITH_V4L
#ifdef WITH_PICASSO
   if (vm.count("picasso"))
   {
      grabberType = picasso;
   }
#endif // WITH_PICASSO
#ifdef WITH_PICASSO_V4L
   if (vm.count("picassoV4L"))
   {
      grabberType = picassoV4L;
   }
#endif // WITH_PICASSO_V4L
   if (vm.count("fake"))
   {
      grabberType = fake;
   }

   if (blurSize != 1)
   {
      blur = true;
   }
   else
   {
      blur = false;
   }

   if (vm.count("upsideDown"))
   {
      upsideDown = true;
   }
   else
   {
      upsideDown = false;
   }

   if (vm.count("flip"))
   {
      flip = true;
   }
   else
   {
      flip = false;
   }
}

void saveImage_r()
{
   static std::string baseName = "grabbedImage_";
   static std::string extension = "_right.ppm";
   char fileName[31];

   sprintf(fileName, "%s%04d%s", baseName.c_str(), savedCounter_r,
           extension.c_str());
   theImage_r.write(fileName);
   savedCounter_r++;
}

void saveImage_l()
{
   static std::string baseName = "grabbedImage_";
   static std::string extension = "_left.ppm";
   char fileName[31];

   sprintf(fileName, "%s%04d%s", baseName.c_str(), savedCounter_l,
           extension.c_str());
   theImage_l.write(fileName);
   savedCounter_l++;
}

inline void flipSave()
{
   save = !save;
}

inline void doSaveOne()
{
   saveOne = true;
}

void quit()
{
   tcsetattr(fileno(stdin), TCSANOW, &origTermios);

   // Some statistics.
   currentTime = time();
   std::cout << "\n" << nbFrames << " frames at "
      << static_cast<double>(nbFrames) / (currentTime - firstTime)
      << " f/s.\n";

   exit(0);
}

void processKeys()
{
   int key = getchar();
   if (key != EOF)
   {
      if (key == 'q')
         quit();
      else if (key == 's')
         doSaveOne();
      else if (key == 'S')
         flipSave();
   }
}

void initNonBlocking()
{
   struct termios newTermios;

   tcgetattr(fileno(stdin), &origTermios);

   tcgetattr(fileno(stdin), &newTermios);

   newTermios.c_cc[VTIME] = 0;
   newTermios.c_cc[VMIN] = 0;
   newTermios.c_lflag &= (~ECHO & ~ICANON);
   tcsetattr(fileno(stdin), TCSANOW, &newTermios);
}

int animate(void* impixmap)
{
   processKeys();

#ifdef WITH_PICASSO
   if (grabberType == picasso)
   {
      dynamic_cast<imalib::picassoImageGrabber*>(theGrabber_l)->
         setInputPort(videoPort_l);
   }
#endif // WITH_PICASSO
   theGrabber_l->getImage(&theImage_l);
   if (saveOne)
   {
      saveImage_l();
   }
   else if (save)
      saveImage_l();

#ifdef WITH_PICASSO
   if (grabberType == picasso)
   {
      dynamic_cast<imalib::picassoImageGrabber*>(theGrabber_r)->
         setInputPort(videoPort_r);
   }
#endif // WITH_PICASSO
   theGrabber_r->getImage(&theImage_r);
   if (saveOne)
   {
      saveImage_r();
   }
   else if (save)
      saveImage_r();

   nbFrames++;

   if (statistics)
   {
      lastTime = currentTime;
      currentTime = time();
      std::cout << "\rFrame: " << nbFrames << " ("
         << 1.0 / (currentTime - lastTime) << " f/s)";
      if (save || saveOne)
         std::cout << " (saved image " << savedCounter_r << ")";
      std::cout << "                              ";
   }
   saveOne = false;

   return(true);
}


int main(int argc, char** argv)
{
   commandLine(argc, argv);

   initNonBlocking();

   switch (grabberType)
   {
#ifdef WITH_V4L
      case v4l:
         theGrabber_l =
            new imalib::v4lImageGrabber(imageWidth, imageHeight,
                                        upsideDown,
                                        videoDevice, videoPort_l,
                                        posX, posY,
                                        sourceWidth, sourceHeight);
         theGrabber_r =
            new imalib::v4lImageGrabber(imageWidth, imageHeight,
                                        upsideDown,
                                        videoDevice, videoPort_r,
                                        posX, posY,
                                        sourceWidth, sourceHeight);
         break;
#endif // WITH_V4L
#ifdef WITH_PICASSO
      case picasso:
         std::cout
            << "Only creating one picassoImageGrabber.  This is normal!\n";
         theGrabber_l =
            new imalib::picassoImageGrabber(imageWidth, imageHeight,
                                            upsideDown,
                                            videoDevice, videoPort_l,
                                            posX, posY,
                                            sourceWidth, sourceHeight);
         theGrabber_r = theGrabber_l;
         break;
#endif // WITH_PICASSO
#ifdef WITH_PICASSO_V4L
      case picassoV4L:
         theGrabber_l =
            new imalib::v4lImageGrabber(imageWidth, imageHeight,
                                        videoDevice, videoPort_l,
                                        posX, posY,
                                        sourceWidth, sourceHeight);
         theGrabber_r =
            new imalib::v4lImageGrabber(imageWidth, imageHeight,
                                        videoDevice, videoPort_r,
                                        posX, posY,
                                        sourceWidth, sourceHeight);
         break;
#endif // WITH_PICASSO_V4L
      case fake:
      default:
         theGrabber_l = NULL;
         theGrabber_r = NULL;
   }

   if (!theGrabber_l || !theGrabber_r
      || !theGrabber_l->isConnected() || !theGrabber_r->isConnected())
   {
      if (!theGrabber_l || !theGrabber_r)
      {
         std::cerr << "\n--> Compiled without image grabber support:";
      }
      else
      {
         std::cerr << "\n--> Could not connect to a frame-grabber:";
         delete theGrabber_l;
         delete theGrabber_r;
      }

      std::cerr << " using a fake one instead.\n\n";
      theGrabber_l = new imalib::imageGrabber(imageWidth, imageHeight,
                                              upsideDown, flip);
      theGrabber_r = new imalib::imageGrabber(imageWidth, imageHeight,
                                              upsideDown, flip);
      grabberType = fake;
   }

//    if (unWrap)
//    {
//       theGrabber_l->doUnwrap(true, imageWidth);
//       theGrabber_r->doUnwrap(true, imageWidth);
//    }
   if (blur)
   {
      theGrabber_l->doBlur(true, blurSize, blurSize);
      theGrabber_r->doBlur(true, blurSize, blurSize);
   }
   if (flip)
   {
      theGrabber_l->setFlip(true);
      theGrabber_r->setFlip(true);
   }

   theGrabber_l->printInfoAboutSource();
   theGrabber_r->printInfoAboutSource();

   theGrabber_l->getImage(&theImage_l);
   theGrabber_r->getImage(&theImage_r);

   firstTime = time();
   currentTime = firstTime;
   lastTime = firstTime;
   nbFrames = 0;

   while(true)
   {
      animate(NULL);
   }

   return(0);
}
