/*
 * imalib
 * 
 * Copyright 2017 Aberystwyth University
 * 
 * This library and program are free software: you can redistribute them
 * and/or modify them under the terms of the GNU General Public License as 
 * published by the Free Software Foundation, either version 3 of the License,
 * or (at your option) any later version.
 * 
 * The program and library are distributed in the hope that they will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*!
 * \file grabSendImages.cpp
 * \brief Grabs from a framegrabber and sends images over a socket.
 * \author Fred Labrosse <ffl@aber.ac.uk>
 * \copyright 2017 Aberystwyth University
 */

#include "imalib/imageTransmitter.h"
#include "imalib/imageGrabber.h"
#ifdef WITH_PICASSO
#  include "imalib/picassoImageGrabber.h"
#endif // WITH_PICASSO
#ifdef WITH_PICASSO_V4L
#  include "imalib/picassoV4LImageGrabber.h"
#endif // WITH_PICASSO_V4L
#ifdef WITH_V4L
#  include "imalib/v4lImageGrabber.h"
#endif // WITH_V4L
#ifdef WITH_V4L2
#   include "imalib/v4l2ImageGrabber.h"
#endif // WITH_V4L2
#ifdef WITH_DC1394
#  include "imalib/dc1394ImageGrabber.h"
#endif // WITH_DC1394
#ifdef WITH_UEYE
#  include "imalib/ueyeImageGrabber.h"
#endif // WITH_UEYE

#include <stdio.h>
#include <termios.h>
#include <unistd.h>

#include <string>
#include <iostream>

#include <boost/program_options.hpp>
namespace po = boost::program_options;
#include <boost/asio.hpp>
using boost::asio::ip::tcp;

imalib::imageTransmitter* imgTransmit;
int port;
int videoDevice;
int videoPort;
int cameraNumber;
int videoInput;
int videoStandard;
#ifdef WITH_V4L2
unsigned short ioMet =
static_cast<unsigned short>(imalib::v4l2ImageGrabber::IO_METHOD_MMAP);
#endif // WITH_V4L2
unsigned short posX, posY, sourceWidth, sourceHeight;
unsigned short imageWidth;
unsigned short imageHeight;
bool unWrap;
short innerRadius = -1;
short angularRes = 1;
short direction = -1;
double power = 1.0;
bool blur;
unsigned short blurSize;
bool upsideDown;
bool flip;

struct termios origTermios;

enum possibleGrabberTypes
{
#ifdef WITH_V4L
   v4l = 0, // For a v4l image grabber.
#endif // WITH_V4L
#ifdef WITH_V4L2
   v4l2, // For a v4l2 image grabber.
#endif // WITH_V4L2
#ifdef WITH_PICASSO
      picasso, // For a picasso image grabber.
#endif // WITH_PICASSO
#ifdef WITH_PICASSO_V4L
      picassoV4L, // For a v4l picasso image grabber.
#endif // WITH_PICASSO_V4L
#ifdef WITH_DC1394
      dc1394, // For a dc1394 (firewire) camera.
#endif // WITH_DC1394
#ifdef WITH_UEYE
      ueye, // For a uEye USB camera.
#endif // WITH_UEYE
      fake // For a fake frame grabber.
} grabberType;

imalib::imageRGB theImage;
imalib::imageGrabber* theGrabber;

boost::asio::io_service ioService;
boost::asio::deadline_timer
keyPressesDelay(ioService,
                boost::posix_time::milliseconds(500));

void commandLine(int argc, char** argv)
{
   std::ostringstream helpKeys;
   helpKeys
      << "While running, up, down, right, left will move the viewport,\n"
      << "'>', '<', '^' and '_' will respectively increase/decrease\n"
      << "the width and increase/decrease the height of the viewport.\n"
      << "'q' will quit.";

   po::options_description generalOptionsDesc("General options");
   generalOptionsDesc.add_options()
      ("help", "Prints this help message and exits")
      ("port,p", po::value<int>(&port)->default_value(7777),
       "Port the server will be listening to")
      ("blurSize,b", po::value<unsigned short>(&blurSize)->default_value(1),
       "To blur the images (blurSize is the size of the square neighbourhood)");

   po::options_description cameraOptionsDesc("Camera options");
   cameraOptionsDesc.add_options()
#ifdef WITH_V4L
      ("v4l", "to grab using a v4l image grabber (default)")
#endif // WITH_V4L
#ifdef WITH_V4L2
   ("v4l2", "to grab using a v4l2 image grabber (default)")
#endif // WITH_V4L2
#ifdef WITH_PICASSO
      ("picasso", "to grab using a picasso image grabber")
#endif // WITH_PICASSO
#ifdef WITH_PICASSO_V4L
      ("picassoV4L", "to grab using a v4l picasso image grabber")
#endif // WITH_PICASSO_V4L
#ifdef WITH_DC1394
      ("dc1394", "to grab using a dc1394 (firewire) camera")
#endif // WITH_DC1394
#ifdef WITH_UEYE
      ("ueye", "to grab using a uEye USB camera")
#endif // WITH_UEYE
      ("fake", "to grab using a fake image grabber"
#ifndef WITH_V4L2
       " (default)"
#endif // WITH_V4L2
      )
      ("videoDevice,d", po::value<int>(&videoDevice)->default_value(0),
       "Number of the video device [v4l, v4l2 picassoV4l, picasso]")
      ("videoPort,v", po::value<int>(&videoPort)->default_value(0),
       "Number of the port on the frame grabber [v4l, picassoV4l, picasso]")
      ("cameraNumber,c", po::value<int>(&cameraNumber)->default_value(0),
       "Number of the camera [dc1394, ueye]")
#ifdef WITH_V4L2
      ("input,i", po::value<int>(&videoInput)->default_value(0),
       "Number of the input to use [v4l2]")
      ("standard,S", po::value<int>(&videoStandard)->default_value(0),
       "Number of the video standard to use [v4l2]")
      ("ioMethod", po::value<unsigned short>(&ioMet)->default_value(ioMet),
       "v4l2 io method:\n0 for READ\n1 for MMAP\n2 for USRPTR")
#endif // WITH_V4L2
      ("posX,x", po::value<unsigned short>(&posX)->default_value(0),
       "X position of the source viewport")
      ("posY,y", po::value<unsigned short>(&posY)->default_value(0),
       "Y position of the source viewport")
      ("sourceWidth,w",
       po::value<unsigned short>(&sourceWidth)->
       default_value(imalib::IMAGE_DEFAULT_INPUTWIDTH),
          "Width of the viewport [v4l, v4l2, picasso, picassoV4L]")
      ("sourceHeight,h",
       po::value<unsigned short>(&sourceHeight)->
       default_value(imalib::IMAGE_DEFAULT_INPUTHEIGHT),
          "Height of the viewport [v4l, v4l2, picasso, picassoV4L]")
      ("imageWidth,W",
       po::value<unsigned short>(&imageWidth)->
       default_value(imalib::IMAGE_DEFAULT_INPUTWIDTH),
       "Width of the image")
      ("imageHeight,H",
       po::value<unsigned short>(&imageHeight)->
       default_value(imalib::IMAGE_DEFAULT_INPUTHEIGHT),
       "Height of the image")
      ("upsideDown,U", "To use if the camera is upsidedown")
      ("flip,f", "To horizontally flip the images")
      ;

   po::options_description
      unwrappingOptionsDesc("Unwrapping options (any will unwrap)");
   unwrappingOptionsDesc.add_options()
      ("innerRadius,r", po::value<short>(&innerRadius),
       "The inner radius of the doughnut used for unwrapping")
      ("power,u", po::value<double>(&power),
       "To unwrap the images (omni-directional to panoramic):"
       " this is the power of the polynomial function along a radial axis"
       " used to scale differently the inside and the outside"
       " (the higher the value, the more the outside is important)")
      ("pixelsPerDegree,R",
       po::value<short>(&angularRes),
       "Number of pixels per degree of angle when unwrapping")
      ("direction,D", po::value<short>(&direction),
       "Direction of the unwrapping from left to right in the images:"
       " 1 counter-clockwise, -1 clockwise");

   po::options_description  optionsDesc("");
   optionsDesc.add(generalOptionsDesc);
   optionsDesc.add(cameraOptionsDesc);
   optionsDesc.add(unwrappingOptionsDesc);

   po::variables_map vm;
   try
   {
      po::store(po::parse_command_line(argc, argv, optionsDesc), vm);
      po::notify(vm);
   }
   catch (const std::exception& e)
   {
      std::cerr << "Error: " << e.what() << std::endl << std::endl
         << optionsDesc << std::endl;
      exit(0);
   }

   if (vm.count("help"))
   {
      std::cout
         << "Some of the options are only valid for some of the grabbers.  This is\n"
         << "indicated when necessary.\n\n"
         << optionsDesc << std::endl
         << helpKeys.str() << std::endl;
      exit(1);
   }

//   innerRadius = 50;
//   power = 1.0;
//   angularRes = 1;
//   direction = -1;
   if(vm.count("innerRadius")
      || vm.count("power")
      || vm.count("pixelsPerDegree")
      || vm.count("direction"))
   {
      unWrap = true;
      if (direction < 0)
         direction = -1;
      else
         direction = 1;
   }

#ifdef WITH_V4L2
   grabberType = v4l2;
#else // WITH_V4L2
   grabberType = fake;
#endif // WITH_V4L
  // Set the requested grabber type, if any.
#ifdef WITH_V4L
   if (vm.count("v4l"))
   {
      grabberType = v4l;
   }
#endif // WITH_V4L
#ifdef WITH_V4L2
if (vm.count("v4l2"))
{
   grabberType = v4l2;
}
#endif // WITH_V4L2
#ifdef WITH_PICASSO
   if (vm.count("picasso"))
   {
      grabberType = picasso;
   }
#endif // WITH_PICASSO
#ifdef WITH_PICASSO_V4L
   if (vm.count("picassoV4L"))
   {
      grabberType = picassoV4L;
   }
#endif // WITH_PICASSO_V4L
#ifdef WITH_DC1394
   if (vm.count("dc1394"))
   {
      grabberType = dc1394;
   }
#endif // WITH_DC1394
#ifdef WITH_UEYE
   if (vm.count("ueye"))
   {
      grabberType = ueye;
   }
#endif // WITH_UEYE
   if (vm.count("fake"))
   {
      grabberType = fake;
   }

   if (blurSize != 1)
   {
      blur = true;
   }
   else
   {
      blur = false;
   }

   if (vm.count("upsideDown"))
   {
      upsideDown = true;
   }
   else
   {
      upsideDown = false;
   }

   if (vm.count("flip"))
   {
      flip = true;
   }
   else
   {
      flip = false;
   }
}

void quit()
{
   tcsetattr(fileno(stdin), TCSANOW, &origTermios);

   delete theGrabber;
}

void processKeys()
{
   bool cont = true;

   int key = getchar();
   if (key != EOF)
   {
      if (key == 'q')
      {
         cont = false;
      }
      else if (key == 68)
      {
#ifdef WITH_PICASSO
         imalib::picassoImageGrabber* tmpPIG =
            dynamic_cast<imalib::picassoImageGrabber*>(theGrabber);
         if (tmpPIG)
            tmpPIG->moveLeft();
#endif // WITH_PICASSO
#ifdef WITH_PICASSO_V4L
         imalib::picassoV4LImageGrabber* tmpPV4LIG =
            dynamic_cast<imalib::picassoV4LImageGrabber*>(theGrabber);
         if (tmpPV4LIG)
            tmpPV4LIG->moveLeft();
#endif // WITH_PICASSO_V4L
#ifdef WITH_V4L
         imalib::v4lImageGrabber* tmpV4LIG =
            dynamic_cast<imalib::v4lImageGrabber*>(theGrabber);
         if (tmpV4LIG)
            tmpV4LIG->moveLeft();
#endif // WITH_V4L
#ifdef WITH_V4L2
         imalib::v4l2ImageGrabber* tmpV4L2IG =
            dynamic_cast<imalib::v4l2ImageGrabber*>(theGrabber);
         if (tmpV4L2IG)
            tmpV4L2IG->moveLeft();
#endif // WITH_V4L2
#ifdef WITH_DC1394
         imalib::dc1394ImageGrabber* tmpDCIG =
            dynamic_cast<imalib::dc1394ImageGrabber*>(theGrabber);
         if (tmpDCIG)
            tmpDCIG->moveLeft();
#endif // WITH_DC1394
#ifdef WITH_UEYE
         imalib::ueyeImageGrabber* tmpUEYEIG =
            dynamic_cast<imalib::ueyeImageGrabber*>(theGrabber);
         if (tmpUEYEIG)
            tmpUEYEIG->moveLeft();
#endif // WITH_UEYE
      }
      else if (key == 67)
      {
#ifdef WITH_PICASSO
         imalib::picassoImageGrabber* tmpPIG =
            dynamic_cast<imalib::picassoImageGrabber*>(theGrabber);
         if (tmpPIG)
            tmpPIG->moveRight();
#endif // WITH_PICASSO
#ifdef WITH_PICASSO_V4L
         imalib::picassoV4LImageGrabber* tmpPV4LIG =
            dynamic_cast<imalib::picassoV4LImageGrabber*>(theGrabber);
         if (tmpPV4LIG)
            tmpPV4LIG->moveRight();
#endif // WITH_PICASSO_V4L
#ifdef WITH_V4L
         imalib::v4lImageGrabber* tmpV4LIG =
            dynamic_cast<imalib::v4lImageGrabber*>(theGrabber);
         if (tmpV4LIG)
            tmpV4LIG->moveRight();
#endif // WITH_V4L
#ifdef WITH_V4L2
         imalib::v4l2ImageGrabber* tmpV4L2IG =
            dynamic_cast<imalib::v4l2ImageGrabber*>(theGrabber);
         if (tmpV4L2IG)
            tmpV4L2IG->moveRight();
#endif // WITH_V4L2
#ifdef WITH_DC1394
         imalib::dc1394ImageGrabber* tmpDCIG =
            dynamic_cast<imalib::dc1394ImageGrabber*>(theGrabber);
         if (tmpDCIG)
            tmpDCIG->moveRight();
#endif // WITH_DC1394
#ifdef WITH_UEYE
         imalib::ueyeImageGrabber* tmpUEYEIG =
            dynamic_cast<imalib::ueyeImageGrabber*>(theGrabber);
         if (tmpUEYEIG)
            tmpUEYEIG->moveRight();
#endif // WITH_UEYE
      }
      else if (key == 65)
      {
#ifdef WITH_PICASSO
         imalib::picassoImageGrabber* tmpPIG =
            dynamic_cast<imalib::picassoImageGrabber*>(theGrabber);
         if (tmpPIG)
            tmpPIG->moveUp();
#endif // WITH_PICASSO
#ifdef WITH_PICASSO_V4L
         imalib::picassoV4LImageGrabber* tmpPV4LIG =
            dynamic_cast<imalib::picassoV4LImageGrabber*>(theGrabber);
         if (tmpPV4LIG)
            tmpPV4LIG->moveUp();
#endif // WITH_PICASSO_V4L
#ifdef WITH_V4L
         imalib::v4lImageGrabber* tmpV4LIG =
            dynamic_cast<imalib::v4lImageGrabber*>(theGrabber);
         if (tmpV4LIG)
            tmpV4LIG->moveUp();
#endif // WITH_V4L
#ifdef WITH_V4L2
         imalib::v4l2ImageGrabber* tmpV4L2IG =
            dynamic_cast<imalib::v4l2ImageGrabber*>(theGrabber);
         if (tmpV4L2IG)
            tmpV4L2IG->moveUp();
#endif // WITH_V4L2
#ifdef WITH_DC1394
         imalib::dc1394ImageGrabber* tmpDCIG =
            dynamic_cast<imalib::dc1394ImageGrabber*>(theGrabber);
         if (tmpDCIG)
            tmpDCIG->moveUp();
#endif // WITH_DC1394
#ifdef WITH_UEYE
         imalib::ueyeImageGrabber* tmpUEYEIG =
            dynamic_cast<imalib::ueyeImageGrabber*>(theGrabber);
         if (tmpUEYEIG)
            tmpUEYEIG->moveUp();
#endif // WITH_UEYE
      }
      else if (key == 66)
      {
#ifdef WITH_PICASSO
         imalib::picassoImageGrabber* tmpPIG =
            dynamic_cast<imalib::picassoImageGrabber*>(theGrabber);
         if (tmpPIG)
            tmpPIG->moveDown();
#endif // WITH_PICASSO
#ifdef WITH_PICASSO_V4L
         imalib::picassoV4LImageGrabber* tmpPV4LIG =
            dynamic_cast<imalib::picassoV4LImageGrabber*>(theGrabber);
         if (tmpPV4LIG)
            tmpPV4LIG->moveDown();
#endif // WITH_PICASSO_V4L
#ifdef WITH_V4L
         imalib::v4lImageGrabber* tmpV4LIG =
            dynamic_cast<imalib::v4lImageGrabber*>(theGrabber);
         if (tmpV4LIG)
            tmpV4LIG->moveDown();
#endif // WITH_V4L
#ifdef WITH_V4L2
         imalib::v4l2ImageGrabber* tmpV4L2IG =
            dynamic_cast<imalib::v4l2ImageGrabber*>(theGrabber);
         if (tmpV4L2IG)
            tmpV4L2IG->moveDown();
#endif // WITH_V4L2
#ifdef WITH_DC1394
         imalib::dc1394ImageGrabber* tmpDCIG =
            dynamic_cast<imalib::dc1394ImageGrabber*>(theGrabber);
         if (tmpDCIG)
            tmpDCIG->moveDown();
#endif // WITH_DC1394
#ifdef WITH_UEYE
         imalib::ueyeImageGrabber* tmpUEYEIG =
            dynamic_cast<imalib::ueyeImageGrabber*>(theGrabber);
         if (tmpUEYEIG)
            tmpUEYEIG->moveDown();
#endif // WITH_UEYE
      }
      else if (key == '>')
      {
#ifdef WITH_PICASSO
         imalib::picassoImageGrabber* tmpPIG =
            dynamic_cast<imalib::picassoImageGrabber*>(theGrabber);
         if (tmpPIG)
            tmpPIG->increaseWidth();
#endif // WITH_PICASSO
#ifdef WITH_PICASSO_V4L
         imalib::picassoV4LImageGrabber* tmpPV4LIG =
            dynamic_cast<imalib::picassoV4LImageGrabber*>(theGrabber);
         if (tmpPV4LIG)
            tmpPV4LIG->increaseWidth();
#endif // WITH_PICASSO_V4L
#ifdef WITH_V4L
         imalib::v4lImageGrabber* tmpV4LIG =
            dynamic_cast<imalib::v4lImageGrabber*>(theGrabber);
         if (tmpV4LIG)
            tmpV4LIG->increaseWidth();
#endif // WITH_V4L
#ifdef WITH_V4L2
         imalib::v4l2ImageGrabber* tmpV4L2IG =
            dynamic_cast<imalib::v4l2ImageGrabber*>(theGrabber);
         if (tmpV4L2IG)
            tmpV4L2IG->increaseWidth();
#endif // WITH_V4L2
#ifdef WITH_DC1394
         imalib::dc1394ImageGrabber* tmpDCIG =
            dynamic_cast<imalib::dc1394ImageGrabber*>(theGrabber);
         if (tmpDCIG)
            tmpDCIG->increaseWidth();
#endif // WITH_DC1394
#ifdef WITH_UEYE
         imalib::ueyeImageGrabber* tmpUEYEIG =
            dynamic_cast<imalib::ueyeImageGrabber*>(theGrabber);
         if (tmpUEYEIG)
            tmpUEYEIG->increaseWidth();
#endif // WITH_UEYE
      }
      else if (key == '<')
      {
#ifdef WITH_PICASSO
         imalib::picassoImageGrabber* tmpPIG =
            dynamic_cast<imalib::picassoImageGrabber*>(theGrabber);
         if (tmpPIG)
            tmpPIG->decreaseWidth();
#endif // WITH_PICASSO
#ifdef WITH_PICASSO_V4L
         imalib::picassoV4LImageGrabber* tmpPV4LIG =
            dynamic_cast<imalib::picassoV4LImageGrabber*>(theGrabber);
         if (tmpPV4LIG)
            tmpPV4LIG->decreaseWidth();
#endif // WITH_PICASSO_V4L
#ifdef WITH_V4L
         imalib::v4lImageGrabber* tmpV4LIG =
            dynamic_cast<imalib::v4lImageGrabber*>(theGrabber);
         if (tmpV4LIG)
            tmpV4LIG->decreaseWidth();
#endif // WITH_V4L
#ifdef WITH_V4L2
         imalib::v4l2ImageGrabber* tmpV4L2IG =
            dynamic_cast<imalib::v4l2ImageGrabber*>(theGrabber);
         if (tmpV4L2IG)
            tmpV4L2IG->decreaseWidth();
#endif // WITH_V4L2
#ifdef WITH_DC1394
         imalib::dc1394ImageGrabber* tmpDCIG =
            dynamic_cast<imalib::dc1394ImageGrabber*>(theGrabber);
         if (tmpDCIG)
            tmpDCIG->decreaseWidth();
#endif // WITH_DC1394
#ifdef WITH_UEYE
         imalib::ueyeImageGrabber* tmpUEYEIG =
            dynamic_cast<imalib::ueyeImageGrabber*>(theGrabber);
         if (tmpUEYEIG)
            tmpUEYEIG->decreaseWidth();
#endif // WITH_UEYE
      }
      else if (key == '^')
      {
#ifdef WITH_PICASSO
         imalib::picassoImageGrabber* tmpPIG =
            dynamic_cast<imalib::picassoImageGrabber*>(theGrabber);
         if (tmpPIG)
            tmpPIG->increaseHeight();
#endif // WITH_PICASSO
#ifdef WITH_PICASSO_V4L
         imalib::picassoV4LImageGrabber* tmpPV4LIG =
            dynamic_cast<imalib::picassoV4LImageGrabber*>(theGrabber);
         if (tmpPV4LIG)
            tmpPV4LIG->increaseHeight();
#endif // WITH_PICASSO_V4L
#ifdef WITH_V4L
         imalib::v4lImageGrabber* tmpV4LIG =
            dynamic_cast<imalib::v4lImageGrabber*>(theGrabber);
         if (tmpV4LIG)
            tmpV4LIG->increaseHeight();
#endif // WITH_V4L
#ifdef WITH_V4L2
         imalib::v4l2ImageGrabber* tmpV4L2IG =
            dynamic_cast<imalib::v4l2ImageGrabber*>(theGrabber);
         if (tmpV4L2IG)
            tmpV4L2IG->increaseHeight();
#endif // WITH_V4L2
#ifdef WITH_DC1394
         imalib::dc1394ImageGrabber* tmpDCIG =
            dynamic_cast<imalib::dc1394ImageGrabber*>(theGrabber);
         if (tmpDCIG)
            tmpDCIG->increaseHeight();
#endif // WITH_DC1394
#ifdef WITH_UEYE
         imalib::ueyeImageGrabber* tmpUEYEIG =
            dynamic_cast<imalib::ueyeImageGrabber*>(theGrabber);
         if (tmpUEYEIG)
            tmpUEYEIG->increaseHeight();
#endif // WITH_UEYE
      }
      else if (key == '_')
      {
#ifdef WITH_PICASSO
         imalib::picassoImageGrabber* tmpPIG =
            dynamic_cast<imalib::picassoImageGrabber*>(theGrabber);
         if (tmpPIG)
            tmpPIG->decreaseHeight();
#endif // WITH_PICASSO
#ifdef WITH_PICASSO_V4L
         imalib::picassoV4LImageGrabber* tmpPV4LIG =
            dynamic_cast<imalib::picassoV4LImageGrabber*>(theGrabber);
         if (tmpPV4LIG)
            tmpPV4LIG->decreaseHeight();
#endif // WITH_PICASSO_V4L
#ifdef WITH_V4L
         imalib::v4lImageGrabber* tmpV4LIG =
            dynamic_cast<imalib::v4lImageGrabber*>(theGrabber);
         if (tmpV4LIG)
            tmpV4LIG->decreaseHeight();
#endif // WITH_V4L
#ifdef WITH_V4L2
         imalib::v4l2ImageGrabber* tmpV4L2IG =
            dynamic_cast<imalib::v4l2ImageGrabber*>(theGrabber);
         if (tmpV4L2IG)
            tmpV4L2IG->decreaseHeight();
#endif // WITH_V4L2
#ifdef WITH_DC1394
         imalib::dc1394ImageGrabber* tmpDCIG =
            dynamic_cast<imalib::dc1394ImageGrabber*>(theGrabber);
         if (tmpDCIG)
            tmpDCIG->decreaseHeight();
#endif // WITH_DC1394
#ifdef WITH_UEYE
         imalib::ueyeImageGrabber* tmpUEYEIG =
            dynamic_cast<imalib::ueyeImageGrabber*>(theGrabber);
         if (tmpUEYEIG)
            tmpUEYEIG->decreaseHeight();
#endif // WITH_UEYE
      }
//       else
//          std::cout << static_cast<int>(key) << std::endl;
   }

   // We make sure it will run again.
   if (cont)
   {
      keyPressesDelay.async_wait(boost::bind(&processKeys));
   }
   else
   {
      ioService.stop();
   }
}

void initNonBlocking()
{
  struct termios newTermios;

  tcgetattr(fileno(stdin), &origTermios);

  tcgetattr(fileno(stdin), &newTermios);

  newTermios.c_cc[VTIME] = 0;
  newTermios.c_cc[VMIN] = 0;
  newTermios.c_lflag &= (~ECHO & ~ICANON);
  tcsetattr(fileno(stdin), TCSANOW, &newTermios);
}

int main(int argc, char** argv)
{
   commandLine(argc, argv);

   initNonBlocking();

   switch (grabberType)
   {
#ifdef WITH_V4L
   case v4l:
      theGrabber =
         new imalib::v4lImageGrabber(imageWidth, imageHeight,
                                     upsideDown, flip,
                                     videoDevice, videoPort,
                                     posX, posY,
                                     sourceWidth, sourceHeight);
      break;
#endif // WITH_V4L
#ifdef WITH_V4L2
      case v4l2:
         theGrabber =
            new imalib::v4l2ImageGrabber(imageWidth, imageHeight,
                                         upsideDown,
                                         flip,
                                         videoDevice,
                                         videoInput,
                                         videoStandard,
                                         static_cast<imalib::v4l2ImageGrabber::possibleIoMethods>(ioMet),
                                         posX, posY,
                                         sourceWidth, sourceHeight);
                                         break;
#endif // WITH_V4L2
#ifdef WITH_PICASSO
   case picasso:
      theGrabber =
         new imalib::picassoImageGrabber(imageWidth, imageHeight,
                                         upsideDown, flip,
                                         videoDevice, videoPort,
                                         posX, posY,
                                         sourceWidth, sourceHeight);
      break;
#endif // WITH_PICASSO
#ifdef WITH_PICASSO_V4L
   case picassoV4L:
      theGrabber =
         new imalib::picassoV4LImageGrabber(imageWidth, imageHeight,
                                            upsideDown, flip,
                                            videoDevice, videoPort,
                                            posX, posY,
                                            sourceWidth, sourceHeight);
      break;
#endif // WITH_PICASSO_V4L
#ifdef WITH_DC1394
   case dc1394:
      theGrabber =
         new imalib::dc1394ImageGrabber(upsideDown, flip,
                                        cameraNumber,
                                        posX, posY,
                                        imageWidth, imageHeight);
      if (theGrabber->isConnected())
         static_cast<imalib::dc1394ImageGrabber*>(theGrabber)->setExposure(-1);
      break;
#endif // WITH_DC1394
#ifdef WITH_UEYE
   case ueye:
      theGrabber =
         new imalib::ueyeImageGrabber(upsideDown, flip,
                                      cameraNumber,
                                      posX, posY,
                                      imageWidth, imageHeight);
      break;
#endif // WITH_UEYE
   case fake:
   default:
      theGrabber = NULL;
   }

   if (!theGrabber || !theGrabber->isConnected())
   {
      if (theGrabber)
      {
         std::cerr << "\n--> Could not connect to the requested grabber.\n"
            << "Using a fake one instead.\n\n";
         delete theGrabber;
      }
      theGrabber = new imalib::imageGrabber(imageWidth, imageHeight,
                                            upsideDown, flip);
   }

   if (unWrap)
      theGrabber->doUnwrap(true, false, innerRadius, angularRes, power,
                           direction);
   if (blur)
      theGrabber->doBlur(true, blurSize, blurSize);
   //   if (flip)
   //      theGrabber->setFlip(true);

   theGrabber->printInfoAboutSource();

   imalib::imageTransmitter* imgTransmit = NULL;

   try
   {
     imgTransmit = new imalib::imageTransmitter(ioService, port, theGrabber);
   }
   catch (const std::exception& e)
   {
      std::cerr << "Exception: " << e.what() << "\n";
      delete imgTransmit;
      quit();
   }

   ioService.post(boost::bind(&processKeys));

   ioService.run();

   delete imgTransmit;
   quit();

   return(0);                   // Even though this is never reached.
}
