/*
 * imalib
 * 
 * Copyright 2017 Aberystwyth University
 * 
 * This library and program are free software: you can redistribute them
 * and/or modify them under the terms of the GNU General Public License as 
 * published by the Free Software Foundation, either version 3 of the License,
 * or (at your option) any later version.
 * 
 * The program and library are distributed in the hope that they will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*!
 * \file imaDistances.cpp
 * \brief A program to calculate the distances between an image and an
 * appearance over the whole image.
 *
 * \author Fred Labrosse <ffl@aber.ac.uk>
 * \copyright 2017 Aberystwyth University
 */

#include "imalib/appearance.h"
#include "imalib/distanceImgAppear.h"
#include "imalib/distanceMetric.h"
#include "imalib/imageLab.h"
#include "imalib/imageRGB.h"
#include "imalib/transformation.h"

#include <iostream>
#include <sstream>
#include <string>
#include <vector>

#include <boost/program_options.hpp>
namespace po = boost::program_options;

std::string imageName;
std::string appearanceName;
std::string outputName;
std::ofstream outputFile;
std::streambuf* outputBuffer;

bool useRobust = false;
bool doRow = false;
unsigned short whichRow;
bool doColumn = false;
unsigned short whichColumn;

void commandLine(int argc, char** argv)
{
   std::ostringstream usage;
   usage
   << "This program computes the distances between an image and an\n"
   << "appearance positioned everywhere on the image.\n"
   << std::endl
   << "The calculated position is that of the centre of the appearance.\n"
   << "Distances are calculated in the _ab colour space.\n"
   << std::endl
   << "Usage:\n" << "\timadistances [options] {image} {appearance} [output]"
   << std::endl << std::endl
   << "Where:\n"
   << "  image        Image file name\n"
   << "  appearance   Appearance file name\n"
   << "  output       Output file name (default standard output)";

   po::options_description distancesOptions("Options");

   imalib::imalibOptions imalibOptions;
   distancesOptions.add(*(imalibOptions.getOptions()));

    distancesOptions.add_options()
   ("help,h", "Prints this help message and exits")
   ("robust", "Use a robust metric")
   ("row", po::value<unsigned short>(&whichRow),
    "Only compute on that row")
   ("column", po::value<unsigned short>(&whichColumn),
    "Only compute on that column")
   ;

   po::options_description fileNamesOptions("Hidden options");
   fileNamesOptions.add_options()
   ("image", po::value<std::string>(&imageName), "Name of the image to use")
   ("appearance", po::value<std::string>(&appearanceName),
    "Name of the image to use as an appearance")
   ("output", po::value<std::string>(&outputName),
    "Name of the output file.  If not given the standard output is used.")
   ;

   po::options_description cmdlineOptions;
   cmdlineOptions.add(distancesOptions).add(fileNamesOptions);

   po::positional_options_description p;
   p.add("image", 1);
   p.add("appearance", 1);
   p.add("output", 1);

   po::variables_map vm;
   try
   {
      po::store(po::command_line_parser(argc, argv).
                options(cmdlineOptions).positional(p).run(), vm);
      notify(vm);
   }
   catch (const std::exception& e)
   {
      std::cerr << "Error: " << e.what() << std::endl << std::endl
         << usage.str() << std::endl << std::endl
         << distancesOptions << std::endl;
      exit(0);
   }

   imalibOptions.commandLine(vm);

   if (vm.count("help"))
   {
      std::cout << usage.str() << std::endl << std::endl
         << distancesOptions << std::endl;
      exit(1);
   }

   if (vm.count("robust"))
   {
      useRobust = true;
   }

   if (vm.count("row"))
   {
      doRow = true;
   }

   if (vm.count("column"))
   {
      doColumn = true;
   }

   if ((imageName == "") || (appearanceName == ""))
   {
      std::cerr << "Both an image and an appearance need to be provided.\n\n"
      << usage.str() << std::endl << std::endl
      << distancesOptions << std::endl;
      exit(0);
   }
}

int main(int argc, char** argv)
{
   commandLine(argc, argv);

   if (outputName != "")
   {
      outputFile.open(outputName);
      if (!outputFile.is_open())
      {
         std::cerr << "Could not open file " << outputName << " for output.\n"
            << "Using the standard output instead\n";
         outputName = "";
      }
      outputBuffer = outputFile.rdbuf();
   }
   if (outputName == "")
   {
      outputBuffer = std::cout.rdbuf();
   }
   std::ostream outputStream(outputBuffer);

   imalib::imageRGB imgAppearRGB;
   if (!imgAppearRGB.read(appearanceName))
   {
      exit(1);
   }

   imalib::image_ab imgAppear_ab;
   imgAppear_ab.copyFromRGB(imgAppearRGB);
   imalib::appearance<double, 2> anAppear_ab(imgAppear_ab);
   anAppear_ab.recentre();

   imalib::imageRGB imgMainRGB;
   if (!imgMainRGB.read(imageName))
   {
      exit(1);
   }
   imalib::image_ab imgMain_ab;
   imgMain_ab.copyFromRGB(imgMainRGB);

//    imalib::euclideanMetric<double, 2> theMetric;
   imalib::distanceImgAppearRectangular<double, 2,
                                      imalib::euclideanMetric<double, 2> >
                                      theDistanceFunction;
//                                       theDistanceFunction(theMetric);
//    imalib::euclideanRobust<double, 2> theRobustMetric;
   imalib::distanceImgAppearRectangular<double, 2,
                                      imalib::euclideanRobust<double, 2> >
                                      theRobustDistanceFunction;
//                                       theRobustDistanceFunction(theRobustMetric);

   imalib::translation transf;
   double transParam[2];
   double transParamMin[2];
   transParamMin[0] = 0.0;
   transParamMin[1] = 0.0;
   double transParamMax[2];
   transParamMax[0] = imgMainRGB.nbColumns();
   transParamMax[1] = imgMainRGB.nbRows();
   transf.setParamsLimits(transParamMin, transParamMax);
   double distMin = 1000000;
   double xMin = 0;
   double yMin = 0;
   double dist;
   imalib::translation transfMin;
   for (double y = (doRow ? whichRow : 0);
        y < (doRow ? whichRow + 1 : imgMainRGB.nbRows());
        y++)
   {
      for (double x = (doColumn ? whichColumn : 0);
           x < (doColumn ? whichColumn + 1 : imgMainRGB.nbColumns());
           x++)
      {
         if ((x < -anAppear_ab.getBBxMin())
            || (x >= imgMainRGB.nbColumns() - anAppear_ab.getBBxMax() - 1)
            || (y < -anAppear_ab.getBByMin())
            || (y >= imgMainRGB.nbRows() - anAppear_ab.getBByMax() - 1))
         {
            if (useRobust)
               dist = 50.0;
            else
               dist = 50.0;
         }
         else
         {
            transParam[0] = x;
            transParam[1] = y;
            transf.setParams(transParam);
            if (useRobust)
            {
               dist = theRobustDistanceFunction.distance(imgMain_ab,
                                                         anAppear_ab,
                                                         transf);
            }
            else
            {
               dist = theDistanceFunction.distanceNormalised(imgMain_ab,
                                                             anAppear_ab,
                                                             transf);
            }
            if (dist < distMin)
            {
               distMin = dist;
               xMin = x;
               yMin = y;
               std::clog << "x: " << x << " y: " << y;
               std::clog << " dist: " << dist << std::endl;
            }
         }
         outputStream << dist << " ";
         outputStream.flush();
      }
      outputStream << std::endl;
   }

   std::clog << "Minimum at x: " << xMin << " y: " << yMin
      << " dist: " << distMin << std::endl;

   std::clog << "Data array is "
      << (doColumn ? 1 : imgMainRGB.nbColumns())
      << " x "
      << (doRow ? 1 : imgMainRGB.nbRows())
      << std::endl;

   if (outputFile.is_open())
   {
      outputFile.close();
   }

   return(0);
}
