/*
 * imalib
 * 
 * Copyright 2017 Aberystwyth University
 * 
 * This library and program are free software: you can redistribute them
 * and/or modify them under the terms of the GNU General Public License as 
 * published by the Free Software Foundation, either version 3 of the License,
 * or (at your option) any later version.
 * 
 * The program and library are distributed in the hope that they will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*!
 * \file imaGrab.cpp
 * \brief A program to grab and display images using the imalib library.
 *
 * If available, GTK+ is used to display the images.
 *
 * \author Fred Labrosse <ffl@aber.ac.uk>
 * \copyright 2017 Aberystwyth University
 */

#include "imalib/imageRGB.h"

#include "imalib/imageGrabber.h"
#ifdef WITH_PICASSO
#  include "imalib/picassoImageGrabber.h"
#endif // WITH_PICASSO
#ifdef WITH_PICASSO_V4L
#  include "imalib/picassoV4LImageGrabber.h"
#endif // WITH_PICASSO_V4L
#ifdef WITH_V4L
#   include "imalib/v4lImageGrabber.h"
#endif // WITH_V4L
#ifdef WITH_V4L2
#   include "imalib/v4l2ImageGrabber.h"
#endif // WITH_V4L2
#ifdef WITH_DC1394
#   include "imalib/dc1394ImageGrabber.h"
#endif // WITH_DC1394
#ifdef WITH_UEYE
#   include "imalib/ueyeImageGrabber.h"
#endif // WITH_UEYE

#include <stdio.h>
#include <stdlib.h>     /* getenv(), etc. */
#include <termios.h>
#include <unistd.h>
#include <sys/time.h>

#ifdef WITH_GTK
#   include <gtk/gtk.h>
#   include <gdk-pixbuf/gdk-pixbuf.h>
GtkWidget* window;
int windowWidth = 384;
int windowHeight = 288;
#endif // WITH_GTK

#include <string>
#include <iostream>
#include <iomanip>

#include <boost/program_options.hpp>
namespace po = boost::program_options;

struct termios origTermios;

double firstTime;
double currentTime;
double lastTime;
unsigned long nbFrames;

int videoDevice;
int videoPort;
int cameraNumber;
int videoInput;
int videoStandard;
#ifdef WITH_V4L2
unsigned short ioMet =
   static_cast<unsigned short>(imalib::v4l2ImageGrabber::IO_METHOD_MMAP);
#endif // WITH_V4L2
unsigned short posX, posY, sourceWidth, sourceHeight;
unsigned short imageWidth;
unsigned short imageHeight;
bool statistics;
bool unWrap;
unsigned short outerRadius = 0;
unsigned short innerRadius = 0;
unsigned short centreX = 0;
unsigned short centreY = 0;
short angularRes = 1;
short direction = 1;
double power = 1.0;
bool blur;
unsigned short blurSize;
bool upsideDown;
bool flip;

bool save;                      // If true, then save the images.
bool saveOne = false;           // If true, save one image.
unsigned short savedCounter = 0;

bool carryOn = true;

std::ostringstream helpKeys;

enum possibleGrabberTypes
{
#ifdef WITH_V4L
   v4l = 0, // For a v4l image grabber.
#endif // WITH_V4L
#ifdef WITH_V4L2
   v4l2, // For a v4l2 image grabber.
#endif // WITH_V4L2
#ifdef WITH_PICASSO
   picasso, // For a picasso image grabber.
#endif // WITH_PICASSO
#ifdef WITH_PICASSO_V4L
   picassoV4L, // For a v4l picasso image grabber.
#endif // WITH_PICASSO_V4L
#ifdef WITH_DC1394
   dc1394, // For a dc1394 (firewire) camera.
#endif // WITH_DC1394
#ifdef WITH_UEYE
   ueye, // For a uEye USB camera.
#endif // WITH_UEYE
   fake // For a fake frame grabber.
} grabberType;


imalib::imageRGB theImage(100, 100);
imalib::imageGrabber* theGrabber;

/*
\brief Retrieves a timestamp in seconds using gettimeofday() for portable
timing
\return The current time since the  Epoch  (see time(2)).
*/
inline double time()
{
   timeval tv;
   gettimeofday(&tv, 0);

   return(static_cast<double>(tv.tv_sec)
         + static_cast<double>(tv.tv_usec) / 1000000.0);
}

void commandLine(int argc, char** argv)
{
   helpKeys
   << "While running:\n"
   << "- up, down, right, left will move the viewport,\n"
   << "- '>', '<', '^' and '_' will respectively increase/decrease\n"
   << "  the width and increase/decrease the height of the viewport\n"
   << "  (if supported).\n"
   << "- 'r' will reset the size of the window to the image size.\n"
   << "- 's' will save one image while 'S' will flip on or off continuous\n"
   << "  image saving.\n"
   << "- 'q' will quit.";

   po::options_description generalOptionsDesc("General options");
   generalOptionsDesc.add_options()
   ("help", "Prints this help message and exits")
   ("blurSize,b", po::value<unsigned short>(&blurSize)->default_value(1),
    "To blur the images (blurSize is the size of the square neighbourhood)")
   ("stats,s", "To display some live statistics")
   ;

   po::options_description cameraOptionsDesc("Camera options");
   cameraOptionsDesc.add_options()
#ifdef WITH_V4L
   ("v4l", "to grab using a v4l image grabber")
#endif // WITH_V4L
#ifdef WITH_V4L2
   ("v4l2", "to grab using a v4l2 image grabber (default)")
#endif // WITH_V4L2
#ifdef WITH_PICASSO
   ("picasso", "to grab using a picasso image grabber")
#endif // WITH_PICASSO
#ifdef WITH_PICASSO_V4L
   ("picassoV4L", "to grab using a v4l picasso image grabber")
#endif // WITH_PICASSO_V4L
#ifdef WITH_DC1394
   ("dc1394", "to grab using a dc1394 (firewire) camera")
#endif // WITH_DC1394
#ifdef WITH_UEYE
   ("ueye", "to grab using a uEye USB camera")
#endif // WITH_UEYE
   ("fake", "to grab using a fake image grabber"
#ifndef WITH_V4L2
   " (default)"
#endif // WITH_V4L2
   )
   ("videoDevice,d", po::value<int>(&videoDevice)->default_value(0),
   "Number of the video device [v4l, v4l2, picassoV4l, picasso]")
   ("videoPort,v", po::value<int>(&videoPort)->default_value(0),
    "Number of the port on the frame grabber [v4l, picassoV4l, picasso]")
   ("cameraNumber,c", po::value<int>(&cameraNumber)->default_value(0),
    "Number of the camera [dc1394, ueye]")
#ifdef WITH_V4L2
   ("input,i", po::value<int>(&videoInput)->default_value(0),
    "Number of the input to use [v4l2]")
   ("standard,S", po::value<int>(&videoStandard)->default_value(0),
    "Number of the video standard to use [v4l2]")
   ("ioMethod", po::value<unsigned short>(&ioMet)->default_value(ioMet),
    "v4l2 io method:\n0 for READ\n1 for MMAP\n2 for USRPTR")
#endif // WITH_V4L2
    ("posX,x", po::value<unsigned short>(&posX)->default_value(0),
    "X position of the source viewport")
   ("posY,y", po::value<unsigned short>(&posY)->default_value(0),
    "Y position of the source viewport")
   ("sourceWidth,w",
    po::value<unsigned short>(&sourceWidth)->
    default_value(imalib::IMAGE_DEFAULT_INPUTWIDTH),
    "Width of the viewport [v4l, v4l2, picasso, picassoV4L]")
   ("sourceHeight,h",
    po::value<unsigned short>(&sourceHeight)->
    default_value(imalib::IMAGE_DEFAULT_INPUTHEIGHT),
    "Height of the viewport [v4l, v4l2, picasso, picassoV4L]")
   ("imageWidth,W",
    po::value<unsigned short>(&imageWidth)->
    default_value(imalib::IMAGE_DEFAULT_INPUTWIDTH),
    "Width of the image")
   ("imageHeight,H",
    po::value<unsigned short>(&imageHeight)->
    default_value(imalib::IMAGE_DEFAULT_INPUTHEIGHT),
    "Height of the image")
   ("upsideDown,U", "To use if the camera is upsidedown")
   ("flip,f", "To horizontally flip the images")
   ;

   po::options_description
      unwrappingOptionsDesc("Unwrapping options (any will unwrap, unspecified options will be guessed)");
   unwrappingOptionsDesc.add_options()
   ("outerRadius", po::value<unsigned short>(&outerRadius),
   "The inner radius of the doughnut to unwrap")
   ("innerRadius", po::value<unsigned short>(&innerRadius),
   "The inner radius of the doughnut to unwrap")
   ("centreX", po::value<unsigned short>(&centreX),
   "X coordinate of the centre of the doughnut to unwrap")
   ("centreY", po::value<unsigned short>(&centreY),
   "Y coordinate of the centre of the doughnut to unwrap")
   ("power,u", po::value<double>(&power),
   "To unwrap the images (omni-directional to panoramic):"
   " this is the power of the polynomial function along a radial axis"
   " used to scale differently the inside and the outside"
   " (the higher the value, the more the outside is important)")
   ("pixelsPerDegree,R",
   po::value<short>(&angularRes),
   "Number of pixels per degree of angle when unwrapping")
   ("direction,D", po::value<short>(&direction),
   "Direction of the unwrapping from left to right in the images:"
   " 1 counter-clockwise, -1 clockwise");

   po::options_description optionsDesc("");
   optionsDesc.add(generalOptionsDesc);
   optionsDesc.add(cameraOptionsDesc);
   optionsDesc.add(unwrappingOptionsDesc);

   po::variables_map vm;
   try
   {
      po::store(po::parse_command_line(argc, argv, optionsDesc), vm);
      po::notify(vm);
   }
   catch (const std::exception& e)
   {
      std::cerr << "Error: " << e.what() << std::endl << std::endl
         << optionsDesc << std::endl;
      exit(0);
   }

   if (vm.count("help"))
   {
      std::cout
         << "Some of the options are only valid for some of the grabbers.  This is\n"
         << "indicated when necessary.\n\n"
         << optionsDesc << std::endl
         << helpKeys.str() << std::endl;
      exit(1);
   }

   if (vm.count("stats"))
   {
      statistics = true;
   }
   else
   {
      statistics = false;
   }

//    innerRadius = 50;
//    power = 1.0;
//    angularRes = 1;
//    direction = -1;
   if(vm.count("outerRadius")
      || vm.count("innerRadius")
      || vm.count("centreX")
      || vm.count("centreY")
      || vm.count("power")
      || vm.count("pixelsPerDegree")
      || vm.count("direction"))
   {
      unWrap = true;
      if (direction < 0)
         direction = -1;
      else
         direction = 1;
      // Other options will be guessed if not set once we have opened the image.
   }

   // Set the default grabber type.
#ifdef WITH_V4L2
   grabberType = v4l2;
#else // WITH_V4L2
   grabberType = fake;
#endif // WITH_V4L
  // Set the requested grabber type, if any.
#ifdef WITH_V4L
   if (vm.count("v4l"))
   {
      grabberType = v4l;
   }
#endif // WITH_V4L
#ifdef WITH_V4L2
if (vm.count("v4l2"))
{
   grabberType = v4l2;
}
#endif // WITH_V4L2
#ifdef WITH_PICASSO
   if (vm.count("picasso"))
   {
#include <iomanip>
      grabberType = picasso;
   }
#endif // WITH_PICASSO
#ifdef WITH_PICASSO_V4L
   if (vm.count("picassoV4L"))
   {
      grabberType = picassoV4L;
   }
#endif // WITH_PICASSO_V4L
#ifdef WITH_DC1394
   if (vm.count("dc1394"))
   {
      grabberType = dc1394;
   }
#endif // WITH_DC1394
#ifdef WITH_UEYE
   if (vm.count("ueye"))
   {
      grabberType = ueye;
   }
#endif // WITH_UEYE
   if (vm.count("fake"))
   {
      grabberType = fake;
   }

   if (blurSize != 1)
   {
      blur = true;
   }
   else
   {
      blur = false;
   }

   if (vm.count("upsideDown"))
   {
      upsideDown = true;
   }
   else
   {
      upsideDown = false;
   }

   if (vm.count("flip"))
   {
      flip = true;
   }
   else
   {
      flip = false;
   }
}

void saveImage()
{
   static std::string baseName = "grabbedImage_";
   static std::string extension = ".ppm";
   char fileName[25];

   sprintf(fileName, "%s%04d%s", baseName.c_str(), savedCounter,
         extension.c_str());
   bool result = theImage.write(fileName);
   if (!result)
   {
     std::cerr << "\nCould not save image " << savedCounter << std::endl;
   }
   else
   {
     savedCounter++;
   }
}

inline void flipSave()
{
   save = !save;
}

inline void doSaveOne()
{
   saveOne = true;
}

void quit()
{
   tcsetattr(fileno(stdin), TCSANOW, &origTermios);

   // Some statistics.
   currentTime = time();
   std::cout.flush();
   std::cout << "\n" << nbFrames << " frames at "
      << static_cast<double>(nbFrames) / (currentTime - firstTime)
      << " f/s.\n";

   carryOn = false;

#ifdef WITH_GTK
   gtk_main_quit();
#endif // WITH_GTK

//    exit(0);
}

#ifdef WITH_GTK
void resetWindowSize()
{
   windowWidth = theImage.nbColumns();
   windowHeight = theImage.nbRows();
   gtk_window_resize(GTK_WINDOW(window), windowWidth, windowHeight);
}
#endif // WITH_GTK

void processKeys()
{
   int key = getchar();
   if (key != EOF)
   {
      if (key == 'q')
         quit();
      else if (key == 's')
         doSaveOne();
      else if (key == 'S')
         flipSave();
#ifdef WITH_GTK
      else if (key == 'r')
         resetWindowSize();
#endif // WITH_GTK
      else if (key == 68)
      {
#ifdef WITH_PICASSO
         imalib::picassoImageGrabber* tmpPIG =
            dynamic_cast<imalib::picassoImageGrabber*>(theGrabber);
         if (tmpPIG)
            tmpPIG->moveLeft();
#endif // WITH_PICASSO
#ifdef WITH_PICASSO_V4L
         imalib::picassoV4LImageGrabber* tmpPV4LIG =
            dynamic_cast<imalib::picassoV4LImageGrabber*>(theGrabber);
         if (tmpPV4LIG)
            tmpPV4LIG->moveLeft();
#endif // WITH_PICASSO_V4L
#ifdef WITH_V4L
         imalib::v4lImageGrabber* tmpV4LIG =
            dynamic_cast<imalib::v4lImageGrabber*>(theGrabber);
         if (tmpV4LIG)
            tmpV4LIG->moveLeft();
#endif // WITH_V4L
#ifdef WITH_V4L2
         imalib::v4l2ImageGrabber* tmpV4L2IG =
            dynamic_cast<imalib::v4l2ImageGrabber*>(theGrabber);
         if (tmpV4L2IG)
            tmpV4L2IG->moveLeft();
#endif // WITH_V4L2
#ifdef WITH_DC1394
         imalib::dc1394ImageGrabber* tmpDCIG =
            dynamic_cast<imalib::dc1394ImageGrabber*>(theGrabber);
         if (tmpDCIG)
            tmpDCIG->moveLeft();
#endif // WITH_DC1394
#ifdef WITH_UEYE
         imalib::ueyeImageGrabber* tmpUEYEIG =
            dynamic_cast<imalib::ueyeImageGrabber*>(theGrabber);
         if (tmpUEYEIG)
            tmpUEYEIG->moveLeft();
#endif // WITH_UEYE
      }
      else if (key == 67)
      {
#ifdef WITH_PICASSO
         imalib::picassoImageGrabber* tmpPIG =
            dynamic_cast<imalib::picassoImageGrabber*>(theGrabber);
         if (tmpPIG)
            tmpPIG->moveRight();
#endif // WITH_PICASSO
#ifdef WITH_PICASSO_V4L
         imalib::picassoV4LImageGrabber* tmpPV4LIG =
            dynamic_cast<imalib::picassoV4LImageGrabber*>(theGrabber);
         if (tmpPV4LIG)
            tmpPV4LIG->moveRight();
#endif // WITH_PICASSO_V4L
#ifdef WITH_V4L
         imalib::v4lImageGrabber* tmpV4LIG =
            dynamic_cast<imalib::v4lImageGrabber*>(theGrabber);
         if (tmpV4LIG)
            tmpV4LIG->moveRight();
#endif // WITH_V4L
#ifdef WITH_V4L2
         imalib::v4l2ImageGrabber* tmpV4L2IG =
            dynamic_cast<imalib::v4l2ImageGrabber*>(theGrabber);
         if (tmpV4L2IG)
            tmpV4L2IG->moveRight();
#endif // WITH_V4L2
#ifdef WITH_DC1394
         imalib::dc1394ImageGrabber* tmpDCIG =
            dynamic_cast<imalib::dc1394ImageGrabber*>(theGrabber);
         if (tmpDCIG)
            tmpDCIG->moveRight();
#endif // WITH_DC1394
#ifdef WITH_UEYE
         imalib::ueyeImageGrabber* tmpUEYEIG =
            dynamic_cast<imalib::ueyeImageGrabber*>(theGrabber);
         if (tmpUEYEIG)
            tmpUEYEIG->moveRight();
#endif // WITH_UEYE
      }
      else if (key == 65)
      {
#ifdef WITH_PICASSO
         imalib::picassoImageGrabber* tmpPIG =
            dynamic_cast<imalib::picassoImageGrabber*>(theGrabber);
         if (tmpPIG)
            tmpPIG->moveUp();
#endif // WITH_PICASSO
#ifdef WITH_PICASSO_V4L
         imalib::picassoV4LImageGrabber* tmpPV4LIG =
            dynamic_cast<imalib::picassoV4LImageGrabber*>(theGrabber);
         if (tmpPV4LIG)
            tmpPV4LIG->moveUp();
#endif // WITH_PICASSO_V4L
#ifdef WITH_V4L
         imalib::v4lImageGrabber* tmpV4LIG =
            dynamic_cast<imalib::v4lImageGrabber*>(theGrabber);
         if (tmpV4LIG)
            tmpV4LIG->moveUp();
#endif // WITH_V4L
#ifdef WITH_V4L2
         imalib::v4l2ImageGrabber* tmpV4L2IG =
            dynamic_cast<imalib::v4l2ImageGrabber*>(theGrabber);
         if (tmpV4L2IG)
            tmpV4L2IG->moveUp();
#endif // WITH_V4L2
#ifdef WITH_DC1394
         imalib::dc1394ImageGrabber* tmpDCIG =
            dynamic_cast<imalib::dc1394ImageGrabber*>(theGrabber);
         if (tmpDCIG)
            tmpDCIG->moveUp();
#endif // WITH_DC1394
#ifdef WITH_UEYE
         imalib::ueyeImageGrabber* tmpUEYEIG =
            dynamic_cast<imalib::ueyeImageGrabber*>(theGrabber);
         if (tmpUEYEIG)
            tmpUEYEIG->moveUp();
#endif // WITH_UEYE
      }
      else if (key == 66)
      {
#ifdef WITH_PICASSO
         imalib::picassoImageGrabber* tmpPIG =
            dynamic_cast<imalib::picassoImageGrabber*>(theGrabber);
         if (tmpPIG)
            tmpPIG->moveDown();
#endif // WITH_PICASSO
#ifdef WITH_PICASSO_V4L
         imalib::picassoV4LImageGrabber* tmpPV4LIG =
            dynamic_cast<imalib::picassoV4LImageGrabber*>(theGrabber);
         if (tmpPV4LIG)
            tmpPV4LIG->moveDown();
#endif // WITH_PICASSO_V4L
#ifdef WITH_V4L
         imalib::v4lImageGrabber* tmpV4LIG =
            dynamic_cast<imalib::v4lImageGrabber*>(theGrabber);
         if (tmpV4LIG)
            tmpV4LIG->moveDown();
#endif // WITH_V4L
#ifdef WITH_V4L2
         imalib::v4l2ImageGrabber* tmpV4L2IG =
            dynamic_cast<imalib::v4l2ImageGrabber*>(theGrabber);
         if (tmpV4L2IG)
            tmpV4L2IG->moveDown();
#endif // WITH_V4L2
#ifdef WITH_DC1394
         imalib::dc1394ImageGrabber* tmpDCIG =
            dynamic_cast<imalib::dc1394ImageGrabber*>(theGrabber);
         if (tmpDCIG)
            tmpDCIG->moveDown();
#endif // WITH_DC1394
#ifdef WITH_UEYE
         imalib::ueyeImageGrabber* tmpUEYEIG =
            dynamic_cast<imalib::ueyeImageGrabber*>(theGrabber);
         if (tmpUEYEIG)
            tmpUEYEIG->moveDown();
#endif // WITH_UEYE
      }
      else if (key == '>')
      {
#ifdef WITH_PICASSO
         imalib::picassoImageGrabber* tmpPIG =
            dynamic_cast<imalib::picassoImageGrabber*>(theGrabber);
         if (tmpPIG)
            tmpPIG->increaseWidth();
#endif // WITH_PICASSO
#ifdef WITH_PICASSO_V4L
         imalib::picassoV4LImageGrabber* tmpPV4LIG =
            dynamic_cast<imalib::picassoV4LImageGrabber*>(theGrabber);
         if (tmpPV4LIG)
            tmpPV4LIG->increaseWidth();
#endif // WITH_PICASSO_V4L
#ifdef WITH_V4L
         imalib::v4lImageGrabber* tmpV4LIG =
            dynamic_cast<imalib::v4lImageGrabber*>(theGrabber);
         if (tmpV4LIG)
            tmpV4LIG->increaseWidth();
#endif // WITH_V4L
#ifdef WITH_V4L2
         imalib::v4l2ImageGrabber* tmpV4L2IG =
            dynamic_cast<imalib::v4l2ImageGrabber*>(theGrabber);
         if (tmpV4L2IG)
            tmpV4L2IG->increaseWidth();
#endif // WITH_V4L2
#ifdef WITH_DC1394
         imalib::dc1394ImageGrabber* tmpDCIG =
            dynamic_cast<imalib::dc1394ImageGrabber*>(theGrabber);
         if (tmpDCIG)
            tmpDCIG->increaseWidth();
#endif // WITH_DC1394
#ifdef WITH_UEYE
         imalib::ueyeImageGrabber* tmpUEYEIG =
            dynamic_cast<imalib::ueyeImageGrabber*>(theGrabber);
         if (tmpUEYEIG)
            tmpUEYEIG->increaseWidth();
#endif // WITH_UEYE
      }
      else if (key == '<')
      {
#ifdef WITH_PICASSO
         imalib::picassoImageGrabber* tmpPIG =
            dynamic_cast<imalib::picassoImageGrabber*>(theGrabber);
         if (tmpPIG)
            tmpPIG->decreaseWidth();
#endif // WITH_PICASSO
#ifdef WITH_PICASSO_V4L
         imalib::picassoV4LImageGrabber* tmpPV4LIG =
            dynamic_cast<imalib::picassoV4LImageGrabber*>(theGrabber);
         if (tmpPV4LIG)
            tmpPV4LIG->decreaseWidth();
#endif // WITH_PICASSO_V4L
#ifdef WITH_V4L
         imalib::v4lImageGrabber* tmpV4LIG =
            dynamic_cast<imalib::v4lImageGrabber*>(theGrabber);
         if (tmpV4LIG)
            tmpV4LIG->decreaseWidth();
#endif // WITH_V4L
#ifdef WITH_V4L2
         imalib::v4l2ImageGrabber* tmpV4L2IG =
            dynamic_cast<imalib::v4l2ImageGrabber*>(theGrabber);
         if (tmpV4L2IG)
            tmpV4L2IG->decreaseWidth();
#endif // WITH_V4L2
#ifdef WITH_DC1394
         imalib::dc1394ImageGrabber* tmpDCIG =
            dynamic_cast<imalib::dc1394ImageGrabber*>(theGrabber);
         if (tmpDCIG)
            tmpDCIG->decreaseWidth();
#endif // WITH_DC1394
#ifdef WITH_UEYE
         imalib::ueyeImageGrabber* tmpUEYEIG =
            dynamic_cast<imalib::ueyeImageGrabber*>(theGrabber);
         if (tmpUEYEIG)
            tmpUEYEIG->decreaseWidth();
#endif // WITH_UEYE
      }
      else if (key == '^')
      {
#ifdef WITH_PICASSO
         imalib::picassoImageGrabber* tmpPIG =
            dynamic_cast<imalib::picassoImageGrabber*>(theGrabber);
         if (tmpPIG)
            tmpPIG->increaseHeight();
#endif // WITH_PICASSO
#ifdef WITH_PICASSO_V4L
         imalib::picassoV4LImageGrabber* tmpPV4LIG =
            dynamic_cast<imalib::picassoV4LImageGrabber*>(theGrabber);
         if (tmpPV4LIG)
            tmpPV4LIG->increaseHeight();
#endif // WITH_PICASSO_V4L
#ifdef WITH_V4L
         imalib::v4lImageGrabber* tmpV4LIG =
            dynamic_cast<imalib::v4lImageGrabber*>(theGrabber);
         if (tmpV4LIG)
            tmpV4LIG->increaseHeight();
#endif // WITH_V4L
#ifdef WITH_V4L2
         imalib::v4l2ImageGrabber* tmpV4L2IG =
            dynamic_cast<imalib::v4l2ImageGrabber*>(theGrabber);
         if (tmpV4L2IG)
            tmpV4L2IG->increaseHeight();
#endif // WITH_V4L2
#ifdef WITH_DC1394
         imalib::dc1394ImageGrabber* tmpDCIG =
            dynamic_cast<imalib::dc1394ImageGrabber*>(theGrabber);
         if (tmpDCIG)
            tmpDCIG->increaseHeight();
#endif // WITH_DC1394
#ifdef WITH_UEYE
         imalib::ueyeImageGrabber* tmpUEYEIG =
            dynamic_cast<imalib::ueyeImageGrabber*>(theGrabber);
         if (tmpUEYEIG)
            tmpUEYEIG->increaseHeight();
#endif // WITH_UEYE
      }
      else if (key == '_')
      {
#ifdef WITH_PICASSO
         imalib::picassoImageGrabber* tmpPIG =
            dynamic_cast<imalib::picassoImageGrabber*>(theGrabber);
         if (tmpPIG)
            tmpPIG->decreaseHeight();
#endif // WITH_PICASSO
#ifdef WITH_PICASSO_V4L
         imalib::picassoV4LImageGrabber* tmpPV4LIG =
            dynamic_cast<imalib::picassoV4LImageGrabber*>(theGrabber);
         if (tmpPV4LIG)
            tmpPV4LIG->decreaseHeight();
#endif // WITH_PICASSO_V4L
#ifdef WITH_V4L
         imalib::v4lImageGrabber* tmpV4LIG =
            dynamic_cast<imalib::v4lImageGrabber*>(theGrabber);
         if (tmpV4LIG)
            tmpV4LIG->decreaseHeight();
#endif // WITH_V4L
#ifdef WITH_V4L2
         imalib::v4l2ImageGrabber* tmpV4L2IG =
            dynamic_cast<imalib::v4l2ImageGrabber*>(theGrabber);
         if (tmpV4L2IG)
            tmpV4L2IG->decreaseHeight();
#endif // WITH_V4L2
#ifdef WITH_DC1394
         imalib::dc1394ImageGrabber* tmpDCIG =
            dynamic_cast<imalib::dc1394ImageGrabber*>(theGrabber);
         if (tmpDCIG)
            tmpDCIG->decreaseHeight();
#endif // WITH_DC1394
#ifdef WITH_UEYE
         imalib::ueyeImageGrabber* tmpUEYEIG =
            dynamic_cast<imalib::ueyeImageGrabber*>(theGrabber);
         if (tmpUEYEIG)
            tmpUEYEIG->decreaseHeight();
#endif // WITH_UEYE
      }
   }
}

void initNonBlocking()
{
   struct termios newTermios;

   tcgetattr(fileno(stdin), &origTermios);

   tcgetattr(fileno(stdin), &newTermios);

   newTermios.c_cc[VTIME] = 0;
   newTermios.c_cc[VMIN] = 0;
   newTermios.c_lflag &= (~ECHO & ~ICANON);
   tcsetattr(fileno(stdin), TCSANOW, &newTermios);
}

#ifdef WITH_GTK
void resize(GtkWidget* /*widget*/, GtkAllocation* size, gpointer /*ptr*/)
{
   windowWidth = size->width;
   windowHeight = size->height;
}
#endif // WITH_GTK

int animate(void* impixmap)
{
   processKeys();

   if (carryOn)
   {
      theGrabber->getImage(&theImage);

      nbFrames++;

#ifdef WITH_GTK
      static GdkPixbuf * pixbuf;
      static GdkPixbuf * prescaled;
      static GdkPixmap * pixmap;
      pixbuf = gdk_pixbuf_new_from_data(theImage.getActualPixel(0, 0),
                                        GDK_COLORSPACE_RGB, FALSE, 8,
                                        theImage.nbColumns(), theImage.nbRows(),
                                        theImage.nbColumns() * 3,
                                        NULL, NULL);
      if (!pixbuf)
      {
         std::cerr << "Could not show image\n";
         return(TRUE);
      }
      if ((windowWidth != theImage.nbColumns())
         || (windowHeight != theImage.nbRows()))
      {
         prescaled = pixbuf;
         pixbuf = gdk_pixbuf_scale_simple(prescaled,
                                          windowWidth, windowHeight,
                                          GDK_INTERP_BILINEAR);
         g_object_unref(prescaled);
         if (!pixbuf)
         {
            std::cerr << "Cannot rescale to size " << windowWidth
               << "x" << windowHeight << ".\n";
            return(TRUE);
         }
      }
      gdk_pixbuf_render_pixmap_and_mask(pixbuf, &pixmap, NULL, 1);
      gdk_draw_pixmap(GTK_WIDGET(impixmap)->window,
                      GTK_WIDGET(impixmap)->style->fg_gc[GTK_WIDGET_STATE(GTK_WIDGET(impixmap))],
                      pixmap, 0, 0, 0, 0, windowWidth, windowHeight);
      g_object_unref(pixbuf);
      g_object_unref(pixmap);
#endif // WITH_GTK

      if (statistics)
      {
         lastTime = currentTime;
         currentTime = time();
         std::cout << "\rFrame: " << std::setw(6) << nbFrames << " ("
            << std::fixed << std::setprecision(2) << std::setw(6)
            << 1.0 / (currentTime - lastTime) << " f/s)";
         if (save || saveOne)
            std::cout << " (saved image " << std::setw(5)
               << savedCounter+1 << ")";
         std::cout.flush();
      }

      if (saveOne)
      {
         saveImage();
         saveOne = false;
      }
      else if (save)
         saveImage();
   }

   return(carryOn);
}


int main(int argc, char** argv)
{
   commandLine(argc, argv);

   switch (grabberType)
   {
#ifdef WITH_V4L
      case v4l:
         theGrabber =
            new imalib::v4lImageGrabber(imageWidth, imageHeight,
                                        upsideDown,
                                        flip,
                                        videoDevice, videoPort,
                                        posX, posY,
                                        sourceWidth, sourceHeight);
         break;
#endif // WITH_V4L
#ifdef WITH_V4L2
      case v4l2:
         theGrabber =
            new imalib::v4l2ImageGrabber(imageWidth, imageHeight,
                                         upsideDown,
                                         flip,
                                         videoDevice,
                                         videoInput,
                                         videoStandard,
                                         static_cast<imalib::v4l2ImageGrabber::possibleIoMethods>(ioMet),
                                         posX, posY,
                                         sourceWidth, sourceHeight);
                                         break;
#endif // WITH_V4L2
#ifdef WITH_PICASSO
      case picasso:
         theGrabber =
            new imalib::picassoImageGrabber(imageWidth, imageHeight,
                                            upsideDown,
                                            flip,
                                            videoDevice, videoPort,
                                            posX, posY,
                                            sourceWidth, sourceHeight);
         break;
#endif // WITH_PICASSO
#ifdef WITH_PICASSO_V4L
      case picassoV4L:
         theGrabber =
            new imalib::v4lImageGrabber(imageWidth, imageHeight,
                                        upsideDown,
                                        flip,
                                        videoDevice, videoPort,
                                        posX, posY,
                                        sourceWidth, sourceHeight);
         break;
#endif // WITH_PICASSO_V4L
#ifdef WITH_DC1394
      case dc1394:
         theGrabber =
            new imalib::dc1394ImageGrabber(upsideDown,
                                           flip,
                                           cameraNumber,
                                           posX, posY,
                                           imageWidth, imageHeight);
         if (theGrabber->isConnected())
            static_cast<imalib::dc1394ImageGrabber*>(theGrabber)->setExposure(-1);
         break;
#endif // WITH_DC1394
#ifdef WITH_UEYE
      case ueye:
         theGrabber =
            new imalib::ueyeImageGrabber(upsideDown,
                                         flip,
                                         cameraNumber,
                                         posX, posY,
                                         imageWidth, imageHeight);
         break;
#endif // WITH_UEYE
      case fake:
      default:
         theGrabber = NULL;
   }

   if (!theGrabber || !theGrabber->isConnected())
   {
      if (grabberType != fake)
      {
         if (!theGrabber)
         {
            std::cerr << "\n--> Compiled without image grabber support:";
         }
         else
         {
            std::cerr << "\n--> Could not connect to a frame-grabber:";
            delete theGrabber;
         }
         std::cerr << " using a fake one instead.\n\n";
      }
      theGrabber = new imalib::imageGrabber(imageWidth, imageHeight,
                                            upsideDown, flip);
   }

   if (unWrap)
   {
     if (centreX == 0)
     {
       centreX = imageWidth / 2;
     }
     if (centreY == 0)
     {
       centreY = imageHeight / 2;
     }
     if (outerRadius == 0)
     {
       outerRadius = fmin(imageWidth, imageHeight) / 2;
       outerRadius = fmin(outerRadius, centreX);
       outerRadius = fmin(outerRadius, imageWidth - centreX);
       outerRadius = fmin(outerRadius, centreY);
       outerRadius = fmin(outerRadius, imageHeight - centreY);
     }
     if (innerRadius == 0)
     {
       innerRadius = outerRadius * 0.4;
     }
     else if (innerRadius >= outerRadius)
     {
       std::cerr << "Inner radius needs to be smaller than the outer radius."
         << "Changed...\n";
       innerRadius = outerRadius - 5;
     }
     theGrabber->doUnwrap(true, outerRadius, innerRadius,
                          centreX, centreY,
                          angularRes,
                          power, direction);
   }
   if (blur)
      theGrabber->doBlur(true, blurSize, blurSize);

   theGrabber->printInfoAboutSource();

   theGrabber->getImage(&theImage);

   initNonBlocking();

#ifdef WITH_GTK
   windowWidth = theImage.nbColumns();
   windowHeight = theImage.nbRows();
   GtkWidget* impixmap;
   gtk_init(&argc, &argv);

   window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
   gtk_window_set_title(GTK_WINDOW(window),
                           "imagrab (from imalib)");
   gtk_container_set_border_width(GTK_CONTAINER(window), 2);
   gtk_signal_connect(GTK_OBJECT(window), "delete-event",
                     GTK_SIGNAL_FUNC(quit), NULL);
   impixmap = GTK_WIDGET(gtk_drawing_area_new());
   gtk_widget_set_usize(GTK_WIDGET(impixmap), windowWidth, windowHeight);
   gtk_signal_connect(GTK_OBJECT(impixmap), "size-allocate",
                     GTK_SIGNAL_FUNC(resize), NULL);
   gtk_container_add(GTK_CONTAINER(window), GTK_WIDGET(impixmap));
   gtk_widget_show_all(GTK_WIDGET(window));
   gtk_idle_add(animate, impixmap);
   // We set a small size so that we can make it small afterwards.
   gtk_widget_set_usize(GTK_WIDGET(impixmap), 4, 3);
#else // WITH_GTK
   std::cout << "\nGTK+ is not available.\n"
            << "--> Images will not be displayed.\n";
#endif // WITH_GTK

   firstTime = time();
   currentTime = firstTime;
   lastTime = firstTime;
   nbFrames = 0;

   std::cout << helpKeys.str() << std::endl;

#ifdef WITH_GTK
   gtk_main();
#else // WITH_GTK
   while(carryOn)
   {
      animate(NULL);
   }
#endif // WITH_GTK

   delete theGrabber;

   return(0);
}
