/*
 * imalib
 * 
 * Copyright 2017 Aberystwyth University
 * 
 * This library and program are free software: you can redistribute them
 * and/or modify them under the terms of the GNU General Public License as 
 * published by the Free Software Foundation, either version 3 of the License,
 * or (at your option) any later version.
 * 
 * The program and library are distributed in the hope that they will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*!
 * \file imaTransf.cpp
 * \brief A program to transform images using the imalib library
 * 
 * \author Fred Labrosse <ffl@aber.ac.uk>
 * \copyright 2017 Aberystwyth University
 */

#include "imalib/imageRGB.h"
#include "imalib/imageLab.h"
#include "imalib/imageMapping.h"

#include <stdio.h>
#include <stdlib.h>     /* getenv(), etc. */
#include <unistd.h>

#include <string>
#include <iostream>

const unsigned short BLURSIZE = 0;
const int INNERRADIUS = 50;

std::string imageName;
std::string newName;
std::string transf;
bool panoramic = false;

void displayUsage(const std::string& commandName)
{
   std::cerr
      << std::endl
      << "Usage:" << std::endl
      << std::endl
      << commandName << " <image> [-p] <transfImage> <transf>\n"
      << std::endl
      << "where:" << std::endl
      << std::endl
      << "<image> is the name of the file containing the image\n"
      << "-p: the image is a panoramic image.\n"
      << "<transfImage> is the name of the file that will contain\n"
      << "              the transformed image\n"
      << "<transf> is a string specifying the transformation:\n"
      << "\tnoop: does nothing ;-)\n"
      << "\tup: translate up\n"
      << "\tdown: translate down\n"
      << "\tright: translate to the right\n"
      << "\tshiftright: shift to the right\n"
      << "\tshiftleft: shift to the left\n"
      << "\tunwrap: unwrap the image (must be 200x200 pixels)\n"
      << "\twarp: warp the image using some sine transformation\n"
      << "\t      of the image border\n"
      << "\tlumin: set the luminance of the image to 90.\n"
      << "\tCIEab: transform the image in CIE L*a*b* and only save\n"
      << "\t       the a* and b* component as R and G (B set to 0)\n"
      << "\tblur: blur the image (mask of " << BLURSIZE << "x"
      << BLURSIZE <<")\n";
}

void commandLine(int argc, char** argv)
{
   // Options
   int option;

   while ((option = getopt(argc, argv, "ph")) != EOF)
   {
      switch(option)
      {
         case 'p':
            panoramic = true;
            break;
         case 'h':
         case '?':
            displayUsage(argv[0]);
            exit(1);
            break;
      }
   }

   if (optind >= argc)
   {
      std::cerr << "Missing needed argument: <image>.\n";
      displayUsage(argv[0]);
      exit(1);
   }
   imageName = argv[optind];
   optind++;
   if (optind >= argc)
   {
      std::cerr << "Missing needed argument: <transfImage>.\n";
      displayUsage(argv[0]);
      exit(1);
   }
   newName = argv[optind];
   optind++;
   if (optind >= argc)
   {
      std::cerr << "Missing needed argument: <transf>.\n";
      displayUsage(argv[0]);
      exit(1);
   }
   transf = argv[optind];
}

double xWarpingForward(double deg, double ampl, double x)
{
   double xp = 0.0;
   if (x <= 90)
   {
      xp = x / 90.0;
   }
   else if ((x > 90) && (x <= 180))
   {
      xp = (-x + 180.0) / 90.0;
      ampl *= -1.0;
   }
   else if ((x > 180) && (x <= 270))
   {
      xp = (x - 180.0) / 90.0;
      ampl *= -1.0;
   }
   else if ((x > 270) && (x <= 360))
   {
      xp = (-x + 360.0) / 90.0;
   }
   return(- ampl * (1 - pow(xp, deg)));
}

double yWarpingForward(double deg, double ampl, double x)
{
   double xp = 0.0;
   if (x <= 90)
   {
      xp = (-x + 90.0) / 90.0;
   }
   else if ((x > 90) && (x <= 180))
   {
      xp = (x - 90.0) / 90.0;
   }
   else if ((x > 180) && (x <= 270))
   {
      xp = (-x + 270.0) / 90.0;
      ampl *= -1.0;
   }
   else if ((x > 270) && (x <= 360))
   {
      xp = (x - 270.0) / 90.0;
      ampl *= -1.0;
   }
   return(- ampl * (1 - pow(xp, deg)));
//    return(- ampl * sin(x * M_PI / 180.0));
}

double xWarpingSide(double deg, double ampl, double x)
{
   double xp = 0.0;
   if (x <= 90)
   {
      xp = (-x + 90.0) / 90.0;
   }
   else if ((x > 90) && (x <= 180))
   {
      xp = (x - 90.0) / 90.0;
   }
   else if ((x > 180) && (x <= 270))
   {
      xp = (-x + 270.0) / 90.0;
      ampl *= -1.0;
   }
   else if ((x > 270) && (x <= 360))
   {
      xp = (x - 270.0) / 90.0;
      ampl *= -1.0;
   }
   return(- ampl * (1 - pow(xp, deg)));
}

double yWarpingSide(double deg, double ampl, double x)
{
   double xp = 0.0;
   if (x <= 90)
   {
      xp = x / 90.0;
   }
   else if ((x > 90) && (x <= 180))
   {
      xp = (-x + 180.0) / 90.0;
      ampl *= -1.0;
   }
   else if ((x > 180) && (x <= 270))
   {
      xp = (x - 180.0) / 90.0;
      ampl *= -1.0;
   }
   else if ((x > 270) && (x <= 360))
   {
      xp = (-x + 360.0) / 90.0;
   }
   return(- ampl * (1 - pow(xp, deg)));
//    return(- ampl * cos(x * M_PI / 180.0));
}

int main(int argc, char** argv)
{
   commandLine(argc, argv);

   imalib::imageRGB theImage;

   if (!theImage.read(imageName))
   {
      exit(1);
   }

   imalib::imageMapping mapping;
   imalib::imageMapping unwrapMapping;

   if (panoramic)
      unwrapMapping.createUnwrappingMapping(theImage.nbColumns());

   if (transf == "up")
   {
      if (panoramic)
      {
         mapping.createVerticalShiftMapping(200, 200,
                                            -10,
                                            75, 125);
      }
      else
      {
         mapping.createVerticalShiftMapping(theImage.nbColumns(),
                                            theImage.nbRows(),
                                            -10,
                                            (theImage.nbRows() / 2
                                             - theImage.nbRows() / 8),
                                            (theImage.nbRows() / 2
                                             + theImage.nbRows() / 8));
      }
   }
   if (transf == "down")
   {
      if (panoramic)
      {
         mapping.createVerticalShiftMapping(200, 200,
                                            10,
                                            75, 125);
      }
      else
      {
         mapping.createVerticalShiftMapping(theImage.nbColumns(),
                                            theImage.nbRows(),
                                            10,
                                            (theImage.nbRows() / 2
                                             - theImage.nbRows() / 8),
                                            (theImage.nbRows() / 2
                                             + theImage.nbRows() / 8));
      }
   }
   else if (transf == "right")
   {
      if (panoramic)
      {
         mapping.createHorizontalShiftMapping(200, 200,
                                              10,
                                              75, 125);
      }
      else
      {
         mapping.createHorizontalShiftMapping(theImage.nbColumns(),
                                              theImage.nbRows(),
                                              10,
                                              (theImage.nbColumns() / 2
                                               - theImage.nbColumns() / 8),
                                              (theImage.nbColumns() / 2
                                               + theImage.nbColumns() / 8));
      }
   }
   else if (transf == "noop")
   {
      mapping.createNoopMapping(theImage.nbColumns(),
                                theImage.nbRows());
   }
   else if (transf == "unwrap")
   {
      mapping.createUnwrappingMapping(theImage.nbColumns(),
                                      INNERRADIUS,
                                      1, // pixel per degree
                                      1.0, // power
                                      -1); // direction
   }
   else if (transf == "warp")
   {
      // Let's create a sine wave with nbSeg + 1 points for the different
      // boundaries.
      imalib::curve top, bottom, right, left;
      imalib::curve topR, bottomR, rightR, leftR; // Re-sampled curves.

      unsigned short nbSeg = 20;
      unsigned short amplY = 8;
      unsigned short amplX = 9;
      double dX = static_cast<double>(theImage.nbColumns() - 1)
         / static_cast<double>(nbSeg);
      for (double ind = 0; ind <= nbSeg; ++ind)
      {
         top.points.
//             push_back(imalib::point((ind * dX
//                                      + xWarpingSide(5, amplX, ind * dX)),
//                                     (-yWarpingSide(2, amplY, ind * dX))));
            push_back(imalib::point((ind * dX
                                     + xWarpingForward(5, amplX, ind * dX)),
                                    yWarpingForward(2, amplY, ind * dX)));
         bottom.points.
//             push_back(imalib::point((theImage.nbColumns() - 1 - ind * dX
//                                      - xWarpingSide(5, amplX, ind * dX)),
//                                     (theImage.nbRows() - 1
//                                      + yWarpingSide(2, amplY, ind * dX))));
            push_back(imalib::point((theImage.nbColumns() - 1 - ind * dX
                                     + xWarpingForward(5, amplX, ind * dX)),
                                    (theImage.nbRows() - 1
                                     + yWarpingForward(2, amplY, ind * dX))));
      }
      right.points.push_back(top.points.back());
      right.points.push_back(bottom.points.front());

      left.points.push_back(bottom.points.back());
      left.points.push_back(top.points.front());

      top.resample(&topR, theImage.nbColumns(), true);
      bottom.resample(&bottomR, theImage.nbColumns(), true);
      right.resample(&rightR, theImage.nbRows(), true);
      left.resample(&leftR, theImage.nbRows(), true);
//       for (unsigned short ind = 0; ind < topR.points.size() - 1; ++ind)
//          std::cout << topR.points[ind].x
//                    << std::endl;
      mapping.createWarpingMapping(theImage.nbColumns(), theImage.nbRows(),
                                   topR, rightR, bottomR, leftR,
                                   true);
   }
   else if (transf == "lumin")
   {
      // Nothing to do here.
   }
   else if (transf == "CIEab")
   {
      // Nothing to do here.
   }
   else if (transf == "blur")
   {
      // Nothing to do here.
   }
//    else
//    {
//       std::cerr << "Unknown transformation: " << transf << std::endl;
//       displayUsage(argv[0]);
//       exit(1);
//    }

   if (panoramic)
   {
      mapping.apply(unwrapMapping);
   }

   imalib::imageRGB newImage;

   if (transf == "shiftright")
      theImage.shift(2, &newImage);
   else if (transf == "shiftleft")
      theImage.shift(-10, &newImage);
   else if (transf == "lumin")
   {
      imalib::imageLab imgLab(theImage);
      imgLab.changeLuminance(90);
      imgLab.toRGB(&newImage);
   }
   else if (transf == "CIEab")
   {
      imalib::imageLab imgLab(theImage);
      imgLab.copyAB(&newImage, 1.5);
   }
   else if (transf == "blur")
   {
      newImage.setSize(theImage.nbColumns(), theImage.nbRows());
      theImage.scaleBlur(BLURSIZE, BLURSIZE, &newImage);
   }
   else
      theImage.apply(mapping, &newImage);

   newImage.write(newName);

   return(0);
}
