/*
 * imalib
 * 
 * Copyright 2017 Aberystwyth University
 * 
 * This library and program are free software: you can redistribute them
 * and/or modify them under the terms of the GNU General Public License as 
 * published by the Free Software Foundation, either version 3 of the License,
 * or (at your option) any later version.
 * 
 * The program and library are distributed in the hope that they will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*!
 * \file imaView.cpp
 * \brief A program to view images using the imalib library
 *
 * It uses imlib and gdk for the visualisation.  This is just an example on
 * how to use the imalib classes.
 *
 * \author Fred Labrosse <ffl@aber.ac.uk>
 * \copyright 2017 Aberystwyth University
 */

#include "imalib/imageRGB.h"
#include "imalib/imageMapping.h"

#include <stdio.h>
#include <stdlib.h>     /* getenv(), etc. */

#include <string>
#include <iostream>

#include <boost/program_options.hpp>
namespace po = boost::program_options;

#include <gtk/gtk.h>
#include <gdk-pixbuf/gdk-pixbuf.h>
#include <gdk/gdkkeysyms.h>

int windowWidth = 384;
int windowHeight = 288;
double scale = 1.0;

std::string imageName;
bool thumbnail = false;
int ratio = 2;
bool blur = false;
int blurWidth = 3;
int blurHeight = 3;
unsigned short thumbWidth = 0;
unsigned short thumbHeight = 0;

bool unWrap = false;
unsigned short outerRadius = 0;
unsigned short innerRadius = 0;
unsigned short centreX = 0;
unsigned short centreY = 0;
short angularRes = 1;
short direction = 1;
double power = 1.0;

imalib::imageRGB imageOrig;
imalib::imageRGB imageTransf;
imalib::imageRGB* imgPtr;
GtkWidget* impixmapImage;
GtkWidget* impixmapThumb;

GtkWidget* windowImage;
GtkWidget* windowThumb;

imalib::imageMapping imgMap;

void commandLine(int argc, char** argv)
{
   std::ostringstream usage;
   usage << "Usage:\n" << "\timaview {options} {file name}";

   std::ostringstream helpKeys;
   helpKeys
      << "While running, 'r' will reload the image.";

   po::options_description optionsDesc("General options");

   imalib::imalibOptions generalOptions;

   optionsDesc.add_options()
   ("help", "Prints this help message and exits")
   ("scale,s",  po::value<double>(&scale),
    "Set the (initial) scale the image by the given value")
   ("thumb,t",  po::value<int>(&ratio),
    "Display a thumbnail of size the image size divided by the given value")
   ("thumbWidth,w",  po::value<unsigned short>(&thumbWidth),
    "Set the width of the thumbnail")
   ("thumbHeight,h",  po::value<unsigned short>(&thumbHeight),
    "Set the height of the thumbnail")
   ("blurWidth,W", po::value<int>(&blurWidth),
    "Width of the window used to blur the thumbnail")
   ("blurHeight,H", po::value<int>(&blurHeight),
    "Height of the window used to blur the thumbnail")
   ;

   po::options_description
      unwrappingOptionsDesc("Unwrapping options (any will unwrap)");
   unwrappingOptionsDesc.add_options()
   ("outerRadius", po::value<unsigned short>(&outerRadius),
   "The inner radius of the doughnut to unwrap")
   ("innerRadius", po::value<unsigned short>(&innerRadius),
   "The inner radius of the doughnut to unwrap")
   ("centreX", po::value<unsigned short>(&centreX),
   "X coordinate of the centre of the doughnut to unwrap")
   ("centreY", po::value<unsigned short>(&centreY),
   "Y coordinate of the centre of the doughnut to unwrap")
   ("power,u", po::value<double>(&power),
   "To unwrap the images (omni-directional to panoramic):"
   " this is the power of the polynomial function along a radial axis"
   " used to scale differently the inside and the outside"
   " (the higher the value, the more the outside is important)")
   ("pixelsPerDegree,R",
   po::value<short>(&angularRes),
   "Number of pixels per degree of angle when unwrapping")
   ("direction,D", po::value<short>(&direction),
   "Direction of the unwrapping from left to right in the images:"
   " 1 counter-clockwise, -1 clockwise");

   po::options_description fileNamesOptions("Hidden options");
   fileNamesOptions.add_options()
   ("file", po::value<std::string>(&imageName)->default_value(""),
    "Name of the image to display")
   ;

   po::options_description cmdlineOptions;
   optionsDesc.add(*(generalOptions.getOptions()));
   optionsDesc.add(unwrappingOptionsDesc);
   
   cmdlineOptions.add(optionsDesc);
   cmdlineOptions.add(fileNamesOptions);
   
   po::positional_options_description p;
   p.add("file", -1);

   po::variables_map vm;
   try
   {
      po::store(po::command_line_parser(argc, argv).
      options(cmdlineOptions).positional(p).run(), vm);
      notify(vm);
   }
   catch (const std::exception& e)
   {
      std::cerr << "Error: " << e.what() << std::endl << std::endl
         << usage.str() << std::endl << std::endl
         << optionsDesc << std::endl;
      exit(0);
   }

   try
   {
      po::store(po::parse_command_line(argc, argv, optionsDesc), vm);
      po::notify(vm);
   }
   catch (const std::exception& e)
   {
      std::cerr << "Error: " << e.what() << std::endl << std::endl
         << usage.str() << std::endl << std::endl
         << optionsDesc << std::endl;
      exit(0);
   }

   generalOptions.commandLine(vm);

   if (vm.count("help"))
   {
      std::cout << usage.str() << std::endl << std::endl
         << optionsDesc << std::endl
         << helpKeys.str() << std::endl;
      exit(1);
   }

   if (vm.count("thumb") || vm.count("thumbHeight") || vm.count("thumbWidth")
       || vm.count("blurWidth") || vm.count("blurHeight"))
   {
      thumbnail = true;
   }

   if (vm.count("blurWidth") || vm.count("blurHeight"))
   {
      blur = true;
   }

   if(vm.count("outerRadius")
      || vm.count("innerRadius")
      || vm.count("centreX")
      || vm.count("centreY")
      || vm.count("power")
      || vm.count("pixelsPerDegree")
      || vm.count("direction"))
   {
      unWrap = true;
      if (direction < 0)
         direction = -1;
      else
         direction = 1;
      // Other options will be guessed if not set once we have opened the image.
   }

   if (imageName == "")
   {
      std::cerr << "Error: missing image name!" << std::endl << std::endl
         << usage.str() << std::endl << std::endl
         << optionsDesc << std::endl;
      exit(0);
   }
}

void paintImage()
{
   static GdkPixbuf * pixbuf;
   static GdkPixbuf * prescaled;
   static GdkPixmap * pixmap;

   pixbuf = gdk_pixbuf_new_from_data(imgPtr->getActualPixel(0, 0),
                                     GDK_COLORSPACE_RGB, FALSE, 8,
                                     imgPtr->nbColumns(), imgPtr->nbRows(),
                                     imgPtr->nbColumns() * 3,
                                     NULL, NULL);
   if (!pixbuf)
   {
      std::cerr << "Could not show image\n";
//       return(TRUE);
   }

   if ((windowWidth != imgPtr->nbColumns())
      || (windowHeight != imgPtr->nbRows()))
   {
      prescaled = pixbuf;
      pixbuf = gdk_pixbuf_scale_simple(prescaled,
                                       windowWidth, windowHeight,
                                       GDK_INTERP_BILINEAR);
      g_object_unref(prescaled);
      if (!pixbuf)
      {
         std::cerr << "Cannot rescale to size " << windowWidth
            << "x" << windowHeight << ".\n";
      }
   }

   gdk_pixbuf_render_pixmap_and_mask(pixbuf, &pixmap, NULL, 1);
   gdk_draw_pixmap(GTK_WIDGET(impixmapImage)->window,
                   GTK_WIDGET(impixmapImage)->style->fg_gc[GTK_WIDGET_STATE(GTK_WIDGET(impixmapImage))],
                   pixmap, 0, 0, 0, 0,
                   windowWidth, windowHeight);

   g_object_unref(pixbuf);
   gdk_pixmap_unref(pixmap);
}

void paintThumb()
{
   static GdkPixbuf * pixbuf;
   static GdkPixmap * pixmap;

   pixbuf = gdk_pixbuf_new_from_data(imgPtr->getActualThumbnailPixel(0, 0),
                                     GDK_COLORSPACE_RGB, FALSE, 8,
                                     imgPtr->nbThumbColumns(),
                                     imgPtr->nbThumbRows(),
                                     imgPtr->nbThumbColumns() * 3,
                                     NULL, NULL);
   if (!pixbuf)
   {
      std::cerr << "Could not show image\n";
//       return(TRUE);
   }
   gdk_pixbuf_render_pixmap_and_mask(pixbuf, &pixmap, NULL, 1);
   gdk_draw_pixmap(GTK_WIDGET(impixmapThumb)->window,
                   GTK_WIDGET(impixmapThumb)->style->fg_gc[GTK_WIDGET_STATE(GTK_WIDGET(impixmapThumb))],
                   pixmap, 0, 0, 0, 0,
                   imgPtr->nbThumbColumns(),
                   imgPtr->nbThumbRows());

   g_object_unref(pixbuf);
   gdk_pixmap_unref(pixmap);
}

void resize(GtkWidget* /* widget */, GtkAllocation* size, gpointer /* ptr */)
{
   windowWidth = size->width;
   windowHeight = size->height;
   paintImage();
}

bool readImage()
{
   if (imageOrig.read(imageName))
   {
//      std::cout << "outerRadius " << outerRadius << std::endl
//        << "innerRadius " << innerRadius << std::endl
//        << "centreX " << centreX << std::endl
//        << "centreY " << centreY << std::endl;
     if (unWrap)
     {
       if (centreX == 0)
       {
         centreX = imageOrig.nbColumns() / 2;
       }
       if (centreY == 0)
       {
         centreY = imageOrig.nbRows() / 2;
       }
       if (outerRadius == 0)
       {
         outerRadius = fmin(imageOrig.nbColumns(), imageOrig.nbRows()) / 2 - 1;
         outerRadius = fmin(outerRadius, centreX);
         outerRadius = fmin(outerRadius, imageOrig.nbColumns() - centreX);
         outerRadius = fmin(outerRadius, centreY);
         outerRadius = fmin(outerRadius, imageOrig.nbRows() - centreY);
       }
       if (innerRadius == 0)
       {
         innerRadius = outerRadius * 0.4;
       }
       else if (innerRadius >= outerRadius)
       {
         std::cerr << "Inner radius needs to be smaller than the outer radius."
         << "Changed...\n";
         innerRadius = outerRadius - 5;
       }
//      std::cout << "outerRadius " << outerRadius << std::endl
//        << "innerRadius " << innerRadius << std::endl
//        << "centreX " << centreX << std::endl
//        << "centreY " << centreY << std::endl;
       imgMap.createUnwrappingMapping(imageOrig.nbColumns(), imageOrig.nbRows(),
                                      outerRadius, innerRadius,
                                      centreX, centreY,
                                      angularRes,
                                      power, direction);
       imageOrig.apply(imgMap, &imageTransf);
       imgPtr = &imageTransf;
     }
     else
     {
       imgPtr = &imageOrig;
     }
     windowWidth = static_cast<int>(imgPtr->nbColumns() * scale + 0.5);
     windowHeight = static_cast<int>(imgPtr->nbRows() * scale + 0.5);
     return(true);
   }
   else
   {
      return(false);
   }
}

gboolean keyPressCB(GtkWidget *widget, GdkEventKey *event, gpointer user_data)
{
   switch (event->keyval)
   {
      case GDK_r:
         if (readImage())
         {
            gtk_window_resize(GTK_WINDOW(windowImage), windowWidth, windowHeight);
            paintImage();
            if (thumbnail)
            {
               gtk_window_resize(GTK_WINDOW(windowThumb),
                                 imgPtr->nbThumbColumns(),
                                 imgPtr->nbThumbRows());
               paintThumb();
            }
         }
         else
         {
            exit(1);
         }
         break;
      default:
         return(FALSE);
   }
   return(TRUE);
}

int main(int argc, char** argv)
{
   commandLine(argc, argv);

   if (!readImage())
      exit(1);

   if (thumbnail)
   {
      if ((thumbWidth != 0) || (thumbHeight != 0))
      {
         if (thumbWidth == 0)
            thumbWidth = imgPtr->nbColumns();
         if (thumbHeight == 0)
            thumbHeight = imgPtr->nbRows();
         imgPtr->setThumbnailSize(thumbWidth, thumbHeight);
      }
      else
         imgPtr->setThumbnailSize(imgPtr->nbColumns() / ratio,
                                  imgPtr->nbRows() / ratio);
      imgPtr->updateThumbnail(blurWidth, blurHeight);
   }

   gtk_init(&argc, &argv);

   windowImage = gtk_window_new(GTK_WINDOW_TOPLEVEL);
   char* windowTitleImage;
   windowTitleImage = new char[imageName.length()+1];
   sprintf(windowTitleImage, "%s", imageName.c_str());
   gtk_window_set_title(GTK_WINDOW(windowImage),
                        windowTitleImage);
   gtk_container_set_border_width(GTK_CONTAINER(windowImage), 2);
   impixmapImage = GTK_WIDGET(gtk_drawing_area_new());
   gtk_widget_set_usize(GTK_WIDGET(impixmapImage),
                        windowWidth, windowHeight);
   gtk_container_add(GTK_CONTAINER(windowImage), GTK_WIDGET(impixmapImage));
   gtk_widget_show_all(GTK_WIDGET(windowImage));

   paintImage();

   g_signal_connect(GTK_OBJECT(windowImage), "delete-event",
                    GTK_SIGNAL_FUNC(gtk_main_quit), NULL);
   g_signal_connect(GTK_OBJECT(windowImage), "expose_event",
                    G_CALLBACK(paintImage), NULL);
   gtk_signal_connect(GTK_OBJECT(impixmapImage), "size-allocate",
                      GTK_SIGNAL_FUNC(resize), NULL);
   g_signal_connect(G_OBJECT(windowImage), "key_press_event",
                    G_CALLBACK(keyPressCB), NULL);

   if (thumbnail)
   {
      windowThumb = gtk_window_new(GTK_WINDOW_TOPLEVEL);
      char* windowTitleThumb;
      windowTitleThumb = new char[imageName.length() + 12];
      sprintf(windowTitleThumb, "%s: thumbnail", imageName.c_str());
      gtk_window_set_title(GTK_WINDOW(windowThumb),
                           windowTitleThumb);
      gtk_container_set_border_width(GTK_CONTAINER(windowThumb), 2);
      impixmapThumb = GTK_WIDGET(gtk_drawing_area_new());
      gtk_widget_set_usize(GTK_WIDGET(impixmapThumb),
                        imgPtr->nbThumbColumns(), imgPtr->nbThumbRows());
      gtk_container_add(GTK_CONTAINER(windowThumb), GTK_WIDGET(impixmapThumb));
      gtk_widget_show_all(GTK_WIDGET(windowThumb));

      paintThumb();

      gtk_signal_connect(GTK_OBJECT(windowThumb), "delete-event",
                         GTK_SIGNAL_FUNC(gtk_main_quit), NULL);
      g_signal_connect(GTK_OBJECT(windowThumb), "expose_event",
                       G_CALLBACK(paintThumb), NULL);
   }

   gtk_main();

   return(0);
}
