/*
 * imalib
 * 
 * Copyright 2017 Aberystwyth University
 * 
 * This library and program are free software: you can redistribute them
 * and/or modify them under the terms of the GNU General Public License as 
 * published by the Free Software Foundation, either version 3 of the License,
 * or (at your option) any later version.
 * 
 * The program and library are distributed in the hope that they will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*!
 * \file imageGrabber.cpp
 * \brief A class to wrap the grabbing of images.
 *
 * The image is just a colourful image generated by the class.  As such, the
 * class should be inherited by a real frame grabber class.
 *
 * \author Fred Labrosse <ffl@aber.ac.uk>
 * \copyright 2017 Aberystwyth University
 */

#include "imalib/image.h"
#include "imalib/imageGrabber.h"
#include "imalib/imageMapping.h"
#include "imalib/imageRGB.h"

#include <iostream>
#include <fstream>

#include <boost/program_options.hpp>
namespace po = boost::program_options;

namespace imalib
{

/*!
 * \brief Default constructor.
 */
imageGrabber::imageGrabber()
   : connected (true),
   imageWidth(0), imageHeight(0),
   _unwrap(false), _unwrappingPower(1.0),
   _blur(false), _blurWidth(0), _blurHeight(0), _upsideDown(false),
   _flip(false),
   _direction(1),
   imgMap(NULL),
   capture(false)
{
   imgMap = new imageMapping;
   updateMapping();                  // So that we have a mapping specified.
}

/*!
 * \brief Constructor.
 * \param[in] imgWidth
 * \param[in] imgHeight The width and height of the images to grab.
 * \param[in] upsideDown Set to true if the camera is upside-down.
 * \param[in] flip Set to tru to horizontally flip the images.
 */
imageGrabber::imageGrabber(unsigned short imgWidth,
                           unsigned short imgHeight,
                           bool upsideDown,
                           bool flip)
   : connected (true),
   imageWidth(imgWidth), imageHeight(imgHeight),
   _unwrap(false), _unwrappingPower(1.0),
   _blur(false), _blurWidth(0), _blurHeight(0), _upsideDown(upsideDown),
   _flip(flip),
   _direction(1),
   imgMap(NULL),
   capture(false)
{
   if ((imgWidth == 0) || (imgHeight == 0))
   {
      std::cout
         << "--> WARNING!\n"
         << "--> WARNING!\n"
         << "--> WARNING!\n"
         << "--> Is that really the size you wanted for your images: ("
         << imgWidth << "x" << imgHeight << ")?!\n";
   }

   imgMap = new imageMapping;
   updateMapping();                  // So that we have a mapping specified.
}

/*!
\brief destructor.
*/
imageGrabber::~imageGrabber()
{
}

/*!
 * \brief Outputs some information about the source of the images in the
 * given stream.
 *
 * You should really override this!
 *
 * \param[in,out] aStream The stream to which the info is added.
 */
void imageGrabber::printInfoAboutSource(std::ostringstream* aStream)
{
   *aStream << "Colourful images generated by a fake frame-grabber.\n";
   printBasicInfoAboutSource(aStream);
}

/*!
 * \brief Output the basic info about the grabber.
 *
 * This should be called by all sub-classes in their
 * printInfoAboutSource() function.
 *
 * \param[in,out] aStream The stream to which the info is added.
 */
void imageGrabber::printBasicInfoAboutSource(std::ostringstream* aStream)
{
   *aStream << "\timage size: " << imgMap->sourceWidth()
      << "x" << imgMap->sourceHeight() << std::endl;
   if (_unwrap)
   {
      *aStream << "Images are unwrapped:\n"
         << "\tsize:               " << imgMap->targetWidth() << "x"
         << imgMap->targetHeight() << std::endl
         << "\touter radius:       " << _outerRadius << std::endl
         << "\tinner radius:       " << _innerRadius << std::endl
         << "\tcentre:             " << _centreX << "," << _centreY << std::endl
         << "\tangular resolution: " << _angularRes << std::endl
         << "\tpower:              " << _unwrappingPower << std::endl
         << "\tdirection:          " << _direction << std::endl;
   }
   if (_blur)
   {
      *aStream << "Images are blurred:\n"
               << "\tblur size: " << _blurWidth << "x" << _blurHeight
               << std::endl;
   }
   if (_upsideDown)
   {
      *aStream << "Camera was upside down\n";
   }
   if (_flip)
   {
      *aStream << "Images are flipped\n";
   }
}


/*!
 * \brief Unwrap or not.
 *
 * \param[in] yes Whether to unwrap or not.
 * \param[in] outerRadius The outer radius of the doughnut.
 * \param[in] innerRadius The inner radius of the doughnut.
 * \param[in] centreX
 * \param[in] centreY The coordinates of the centre of the doughnut
 * \param[in] angularRes The angular resolution (number of pixels per
 * degree of angle).
 * \param[in] power The power of the function to convert values along a
 * radial axis in the omni-directional image to a column in the panoramic
 * image (1.0 is linear, 2.0 is squared, etc).
 * \param[in] direction 1 is turn counter-clockwise (positive
 * trigonometric), -1 is turn clockwise (negative trigonometric) from
 * left to right in the unwrapped image.
 * \sa createUnwrappingMapping(), imageMapping::createUnwrappingMapping().
 */
void imageGrabber::doUnwrap(bool yes,
                            unsigned short outerRadius,
                            unsigned short innerRadius,
                            unsigned short centreX,
                            unsigned short centreY,
                            unsigned short angularRes, double power,
                            short direction)
{
   _unwrap = yes;
   _unwrappingPower = power;
   _angularRes = angularRes;
   _outerRadius = outerRadius;
   _innerRadius = innerRadius;
   _centreX = centreX,
   _centreY = centreY;
   _direction = direction;

   updateMapping();
}

/*!
 * \brief Unwrap or not.
 *
 * This function actually reads the configuration file, despite the fact that
 * the imageMapping class can do it as well.  This is because we need the
 * info here.
 *
 * \param[in] yes Whether to unwrap or not.
 * \param[in] fileName Name of the configuration file of the unwrapping.
 */
void imageGrabber::doUnwrap(bool yes,
                            const std::string& fileName)
{
   _innerRadius = 40;
   _angularRes = 1;
   _unwrappingPower = 1.0;
   _direction = -1;
   po::options_description optionsDesc("unwrappingMapping options");
   optionsDesc.add_options()
      ("outerRadius", po::value<unsigned short>(&_outerRadius), "")
      ("innerRadius", po::value<unsigned short>(&_innerRadius), "")
      ("centreX", po::value<unsigned short>(&_centreX), "")
      ("centreY", po::value<unsigned short>(&_centreY), "")
      ("pixelsPerDegree",
       po::value<unsigned short>(&_angularRes)->default_value(_angularRes),
       "")
      ("power",
       po::value<double>(&_unwrappingPower)->default_value(_unwrappingPower),
       "")
      ("direction",
       po::value<short>(&_direction)->default_value(_direction), "")
      ;

   po::variables_map vm;
   std::ifstream configFile(fileName.c_str());
   if (configFile.is_open())
   {
      po::store(po::parse_config_file(configFile, optionsDesc), vm);
      po::notify(vm);
      configFile.close();
   }
   else
   {
      std::cerr << "imageGrabber::doUnwrap():\n"
         << "WARNING: could not open configuration file " << fileName
         << "!\n"
         << "\tUsing defaults.\n";
   }

   _unwrap = yes;

   updateMapping();
}


/*!
 * \brief Blur or not.
 * \param yes (IN) Whether to blur or not.
 * \param blurWidth
 * \param blurHeight (IN) The size of the window used to blur.
 * \sa scaleBlur().
 */
void imageGrabber::doBlur(bool yes,
                          unsigned short blurWidth,
                          unsigned short blurHeight)
{
   _blur = yes;
   _blurWidth = blurWidth;
   _blurHeight = blurHeight;
}

/*!
 * \brief Capture the image from a grabber.
 *
 * This is there because the capture can take a while.  We thus first
 * capture and then use (with getImage()).
 */
void imageGrabber::captureImage()
{
   capture = false;          // Since we are actually not capturing here.
}

/*!
 * \brief Get an image from whatever image source.
 *
 * The default implementation generates a moving image.  You may want to
 * override this.
 *
 * The images are resized here.
 *
 * \param[out] theImage The image in which to save the grabbed image, if
 * not NULL.
 * \param[out] rawImage The image in which to save the raw image (before any
 * transformation is applied), if not NULL.
 */
void imageGrabber::getImage(imageRGB* theImage, imageRGB* rawImage)
{
   static imageRGB tmpImg;
   bool deleteImage = false;

   imageRGB* grabbedImage = NULL;
   if (rawImage != NULL)
   {
      grabbedImage = rawImage;
   }
   else
   {
      deleteImage = true;
      grabbedImage = new imageRGB;
   }
   grabbedImage->setSize(imgMap->sourceWidth(),
                         imgMap->sourceHeight());

   tmpImg.setSize(imgMap->targetWidth(),
                  imgMap->targetHeight());

   if (!capture)
   {
      unsigned long nbPix
         = imgMap->sourceWidth() * imgMap->sourceHeight();

      unsigned char* data = grabbedImage->getActualPixel(0, 0);

      static unsigned short counter = 0;
      for (unsigned long i = 0; i < nbPix * 3; i += 3)
      {
//          data[((i / 3 * imageWidth) % mod) * 3] =
         data[i] =
            (counter
             + static_cast<unsigned char>(
                       static_cast<double>(i)
                       / static_cast<double>(imgMap->sourceWidth()
                                             * imgMap->sourceHeight()
                                             * 3)
                                             * 255.0)) % 255;
         data[i+1] = (data[i] + 100) % 255;
         data[i+2] = (data[i] + 200) % 255;
            //       data[i+2] = 0;
      }
      counter++;
   }

   if (theImage)
   {
      theImage->setSize(imgMap->targetWidth(),
                        imgMap->targetHeight());

      // Apply the mapping (which might be a noop).
      imageRGB* imgPtr;
      if (_blur)
      {
         imgPtr = &tmpImg;
      }
      else
      {
         imgPtr = theImage;
      }

      std::vector<imageMapping::coordsPair>::iterator iter;
      for (iter = imgMap->map.begin();
           iter < imgMap->map.end();
           ++iter)
      {
         imgPtr->setPixel(iter->tX, iter->tY,
                          grabbedImage->getActualPixel(iter->sX,
                             iter->sY));
      }

      // Possibly blur the image.
      if (_blur)
      {
         tmpImg.scaleBlur(_blurWidth, _blurHeight, theImage);
      }
   }

   capture = false;          // We have used the captured image.
   if (deleteImage)
      delete grabbedImage;
}

/*!
 * \brief Update the mapping.
 */
void imageGrabber::updateMapping()
{
   if (_upsideDown)
   {
      imgMap->createUpsideDownMapping(imageWidth, imageHeight);
   }
   else
   {
      imgMap->createNoopMapping(imageWidth, imageHeight);
   }

   if (_flip)
   {
      imageMapping flipMapping;
      flipMapping.createHorizontalFlipMapping(imageWidth, imageHeight);
      imgMap->concatenate(flipMapping);
   }

   if (_unwrap)
   {
      imageMapping unwrapMap;
      unwrapMap.createUnwrappingMapping(imageWidth, imageHeight,
                                        _outerRadius, _innerRadius,
                                        _centreX, _centreY,
                                        _angularRes, _unwrappingPower,
                                        _direction);

      imgMap->concatenate(unwrapMap);
   }
}


} // namespace imalib
