/*
 * imalib
 * 
 * Copyright 2017 Aberystwyth University
 * 
 * This library and program are free software: you can redistribute them
 * and/or modify them under the terms of the GNU General Public License as 
 * published by the Free Software Foundation, either version 3 of the License,
 * or (at your option) any later version.
 * 
 * The program and library are distributed in the hope that they will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*!
 * \file imageHSV.cpp
 * \brief Defines classes to play with HSV images.  Implementation.
 * \author Fred Labrosse <ffl@aber.ac.uk>
 * \copyright 2017 Aberystwyth University
 */

#include "imalib/imageHSV.h"

#include <math.h>

#include <limits>

namespace imalib
{
   //-----------------------------------------------------
   // Class imageHSV.
   //-----------------------------------------------------

   /*!
    * \brief The value H should be set to if it is undefined for
    * a particular colour.
    */   
   const double imageHSV::HSV_H_UNDEFINED = -1.0;

   /*!
    * \brief Convert a pixel in the RGB colour space to a pixel in the
    * HSV colour space.
    * \param rgb (IN) the pixel in RGB colour space.
    * \param hsv (OUT) The pixel in HSV colour space (must be
    * allocated beforehand).
    */         
   void imageHSV::RGBToHSV(const unsigned char* rgb, double* hsv) 
   {
      double max;
      double min;
      double r, g, b;
      r = static_cast<double>(rgb[0]) / 255.0;
      g = static_cast<double>(rgb[1]) / 255.0;
      b = static_cast<double>(rgb[2]) / 255.0;
      if ((rgb[0] >= rgb[1]) && (rgb[0] >= rgb[2]))
         max =  r;
      else if ((rgb[1] >= rgb[0]) && (rgb[1] >= rgb[2]))
         max = g;
      else
         max = b;
      if ((rgb[0] <= rgb[1]) && (rgb[0] <= rgb[2]))
         min = r;
      else if ((rgb[1] <= rgb[0]) && (rgb[1] <= rgb[2]))
         min = g;
      else
         min = b;

      double v = max;
      double s;
      if (max != 0)
         s = (max - min) / max;
      else
         s = 0.0;
      double h = HSV_H_UNDEFINED;
      if (s == 0)
         h = HSV_H_UNDEFINED;
      else
      {
         double delta = max - min;
         if (r == max)
            h = (g - b) / delta;
         else if (g == max)
            h = 2.0 + (b - r) / delta;
         else if (b == max)
            h = 4.0 + (r - g) / delta;
         h *= 60.0;
         if (h < 0)
            h += 360.0;
      }
      
      
      //setHSV(h, s, v);
      hsv[0] = h;
      hsv[1] = s;
      hsv[2] = v;
   }
   
//-----------------------------------------------------
// Class imageHS_.
//-----------------------------------------------------


   /*!
      \brief The value H should be set to if it is undefined for
      a particular colour. 
   */   
   const double imageHS_::HSV_H_UNDEFINED = -1.0;

   /*!
     \brief Convert a pixel in the RGB colour space to a pixel in the
     HSV colour space with only H and S components.
     \param rgb (IN) the pixel in RGB colour space.
     \param hsv (OUT) The pixel in HSV colour space (must be
     allocated beforehand).
   */         
   void imageHS_::RGBToHS_(const unsigned char* rgb, double* hsv) 
   {
      double max;
      double min;
      double r, g, b;
      r = static_cast<double>(rgb[0]) / 255.0;
      g = static_cast<double>(rgb[1]) / 255.0;
      b = static_cast<double>(rgb[2]) / 255.0;
      if ((rgb[0] >= rgb[1]) && (rgb[0] >= rgb[2]))
         max =  r;
      else if ((rgb[1] >= rgb[0]) && (rgb[1] >= rgb[2]))
         max = g;
      else
         max = b;
      if ((rgb[0] <= rgb[1]) && (rgb[0] <= rgb[2]))
         min = r;
      else if ((rgb[1] <= rgb[0]) && (rgb[1] <= rgb[2]))
         min = g;
      else
         min = b;

      double s;
      if (max != 0)
         s = (max - min) / max;
      else
         s = 0.0;
      double h = HSV_H_UNDEFINED;
      if (s != 0)
      {
         double delta = max - min;
         if (r == max)
            h = (g - b) / delta;
         else if (g == max)
            h = 2.0 + (b - r) / delta;
         else if (b == max)
            h = 4.0 + (r - g) / delta;
         h *= 60.0;
         if (h < 0)
            h += 360.0;
      }
      
      
      //setHS_(h, s);
      hsv[0] = h;
      hsv[1] = s;
   }

} // namespace imalib

//  LocalWords:  imageYUV param imageHSV
