/*
 * imalib
 * 
 * Copyright 2017 Aberystwyth University
 * 
 * This library and program are free software: you can redistribute them
 * and/or modify them under the terms of the GNU General Public License as 
 * published by the Free Software Foundation, either version 3 of the License,
 * or (at your option) any later version.
 * 
 * The program and library are distributed in the hope that they will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*!
 * \file imageLab.cpp
 * \brief Defines classes to play with CIEL*a*b* images.  Implementation.
 * \author Fred Labrosse <ffl@aber.ac.uk>
 * \copyright 2017 Aberystwyth University
 */

#include "imalib/imageLab.h"

#include <math.h>

#include <limits>

namespace imalib
{
//-----------------------------------------------------
// Class imageLab.
//-----------------------------------------------------

/*!
 * \brief Convert a pixel in the
 * CIE&nbsp;L<sup>*</sup>a<sup>*</sup>b<sup>*</sup> colour space to a
 * pixel in the RGB colour space.
 *
 * \param[in] lab The pixel in
 * CIE&nbsp;L<sup>*</sup>a<sup>*</sup>b<sup>*</sup> colour space.
 * \param[out] rgb The pixel in RGB colour space (must be allocated
 * beforehand).
 */
void imageLab::LabToRGB(const double* lab, unsigned char* rgb)
{
   static const double Xn = 2.7690 * 255.0 + 1.7518 * 255.0 + 1.1300 * 255.0;
   static const double Yn = 1.0000 * 255.0 + 4.5907 * 255.0 + 0.0601 * 255.0;
   static const double Zn =                + 0.0565 * 255.0 + 5.5943 * 255.0;
   double X;
   double Y;
   double Z;
   double C, C2, C3;
   double L, L2, L3;
   double a, a2, a3;
   double b, b2, b3;
   double R;
   double G;
   double B;

   L = lab[0];
   L2 = L * L;
   L3 = L2 * L;
   a = lab[1];
   a2 = a * a;
   a3 = a2 * a;
   b = lab[2];
   b2 = b * b;
   b3 = b2 * b;

   Y = 1.0/1560896.0 * Yn * L3 + 3.0/97556.0 * Yn * L2
      + 12.0/24389.0 * Yn * L + 64.0/24389.0 * Yn;
   if (Y < 0)
   {
      // Y must be positive.  If it's negative, then it's because it is
      // something like -1e-14 because of round-offs in the previous
      // equations. (Has to be done because of the pow following.)
      Y = 0;
   }
//    else if (Y > Yn)
//    {
//       // Y must be <= Yn.  If it's not, then it's because it is
//       // something like Yn + 1e-14 because of round-offs in the previous
//       // equations.
//       Y = Yn;
//    }

   C = cbrt(Y/ Yn);
   C2 = C * C;
   C3 = C2 * C;
   X = 1.0/125000000.0 * Xn * a3 + 3.0/250000.0 * Xn * a2 * C
      +3.0/500.0 * Xn * a * C2 + Xn * C3;
//    if (X < 0)
//    {
//       // X must be positive.  If it's negative, then it's because it is
//       // something like -1e-14 because of round-offs in the previous
//       // equations.
//       X = 0;
//    }
//    else if (X > Xn)
//    {
//       // X must be <= Xn.  If it's not, then it's because it is
//       // something like Xn + 1e-14 because of round-offs in the previous
//       // equations.
//       X = Xn;
//    }
   Z = -1.0/8000000.0 * Zn * b3 + 3.0/40000.0 * Zn * b2 * C
      - 3.0/200.0 * Zn * b * C2 + Zn * C3;
//    if (Z < 0)
//    {
//       // Z must be positive.  If it's negative, then it's because it is
//       // something like -1e-14 because of round-offs in the previous
//       // equations.
//       Z = 0;
//    }
//    else if (Z > Zn)
//    {
//       // Z must be <= Zn.  If it's not, then it's because it is
//       // something like Zn + 1e-14 because of round-offs in the previous
//       // equations.
//       Z = Zn;
//    }

   R =  0.4184383528    * X - 0.1586557990   * Y - 0.08281646054 * Z;
   G = -0.09116119246   * X + 0.2524253419   * Y + 0.01570194384 * Z;
   B =  0.0009206884461 * X - 0.002549386307 * Y + 0.1785947912  * Z;
   if (R < 0)
   {
         // If R is not positive, it's because of round-offs in the
         // previous equations.
      R = 0;
   }
   else if (R > 255)
   {
         // If R is not <= 255, it's because of round-offs in the previous
         // equations.
      R = 255;
   }
   if (G < 0)
   {
         // If G is not positive, it's because of round-offs in the
         // previous equations.
      G = 0;
   }
   else if (G > 255)
   {
         // If G is not <= 255, it's because of round-offs in the previous
         // equations.
      G = 255;
   }
   if (B < 0)
   {
         // If B is not positive, it's because of round-offs in the
         // previous equations.
      B = 0;
   }
   else if (B > 255)
   {
         // If B is not <= 255, it's because of round-offs in the previous
         // equations.
      B = 255;
   }

   rgb[0] = static_cast<unsigned char>(R + 0.5);
   rgb[1] = static_cast<unsigned char>(G + 0.5);
   rgb[2] = static_cast<unsigned char>(B + 0.5);
}

/*!
 * \brief Convert a pixel in the RGB colour space to a pixel in the
 * CIE&nbsp;L<sup>*</sup>a<sup>*</sup>b<sup>*</sup> colour space.
 * \param[in] rgb The pixel in RGB colour space.
 * \param[out] lab The pixel in
 * CIE&nbsp;L<sup>*</sup>a<sup>*</sup>b<sup>*</sup> colour space (must
 * be allocated beforehand).
 */
void imageLab::RGBToLab(const unsigned char* rgb, double* lab)
{
   lab[0] = static_cast<double>(rgb[0]);
   lab[1] = static_cast<double>(rgb[1]);
   lab[2] = static_cast<double>(rgb[2]);
   RGBToLab(lab, lab);
}

/*!
 * \brief Convert a pixel in the RGB colour space to a pixel in the
 * CIE&nbsp;L<sup>*</sup>a<sup>*</sup>b<sup>*</sup> colour space.
 *
 * Note that the RGB values are in [0, 255].
 *
 * \param[in] rgb The pixel in RGB colour space.
 * \param[out] lab The pixel in
 * CIE&nbsp;L<sup>*</sup>a<sup>*</sup>b<sup>*</sup> colour space (must
 * be allocated beforehand).
 */
void imageLab::RGBToLab(const double* rgb, double* lab)
{
   static const double Xn = 2.7690 * 255.0 + 1.7518 * 255.0 + 1.1300 * 255.0;
   static const double Yn = 1.0000 * 255.0 + 4.5907 * 255.0 + 0.0601 * 255.0;
   static const double Zn =                + 0.0565 * 255.0 + 5.5943 * 255.0;
   double X;
   double Y;
   double Z;

   X = 2.7690 * rgb[0]
      + 1.7518 * rgb[1]
      + 1.1300 * rgb[2];
   Y = 1.0000 * rgb[0]
      + 4.5907 * rgb[1]
      + 0.0601 * rgb[2];
   Z = 0.0565 * rgb[1]
      + 5.5943 * rgb[2];

   X = cbrt(X / Xn);
   Y = cbrt(Y / Yn);
   Z = cbrt(Z / Zn);

   lab[0] = 116.0 * Y - 16.0;
   lab[1] = 500.0 * (X - Y);
   lab[2] = 200.0 * (Y - Z);
}


/*!
 * \brief Copy the a<sup>*</sup> and b<sup>*</sup> components in an
 * RGB image using the R and G components only (B is set to 124).
 *
 * The a<sup>*</sup> and b<sup>*</sup> components are normalised to
 * the interval [0; 255] using the minimum and maximum values
 * obtainable for them by converting the RGB values (255, 0, 0) (red),
 * (0, 255, 0) (green), (255, 255, 0) (yellow) and (0, 0, 255) (blue)
 * which roughly correspond to the axes defined by the components
 * a<sup>*</sup> and b<sup>*</sup>.  This range can be altered.
 *
 * \param img (OUT) The image where to save the a<sup>*</sup> and
 * b<sup>*</sup> components (its size is set here).
 * \param coeff (IN) Divide the "theoretical" range of a<sup>*</sup>
 * and b<sup>*</sup> by this coefficient (to perform some kind of
 * contrast modification --- enhancement for coeff > 1.0, reduction
 * for coeff < 1.0).
 */
void imageLab::copyAB(imageRGB* img, double coeff) const
{
   unsigned char RGBCol[3];
   double LabCol[3];

   // a* max (red)
   RGBCol[0] = 255;
   RGBCol[1] = 0;
   RGBCol[2] = 0;
   RGBToLab(RGBCol, LabCol);
   const double aMax = LabCol[1];
   // a* min (green)
   RGBCol[0] = 0;
   RGBCol[1] = 255;
   RGBCol[2] = 0;
   RGBToLab(RGBCol, LabCol);
   const double aMin = LabCol[1];
   // b* max (yellow)
   RGBCol[0] = 255;
   RGBCol[1] = 255;
   RGBCol[2] = 0;
   RGBToLab(RGBCol, LabCol);
   const double bMax = LabCol[2];
   // b* min (blue)
   RGBCol[0] = 0;
   RGBCol[1] = 0;
   RGBCol[2] = 255;
   RGBToLab(RGBCol, LabCol);
   const double bMin = LabCol[2];

   const double aFactor = 1.0 / ((aMax - aMin) / coeff) * 255.0;
   const double bFactor = 1.0 / ((bMax - bMin) / coeff) * 255.0;

   img->setSize(nbColumns(), nbRows());
   unsigned char* RGBPixel = img->getActualPixel(0, 0);
   const double* LabPixel = getActualPixel(0, 0);

   // Copy the a* and b* (scaled) as the R and G, B being set to 124.
   for (long ind = 0; ind < nbColumns() * nbRows(); ++ind)
   {
      RGBPixel[0] = static_cast<unsigned char>((LabPixel[1]
                                                - aMin) * aFactor);
      RGBPixel[1] = static_cast<unsigned char>((LabPixel[2]
                                                - bMin) * bFactor);
      RGBPixel[2] = 124;
      RGBPixel += 3;
      LabPixel += 3;
   }
}

//-----------------------------------------------------
// Class image_ab.
//-----------------------------------------------------

/*!
 * \brief Convert a pixel in the
 * CIE&nbsp;L<sup>*</sup>a<sup>*</sup>b<sup>*</sup> colour space with only
 * the a<sup>*</sup> and b<sup>*</sup> components to a pixel in the RGB
 * colour space.
 *
 * The value for L<sup>*</sup> is taken to be the default L;
 *
 * \param[in] ab The pixel in
 * CIE&nbsp;_a<sup>*</sup>b<sup>*</sup> colour space.
 * \param[out] rgb The pixel in RGB colour space (must be allocated
 * beforehand).
 * \sa getdefaultL(), setDefaultL();
 */
void image_ab::_abToRGB(const double* ab, unsigned char* rgb) const
{
   static const double Xn = 2.7690 * 255.0 + 1.7518 * 255.0 + 1.1300 * 255.0;
   static const double Yn = 1.0000 * 255.0 + 4.5907 * 255.0 + 0.0601 * 255.0;
   static const double Zn =                + 0.0565 * 255.0 + 5.5943 * 255.0;
   double X;
   double Y;
   double Z;
   double C, C2, C3;
   double L, L2, L3;
   double a, a2, a3;
   double b, b2, b3;
   double R;
   double G;
   double B;

   L = defaultL;
   L2 = L * L;
   L3 = L2 * L;
   a = ab[0];
   a2 = a * a;
   a3 = a2 * a;
   b = ab[1];
   b2 = b * b;
   b3 = b2 * b;

   Y = 1.0/1560896.0 * Yn * L3 + 3.0/97556.0 * Yn * L2
      + 12.0/24389.0 * Yn * L + 64.0/24389.0 * Yn;
   if (Y < 0)
   {
      // Y must be positive.  If it's negative, then it's because it is
      // something like -1e-14 because of round-offs in the previous
      // equations. (Has to be done because of the pow following.)
      Y = 0;
   }
//    else if (Y > Yn)
//    {
//       // Y must be <= Yn.  If it's not, then it's because it is
//       // something like Yn + 1e-14 because of round-offs in the previous
//       // equations.
//       Y = Yn;
//    }

   C = cbrt(Y/ Yn);
   C2 = C * C;
   C3 = C2 * C;
   X = 1.0/125000000.0 * Xn * a3 + 3.0/250000.0 * Xn * a2 * C
      +3.0/500.0 * Xn * a * C2 + Xn * C3;
//    if (X < 0)
//    {
//       // X must be positive.  If it's negative, then it's because it is
//       // something like -1e-14 because of round-offs in the previous
//       // equations.
//       X = 0;
//    }
//    else if (X > Xn)
//    {
//       // X must be <= Xn.  If it's not, then it's because it is
//       // something like Xn + 1e-14 because of round-offs in the previous
//       // equations.
//       X = Xn;
//    }
   Z = -1.0/8000000.0 * Zn * b3 + 3.0/40000.0 * Zn * b2 * C
      - 3.0/200.0 * Zn * b * C2 + Zn * C3;
//    if (Z < 0)
//    {
//       // Z must be positive.  If it's negative, then it's because it is
//       // something like -1e-14 because of round-offs in the previous
//       // equations.
//       Z = 0;
//    }
//    else if (Z > Zn)
//    {
//       // Z must be <= Zn.  If it's not, then it's because it is
//       // something like Zn + 1e-14 because of round-offs in the previous
//       // equations.
//       Z = Zn;
//    }

   R =  0.4184383528    * X - 0.1586557990   * Y - 0.08281646054 * Z;
   G = -0.09116119246   * X + 0.2524253419   * Y + 0.01570194384 * Z;
   B =  0.0009206884461 * X - 0.002549386307 * Y + 0.1785947912  * Z;
   if (R < 0)
   {
         // If R is not positive, it's because of round-offs in the
         // previous equations.
      R = 0;
   }
   else if (R > 255)
   {
         // If R is not <= 255, it's because of round-offs in the previous
         // equations.
      R = 255;
   }
   if (G < 0)
   {
         // If G is not positive, it's because of round-offs in the
         // previous equations.
      G = 0;
   }
   else if (G > 255)
   {
         // If G is not <= 255, it's because of round-offs in the previous
         // equations.
      G = 255;
   }
   if (B < 0)
   {
         // If B is not positive, it's because of round-offs in the
         // previous equations.
      B = 0;
   }
   else if (B > 255)
   {
         // If B is not <= 255, it's because of round-offs in the previous
         // equations.
      B = 255;
   }

   rgb[0] = static_cast<unsigned char>(R + 0.5);
   rgb[1] = static_cast<unsigned char>(G + 0.5);
   rgb[2] = static_cast<unsigned char>(B + 0.5);
}

/*!
 * \brief Convert a pixel in the RGB colour space to a pixel in the
 * CIE&nbsp;L<sup>*</sup>a<sup>*</sup>b<sup>*</sup> colour space with only
 * the a<sup>*</sup> and b<sup>*</sup> components.
 * \param[out] rgb The pixel in RGB colour space.
 * \param[in] ab The pixel in
 * CIE&nbsp;_a<sup>*</sup>b<sup>*</sup> colour space (must
 * be allocated beforehand).
 */
void image_ab::RGBTo_ab(const unsigned char* rgb, double* ab) const
{
   static const double Xn = 2.7690 * 255.0 + 1.7518 * 255.0 + 1.1300 * 255.0;
   static const double Yn = 1.0000 * 255.0 + 4.5907 * 255.0 + 0.0601 * 255.0;
   static const double Zn =                + 0.0565 * 255.0 + 5.5943 * 255.0;
   double X;
   double Y;
   double Z;

   X = 2.7690 * static_cast<double>(rgb[0])
      + 1.7518 * static_cast<double>(rgb[1])
      + 1.1300 * static_cast<double>(rgb[2]);
   Y = 1.0000 * static_cast<double>(rgb[0])
      + 4.5907 * static_cast<double>(rgb[1])
      + 0.0601 * static_cast<double>(rgb[2]);
   Z = 0.0565 * static_cast<double>(rgb[1])
      + 5.5943 * static_cast<double>(rgb[2]);

   X = cbrt(X / Xn);
   Y = cbrt(Y / Yn);
   Z = cbrt(Z / Zn);

   // lab[0] = 116.0 * Y - 16.0; L is not set.
   ab[0] = 500.0 * (X - Y);
   ab[1] = 200.0 * (Y - Z);
}

} // namespace imalib
