/*
 * imalib
 * 
 * Copyright 2017 Aberystwyth University
 * 
 * This library and program are free software: you can redistribute them
 * and/or modify them under the terms of the GNU General Public License as 
 * published by the Free Software Foundation, either version 3 of the License,
 * or (at your option) any later version.
 * 
 * The program and library are distributed in the hope that they will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*!
 * \file imageMapping.cpp
 * \brief A class to represent the transformation of an image into another
 * (implementation).
 *
 * \author Fred Labrosse
 * \copyright 2017 Aberystwyth University
 */

#include "imalib/imageMapping.h"
#include "imalib/utilities.h"


#include <algorithm>
#include <cassert>
#include <fstream>
#include <functional>
#include <iostream>

#include <math.h>

#include <boost/program_options.hpp>
namespace po = boost::program_options;

namespace imalib
{
//-----------------------
// Class imageMapping
//-----------------------

/*!
 * \brief Create the imageMapping to unwrap an omni-directional into a
 * panoramic image.
 *
 * So far, some of the parameters of the mapping are hard-coded:
 * - omni-directional image must be square,
 * - size of the target (panoramic image): (360 * angularRes) x
 * imgSize / 2 - innerRadius,
 *
 * This mapping is one-to-many.
 *
 * \param[in] imgSize The size of the (square) image this mapping
 * will be applied to.
 * \param[in] innerRadius The inner radius of the considered doughnut as a
 * percentage of the image size (the outer radius is 95% of the image
 * size).  When negative, 40% of the image size is used.
 * \param[in] angularRes The angular resolution (number of pixels per
 * degree of angle).
 * \param[in] power The power of the function to convert values along a
 * radial axis in the omni-directional image to a column in the panoramic
 * image (1.0 is linear, 2.0 is squared, etc).
 * \param[in] direction 1 is turn counter-clockwise (positive
 * trigonometric), -1 is turn clockwise (negative trigonometric) from left
 * to right in the unwrapped image.
 */
void imageMapping::createUnwrappingMapping(unsigned short imgSize,
                                           short innerRadius,
                                           unsigned short angularRes,
                                           double power,
                                           short direction)
{
   clear();

   double convFact = static_cast<double>(imgSize) / 200.0;
                                // The hard coded sizes work for 200x200
                                // images.  They are thus multiplied by
                                // convFact to get the desired ones.

   double radius = 99.49 * convFact; // Radius of the useful area
   double xCentre = 99.5 * convFact; // Centre of the image.
   double yCentre = 99.5 * convFact;
   unsigned short xPos, yPos;

   double robotRadius = 40.0 * convFact;
                                // Radius of the robot in the centre of the
                                // image.
   if (innerRadius >= 0)
      robotRadius = static_cast<double>(innerRadius) * convFact;

   double maxRadius = 95.0 * convFact;
                                // Maximum radius of the useful area in the
                                // omni-directional image.

   sourceWidth_ = static_cast<unsigned short>(200.0 * convFact + 0.5);
   sourceHeight_ = sourceWidth_;  // Square images.
   targetWidth_ = 360 * angularRes;
   targetHeight_ = static_cast<unsigned short>(maxRadius - robotRadius + 0.5);

   sourceMinCol_ = 0;
   sourceMaxCol_ = sourceWidth_ - 1;
   sourceMinRow_ = 0;
   sourceMaxRow_ = sourceHeight_ - 1;
   targetMinCol_ = 0;
   targetMaxCol_ = targetWidth_ - 1;
   targetMinRow_ = 0;
   targetMaxRow_ = targetHeight_ - 1;

   double radiusStep = 1.0 / static_cast<double>(targetHeight_);
                                // We want targetHeight values.
   double rad;
   double degToRad = M_PI * 2.0 / static_cast<double>(targetWidth_);

   double ind = 1.0;
   for (unsigned short rows = 0; rows < targetHeight_; rows++)
   {
      radius = robotRadius + ((1.0 - pow((1.0 - ind), power))
                              * (maxRadius - robotRadius));
      for (unsigned short cols = 0; cols < targetWidth_; cols++)
      {
         rad = static_cast<double>(- cols * degToRad * direction);
         xPos = static_cast<unsigned short>(xCentre
                                            + (radius * cos(rad)) + 0.5);
         if (xPos >= sourceWidth_)
            xPos = sourceWidth_ - 1;
         yPos = static_cast<unsigned short>(yCentre
                                            + (radius * sin(rad)) + 0.5);
         if (yPos >= sourceHeight_)
            yPos = sourceHeight_ - 1;
         map.push_back(imageMapping::coordsPair(xPos, yPos, cols, rows));
      }
      ind -= radiusStep;
   }
}

/*!
 * \brief Create the imageMapping to unwrap an omni-directional into a
 * panoramic image.
 *
 * The position of the doughnut is verified only if not compiled with NDBUG.
 *
 * This mapping is one-to-many.
 *
 * \param[in] imgWidth
 * \param[in] imgHeight The size of the images this mapping
 * will be applied to.
 * \param[in] outerRadius The outer radius of the doughnut.
 * \param[in] innerRadius The inner radius of the doughnut.
 * \param[in] centreX
 * \param[in] centreY The coordinates of the centre of the doughnut
 * \param[in] angularRes The angular resolution (number of pixels per
 * degree of angle).
 * \param[in] power The power of the function to convert values along a
 * radial axis in the omni-directional image to a column in the panoramic
 * image (1.0 is linear, 2.0 is squared, etc).
 * \param[in] direction 1 is turn counter-clockwise (positive
 * trigonometric), -1 is turn clockwise (negative trigonometric) from left
 * to right in the unwrapped image.
 */
void imageMapping::createUnwrappingMapping(unsigned short imgWidth,
                                           unsigned short imgHeight,
                                           unsigned short outerRadius,
                                           unsigned short innerRadius,
                                           unsigned short centreX,
                                           unsigned short centreY,
                                           unsigned short angularRes,
                                           double power,
                                           short direction)
{
  assert((static_cast<long>(centreX) - static_cast<long>(outerRadius) >= 0)
         && (centreX + outerRadius < imgWidth)
         && (static_cast<long>(centreY) - static_cast<long>(outerRadius) >= 0)
         && (centreY + outerRadius < imgHeight)
         && (innerRadius < outerRadius));

  clear();

  sourceWidth_ = imgWidth;
  sourceHeight_ = imgHeight;
  targetWidth_ = 360 * angularRes;
  targetHeight_ = static_cast<unsigned short>(outerRadius - innerRadius + 0.5);

  sourceMinCol_ = 0;
  sourceMaxCol_ = sourceWidth_ - 1;
  sourceMinRow_ = 0;
  sourceMaxRow_ = sourceHeight_ - 1;
  targetMinCol_ = 0;
  targetMaxCol_ = targetWidth_ - 1;
  targetMinRow_ = 0;
  targetMaxRow_ = targetHeight_ - 1;

  double radiusStep = 1.0 / static_cast<double>(targetHeight_);
                                // We want targetHeight values.
  double rad;
  double degToRad = M_PI * 2.0 / static_cast<double>(targetWidth_);

  double ind = 1.0;
  double radius;
  unsigned short xPos, yPos;
  for (unsigned short rows = 0; rows < targetHeight_; rows++)
  {
    radius = innerRadius + ((1.0 - pow((1.0 - ind), power))
                            * (outerRadius - innerRadius));
    for (unsigned short cols = 0; cols < targetWidth_; cols++)
    {
      rad = static_cast<double>(- cols * degToRad * direction);
      xPos = static_cast<unsigned short>(centreX
                                         + (radius * cos(rad)) + 0.5);
      if (xPos >= sourceWidth_)
        xPos = sourceWidth_ - 1;
      yPos = static_cast<unsigned short>(centreY
                                         + (radius * sin(rad)) + 0.5);
      if (yPos >= sourceHeight_)
        yPos = sourceHeight_ - 1;
      map.push_back(imageMapping::coordsPair(xPos, yPos, cols, rows));
    }
    ind -= radiusStep;
  }
}

/*!
 * \brief Create the imageMapping to unwrap an omni-directional into a
 * panoramic image.
 *
 * The parameters of the mapping are taken from the configuration file
 * specified.
 *
 * \param[in] imgSize The size of the (square) image this mapping will be
 * applied to.
 * \param[in] fileName Name of the configuration file.
 */
void imageMapping::createUnwrappingMapping(unsigned short imgSize,
                                           const std::string& fileName)
{
   po::options_description optionsDesc("unwrappingMapping options");
   int inRad;
   int ppd;
   double powe;
   int direct;
   optionsDesc.add_options()
      ("innerRadius", po::value<int>(&inRad)->default_value(40), "")
      ("pixelsPerDegree", po::value<int>(&ppd)->default_value(1), "")
      ("power", po::value<double>(&powe)->default_value(1.0), "")
      ("direction", po::value<int>(&direct)->default_value(-1), "")
      ;

   po::variables_map vm;
   std::ifstream configFile(fileName.c_str());
   if (configFile.is_open())
   {
      po::store(po::parse_config_file(configFile, optionsDesc), vm);
      po::notify(vm);
      configFile.close();
   }
   else
   {
      std::cerr << "imageMapping::createUnwrappingMapping():\n"
         << "WARNING: could not open configuration file " << fileName
         << "!\n"
         << "\tUsing defaults.\n";
   }

   createUnwrappingMapping(imgSize,
                           inRad,
                           ppd,
                           powe,
                           direct);
}

/*!
 * \brief Create the imageMapping to unwrap an omni-directional into a
 * panoramic image.
 *
 * The parameters of the mapping are taken from the configuration file
 * specified.
 *
 * \param[in] imgWidth
 * \param[in] imgHeight The size of the image this mapping will be
 * applied to.
 * \param[in] fileName Name of the configuration file.
 */
void imageMapping::createUnwrappingMapping(unsigned short imgWidth,
                                           unsigned short imgHeight,
                                           const std::string& fileName)
{
   po::options_description optionsDesc("unwrappingMapping options");
   unsigned short outRad, inRad;
   unsigned short centreX, centreY;
   unsigned short ppd;
   double powe;
   int direct;
   optionsDesc.add_options()
      ("outerRadius", po::value<unsigned short>(&outRad), "")
      ("innerRadius", po::value<unsigned short>(&inRad), "")
      ("centreX", po::value<unsigned short>(&centreX), "")
      ("centreY", po::value<unsigned short>(&centreY), "")
      ("pixelsPerDegree", po::value<unsigned short>(&ppd)->default_value(1), "")
      ("power", po::value<double>(&powe)->default_value(1.0), "")
      ("direction", po::value<int>(&direct)->default_value(-1), "")
      ;

   po::variables_map vm;
   std::ifstream configFile(fileName.c_str());
   if (configFile.is_open())
   {
      po::store(po::parse_config_file(configFile, optionsDesc), vm);
      po::notify(vm);
      configFile.close();
   }
   else
   {
      std::cerr << "imageMapping::createUnwrappingMapping():\n"
         << "WARNING: could not open configuration file " << fileName
         << "!\n"
         << "\tUsing defaults.\n";
   }

   createUnwrappingMapping(imgWidth, imgHeight,
                           outRad, inRad,
                           centreX, centreY,
                           ppd,
                           powe,
                           direct);
}

/*!
 * \brief Create the imageMapping to unwrap a transformed
 * omni-directional image into a panoramic image.
 *
 * So far, some of the parameters of the mapping are hard-coded:
 * - omni-directional image must be square,
 * - size of the target (panoramic image): 360 x 60,
 * - radius of the ignored centre: 40 * omni-directional image size / 200.
 *
 * This mapping is one-to-many.
 *
 * The original mapping must use and produce square images of the same
 * size.  This is only verified if compiled without NDEBUG.
 *
 * \param[in] mapping The mapping to apply to the omni-directional
 * image before unwrapping.
 * \param[in] power The power of the function to convert values along
 * a radial axis in the omni-directional image to a column in the
 * panoramic image (1.0 is linear, 2.0 is squared, etc).
 * \todo Is that function used anywhere?
 */
void imageMapping::createUnwrappingMapping(imageMapping& mapping,
                                           double power)
{
//    assert((sourceWidth_ == sourceHeight_ == targetWidth_ == targetHeight_));
// Why the line above, which does not even make much sense?

   static double invOneEighty = 1.0 / 180.0;

   double convFact = static_cast<double>(sourceWidth_) / 200.0;
                                // The hard coded sizes work for 200x200
                                // images.  They are thus multiplied by
                                // convFact to get the desired ones.
   sourceWidth_ = static_cast<unsigned short>(200 * convFact + 0.5);
   sourceHeight_ = static_cast<unsigned short>(200 * convFact + 0.5);
   targetWidth_ = 360;
   targetHeight_ = 60;

   sourceMinCol_ = 0;
   sourceMaxCol_ = sourceWidth_ - 1;
   sourceMinRow_ = 0;
   sourceMaxRow_ = sourceHeight_ - 1;
   targetMinCol_ = 0;
   targetMaxCol_ = targetWidth_ - 1;
   targetMinRow_ = 0;
   targetMaxRow_ = targetHeight_ - 1;

   double radius = 99.49 * convFact; // Radius of the useful area
   double xCentre = 99.5 * convFact; // Centre of the image.
   double yCentre = 99.5 * convFact;
   unsigned short xPos, yPos;

   double robotRadius = 40.0 * convFact;
                                // Radius of the robot in the centre of the
                                // image.
   double maxRadius = 95.0 * convFact;
                                // Maximum radius of the useful area in the
                                // omni-directional image.

   double radiusStep = 1.0 / static_cast<double>(targetHeight_);
                                // We want targetHeight values.
   double tmpDouble;

   clear();

   double ind = 1.0;
   for (unsigned short rows = 0; rows < targetHeight_; rows++)
   {
      for (unsigned short cols = 0; cols < targetWidth_; cols++)
      {
         radius = robotRadius + ((1.0 - pow((1.0 - ind), power))
                                 * (maxRadius - robotRadius));
         tmpDouble = static_cast<double>(cols * M_PI * invOneEighty);
         xPos = static_cast<unsigned short>(xCentre
                                            + (radius * cos(tmpDouble))
                                            + 0.5);
         if (xPos >= sourceWidth_)
            xPos = sourceWidth_ - 1;
         yPos = static_cast<unsigned short>(yCentre
                                            + (radius * sin(tmpDouble))
                                            + 0.5);
         if (yPos >= sourceHeight_)
            yPos = sourceHeight_ - 1;

         hashTableType::const_iterator iterMapHash =
            mapping.findTarget(xPos, yPos);
         if (iterMapHash != mapping.endHashTargets())
            map.push_back(imageMapping::coordsPair(
               mapping.map[iterMapHash->second].sX,
               mapping.map[iterMapHash->second].sY,
               cols,
               rows));
      }
      ind -= radiusStep;
   }
}


/*!
 * \brief Create a mapping that does nothing!
 *
 * The mapping will map an image onto itself.  It is a one-to-one mapping.
 *
 * \param[in] width The width if the image.
 * \param[in] height The height of the image.
 */
void imageMapping::createNoopMapping(unsigned short width,
                                     unsigned short height)
{
   clear();

   sourceWidth_ = width;
   sourceHeight_ = height;
   targetWidth_ = sourceWidth_;
   targetHeight_ = sourceHeight_;

   sourceMinCol_ = 0;
   sourceMaxCol_ = sourceWidth_ - 1;
   sourceMinRow_ = 0;
   sourceMaxRow_ = sourceHeight_ - 1;
   targetMinCol_ = 0;
   targetMaxCol_ = targetWidth_ - 1;
   targetMinRow_ = 0;
   targetMaxRow_ = targetHeight_ - 1;

   for (unsigned short rows = 0; rows < targetHeight_; rows++)
   {
      for (unsigned short cols = 0; cols < targetWidth_; cols++)
      {
         map.push_back(imageMapping::coordsPair(cols, rows, cols, rows));
      }
   }
}

/*!
 * \brief Create a mapping that will vertically shift pixels.
 *
 * If a shift moves a pixel outside of the target image, then the pixel
 * is ignored.
 *
 * It is a one-to-one mapping.
 *
 * \param[in] width
 * \param[in] height The size of the images.
 * \param[in] shift By how much to shift (positive shifts down).
 * \param[in] minRow
 * \param[in] maxRow Minimum and maximum (inclusive) rows that will
 * be shifted.  The other pixels will not be shifted.
 */
void  imageMapping::createVerticalShiftMapping(unsigned short width,
                                               unsigned short height,
                                               long shift,
                                               unsigned short minRow,
                                               unsigned short maxRow)
{
   clear();

   sourceWidth_ = width;
   sourceHeight_ = height;
   targetWidth_ = sourceWidth_;
   targetHeight_ = sourceHeight_;

   sourceMinCol_ = 0;
   sourceMaxCol_ = sourceWidth_ - 1;
   sourceMinRow_ = 0;
   sourceMaxRow_ = sourceHeight_ - 1;
   targetMinCol_ = 0;
   targetMaxCol_ = targetWidth_ - 1;
   targetMinRow_ = 0;
   targetMaxRow_ = targetHeight_ - 1;

   if (maxRow + shift >= targetHeight_)
      maxRow = targetHeight_ - shift - 1;
   if (static_cast<long>(minRow) + shift < 0)
      minRow = shift;

   for (unsigned short indR = minRow; indR <= maxRow; ++indR)
   {
      for (unsigned short indC = 0; indC < sourceWidth_; ++indC)
      {
         map.push_back(imageMapping::coordsPair(indC, indR,
                                                indC, indR + shift));
      }
   }
}

/*!
 * \brief Create a mapping that will horizontally shift pixels.
 *
 * If a shift moves a pixel outside of the target image, then the pixel
 * is ignored.
 *
 * It is a one-to-one mapping.
 *
 * \param[in] width
 * \param[in] height The size of the images.
 * \param[in] shift By how much to shift (positive shifts to the
 * right).
 * \param[in] minColumn
 * \param[in] maxColumn Minimum and maximum (inclusive) columns that
 * will be shifted.  The other pixels will not be shifted.
 */
void imageMapping::createHorizontalShiftMapping(unsigned short width,
                                                unsigned short height,
                                                long shift,
                                                unsigned short minColumn,
                                                unsigned short maxColumn)
{
   clear();

   sourceWidth_ = width;
   sourceHeight_ = height;
   targetWidth_ = sourceWidth_;
   targetHeight_ = sourceHeight_;

   sourceMinCol_ = 0;
   sourceMaxCol_ = sourceWidth_ - 1;
   sourceMinRow_ = 0;
   sourceMaxRow_ = sourceHeight_ - 1;
   targetMinCol_ = 0;
   targetMaxCol_ = targetWidth_ - 1;
   targetMinRow_ = 0;
   targetMaxRow_ = targetHeight_ - 1;

   if (maxColumn + shift >= targetWidth_)
      maxColumn = targetWidth_ - shift - 1;
   if (static_cast<long>(minColumn) + shift < 0)
      minColumn = shift;

   for (unsigned short indC = minColumn; indC <= maxColumn; ++indC)
   {
      for (unsigned short indR = 0; indR < sourceHeight_; ++indR)
      {
         map.push_back(imageMapping::coordsPair(indC, indR,
                                                indC + shift, indR));
      }
   }
}

/*!
 * \brief Create a warping mapping.
 *
 * It uses the algorithm (and the code is based on the one) described
 * in [\ref Hec94].
 *
 * The boundaries of the target in the source define the size of the
 * target.  Top and bottom should thus have the same size.  The same
 * applies to right and left.  This is only verified if compiled
 * without NDEBUG.
 *
 * The boundaries must run clockwise from the top left corner of the
 * source.
 *
 * Coordinates in the source image (implied by the warping boundaries)
 * will be truncated to the limits of the source image, unless wrapX
 * is true, in which case the x coordinates will be wrapped around
 * (useful if the image is a panoramic image).
 *
 * \param[in] widthS
 * \param[in] heightS The width and height of the source image.
 * \param[in] top
 * \param[in] right
 * \param[in] bottom
 * \param[in] left The boundaries of the target in the source.
 * \param[in] wrapX If true, then wrap the X coordinates in the
 * source image.
 */
void imageMapping::createWarpingMapping(unsigned short widthS,
                                        unsigned short heightS,
                                        const curve& top,
                                        const curve& right,
                                        const curve& bottom,
                                        const curve& left,
                                        bool wrapX)
{
   assert((top.points.size() == bottom.points.size())
          && (right.points.size() == left.points.size()));

   long u0 = 0;
   long v0 = 0;

   clear();

   // What is the source size?
   sourceWidth_ = widthS;
   sourceHeight_ = heightS;
   targetWidth_ = top.points.size();
   targetHeight_ = right.points.size();

   sourceMinCol_ = 0;
   sourceMaxCol_ = sourceWidth_ - 1;
   sourceMinRow_ = 0;
   sourceMaxRow_ = sourceHeight_ - 1;
   targetMinCol_ = 0;
   targetMaxCol_ = targetWidth_ - 1;
   targetMinRow_ = 0;
   targetMaxRow_ = targetHeight_ - 1;

#define SHIFT 20                // number of fractional bits in fixed point
                                // coords
#define SCALE (1<<SHIFT)

   long x, y, qx, qy, dqx, dqy, tX, tY;
   unsigned long u, v, nu, nv;
   double du, dv;

   nu = top.points.size() - 1; // nu = destination width - 1
   nv = right.points.size() - 1; // nv = destination height - 1

   // Bottom and left boundaries are in the reverse of the desired
   // order, running from right to left and bottom to top, resp., so we
   // index them from the end.

   // Upper left patch corner.
   point p00((left.points[nv].x + top.points[0].x) / 2.0,
             (left.points[nv].y + top.points[0].y) / 2.0);
   // Upper right patch corner.
   point p10((top.points[nu].x + right.points[0].x) / 2.0,
             (top.points[nu].y + right.points[0].y) / 2.0);
   // Lower right patch corner.
   point p11((right.points[nv].x + bottom.points[0].x) / 2.0,
             (right.points[nv].y + bottom.points[0].y) / 2.0);
   // Lower left patch corner.
   point p01((bottom.points[nu].x + left.points[0].x) / 2.0,
             (bottom.points[nu].y + left.points[0].y) / 2.0);

   du = 1.0 / nu;
   dv = 1.0 / nv;
   double fv;

   std::vector<segment> pua;
   pua.reserve(nu + 1);
   for (u = 0; u <= nu; ++u)
   {
      pua.push_back(segment(lround(top.points[u].x * SCALE),
                            lround(top.points[u].y * SCALE),
                            lround((bottom.points[nu-u].x
                                           - top.points[u].x)
                                          * dv * SCALE),
                            lround((bottom.points[nu-u].y
                                           - top.points[u].y)
                                          * dv * SCALE)));
   }

   for (fv = 0.0, v = 0; v <= nv; ++v, fv += dv)
   {
      qx = lround((left.points[nv - v].x
                   - (1.0 - fv) * p00.x - fv * p01.x + 0.5)
                  * SCALE);
      qy = lround((left.points[nv - v].y
                   - (1.0 - fv) * p00.y - fv * p01.y + 0.5)
                  * SCALE);
      dqx = lround((right.points[v].x - left.points[nv-v].x
                    - (1.0 - fv) * (p10.x - p00.x)
                    - fv * (p11.x - p01.x))
                   * du * SCALE);
      dqy = lround((right.points[v].y - left.points[nv-v].y
                    - (1.0 - fv) * (p10.y - p00.y)
                    - fv * (p11.y - p01.y))
                   * du * SCALE);
      for (u = 0; u <= nu; ++u)
      {
         x = (pua[u].x + qx) >> SHIFT;
         y = (pua[u].y + qy) >> SHIFT;

         tX = u0 + u;
         tY = v0 + v;
         if (wrapX)
         {
            x = modulo(x, static_cast<long>(sourceWidth_));
         }
         if ((x >= 0) && (x < sourceWidth_)
             && (y >= 0) && (y < sourceHeight_)
             && (tX >= 0) && (tX < targetWidth_)
             && (tY >= 0) && (tY < targetHeight_))
         {
            map.push_back(imageMapping::coordsPair(x, y, tX, tY));
         }
         qx += dqx;
         qy += dqy;
         pua[u].x += pua[u].dx;
         pua[u].y += pua[u].dy;
      }
   }
#undef SHIFT
#undef SCALE
}

/*!
 * \brief Create a mapping that rotates images by 180deg, for when the
 * camera is upside-down.
 *
 * \param[in] width The width if the image.
 * \param[in] height The height of the image.
 */
void imageMapping::createUpsideDownMapping(unsigned short width,
                                           unsigned short height)
{
   clear();

   sourceWidth_ = width;
   sourceHeight_ = height;
   targetWidth_ = sourceWidth_;
   targetHeight_ = sourceHeight_;

   sourceMinCol_ = 0;
   sourceMaxCol_ = sourceWidth_ - 1;
   sourceMinRow_ = 0;
   sourceMaxRow_ = sourceHeight_ - 1;
   targetMinCol_ = 0;
   targetMaxCol_ = targetWidth_ - 1;
   targetMinRow_ = 0;
   targetMaxRow_ = targetHeight_ - 1;

   for (unsigned short rows = 0; rows < targetHeight_; rows++)
   {
      for (unsigned short cols = 0; cols < targetWidth_; cols++)
      {
         map.push_back(imageMapping::coordsPair(cols,
                                                rows,
                                                targetWidth_ - 1 - cols,
                                                targetHeight_ - 1 - rows));
      }
   }
}

/*!
 * \brief Create a mapping that does a horizontal flip.
 *
 * \param[in] width The width if the image.
 * \param[in] height The height of the image.
 */
void imageMapping::createHorizontalFlipMapping(unsigned short width,
                                               unsigned short height)
{
   clear();

   sourceWidth_ = width;
   sourceHeight_ = height;
   targetWidth_ = sourceWidth_;
   targetHeight_ = sourceHeight_;

   sourceMinCol_ = 0;
   sourceMaxCol_ = sourceWidth_ - 1;
   sourceMinRow_ = 0;
   sourceMaxRow_ = sourceHeight_ - 1;
   targetMinCol_ = 0;
   targetMaxCol_ = targetWidth_ - 1;
   targetMinRow_ = 0;
   targetMaxRow_ = targetHeight_ - 1;

   for (unsigned short rows = 0; rows < targetHeight_; rows++)
   {
      for (unsigned short cols = 0; cols < targetWidth_; cols++)
      {
         map.push_back(imageMapping::coordsPair(cols,
                                                rows,
                                                targetWidth_ - 1 - cols,
                                                rows));
      }
   }
}

/*!
 * \brief Apply to the mapping the given one.
 *
 * The source and target pixels will be transformed according to the
 * given mapping.
 *
 * If a source or target pixel is not found in the given mapping, that
 * pair is removed from the original mapping and false is returned.
 *
 * The application of the mapping can be strict (the default) or
 * approximative.  The problem is that a strict application can result
 * in holes left between pixels from the current mapping if the second
 * mapping does not necessarily use contiguous image pixels (such as
 * the mapping created by createUnwrappingMapping()).  The
 * approximative application will look for neighbours of the source
 * and target pixels of the current mapping in the mapping to be
 * applied if these pixels are not handled by this mapping.  The
 * considered neighbours are the four closest pixels.
 *
 * This should work for one-to-one and one-to-many mappings.
 *
 * \param[in] mapping The mapping to apply.
 * \param[in] approx If true, then allow an approximative application
 * of the mapping.
 * \return true if all points of the current mapping have been
 * transformed, false otherwise.
 * \todo Some of this function could use the hashtables for more efficient
 * searches.  Is it worth doing it?
 */
bool imageMapping::apply(imageMapping& mapping, bool approx)
{
   std::cout << "Applying a mapping to another one.  This may take a while...";
   std::cout.flush();

   bool ok = true;

   std::vector<imageMapping::coordsPair> newMap;

   // For all pairs of the current mapping.
   std::vector<imageMapping::coordsPair>::iterator iterMap1;
   for (iterMap1 = map.begin();
        iterMap1 < map.end();
        ++iterMap1)
   {
      // Look for the first source of the current pair of the current
      // mapping in the second mapping.
      std::vector<imageMapping::coordsPair>::iterator iterMap2S =
         std::find_if(mapping.map.begin(),
                      mapping.map.end(),
                      std::bind2nd(std::mem_fun_ref(&coordsPair::sameSources),
                                   *iterMap1));
      if ((iterMap2S == mapping.map.end()) && approx)
      {                      // The approximative matching.
         // This can't use the hash tables because of the approximative
         // search.
         iterMap2S =
            std::find_if(mapping.map.begin(),
                         mapping.map.end(),
                         std::bind2nd(std::mem_fun_ref(&coordsPair::sameSourcesApprox),
                                      *iterMap1));
      }
      if (iterMap2S == mapping.map.end())
      {
         ok = false;
      }
      else
      {
         // Look for all pairs in the new mapping that have for source
         // the target of the current pair of the current mapping.  If
         // none is found and approx is true, look for ONE approximative
         // match.
         std::vector<imageMapping::coordsPair>::iterator iterMap2T =
            mapping.map.begin();
         bool foundOne = false;
         while (iterMap2T < mapping.map.end())
         {
            iterMap2T =
               std::find_if(iterMap2T,
                            mapping.map.end(),
                            std::bind2nd(std::mem_fun_ref(&coordsPair::sameSourceTarget),
                                         *iterMap1));
            if (iterMap2T < mapping.map.end())
            {
               foundOne = true;
               newMap.push_back(imageMapping::coordsPair(iterMap2S->tX,
                  iterMap2S->tY,
                  iterMap2T->tX,
                  iterMap2T->tY));
               ++iterMap2T;
            }
         }
         if (!foundOne && approx)
         {                // The approximative matching.
            iterMap2T =
               std::find_if(mapping.map.begin(),
                            mapping.map.end(),
                            std::bind2nd(std::mem_fun_ref(&coordsPair::sameSourceTargetApprox),
                                         *iterMap1));
            if (iterMap2T < mapping.map.end())
            {
               newMap.push_back(imageMapping::coordsPair(iterMap2S->tX,
                  iterMap2S->tY,
                  iterMap2T->tX,
                  iterMap2T->tY));
            }
         }
      }
   }

   map = newMap;
   sourceWidth_ = mapping.targetWidth_;
   sourceHeight_ = mapping.targetHeight_;
   targetWidth_ = mapping.targetWidth_;
   targetHeight_ = mapping.targetHeight_;
   sourceMinCol_ = 0;
   sourceMaxCol_ = sourceWidth_ - 1;
   sourceMinRow_ = 0;
   sourceMaxRow_ = sourceHeight_ - 1;
   targetMinCol_ = 0;
   targetMaxCol_ = targetWidth_ - 1;
   targetMinRow_ = 0;
   targetMaxRow_ = targetHeight_ - 1;

   hashTablesValid = false;

   std::cout << " done.\n";

   return(ok);
}

/*!
 * \brief Apply a pixelSelector to the mapping.
 *
 * Only those pixels that are mapped by the mapping AND in the selector are
 * kept.
 *
 * The function is certainly not efficient and is therefore not meant to be
 * used repeatedly.  In particular, the current mapping is copied into a
 * temporary one and then filtered back into the original.
 *
 * \param[in] selector The pixel selector to use.
 */
void imageMapping::apply(const pixelSelector& selector)
{
   std::vector<coordsPair> tmpMap;

   std::vector<imageMapping::coordsPair>::iterator iterMap;
   std::vector<pixelSelector::coords>::iterator iterSelect;

   for (iterMap = map.begin();
        iterMap < map.end();
        ++iterMap)
   {
      pixelSelector::hashTableType::const_iterator iterHashSelector =
         selector.find(iterMap->sX, iterMap->sY);
      if (iterHashSelector != selector.endHash())
         tmpMap.push_back(*iterMap);
   }

   map = tmpMap;
   hashTablesValid = false;
}

/*!
 * \brief Concatenate two image mappings.
 *
 * A new mapping is created that will have all the pairs of the second mapping
 * that have a source equal to a target in the first (original) mapping.  Some
 * pairs of the second or first mappings might not be represented in the new
 * mapping.  The new mapping replaces the first (original) mapping.
 *
 * The source size of the new mapping is the source size of the original
 * mapping while the target size of the new mapping is the target size of the
 * second mapping.
 *
 * \param[in] mapping The second image mapping.
 */
void imageMapping::concatenate(const imageMapping& mapping)
{
   std::vector<imageMapping::coordsPair> newMap;

   // For all pairs of the second mapping.
   std::vector<imageMapping::coordsPair>::const_iterator iterMap2;
   for (iterMap2 = mapping.map.begin();
        iterMap2 < mapping.map.end();
        ++iterMap2)
   {
      // Look for the first source of the current pair of the second
      // mapping in the first (original) mapping.
      hashTableType::const_iterator iterHashTargetMap1 =
         findTarget(iterMap2->sX, iterMap2->sY);
      if (iterHashTargetMap1 != endHashTargets())
      {
         newMap.push_back(imageMapping::coordsPair(
            map[iterHashTargetMap1->second].sX,
            map[iterHashTargetMap1->second].sY,
            iterMap2->tX,
            iterMap2->tY));
      }
   }

   map = newMap;
   targetWidth_ = mapping.targetWidth_;
   targetHeight_ = mapping.targetHeight_;



   sourceMinCol_ = 0;
   sourceMaxCol_ = sourceWidth_ - 1;
   sourceMinRow_ = 0;
   sourceMaxRow_ = sourceHeight_ - 1;
   targetMinCol_ = 0;
   targetMaxCol_ = targetWidth_ - 1;
   targetMinRow_ = 0;
   targetMaxRow_ = targetHeight_ - 1;




   hashTablesValid = false;
}

/*!
 * \brief Updates the hash tables used for efficient searching.
 *
 * This function should be called at the end of each mapping modifications.
 */
void imageMapping::updateHashTables()
{
//    std::clog << "Updating hash tables\n";

   // Empty the current hash tables.
   hashSources.clear();
   hashTargets.clear();

   // Populate the hash tables.
   for (unsigned int ind = 0; ind < map.size(); ++ind)
   {
      hashSources.insert(std::pair<unsigned long, unsigned long>(
                            map[ind].sY * sourceWidth_ + map[ind].sX, ind));
      hashTargets.insert(std::pair<unsigned long, unsigned long>(
                            map[ind].tY * targetWidth_ + map[ind].tX, ind));
   }
   hashTablesValid = true;
}

} // namespace imalib
