/*
 * imalib
 * 
 * Copyright 2017 Aberystwyth University
 * 
 * This library and program are free software: you can redistribute them
 * and/or modify them under the terms of the GNU General Public License as 
 * published by the Free Software Foundation, either version 3 of the License,
 * or (at your option) any later version.
 * 
 * The program and library are distributed in the hope that they will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*!
 * \file imageRGB.cpp
 * \brief Defines classes to play with RGB and rgb images.  Implementation.
 * \author Fred Labrosse <ffl@aber.ac.uk>
 * \copyright 2017 Aberystwyth University
 */

#include "imalib/imageRGB.h"

#include <math.h>

#include <limits>

#ifdef WITH_PNG
#  include <png.h>
#endif // WITH_PNG

#ifdef WITH_JPEG
#  include <jpeglib.h>
#  include <jerror.h>
#  include <setjmp.h>
#endif // WITH_JPEG

#ifdef WITH_OPENEXR
#  include <OpenEXR/ImfTestFile.h>
#  include <OpenEXR/ImfArray.h>
#  include <OpenEXR/ImfPixelType.h>
#  include <OpenEXR/ImfFrameBuffer.h>
#  include <OpenEXR/ImfInputFile.h>
#  include <OpenEXR/ImathBox.h>
#endif // WITH_OPENEXR


namespace imalib
{
   //-----------------------------------------------------
   // Class imageRGB.
   //-----------------------------------------------------

   /*!
    * \brief Reads an image in the PPM format.
    * \param[in] fileName The name of the file containing the image.
    * \return true if the operation is successful.
    */
   bool imageRGB::readPPM(const std::string& fileName)
   {
      const std::string
         errorMsg("In function: void imageRGB::readPPM(const std::string&)\n");

      std::ifstream file(fileName.c_str());
      if (!file.is_open())
      {
         std::cerr << errorMsg
                   << "Could not open file " << fileName << ".\n";
         return(false);
      }

      char aChar;
      std::string word;
      file >> word;
      file.get(aChar); // To read the separator.
      if ("P3" == word)
      {
         std::cerr << errorMsg
                   << "Plain PPM format is not supported.\n";
         file.close();
         return(false);
      }
      else if ("P6" == word)
      {
         file >> word;
         // Possible comment.
         while ('#' == word[0])
         {
            getline(file, word); // To read to the end of the line.
            file >> word;
         }

         // Width (columns) and height (rows).
         unsigned short col = atoi(word.c_str());
         file >> word;
         unsigned short row = atoi(word.c_str());

         // Max colour.
         file >> word;
         unsigned short maxCol = atoi(word.c_str());
         if (255 < maxCol)
         {
            std::cerr << errorMsg
                      << "Only three bytes per pixel images are supported.\n";
            file.close();
            return(false);
         }

         // Let's read the data.
         setSize(col, row);

         unsigned char pix[3];
         pix[0] = file.get();     // To skip the eol of the previous line.

         for (row = 0; row < nbRows(); ++row)
            for (col = 0; col < nbColumns(); ++col)
            {
//                file.read(pix, 3);
//                file >> pix[0];
//                file >> pix[1];
//                file >> pix[2];
               pix[0] = file.get();
               pix[1] = file.get();
               pix[2] = file.get();

               setPixel(col, row, static_cast<unsigned char*>(pix));
            }
      }
      else
      {
         std::cerr << errorMsg
                   << "The file " << fileName
                   << " does not contain a PPM image.\n";
         file.close();
         return(false);
      }

      file.close();
      return(true);
   }

   /*!
    * \brief Write an image in the PPM format.
    * \param[in] fileName The name of the file containing the image.
    * \param[in] comment The string to be embedded in the image.
    * \return true if the operation is successful.
    */
   bool imageRGB::writePPM(const std::string& fileName, const std::string& comment) const
   {
      const std::string
         errorMsg("In function: void imageRGB::writePPM(const std::string&)\n");

      std::istringstream comment_stream(comment) ;
      std::ofstream file(fileName.c_str());
      if (!file.is_open())
      {
         std::cerr << errorMsg
                   << "Could not open file " << fileName << " for writing.\n";
         return(false);
      }

      file << "P6\n";
      file << "# Generated by imalib v." << PACKAGE_VERSION << ".\n";
      for (std::string line; std::getline( comment_stream, line, '\n');){
         file << "#" << line << "\n";
      }
      file << nbColumns() << " " << nbRows() << "\n255\n";

      const unsigned char* data = getActualPixel(0, 0);

      for (unsigned long ind = 0;
           ind < (static_cast<unsigned long>(nbColumns())
                   * static_cast<unsigned long>(nbRows())
                   * 3);
           ++ind)
         file << data[ind];

      file.close();
      return(true);
   }

/*!
    * \brief Write an image in the PPM format.
    * \param[in] fileName The name of the file containing the image.
    * \return true if the operation is successful.
    */
   bool imageRGB::writePPM(const std::string& fileName) const
   {
      const std::string
         errorMsg("In function: void imageRGB::writePPM(const std::string&)\n");

      std::ofstream file(fileName.c_str());
      if (!file.is_open())
      {
         std::cerr << errorMsg
                   << "Could not open file " << fileName << " for writing.\n";
         return(false);
      }

      file << "P6\n";
      file << "# Generated by imalib v." << PACKAGE_VERSION << ".\n";
      file << nbColumns() << " " << nbRows() << "\n255\n";

      const unsigned char* data = getActualPixel(0, 0);

      for (unsigned long ind = 0;
           ind < (static_cast<unsigned long>(nbColumns())
                   * static_cast<unsigned long>(nbRows())
                   * 3);
           ++ind)
         file << data[ind];

      file.close();
      return(true);
   }


   /*!
    * \brief Write an image in the XPM format.
    * \param[in] fileName The name of the file that will contain the image.
    * \return true if the operation is successful.
    */
   bool imageRGB::writeXPM(const std::string& fileName) const
   {
      std::string
         errorMsg = "In function: bool imageRGB::writeXPM(const std::string& fileName)";

      std::ofstream file(fileName.c_str());
      if (!file.is_open())
      {
         std::cerr << errorMsg
                   << "Could not open file " << fileName << ".\n";
         return(false);
      }

      std::string::size_type extPos = fileName.rfind('.');
      if (extPos == std::string::npos)
      {
         extPos = fileName.length();
      }
      std::string name = fileName.substr(0, extPos);

      std::map<std::string, std::string> colours;
      unsigned short nbChar;
      int index;
//       std::string colour, code;
      std::ostringstream aStream;
      unsigned char pixel[3];

      // Let's save all colours...
      for (short col = 0; col < nbColumns(); ++col)
      {
         for (short row = 0; row < nbRows(); ++row)
         {
            getPixel(col, row, pixel);
            aStream << "#"
                    << std::hex << static_cast<int>(pixel[0])
                    << std::hex << static_cast<int>(pixel[1])
                    << std::hex << static_cast<int>(pixel[2]);
            colours.insert(make_pair(aStream.str(),
                                     std::string("")));
         }
      }
      // and give them a code.
      nbChar = 1;
      while (pow(26, nbChar) < colours.size())
         ++nbChar;
                                // Yes, I know, there is probably
                                // better... but I can't remember for now.
      std::string code;
      code.assign(nbChar, 'a');

      bool cont;
      std::map<std::string, std::string>::iterator iter;
      for (iter = colours.begin(); iter != colours.end(); ++iter)
      {
         iter->second = code;
         cont = true;
         index = nbChar - 1;
         while (cont)
         {
            ++code[index];
            if (code[index] > 'z')
            {                      // We need to wrap
               code[index] = 'a';
               --index;
               if (index < 0)
                  cont = false;
            }
            else
               cont = false;
         }
      }

      //
      // Let's create the pixmap.
      //

      // Let's write the specs.
      file << "/* XPM */\n"
           << "static char *" << name << "[] = {\n"
           << "/* columns rows colors chars-per-pixel */\n"
           << "\"" << nbColumns() << " " << nbRows() << " "
           << colours.size() << " " << nbChar << "\"";

      // Let's write the colour map.
      for (iter = colours.begin(); iter != colours.end(); ++iter)
      {
         file << ",\n\"" << iter->second << " c " << iter->first << "\"";
      }

      // We now just need to loop over the pixels and get the colour code from
      // colours.
      for (unsigned short row = 0; row < nbRows(); ++row)
      {
         for (unsigned short col = 0; col < nbColumns(); ++col)
         {
            getPixel(col, row, pixel);
//          aStream.str("");
            aStream << "#"
                    << std::hex << static_cast<int>(pixel[0])
                    << std::hex << static_cast<int>(pixel[1])
                    << std::hex << static_cast<int>(pixel[2]);
            iter = colours.find(aStream.str());
            file << ",\n\"" << iter->second << "\"";
         }
      }

      file << "\n"
           << "};";

      file.close();

      return(true);
   }

   /*!
    * \brief Write an image in txt format.
    * 
    * This write a desaturated image (grey) where each line of the txt file
    * corresponds to a row of the image and therefore has nbRows() lines.
    *  Values are between 0 and 255.
    * \param[in] fileNmae The name of the file that will contain the image.
    * \return true if the operation is successful.
    */
   bool imageRGB::writeTxt(const std::string& fileName) const
   {
      std::string
         errorMsg = "In function: bool imageRGB::writeTxt(const std::string& fileName)";

      std::ofstream outfile(fileName.c_str());
      if (!outfile.is_open())
      {
         std::cerr << errorMsg
            << "Could not open file " << fileName << ".\n";
         return(false);
      }

      unsigned char pixel[3];
      for (unsigned short y = 0; y < nbRows(); y++)
      {
         for (unsigned short x = 0; x < nbColumns(); x++)
         {
            getPixel(x, y, pixel);
            outfile << pixel[0] * 0.3
                       + pixel[1] * 0.59
                       + pixel[2] * 0.11
                    << " ";
         }
         outfile << std::endl;
      }

      outfile.close();

      return(true);
   }

#ifdef WITH_PNG
   const std::string imageRGB::readFormats = "*.png *.ppm";
   const std::string imageRGB::writeFormats = "*.png *.xpm";
#else
   const std::string imageRGB::readFormats = "* ppm";
   const std::string imageRGB::writeFormats = "*.xpm";
#endif // WITH_PNG

   /*
     \brief Output the supported image formats.
     \param output (IN/OUT) The stream where the formats are written.
     \sa getReadFormats(), getWriteFormats().
   */
   void imageRGB::outputSupportedImageFormats(std::ostream& output)
   {
      output << "- Image formats currently supported are:" << std::endl
#ifdef WITH_PNG
          << "\tping: read/write (png)," << std::endl
#endif // WITH_PNG
#ifdef WITH_JPEG
          << "\tjpeg: read (jpg)," << std::endl
#endif // WITH_JPEG
          << "\tportable pixmap: read (ppm)," << std::endl
          << "\tX PixMap: write (xpm)." << std::endl
          << "\ttext: write (txt)";
   }


   /*
    * \brief Read an imageRGB from a file.
    * \param[in] fileName The name of the file containing the image.
    * \return true if the operation is successful.
    */
   bool imageRGB::read(const std::string& fileName)
   {
      bool success = false;
      std::string extension;

      std::string::size_type extPos = fileName.rfind('.');
      if (extPos != std::string::npos)
         extension = fileName.substr(extPos + 1);

      if (extension == "ppm")
         success = readPPM(fileName);
#ifdef WITH_PNG
      else if (extension == "png")
         success = readPNG(fileName);
#endif
#ifdef WITH_JPEG
      else if ((extension == "jpg") || (extension == "jpeg"))
         success = readJPEG(fileName);
#endif // WITH_JPEG
#ifndef NDEBUG
      else
      {
         std::cerr
            << "In function: bool imageRGB::read(const std::string&)\n"
            << "\t" << extension << " is an unsupported image format\n"
            << "\tin image name " << fileName << std::endl;
      }
#endif // !NDEBUG

      return(success);
   }

   /*
     \brief Write the image into a file.
     \param fileName (IN) The name of the file.
     \return true if successful, false otherwise.
   */
   bool imageRGB::write(const std::string& fileName) const
   {
      bool success = false;
      std::string extension;

      std::string::size_type extPos = fileName.rfind('.');
      if (extPos != std::string::npos)
      {
         extension = fileName.substr(extPos + 1);
      }

      if (extension == "xpm")
         success = writeXPM(fileName);
      else if (extension == "ppm")
         success = writePPM(fileName);
#ifdef WITH_PNG
      else if (extension == "png")
         success = writePNG(fileName);
#endif // WITH_PNG
      else if (extension == "txt")
         success = writeTxt(fileName);

#ifndef NDEBUG
      else
      {
         std::cerr
            << "In function: bool imageRGB::write(const std::string&)\n"
            << extension << " is an unsupported image format.\n";
      }
#endif // !NDEBUG

      return(success);
   }

   /*!
    \brief Write the image into a file.
    \param[in] fileName The name of the file.
    \param[in] comment The comment to add to the file, new line has to start with a # so '\n#'
    \return true if successful, false otherwise.
   */
   bool imageRGB::write(const std::string& fileName, const std::string& comment) const
   {
      bool success = false;
      std::string extension;

      std::string::size_type extPos = fileName.rfind('.');
      if (extPos != std::string::npos)
      {
         extension = fileName.substr(extPos + 1);
      }

      if (extension == "ppm")
         success = writePPM(fileName,comment);

#ifndef NDEBUG
      else
      {
         std::cerr
            << "In function: bool imageRGB::write(const std::string&, const std::string&)\n"
            << extension << " is an unsupported image format, only ppm support for comments.\n";
      }
#endif // !NDEBUG

      return(success);
   }



#ifdef WITH_PNG
   /*
     \brief Reads an image in the PNG format.
     \param fileName  (IN) The name of the file containing the image.
     \return true if the operation is successful.
   */
   bool imageRGB::readPNG(const std::string& fileName)
   {
      const unsigned short headerRead = 8;
      int color_type, bit_depth;
      png_uint_32 width, height;
//       double screen_gamma;
//       std::string gamma_str;
//       double gamma;
      const std::string errorMsg("In function: void imageRGB::readPNG(const std::string&)\n");

      // Open the file.
      FILE* file = fopen(fileName.c_str(), "rb");
      if (!file)
      {
         std::cerr << errorMsg
                   << "Cannot open the file " << fileName << ".\n";
         return(false);
      }

      // Verify that the file is a PNG file.
      unsigned char header[headerRead];
      size_t ret = fread(header, 1, headerRead, file);
      if (ret != headerRead)
      {
         std::cerr << "Could not read PNG header.\n";
         return(false);
      }

      if (png_sig_cmp(header, 0, headerRead))
//       if (png_check_sig(header, headerRead))
      {
         std::cerr << errorMsg
                   << "The file " << fileName
                   << " is not a PNG file.\n";
         return(false);
      }

      // Allocate and initialise png_struct and png_info.
      png_structp png_ptr =
         png_create_read_struct(PNG_LIBPNG_VER_STRING,
                                NULL, NULL, NULL);
      if (!png_ptr)
      {
         std::cerr << errorMsg
                   << "Could not create a libpng structure.\n";
         return(false);
      }

      png_infop info_ptr = png_create_info_struct(png_ptr);
      if (!info_ptr)
      {
         png_destroy_read_struct(&png_ptr,
                                 (png_infopp)NULL, (png_infopp)NULL);
         std::cerr << errorMsg
                   << "Could not create a libpng structure.\n";
         return(false);
      }
//    png_infop end_info = png_create_info_struct(png_ptr);

      // Reading the info.
      png_init_io(png_ptr, file);
      png_set_sig_bytes(png_ptr, headerRead);
      png_read_info(png_ptr, info_ptr);
      png_get_IHDR(png_ptr, info_ptr,
                   &width, &height, &bit_depth, &color_type,
                   NULL, NULL, NULL);

      // Expand paletted colors into true RGB triplets
      if (color_type == PNG_COLOR_TYPE_PALETTE)
         png_set_expand(png_ptr);
      // Expand grayscale images to the full 8 bits from 1, 2, or 4
      // bits/pixel
      if (color_type == PNG_COLOR_TYPE_GRAY && bit_depth < 8)
         png_set_expand(png_ptr);
      // Strip to 8 bits/pixel.
      if (bit_depth == 16)
         png_set_strip_16(png_ptr);
      // Remove the transparency.
      if (color_type & PNG_COLOR_MASK_ALPHA)
         png_set_strip_alpha(png_ptr);
      // Expand to 1 pixel per byte.
      if (bit_depth < 8)
         png_set_packing(png_ptr);
      // Transform grayscale to RGB.
      if (color_type == PNG_COLOR_TYPE_GRAY ||
          color_type == PNG_COLOR_TYPE_GRAY_ALPHA)
         png_set_gray_to_rgb(png_ptr);

//       // Add an alpha channel if not already there.
//       png_set_filler(png_ptr, 0x000000ff, PNG_FILLER_AFTER);

      // Composite the alpha channel with a background colour.
//       png_color_16 myBackground;
//       myBackground.red = 127;
//       myBackground.green = 127;
//       myBackground.blue = 127;

//       png_color_16p image_background;
//       if (png_get_bKGD(png_ptr, info_ptr, &image_background))
//          png_set_background(png_ptr, image_background,
//                             PNG_BACKGROUND_GAMMA_FILE, 1, 1.0);
//       else
//          png_set_background(png_ptr, &myBackground,
//                             PNG_BACKGROUND_GAMMA_SCREEN, 0, 1.0);

#if false
      // Do gamma correction.
      if (getenv("SCREEN_GAMMA") != NULL)
      {
         gamma_str = getenv("SCREEN_GAMMA");
         screen_gamma = atof(gamma_str.c_str());
      }
      else
      {
#if defined(Linux)
         screen_gamma = 2.5; // Guess for a PC.
#elif defined(SunOS)
         screen_gamma = 2.5; // Guess for a Sun.
#elif defined(IRIX) || defined(IRIX64)
         screen_gamma = 1.7; // Guess for an SGI.
#else
         screen_gamma = 2.5;
#endif
      }
      if (png_get_gAMA(png_ptr, info_ptr, &gamma))
         png_set_gamma(png_ptr, screen_gamma, gamma);
      else
         png_set_gamma(png_ptr, screen_gamma, 0.45455);
#endif // false

      /* Turn on interlace handling.  REQUIRED if you are not using
       * png_read_image().  To see how to handle interlacing passes,
       * see the png_read_row() method below.
       */
      /*int number_passes = */png_set_interlace_handling(png_ptr);

      // Reread the info.
      png_read_update_info(png_ptr, info_ptr);

      // Set the size of the image.
      setSize(width, height);

      // Allocate the memory to hold the image using the fields of
      // info_ptr.
      png_uint_32 rowbyte = png_get_rowbytes(png_ptr, info_ptr);

      png_bytep* row_pointers
         = static_cast<png_bytep*>(png_malloc(png_ptr,
                                              sizeof(png_bytep) * height));
      if (!row_pointers)
      {
         png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);
         std::cerr << errorMsg
                   << "Could not allocate memory for the image.\n";
         return(false);
      }

      row_pointers[0] = getActualPixel(0, 0);
      for (unsigned short indR = 1; indR < height; indR++)
      {
         row_pointers[indR] = row_pointers[indR-1] + rowbyte;
//          row_pointers[indR] = png_malloc(png_ptr,
//                                          png_get_rowbytes(png_ptr, info_ptr));
//          if (!row_pointers[indR])
//          {
//             png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);
//             for (int ind = 0; ind < indR; ++ind)
//                free(row_pointers[ind]);
//             free(row_pointers);
//             std::cerr << errorMsg
//                       << "Could not allocate memory for the image.\n";
//             return(false);
//          }
      }

      // Read the image.
      png_read_image(png_ptr, row_pointers);

      // read rest of file, and get additional chunks in info_ptr -
      // REQUIRED
      png_read_end(png_ptr, info_ptr);

      fclose(file);

      // Transfer into the imageRGB.
//       for (short indR = 0; indR < height; indR++)
//          for (short indC = 0; indC < width; indC++)
//          {
//             setPixel(indC, indR,
//                      T(colourUChar(row_pointers[indR][4*indC+0],
//                                    row_pointers[indR][4*indC+1],
//                                    row_pointers[indR][4*indC+2])));
//             if (haveAlphaData)
//                setAlpha(indC, indR,
//                         double(row_pointers[indR][4*indC+3]) / 255.0);
//          }

      // Free the allocated memory.
      png_free(png_ptr, row_pointers);
      png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);

      return(true);
   }

   /*
     \brief Writes an image in the PNG format.
     \param fileName (IN) The name of the file that will contain the
     image.
     \return true if the operation is successful.
   */
   bool imageRGB::writePNG(const std::string& fileName) const
   {
//       double screen_gamma;
      std::string gamma_str;
//       unsigned short indC, indR;
      const std::string errorMsg("In function: void imageRGB::writePNG(const std::string&)\n");

      FILE *fp = fopen(fileName.c_str(), "wb");
      if (!fp)
      {
         std::cerr << errorMsg
                   << "Cannot open file " << fileName << ".\n"
                   << "The image is not saved.\n";
         return(false);
      }

      png_structp png_ptr =
         png_create_write_struct(PNG_LIBPNG_VER_STRING,
                                 NULL, NULL, NULL);
      if (!png_ptr)
      {
         std::cerr << errorMsg
                   << "Could not create a libpng structure.\n";
         return(false);
      }

      png_infop info_ptr = png_create_info_struct(png_ptr);
      if (!info_ptr)
      {
         png_destroy_write_struct(&png_ptr, (png_infopp)NULL);
         std::cerr << errorMsg
                   << "Could not create a libpng structure.\n";
         return(false);
      }

      png_init_io(png_ptr, fp);

      int colourType = PNG_COLOR_TYPE_RGB;
//       short bpp = 3;                // How many bytes per pixel do we have.

      // Fillin the info structure.
      png_set_IHDR(png_ptr, info_ptr, nbColumns(), nbRows(), 8,
                   colourType, PNG_INTERLACE_ADAM7,
                   PNG_COMPRESSION_TYPE_DEFAULT,
                   PNG_FILTER_TYPE_DEFAULT);

#if false
      if (getenv("SCREEN_GAMMA") != NULL)
      {
         gamma_str = getenv("SCREEN_GAMMA");
         screen_gamma = atof(gamma_str.c_str());
      }
      else
      {
#if defined(Linux)
         screen_gamma = 2.5;    // Guess for a PC.
#elif defined(SunOS)
         screen_gamma = 2.5;    // Guess for a Sun.
#elif defined(IRIX) || defined(IRIX64)
         screen_gamma = 1.7;    // Guess for an SGI.
#else
         screen_gamma = 2.5;
#endif
      }
//       png_set_gAMA(png_ptr, info_ptr, 0.45455);
      png_set_gAMA(png_ptr, info_ptr, 1.0 / screen_gamma);
#endif // false

      png_write_info(png_ptr, info_ptr);

      // Create the row pointers to point to the relevant bits in the image.
      png_uint_32 rowbyte = png_get_rowbytes(png_ptr, info_ptr);
      png_bytep* row_pointers
         = static_cast<png_bytep*>(png_malloc(png_ptr,
                                              sizeof(png_bytep) * nbRows()));
      if (!row_pointers)
      {
         png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);
         std::cerr << errorMsg
                   << "Could not allocate memory for the image.\n";
         return(false);
      }

      row_pointers[0] = static_cast<png_bytep>(
         const_cast<unsigned char*>(getActualPixel(0, 0)));
      for (short indR = 1; indR < nbRows(); indR++)
      {
         row_pointers[indR] = row_pointers[indR - 1] + rowbyte;
      }

      // Write the image.
      png_write_image(png_ptr, row_pointers);
      png_write_end(png_ptr, info_ptr);

      // Close the file.
      fclose(fp);

      // Free the allocated memory.
      png_free(png_ptr, row_pointers);
      png_destroy_write_struct(&png_ptr, &info_ptr);

      return(true);
   }
#endif // WITH_PNG

#ifdef WITH_JPEG
/*!
 * \brief Reads an image in the JPEG format.
 *
 * Based on example.c from libjpeg-turbo.
 *
 * \param[in] fileName  The name of the file containing the image.
 * \return true if the operation is successful.
 */
bool imageRGB::readJPEG(const std::string& fileName)
{
   const std::string errorMsg("In function: void imageRGB::readJPEG(const std::string&)\n");

   /* This struct contains the JPEG decompression parameters and pointers to
    * working space (which is allocated as needed by the JPEG library).
    */
   struct jpeg_decompress_struct cinfo;
   struct jpeg_error_mgr jerr;
   cinfo.err = jpeg_std_error(&jerr);
   /* More stuff */
   FILE * infile;                /* source file */
   JSAMPARRAY buffer;            /* Output row buffer */
   int row_stride;               /* physical row width in output buffer */

   /* In this example we want to open the input file before doing anything else,
    * so that the setjmp() error recovery below can assume the file is open.
    * VERY IMPORTANT: use "b" option to fopen() if you are on a machine that
    * requires it in order to read binary files.
    */

   if ((infile = fopen(fileName.c_str(), "rb")) == NULL)
   {
      std::cerr << errorMsg
         << "can't open file " << fileName << std::endl;
      return(false);
   }

   /* Step 1: allocate and initialize JPEG decompression object */

   /* Now we can initialize the JPEG decompression object. */
   jpeg_create_decompress(&cinfo);

   /* Step 2: specify data source (eg, a file) */

   jpeg_stdio_src(&cinfo, infile);

   /* Step 3: read file parameters with jpeg_read_header() */

   jpeg_read_header(&cinfo, TRUE);
   /* We can ignore the return value from jpeg_read_header since
    *   (a) suspension is not possible with the stdio data source, and
    *   (b) we passed TRUE to reject a tables-only JPEG file as an error.
    * See libjpeg.txt for more info.
    */

   /* Step 4: set parameters for decompression */

   /* In this example, we don't need to change any of the defaults set by
    * jpeg_read_header(), so we do nothing here.
    */

   /* Step 5: Start decompressor */

   (void) jpeg_start_decompress(&cinfo);
   /* We can ignore the return value since suspension is not possible
    * with the stdio data source.
    */

   // This can only be done after decompression has started.
   if (cinfo.output_components != 3)
   {
      std::cerr << errorMsg
         << "only 3 component images can be dealt with at the moment!\n";
      jpeg_destroy_decompress(&cinfo);
      fclose(infile);
      return(false);
   }
   if (cinfo.out_color_space != JCS_RGB)
   {
      std::cerr << errorMsg
         << "only RGB images can be dealt with at the moment!\n";
      jpeg_destroy_decompress(&cinfo);
      fclose(infile);
      return(false);
   }

   /* We may need to do some setup of our own at this point before reading
    * the data.  After jpeg_start_decompress() we have the correct scaled
    * output image dimensions available, as well as the output colormap
    * if we asked for color quantization.
    * In this example, we need to make an output work buffer of the right size.
    */
   /* JSAMPLEs per row in output buffer */
   row_stride = cinfo.output_width * cinfo.output_components;

   setSize(cinfo.output_width, cinfo.output_height);

   /* Make a one-row-high sample array that will go away when done with image */
   buffer = (*cinfo.mem->alloc_sarray)
                      ((j_common_ptr) &cinfo, JPOOL_IMAGE, row_stride, 1);

   /* Step 6: while (scan lines remain to be read) */
   /*           jpeg_read_scanlines(...); */

   /* Here we use the library's state variable cinfo.output_scanline as the
    * loop counter, so that we don't have to keep track ourselves.
    */
   unsigned short row = 0;
   while (cinfo.output_scanline < cinfo.output_height)
   {
      /* jpeg_read_scanlines expects an array of pointers to scanlines.
       * Here the array is only one element long, but you could ask for
       * more than one scanline at a time if that's more convenient.
       */
      (void) jpeg_read_scanlines(&cinfo, buffer, 1);
      /* Assume put_scanline_someplace wants a pointer and sample count. */
      memcpy(getActualPixel(0, row), buffer[0],
             row_stride * sizeof(unsigned char));
      row++;
//       put_scanline_someplace(buffer[0], row_stride);
   }

   /* Step 7: Finish decompression */

   (void) jpeg_finish_decompress(&cinfo);
   /* We can ignore the return value since suspension is not possible
    * with the stdio data source.
    */

   /* Step 8: Release JPEG decompression object */

   /* This is an important step since it will release a good deal of memory. */
   jpeg_destroy_decompress(&cinfo);

   /* After finish_decompress, we can close the input file.
    * Here we postpone it until after no more JPEG errors are possible,
    * so as to simplify the setjmp error logic above.  (Actually, I don't
    * think that jpeg_destroy can do an error exit, but why assume anything...)
    */
   fclose(infile);

   /* And we're done! */
   return(true);
}
#endif // WITH_JPEG

   //-----------------------------------------------------
   // Class imageRgb.
   //-----------------------------------------------------

   /*!
     \brief Convert a pixel in the rgb colour space (0.0 to 1.0) to a
     pixel in the RGB colour space (0 to 255).

     rgb values are checked only if compiled without NDEBUG.

     \param lab (IN) The pixel in rgb colour space.
     \param rgb (OUT) The pixel in RGB colour space (must be allocated
     beforehand).
   */
   void imageRgb::rgbToRGB(const double* rgb, unsigned char* RGB)
   {
      assert((rgb[0] >= 0.0) && (rgb[0] <= 1.0)
             && (rgb[1] >= 0.0) && (rgb[1] <= 1.0)
             && (rgb[2] >= 0.0) && (rgb[2] <= 1.0));

      RGB[0] = static_cast<unsigned char>(rgb[0] * 255.0);
      RGB[1] = static_cast<unsigned char>(rgb[1] * 255.0);
      RGB[2] = static_cast<unsigned char>(rgb[2] * 255.0);
   }

   /*!
     \brief Convert a pixel in the RGB colour space (0 to 255) to a
     pixel in the rgb colour space (0.0 to 1.0).
     \param rgb (OUT) The pixel in RGB colour space.
     \param lab (IN) The pixel in rgb colour space (must
     be allocated beforehand).
   */
   void imageRgb::RGBToRgb(const unsigned char* RGB, double* rgb)
   {
      rgb[0] = static_cast<double>(RGB[0]) / 255.0;
      rgb[1] = static_cast<double>(RGB[1]) / 255.0;
      rgb[2] = static_cast<double>(RGB[2]) / 255.0;
   }


   /*
    * \brief Read an imageRgb from a file.
    * \param[in] fileName The name of the file containing the image.
    * \return true if the operation is successful.
    */
   bool imageRgb::read(const std::string& fileName)
   {
      bool success = false;
      std::string extension;

      const std::string::size_type extPos = fileName.rfind('.');
      if (extPos != std::string::npos)
         extension = fileName.substr(extPos + 1);

      if ((extension == "ppm")
#ifdef WITH_PNG
         || (extension == "png")
#endif
          )
      {
         imageRGB temp;
         success = temp.read(fileName);
         if (success)
            copyFromRGB(temp);
      }
#ifdef WITH_OPENEXR
      else if (extension == "exr")
         success = readOpenEXR(fileName);
#endif
#ifndef NDEBUG
      else
      {
         std::cerr
            << "In function: bool imageRgb::read(const std::string&)\n"
            << extension << " is an unsupported image format\n"
            << "in image name " << fileName << std::endl;
      }
#endif // !NDEBUG

      return(success);
   }

   /*
     \brief Write the image into a file.
     \param fileName (IN) The name of the file.
     \return true if successful, false otherwise.
   */
   bool imageRgb::write(const std::string& fileName)
   {
      bool success = false;
      std::string extension;

      const std::string::size_type extPos = fileName.rfind('.');
      if (extPos != std::string::npos)
      {
         extension = fileName.substr(extPos + 1);
      }

      if (0)
         ; // Botch to ensure future extensibility without a default format
#ifdef WITH_OPENEXR
      else if (extension == "exr")
         success = writeOpenEXR(fileName);
#endif // WITH_OPENEXR
#ifndef NDEBUG
      else
      {
         std::cerr
            << "In function: bool imageRgb::write(const std::string&)\n"
            << extension << " is an unsupported image format.\n";
      }
#endif // !NDEBUG

      return(success);
   }


#ifdef WITH_OPENEXR
   /*!
    * \brief Reads an image in the OpenEXR format.
    *
    * Heavily based on "Reading and Writing OpenEXR Image Files with the IlmImf Library"
    * \param fileName  (IN) The name of the file containing the image.
    * \return true if the operation is successful.
    * \author Alan Woodland
    */
   bool imageRgb::readOpenEXR(const std::string& fileName)
   {
      if (!Imf::isOpenExrFile(fileName.c_str()))
         return(false);

      Imf::InputFile file(fileName.c_str());
      if (!file.isComplete())
         return(false);

      const Imath::Box2i& dw = file.header().dataWindow();
      const unsigned int width = dw.max.x - dw.min.x + 1;
      const unsigned int height = dw.max.y - dw.min.y + 1;

      // I'm almost certain there ought to be a better way to do this, but it
      // works and if it hurts it can be fixed later.
      Imf::Array2D<float> rPixels(height,width);
      Imf::Array2D<float> gPixels(height,width);
      Imf::Array2D<float> bPixels(height,width);

      // TODO: Reading like this isn't exactly robust to variations in channel
      // names. (Iterate ChannelList)
      // TODO: Read CIEL*a*b images like this?

      Imf::FrameBuffer frameBuffer;
      frameBuffer.insert("R",                                                         // name
                         Imf::Slice(Imf::FLOAT,                                       // type
                                    // Who designed this interface?!?!
                                    reinterpret_cast<char*>(&rPixels[0][0] -          // base
                                    dw.min.x -
                                    dw.min.y * width),
                                    sizeof(rPixels[0][0]) * 1,                        // xStride
                                    sizeof(rPixels[0][0]) * width,                    // yStride
                                    1, 1,                                             // x/y sampling
                                    0.0));                                            // fillValue

      frameBuffer.insert("G",                                                         // name
                         Imf::Slice(Imf::FLOAT,                                       // type
                                    reinterpret_cast<char*>(&gPixels[0][0] -          // base
                                    dw.min.x -
                                    dw.min.y * width),
                                    sizeof(gPixels[0][0]) * 1,                        // xStride
                                    sizeof(gPixels[0][0]) * width,                    // yStride
                                    1, 1,                                             // x/y sampling
                                    0.0));                                            // fillValue

      frameBuffer.insert("B",                                                         // name
                         Imf::Slice(Imf::FLOAT,                                       // type
                                    reinterpret_cast<char*>(&bPixels[0][0] -          // base
                                    dw.min.x -
                                    dw.min.y * width),
                                    sizeof(bPixels[0][0]) * 1,                        // xStride
                                    sizeof(bPixels[0][0]) * width,                    // yStride
                                    1, 1,                                             // x/y sampling
                                    0.0));                                            // fillValue

      file.setFrameBuffer(frameBuffer);
      file.readPixels(dw.min.y, dw.max.y);

      // Set the size of the image.
      setSize(width, height);

      // This copy could be implemented better.
      for (unsigned int i = 0; i < width; ++i)
      {
         for (unsigned int j = 0; j < height; ++j)
         {
            double *pixel = getActualPixel(i,j);
            pixel[0] = rPixels[j][i];
            pixel[1] = gPixels[j][i];
            pixel[2] = bPixels[j][i];
         }
      }

      return(true);
   }

   /*!
    * \brief Writes an image in the OpenEXR format.
    *
    * Heavily based on "Reading and Writing OpenEXR Image Files with the IlmImf Library"
    * \param fileName (IN) The name of the file that will contain the
    * image.
    * \return true if the operation is successful.
    * \author Alan Woodland
    */
   bool imageRgb::writeOpenEXR(const std::string& fileName)
   {
      // Unimplemented!
      return(false);
   }
#endif // WITH_OPENEXR


} // namespace imalib
