/*
 * imalib
 * 
 * Copyright 2017 Aberystwyth University
 * 
 * This library and program are free software: you can redistribute them
 * and/or modify them under the terms of the GNU General Public License as 
 * published by the Free Software Foundation, either version 3 of the License,
 * or (at your option) any later version.
 * 
 * The program and library are distributed in the hope that they will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*!
 * \file imageReceiver.cpp
 * \brief A class to wrap the reception of an image from a socket
 * (implementation).
 *
 * \author Fred Labrosse <ffl@aber.ac.uk>
 * \copyright 2017 Aberystwyth University
 */

#include "imalib/imageReceiver.h"

#include "imalib/commProtocol.h"

#include <iostream>

namespace imalib
{
/*!
 * \brief Constructor.
 * \param[in] ioService The boost::asio IO service.
 * \param[in] server The name of the computer where the server is running.
 * \param[in] port The port number the server is listening to.
 */
imageReceiver::imageReceiver(boost::asio::io_service& ioService,
                             const std::string& server,
                             const std::string& port)
   : connected(false), sock(ioService)
{
   tcp::resolver resolver(ioService);
   tcp::resolver::query query(tcp::v4(), server, port);
   tcp::resolver::iterator iterator = resolver.resolve(query);

   try
   {
      sock.connect(*iterator);
      connected = true;
   }
   catch (const std::exception& e)
   {
      connected = false;
      std::cerr << "Exception: " << e.what() << "\n";
   }
}

/*!
 * \brief To receive an image.
 *
 * Should follow a successful call to askForImage().
 *
 * \param[out] img The received image.
 * \sa askForImage().
 */
void imageReceiver::receiveImage(imageRGB* img) throw(std::runtime_error)
{
   // First the size.  This is 2 pairs of bytes as unsigned shorts.
   unsigned short width, height;
   try
   {
      boost::asio::read(sock, boost::asio::buffer(&width, 2));
      width = ntohs(width);
      boost::asio::read(sock, boost::asio::buffer(&height, 2));
      height = ntohs(height);
   }
   catch (const std::exception& e)
   {
      connected = false;
      std::cerr << "Exception: " << e.what() << "\n";
      throw(std::runtime_error("Not connected"));
   }

   // Then the data.
   size_t size = width * height * 3;
   img->setSize(width, height);
   try
   {
      boost::asio::read(sock, boost::asio::buffer(img->getActualPixel(0, 0),
         size));
   }
   catch (const std::exception& e)
   {
      connected = false;
      std::cerr << "Exception: " << e.what() << "\n";
      throw(std::runtime_error("Not connected"));
   }
}

/*!
 * \brief Tell the server we want an image.
 *
 * Use askForImage() first, then use receiveImage().
 *
 * \sa receiveImage().
 */
void imageReceiver::askForImage() throw(std::runtime_error)
{
   try
   {
      boost::asio::write(sock,
                         boost::asio::buffer(&commImg::getImageCmd, 1));
   }
   catch (const std::exception& e)
   {
      std::cerr << "Exception was caught: " << e.what() << std::endl;
         connected = false;
      throw(std::runtime_error("Not connected"));
   }
}

/*!
 * \brief Tell the server we are done.
 */
void imageReceiver::tellToFinish() throw(std::runtime_error)
{
   try
   {
      boost::asio::write(sock, boost::asio::buffer(&commImg::endCmd, 1));
   }
   catch (const std::exception& e)
   {
      std::cerr << "Exception was caught: " << e.what() << std::endl;
      connected = false;
      throw(std::runtime_error("Not connected"));
   }
}

} // namespace imalib
