/*
 * imalib
 * 
 * Copyright 2017 Aberystwyth University
 * 
 * This library and program are free software: you can redistribute them
 * and/or modify them under the terms of the GNU General Public License as 
 * published by the Free Software Foundation, either version 3 of the License,
 * or (at your option) any later version.
 * 
 * The program and library are distributed in the hope that they will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*!
 * \file imageTransmitter.cpp
 * \brief A class to wrap the serving of images on a socket (implementation).
 *
 * \author Fred Labrosse <ffl@aber.ac.uk>
 * \copyright 2017 Aberystwyth University
 */

#include "imalib/imageTransmitter.h"
#include "imalib/commProtocol.h"

#include <iostream>

namespace imalib
{
/*!
 * \brief Constructor.
 * \param[in] ioService The boost::asio IO service.
 * \param[in] sockPort The port to use for the server to wait on.
 * \param[in,out] imgGrabber Where the images come from.
 */
imageTransmitter::imageTransmitter(boost::asio::io_service& ioService,
                                   int sockPort,
                                   imageGrabber* imgGrabber)
   : //serverRunning(true),
   theGrabber(imgGrabber),
//    connected(false),
   ioService_(ioService),
   acceptor_(ioService_, tcp::endpoint(tcp::v4(), sockPort))
{
   session* newSession = new session(ioService_, theGrabber);
   acceptor_.async_accept(newSession->socket(),
                          boost::bind(&imageTransmitter::handleAccept, this,
                                      newSession,
                                      boost::asio::placeholders::error));
   std::cout << "Accepting connections on port " << sockPort << ".\n";
}

/*!
 * \brief Handle the reading of a command.
 *
 * The commands are defined in commProtocol.h.
 *
 * \param[in] error The error code.
 */
void
   imageTransmitter::session::handleRead(const boost::system::error_code& error)
{
//    std::cout << "Command: " << command_ << std::endl;
   if (!error)
   {
      // We do the actions based on the command received (see the protocol).
      if (commImg::endCmd == command_) // This is the end!
      {
         delete this;
      }
      else if (commImg::getImageCmd == command_) // We must send an image.
      {
         // Get an image
         grabber_->getImage(&image_);

         // Send the image size.
         unsigned short buff = htons(image_.nbColumns());
         boost::asio::write(socket_,
                            boost::asio::buffer(&buff, 2));
         buff = htons(image_.nbRows());
         boost::asio::write(socket_,
                            boost::asio::buffer(&buff, 2));
         // Send the data
         boost::asio::async_write(socket_,
                               boost::asio::buffer(image_.getActualPixel(0,0),
                                  image_.nbColumns() * image_.nbRows() * 3),
                               boost::bind(&session::handleWrite, this,
                                           boost::asio::placeholders::error));
      }
      else
      {
         std::cerr << "Unrecognised command '" << command_ << "'.\n";
      }
   }
   else
   {
      delete this;
   }
}

} // namespace imalib
