/*
 * imalib
 * 
 * Copyright 2017 Aberystwyth University
 * 
 * This library and program are free software: you can redistribute them
 * and/or modify them under the terms of the GNU General Public License as 
 * published by the Free Software Foundation, either version 3 of the License,
 * or (at your option) any later version.
 * 
 * The program and library are distributed in the hope that they will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*!
 * \file appearance.h
 * \brief Defines classes to play with appearances.
 * \author Fred Labrosse <ffl@aber.ac.uk>
 * \copyright 2017 Aberystwyth University
 */

#ifndef IMALIB_APPEARANCE_H
#define IMALIB_APPEARANCE_H

#include <string.h>

#include <iterator>
#include <limits>
#include <vector>

#include "imalib/appearancePixel.h"
#include "imalib/image.h"
#include "imalib/pixelCoord.h"
#include "imalib/pixelValue.h"
#include "imalib/transformation.h"

namespace imalib
{

/*!
 * \brief The type for the pixels of the appearance.
 */
template <typename typeCol, int nbColComp>
using appearancePixels = std::map<pixelCoord, pixelValue<typeCol, nbColComp>>;

/*!
 * \class appearance appearance.h imalib/appearance.h
 * \brief A class to represent appearances.
 *
 * An appearance is a bunch of "pixels", i.e. real valued positions with a
 * pixel value associated to them.  They don't necessarily define a complete
 * coverage of a rectangular area.
 *
 * Pixels of the appearance have nbColComp components of type typeCol, both
 * defined at compilation time (using templates).
 *
 * \author Fred Labrosse <ffl@aber.ac.uk>
 */
template <typename typeCol, int nbColComp>
class appearance
{
public:

   //Just to make life easier!
   friend class appearPixelsUsedIter;
//    friend class appearPixelsUsedIter<typeCol, nbColComp>;

   /*!
    * \brief Default constructor.
    */
   appearance()
      : boundingBoxValid_(true),
      minX_(0.0), maxX_(0.0), minY_(0.0), maxY_(0.0)
   {
   }

   /*!
    * \brief Constructor from an image.
    *
    * The appearance is centred.
    *
    * \param[in] img The image to copy.
    * \sa fromImage().
    */
   appearance(const image<typeCol, nbColComp>& img)
   {
      fromImage(img);
   }

   /*!
    * \brief Copy constructor.
    *
    * \param[in] appear The appearance to copy.
    */
   appearance(const appearance<typeCol, nbColComp>& appear)
   {
      *this = appear;
   }

   /*!
    * \brief Destructor
    */
   ~appearance()
   {
      pixels_.clear();
   }

   /*!
    * \brief Assignment operator.
    *
    * \param[in] appear The appearance to copy.
    */
   appearance<typeCol, nbColComp>& operator=(const appearance<typeCol, nbColComp>& appear)
   {
      pixels_ = appear.pixels_;
      minX_ = appear.minX_;
      minY_ = appear.minY_;
      maxX_ = appear.maxX_;
      maxY_ = appear.maxY_;
      boundingBoxValid_ = appear.boundingBoxValid_;

      return(*this);
   }

   /*!
    * \brief Load (part of) an image into an appearance.
    *
    * The specified area is clamped to the size of the image.
    *
    * \param[in] img The image to copy.
    * \param[in] minX
    * \param[in] minY Coordinates of the top left corner.
    * \param[in] maxX
    * \param[in] maxY Coordinates of the bottom right corner.
    * \param[in] recentre If true, then recentre the appearance.  This is
    * cheaper than re-centring afterwards.
    * \sa recentre().
    */
   void fromImage(const image<typeCol, nbColComp>& img,
                  unsigned short minX = 0,
                  unsigned short minY = 0,
                  unsigned short maxX =
                  std::numeric_limits<unsigned short>::max(),
                  unsigned short maxY =
                  std::numeric_limits<unsigned short>::max(),
                  bool recentre = true)
   {
      minX_ = static_cast<double>(minX);
      minY_ = static_cast<double>(minY);
      if (maxX > img.nbColumns() - 1)
         maxX_ = static_cast<double>(img.nbColumns()) - 1.0;
      else
         maxX_ = static_cast<double>(maxX);
      if (maxY > img.nbRows() - 1)
         maxY_ = static_cast<double>(img.nbRows()) - 1.0;
      else
         maxY_ = static_cast<double>(maxY);
      boundingBoxValid_ = true;

      pixels_.clear();

      double Xshift = maxX_ - ((maxX_ - minX_) / 2.0);
      double Yshift = maxY_ - ((maxY_ - minY_) / 2.0);

      for (unsigned short row = static_cast<unsigned short>(minY_);
           row <= static_cast<unsigned short>(maxY_);
           ++row)
      {
         for (unsigned short col = static_cast<unsigned short>(minX_);
              col <= static_cast<unsigned short>(maxX_);
              ++col)
         {
            if (!recentre)
            {
               pixels_.insert(std::make_pair(pixelCoord(static_cast<double>(col),
                                                        static_cast<double>(row)),
                                             pixelValue<typeCol, nbColComp>(
                                                img.getActualPixel(col,
                                                                   row))));
            }
            else
            {
               pixels_.insert(std::make_pair(pixelCoord(static_cast<double>(col-Xshift),
                                                        static_cast<double>(row-Yshift)),
                                             pixelValue<typeCol, nbColComp>(
                                                img.getActualPixel(round(col),
                                                                   round(row)))));
            }
         }
      }

      if (recentre)
      {
         minX_ -= Xshift;
         maxX_ -= Xshift;
         minY_ -= Yshift;
         maxY_ -= Yshift;
      }
   }

   /*!
    * \brief Create an appearance from positions in an image.
    *
    * If given positions fall outside the image, they are ignored.
    * \param[in] img The image to use.
    * \param[in] positions The pixel positions to use.
    * \param[in] transf A transformation to apply to the given position.
    */
   void fromImage(const image<typeCol, nbColComp>& img,
                  const std::vector<pixelCoord>& positions,
                  const transformation& transf)
   {
      minX_ = std::numeric_limits<double>::max();
      minY_ = std::numeric_limits<double>::max();
      maxX_ = std::numeric_limits<double>::min();
      maxY_ = std::numeric_limits<double>::min();

      pixels_.clear();

      std::vector<pixelCoord>::const_iterator pixIter;
      double transfX, transfY;
      for (pixIter = positions.begin(); pixIter != positions.end(); pixIter++)
      {
         transf.apply(pixIter->getX(), pixIter->getY(), &transfX, &transfY);
         if ((transfX > 0.0) && (transfX < img.nbColumns())
            && (transfY > 0.0) && (transfY < img.nbRows()))
         {
            if (minX_ > transfX)
               minX_ = transfX;
            if (minY_ > transfY)
               minY_ = transfY;
            if (maxX_ < transfX)
               maxX_ = transfX;
            if (maxY_ < transfY)
               maxY_ = transfY;

            pixels_.insert(std::make_pair(pixelCoord(transfX, transfY),
                                          pixelValue<typeCol, nbColComp>(
                                             img.getActualPixel(round(transfX),
                                                                round(transfY)))));
         }
         boundingBoxValid_ = true;
      }
   }

   /*!
    * \brief Save an appearance into an image.
    *
    * The appearance is re-centred so that the image covers the bounding box
    * of the appearance.
    *
    * The background of the image is set to the default value of typeCol
    * and only the pixels of the appearance are set.  The pixels of the
    * appearance are placed onto the nearest neighbour integer image position.
    *
    * The image is resized here.
    *
    * \param[out] img The image into which to save the appearance.
    */
   void toImage(image<typeCol, nbColComp>* img) const
   {
      img->setSize((static_cast<unsigned short>(round(maxX_ - minX_)) + 1),
                   (static_cast<unsigned short>(round(maxY_ - minY_)) + 1));
      typeCol defaultColour[nbColComp];
      for (unsigned short ind = 0; ind < nbColComp; ++ind)
         defaultColour[ind] = typeCol();
      img->setPixels(defaultColour);

      typename appearancePixels<typeCol, nbColComp>::const_iterator iter;
      for (iter = pixels_.begin();
           iter != pixels_.end();
           ++iter)
      {
         img->setPixel(static_cast<unsigned short>(round(iter->first.getX()-minX_)),
                       static_cast<unsigned short>(round(iter->first.getY()-minY_)),
                       iter->second.get());
      }
   }

   /*!
    * \brief Update the appearance (pixel values) from an image and the
    * transformed position of the appearance.
    *
    * Parts of the transformed appearance that fall outside the image are
    * ignored (not updated).
    *
    * \param[in] img The image to use.
    * \param[in] transf A transformation to apply to the given position.
    * \param[in] usedInDistance If true, then only use the positions that
    * have been used in the last distance computation.
    * \sa pixelCoord::wasUsedInDistance(), pixelCoord:setUsedInDistance().
    */
   void update(const image<typeCol, nbColComp>& img,
               const transformation& transf,
               bool usedInDistance = false)
   {
      typename appearancePixels<typeCol, nbColComp>::iterator iter;
      double transfX, transfY;
      for (iter = pixels_.begin(); iter != pixels_.end(); ++iter)
      {
         transf.apply(iter->first.getX(), iter->first.getY(),
                      &transfX, &transfY);
         if ((transfX > 0.0) && (transfX < img.nbColumns())
            && (transfY > 0.0) && (transfY < img.nbRows()))
         {
            if ((usedInDistance && iter->first.wasUsedInDistance())
                || !usedInDistance)
            {
               iter->second.set(img.getActualPixel(round(transfX),
                                                   round(transfY)));
            }
         }
      }
   }

   /*!
    * \brief Update the appearance (pixel values) from an image and the
    * transformed position of the appearance.  This also updates a
    * "parallel" appearance.
    *
    * A "parallel" appearance is updated.  This parallel appearance is useful
    * to update in the same manner an appearance in another colour space.
    *
    * It is assumed that the parallel appearance has the same
    * pixel positions as this appearance, but only the fact that they have the
    * same bounding box and number of pixels is tested (would be too expensive
    * to do otherwise).  The two images to draw pixels from must have the same
    * size.  These conditions are only tested if the code is compiled without
    * -DNDEBUG.
    *
    * Parts of the transformed appearances that fall outside the image are
    * ignored (not updated).
    *
    * \param[in] img The image to use.
    * \param[in] imgPar The "parallel" image, the one used to update the
    * "parallel" appearance.
    * \param[in,out] appearPar The "parallel" appearance.
    * \param[in] transf A transformation to apply to the given position.
    * \param[in] usedInDistance If true, then only use the positions that
    * have been used in the last distance computation.
    * \sa pixelCoord::wasUsedInDistance(), pixelCoord:setUsedInDistance().
    */
   template <typename typeCol2, int nbColComp2>
   void update(const image<typeCol, nbColComp>& img,
               const image<typeCol2, nbColComp2>& imgPar,
               appearance<typeCol2, nbColComp2>* appearPar,
               const transformation& transf,
               bool usedInDistance = false)
   {
      assert((getBBxMax() == appearPar->getBBxMax())
             && (getBBxMin() == appearPar->getBBxMin())
             && (getBByMax() == appearPar->getBByMax())
             && (getBByMin() == appearPar->getBByMin())
             && (img.nbColumns() == imgPar.nbColumns())
             && (img.nbRows() == imgPar.nbRows()));

      typename appearancePixels<typeCol, nbColComp>::iterator iter;
      typename appearancePixels<typeCol2, nbColComp2>::iterator iterPar;
      double transfX, transfY;
      for (iter = pixels_.begin(), iterPar = appearPar->pixels()->begin();
           iter != pixels_.end();
           ++iter, ++iterPar)
      {
         if ((usedInDistance && iter->first.wasUsedInDistance())
            || !usedInDistance)
         {
            transf.apply(iter->first.getX(), iter->first.getY(),
                         &transfX, &transfY);
            if ((transfX > 0.0) && (transfX < img.nbColumns())
               && (transfY > 0.0) && (transfY < img.nbRows()))
            {
               iter->second.set(img.getActualPixel(round(transfX),
                                                   round(transfY)));
               iterPar->second.set(imgPar.getActualPixel(round(transfX),
                                                         round(transfY)));
            }
         }
      }
   }

   /*!
    * \brief Update the appearance (pixel values) from an image and the
    * transformed position of the appearance, but only for the given positions.
    *
    * Parts of the transformed appearance that fall outside the image are
    * ignored (not updated).  Given positions that are not in the appearance
    * are ignored.
    *
    * This is an expensive function because each position is searched for in the
    * appearance.
    *
    * \param[in] img The image to use.
    * \param[in] transf A transformation to apply to the given position.
    * \param[in] positions The list of positions that need to be updated.
    */
   void update(const image<typeCol, nbColComp>& img,
               const transformation& transf,
               const std::vector<imalib::pixelCoord>& positions)
   {
      typename appearancePixels<typeCol, nbColComp>::iterator appearIter;
      std::vector<imalib::pixelCoord>::const_iterator posIter;
      double transfX, transfY;

      for (posIter = positions.begin(); posIter != positions.end(); posIter++)
      {
         // Is the position indeed contained in the appearance?
         appearIter = pixels_.find(*posIter);
         if (appearIter != pixels_.end())
         {
            transf.apply(posIter->getX(), posIter->getY(), &transfX, &transfY);
            if ((transfX > 0.0) && (transfX < img.nbColumns())
               && (transfY > 0.0) && (transfY < img.nbRows()))
            {
               appearIter->second.set(img.getActualPixel(round(transfX),
                                                         round(transfY)));
            }
         }
      }
   }

   /*!
    * \brief Update the bounding box of the appearance.
    */
   void updateBoundingBox()
   {
      typename appearancePixels<typeCol, nbColComp>::iterator iter;

      minX_ = std::numeric_limits<double>::max();
      maxX_ = std::numeric_limits<double>::min();
      minY_ = std::numeric_limits<double>::max();
      maxY_ = std::numeric_limits<double>::min();
      for (iter = pixels_.begin();
           iter != pixels_.end();
           ++iter)
      {
         if (iter->first.getX() < minX_)
            minX_ = iter->first.getX();
         if (iter->first.getX() > maxX_)
            maxX_ = iter->first.getX();
         if (iter->first.getY() < minY_)
            minY_ = iter->first.getY();
         if (iter->first.getY() > maxY_)
            maxY_ = iter->first.getY();
      }
      boundingBoxValid_ = true;
   }

   /*!
    * \brief Get the minimum X coordinate of the bounding box.
    * \return The X coordinate.
    */
   double getBBxMin() const
   {
      return(minX_);
   }

   /*!
    * \brief Get the minimum Y coordinate of the bounding box.
    * \return The Y coordinate.
    */
   double getBByMin() const
   {
      return(minY_);
   }

   /*!
    * \brief Get the maximum X coordinate of the bounding box.
    * \return The X coordinate.
    */
   double getBBxMax() const
   {
      return(maxX_);
   }

   /*!
    * \brief Get the maximum Y coordinate of the bounding box.
    * \return The Y coordinate.
    */
   double getBByMax() const
   {
      return(maxY_);
   }

   /*!
    * \brief Re-centre the appearance.
    *
    * This translates the positions such that the bounding box is centred
    * around the (0,0) coordinate.
    *
    * This is an expensive operation because of the internal representation (a
    * new re-centred appearance is created and then copied).
    */
   void recentre()
   {
      appearancePixels<typeCol, nbColComp> newPixels;

      if (!boundingBoxValid_)
         updateBoundingBox();

      double Xshift = maxX_ - ((maxX_ - minX_) / 2.0);
      double Yshift = maxY_ - ((maxY_ - minY_) / 2.0);

      typename appearancePixels<typeCol, nbColComp>::iterator iter;
      for (iter = pixels_.begin();
           iter != pixels_.end();
           ++iter)
      {
         newPixels.insert(std::make_pair(pixelCoord(iter->first.getX() - Xshift,
                                                    iter->first.getY() - Yshift),
                                         pixelValue<typeCol, nbColComp>(iter->second.get())));
      }
      pixels_ = newPixels; // Expensive!
      maxX_ -= Xshift;
      minX_ -= Xshift;
      maxY_ -= Yshift;
      minY_ -= Yshift;
   }

   /*!
    * \brief Geometrically transform the appearance.
    *
    * This is an expensive operation because of the internal representation (a
    * new transformed appearance is created and then copied).
    *
    * \param[in] transf The transformation to use.
    */
   void transform(const transformation& transf)
   {
      minX_ = std::numeric_limits<double>::max();
      maxX_ = std::numeric_limits<double>::min();
      minY_ = std::numeric_limits<double>::max();
      maxY_ = std::numeric_limits<double>::min();

      appearancePixels<typeCol, nbColComp> newPixels;
      pixelCoord newCoord;

      typename appearancePixels<typeCol, nbColComp>::iterator iter;
      for (iter = pixels_.begin();
           iter != pixels_.end();
           ++iter)
      {
         transf.apply(iter->first, &newCoord);
         newPixels.insert(std::make_pair(newCoord,
                                         pixelValue<typeCol, nbColComp>(iter->second.get())));
         if (newCoord.getX() < minX_)
            minX_ = newCoord.getX();
         if (newCoord.getX() > maxX_)
            maxX_ = newCoord.getX();
         if (newCoord.getY() < minY_)
            minY_ = newCoord.getY();
         if (newCoord.getY() > maxY_)
            maxY_ = newCoord.getY();
      }
      pixels_ = newPixels; // Expensive!
      boundingBoxValid_ = true;
   }

   /*!
    * \brief Get a constant handle on the pixels of the appearance.
    * \return The pixels of the appearance.
    */
   const appearancePixels<typeCol, nbColComp>& constPixels() const
   {
      return(pixels_);
   }

   /*!
    * \brief Get a <b>non-constant</b> handle on the pixels of the appearance.
    * \return The pixels of the appearance.
    */
   appearancePixels<typeCol, nbColComp>* pixels()
   {
      return(&pixels_);
   }

   /*!
    * \brief Get the number of pixels in the appearance.
    * \return The number of pixels.
    */
   unsigned long getNbPixels() const
   {
      return(pixels_.size());
   }

   /*!
    * \class appearPixelsUsedIter appearance.h imalib/appearance.h
    * \brief A class that provides an iterator over all pixels of an appearance
    * that have been used by a previous distance calculation.
    *
    * \author Fred Labrosse <ffl@aber.ac.uk>
    */
   class appearPixelsUsedIter
      : public std::iterator<std::bidirectional_iterator_tag,
                             appearance<typeCol, nbColComp>>
   {
   public:

      friend class appearance<typeCol, nbColComp>;

      /*!
       * \brief Constructor.
       * \param[in] theAppear The appearance used.
       */
      appearPixelsUsedIter(appearance<typeCol, nbColComp>& theAppear)
      : theAppear_(theAppear)
      {
         begin();
      }

      /*!
       * \brief Different from operator.
       * \param[in] rIter The right hand side iterator to which this is
       * compared.
       * \return true if the two iterators are different.
       */
      bool operator!=(const appearPixelsUsedIter& rIter)
      {
         return(this->appearanceIter_ != rIter.appearanceIter_);
      }

      /*!
       * \brief Increment iterator.
       * \return Itself, incremented.
       */
      appearPixelsUsedIter& operator++()
      {
         // We increment, and then see if we need to increment more to skip 
         // non used pixels.
         do
         {
            ++appearanceIter_;
         }
         while ((appearanceIter_ != theAppear_.pixels_.end())
            && !appearanceIter_->first.wasUsedInDistance());

         return(*this);
      }

      /*!
       * \brief Dereferencing operator.
       *
       * Elements of the iterator are pairs of pixelCoord and
       * pixelValue<typeCol, nbColComp>.
       *
       * \return The content pointed to by the iterator.
       */
      const std::pair<const pixelCoord, pixelValue<typeCol, nbColComp>>& operator*()
      {
         return(*appearanceIter_);
      }

      /*!
       * \brief Pointer (dereferencing) operator.
       *
       * Elements of the iterator are pairs of pixelCoord and
       * pixelValue<typeCol, nbColComp>.
       *
       * \return The content pointed to by the iterator.
       */
      const std::pair<const pixelCoord, pixelValue<typeCol, nbColComp>>* operator->()
      {
         return(&(*appearanceIter_)); // We take the pair and then its address!
      }

   protected:

      /*!
       * \brief Create an iterator that points to the beginning.
       */
      appearPixelsUsedIter begin()
      {
         appearanceIter_ = std::begin(theAppear_.pixels_);
         while ((appearanceIter_ != std::end(theAppear_.pixels_))
            && !appearanceIter_->first.wasUsedInDistance())
         {
            ++appearanceIter_;
         }
         return(*this);
      }

      /*!
       * \brief Create an iterator that points to the end.
       */
      appearPixelsUsedIter end()
      {
         appearanceIter_ = theAppear_.pixels_.end();
         return(*this);
      }

   private:

      /*!
       * \brief Default constructor.
       *
       * We do not provide an implementation of that on purpose.  This is to
       * prevent its use.
       */
      appearPixelsUsedIter();

      /*!
       * The appearance this iterator iterates over.
       */
      appearance<typeCol, nbColComp>& theAppear_;

      /*!
       * \brief A "normal" iterator on the appearance.
       */
      typename appearancePixels<typeCol, nbColComp>::const_iterator appearanceIter_;
   };

   /*!
    * \brief Create an iterator that points to the beginning of the collection
    * and iterates over pixels used during the previous distance calculation.
    * \return The iterator.
    * \sa appearPixelsUsedIter.
    */
   appearPixelsUsedIter begin()
   {
      return(appearPixelsUsedIter(*this));
   }

   /*!
    * \brief Create an iterator that points to the end of the collection
    * and iterates over pixels used during the previous distance calculation.
    * \return The iterator.
    * \sa appearPixelsUsedIter.
    */
   appearPixelsUsedIter end()
   {
      appearPixelsUsedIter theIter(*this);
      return(theIter.end());
   }

   /*!
    * Save into an image a mask of the pixels used in the last distance
    * calculation.
    *
    * Only the pixels used are changed (set to col) in the given image.
    *
    * The image must be big enough to contain the appearance.  This is only
    * checked if compiled without NDEBUG set.
    *
    * \param[in] col Colour to use (array assumed to be of the right size).
    * \param[in,out] maskImg The mask image.
    */
   template <typename typeCol2, int nbColComp2>
   void saveUsedPixelsMask(typeCol2* col,
                           image<typeCol2, nbColComp2>* maskImg)
   {
      assert((maskImg->nbColumns() >= (getBBxMax() - getBBxMin()))
             && (maskImg->nbRows() >= (getBByMax() - getBByMin())));

      for (appearPixelsUsedIter iter = begin();
           iter != end();
           ++iter)
      {
         maskImg->setPixel(round(iter->first.getX() - getBBxMin()),
                           round(iter->first.getY() - getBByMin()),
                           col);
      }
   }

protected:

private:

   /*!
    * \brief The appearance pixels that make up the appearance.
    */
   appearancePixels<typeCol, nbColComp> pixels_;

   /*!
    * \brief If true, then the bounding box information is valid.
    * \sa minX_, maxX_, minY_, maxY_.
    */
   bool boundingBoxValid_;

   /*!
    * \brief The minimum horizontal pixel position of the appearance.
    * \sa boundingBoxValid_.
    */
   double minX_;

   /*!
    * \brief The maximum horizontal pixel position of the appearance.
    * \sa boundingBoxValid_.
    */
   double maxX_;

   /*!
    * \brief The minimum vertical pixel position of the appearance.
    * \sa boundingBoxValid_.
    */
   double minY_;

   /*!
    * \brief The maximum vertical pixel position of the appearance.
    * \sa boundingBoxValid_.
    */
   double maxY_;

};

} // namespace imalib

#endif // IMALIB_APPEARANCE_H

