/*
 * imalib
 * 
 * Copyright 2017 Aberystwyth University
 * 
 * This library and program are free software: you can redistribute them
 * and/or modify them under the terms of the GNU General Public License as 
 * published by the Free Software Foundation, either version 3 of the License,
 * or (at your option) any later version.
 * 
 * The program and library are distributed in the hope that they will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*!
 * \file appearancePixel.h
 * \brief Defines a class to represent an appearance pixel.
 * \author Fred Labrosse <ffl@aber.ac.uk>
 * \copyright 2017 Aberystwyth University
 */

#ifndef IMALIB_APPEARANCEPIXEL_H
#define IMALIB_APPEARANCEPIXEL_H

#include <string.h>

namespace imalib
{

/*!
 * \class appearancePixel appearancePixel.h imalib/appearancePixel.h
 * \brief A class to represent an appearance pixel.
 *
 * An appearance pixel is a real valued 2D position with an associated pixel
 * value having nbColComp values of type typeCol.
 */
template <typename typeCol, int nbColComp>
class appearancePixel
{
public:

   /*!
    * \brief Constructor.
    */
   appearancePixel()
      : x_(0.0), y_(0.0), value_(NULL)
   {
      value_ = new typeCol[nbColComp];
      for (unsigned short ind = 0; ind < nbColComp; ind++)
      {
         value_[ind] = typeCol();
      }
   }

   /*!
    * \brief Copy constructor.
    * \param[in] appPix The appearancePixel to copy.
    */
   appearancePixel(const appearancePixel& appPix)
      : x_(appPix.x_), y_(appPix.y_), value_(NULL)
   {
      value_ = new typeCol[nbColComp];
      memcpy(value_, appPix.value_, sizeof(typeCol) * nbColComp);
   }

   /*!
    * \brief Constructor.
    * \param[in] x The new X position of the pixel.
    * \param[in] y The new Y position of the pixel.
    * \param[in] value The new pixel value (as an array of nbColComp values
    * of type typeCol).
    */
   appearancePixel(double x, double y, const typeCol* value)
      : x_(x), y_(y), value_(NULL)
   {
      value_ = new typeCol[nbColComp];
      memcpy(value_, value, sizeof(typeCol) * nbColComp);
   }

   /*!
    * \brief Destructor.
    */
   ~appearancePixel()
   {
      delete [] value_;
      value_ = NULL;
   }

   /*!
    * \brief Set a pixel.
    * \param[in] x The new X position of the pixel.
    * \param[in] y The new Y position of the pixel.
    * \param[in] value The new pixel value (as an array of nbColComp values
    * of type typeCol).
    */
   void set(double x, double y, const typeCol* value)
   {
      x_ = x;
      y_ = y;
      memcpy(value_, value, sizeof(typeCol) * nbColComp);
   }

   /*!
    * \brief Set the X position of a pixel.
    * \param[in] x The new X position of the pixel.
    */
   void setX(double x)
   {
      x_ = x;
   }

   /*!
    * \brief Set the Y position of a pixel.
    * \param[in] y The new Y position of the pixel.
    */
   void setY(double y)
   {
      y_ = y;
   }


   /*!
    * \brief Set the value of a pixel.
    * \param[in] value The new pixel value (as an array of nbColComp values of
    * type typeCol).
    */
   void setValue(typeCol* value)
   {
      memcpy(value_, value, sizeof(typeCol) * nbColComp);
   }

   /*!
    * \brief Get the X position of the pixel.
    * \return The X position.
    */
   double getX() const
   {
      return(x_);
   }

   /*!
    * \brief Get the Y position of the pixel.
    * \return The Y position.
    */
   double getY() const
   {
      return(y_);
   }

   /*!
    * \brief Get the value of the pixel.
    * \return The value.
    */
   const typeCol* getValue() const
   {
      return(value_);
   }

   /*!
    * \brief Assignment operator.
    * \param[in] appPix The appearancePixel to copy.
    * \return The assigned appearancePixel.
    */
   appearancePixel& operator=(const appearancePixel& appPix)
   {
      if (this != &appPix)
      {
         x_ = appPix.x_;
         y_ = appPix.y_;
         memcpy(value_, appPix.value_, sizeof(typeCol) * nbColComp);
      }
      return(*this);
   }

protected:

private:

   /*!
    * \brief X position of the pixel.
    */
   double x_;

   /*!
    * \brief Y position of the pixel.
    */
   double y_;

   /*!
    * \brief pixel value.
    */
   typeCol* value_;
};

} // namespace imalib

#endif // IMALIB_APPEARANCEPIXEL_H
