/*
 * imalib
 * 
 * Copyright 2017 Aberystwyth University
 * 
 * This library and program are free software: you can redistribute them
 * and/or modify them under the terms of the GNU General Public License as 
 * published by the Free Software Foundation, either version 3 of the License,
 * or (at your option) any later version.
 * 
 * The program and library are distributed in the hope that they will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*!
 * \file dc1394ImageGrabber.h
 * \brief Image grabber classe for firewire (dc1394) frame grabber.
 *
 * \author Fred Labrosse <ffl@aber.ac.uk>
 * \copyright 2017 Aberystwyth University
 */

#ifndef DC1394IMAGEGRABBER_H
#define DC1394IMAGEGRABBER_H

#include "imageGrabber.h"

#include <dc1394/dc1394.h>
#include "image.h"

namespace imalib
{

/*!
 * \class dc1394ImageGrabber dc1394ImageGrabber.h imalib/dc1394ImageGrabber.h
 * \brief A class to interface a dc1394 (firewire) camera.
 *
 * If the class is available, WITH_DC1394 is defined (if you use
 * imalib-config --cflags when compiling, see \ref autotools).
 */
class dc1394ImageGrabber : public imageGrabber
{
public:
   /*!
    * \brief Constructor.
    *
    * \param[in] upsideDown Set to true if the camera is upside-down.
    * \param[in] flip Set to true to horizontally flip the images.
    * \param[in] cameraNumber The number of the camera on the firewire system.
    * \param[in] posX
    * \param[in] posY The position of the viewport of the grabbed part.
    * \param[in] width
    * \param[in] height The width and height of the grabbed area and of the image.
    */
   dc1394ImageGrabber(bool upsideDown,
                      bool flip,
                      unsigned short cameraNumber,
                      unsigned short posX = 0,
                      unsigned short posY = 0,
                      unsigned short width = IMAGE_DEFAULT_INPUTWIDTH,
                      unsigned short height = IMAGE_DEFAULT_INPUTHEIGHT);

   /*!
    * \brief Constructor.
    * \param[in] fileName Name of the configuration file.
    */
   dc1394ImageGrabber(const std::string& fileName);

   /*!
    * \brief Destructor.
    */
   virtual ~dc1394ImageGrabber();

   /*!
    * \brief Display on the standard output some information about the
    * source of the images.
    * \param[in,out] aStream The stream to which the info is added.
    */
   virtual void printInfoAboutSource(std::ostringstream* aStream);

   /*!
    * \brief Capture the image from a grabber.
    *
    * This doesn't do anything for that type of camera.
    */
   virtual void captureImage()
   {
      capture = true;  // Not used, but just in case.
   }

   /*!
    * \brief Get an image from the frame grabber.
    *
    * The image might have been pre-captured (with captureImage()), but
    * if not will be captured here.
    *
    * The images are resized here.
    *
    * \param[out] theImage The image in which to save the grabbed image, if
    * not NULL.
    * \param[out] rawImage The image in which to save the raw image (before any
    * transformation is applied), if not NULL.
    */
   virtual void getImage(imageRGB* theImage, imageRGB* rawImage = NULL);

   /*!
    * \brief Set the size of the image to be captured.
    * \param[in] imgWidth
    * \param[in] imgHeight The width and height of the images to grab.
    * \return true if successful, false otherwise.
    */
   bool setImageSize(unsigned short imgWidth, unsigned short imgHeight);

   /*!
    * \brief Specify the input rectangle (viewport on the grabbed image).
    * \param[in] x
    * \param[in] y Coordinates of the top-left corner of the rectangle.
    * \param[in] w
    * \param[in] h Size of the rectangle.
    * \return true is successful, false otherwise.
    * \sa useInputRectangle().
    */
   bool setInputRectangle(unsigned short x, unsigned short y,
                          unsigned short w, unsigned short h);

   /*!
    * \brief Use the input rectangle currently set.
    * \return true is successful, false otherwise.
    * \sa setInputRectangle().
    */
   bool useInputRectangle();

   /*!
    * \brief Move the viewport to the right.
    * \sa setInputRectangle().
    */
   void moveRight();

   /*!
    * \brief Move the viewport to the left.
    * \sa setInputRectangle().
    */
   void moveLeft();

   /*!
    * \brief Move the viewport up.
    * \sa setInputRectangle().
    */
   void moveUp();

   /*!
    * \brief Move the viewport down.
    * \sa setInputRectangle().
    */
   void moveDown();

   /*!
    * \brief Increase the width of the viewport.
    * \sa setInputRectangle().
    */
   void increaseWidth();

   /*!
    * \brief Decrease the width of the viewport.
    * \sa setInputRectangle().
    */
   void decreaseWidth();

   /*!
    * \brief Increase the height of the viewport.
    * \sa setInputRectangle().
    */
   void increaseHeight();

   /*!
    * \brief Decrease the height of the viewport.
    * \sa setInputRectangle().
    */
   void decreaseHeight();

   /*!
    * \brief Set the gain.
    *
    * \param[in] value The new gain value.  A negative value means automatic
    * gain control.
    */
   void setGain(long value);

   /*!
    * \brief Get the current gain value.
    * \return The value.
    */
   long getGain();

   /*!
    * \brief Get the minimum value for the gain.
    * \return The minimum value.
    */
   long getMinGain()
   {
      return(static_cast<long>(minGain));
   }

   /*!
    * \brief Get the maximum value for the gain.
    * \return The maximum value.
    */
   long getMaxGain()
   {
      return(static_cast<long>(maxGain));
   }

   /*!
    * \brief Set the exposure.
    *
    * \param[in] value The new exposure value.  A negative value means automatic
    * exposure control.
    */
   void setExposure(long value);

   /*!
    * \brief Get the current exposure value.
    * \return The value.
    */
   long getExposure();

   /*!
    * \brief Get the minimum value for the exposure.
    * \return The minimum value.
    */
   long getMinExposure()
   {
      return(static_cast<long>(minExposure));
   }

   /*!
    * \brief Get the maximum value for the exposure.
    * \return The maximum value.
    */
   long getMaxExposure()
   {
      return(static_cast<long>(maxExposure));
   }

   /*!
    * \brief Set the shutter.
    *
    * \param[in] value The new shutter value.  A negative value means automatic
    * shutter control.
    */
   void setShutter(long value);

   /*!
    * \brief Get the current shutter value.
    * \return The value.
    */
   long getShutter();

   /*!
    * \brief Get the minimum value for the shutter.
    * \return The minimum value.
    */
   long getMinShutter()
   {
      return(static_cast<long>(minShutter));
   }

   /*!
    * \brief Get the maximum value for the shutter.
    * \return The maximum value.
    */
   long getMaxShutter()
   {
      return(static_cast<long>(maxShutter));
   }

protected:

   /*!
    * \brief Initialise the grabber.
    * \param[in] upsideDown Set to true if the camera is upside-down.
    * \param[in] flip Set to true to horizontally flip the images.
    * \param[in] cameraNumber The number of the camera on the firewire system.
    * \param[in] posX
    * \param[in] posY The position of the viewport of the grabbed part.
    * \param[in] width
    * \param[in] height The width and height of the grabbed area and of the
    * images.
    */
   void init(bool upsideDown,
             bool flip,
             unsigned short cameraNumber,
             unsigned short posX = 0,
             unsigned short posY = 0,
             unsigned short width = IMAGE_DEFAULT_INPUTWIDTH,
             unsigned short height = IMAGE_DEFAULT_INPUTHEIGHT);

private:

   /*!
    * \brief dc1394 context.
    */
   dc1394_t* context;

   /*!
    * \brief The camera we use.
    */
   dc1394camera_t* camera;

   /*!
    * \brief Tells if the capture is running.
    */
   bool captureRunning;

   /*!
    * \brief Input rectangle X position.
    */
   unsigned int inputX;

   /*!
    * \brief Input X position unit.
    */
   unsigned int xUnit;

   /*!
    * \brief Input rectangle Y position.
    */
   unsigned int inputY;

   /*!
    * \brief Input Y position unit.
    */
   unsigned int yUnit;

   /*
    * \brief Input rectangle width.
    */
//    unsigned int imageWidth;

   /*!
    * \brief Input width unit.
    */
   unsigned int widthUnit;

   /*!
    * \brief Input max width.
    */
   unsigned int widthMax;

   /*
    * \brief Input rectangle height.
    */
//    unsigned int imageHeight;

   /*!
    * \brief Input height unit.
    */
   unsigned int heightUnit;

   /*!
    * \brief Input max height.
    */
   unsigned int heightMax;

   /*
    * \brief A temporary "image" to convert the grabbed image into a proper
    * image.
    */
//    unsigned char* grabbedImage;

   /*!
    * \brief Temporary frame.
    */
   dc1394video_frame_t* grabbedFrame;

   /*!
    * \brief Min gain.
    */
   uint32_t minGain;

   /*!
    * \brief Max gain.
    */
   uint32_t maxGain;

   /*!
    * \brief Min shutter.
    */
   uint32_t minShutter;

   /*!
    * \brief Max shutter.
    */
   uint32_t maxShutter;

   /*!
    * \brief Min exposure.
    */
   uint32_t minExposure;

   /*!
    * \brief Max exposure.
    */
   uint32_t maxExposure;

}; // class dc1394ImageGrabber

} // namespace imalib

#endif // DC1394IMAGEGRABBER_H
