/*
 * imalib
 * 
 * Copyright 2017 Aberystwyth University
 * 
 * This library and program are free software: you can redistribute them
 * and/or modify them under the terms of the GNU General Public License as 
 * published by the Free Software Foundation, either version 3 of the License,
 * or (at your option) any later version.
 * 
 * The program and library are distributed in the hope that they will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*!
 * \file distance.h
 * \brief Defines classes to manage a hierarchy of distance functions.
 *
 * The distance functions are used when comparing images and correspond to
 * different metrics (colour spaces, restricted zones, Euclidean, Manhattan,
 * etc.).
 *
 * \author Fred Labrosse <ffl@aber.ac.uk>
 * \copyright 2017 Aberystwyth University
 */

#ifndef IMALIB_DISTANCE_H
#define IMALIB_DISTANCE_H

#include <math.h>

#include "imalib/image.h"
#include "imalib/pixelSelector.h"

namespace imalib
{

/*!
 * \class distance_base distance.h imalib/distance.h
 * \brief Base class for all distance functions.
 *
 * As such, this class doesn't do much (if anything at all).  It is the base
 * class to compute the distance between an image and another image or
 * appearance.
 *
 * Classes inheriting from the distance class should contain a number of
 * attributes that are used when computing the distance between images.  Some
 * of these will have the interfaces and will have a super class that will
 * define their interface with virtual functions.  They can therefore be used
 * interchangeably from a syntax %point of view.  Different branches will have
 * different interfaces.
 *
 * The colour space is mostly tackled by the images themselves, although it
 * could be conceivable that some of the distance functions only consider a
 * subset of the colour components.  The type <b>typeCol</b> is the type of
 * the components of the colour space while the int <b>nbColComp</b> is the
 * number of components, both as defined in the image class.
 *
 * The metrics are handled by the classes inherited from the class
 * distanceMetric_base.  The type <b>distMetr</b> is the actual type used for
 * the pixel comparisons.
 *
 * \author Fred Labrosse <ffl@aber.ac.uk>
 */
template <class typeCol, int nbColComp, class distMetr>
class distance_base
{
public:

   /*!
    * \brief Default constructor.
    */
   distance_base()
   {
   }

   /*
    * \brief Constructor.
    * \param[in] metr The functor implementing the distance metric.
    */
//    distance_base(const distMetr& metr);
//       : metric(metr)
//    {
//    }

   /*!
    * \brief Destructor.
    */
   virtual ~distance_base()
   {
   }

   /*!
    * \brief Get the number of pixels that were involved in the last distance
    * calculation.
    * 
    * This may depend on the metric used and how it works.
    * \return The number of pixels involved in the last distance calculation.
    * \sa distanceMetric_base::getNbPixels().
    */
   unsigned long getNbPixels()
   {
      return(metric.getNbPixels());
   }

protected:

   /*!
    * \brief The functor used to compare pixels, therefore implementing the
    * actual distance metric used to compare images.
    * \sa distanceMetric_base.
    */
   distMetr metric;

private:
};


} // namespace imalib

#endif // IMALIB_DISTANCE_H
