/*
 * imalib
 * 
 * Copyright 2017 Aberystwyth University
 * 
 * This library and program are free software: you can redistribute them
 * and/or modify them under the terms of the GNU General Public License as 
 * published by the Free Software Foundation, either version 3 of the License,
 * or (at your option) any later version.
 * 
 * The program and library are distributed in the hope that they will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*!
 * \file distanceImgAppear.h
 * \brief Defines classes to compute the distance between an image and an
 * appearance.
 *
 * The distance functions are used when comparing images and correspond to
 * different metrics (colour spaces, restricted zones, Euclidean, Manhattan,
 * etc.).
 *
 * \author Fred Labrosse <ffl@aber.ac.uk>
 * \copyright 2017 Aberystwyth University
 */

#ifndef IMALIB_DISTANCEIMGAPPEAR_H
#define IMALIB_DISTANCEIMGAPPEAR_H

#include <gsl/gsl_multimin.h>

#include "imalib/appearance.h"
#include "imalib/distance.h"
#include "imalib/utilities.h"

namespace imalib
{

/*!
 * \class distanceImgAppear_base distanceImgAppear.h imalib/distanceImgAppear.h
 * \brief A base class to compute the distance between an image and an
 * appearance.
 *
 * \author Fred Labrosse <ffl@aber.ac.uk>
 */
template <typename typeCol, int nbColComp, typename distMetr>
class distanceImgAppear_base
         : public distance_base<typeCol, nbColComp, distMetr>
{

public:

   /*!
    * \brief Default constructor.
    */
   distanceImgAppear_base()
   : dofMinimisation(0), gsl_x_(NULL), gsl_dx_(NULL),
   gsl_Tf_(gsl_multimin_fminimizer_nmsimplex2), gsl_sf_(NULL),
   paramsTransf(NULL),
   paramsTransfMin(NULL),
   paramsTransfMax(NULL),
   paramsTransfSteps(NULL)
   {
   }

//    /*
//     * \brief Constructor.
//     * \param[in] metr The functor implementing the distance metric.
//     */
//     distanceImgAppear_base(const distMetr& metr);
//          : distance_base<typeCol, nbColComp, distMetr>(metr),
//          dofMinimisation(0), gsl_x_(NULL), gsl_dx_(NULL),
//          gsl_Tf_(gsl_multimin_fminimizer_nmsimplex2), gsl_sf_(NULL),
//          paramsTransf(NULL),
//          paramsTransfMin(NULL),
//          paramsTransfMax(NULL),
//          paramsTransfSteps(NULL)
//    {
//    }

   /*!
    * \brief Destructor.
    */
   virtual ~distanceImgAppear_base()
   {
      gslFree();
   }

   /*!
    * \brief Compute the distance between a panoramic image and an
    * appearance.
    *
    * The apearance is positioned on the panoramic image using a
    * transformation.  The bounding box of the transformed appearance must
    * be inside the image in the vertical direction (this is only verified if
    * compiled without NDEBUG).  In the horizontal direction, the pixels are
    * wrapped around.
    *
    * \param[in] img The panoramic image.
    * \param[in] appear The appearance.
    * \param[in] transf The transformation.
    * \return The distance.
    */
   virtual double distance(const image<typeCol, nbColComp>& img,
                           const appearance<typeCol, nbColComp>& appear,
                           const transformation& transf) = 0;

   /*!
    * \brief Compute the normalised distance between a panoramic image and an
    * appearance.
    *
    * The apearance is positioned on the panoramic image using a
    * transformation.  The bounding box of the transformed appearance must
    * be inside the image in the vertical direction (this is only verified if
    * compiled without NDEBUG).  In the horizontal direction, the pixels are
    * wrapped around.
    *
    * \param[in] img The panoramic image.
    * \param[in] appear The appearance.
    * \param[in] transf The transformation.
    * \return The normalised distance.
    */
   virtual double distanceNormalised(const image<typeCol, nbColComp>& img,
                                     const appearance<typeCol, nbColComp>& appear,
                                     const transformation& transf) = 0;

   /*!
    * \brief Compute the distance between a panoramic image and an
    * appearance.
    *
    * The appearance is positioned on the panoramic image using a
    * transformation.  The transformed appearance can be outside the image
    * but only the pixels inside the image are used.
    *
    * The number of appearance pixels actually used in the distance computation
    * is returned.
    *
    * The normalised version of the metric is used because the number of pixels
    * used can vary.
    *
    * \param[in] img The panoramic image.
    * \param[in] appear The appearance.
    * \param[in] transf The transformation.
    * \param[out] nbPix The number of appearance pixels used in the distance
    * computation.
    * \return The normalised distance.
    */
   virtual double distance(const image<typeCol, nbColComp>& img,
                           const appearance<typeCol, nbColComp>& appear,
                           const transformation& transf,
                           unsigned long* nbPix) = 0;

   /*!
    * \brief Compute the minimum (normalised) distance between an image and an
    * appearance.
    *
    * The appearance is positioned on the image using a
    * transformation, which is used as the starting point for the local
    * optimisation.  The transformation is modified to contain the
    * transformation that gives the minimum distance.  The appearance might not
    * be fully contained in the image, in which case only the pixels inside the
    * image are considered.  Because of that, the normalised distance is used.
    * However, the minimisation will however favour appearances fully
    * contained in the image.
    *
    * \param[in] img The image.
    * \param[in] appear The appearance.
    * \param[in,out] transf The transformation used as the starting point for
    * the minimisation.  Will then contain the transformation that gives the
    * minimum.
    * \param[in] size The acceptable size of the simplex to finish.
    * \return The minimum normalised distance.  It is negative if the
    * minimisation did not successfully finish.
    * \sa distance(const image<typeCol, nbColComp>&, const appearance<typeCol, nbColComp>&, const transformation&, unsigned long*).
    */
   double distanceMin(const image<typeCol, nbColComp>& img,
                      const appearance<typeCol, nbColComp>& appear,
                      transformation* transf,
                      double size = 0.5)
   {
      if (dofMinimisation != transf->getDoF())
      {
         gslAlloc(transf->getDoF());
      }
      transf->getParams(paramsTransf);
      transf->getParamsSteps(paramsTransfSteps);
      for (unsigned short ind = 0; ind < dofMinimisation; ind++)
      {
         gsl_vector_set(gsl_x_, ind, paramsTransf[ind]);
         gsl_vector_set(gsl_dx_, ind, paramsTransfSteps[ind]);
      }

      // Lambda function used to wrap the objective function to be minimised
      // by GSL.
      auto fctptr = [this, transf, &img, &appear](const gsl_vector* x)->double
      {
         return gslDistance(x,
                            transf,
                            img,
                            appear);
      };
      // Create the gsl_multimin_function needed for the minimisation.
      imalib::gslMultiminFunction<decltype(fctptr)> Fp(fctptr, transf->getDoF());
      gsl_multimin_function *F = static_cast<gsl_multimin_function*>(&Fp);
      gsl_multimin_fminimizer_set(gsl_sf_, F, gsl_x_, gsl_dx_);

      int iter = 0;
      int maxIter = 100;
      int status = 0;
      do
      {
         iter++;
         status = gsl_multimin_fminimizer_iterate(gsl_sf_);
         if (status)
         {
            printf("error: %s\n", gsl_strerror(status));
            break;
         }

         status = gsl_multimin_test_size(gsl_multimin_fminimizer_size(gsl_sf_),
                                         size);
      }
      while ((status == GSL_CONTINUE) && (iter < maxIter));

      //          if (status == GSL_SUCCESS)
      //          {
      //             std::cout << "Found minimum in " << iter
      //             << " iterations: error: "
      //             << s_f->fval
      //             << " size: " << gsl_multimin_fminimizer_size(s_f) << "."
      //             << std::endl;
      //          }

      if ((verboseLevel() >= 3) && (iter >= maxIter))
      {
         std::cout << "Reached iteration number limit: " << maxIter
                   << std::endl;
      }

      for (unsigned short ind = 0; ind < dofMinimisation; ind++)
      {
         paramsTransf[ind] = gsl_vector_get(gsl_sf_->x, ind);
      }
      transf->setParams(paramsTransf);

      if (verboseLevel() >= 2)
      {
         std::cout << "Distance: " << gsl_sf_->fval
            << " size: " << gsl_multimin_fminimizer_size(gsl_sf_) << " ("
            << paramsTransf[0];
         for (unsigned short ind = 1; ind < dofMinimisation; ind++)
         {
            std::cout << "x" << paramsTransf[ind];
         }
         std::cout << ")" << std::endl;
      }

      return(status == GSL_SUCCESS ? gsl_sf_->fval : - (gsl_sf_->fval));
   }

protected:

   /*!
    * \brief Function to compute the distance between an image and a
    * transformed appearance for the minimisation in GSL.
    *
    * \param[in] x Point at which the function is evaluated.
    * \param[in] transf The transformation to use.
    * \param[in] img The image.
    * \param[in] appear The appearance.
    * \return The error at x.
    */
   double gslDistance(const gsl_vector* x,
                            transformation* transf,
                            const image<typeCol, nbColComp>& img,
                            const appearance<typeCol, nbColComp>& appear)
   {
      const double coef = 1000.0;
      for (unsigned short ind = 0;
           ind < transf->getDoF();
           ind++)
      {
         paramsTransf[ind] = gsl_vector_get(x, ind);
      }
      transf->setParams(paramsTransf);

      unsigned long nbPix;
      double result = distance(img, appear, *transf, &nbPix);

      // Adding some value if we are outside of the limits.
//       if (nbPix < appear.getNbPixels())
//       {
//          result += coef * pow(appear.getNbPixels() - nbPix, 2.0);
//       }
      transf->getParamsLimits(paramsTransfMin, paramsTransfMax);
      for (unsigned short ind = 0;
           ind < transf->getDoF();
           ++ind)
      {
         if (paramsTransf[ind] < paramsTransfMin[ind])
            result += coef * pow(paramsTransf[ind] - paramsTransfMin[ind],
                                 2.0);
         if (paramsTransf[ind] > paramsTransfMax[ind])
            result += coef * pow(paramsTransfMax[ind] - paramsTransf[ind],
                                 2.0);
      }

      return(result);
   }

   /*!
    * \brief Allocate the memory for the GSL minimisation.
    * \param[in] dof The new number of degrees of freedom of the minimisation.
    */
   void gslAlloc(unsigned short dof)
   {
      gslFree();

      dofMinimisation = dof;
      gsl_x_ = gsl_vector_alloc(dofMinimisation);
      gsl_dx_ = gsl_vector_alloc(dofMinimisation);
      gsl_sf_ = gsl_multimin_fminimizer_alloc(gsl_Tf_, dofMinimisation);

      paramsTransf = new double[dofMinimisation];
      paramsTransfMin = new double[dofMinimisation];
      paramsTransfMax = new double[dofMinimisation];
      paramsTransfSteps = new double[dofMinimisation];
   }

   /*!
    * \brief Free the memory used by GSL for the minimisation.
    */
   void gslFree()
   {
      if (gsl_x_)
      {
         gsl_vector_free(gsl_x_);
         gsl_x_ = NULL;
      }
      if (gsl_dx_)
      {
         gsl_vector_free(gsl_dx_);
         gsl_dx_ = NULL;
      }
      if (gsl_sf_)
      {
         gsl_multimin_fminimizer_free(gsl_sf_);
         gsl_sf_ = NULL;
      }

      delete [] paramsTransf;
      paramsTransf = NULL;
      delete [] paramsTransfMin;
      paramsTransfMin = NULL;
      delete [] paramsTransfMax;
      paramsTransfMax = NULL;
      delete [] paramsTransfSteps;
      paramsTransfSteps = NULL;
   }

private:

   /*!
    * \class minimParams distancePanoramic.h imalib/distancePanoramic.h
    * \brief A class to hold the parameters needed for a minimisation.
    *
    * It's all public but the class is private to distanceImgAppear_base.
    * 
    * \todo Can this not be removed?
    */
   class minimParams
   {
   public:
      distanceImgAppear_base* distanceFunc;
      transformation* transf;
      const image<typeCol, nbColComp>* img;
      const appearance<typeCol, nbColComp>* appear;
   } minParams;

   /*!
    * \brief Degrees of freedom of the minimisation.
    */
   unsigned short dofMinimisation;

   /*!
    * \brief GSL vector of minimisation positions.
    */
   gsl_vector* gsl_x_;

   /*!
    * \brief GSL vector of minimisation step size.
    */
   gsl_vector* gsl_dx_;

   /*!
    * \brief GSL minimiser type.
    */
   const gsl_multimin_fminimizer_type* gsl_Tf_;

   /*!
    * \brief GSL minimiser.
    */
   gsl_multimin_fminimizer* gsl_sf_;

   /*!
    * \brief A vector of the parameters of the transformation to optimise.
    */
   double* paramsTransf;

   /*!
    * \brief A vector of the minimum values for the parameters of the
    * transformation to use in the optimisation.
    */
   double* paramsTransfMin;

   /*!
    * \brief A vector of the maximum values for the parameters of the
    * transformation to use in the optimisation.
    */
   double* paramsTransfMax;

   /*!
    * \brief A vector of the steps for the parameters of the transformation
    * to use in the optimisation.
    */
   double* paramsTransfSteps;

};

/*!
 * \class distanceImgAppearPanoramic distanceImgAppear.h imalib/distanceImgAppear.h
 * \brief A class to compute the distance between a panoramic image and an
 * appearance.
 *
 * \author Fred Labrosse <ffl@aber.ac.uk>
 */
template <typename typeCol, int nbColComp, typename distMetr>
class distanceImgAppearPanoramic
         : public distanceImgAppear_base<typeCol, nbColComp, distMetr>
{

public:

   /*!
    * \brief Default constructor.
    */
   distanceImgAppearPanoramic()
   {
   }

   /*
    * \brief Constructor.
    * \param[in] metr The functor implementing the distance metric.
    */
//    distanceImgAppearPanoramic(const distMetr& metr)
//          : distanceImgAppear_base<typeCol, nbColComp, distMetr>(metr)
//    {
//    }

   /*!
    * \brief Compute the distance between a panoramic image and an
    * appearance.
    *
    * The appearance is positioned on the panoramic image using a
    * transformation.  The bounding box of the transformed appearance must
    * be inside the image in the vertical direction (this is only verified if
    * compiled without NDEBUG).  In the horizontal direction, the pixels are
    * wrapped around.
    *
    * \param[in] img The panoramic image.
    * \param[in] appear The appearance.
    * \param[in] transf The transformation.
    * \return The distance.
    */
   virtual double distance(const image<typeCol, nbColComp>& img,
                           const appearance<typeCol, nbColComp>& appear,
                           const transformation& transf)
   {
#ifndef NDEBUG
      double xBBMin, yBBMin, xBBMax, yBBMax;
      transf.apply(appear.getBBxMin(), appear.getBByMin(), &xBBMin, &yBBMin);
      transf.apply(appear.getBBxMax(), appear.getBByMax(), &xBBMax, &yBBMax);
      assert((static_cast<long>(round(yBBMax)) < img.nbRows())
             && (static_cast<long>(round(yBBMin)) >= 0));
#endif // ndef NDEBUG

      this->metric.reset();

//       typename appearance<typeCol, nbColComp>::
//       appearancePixels::const_iterator iter;
      typename appearancePixels<typeCol, nbColComp>::const_iterator iter;
      double pixX, pixY;

      for (iter = appear.constPixels().begin();
            iter != appear.constPixels().end();
            ++iter)
      {
         transf.apply(iter->first.getX(), iter->first.getY(), &pixX, &pixY);
         this->metric.
         comparePixels(img.
                       getActualPixel(static_cast<unsigned short>(
                                    modulo(static_cast<long>(round(pixX)),
                                           static_cast<long>(img.nbColumns()))),
                                      static_cast<unsigned short>(round(pixY))),
                       iter->second.get(),
                       &(iter->first));
      }

      return(this->metric.distance());
   }

   /*!
    * \brief Compute the normalised distance between a panoramic image and an
    * appearance.
    *
    * The appearance is positioned on the panoramic image using a
    * transformation.  The bounding box of the transformed appearance must
    * be inside the image in the vertical direction (this is only verified if
    * compiled without NDEBUG).  In the horizontal direction, the pixels are
    * wrapped around.
    *
    * \param[in] img The panoramic image.
    * \param[in] appear The appearance.
    * \param[in] transf The transformation.
    * \return The normalised distance.
    */
   double distanceNormalised(const image<typeCol, nbColComp>& img,
                             const appearance<typeCol, nbColComp>& appear,
                             const transformation& transf)
   {
      distance(img, appear, transf);
      return(this->metric.distanceNormalised());
   }

   /*!
    * \brief Compute the distance between a panoramic image and an
    * appearance.
    *
    * The appearance is positioned on the panoramic image using a
    * transformation.  The transformed appearance can be outside the image
    * but only the pixels inside the image are used.
    *
    * The number of appearance pixels actually used in the distance computation
    * is returned.
    *
    * The normalised version of the metric is used because the number of pixels
    * used can vary.
    *
    * \param[in] img The panoramic image.
    * \param[in] appear The appearance.
    * \param[in] transf The transformation.
    * \param[out] nbPix The number of appearance pixels used in the distance
    * computation.
    * \return The distance.
    */
   virtual double distance(const image<typeCol, nbColComp>& img,
                           const appearance<typeCol, nbColComp>& appear,
                           const transformation& transf,
                           unsigned long* nbPix)
   {
      this->metric.reset();

      typename appearancePixels<typeCol, nbColComp>::const_iterator iter;
      double pixX, pixY;

      for (iter = appear.constPixels().begin();
            iter != appear.constPixels().end();
            ++iter)
      {
         transf.apply(iter->first.getX(), iter->first.getY(), &pixX, &pixY);
         if ((static_cast<long>(round(pixY)) >= 0)
               && (static_cast<unsigned short>(round(pixY)) < img.nbRows()))
         {
            this->metric.
            comparePixels(img.
                          getActualPixel(static_cast<unsigned short>(
                                    modulo(static_cast<long>(round(pixX)),
                                           static_cast<long>(img.nbColumns()))),
                                         static_cast<unsigned short>(round(pixY))),
                          iter->second.get(),
                          &(iter->first));
         }
      }

      double dist = this->metric.distanceNormalised();
      *nbPix = this->metric.getNbPixels();

      return(dist);
   }

protected:

private:

};

/*!
 * \class distanceImgAppearRectangular distanceImgAppear.h imalib/distanceImgAppear.h
 * \brief A class to compute the distance between a "normal" rectangular image
 * and an appearance.
 *
 * \author Fred Labrosse <ffl@aber.ac.uk>
 */
template <typename typeCol, int nbColComp, typename distMetr>
class distanceImgAppearRectangular
   : public distanceImgAppear_base<typeCol, nbColComp, distMetr>
{

public:

   /*!
    * \brief Default constructor.
    */
   distanceImgAppearRectangular()
   {
   }

   /*
    * \brief Constructor.
    * \param[in] metr The functor implementing the distance metric.
    */
//    distanceImgAppearRectangular(const distMetr& metr);
//       : distanceImgAppear_base<typeCol, nbColComp, distMetr>(metr)
//    {
//    }

   /*!
    * \brief Compute the distance between a "normal" rectangular image and an
    * appearance.
    *
    * The appearance is positioned on the rectangular image using a
    * transformation.  The bounding box of the transformed appearance must
    * be inside the image (this is only verified if compiled without NDEBUG).
    *
    * \param[in] img The image.
    * \param[in] appear The appearance.
    * \param[in] transf The transformation.
    * \return The distance.
    */
   virtual double distance(const image<typeCol, nbColComp>& img,
                           const appearance<typeCol, nbColComp>& appear,
                           const transformation& transf)
   {
#ifndef NDEBUG
      double xBBMin, yBBMin, xBBMax, yBBMax;
      transf.apply(appear.getBBxMin(), appear.getBByMin(), &xBBMin, &yBBMin);
      transf.apply(appear.getBBxMax(), appear.getBByMax(), &xBBMax, &yBBMax);
      assert((static_cast<long>(round(xBBMax)) < img.nbColumns())
             && (static_cast<long>(round(xBBMin)) >= 0)
             && (static_cast<long>(round(yBBMax)) < img.nbRows())
             && (static_cast<long>(round(yBBMin)) >= 0));
#endif // ndef NDEBUG

      this->metric.reset();

      typename appearancePixels<typeCol, nbColComp>::const_iterator iter;
      double pixX, pixY;

      for (iter = appear.constPixels().begin();
           iter != appear.constPixels().end();
           ++iter)
      {
         transf.apply(iter->first.getX(), iter->first.getY(), &pixX, &pixY);
         this->metric.
            comparePixels(img.
                       getActualPixel(static_cast<unsigned short>(round(pixX)),
                                      static_cast<unsigned short>(round(pixY))),
                          iter->second.get(),
                          &(iter->first));
      }

      return(this->metric.distance());
   }

   /*!
    * \brief Compute the normalised distance between a "normal" rectangular
    * image and an appearance.
    *
    * The appearance is positioned on the rectangular image using a
    * transformation.  The bounding box of the transformed appearance must
    * be inside the image (this is only verified if compiled without NDEBUG).
    *
    * \param[in] img The image.
    * \param[in] appear The appearance.
    * \param[in] transf The transformation.
    * \return The normalised distance.
    */
   double distanceNormalised(const image<typeCol, nbColComp>& img,
                             const appearance<typeCol, nbColComp>& appear,
                             const transformation& transf)
   {
      distance(img, appear, transf);
      return(this->metric.distanceNormalised());
   }

   /*!
    * \brief Compute the distance between a "normal" rectangular image and an
    * appearance.
    *
    * The appearance is positioned on the rectangular image using a
    * transformation.  The transformed appearance can be outside the image
    * but only the pixels inside the image are used.
    *
    * The number of appearance pixels actually used in the distance computation
    * is returned.
    *
    * The normalised version of the metric is used because the number of pixels
    * used can vary.
    *
    * \param[in] img The image.
    * \param[in] appear The appearance.
    * \param[in] transf The transformation.
    * \param[out] nbPix The number of appearance pixels used in the distance
    * computation.
    * \return The distance.
    * \todo Can we get rid of that function, given that the metric can give us
    * the number of pixels?
    */
   virtual double distance(const image<typeCol, nbColComp>& img,
                           const appearance<typeCol, nbColComp>& appear,
                           const transformation& transf,
                           unsigned long* nbPix)
   {
      this->metric.reset();

      typename appearancePixels<typeCol, nbColComp>::const_iterator iter;
      double pixX, pixY;

      for (iter = appear.constPixels().begin();
            iter != appear.constPixels().end();
            ++iter)
      {
         transf.apply(iter->first.getX(), iter->first.getY(), &pixX, &pixY);
         if ((static_cast<long>(round(pixY)) >= 0)
              && (static_cast<unsigned short>(round(pixY)) < img.nbRows())
              && (static_cast<long>(round(pixX)) >= 0)
              && (static_cast<unsigned short>(round(pixX)) < img.nbColumns()))
         {
            this->metric.
               comparePixels(img.
                       getActualPixel(static_cast<unsigned short>(round(pixX)),
                                      static_cast<unsigned short>(round(pixY))),
                             iter->second.get(),
                             &(iter->first));
         }
      }

      double dist = this->metric.distanceNormalised();
      *nbPix = this->metric.getNbPixels();
      return(dist);
   }

protected:

private:

};

} // namespace imalib

#endif // IMALIB_DISTANCEIMGAPPEAR_H
