/*
 * imalib
 * 
 * Copyright 2017 Aberystwyth University
 * 
 * This library and program are free software: you can redistribute them
 * and/or modify them under the terms of the GNU General Public License as 
 * published by the Free Software Foundation, either version 3 of the License,
 * or (at your option) any later version.
 * 
 * The program and library are distributed in the hope that they will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*!
 * \file distanceMetric.h
 * \brief Defines classes to manage a hierarchy of distance metrics for images.
 *
 * \author Fred Labrosse <ffl@aber.ac.uk>
 * \copyright 2017 Aberystwyth University
 * 
 * \todo Split this into several files, one for each class.
 */

#ifndef IMALIB_DISTANCEMETRIC_H
#define IMALIB_DISTANCEMETRIC_H

#include <math.h>

#include "imalib/image.h"
#include "imalib/pixelSelector.h"
#include "imalib/pixelCoord.h"

namespace imalib
{


/*!
 * \class distanceMetric_base distanceMetric.h imalib/distanceMetric.h
 * \brief A base class to handle distance metrics.
 *
 * The class provides methods to compare pixels and to post-process the result
 * of the pixel comparisons.  The contributions from all pixels are computed
 * and accumulated in comparePixels().  The accumulation must be reset
 * beforehand with reset().  Once all pixels have been compared, the final
 * distance is obtained with distance() or distanceNormalised().
 * 
 * The functions distance() and distanceNormalised() can be used as
 * intermediate results and further pixels accumulated after.  In this case,
 * reset() should not be called.
 *
 * The functions defined can be used through a pointer and using run-time
 * resolution but this is not recommended in tight loops because of the
 * overhead of the run-time resolution and the lack of inlining.  Instead, in
 * such situations, the classes should be used using templates for the
 * resolution to be made at compile time.  This is what is done in the class
 * distance_base and its sub-classes.
 *
 * The distance between two pixels is always a double, whichever type the
 * pixels are.
 *
 * \author Fred Labrosse <ffl@aber.ac.uk>
 */
template <class typeCol, int nbColComp>
class distanceMetric_base
{
public:

   /*!
    * \brief Constructor.
    */
   distanceMetric_base()
   : accumulator_(0),
     nbPixels_(0)
   {}

   /*!
    * \brief Destructor.
    */
   virtual ~distanceMetric_base()
   {}

   /*!
    * \brief Reset the comparison.
    *
    * This must be done at the beginning of a new distance computation.
    *
    * If given the images can be used.
    *
    * \param[in] img1 The first image.
    * \param[in] img2 The second image.
    * \todo This is not virtual so having the unused parameters is not useful.
    * Can we get rid of them?
    */
   void reset(const image<typeCol, nbColComp>* img1 = NULL,
              const image<typeCol, nbColComp>* img2 = NULL)
   {
      accumulator_ = 0.0;
      nbPixels_ = 0;
   }

   /*!
    * \brief Compare two pixels.
    *
    * The pixels are provided as a pointer to the first component of each.
    * nbColComp specifies the number of components.
    *
    * The given coordinates are that of one of the two pixels.
    *
    * This has to be implemented in sub-classes.
    *
    * \param[in] pixel1
    * \param[in] pixel2 The two pixels to compare.
    * \param[in] coord The coordinates of the second pixel.
    */
   virtual void comparePixels(const typeCol* pixel1,
                              const typeCol* pixel2,
                              const pixelCoord* coord = NULL) = 0;

   /*!
    * \brief Post-process the result of the comparison of the pixels.
    *
    * After comparing all the pixels of the image and accumulating the values,
    * the result might be post-processed.
    *
    * This does not have to be implemented in sub-classes and the default only
    * returns the current value of the accumulator.  Note that the function
    * does not change the value of the accumulator (is const).
    *
    * \return The distance according to the implemented metric.
    * \sa distanceNormalised().
    */
   double distance() const
   {
      return(accumulator_);
   }

   /*!
    * \brief Normalised version of the distance.
    *
    * This value is independent of the number of pixels used in the comparison.
    * 
    * This is not a normalisation as in average distance over several points
    * but a normalisation over the number of dimensions of the image space.
    *
    * \return Normalised distance value.
    * \sa distance().
    */
   double distanceNormalised() const
   {
      if (this->nbPixels_ != 0)
         return(this->accumulator_ / static_cast<double>(this->nbPixels_));
      else
         return(0.0);
   }

   /*!
    * \brief Get the number of accumulated comparisons (pixels).
    *
    * \return The number of accumulated comparisons (pixels).
    */
   virtual unsigned long getNbPixels() const
   {
      return(this->nbPixels_);
   }

   /*!
    * \brief Merge an accumulated distance into the current one.
    * 
    * This is useful for when two areas of two images are compared
    * independently but need to be combined in a single distance measure.
    */
   virtual
   void merge(const distanceMetric_base<typeCol, nbColComp>* other) = 0;

protected:

   /*!
    * \brief Local variable in which the computation of the distance is
    * accumulated.
    */
   double accumulator_;

   /*!
    * \brief Number of accumulated comparisons (pixels).
    *
    * This needs to be increased in comparePixels().
    */
   unsigned long nbPixels_;

private:
};

/*!
 * \class euclideanMetric distanceMetric.h imalib/distanceMetric.h
 * \brief A distance metric implementing the Euclidean distance.
 *
 * First reset() the distance, then for each pair of pixels, call
 * comparePixels(). Finally, call distance() to get the result.
 *
 * \sa distanceMetric_base.
 *
 * \author Fred Labrosse <ffl@aber.ac.uk>
 */
template <class typeCol, int nbColComp>
class euclideanMetric : public distanceMetric_base<typeCol, nbColComp>
{
public:

   /*!
    * \brief Compare two pixels.
    *
    * The pixels are provided as a pointer to the first component of each.
    * nbColComp specifies the number of components.
    *
    * \param[in] pixel1
    * \param[in] pixel2 The two pixels to compare.
    */
   virtual void comparePixels(const typeCol* pixel1, const typeCol* pixel2,
                              const pixelCoord* /* coord */ = NULL)
   {
      this->nbPixels_++;
      for (int ind = 0; ind < nbColComp; ++ind)
      {
         this->accumulator_ += ((static_cast<double>(pixel1[ind])
                                   - static_cast<double>(pixel2[ind]))
                                  * (static_cast<double>(pixel1[ind])
                                     - static_cast<double>(pixel2[ind])));
      }
   }

   /*!
    * \brief Euclidean distance.
    * \return The Euclidean distance.
    */
   double distance() const
   {
      return(sqrt(this->accumulator_));
   }

   /*!
    * \brief Normalised Euclidean distance.
    *
    * This is not a normalisation as in average distance over several points
    * but a normalisation over the number of dimensions of the image space.
    * This is why it is normalised by the square root of the number of pixels.
    * This means that two pairs of images that are equal but for the number of
    * pixels will return the same value (eg a black and a white image).
    *
    * \return The normalised Euclidean distance.
    */
   double distanceNormalised() const
   {
      if (this->nbPixels_ != 0)
         return(sqrt(this->accumulator_ / static_cast<double>(this->nbPixels_)));
      else
         return(0.0);
   }

   /*!
    * \brief Merge an accumulated distance into the current one.
    * 
    * This is useful for when two areas of two images are compared
    * independently but need to be combined in a single distance measure.
    */
   virtual void merge(const distanceMetric_base<typeCol, nbColComp>* other)
   {
      const euclideanMetric<typeCol, nbColComp>* otherEuclid =
         dynamic_cast<const euclideanMetric<typeCol, nbColComp>*>(other);
      assert(otherEuclid);
      this->nbPixels_ += otherEuclid->nbPixels_;
      this->accumulator_ += otherEuclid->accumulator_;
   }

protected:
private:
};

/*!
 * \class euclideanRobust distanceMetric.h imalib/distanceMetric.h
 * \brief A robust distance metric implementing the Euclidean distance.
 *
 * Pixels that are very different are not kept in the result.  This implies
 * that the distance is always normalised.
 *
 * More specifically, \f$\mathcal{P}\f$ is the set of pixels \f$p_i\f$ of
 * image \f$I\f$
 * that fall under the pixels \f$a_i\f$ of appearance \f$A\f$ after it has been
 * transformed.  For each pixel \f$p_i \in \mathcal{P}\f$ and corresponding
 * appearance pixel \f$a_i\f$, the Sum of Squared differences (SSD, Euclidean
 * distance squared,
 * "distance" in what follows) \f$d_i\f$ is calculated:
 * \f[
 * d_i = \sum_{c=1}^C (p_i(c) - a_i(c))^2,
 * \f]
 * where \f$c\f$ is the \f$c\f$<sup>th</sup> colour component and \f$C\f$ is
 * the number of colour components.  The minimum distance is kept during this
 * calculation:
 * \f[
 * d_m = \min_i d_i.
 * \f]
 * The mean distance over all pixels is calculated:
 * \f[
 * \bar{d} = \frac{1}{N \times C} \sum_i d_i,
 * \f]
 * where \f$N\f$ is the number of pixels in the appearance.
 *
 * The robust Euclidean distance is calculated using the set \f$\mathcal{S}\f$
 * of pixels for which \f$d_i\f$ is close enough to the mean \f$\bar{d}\f$:
 * \f[
 * \mathcal{S} = \{p_i \in \mathcal{P} \mid |d_i - \bar{d}| \leq \bar{d} - d_m\}.
 * \f]
 * The robust Euclidean distance is therefore:
 * \f[
 * E_R(I, A) = \sqrt{\sum_{p_i \in \mathcal{S}} (p_i(c) - a_i(c))^2}.
 * \f]
 *
 * This class is much slower than the non robust version (euclideanDistance).
 * In particular the final accumulation is done in distance() where the
 * necessary statistics can be built.
 *
 * \sa euclideanDistance, distanceMetric_base.
 *
 * \author Fred Labrosse <ffl@aber.ac.uk>
 */
template <class typeCol, int nbColComp>
class euclideanRobust : public distanceMetric_base<typeCol, nbColComp>
{
public:

   /*!
    * \brief Reset the comparison.
    *
    * This must be done at the beginning of a new distance computation.
    * 
    * \param[in] maxNbPixels If not 0 (default), then use this as a hint for
    * how many pixels will be involved in the distance computation. It is safe
    * to leave it as 0 but some speed up might be seen if giving an appropriate
    * value.
    */
   void reset(unsigned long maxNbPixels = 0)
   {
      distanceMetric_base<typeCol, nbColComp>::reset();
      compMax_ = std::numeric_limits<double>::min();
      compMin_ = std::numeric_limits<double>::max();
      comparisons_.clear();
      coordinates_.clear();
      if (maxNbPixels > 0)
      {
         comparisons_.reserve(maxNbPixels);
         coordinates_.reserve(maxNbPixels);
      }
   }

   /*!
    * \brief Compare two pixels.
    *
    * The pixels are provided as a pointer to the first component of each.
    * nbColComp specifies the number of components.
    *
    * The provided coordinates are stored.  This could be the coordinates of
    * one of the two pixels, used for display purposes or set an 
    * imalib::appearance.
    *
    * \param[in] pixel1
    * \param[in] pixel2 The two pixels to compare.
    * \param[in] coord The coordinates of a pixel.
    */
   virtual void comparePixels(const typeCol* pixel1, const typeCol* pixel2,
                              const pixelCoord* coord = NULL)
   {
      this->nbPixels_++;
      double pixelDist = 0;
      for (int ind = 0; ind < nbColComp; ++ind)
      {
         pixelDist += ((static_cast<double>(pixel1[ind])
                        - static_cast<double>(pixel2[ind]))
                       * (static_cast<double>(pixel1[ind])
                        - static_cast<double>(pixel2[ind])));
      }
      comparisons_.push_back(pixelDist);
      if (compMax_ < pixelDist)
         compMax_ = pixelDist;
      if (compMin_ > pixelDist)
         compMin_ = pixelDist;
      this->accumulator_ += pixelDist;
      if (coord)
      {
         coord->setUsedInDistance(false);
         coordinates_.push_back(coord);
      }
   }

   /*!
    * \brief Euclidean distance.
    * 
    * This is the same as the normalised version as the number of pixels used
    * can vary.
    *
    * \return The Euclidean distance.
    * \sa distanceNormalised().
    */
   double distance() const
   {
      return(distanceNormalised());
   }

   /*!
    * \brief Normalised Euclidean distance.
    *
    * This is not a normalisation as in average distance over several points
    * but a normalisation over the number of dimensions of the image space.
    * This is why it is normalised by the square root of the number of pixels.
    * This means that two pairs of images that are equal but for the number of
    * pixels will return the same value (eg a black and a white image).
    * 
    * If no pixels are retained then 0 is returned.
    *
    * \return The robust normalised Euclidean distance.
    */
   double distanceNormalised() const
   {
      nbPixelsUsed_ = 0;
      double finalAccum = 0;
      double mean = this->accumulator_ / (this->nbPixels_ * nbColComp);
      double maxDev = mean - compMin_;
      for (unsigned long ind = 0; ind < comparisons_.size(); ind++)
      {
         if (fabs(comparisons_[ind] - mean) <= maxDev)
         {
            nbPixelsUsed_++;
            finalAccum += comparisons_[ind];
            coordinates_[ind]->setUsedInDistance(true);
         }
         else
         {
            coordinates_[ind]->setUsedInDistance(false);
         }
      }
      if (nbPixelsUsed_ != 0)
         return(sqrt(finalAccum / static_cast<double>(nbPixelsUsed_)));
      else
         return(0.0);
   }

   /*!
    * \brief Get the number of pixels used in the robust distance.
    *
    * \return The number of pixels used.
    */
   virtual unsigned long getNbPixels() const
   {
      return(this->nbPixelsUsed_);
   }

   /*!
    * \brief Merge an accumulated distance into the current one.
    * 
    * This is useful for when two areas of two images are compared
    * independently but need to be combined in a single distance measure.
    */
   virtual void merge(const distanceMetric_base<typeCol, nbColComp>* other)
   {
      const euclideanRobust<typeCol, nbColComp>* otherEuclid =
         dynamic_cast<const euclideanRobust<typeCol, nbColComp>*>(other);
      assert(otherEuclid);
      this->nbPixels_ += otherEuclid->nbPixels_;
      this->accumulator_ += otherEuclid->accumulator_;
      comparisons_.insert(comparisons_.end(),
                          otherEuclid->comparisons_.begin(),
                          otherEuclid->comparisons_.end() );
   }

protected:
private:

   /*!
    * \brief Vector of all pixel comparisons.
    * 
    * We store all pixel comparisons so that we can select the "good" ones
    * later.
    */
   std::vector<double> comparisons_;

   /*!
    * \brief Maximum value of the comparison.
    */
   double compMax_;

   /*!
    * \brief Minimum value of the comparison.
    */
   double compMin_;

   /*!
    * \brief A vector of the pixel coordinates used.
    *
    * This is only used by the functions that pass and get coordinates.
    * Otherwise this is not used.  These functions are much slower!
    */
   std::vector<const pixelCoord*> coordinates_;

   /*!
    * \brief Number of pixels actually used in the distance.
    */
   mutable unsigned long nbPixelsUsed_;
};

/*!
 * \class chiSquareMetric distanceMetric.h imalib/distanceMetric.h
 * \brief A distance metric implementing the chi-square distance.
 *
 * First reset() the distance, then for each pair of pixels, call
 * comparePixels(). Finally, call distance() to get the result.
 *
 * \author Juan Cao <juc3@aber.ac.uk>
 */
template <class typeCol, int nbColComp>
class chiSquareMetric : public distanceMetric_base<typeCol, nbColComp>
{
public:

   /*!
    * \brief Compare two pixels.
    *
    * The pixels are provided as a pointer to the first component of each.
    * nbColComp specifies the number of components.
    *
    * \param[in] pixel1
    * \param[in] pixel2 The two pixels to compare.
    */
   void comparePixels(const typeCol* pixel1, const typeCol* pixel2,
                      const pixelCoord* /* coord */ = NULL)
   {
      this->nbPixels_++;
      for (int ind = 0; ind < nbColComp; ++ind)
      {
         double vsum=(static_cast<double>(pixel1[ind])
                      + static_cast<double>(pixel2[ind]));
         double vsub=(static_cast<double>(pixel1[ind])
                      - static_cast<double>(pixel2[ind]));
         if (vsum != 0.0)
         {
            this->accumulator_ += (vsub * vsub) / (2 * vsum);
         }
      }
   }

   /*!
    * \brief Chi square distance.
    *
    * \return The Chi square distance.
    */
   double distance() const
   {
      return(this->accumulator_);
   }

   /*!
    * \brief Normalised Chi square distance.
    * \return The normalised Chi square distance.
    */
   double distanceNormalised() const
   {
      if (this->nbPixels_ != 0)
         return(this->accumulator_ / static_cast<double>(this->nbPixels_));
      else
         return(0.0);
   }

   /*!
    * \brief Merge an accumulated distance into the current one.
    * 
    * This is useful for when two areas of two images are compared
    * independently but need to be combined in a single distance measure.
    */
   virtual void merge(const distanceMetric_base<typeCol, nbColComp>* other)
   {
      const chiSquareMetric<typeCol, nbColComp>* otherChi =
      dynamic_cast<const chiSquareMetric<typeCol, nbColComp>*>(other);
      assert(otherChi);
      this->nbPixels_ += otherChi->nbPixels_;
      this->accumulator_ += otherChi->accumulator_;
   }

protected:
private:
};

/*!
 * \class correlationMetric distanceMetric.h imalib/distanceMetric.h
 * \brief A distance metric implementing the correlation distance.
 *
 * First reset() the distance, then for each pair of pixels, call
 * comparePixels(). Finally, call distance() to get the result.
 *
 * The pixel values are accumulated and the correlation coefficient computed
 * at the end, when the distance is requested.  This is not very
 * efficient, but this metric is anyway slow to compute.  The class works like
 * that because the normalisation step requires that all the pixels accumulated
 * are known.
 *
 * \author Fred Labrosse <ffl@aber.ac.uk>
 * \author Juan Cao <juc3@aber.ac.uk>
 */
template <class typeCol, int nbColComp>
class correlationMetric : public distanceMetric_base<typeCol, nbColComp>
{
public:

   /*!
    * \brief Constructor.
    */
   correlationMetric()
   {
      colours1_.resize(nbColComp);
      colours2_.resize(nbColComp);
   }

   /*!
    * \brief Destructor
    */
   ~correlationMetric()
   {
   }

   /*!
    * \brief Reset the comparison.
    *
    * This must be done at the beginning of a new distance computation.
    *
    * \param[in] img1 The first image.
    * \param[in] img2 The second image.
    */
   void reset(const image<typeCol, nbColComp>* img1,
              const image<typeCol, nbColComp>* img2)
   {
      assert(img1 && img2);

      distanceMetric_base<typeCol, nbColComp>::reset();

      for (int ind = 0; ind < nbColComp; ++ind)
      {
         if (colours1_[ind].capacity() < (img1->nbColumns() * img1->nbRows()))
         {
            colours1_[ind].reserve(img1->nbColumns() * img1->nbRows());
            colours2_[ind].reserve(img2->nbColumns() * img2->nbRows());
            // This assumes that pixels are accumulated only once.
         }
         colours1_[ind].clear();
         colours2_[ind].clear();
      }
   }

   /*!
    * \brief Compare two pixels.
    *
    * The pixels are provided as a pointer to the first component of each.
    * nbColComp specifies the number of components.
    *
    * \param[in] pixel1
    * \param[in] pixel2 The two pixels to compare.
    */
   void comparePixels(const typeCol* pixel1, const typeCol* pixel2,
                      const pixelCoord* /* coord */ = NULL)
   {
      for (int ind = 0; ind < nbColComp; ++ind)
      {
         colours1_[ind].push_back(pixel1[ind]);
         colours2_[ind].push_back(pixel2[ind]);
      }
   }

   /*!
    * \brief Correlation coefficient.
    * \return The correlation coefficient.
    */
   double distance() const
   {
      double accumulator0 = 0.0;
      double accumulator1 = 0.0;
      double accumulator2 = 0.0;
      double average1[nbColComp];
      double average2[nbColComp];

      // We compute the averages.
      for (int ind = 0; ind < nbColComp; ++ind)
      {
         double sum1 = 0;
         double sum2 = 0;
         for (unsigned long indPix = 0; indPix < colours1_[ind].size();
              indPix++)
         {
            sum1 += colours1_[ind][indPix];
            sum2 += colours2_[ind][indPix];
         }
         average1[ind] = sum1 / colours1_[ind].size();
         average2[ind] = sum2 / colours2_[ind].size();
      }

      // We accumulate all the values.
      for (int ind = 0; ind < nbColComp; ++ind)
      {
         for (unsigned long indPix = 0; indPix < colours1_[ind].size();
              indPix++)
         {
            double vsub1 = colours1_[ind][indPix] - average1[ind];
            double vsub2 = colours2_[ind][indPix] - average2[ind];
            accumulator0 += vsub1 * vsub2;
            accumulator1 += vsub1 * vsub1;
            accumulator2 += vsub2 * vsub2;
         }
      }

      // Normalisation.
      if ((accumulator1 != 0.0) && (accumulator2 != 0.0))
      {
         return(accumulator0 / (sqrt(accumulator1 * accumulator2)));
      }
      else
      {
         std::cerr
           << "correlationMetric::distance(): One of the two images was uniform in colour.\n";
         return(0.0);
      }
   }

   /*!
    * \brief Normalised correlation coefficient.
    *
    * This is the same as the distance() function as the correlation
    * coefficient is normalised anyway.
    *
    * \return The correlation coefficient.
    */
   double distanceNormalised() const
   {
      return(distance());
   }

   /*!
    * \brief Merge an accumulated distance into the current one.
    * 
    * This is useful for when two areas of two images are compared
    * independently but need to be combined in a single distance measure.
    */
   virtual void merge(const distanceMetric_base<typeCol, nbColComp>* other)
   {
      const correlationMetric<typeCol, nbColComp>* otherCorrel =
         dynamic_cast<const correlationMetric<typeCol, nbColComp>*>(other);
      assert(otherCorrel);
      this->nbPixels_ += otherCorrel->nbPixels_;
      for (int ind = 0; ind < nbColComp; ++ind)
      {
         colours1_[ind].insert(colours1_[ind].end(),
                               otherCorrel->colours1_[ind].begin(),
                               otherCorrel->colours1_[ind].end());
         colours2_[ind].insert(colours2_[ind].end(),
                               otherCorrel->colours2_[ind].begin(),
                               otherCorrel->colours2_[ind].end());
      }
   }

protected:

private:

   /*!
    * \brief To store the colour values to accumulate from the first image.
    */
   std::vector<std::vector<double> > colours1_;

   /*!
    * \brief To store the colour values to accumulate from the second image.
    */
   std::vector<std::vector<double> > colours2_;
};

} // namespace imalib

#endif // IMALIB_DISTANCEMETRIC_H
