/*
 * imalib
 * 
 * Copyright 2017 Aberystwyth University
 * 
 * This library and program are free software: you can redistribute them
 * and/or modify them under the terms of the GNU General Public License as 
 * published by the Free Software Foundation, either version 3 of the License,
 * or (at your option) any later version.
 * 
 * The program and library are distributed in the hope that they will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*!
 * \file distancePanoramic.h
 * \brief Defines classes to manage a hierarchy of distance functions for
 * panoramic images.
 *
 * The distance functions are used when comparing images and correspond to
 * different metrics (colour spaces, restricted zones, Euclidean, Manhattan,
 * etc.).
 *
 * \author Fred Labrosse <ffl@aber.ac.uk>
 * \copyright 2017 Aberystwyth University
 */

#ifndef IMALIB_DISTANCEPANORAMIC_H
#define IMALIB_DISTANCEPANORAMIC_H

#include <math.h>
#include <assert.h>

#include <gsl/gsl_multimin.h>

#include "imalib/distance.h"
#include "imalib/pixelSelector.h"
#include "imalib/transformation.h"
#include "imalib/utilities.h"
#include "imalib/shape.h"

namespace imalib
{

/*!
 * \class distancePanoramic_base distancePanoramic.h imalib/distancePanoramic.h
 * \brief A base class to compute the distance for panoramic images.
 *
 * This will typically support "infinitely" wide images by wrapping them
 * around.
 *
 * \author Fred Labrosse <ffl@aber.ac.uk>
 */
template <typename typeCol, int nbColComp, typename distMetr>
class distancePanoramic_base
         : public distance_base<typeCol, nbColComp, distMetr>
{
public:

   /*!
    * \brief Default constructor.
    */
   distancePanoramic_base()
   : minimiser(this)
   {
   }

   /*
    * \brief Constructor.
    *
    * \param[in] metr The functor implementing the distance metric.
    */
//    distancePanoramic_base(const distMetr& metr)
//          : distance_base<typeCol, nbColComp, distMetr>(metr),
//          minimiser(this)
//    {
//    }

   /*!
    * \brief Compute the distance between two panoramic images.
    *
    * The second image can be shifted column-wise to represent a rotation of
    * the camera.
    *
    * \param[in] image1 The first image.
    * \param[in] image2 The second image.
    * \param[in] shift The shift, positive to the right.
    * \return The distance.
    */
   virtual double distance(const image<typeCol, nbColComp>& image1,
                           const image<typeCol, nbColComp>& image2,
                           short shift) = 0;

   /*!
    * \brief Compute the distance between a panoramic image and a set of
    * pixels from a second image.
    *
    * This is similar in idea to what is performed in the class
    * distancePanoramicSelector but takes the pixelSelector as a parameter
    * rather than having it as a class member (therefore avoiding to have
    * to copy it if it changes).
    *
    * The set of pixels can be moved vertically (to cope with a change in
    * attitude of the panoramic camera) and horizontally (to cope with a
    * rotation of the panoramic camera).  The set of pixels must remain within
    * the panoramic image (which is only tested if compiled without -NDEBUG).
    *
    * \param[in] image1 The panoramic image.
    * \param[in] image2 The second image from which the set of pixels is taken.
    * \param[in] selector The pixels of image2 to use.
    * \param[in] shiftX The horizontal shift, positive to the right.
    * \param[in] shiftY The vertical shift, positive downwards.
    * \return The distance.
    */
   virtual double distance(const image<typeCol, nbColComp>& image1,
                           const image<typeCol, nbColComp>& image2,
                           const pixelSelector& selector,
                           short shiftX,
                           short shiftY)
   {
      std::cerr << "Function \n"
         << "\tdouble distance(const image<typeCol, nbColComp>&,\n"
         << "\t                const image<typeCol, nbColComp>&,\n"
         << "\t                const pixelSelector&,\n"
         << "\t                short,\n"
         << "\t                short)\n"
         << "must be implemented for your class.\n";

      return(nan("char-sequence"));
   }

   /*!
    * \brief Compute the distance between two panoramic images with a shift
    * in the X and Y directions of the second image.
    *
    * This is obviously only possible for subclasses where not all pixels
    * are included (such as distancePanoramicMapping).
    *
    * The set of pixels can be moved vertically (to cope with a change in
    * attitude of the panoramic camera) and horizontally (to cope with a
    * rotation of the panoramic camera).  The set of pixels must remain within
    * the panoramic image (which is only tested if compiled without -NDEBUG).
    *
    * \param[in] image1 The panoramic image.
    * \param[in] image2 The second image from which the set of pixels is taken.
    * \param[in] shiftX The horizontal shift, positive to the right.
    * \param[in] shiftY The vertical shift, positive downwards.
    * \return The distance.
    */
   virtual double distance(const image<typeCol, nbColComp>& image1,
                           const image<typeCol, nbColComp>& image2,
                           short shiftX,
                           short shiftY)
   {
      std::cerr << "Function \n"
         << "\tdouble distance(const image<typeCol, nbColComp>&,\n"
         << "\t                const image<typeCol, nbColComp>&,\n"
         << "\t                short,\n"
         << "\t                short)\n"
         << "must be implemented for your class.\n";

      return(nan("char-sequence"));
   }


   /*!
    * \brief Compute the minimum distance between two images for all possible
    * rotations of the second image.
    *
    * This only makes sense for panoramic images (see
    * imageMapping::createUnwrappingMapping()).
    *
    * The search can be global or local.  In the latter case, once a local
    * minimum is found, the distance at the local minimum +/- 5 in skip is
    * measured and a new search is initiated from these if they are lower, to
    * try to get out of local minima.
    *
    * \param[in] image1 The first image.
    * \param[in] image2 The second image.
    * \param[in,out] bestShift Used as the starting point of a local search and
    * set to the best shift if non NULL.
    * \param[in] local If true, then look for the local minimum.
    * \return The minimum distance.
    */
   virtual double distanceMin(const image<typeCol, nbColComp>& image1,
                              const image<typeCol, nbColComp>& image2,
                              short* bestShift,
                              bool local = false)
   {
      if (!local)
      {                         // Absolute minimum.
         double bestDistance = distance(image1, image2, 0);
         if (bestShift)
            *bestShift = 0;

         for (unsigned short index = 1; index < image1.nbColumns(); ++index)
         {
            double dist = distance(image1, image2, index);
            if (dist < bestDistance)
            {
               bestDistance = dist;
               if (bestShift)
                  *bestShift = index;
            }
         }

         return(bestDistance);
      }
      else
      {                         // Local minimum.
         double found = false;
         short direction = 1;
         short shiftMin = image2.nbColumns(), shiftMax = -image2.nbColumns();
         short currShift = (bestShift ? *bestShift : 0);
         double distanceZero = distance(image1, image2, currShift);
         double currentDist = distanceZero;
         double minDist = currentDist;
         short shiftAroundMax = image1.nbColumns() / 36;
         short nbAround = 4;

         while (!found)
         {
            double previousDist;
            bool valleyMin = false, valleyMax = false, valley = false;
            bool goingDown = false;

            while (!valleyMin && !valleyMax)
            {
               currShift += direction;
               previousDist = currentDist;
               currentDist = distance(image1, image2, currShift);
               if (currentDist < previousDist)
               {
                  goingDown = true;
                  valley = valleyMin = valleyMax = false;
               }
               else if (currentDist == previousDist)
               {
                  if (direction == 1)
                  {
                     shiftMax = currShift - 1;
                     valleyMax = true;
                  }
                  else
                  {
                     shiftMin = currShift + 1;
                     valleyMin = true;
                  }
                  valley = true;
                  minDist = currentDist;
               }
               else                // currentDist > previousDist
               {
                  if (valley)      // We were in a valley.
                  {
                     if (direction == 1)
                     {
                        shiftMax = currShift - 1;
                        valleyMax = true;
                     }
                     else
                     {
                        shiftMin = currShift + 1;
                        valleyMin = true;
                     }
                  }
                  else if (goingDown) // We were not in a valley, but were
                     // going down previously.
                  {
                     shiftMin = shiftMax = currShift - direction;
                     valleyMin = valleyMax = true;
                     minDist = previousDist;
                  }
                  else             // The distance is increasing.  Let's look
                  {                // for the minimum.
                     direction = -direction;
                     goingDown = true;
                     currShift = (bestShift ? *bestShift : 0);
                     currentDist = distanceZero;
                  }
               }
            }

            // Look for lower points around.
            currShift = (shiftMax + shiftMin) / 2;
            short shiftAround = 0;
            short ind = 1;
            found = true;
            while (found && (ind < nbAround))
            {
               shiftAround += shiftAroundMax / nbAround;
               currentDist = distance(image1, image2,
                              modulo(static_cast<long>(currShift + shiftAround),
                                     static_cast<long>(image1.nbColumns())));
               if (currentDist < minDist)
               {
                  found = false;
                  currShift = modulo(static_cast<long>(currShift + shiftAround),
                                     static_cast<long>(image1.nbColumns()));
                  if (bestShift)
                     *bestShift = currShift;
                  direction = 1;
                  distanceZero = currentDist;
               }
               else
               {
                  currentDist = distance(image1, image2,
                           modulo(static_cast<long>(currShift - shiftAround),
                                  static_cast<long>(image1.nbColumns())));
                  if (currentDist < minDist)
                  {
                     found = false;
                     currShift =
                        modulo(static_cast<long>(currShift - shiftAround),
                               static_cast<long>(image1.nbColumns()));
                     if (bestShift)
                        *bestShift = currShift;
                     direction = -1;
                     distanceZero = currentDist;
                  }
               }
               ++ind;
            }
         }

         if (bestShift)
            *bestShift = (shiftMax + shiftMin) / 2;

         return(minDist);
      }
   }

   /*!
    * \brief Compute the distance between two images with a sub-pixel
    * resolution.
    *
    * This function uses the pixel-resolution distance function to compute 3
    * values (at shift, shift - delta (\f$ \delta \f$) and shift + delta).
    * Depending on the spatial configuration of these 3 values and the height
    * of the minimum, a parabolic interpolation or a bi-linear interpolation
    * is performed to find the sub-pixel local minimum.
    *
    * The parabolic interpolation is used when the 3 points form a sharp
    * minimum (outer values roughly at the same height --- lower of the 2
    * above the mid-height of the other and centre value ---, and centre
    * value below) or when the pixel resolution minimum is above 1/4 of the
    * height of the self-distance of the reference image above 0. In this
    * case, the following algorithm is used.
    *
    * For 3 values of \f$ f(x) \f$ at \f$ a \f$, \f$ b \f$ and \f$ c \f$ (in
    * that order), the minimum (where the parabola's derivative is 0) is at
    * position \f$ m \f$:
    *
    * \f[
    *    m = b - \frac{1}{2}
    *    \frac{(b - a)^2(f(b) - f(c)) - (b - c)^2(f(b) - f(a))}
    *    {(b - a)(f(b) - f(c)) - (b - c)(f(b) - f(a))}.
    * \f]
    *
    * Here, \f$ b = \f$ <i>shift</i>, \f$ a = b - \delta\f$ and \f$ c = b +
    * \delta\f$, which simplifies the above equation quite a lot!:
    *
    * \f[
    *    m = b - \frac{1}{2}
    *    \frac{\delta(f(a) - f(c))}{2f(b) - f(a) - f(c)}.
    * \f]
    *
    * This can be put back in the parabolic function:
    *
    * \f[
    *    f(x) = f(a)\frac{(x-b)(x-c)}{(a-b)(a-c)}
    *    + f(b)\frac{(x-c)(x-a)}{(b-c)(b-a)}
    *    + f(c)\frac{(x-a)(x-b)}{(c-a)(c-b)},
    * \f]
    *
    * which leads to after simplifications:
    *
    * \f[
    *    f(m) = \frac{f(a)^2+16f(b)^2+f(c)^2 - f(c)[8f(b)+2f(a)]-8f(a)f(b)}
    *    {16f(b) - 8[f(a)+f(c)]}.
    * \f]
    *
    * When the 3 values form a square minimum (opposite case of the sharp
    * minimum) and the pixel-resolution minimum is below 1/4 of the height of
    * the self-distance of the reference image from 0, a bi-linear
    * interpolation is used. The 2 lower points on either side of the 2
    * minima (forming the "square") \f$(x_1,y_1)\f$, \f$(x_2,y_2)\f$ and
    * \f$(x_3,y_3)\f$, \f$(x_4,y_4)\f$, are used to create 2 straight lines,
    * the intersection of which is an approximation of the sub-pixel minimum.
    * The x and y (shift and distance) of the minimum are given by:
    *
    * \f[
    *    \left[
    *    x=-{{\left(x_2-x_1\right)\,x_4y_3+\left(x_1y_2-x_2y_1\right)\,
    *    x_4+\left(x_1-x_2\right)\,x_3y_4+\left(x_2y_1-x_1y_2\right)\,x_3}
    *    \over{\left(x_2-x_1\right)\,y_4+\left(x_1-x_2\right)\,y_3
    *    +\left(x_3-x_4\right)\,y_2+\left(x_4-x_3\right)\,y_1}},
    *    y={{\left(x_2y_1-x1y_2\right)\,y_4+\left(x_1y_2-x_2y_1\right)\,
    *    y_3+\left(x_3y_4-x_4y_3\right)\,y_2+\left(x_4y_3-x_3y_4\right)\,
    *    y_1}\over{\left(x_2-x_1\right)\,y_4+\left(x_1-x_2\right)\,
    *    y_3+\left(x_3-x_4\right)\,y_2+\left(x_4-x_3\right)\,y_1}} \right]
    * \f]
    *
    * The nominal \f$ x \f$ above is moved up by a proportion of itself such
    * that for pixel resolution minima close to the 1/4 threshold above the
    * value is close to that minimum (and hence close to the value returned by
    * parabolic interpolation), for continuity reasons.
    *
    * \todo The function assumes that there are no shift wrapping problem.
    * Fix this!
    *
    * \param[in] image1 The first image.
    * \param[in] image2 The second image.
    * \param[in,out] bestShift The shift, positive to the right.  Must contain
    * the pixel-resolution position of the minimum.  Will contain the
    * sub-pixel optimum.
    * \param[in] delta The difference between the centre point and the
    * two other points.
    * \param[out] shiftLeft The shift of the left-most point used to compute
    * the sub-pixel minimum, if not NULL.
    * \param[out] heightLeft The height (Euclidean distance) of the left-most
    * point used to compute the sub-pixel minimum, if not NULL.
    * \param[out] shiftRight The shift of the right-most point used to compute
    * the sub-pixel minimum, if not NULL.
    * \param[out] heightRight The height (Euclidean distance) of the
    * right-most point used to compute the sub-pixel minimum, if not NULL.
    * \return The sub-pixel distance.
    */
   inline double distanceMinSubPixel(const image<typeCol, nbColComp>& image1,
                                     const image<typeCol, nbColComp>& image2,
                                     double* bestShift,
                                     double delta = 1,
                                     short* shiftLeft = NULL,
                                     double* heightLeft = NULL,
                                     short* shiftRight = NULL,
                                     double* heightRight = NULL)
   {
      double valA = distance(image1, image2,
                             static_cast<short>(*bestShift - delta));
      double valB = distance(image1, image2,
                             static_cast<short>(*bestShift));
      double valC = distance(image1, image2,
                             static_cast<short>(*bestShift + delta));
      short shift = static_cast<short>(*bestShift);
      // This probably does not make much difference, but shift above
      // should be 0 given that the distance is measured using the
      // reference image and itself.
      double maxRef = distanceMax(image1, image1, shift);

      double x1, x2, x3, x4, y1, y2, y3, y4;
      bool linear;

      double relativeHeight = (maxRef - valB) / maxRef;
      static double thresh = 0.5;
      if (relativeHeight < (1.0 - thresh))
      {
         linear = false;
      }
      else
      {
         if (valC > valA)
         {                  // The fourth value (if needed) will be at -2delta.
            if (valA < ((valC - valB) / 2.0) + valB)
            {               // Square minimum
               linear = true;
               x1 = *bestShift - 2 * delta;
               y1 = distance(image1, image2, static_cast<short>(x1));
               x2 = *bestShift - delta;
               y2 = valA;
               x3 = *bestShift;
               y3 = valB;
               x4 = *bestShift + delta;
               y4 = valC;
            }
            else
            {               // Pointy minimum
               linear = false;
            }
         }
         else
         {               // The fourth value (if needed) will be at +2delta.
            if (valC < ((valA - valB) / 2.0) + valB)
            {               // Square minimum
               linear = true;
               x1 = *bestShift - delta;
               y1 = valA;
               x2 = *bestShift;
               y2 = valB;
               x3 = *bestShift + delta;
               y3 = valC;
               x4 = *bestShift + 2 * delta;
               y4 = distance(image1, image2, static_cast<short>(x4));
            }
            else
            {               // Pointy minimum
               linear = false;
            }
         }
      }
//          linear = false;
      if (!linear)
      {
         if (shiftLeft)
            *shiftLeft = static_cast<short>(*bestShift - delta);
         if (heightLeft)
            *heightLeft = valA;
         if (shiftRight)
            *shiftRight = static_cast<short>(*bestShift + delta);
         if (heightRight)
            *heightRight = valC;

         *bestShift = *bestShift - (0.5 * delta * (valA - valC)
                                    / (2 * valB - valC - valA));

         return((valA*valA + 16*valB*valB + valC*valC
                 - valC*(8*valB + 2*valA) - 8*valA*valB)
                / (16*valB - 8*(valA + valC)));
      }
      else
      {
//             double bestShiftLin = -(((x2 - x1) * x4 * y3
//                                     + (x1 * y2 - x2 * y1) * x4
//                                     + (x1 - x2) * x3 * y4
//                                     + (x2* y1 - x1 * y2) * x3)
//                                     / ((x2 - x1) * y4
//                                        + (x1 - x2) * y3
//                                        + (x3 - x4) * y2
//                                        + (x4 - x3) * y1));
         double valLin = (((x2 * y1 - x1 * y2) * y4
                           + (x1 * y2 - x2 * y1) * y3
                           + (x3 * y4 - x4 * y3) * y2
                           + (x4 * y3 - x3 * y4) * y1)
                          / ((x2 - x1) * y4
                             + (x1 - x2) * y3
                             + (x3 - x4) * y2
                             + (x4 - x3) * y1));
         double bestShiftPara = *bestShift - (0.5 * delta * (valA - valC)
                                              / (2 * valB - valC - valA));
         double valPara = (valA * valA + 16 * valB * valB + valC * valC
                           - valC * (8 * valB + 2 * valA) - 8 * valA * valB)
                          / (16 * valB - 8 * (valA + valC));

         if (shiftLeft)
            *shiftLeft = static_cast<short>(x1);
         if (heightLeft)
            *heightLeft = y1;
         if (shiftRight)
            *shiftRight = static_cast<short>(x4);
         if (heightRight)
            *heightRight = y4;

         double prop = (relativeHeight - thresh) / (1.0 - thresh);
//             *bestShift = prop * bestShiftLin + (1.0 - prop) * bestShiftPara;
         *bestShift = bestShiftPara;

         return(prop * valLin + (1.0 - prop) * valPara);
      }
   }

   /*!
    * \brief Compute the maximum of the distance between two images.
    *
    * This maximum is the value of the distance when the images are least
    * correlated.  Only an approximation of that is computed for efficiency
    * reasons.  This is the value of the distance function at the position
    * opposite the minimum (best correlation).
    *
    * \param[in] image1 The first image.
    * \param[in] image2 The second image.
    * \param[in] bestShift The shift of the pixel-resolution minimum.
    * \return The maximum distance value.
    */
   inline double distanceMax(const image<typeCol, nbColComp>& image1,
                             const image<typeCol, nbColComp>& image2,
                             short bestShift)
   {
      short shift = modulo(static_cast<long>(bestShift)
                           + static_cast<long>(image1.nbColumns()) / 2,
                           static_cast<long>(image1.nbColumns()));
      // 180 for a resolution of one pixel per
      // degree.
      return(distance(image1, image2, shift));
   }

   /*!
    * \brief Compute the amplitude of the distance between two images.
    *
    * The amplitude is the difference in height of the maximum of the distance
    * function and the minimum.
    *
    * \param[in] image1 The first image.
    * \param[in] image2 The second image.
    * \param[in,out] shift Used as the starting point of a local search and
    * set to the best shift.
    * \param[in] local If true, then look for the local minimum.
    * \return The distance amplitude.
    */
   inline double distanceAmplitude(const image<typeCol, nbColComp>& image1,
                                   const image<typeCol, nbColComp>& image2,
                                   double* shift,
                                   bool local = false)
   {
      short aShift = static_cast<short>(round(*shift));
      double minDist = distanceMin(image1, image2, &aShift, local);
      *shift = aShift;
      minDist = distanceMinSubPixel(image1, image2, shift, 1);
      double heightDist = distanceMax(image1, image2, aShift);
      return (heightDist - minDist);
   }

   /*!
    * \brief Compute the minimum distance between a panoramic image and a set
    * of pixels from a second image.
    *
    * This is similar in idea to what is performed in the class
    * distancePanoramicSelector but takes the pixelSelector as a parameter
    * rather than having it as a class member (therefore avoiding to have
    * to copy it if it changes).
    *
    * The set of pixels is moved vertically (to cope with a change in
    * attitude of the panoramic camera) and horizontally (to cope with a
    * rotation of the panoramic camera).  The set of pixels must remain within
    * the panoramic image (which is only tested if compiled without -NDEBUG).
    *
    * \param[in] image1 The panoramic image.
    * \param[in] image2 The second image from which the set of pixels is taken.
    * \param[in] selector The pixels of image2 to use.
    * \param[in,out] shiftX The horizontal shift, positive to the right.
    * Initial guess and optimum value.
    * \param[in,out] shiftY The vertical shift, positive downwards.  Initial
    * guess and optimum value.
    * \return The minimum distance.
    * \sa distance(const image<typeCol, nbColComp>&, const image<typeCol, nbColComp>&, const pixelSelector&, short, short).
    */
   virtual double distanceMin(const image<typeCol, nbColComp>& image1,
                              const image<typeCol, nbColComp>& image2,
                              const pixelSelector& selector,
                              short* shiftX,
                              short* shiftY)
   {
      return (fabs(minimiser.minimise(image1, image2,
                                      selector,
                                      shiftX, shiftY)));
   }

   /*!
    * \brief Compute the minimum distance between a panoramic image and a set
    * of pixels from a second image.
    *
    * This is an overloaded version of distanceMin(const image<typeCol, nbColComp>&, const image<typeCol, nbColComp>&, const pixelSelector&, short*, short*).
    *
    * \param[in] image1 The panoramic image.
    * \param[in] image2 The second image from which the set of pixels is taken.
    * \param[in] selector The pixels of image2 to use.
    * \param[in,out] shift The shift, positive to the right for the horizontal
    * component, positive downwards for the vertical component.  Initial
    * guess and optimum value.
    * \return The minimum distance.
    */
   double distanceMin(const image<typeCol, nbColComp>& image1,
                      const image<typeCol, nbColComp>& image2,
                      const pixelSelector& selector,
                      point* shift)
   {
      short shiftX = static_cast<short>(round(shift->x));
      short shiftY = static_cast<short>(round(shift->y));
      double distMin = distanceMin(image1, image2, selector,
                                   &shiftX, &shiftY);
      shift->x = static_cast<double>(shiftX);
      shift->y = static_cast<double>(shiftY);
      return(distMin);
   }

protected:

private:

   /*!
    * \class xyMinimisation distancePanoramic.h imalib/distancePanoramic.h
    * \brief A class to perform the minimisation in
    * distancePanoramic_base::distanceMin(const image<typeCol, nbColComp>&, const image<typeCol, nbColComp>&, const pixelSelector&, short*, short*).
    *
    * The minimisation is done using GSL's gsl_multimin_fminimizer_nmsimplex2
    * method.
    */
   class xyMinimisation
   {
   public:

      /*!
       * \brief Constructor.
       * \param[in] distFunc The distance function to use.
       */
      xyMinimisation(distancePanoramic_base* distFunc)
      : x(NULL), dx(NULL), T_f(gsl_multimin_fminimizer_nmsimplex2), s_f(NULL)
      {
         gslAlloc();
         errorFunc_f.f = &distancePanoramic_base::xyMinimisation::error;

         minParams.distanceFunc = distFunc;
      }

      /*!
       * \brief Destructor.
       */
      ~xyMinimisation()
      {
         gslFree();
      }

      /*!
       * \brief Perform the minimisation.
       * \param[in] image1 The panoramic image.
       * \param[in] image2 The second image from which the set of pixels is taken.
       * \param[in] selector The pixels of image2 to use.
       * \param[in,out] shiftX The horizontal shift, positive to the right.
       * Initial guess and optimum value.
       * \param[in,out] shiftY The vertical shift, positive downwards.  Initial
       * guess and optimum value.
       * \return The minimum distance.
       * \sa distancePanoramic_base::distanceMin(const image<typeCol, nbColComp>&, const image<typeCol, nbColComp>&, const pixelSelector&, short*, short*).
       */
      double minimise(const image<typeCol, nbColComp>& image1,
                      const image<typeCol, nbColComp>& image2,
                      const pixelSelector& selector,
                      short* shiftX,
                      short* shiftY)
      {
         gsl_vector_set(x, 0, static_cast<double>(*shiftX));
         gsl_vector_set(x, 1, static_cast<double>(*shiftY));
         gsl_vector_set(dx, 0, 1.0); // delta of 1.
         gsl_vector_set(dx, 1, 1.0);

         shiftYMin = - static_cast<double>(selector.topLeftY());
         shiftYMax = static_cast<double>(image2.nbRows()
                                         - selector.bottomRightY());

         minParams.img1 = &image1;
         minParams.img2 = &image2;
         minParams.selector = &selector;
         errorFunc_f.params = &minParams;

         gsl_multimin_fminimizer_set(s_f, &errorFunc_f, x, dx);

         int iter = 0;
         int maxIter = 10;
         int status = 0;
         do
         {
            iter++;
            status = gsl_multimin_fminimizer_iterate(s_f);
            if (status)
            {
               printf("error: %s\n", gsl_strerror(status));
               break;
            }

            status = gsl_multimin_test_size(gsl_multimin_fminimizer_size(s_f),
                                            0.5);
         }
         while ((status == GSL_CONTINUE) && (iter < maxIter));

         //          if (status == GSL_SUCCESS)
         //          {
            //             std::cout << "Found minimum in " << iter
            //             << " iterations: error: "
            //             << s_f->fval
            //             << " size: " << gsl_multimin_fminimizer_size(s_f) << "."
            //             << std::endl;
            //          }

//          if (iter >= maxIter)
//          {
//             std::cout << "Reached iteration number limit: " << maxIter
//                << std::endl;
//          }

         *shiftX = static_cast<short>(round(gsl_vector_get(s_f->x, 0)));
         *shiftY = static_cast<short>(round(gsl_vector_get(s_f->x, 1)));

//          std::cout << "Error: " << s_f->fval
//             << " size: " << gsl_multimin_fminimizer_size(s_f) << " ("
//             << *shiftX << "," << *shiftY << ")" << std::endl;

         return(status == GSL_SUCCESS ? s_f->fval : - (s_f->fval));
      }

   private:

      /*!
       * \brief A structure to hold the parameters needed for the minimisation.
       *
       * It's all public but the class is private to
       * distancePanoramic_base::xyMinimisation.
       */
      class minimParams
      {
         public:

         /*!
          * \brief The distance function to use.
          */
         distancePanoramic_base* distanceFunc;

         /*!
          * \brief The first image.
          */
         const image<typeCol, nbColComp>* img1;

         /*!
          * \brief The second image.
          */
         const image<typeCol, nbColComp>* img2;

         /*!
          * \brief The pixel selector to select pixels from the second image.
          */
         const pixelSelector* selector;
      } minParams;

      /*!
       * \brief Function to compute the error (distance) between an image and
       * a selection of images from a second image with a horizontal and
       * vertical shift.
       *
       * Used by the minimisation in GSL.
       *
       * \param[in] v Point at which the function is evaluated.
       * \param[in] params Parameters of the optimisation (a minimParams).
       * \return The error at v.
       */
      static double error(const gsl_vector* v, void* params)
      {
         const double coef = 1000.0;
         minimParams* minParams = static_cast<minimParams*>(params);
         currentShiftX = gsl_vector_get(v, 0);
         currentShiftY = gsl_vector_get(v, 1);

         static double result = std::numeric_limits<double>::max();

         if (currentShiftY < shiftYMin)
         {
            result += coef * pow(shiftYMin - currentShiftY, 2.0);
         }
         else if (currentShiftY >= shiftYMax)
            result += coef * pow(currentShiftY - shiftYMax, 2.0);
         else
            result = minParams->distanceFunc->distance(*(minParams->img1),
                                                       *(minParams->img2),
                                                       *(minParams->selector),
                                                       currentShiftX,
                                                       currentShiftY);

         return(result);
      }

      /*!
       * \brief Allocate the memory for the GSL minimisation.
       */
      void gslAlloc()
      {
         gslFree();

         x = gsl_vector_alloc(2);
         dx = gsl_vector_alloc(2);
         s_f = gsl_multimin_fminimizer_alloc(T_f, 2);
         errorFunc_f.n = 2;
      }

      /*!
       * \brief Free the memory used by GSL for the minimisation.
       */
      void gslFree()
      {
         if (x)
         {
            gsl_vector_free(x);
            x = NULL;
         }
         if (dx)
         {
            gsl_vector_free(dx);
            dx = NULL;
         }
         if (s_f)
         {
            gsl_multimin_fminimizer_free(s_f);
            s_f = NULL;
         }
      }

      /*!
       * \brief GSL vector of minimisation positions.
       */
      gsl_vector* x;

      /*!
       * \brief GSL vector of minimisation step size.
       */
      gsl_vector* dx;

      /*!
       * \brief function to minimise.
       */
      gsl_multimin_function errorFunc_f;

      /*!
       * \brief GSL minimiser type.
       */
      const gsl_multimin_fminimizer_type* T_f;

      /*!
       * \brief GSL minimiser.
       */
      gsl_multimin_fminimizer* s_f;

      /*!
       * \brief The current value of shiftX.
       */
      static double currentShiftX;

      /*!
       * \brief The current value of shiftY.
       */
      static double currentShiftY;

      /*!
       * \brief The minimum value of shiftY
       */
      static double shiftYMin;

      /*!
       * \brief The maximum value of shiftY
       */
      static double shiftYMax;
   };

   /*!
    * \brief the minimiser to use for in
    * distancePanoramic_base::distanceMin(const image<typeCol, nbColComp>&, const image<typeCol, nbColComp>&, const pixelSelector&, short*, short*).
    */
   xyMinimisation minimiser;

};

/*
 * Declaration and initialisation of the static members of
 * distancePanoramic_base<typeCol, nbColComp, distMetr>::xyMinimisation
 */
template<typename typeCol, int nbColComp, typename distMetr>
double distancePanoramic_base<typeCol, nbColComp, distMetr>::
   xyMinimisation::currentShiftX = 0.0;

template<typename typeCol, int nbColComp, typename distMetr>
double distancePanoramic_base<typeCol, nbColComp, distMetr>::
   xyMinimisation::currentShiftY = 0.0;

template<typename typeCol, int nbColComp, typename distMetr>
double distancePanoramic_base<typeCol, nbColComp, distMetr>::
   xyMinimisation::shiftYMin = 0.0;

template<typename typeCol, int nbColComp, typename distMetr>
double distancePanoramic_base<typeCol, nbColComp, distMetr>::
   xyMinimisation::shiftYMax = 0.0;



/*!
 * \class distancePanoramic distancePanoramic.h imalib/distancePanoramic.h
 * \brief A class to compute the distance for entire panoramic images.
 * \author Fred Labrosse <ffl@aber.ac.uk>
 */
template <typename typeCol, int nbColComp, typename distMetr>
class distancePanoramic
         : public distancePanoramic_base<typeCol, nbColComp, distMetr>
{
public:

   /*!
    * \brief Default constructor.
    */
   distancePanoramic()
   {
   }

   /*
    * \brief Constructor.
    * \param[in] metr The functor implementing the distance metric.
    */
//    distancePanoramic(const distMetr& metr)
//          : distancePanoramic_base<typeCol, nbColComp, distMetr>(metr)
//    {
//    }

   /*!
    * \brief Destructor.
    */
   virtual ~distancePanoramic()
   {
   }

   /*!
    * \brief Compute the distance between two images.
    *
    * The second image can be shifted column-wise to represent a rotation of
    * the camera.
    *
    * The type typeCol must be casteble into a double.
    *
    * The size of the two images must be equal (this is only verified if
    * compiled without NDEBUG).
    *
    * \param[in] image1 The first image.
    * \param[in] image2 The second image.
    * \param[in] shift The shift, positive to the right.
    * \return The distance.
    */
   virtual double distance(const image<typeCol, nbColComp>& image1,
                           const image<typeCol, nbColComp>& image2,
                           short shift)
   {
      assert((image1.nbColumns() == image2.nbColumns())
             && (image1.nbRows() == image2.nbRows()));

      if (shift < 0)
         shift = image1.nbColumns() - ((-shift) % image1.nbColumns());
      else
         shift = shift % image1.nbColumns();

      unsigned long indShift, ind2, row2;
      unsigned long nbValWidth = image1.nbColumns() * nbColComp;
      unsigned long start = 0;

      short shiftVal = shift * nbColComp;

      this->metric.reset(&image1,&image2);

      const typeCol* pixels1 = image1.getActualPixel(0, 0);
      const typeCol* pixels2 = image2.getActualPixel(0, 0);

      for (unsigned short row = 0; row < image1.nbRows(); ++row)
      {
         row2 = row * nbValWidth;
         for (unsigned long ind = start; ind < nbValWidth; ind += nbColComp)
         {
            ind2 = ind + row2;
            indShift = ((ind + shiftVal) % nbValWidth) + row2;
            this->metric.comparePixels(&(pixels1[ind2]),
                                       &(pixels2[indShift]));
         }
      }

      return(this->metric.distance());
   }

   /*!
    * \brief Compute the distance between a panoramic image and a set of
    * pixels from a second image.
    *
    * This is similar in idea to what is performed in the class
    * distancePanoramicSelector but takes the pixelSelector as a parameter
    * rather than having it as a class member (therefore avoiding to have
    * to copy it if it changes).
    *
    * The set of pixels can be moved vertically (to cope with a change in
    * attitude of the panoramic camera) and horizontally (to cope with a
    * rotation of the panoramic camera).  The set of pixels has to be
    * vertically contained in the images but can wrap around horizontally.
    * The two images must have the same sizes.  This is only tested if
    * compiled without -NDEBUG.
    *
    * \param[in] image1 The panoramic image.
    * \param[in] image2 The second image from which the set of pixels is taken.
    * \param[in] selector The pixels of image2 to use.
    * \param[in] shiftX The horizontal shift, positive to the right.
    * \param[in] shiftY The vertical shift, positive downwards.
    * \return The distance.
    */
   virtual double distance(const image<typeCol, nbColComp>& image1,
                           const image<typeCol, nbColComp>& image2,
                           const pixelSelector& selector,
                           short shiftX,
                           short shiftY)
   {
      assert((image1.nbColumns() == image2.nbColumns())
             && (image1.nbRows() == image2.nbRows())
             && (image1.nbRows() >= selector.bottomRightY()));

      this->metric.reset();

      std::vector<pixelSelector::coords>::const_iterator iter;

      for (iter = selector.pixels().begin();
            iter < selector.pixels().end();
            ++iter)
      {
         this->metric.
            comparePixels(image1.getActualPixel(modulo(iter->X,
                                                       image1.nbColumns()),
                                                iter->Y),
                          image2.getActualPixel(
                             static_cast<unsigned short>(
                                modulo(static_cast<long>(iter->X) + shiftX,
                                       static_cast<long>(image2.nbColumns()))),
                                                iter->Y + shiftY));
      }

      return(this->metric.distance());
   }

}; // class distancePanoramic

/*!
 * \class distancePanoramicFrontBack distancePanoramic.h imalib/distancePanoramic.h
 * \brief A class to compute the distance between two panoramic images only
 * considering the parts corresponding to the front and back of the camera.
 *
 * This assumes that the front of the camera is centred at 3/4 of the width
 * of the panramic image from the left while the back of the camera is
 * centred at 1/4 of the width of the image from the left.
 *
 * \author Fred Labrosse <ffl@aber.ac.uk>
 */
template <class typeCol, int nbColComp, class distMetr>
class distancePanoramicFrontBack
         : public distancePanoramic_base<typeCol, nbColComp, distMetr>
{
public:

   /*!
    * \brief Constructor.
    *
    * The width of the field of view is in pixels.
    *
    * \param[in] width Width of the field of view.
    */
   distancePanoramicFrontBack(//const distMetr& metr,
                              unsigned short width = 60)
         : //distancePanoramic_base<typeCol, nbColComp, distMetr>(metr),
         frontBackWidth(width)
   {
   }

   /*!
    * \brief Set the field of view of the front and back of the camera.
    *
    * The width is in pixels.
    *
    * \param[in] width New width of the field of view.
    */
   void setFrontBackWidth(unsigned short width)
   {
      frontBackWidth = width;
   }

   /*!
    * \brief Get the field of view of the front and back of the camera.
    *
    * The width is in pixels.
    *
    * \return The width of the field of view.
    */
   unsigned short getFrontBackWidth()
   {
      return(frontBackWidth);
   }

   /*!
    * \brief Compute the distance between two images.
    *
    * The second image can be shifted column-wise to represent a rotation of
    * the camera.
    *
    * The type typeCol must be casteble into a double.
    *
    * The size of the two images must be equal (this is only verified if
    * compiled without NDEBUG).
    *
    * \param[in] image1 The first image.
    * \param[in] image2 The second image.
    * \param[in] shift The shift, positive to the right.
    * \return The distance.
    */
   virtual double distance(const image<typeCol, nbColComp>& image1,
                           const image<typeCol, nbColComp>& image2,
                           short shift)
   {
      assert((image1.nbColumns() == image2.nbColumns())
             && (image1.nbRows() == image2.nbRows()));

      if (shift < 0)
         shift = image1.nbColumns() - ((-shift) % image1.nbColumns());
      else
         shift = shift % image1.nbColumns();

      unsigned long indexShift, index, row2;
      unsigned long nbValWidth = image1.nbColumns() * nbColComp;

      short shiftVal = shift * nbColComp;

      this->metric.reset();

//       unsigned short maxRow = (this->rowMax > image1.nbRows())
//          ? image1.nbRows() : this->rowMax;
      // Note on the above line (and many subsequent ones): the "this->"
      // to access rowMax is there to solve the "two-phase name lookup"
      // problem.

      unsigned long start, end;

      unsigned short nbCols34 = image1.nbColumns() * 3 / 4;
      unsigned short nbCols4 = image1.nbColumns() / 4;
      unsigned short frontBackWidth2 = frontBackWidth / 2;

      const typeCol* pixels1 = image1.getActualPixel(0, 0);
      const typeCol* pixels2 = image2.getActualPixel(0, 0);

      for (unsigned short row = 0; row < image1.nbRows(); ++row)
      {
         row2 = row * nbValWidth;
         // Front.
         start = (nbCols34 - frontBackWidth2) * nbColComp;
         end = start + frontBackWidth * nbColComp;
         for (unsigned long ind = start; ind < end; ind += nbColComp)
         {
//             std::cout << ind / nbColComp << " " << row << std::endl;
            index = ind + row2;
            indexShift = ((ind + shiftVal) % nbValWidth) + row2;
            this->metric.comparePixels(&(pixels1[index]),
                                       &(pixels2[indexShift]));
         }
         // Back.
         start = (nbCols4 - frontBackWidth2) * nbColComp;
         end = start + frontBackWidth * nbColComp;
         for (unsigned long ind = start; ind < end; ind += nbColComp)
         {
//             std::cout << ind / nbColComp << " " << row << std::endl;
            index = ind + row2;
            indexShift = ((ind + shiftVal) % nbValWidth) + row2;
            this->metric.comparePixels(&(pixels1[index]),
                                       &(pixels2[indexShift]));
         }
      }

      return(this->metric.distance());
   }

protected:
private:

   /*!
    * \brief Width of the field of view of the front and back of the camera.
    *
    * The width is in pixels.
    */
   unsigned short frontBackWidth;

};

/*!
 * \class distancePanoramicSelector distancePanoramic.h imalib/distancePanoramic.h
 * \brief A class to compute the distance between two subsets of pixels of
 * panoramic images.
 *
 * This class uses a pixelSelector for the selection of the pixels making up
 * the subset.  Example selectors include random over the whole image or
 * random over the front/back areas of panoramic images.
 *
 * \author Fred Labrosse <ffl@aber.ac.uk>
*/
template <class typeCol, int nbColComp, class distMetr>
class distancePanoramicSelector
         : public distancePanoramic_base<typeCol, nbColComp, distMetr>
{
public:

   /*!
    * \brief Constructor.
    * \param[in] select The pixel selector to use.
    */
   distancePanoramicSelector(const pixelSelector& select)
         : selector(select)
   {
   }

   /*!
    * \brief Default Constructor.
    * 
    * A pixelSelector should be specified later otherwise no pixels will be
    * used in the distance computation!
    * 
    * \sa setSelector().
    */
   distancePanoramicSelector()
         : selector()
   {
   }

   /*!
    * \brief Set a new pixelSelector.
    * \param[in] select The new pixelSelector.
    */
   void setSelector(const pixelSelector& select)
   {
      selector = select;
   }

   /*!
    * \brief Compute the distance between subsets of pixels of two images.
    *
    * The second image can be shifted column-wise to represent a rotation of
    * the camera.
    *
    * \param[in] image1 The first image.
    * \param[in] image2 The second image.
    * \param[in] shift The shift, positive to the right.
    * \return The distance.
    */
   virtual double distance(const image<typeCol, nbColComp>& image1,
                           const image<typeCol, nbColComp>& image2,
                           short shift)
   {
      assert((image1.nbColumns() == image2.nbColumns())
             && (image1.nbRows() == image2.nbRows())
//              && (image1.nbColumns() > selector.bottomRightX())
             && (image1.nbRows() > selector.bottomRightY()));

      this->metric.reset(&image1,&image2);

      std::vector<pixelSelector::coords>::const_iterator iter;

      for (iter = selector.pixels().begin();
            iter < selector.pixels().end();
            ++iter)
      {
         this->metric.
            comparePixels(image1.getActualPixel(iter->X, iter->Y),
                          image2.getActualPixel(
                             static_cast<unsigned short>(
                                modulo(static_cast<long>(iter->X) + shift,
                                       static_cast<long>(image2.nbColumns()))),
                                                iter->Y));
      }

      return(this->metric.distance());
   }

   /*!
    * \brief Compute the distance between subsets of pixels of two images with
    * some specific parts being masked out.
    *
    * The second image can be shifted column-wise to represent a rotation of
    * the camera.
    *
    * \param[in] image1 The first image.
    * \param[in] image2 The second image.
    * \param[in] imageMask The image mask.
    * \param[in] shift The shift, positive to the right.
    * \return The normalised distance.
    * 
    * \author Juan Cao <juc3@aber.ac.uk>
    */
   double distance(const image<typeCol, nbColComp> &image1,
                   const image<typeCol, nbColComp> &image2,
                   const image<unsigned char, 3> &imageMask,
                   short shift)
   {
      assert((image1.nbColumns () == image2.nbColumns ())
             && (image1.nbRows () == image2.nbRows ())
             && (image1.nbRows () > selector.bottomRightY()));

      this->metric.reset(&image1, &image2);

      std::vector<pixelSelector::coords>::const_iterator iter;
      const unsigned char *pixel1, *pixel2;
//       double pixelValue1, pixelValue2;

      for (iter = selector.pixels().begin ();
           iter < selector.pixels().end ();
           ++iter)
      {
         pixel1 = imageMask.getActualPixel(iter->X, iter->Y);
         pixel2 =
            imageMask.getActualPixel(static_cast<unsigned short>(
               modulo(static_cast<long>(iter->X) + shift,
                      static_cast<long>(imageMask.nbColumns ()))),
                                     iter->Y);
            if ((pixel1[0] != 0) && (pixel2[0] != 0))
            {
               this->metric.comparePixels(
                  image1.getActualPixel(iter->X, iter->Y),
                  image2.getActualPixel(static_cast<unsigned short>(
                     modulo(static_cast<long>(iter-> X) + shift,
                            static_cast<long>(image2.nbColumns()))),
                                        iter->Y));
            }
      }

      return(this->metric.distanceNormalised());
   }

   /*!
    * \brief Get the number of accumulated comparisons (pixels).
    *
    * \return The number of accumulated comparisons (pixels)
    */
   unsigned long getNbPixels() const
   {
      return(this->metric.getNbPixels());
   }

   /*!
    * \brief Compute the distance between a panoramic image and a set of
    * pixels from a second image.
    *
    * The set of pixels can be moved vertically (to cope with a change in
    * attitude of the panoramic camera) and horizontally (to cope with a
    * rotation of the panoramic camera).  The set of pixels must remain within
    * the panoramic image (which is only tested if compiled without -NDEBUG).
    *
    * \param[in] image1 The panoramic image.
    * \param[in] image2 The second image from which the set of pixels is taken.
    * \param[in] shiftX The horizontal shift, positive to the right.
    * \param[in] shiftY The vertical shift, positive downwards.
    * \return The distance.
    */
   double distance(const image<typeCol, nbColComp>& image1,
                   const image<typeCol, nbColComp>& image2,
                   short shiftX,
                   short shiftY)
   {
      assert((static_cast<long>(selector.topLeftY()) + shiftY >= 0)
             && (static_cast<long>(selector.bottomRightY() + shiftY)
                 <= image2.nbRows())
             && (image1.nbColumns() >= selector.bottomRightX())
             && (image1.nbRows() >= selector.bottomRightY()));

      this->metric.reset();

      std::vector<pixelSelector::coords>::const_iterator iter;

      for (iter = selector.pixels().begin();
            iter < selector.pixels().end();
            ++iter)
      {
         this->metric.
            comparePixels(image1.getActualPixel(iter->X, iter->Y),
                          image2.getActualPixel(
                             static_cast<unsigned short>(
                                modulo(static_cast<long>(iter->X) + shiftX,
                                       static_cast<long>(image2.nbColumns()))),
                                                iter->Y + shiftY));
      }

      return(this->metric.distance());
   }


   /*!
    * \brief Compute the distance between two images with scaling factor
    *
    * The second image can be moved horizontally and vertically and scaled but
    * motion should be restricted so that the second image remains within the
    * first.  This is only checked if compiled without -DNDEBUG.
    *
    * The scaling is done around position (0,0), not around the centre of the
    * pixel selector.  In other words, the scaling displaces the centre of the
    * pixel selector.
    *
    * \param[in] image1 The first image.
    * \param[in] image2 The second image.
    * \param[in] shiftX
    * \param[in] shiftY The motion of the second image.
    * \param[in] shiftZ The scaling of the second image.
    * \return The distance.
    * \author James McIntyre <jvm5@aber.ac.uk>
    */
   double distanceXYScale(const image<typeCol, nbColComp>& image1,
                          const image<typeCol, nbColComp>& image2,
                          long shiftX,
                          long shiftY,
                          double shiftZ)
   {
      assert((static_cast<long>(selector.topLeftY()) + shiftY >= 0)
             && (static_cast<long>((selector.bottomRightY() * shiftZ)
                                   + shiftY)
                 <= image2.nbRows())
             && (image1.nbColumns() >= selector.bottomRightX())
             && (image1.nbRows() >= selector.bottomRightY()));

      this->metric.reset();

      std::vector<pixelSelector::coords>::const_iterator iter;

      for (iter = selector.pixels().begin();
            iter < selector.pixels().end();
            ++iter)
      {
         this->metric.
         comparePixels(image1.getActualPixel(iter->X, iter->Y),
                       image2.getActualPixel(
                          static_cast<unsigned short>(
                             modulo(round(static_cast<double>(iter->X)
                                          * shiftZ + shiftX),
                                    static_cast<long>(image2.nbColumns()))),
                          static_cast<unsigned short>(
                             round(static_cast<double>(iter->Y) * shiftZ
                                   + shiftY))));
      }

      return(this->metric.distance());
   }

protected:

private:

   /*!
    * \brief The pixelSelector to use in the distance computation.
    */
   pixelSelector selector;
};


/*!
 * \class distancePanoramicMapping distancePanoramic.h imalib/distancePanoramic.h
 * \brief A class to compute the distance between two panoramic images, with
 * an imageMapping applied to one of the images.
 *
 * A subset of the pixels can also be used by applying a pixelSelector to the
 * imageMapping.
 *
 * \author Fred Labrosse <ffl@aber.ac.uk>
 */
template <class typeCol, int nbColComp, class distMetr>
class distancePanoramicMapping
         : public distancePanoramic_base<typeCol, nbColComp, distMetr>
{
public:

   /*!
    * \brief Constructor.
    * \param[in] mapp The image mapping representing the
    * transformation.
    * \param[in] first If true, apply the mapping to the first image,
    * otherwise apply it to the second image.
    */
   distancePanoramicMapping(const imageMapping& mapp,
                            bool first = false)
         : mapping_(mapp),
          shiftFirstImage_(first)
   {
   }

   /*!
    * \brief Set a new mapping.
    * \param[in] mapp The new mapping.
    */
   void setMapping(const imageMapping& mapp)
   {
        mapping_.copy(mapp);
   }

   /*!
    * \brief Apply the shift to the first image.
    * \param[in] yes If true, then apply the shift to the first image,
    * otherwise, apply the shift to the second image.
    */
   void shiftFirstImage(bool yes = true)
   {
        shiftFirstImage_ = yes;
   }

   /*!
    * \brief Apply the shift to the second image.
    * \param[in] yes If true, then apply the shift to the second image,
    * otherwise, apply the shift to the first image.
    */
   void shiftSecondImage(bool yes = true)
   {
        shiftFirstImage_ = !yes;
   }

   /*!
    * \brief Compute the distance between two images with the mapping applied
    * to one of the two images.
    *
    * The type typeCol must be castable into a double.
    *
    * The size of the two images must be equal and compatible with the mapping
    * (this is only verified if compiled without NDEBUG).
    *
    * \sa applyToFirstImage(), applyToSecondImage().
    *
    * \param[in] image1 The first image.
    * \param[in] image2 The second image.
    * \param[in] shift The shift, positive to the right.
    * \return The distance.
    */
   virtual double distance(const image<typeCol, nbColComp>& image1,
                           const image<typeCol, nbColComp>& image2,
                           short shift)
   {
      return(distance(image1, image2, shift, true));
   }

   /*!
    * \brief Compute the distance between the two images with the mapping
    * applied or not to the second image.
    *
    * One of the two images can be shifted horizontally.
    *
    * The type typeCol must be castable into a double.
    *
    * The size of the two images must be equal and compatible with the mapping
    * (this is only verified if compiled without NDEBUG).
    *
    * \todo I have changed the "- shift"s for "+ shift"s because that makes
    * more sense. This may also mean that some stuff won't be working anymore.
    * Verify this.
    *
    * \sa shiftFirstImage(), shiftSecondImage().
    *
    * \param[in] image1 The first image.
    * \param[in] image2 The second image.
    * \param[in] shift The shift, positive to the right.
    * \param[in] apply If true, then apply the mapping.  Otherwise, don't apply
    * the mapping (but still only consider the pixels that would have been used
    * by the mapping).
    * \return The distance.
    */
   double distance(const image<typeCol, nbColComp>& image1,
                   const image<typeCol, nbColComp>& image2,
                   short shift,
                   bool apply)
   {
      assert((image1.nbColumns() == mapping_.sourceWidth())
             && (image1.nbRows() == mapping_.sourceHeight())
             && (mapping_.targetWidth() == image2.nbColumns())
             && (mapping_.targetHeight() == image2.nbRows()));

      if (shift < 0)
         shift = image1.nbColumns() - ((-shift) % image1.nbColumns());
      else
         shift = shift % image1.nbColumns();

      this->metric.reset();

      long xImageShift = 0;

      std::vector<imageMapping::coordsPair>::iterator iter;

      if (apply)
      {
         if (shiftFirstImage_)
         {
            for (iter = mapping_.map.begin();
                 iter < mapping_.map.end();
                 ++iter)
            {
               xImageShift =
                  static_cast<unsigned short>(modulo(
                                    static_cast<long>(iter->sX) + shift,
                                    static_cast<long>(image2.nbColumns())));
               this->metric.comparePixels(
                  image1.getActualPixel(xImageShift, iter->sY),
                  image2.getActualPixel(iter->tX, iter->tY));
            }
         }
         else
         {
            for (iter = mapping_.map.begin();
                 iter < mapping_.map.end();
                 ++iter)
            {
               xImageShift =
                  static_cast<unsigned short>(modulo(
                                 static_cast<long>(iter->tX) + shift,
                                 static_cast<long>(image2.nbColumns())));
               this->metric.comparePixels(
                  image1.getActualPixel(iter->sX, iter->sY),
                  image2.getActualPixel(xImageShift, iter->tY));
            }
         }
      }
      else
      {
         if (shiftFirstImage_)
         {
            for (iter = mapping_.map.begin();
                  iter < mapping_.map.end();
                  ++iter)
            {
               xImageShift =
                  static_cast<unsigned short>(modulo(
                                    static_cast<long>(iter->sX) + shift,
                                    static_cast<long>(image2.nbColumns())));
               this->metric.comparePixels(
                  image1.getActualPixel(xImageShift, iter->sY),
                  image2.getActualPixel(iter->sX, iter->sY));
            }
         }
         else
         {
            for (iter = mapping_.map.begin();
                  iter < mapping_.map.end();
                  ++iter)
            {
               xImageShift =
                  static_cast<unsigned short>(modulo(
                                       static_cast<long>(iter->sX) + shift,
                                       static_cast<long>(image2.nbColumns())));
               this->metric.comparePixels(
                  image1.getActualPixel(iter->sX, iter->sY),
                  image2.getActualPixel(xImageShift, iter->sY));
            }
         }
      }

      return(this->metric.distance());
   }

   /*!
    * \brief Compute the distance between two panoramic images with a pixel
    * mapping between the two.
    * 
    * The pixels from the second image can be shifted horizontally to simulate
    * a rotation of the camera and vertically to simulate a tilt of the camera.
    *
    * The pixels involved in the mapping must fit in both images with the
    * shift applied for the second image.  The horizontal coordinates are
    * wrapped around.  The two images must have the same width. This is only
    * verified if compiled without -DNDEBUG.
    *
    * \param[in] image1 The first image.
    * \param[in] image2 The second image.
    * \param[in] shiftX The horizontal shift, positive to the right.
    * \param[in] shiftY The vertical shift, positive downwards.
    * \return The distance between the two images.
    */
   double distance(const image<typeCol, nbColComp>& image1,
                   const image<typeCol, nbColComp>& image2,
                   short shiftX,
                   short shiftY)
   {
      assert((image1.nbColumns() == image2.nbColumns())
             && (image1.nbColumns() > mapping_.sourceMaxCol())
             && (image1.nbRows() > mapping_.sourceMaxRow())
             && (image2.nbColumns() > mapping_.targetMaxCol() + shiftY)
             && (static_cast<long>(mapping_.targetMinCol()) + shiftY > 0));

      if (shiftX < 0)
         shiftX = image1.nbColumns() - ((-shiftX) % image1.nbColumns());
      else
         shiftX = shiftX % image1.nbColumns();

      this->metric.reset();

      long xImageShift = 0;

      std::vector<imageMapping::coordsPair>::iterator iter;

      if (shiftFirstImage_)
      {
         for (iter = mapping_.map.begin();
              iter < mapping_.map.end();
              ++iter)
         {
            xImageShift =
            static_cast<unsigned short>(modulo(
               static_cast<long>(iter->sX) + shiftX,
               static_cast<long>(image2.nbColumns())));
            this->metric.comparePixels(
               image1.getActualPixel(xImageShift, iter->sY + shiftY),
               image2.getActualPixel(iter->tX, iter->tY + shiftY));
         }
      }
      else
      {
         for (iter = mapping_.map.begin();
              iter < mapping_.map.end();
              ++iter)
         {
            xImageShift =
            static_cast<unsigned short>(modulo(
               static_cast<long>(iter->tX) + shiftX,
               static_cast<long>(image2.nbColumns())));
            this->metric.comparePixels(
               image1.getActualPixel(iter->sX, iter->sY + shiftY),
               image2.getActualPixel(xImageShift, iter->tY + shiftY));
         }
      }

      return(this->metric.distance());
   }

   /*!
    * \brief Compute the difference between distances computed with and without
    * the mapping applied.
    *
    * This is useful to compute the change in distance when the mapping
    * simulates a displacement (as in [\ref BL06,\ref Lab07]).
    *
    * \param[in] img1 The first image.
    * \param[in] img2 The second image.
    * \return The gradient (difference between distance with and without
    * mapping applied.
    * \param[out] dist If non NULL, will contain the distance between the two
    * images without the mapping being applied.
    * \sa double gradientMapping(const image<typeCol, nbColComp>&,
    * const image<typeCol, nbColComp>&, short, bool)
    */
   double gradientMapping(const image<typeCol, nbColComp>& img1,
                          const image<typeCol, nbColComp>& img2,
                          double* dist = NULL)
   {
      return(gradientMapping(img1, img2, 0, dist));
   }

   /*!
    * \brief Compute the difference between distances computed with and without
    * the mapping applied.
    *
    * This is useful to compute the change in distance when the mapping
    * simulates a displacement (as in [\ref BL06,\ref Lab07]).
    *
    * This function is the same as
    * gradientMapping()
    * but allows the specification of a horizontal shift (corresponding to a
    * rotation of the camera) of the second image.
    *
    * \param[in] img1 The first image.
    * \param[in] img2 The second image.
    * \param[in] shift The shift to apply to the second image while computing
    * the gradient.
    * \param[out] dist If non NULL, will contain the distance between the two
    * images without the mapping being applied.
    * \return The gradient (difference between distance with and without
    * mapping applied.
    * \sa double gradientMapping(const image<typeCol, nbColComp>&,
    * const image<typeCol, nbColComp>&, bool)
    */
   double gradientMapping(const image<typeCol, nbColComp>& img1,
                          const image<typeCol, nbColComp>& img2,
                          short shift,
                          double* dist = NULL)
   {
      double aDist = distance(img1, img2, shift, false);
      if (dist)
         *dist = aDist;
      return(distance(img1, img2, shift, true)
             - aDist);
   }

protected:

private:

   /*!
    * \brief The imageMapping to apply to the second image when calculating
    * distances.
    */
   imageMapping mapping_;

   /*!
    * \brief If true, then apply the shift to the first image.  If false,
    * then apply the shift to the second image.
    * \todo This seems to be always true (even in visualNavigation).  Who/what
    * uses that?
    */
   bool shiftFirstImage_;

};


} // namespace imalib

#endif // IMALIB_DISTANCEPANORAMIC_H
