/*
 * imalib
 * 
 * Copyright 2017 Aberystwyth University
 * 
 * This library and program are free software: you can redistribute them
 * and/or modify them under the terms of the GNU General Public License as 
 * published by the Free Software Foundation, either version 3 of the License,
 * or (at your option) any later version.
 * 
 * The program and library are distributed in the hope that they will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*!
 * \file distanceRectangular.h
 * \brief Defines classes to manage a hierarchy of distance functions for
 * rectangular images.
 *
 * The distance functions are used when comparing images and correspond to
 * different metrics (colour spaces, restricted zones, Euclidean, Manhattan,
 * etc.).
 *
 * \author Fred Labrosse <ffl@aber.ac.uk>
 * \copyright 2017 Aberystwyth University
 */

#ifndef IMALIB_DISTANCERECTANGULAR_H
#define IMALIB_DISTANCERECTANGULAR_H

#include <math.h>

#include "imalib/appearance.h"
#include "imalib/distance.h"
#include "imalib/pixelSelector.h"
#include "imalib/transformation.h"

namespace imalib
{

/*!
 * \class distanceRectangular_base distanceRectangular.h imalib/distanceRectangular.h
 * \brief A base class to compute distances between "normal" rectangular
 * images.
 *
 * \author Fred Labrosse <ffl@aber.ac.uk>
 */
template <class typeCol, int nbColComp, class distMetr>
class distanceRectangular_base
            : public distance_base<typeCol, nbColComp, distMetr>
{
public:

   /*!
    * \brief Default constructor.
    */
   distanceRectangular_base()
   {
   }

   /*!
    * \brief Compute the distance between two images
    *
    * The second image can be moved horizontally and vertically but motion
    * should be restricted so that the second image remains within the first.
    * This is only checked if compiled without -DNDEBUG.
    *
    * \param[in] img1 The first image.
    * \param[in] img2 The second image.
    * \param[in] shiftX
    * \param[in] shiftY The motion of the second image.
    * \return The distance.
    */
   virtual double distance(const image<typeCol, nbColComp>& img1,
                           const image<typeCol, nbColComp>& img2,
                           long shiftX,
                           long shiftY) = 0;

   /*!
    * \brief Gradient in X and Y directions of the distance between the images
    * at a given position of the second image on the first image.
    *
    * The size, position and displacement of the second image must be
    * compatible with the size of the first image (see distance()).
    *
    * \param[in] img1 The first image.
    * \param[in] img2 The second image.
    * \param[in] shiftX
    * \param[in] shiftY The motion of the second image.
    * \param[in] deltaX
    * \param[in] deltaY The change in position to compute the gradient.
    * \param[out] gradX
    * \param[out] gradY The X and Y gradient.
    * \return The distance at the position.
    */
   double gradient(const image<typeCol, nbColComp>& img1,
                   const image<typeCol, nbColComp>& img2,
                   long shiftX,
                   long shiftY,
                   long deltaX,
                   long deltaY,
                   double* gradX,
                   double* gradY)
   {
      double distCentre = distance(img1, img2, shiftX, shiftY);
      double distDeltaX = distance(img1, img2,
                                   shiftX + deltaX,
                                   shiftY);
      double distDeltaY = distance(img1, img2,
                                   shiftX,
                                   shiftY + deltaY);

      *gradX = distCentre - distDeltaX;
      *gradY = distCentre - distDeltaY;

      return(distCentre);
   }

protected:

private:

};

/*!
 * \class distanceRectangular distanceRectangular.h imalib/distanceRectangular.h
 * \brief A class to compute the distance to "normal" rectangular images.
 * \author Fred Labrosse <ffl@aber.ac.uk>
 */
template <class typeCol, int nbColComp, class distMetr>
class distanceRectangular
            : public distanceRectangular_base<typeCol, nbColComp, distMetr>
{
public:

   /*!
    * \brief Default constructor.
    */
   distanceRectangular()
   {
   }

   /*!
    * \brief Compute the distance between two images
    *
    * The first image can be moved horizontally and vertically but motion
    * should be restricted so that it remains within the second.  The
    * first image is typically smaller than the second.
    * This is only checked if compiled without -DNDEBUG.
    *
    * \param[in] img1 The first image.
    * \param[in] img2 The second image.
    * \param[in] shiftX
    * \param[in] shiftY The motion of the first image.
    * \return The distance.
    */
   virtual double distance(const image<typeCol, nbColComp>& img1,
                           const image<typeCol, nbColComp>& img2,
                           long shiftX,
                           long shiftY)
   {
      assert((shiftX >= 0)
              && (shiftY >= 0)
              && (static_cast<long>(img1.nbColumns()) + shiftX
                  <= img2.nbColumns())
              && (static_cast<long>(img1.nbRows()) + shiftY
                  <= img2.nbRows()));

      this->metric.reset(&img1, &img2);

      for (unsigned short indY = 0; indY < img1.nbRows(); indY++)
      {
         for (unsigned short indX = 0; indX < img1.nbColumns(); indX++)
         {
            this->metric.
            comparePixels(img1.getActualPixel(indX, indY),
                          img2.getActualPixel(indX + shiftX,
                                              indY + shiftY));
         }
      }

      return(this->metric.distance());
   }

protected:

private:

};

/*!
 * \class distanceRectangularSelector distanceRectangular.h imalib/distanceRectangular.h
 * \brief A class to compute the distance between two subsets of pixels
 * of "normal" rectangular images.
 *
 * This class uses a pixelSelector for the selection of the pixels making up
 * the subset.  In the distance functions, the pixelSelector is used to select
 * a subset of the pixels of the first image and must be completly contained in
 * the second image.  In other words, the selection of the first image is
 * typically smaller than the second image.  This is only tested if not
 * compiled with -DNDEBUG.
 *
 * Example selectors include random over the whole image or random over the
 * front/back areas of panoramic images.
 *
 * \author Fred Labrosse <ffl@aber.ac.uk>
 */
template <class typeCol, int nbColComp, class distMetr>
class distanceRectangularSelector
   : public distanceRectangular_base<typeCol, nbColComp, distMetr>
{
public:

   /*!
    * \brief Constructor.
    *
    * The pixelSelector must be completely contained in the second image to
    * use in the distance functions.  This is only tested if not compiled
    * with -DNDEBUG.
    *
    * \param[in] select The pixelSelector to use.
    */
   distanceRectangularSelector(const pixelSelector& select)
   : selector(select)
   {
   }

   /*!
    * \brief Compute the distance between two images
    *
    * The selector positions can be moved horizontally and vertically on the
    * second image but motion should be restricted so that the positions remain
    * within the second image.  This is only checked if compiled without
    * -DNDEBUG.
    *
    * \param[in] img1 The first image.
    * \param[in] img2 The second image.
    * \param[in] shiftX
    * \param[in] shiftY The motion of the pixelSelector of the first image
    * (positive to the right/bottom).
    * \return The distance.
    */
   double distance(const image<typeCol, nbColComp>& img1,
                   const image<typeCol, nbColComp>& img2,
                   long shiftX,
                   long shiftY)
   {
      assert((static_cast<long>(selector.topLeftX()) + shiftX >= 0)
             && (static_cast<long>(selector.topLeftY()) + shiftY >= 0)
             && (static_cast<long>(selector.bottomRightX()) + shiftX
                       < img2.nbColumns())
             && (static_cast<long>(selector.bottomRightY()) + shiftY
                       < img2.nbRows()));

      this->metric.reset(&img1, &img2);

      std::vector<pixelSelector::coords>::iterator iter;

      for (iter = selector.pixels().begin();
           iter < selector.pixels().end();
           ++iter)
      {
         this->metric.
         comparePixels(img1.getActualPixel(iter->X, iter->Y),
                       img2.getActualPixel(static_cast<unsigned short>(
                          static_cast<long>(iter->X) + shiftX),
                                           static_cast<unsigned short>(
                                              static_cast<long>(iter->Y)
                                              + shiftY)));
      }

      return(this->metric.distance());
   }

   /*!
    * \brief Compute the Normalised distance between two images
    *
    * The selector positions can be moved horizontally and vertically on the
    * second image but motion should be restricted so that the positions remain
    * within the second image.  This is only checked if compiled without
    * -DNDEBUG.
    *
    * \param[in] img1 The first image.
    * \param[in] img2 The second image.
    * \param[in] shiftX
    * \param[in] shiftY The motion of the pixelSelector of the first image
    * (positive to the right/bottom).
    * \return The Normalised distance.
    */
   double distanceNormalised(const image<typeCol, nbColComp>& img1,
                             const image<typeCol, nbColComp>& img2,
                             long shiftX,
                             long shiftY)
   {
      assert((static_cast<long>(selector.topLeftX()) + shiftX >= 0)
             && (static_cast<long>(selector.topLeftY()) + shiftY >= 0)
             && (static_cast<long>(selector.bottomRightX()) + shiftX
                   < img2.nbColumns())
             && (static_cast<long>(selector.bottomRightY()) + shiftY
                       < img2.nbRows()));

      this->metric.reset(&img1, &img2);

      std::vector<pixelSelector::coords>::iterator iter;

      for (iter = selector.pixels().begin();
           iter < selector.pixels().end();
           ++iter)
      {
         this->metric.
             comparePixels(img1.getActualPixel(iter->X, iter->Y),
                           img2.getActualPixel(static_cast<unsigned short>(
                              static_cast<long>(iter->X) + shiftX),
                                               static_cast<unsigned short>(
                                                  static_cast<long>(iter->Y)
                                                  + shiftY)));
      }

      return(this->metric.distanceNormalised());
   }

   /*!
    * \brief Compute the distance between two images with scaling factor.
    *
    * The second image can be moved horizontally and vertically and scaled but
    * motion should be restricted so that the second image remains within the
    * first.  This is only checked if compiled without -DNDEBUG.
    *
    * The scaling is done around position (0,0), not around the centre of the
    * pixel selector.  In other words, the scaling displaces the centre of the
    * pixel selector.
    *
    * \param[in] img1 The first image.
    * \param[in] img2 The second image.
    * \param[in] shiftX
    * \param[in] shiftY The motion of the second image.
    * \param[in] shiftZ The scaling of the second image.
    * \return The distance.
    * \author James McIntyre <jvm5@aber.ac.uk>
    */
   double distanceXYScale(const image<typeCol, nbColComp>& img1,
                          const image<typeCol, nbColComp>& img2,
                          long shiftX,
                          long shiftY,
                          double shiftZ)
   {
      assert((static_cast<long>(selector.topLeftX()) + shiftX >= 0)
             && (static_cast<long>(selector.topLeftY()) + shiftY >= 0)
             && (static_cast<long>(selector.bottomRightX()) + shiftX
                 < img2.nbColumns())
             && (static_cast<long>(selector.bottomRightY()) + shiftY
                 < img2.nbRows()));

      this->metric.reset();

      std::vector<pixelSelector::coords>::iterator iter;

      for (iter = selector.pixels().begin();
           iter < selector.pixels().end();
           ++iter)
      {
         this->metric.
         comparePixels(img1.getActualPixel(iter->X, iter->Y),
                       img2.getActualPixel(
                          static_cast<unsigned short>(
                             round(static_cast<double>(iter->X) * shiftZ
                                   + shiftX)),
                           static_cast<unsigned short>(
                              round(static_cast<double>(iter->Y) * shiftZ
                              + shiftY))));
      }

      return(this->metric.distance());
   }

   /*!
    * \brief Compute the distance between two images
    *
    * The selector positions can be moved horizontally and vertically on the
    * second image but motion should be restricted so that the positions remain
    * within the second image. The selector positions can be rotated, this
    * should also be restricted so that the positions remain within the second
    * image. This is only checked if compiled without -DNDEBUG.
    *
    * \param[in] img1 The first image.
    * \param[in] img2 The second image.
    * \param[in] shiftX
    * \param[in] shiftY The motion of the second image (positive to the
    * right/bottom).
    * \param[in] rotation The degrees rotation of the second image
    * (anti-clockwise about the centre in degrees)
    * \return The distance.
    * \author Andrew Church <aec4@aber.ac.uk>
    */
   double distanceXYRotation(const image<typeCol, nbColComp>& img1,
                             const image<typeCol, nbColComp>& img2,
                             long shiftX,
                             long shiftY,
                             double rotation)
   {
      // preparation
      const double rotRadian = rotation * M_PI / 180.0;
      const double sineResult = sin(rotRadian);
      const double cosineResult = cos(rotRadian);
      const double midX = static_cast<double>(img1.nbColumns()) / 2.0;
      const double midY = static_cast<double>(img1.nbRows()) / 2.0;

      double xMinusMidX;
      double yMinusMidY;

      //assertions
#ifndef NDEBUG
      //top-left
      xMinusMidX = static_cast<double>(selector.topLeftX()) - midX;
      yMinusMidY = static_cast<double>(selector.topLeftY()) - midY;
      double x = (xMinusMidX * cosineResult) + (yMinusMidY * sineResult)
                 + midX + shiftX;
      double y = - ((xMinusMidX * sineResult) - (yMinusMidY * cosineResult))
                       + midY + shiftY;
      assert((x >= 0) && (x < img2.nbColumns())
              && (y >= 0) && (y < img2.nbRows()));
      //top-right
      xMinusMidX = static_cast<double>(selector.bottomRightX()) - midX;
      yMinusMidY = static_cast<double>(selector.topLeftY()) - midY;
      x = (xMinusMidX * cosineResult) + (yMinusMidY * sineResult)
           + midX + shiftX;
      y = - ((xMinusMidX * sineResult) - (yMinusMidY * cosineResult))
                + midY + shiftY;
      assert((x >= 0) && (x < img2.nbColumns())
               && (y >= 0) && (y < img2.nbRows()));
      //bottom-left
      xMinusMidX = static_cast<double>(selector.topLeftX()) - midX;
      yMinusMidY = static_cast<double>(selector.bottomRightY()) - midY;
      x = (xMinusMidX * cosineResult) + (yMinusMidY * sineResult)
           + midX + shiftX;
      y = -((xMinusMidX * sineResult) - (yMinusMidY * cosineResult))
                + midY + shiftY;
      assert((x >= 0) && (x < img2.nbColumns())
                   && (y >= 0) && (y < img2.nbRows()));
      //bottom-right
      xMinusMidX = static_cast<double>(selector.bottomRightX()) - midX;
      yMinusMidY = static_cast<double>(selector.bottomRightY()) - midY;
      x = (xMinusMidX * cosineResult) + (yMinusMidY * sineResult)
                + midX + shiftX;
      y = -((xMinusMidX * sineResult) - (yMinusMidY * cosineResult))
                + midY + shiftY;
      assert((x >= 0) && (x < img2.nbColumns())
                   && (y >= 0) && (y < img2.nbRows()));
#endif // ifndef NDEBUG

      this->metric.reset();

      std::vector<pixelSelector::coords>::iterator iter;
      unsigned short xTransf;
      unsigned short yTransf;

      for (iter = selector.pixels().begin();
           iter < selector.pixels().end();
           ++iter)
      {
         // Convert to mathematical coords with central origin and apply
         // rotation
         //
         // Convert back to computing coordinate system with origin at top left
         // and apply the shift
         //
         // Compare the pixels after rounding down.
         xMinusMidX = static_cast<double>(iter->X) - midX;
         yMinusMidY = static_cast<double>(iter->Y) - midY;
         xTransf =
            static_cast<unsigned short>(round((xMinusMidX * cosineResult
            + yMinusMidY * sineResult)
            + midX + shiftX));
         yTransf =
            static_cast<unsigned short>(round(-(xMinusMidX * sineResult
            - yMinusMidY * cosineResult)
            + midY + shiftY));

         this->metric.comparePixels(img1.getActualPixel(iter->X, iter->Y),
                                    img2.getActualPixel(xTransf, yTransf));
      }

      return(this->metric.distance());
   }

protected:

private:

   /*!
    * \brief The pixelSelector to use in the distance computation.
    */
   pixelSelector selector;
};


} // namespace imalib

#endif // IMALIB_DISTANCERECTANGULAR_H
