/*
 * imalib
 * 
 * Copyright 2017 Aberystwyth University
 * 
 * This library and program are free software: you can redistribute them
 * and/or modify them under the terms of the GNU General Public License as 
 * published by the Free Software Foundation, either version 3 of the License,
 * or (at your option) any later version.
 * 
 * The program and library are distributed in the hope that they will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*!
 * \file fileGrabber.h
 * \brief A class create a "frame grabber" that grabs images from files.
 *
 * \author Fred Labrosse <ffl@aber.ac.uk>
 * \copyright 2017 Aberystwyth University
 */

#ifndef FILEGRABBER_H
#define FILEGRABBER_H

#include "imalib/imageMapping.h"
#include "imalib/imageGrabber.h"

#include <string>
#include <vector>

namespace imalib
{
   class imageRGB;

   /*!
    * \class fileGrabber fileGrabber.h imalib/fileGrabber.h
    * \brief A class to instantiate an imageGrabber that takes the data from
    * files rather than an actual frame grabber.
    */
   class fileGrabber : public imageGrabber
   {
   public:
      /*!
       * \brief Constructor.
       *
       * A camera configuration file can be given to take relevant camera
       * settings (only upsideDown for now).  If the name is
       * empty, then the file is not used and default values are used.
       *
       * \param[in] names The names of the files.
       * \param[in] configName Name of the configuration file.
       */
      fileGrabber(const std::vector<std::string>& names,
                  const std::string& configName = "");

      /*!
       * \brief destructor.
       */
      virtual ~fileGrabber();

      /*!
       * \brief Display on the standard output some information about the
       * source of the images.
       *
       * \param[in,out] aStream The stream to which the info is added.
       */
      virtual void printInfoAboutSource(std::ostringstream* aStream);

      /*!
       * \brief Tell if we are connected to the required frame grabber.
       *
       * Here, we loose the connection when we are the end of the list of
       * names.
       *
       * \return true if we are connected, false otherwise.
       */
      virtual bool isConnected()
      {
         return (connected);
      }

      /*!
       * \brief Get an image from the next file.
       *
       * The image is resized here.
       *
       * \param[out] theImage The image in which to save the grabbed image,
       * if not NULL.
       * \param[out] rawImage The image in which to save the raw image (before
       * any transformation is applied), if not NULL.
       */
      virtual void getImage(imageRGB* theImage, imageRGB* rawImage = NULL);

      /*!
       * \brief Reset the next image to be grabbed to be the one at the
       * given index in the list of names.
       *
       * If the given index is out of range, then the grabber becomes
       * disconnected.
       *
       * \param[in] index The index of the next image to be grabbed.
       * \return true if we are connected, false otherwise.
       */
      bool resetImage(unsigned long index);

      /*!
       * \brief Get the number of images in the file grabber.
       * \return The number of images.
       */
      unsigned short getNbImages()
      {
         return(fileNames.size());
      }

      /*!
       * \brief Get the name of the file containing the image to be "grabbed"
       * next.
       *
       * I told you, this is the name of the image to be "grabbed" <b>NEXT</b>.
       *
       * \return The name of the image to be "grabbed" next.
       */
      std::string getNextImageName()
      {
         return(fileNames[imageIndex]);
      }

   protected:

      fileGrabber(const fileGrabber* imgTrans);
                                // We don't provide an implementation of
                                // that on purpose!
      fileGrabber& operator=(const fileGrabber& imgTrans);
                                // Same as above!

      std::vector<std::string> fileNames; // The names of the files from
                                          // which the images will be read.
      unsigned long imageIndex; // Index in the list of names of the last
                                // image that has been sent.
   };
} // namespace imalib


#endif

//  LocalWords:  param theImage imgWidth imgHeight imageMapping imageGrabber
//  LocalWords:  saveUnwrappingMapping fileGrabber
