/*
 * imalib
 * 
 * Copyright 2017 Aberystwyth University
 * 
 * This library and program are free software: you can redistribute them
 * and/or modify them under the terms of the GNU General Public License as 
 * published by the Free Software Foundation, either version 3 of the License,
 * or (at your option) any later version.
 * 
 * The program and library are distributed in the hope that they will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*!
 * \file image.h
 * \brief Defines classes to play with images.
 * \author Fred Labrosse <ffl@aber.ac.uk>
 * \copyright 2017 Aberystwyth University
 */

#ifndef IMALIB_IMAGE_H
#define IMALIB_IMAGE_H

#include <cassert>
#include <cstdlib>
#include <exception>
#include <fstream>
#include <iostream>
#include <map>
#include <sstream>

#include <string.h>

#ifdef WITH_PNG
#  include <png.h>
#endif // WITH_PNG

#include <imalib/imageMapping.h>
#include <imalib/utilities.h>

// Forward declaration.
class imageRGB;

namespace imalib
{
/*!
 * \class image image.h imalib/image.h
 * \brief A class to represent images.
 *
 * The class can represent the image and a thumbnail of the image.
 *
 * Pixels of the image have nbColComp components of type typeCol, both defined
 * at compilation time (using templates).  The actual data can be retrieved
 * using getActualPixel() and getActualThumbnailPixel().  This is not very
 * good OO programming, but this allows faster manipulation of the data,
 * including horrible things like memcpying data or even memmapping data.
 *
 * \author Fred Labrosse <ffl@aber.ac.uk>
 */
template <class typeCol, int nbColComp>
class image
{
public:

   /*!
    * \brief Default constructor.
    */
   image()
      : pixels_(NULL), thumbPixels_(NULL),
      nbColumns_(0), nbRows_(0),
      thumbNbColumns_(0), thumbNbRows_(0)
   {}

   /*!
    * \brief Copy constructor.
    *
    * \param[in] img The image to copy.
    */
   image(const image<typeCol, nbColComp>& img)
   : image()
   {
      copy(img);
   }

   /*!
    * \brief A constructor that specifies the size of the image.
    *
    * If any of the size parameters are negative or null, the corresponding
    * memory chunk will not be allocated.
    *
    * \param cols (IN) The number of columns.
    * \param rows (IN) The number of rows.
    * \param tCols (IN) The number of columns in the thumbnail.
    * \param tRows (IN) The number of rows in the thumbnail.
    * \sa setSize().
    */
   image(short cols, short rows, short tCols = 0, short tRows = 0)
         : pixels_(NULL), thumbPixels_(NULL),
         nbColumns_(0), nbRows_(0),
         thumbNbColumns_(0), thumbNbRows_(0)
   {
      setSize(cols, rows, tCols, tRows);
   }

   /*!
    * \brief Destructor.
    */
   virtual ~image()
   {
      delete [] pixels_;
      delete [] thumbPixels_;
   }

   /*!
    * \brief Assignment operator.
    * \param[in] img The image to copy.
    */
   image<typeCol, nbColComp>& operator=(const image<typeCol, nbColComp>& img)
   {
      copy(img);
      return(*this);
   }

   /*!
    * \brief Set the size of the image.
    *
    * The size of the image must strictly positive.  If it is null, the memory
    * is de-allocated.  The size of the thumbnail can be null, in which case no
    * thumbnail is created or the memory is de-allocated.
    *
    * If the size is different from the current size, memory is re-allocated.
    *
    * In any case, the image is cleared (THIS IS NOT TRUE ANYMORE, for efficiency
    * reasons!!!!).
    *
    * \param cols (IN) The number of columns.
    * \param rows (IN) The number of rows.
    * \param tCols (IN) The number of columns in the thumbnail.
    * \param tRows (IN) The number of rows in the thumbnail.
    * \sa setThumbnailSize().
    */
   void setSize(unsigned short cols, unsigned short rows,
                unsigned short tCols = 0, unsigned short tRows = 0)
   {
      if ((nbColumns_ != cols) || (nbRows_ != rows))
      {
         delete [] pixels_;
         if ((cols > 0) && (rows > 0))
         {
            nbColumns_ = cols;
            nbRows_ = rows;
            try
            {
               pixels_ = new typeCol[nbColumns_ * nbRows_ * nbColComp];
            }
            catch (std::bad_alloc& execp)
            {
               std::cerr << execp.what();
               abort();
            }
         }
         else
         {
            pixels_ = NULL;
            nbColumns_ = 0;
            nbRows_ = 0;
         }
      }
//    for (long ind = 0;
//          ind < nbColumns_ * nbRows_ * nbColComp;
//          ++ind)
//       pixels_[ind] = typeCol();

      setThumbnailSize(tCols, tRows);
   }

   /*!
    * \brief Set the size of the thumbnail image.
    *
    * The size of the image must strictly positive.  If it is null, the memory
    * is de-allocated.
    *
    * If the size is different from the current size, memory is re-allocated.
    *
    * \param tCols (IN) The number of columns in the thumbnail.
    * \param tRows (IN) The number of rows in the thumbnail.
    * \sa setSize().
    */
   void setThumbnailSize(unsigned short tCols,
                         unsigned short tRows)
   {
      if ((thumbNbColumns_ != tCols) || (thumbNbRows_ != tRows))
      {
         delete [] thumbPixels_;
         if ((tCols > 0) && (tRows > 0))
         {
            thumbNbColumns_ = tCols;
            thumbNbRows_ = tRows;
            try
            {
               thumbPixels_ = new typeCol[thumbNbColumns_ * thumbNbRows_
                                          * nbColComp];
            }
            catch (std::bad_alloc& execp)
            {
               std::cerr << execp.what();
               abort();
            }
         }
         else
         {
            thumbPixels_ = NULL;
            thumbNbColumns_ = 0;
            thumbNbRows_ = 0;
         }
      }
   }

   /*!
    * \brief Return the number of columns.
    */
   unsigned short nbColumns() const
   {
      return(nbColumns_);
   }

   /*!
    * \brief Return the number of rows.
    */
   unsigned short nbRows() const
   {
      return(nbRows_);
   }

   /*!
   \brief Return the number of columns of the thumbnail.
   */
   unsigned short nbThumbColumns() const
   {
      return(thumbNbColumns_);
   }

   /*!
   \brief Return the number of rows of the thumbnail.
   */
   unsigned short nbThumbRows() const
   {
      return(thumbNbRows_);
   }

   /*!
    * \brief Copy an image into the current one.
    *
    * \warning This function does not perform any conversion of format.
    *
    * \param[in] img The image to copy.
    */
   virtual void copy(const image<typeCol, nbColComp>& img)
   {
      setSize(img.nbColumns(), img.nbRows(),
              img.nbThumbColumns(), img.nbThumbRows());

      if (img.pixels_)
         memcpy(pixels_, img.getActualPixel(0, 0),
                nbColumns_ * nbRows_ * nbColComp * sizeof(typeCol));
      if (img.thumbPixels_)
         memcpy(thumbPixels_, img.getActualThumbnailPixel(0, 0),
                thumbNbColumns_ * thumbNbRows_ * nbColComp * sizeof(typeCol));
   }

   /*!
    * \brief Set a particular pixel of the image to the given value.
    *
    * Checks on the coordinates are only performed if compiled without NDEBUG.
    *
    * \param[in] col The column of the pixel.
    * \param[in] row The row of the pixel.
    * \param[in] vals The new value.
    * \sa setPixel(unsigned short, unsigned short, const typeCol*, bool).
    */
   void setPixel(unsigned short col, unsigned short row,
                 const typeCol* vals)
   {
      assert((col < nbColumns_) && (row < nbRows_));
      memcpy(&pixels_[(row * nbColumns_ + col) * nbColComp],
             vals,
             nbColComp * sizeof(typeCol));
   }

   /*!
    * \brief Set a particular pixel of the image to the given value.
    *
    * This assumes a panoramic (circular in x) image.
    *
    * Checks on the coordinates are only performed if compiled without NDEBUG.
    *
    * \param[in] col The column of the pixel.
    * \param[in] row The row of the pixel.
    * \param[in] vals The new value.
    * \sa setPixel(unsigned short, unsigned short, const typeCol*, bool).
    */
   void setPixel(long col, unsigned short row, const typeCol* vals)
   {
      assert(row < nbRows_);
      memcpy(&pixels_[(row * nbColumns_
                       + modulo(col, static_cast<long>(nbColumns_)))
                      * nbColComp],
             vals,
             nbColComp * sizeof(typeCol));
   }

   /*!
    * Set a particular pixel in the image to the given value.
    * 
    * This assumes a panoramic (circular in x) image.
    *
    * \param[in] col The column of the pixel.
    * \param[in] row The row of the pixel.
    * \param[in] vals The new value.
    * \param[in] swap If true then swap column and row.
    * \sa setPixel(unsigned short, unsigned short, const typeCol*).
    */
   void setPixel(long col, long row,
                 const typeCol* vals, bool swap)
   {
      if (swap)
         setPixel(row, col, vals);
      else
         setPixel(col, row, vals);
   }

   /*!
    * \brief Get the value of a particular pixel of the image.
    *
    * Checks on the coordinates are only performed if compiled without NDEBUG.
    *
    * \param[in] col The column of the pixel.
    * \param[in] row The row of the pixel.
    * \param[out] vals The array that will contain the values.  It must
    * be allocated somewhere else.
    */
   void getPixel(unsigned short col, unsigned short row,
                 typeCol* const vals) const
   {
      assert((col < nbColumns_) && (row < nbRows_));
      memcpy(vals,
             &pixels_[(row * nbColumns_ + col) * nbColComp],
             nbColComp * sizeof(typeCol));
   }


   /*!
   \brief Get the actual pixel of the image at a given position.

   \warning A pointer to the actual pixel is returned. Beware of
   aliasing problems!  <b>YOU HAVE BEEN WARNED!</b>

   This is not very good OO programming.  However, this allows efficient
   programs ;-).  In particular, the address returned for pixel (0,0) is the
   address of the array storing the image. This address can thus be used e.g. to
   memcpy or worse!

   Checks on the coordinates are only performed if compiled without NDEBUG.

   \param[in] col The column of the pixel.
   \param[in] row The row of the pixel.
   \return Address of the pixel.  The nbColComp positions from that
        address contain the colour components.
   */
   inline typeCol* getActualPixel(unsigned short col,
                                  unsigned short row)
   {
      assert((col < nbColumns_) && (row < nbRows_));
      return(&(pixels_[(row * nbColumns_ + col) * nbColComp]));
   }

   /*!
    * \brief Get the actual pixel of the image at a given position.
    *
    * \warning A pointer to the actual pixel is returned. Beware of
    * aliasing problems!  <b>YOU HAVE BEEN WARNED!</b>
    *
    * Checks on the coordinates are only performed if compiled without NDEBUG.
    *
    * \param[in] col The column of the pixel.
    * \param[in] row The row of the pixel.
    * \return Address of the pixel.  The nbColComp positions from that address
    * contain the colour components.
    */
   const typeCol* getActualPixel(unsigned short col,
                                 unsigned short row) const
   {
      assert((col < nbColumns_) && (row < nbRows_));
      return(&pixels_[(row * nbColumns_ + col) * nbColComp]);
   }

   /*!
    * \brief Set all the pixels of the image to be the given argument.
    * \param[in] newVal The new colour.
    */
   void setPixels(const typeCol* newVal)
   {
      for (long ind = 0;
           ind < (static_cast<long>(nbColumns_)
                  * static_cast<long>(nbRows_) * nbColComp);
           ind += nbColComp)
      {
         memcpy(&(pixels_[ind]), newVal, nbColComp * sizeof(typeCol));
      }
   }

   /*!
    * \brief Set a particular pixel of the thumbnail to the given value.
    *
    * Checks on the coordinates are only performed if compiled without NDEBUG.
    *
    * \param[in] col The column of the pixel.
    * \param[in] row The row of the pixel.
    * \param[in] vals The new value.
    */
   void setThumbnailPixel(unsigned short col,
                          unsigned short row,
                          const typeCol* vals)
   {
      assert((col < thumbNbColumns_) && (row < thumbNbRows_));
      memcpy(thumbPixels_[(row * thumbNbColumns_ + col) * nbColComp],
             vals,
             nbColComp * sizeof(typeCol));
   }

   /*!
   \brief Update the thumbnail.

   \param[in] blurWidth
   \param[in] blurHeight The amount of blur to use.
   \sa scaleBlur().
   */
   inline void updateThumbnail(unsigned short blurWidth = 1,
                               unsigned short blurHeight = 1)
   {
      scaleBlur(blurWidth, blurHeight,
                thumbNbColumns_, thumbNbRows_, thumbPixels_);
   }

   /*!
    * \brief Get the value of a particular pixel of the thumbnail.
    *
    * Checks on the coordinates are only performed if compiled without NDEBUG.
    *
    * \param col (IN) The column of the pixel.
    * \param row (IN) The row of the pixel.
    * \param vals (OUT) The array that will contain the values.
    */
   void getThumbnailPixel(unsigned short col,
                          unsigned short row,
                          const typeCol* const vals) const
   {
      assert((col < thumbNbColumns_) && (row < thumbNbRows_));
      memcpy(vals,
             thumbPixels_[(row * thumbNbColumns_ + col) * nbColComp],
             nbColComp * sizeof(typeCol));
   }

   /*!
    * \brief Get the actual pixel of the thumbnail at a given position.
    *
    * <b>WARNING!</b>: A pointer to the actual pixel is returned.  Beware of
    * aliasing problems!  <b>YOU HAVE BEEN WARNED!</b>
    *
    * This is not very good OO programming.  However, this allows efficient
    * programs ;-).  In particular, the address returned for pixel (0,0) is the
    * address of the array storing the thumbnail. This address can thus be used
    * e.g. to memcpy or worse!
    *
    * Checks on the coordinates are only performed if compiled without NDEBUG.
    *
    * \param col (IN) The column of the pixel.
    * \param row (IN) The row of the pixel.
    * \return A reference to the pixel.
    */
   typeCol* getActualThumbnailPixel(unsigned short col,
                                    unsigned short row)
   {
      assert((col < thumbNbColumns_) && (row < thumbNbRows_));
      return(&(thumbPixels_[(row * thumbNbColumns_ + col) * nbColComp]));
   }

   /*!
    * \brief Get the actual pixel of the thumbnail at a given position.
    *
    * <b>WARNING!</b>: A pointer to the actual pixel is returned.  Beware of
    * aliasing problems!  <b>YOU HAVE BEEN WARNED!</b>
    *
    * This is not very good OO programming.  However, this allows efficient
    * programs ;-).  In particular, the address returned for pixel (0,0) is the
    * address of the array storing the thumbnail. This address can thus be used
    * e.g. to memcpy or worse!
    *
    * Checks on the coordinates are only performed if compiled without NDEBUG.
    *
    * \param col (IN) The column of the pixel.
    * \param row (IN) The row of the pixel.
    * \return A reference to the pixel.
    */
   const typeCol* getActualThumbnailPixel(unsigned short col,
                                          unsigned short row) const
   {
      assert((col < thumbNbColumns_) && (row < thumbNbRows_));
      return(&(thumbPixels_[(row * thumbNbColumns_ + col) * nbColComp]));
   }

   /*!
    * \brief Apply a mapping to the image.
    *
    * The size imposed by the mapping on the source image is only checked if
    * compiled without NDEBUG.
    *
    * \param[in] mapping The imageMapping.
    * \param[out] result The result of the transformation.
    * \param[in] doit Apply the mapping only if doit is true.  This can be
    * useful to only keep the pixels transformed by the mapping without
    * actually changing the pixels.
    */
   void apply(const imageMapping& mapping,
              image<typeCol, nbColComp>* result,
              bool doit = true) const
   {
      assert((nbColumns_ == mapping.sourceWidth())
             && (nbRows_ == mapping.sourceHeight()));

      result->setSize(mapping.targetWidth(), mapping.targetHeight());

      std::vector<imageMapping::coordsPair>::const_iterator iter;
      if (doit)
      {
         for (iter = mapping.map.begin();
              iter < mapping.map.end();
              ++iter)
         {
            result->setPixel(iter->tX, iter->tY,
                             getActualPixel(iter->sX, iter->sY));
         }
      }
      else
      {
         for (iter = mapping.map.begin();
              iter < mapping.map.end();
              ++iter)
         {
            result->setPixel(iter->sX, iter->sY,
                             getActualPixel(iter->sX, iter->sY));
         }
      }
   }

   /*!
    * \brief Shift the image column-wise.
    *
    * This can simulate the rotation of an omni-directional camera around
    * its optical axis if the image is panoramic [\ref Lab04].
    *
    * \param[in] nbCol The number of columns to shift it by.  Negative
    * values shift to the left.  It is kept into the width of the image using
    * modulo().
    * \param[out] img2 The result.  It is resized here.
    */
   void shift(long nbCol, image<typeCol, nbColComp>* img2) const
   {
      img2->setSize(nbColumns_, nbRows_);

      nbCol = modulo(nbCol, static_cast<long>(nbColumns_));

      if (nbCol == 0)
      {
         img2->copy(*this);
      }
      else
      {
         unsigned short sh = (nbCol > 0 ? nbCol : nbColumns_ + nbCol);
         unsigned short shComp = nbColumns_ - sh;
         for (unsigned short row = 0; row < nbRows_; ++row)
         {
            memcpy(img2->getActualPixel(sh, row),
                   getActualPixel(0, row),
                   shComp * sizeof(typeCol) * nbColComp);
            memcpy(img2->getActualPixel(0, row),
                   getActualPixel(shComp, row),
                   sh * sizeof(typeCol) * nbColComp);
         }
      }
   }

   /*!
    * \brief Crop an image into the given one.
    *
    * The crop position and size are only checked if compiled without NDEBUG.
    *
    * \param[in] x
    * \param[in] y Position of the top left corner of the area to copy.
    * \param[in] w
    * \param[in] h Width and height of the area to copy.
    * \param[out] img2 The result.  It is resized here.
    */
   void crop(unsigned short x, unsigned short y,
             unsigned short w, unsigned short h,
             image<typeCol, nbColComp>* img2) const
   {
//       assert((x < nbColumns_) && (x + w < nbColumns_)
//              && (y < nbRows_) && (y + h < nbRows_));
     // Not sure why the assert above.  The one below seems more appropriate.
     assert((x + w < nbColumns_) && (y + h < nbRows_));

      img2->setSize(w, h);

      for (unsigned short indY = y; indY < y + h; indY++)
      {
         memcpy(img2->getActualPixel(0, indY - y),
                getActualPixel(x, indY),
                w * sizeof(typeCol) * nbColComp);
      }
   }

   /*!
    * \brief Crop a panoramic image into the given one.
    *
    * The crop position and size are only checked if compiled without NDEBUG.
    *
    * \param[in] x
    * \param[in] y Position of the top left corner of the area to copy.
    * \param[in] w
    * \param[in] h Width and height of the area to copy.
    * \param[out] img2 The result.  It is resized here.
    */
   void cropPanoramic(unsigned short x, unsigned short y,
                      unsigned short w, unsigned short h,
                      image<typeCol, nbColComp>* img2) const
   {
     assert(y + h < nbRows_);

     img2->setSize(w, h);

     long indX, indY;
     unsigned short indXCrop, indYCrop;
     for (indX = x, indXCrop = 0;
          indX < x + w;
          indX++, indXCrop++)
     {
       for (indY = y, indYCrop = 0;
            indY < y + h;
            indY++, indYCrop++)
       {
         img2->setPixel(indXCrop, indYCrop,
                        getActualPixel(modulo(indX,
                                              static_cast<long>(nbColumns_)),
                                       indY));
       }
     }
   }

   /*!
    * \brief Scale and/or blur an image.
    *
    * Transform the image so that it has a different size and/or is
    * blurred.  The blurring is done by taking the mean of the values in
    * a rectangle of the given size (reduced along the edges of the
    * image).  A size of 1 corresponds to no blurring, so a nearest
    * neighbour resampling in case of a change in scale.
    *
    * The type typeCol must be castable into a double and back.
    *
    * This member function cannot be constant (because it is used
    * internally to create the thumbnail).
    *
    * \param blurWidth (IN) The width of the neighbourhood for the
    * blurring (changed to the above odd number if even).
    * \param blurHeight (IN) The height of the neighbourhood for the
    * blurring (changed to the above odd number if even).
    * \param img (IN/OUT) The result.  The size of the result is the size
    * of img.
    */
   inline void scaleBlur(unsigned short blurWidth,
                         unsigned short blurHeight,
                         image<typeCol, nbColComp>* img) const
   {
      scaleBlur(blurWidth, blurHeight,
                img->nbColumns(), img->nbRows(),
                img->getActualPixel(0, 0));
   }

   /*!
    * \brief Draw a line in the image.
    *
    * This function comes from "Symmetric Double Step Line Algorithm" by Brian
    * Wyvill from "Graphics Gems", Academic Press, 1990
    *
    * \param[in] x1
    * \param[in] y1 Coordinates of the first point.
    * \param[in] x2
    * \param[in] y2 Coordinates of the second point.
    * \param[in] colour The colour.
    */
   void drawLine(unsigned short x1, unsigned short y1,
                 unsigned short x2, unsigned short y2,
                 const typeCol* colour)
   {
#define swap(a, b) {a ^= b; b^= a; a^= b;}
#define absolute(i, j, k) ((i - j) * (k = ((i - j) < 0 ? -1 : 1)))

      short incr1, incr2, D, c, xend, pixelsLeft;
      unsigned short x, y, dx, dy;
      short signX, signY, step, i;
      bool reverse;

      dx = (unsigned short)absolute(long(x2), long(x1), signX);
      dy = (unsigned short)absolute(long(y2), long(y1), signY);

      // decide increment sign by the slope sign
      if (signX == signY)
         step = 1;
      else
         step = -1;

      if (dy > dx)
      {                // chooses axis of greatest movement (make dx)
         swap(x1, y1);
         swap(x2, y2);
         swap(dx, dy);
         reverse = true;
      }
      else
         reverse = false;
      // note error check for dx==0 should be included here
      if (x1 > x2)
      {                            // start from the smaller coordinate
         x = x2;
         y = y2;
      }
      else
      {
         x = x1;
         y = y1;
         x1 = x2;
         y1 = y2;
      }

      // Note dx=n implies 0 - n or (dx+1) pixels to be set.  Go round loop
      // dx/4 times then plot last 0,1,2 or 3 pixels.  In fact (dx-1)/4 as 2
      // pixels are already plotted
      xend = (dx - 1) / 4;
      pixelsLeft = (dx - 1) % 4;   // number of pixels left over at the end.
      setPixel(x, y, colour, reverse);
      if (pixelsLeft < 0)
         return ;       // plot only one pixel for zero
         // length vectors.
         setPixel(x1, y1, colour, reverse);
      // plot first two points
      incr2 = 4 * dy - 2 * dx;
      if (incr2 < 0)
      {                            // slope less than 1/2.
         c = 2 * dy;
         incr1 = 2 * c;
         D = incr1 - dx;

         for (i = 0; i < xend; i++)
         {                         // plotting loop.
            ++x;
            --x1;
            if (D < 0)
            {
               // pattern 1 forward.
               setPixel(x, y, colour, reverse);
               setPixel(++x, y, colour, reverse);
               // pattern 1 backward.
               setPixel(x1, y1, colour, reverse);
               setPixel(--x1, y1, colour, reverse);
               D += incr1;
            }
            else
            {
               if (D < c)
               {
                  // pattern 2 forward.
                  setPixel(x, y, colour, reverse);
                  setPixel(++x, y += step, colour, reverse);
                  // pattern 2 backward.
                  setPixel(x1, y1, colour, reverse);
                  setPixel(--x1, y1 -= step, colour, reverse);
               }
               else
               {
                  // pattern 3 forwards.
                  setPixel(x, y += step, colour, reverse);
                  setPixel(++x, y, colour, reverse);
                  // pattern 3 backwards.
                  setPixel(x1, y1 -= step, colour, reverse);
                  setPixel(--x1, y1, colour, reverse);
               }
               D += incr2;
            }
         }         // end for.

         // plot last pattern.
         if (pixelsLeft)
         {
            if (D < 0)
            {
               setPixel(++x, y, colour, reverse);
               // pattern 1.
               if (pixelsLeft > 1)
                  setPixel(++x, y, colour, reverse);
               if (pixelsLeft > 2)
                  setPixel(--x1, y1, colour, reverse);
            }
            else
            {
               if (D < c)
               {
                  setPixel(++x, y, colour, reverse);
                  // pattern 2.
                  if (pixelsLeft > 1)
                     setPixel(++x, y += step, colour, reverse);
                  if (pixelsLeft > 2)
                     setPixel(--x1, y1, colour, reverse);
               }
               else
               {
                  // pattern 3.
                  setPixel(++x, y += step, colour, reverse);
                  if (pixelsLeft > 1)
                     setPixel(++x, y, colour, reverse);
                  if (pixelsLeft > 2)
                     setPixel(--x1, y1 -= step, colour, reverse);
               }
            }
         } // end if pixelsLeft.
      } // end slope < 1/2.
      else
      {                            // slope greater than 1/2.
         c = 2 * (dy - dx);
         incr1 = 2 * c;
         D = incr1 + dx;
         for (i = 0; i < xend; i++)
         {
            ++x;
            --x1;
            if (D > 0)
            {
               // pattern 4 forwards.
               setPixel(x, y += step, colour, reverse);
               setPixel(++x, y += step, colour, reverse);
               // pattern 4 backwards.
               setPixel(x1, y1 -= step, colour, reverse);
               setPixel(--x1, y1 -= step, colour, reverse);
               D += incr1;
            }
            else
            {
               if (D < c)
               {
                  // pattern 2 forwards
                  setPixel(x, y, colour, reverse);
                  setPixel(++x, y += step, colour, reverse);

                  // pattern 2 backwards.
                  setPixel(x1, y1, colour, reverse);
                  setPixel(--x1, y1 -= step, colour, reverse);
               }
               else
               {
                  // pattern 3 forwards.
                  setPixel(x, y += step, colour, reverse);
                  setPixel(++x, y, colour, reverse);
                  // pattern 3 backwards.
                  setPixel(x1, y1 -= step, colour, reverse);
                  setPixel(--x1, y1, colour, reverse);
               }
               D += incr2;
            }
         } // end for
         // plot last pattern.
         if (pixelsLeft)
         {
            if (D > 0)
            {
               setPixel(++x, y += step, colour, reverse);
               // pattern 4.
               if (pixelsLeft > 1)
                  setPixel(++x, y += step, colour, reverse);
               if (pixelsLeft > 2)
                  setPixel(--x1, y1 -= step, colour, reverse);
            }
            else
            {
               if (D < c)
               {
                  setPixel(++x, y, colour, reverse);
                  // pattern 2.
                  if (pixelsLeft > 1)
                     setPixel(++x, y += step, colour, reverse);
                  if (pixelsLeft > 2)
                     setPixel(--x1, y1, colour, reverse);
               }
               else
               {
                  // pattern 3.
                  setPixel(++x, y += step, colour, reverse);
                  if (pixelsLeft > 1)
                     setPixel(++x, y, colour, reverse);
                  if (pixelsLeft > 2)
                  {
                     if (D > c)    // step 3.
                        setPixel(--x1, y1 -= step, colour, reverse);
                     else          // step 2.
                        setPixel(--x1, y1, colour, reverse);
                  }
               }
            }
         }
      }
#undef swap
#undef absolute
   }

   /*!
    * \brief Draw a line in the image.
    * 
    * This assumes the image is panoramic (circular x) so lines are drawn as
    * such.
    *
    * This function comes from "Symmetric Double Step Line Algorithm" by Brian
    * Wyvill from "Graphics Gems", Academic Press, 1990
    *
    * \param[in] x1
    * \param[in] y1 Coordinates of the first point.
    * \param[in] x2
    * \param[in] y2 Coordinates of the second point.
    * \param[in] colour The colour.
    */
   void drawLine(long x1, unsigned short y1,
                 long x2, unsigned short y2,
                 const typeCol* colour)
   {
#define swap(a, b) {a ^= b; b^= a; a^= b;}
#define absolute(i, j, k) ((i - j) * (k = ((i - j) < 0 ? -1 : 1)))

      short incr1, incr2, D, c, xend, pixelsLeft;
      long x, y, dx, dy;
      short signX, signY, step, i;
      bool reverse;
      long x1L = x1;
      long y1L = y1;
      long x2L = x2;
      long y2L = y2;

      dx = absolute(x2L, x1L, signX);
      dy = absolute(y2L, y1L, signY);

      // decide increment sign by the slope sign
      if (signX == signY)
         step = 1;
      else
         step = -1;

      if (dy > dx)
      {                // chooses axis of greatest movement (make dx)
         swap(x1L, y1L);
         swap(x2L, y2L);
         swap(dx, dy);
         reverse = true;
      }
      else
         reverse = false;
      // note error check for dx==0 should be included here
      if (x1L > x2L)
      {                            // start from the smaller coordinate
         x = x2L;
         y = y2L;
      }
      else
      {
         x = x1L;
         y = y1L;
         x1L = x2L;
         y1L = y2L;
      }

      // Note dx=n implies 0 - n or (dx+1) pixels to be set.  Go round loop
      // dx/4 times then plot last 0,1,2 or 3 pixels.  In fact (dx-1)/4 as 2
      // pixels are already plotted
      xend = (dx - 1) / 4;
      pixelsLeft = (dx - 1) % 4;   // number of pixels left over at the end.
      setPixel(x, y, colour, reverse);
      if (pixelsLeft < 0)
         return ;       // plot only one pixel for zero
      // length vectors.
      setPixel(x1L, y1L, colour, reverse);
      // plot first two points
      incr2 = 4 * dy - 2 * dx;
      if (incr2 < 0)
      {                            // slope less than 1/2.
         c = 2 * dy;
         incr1 = 2 * c;
         D = incr1 - dx;

         for (i = 0; i < xend; i++)
         {                         // plotting loop.
            ++x;
            --x1L;
            if (D < 0)
            {
               // pattern 1 forward.
               setPixel(x, y, colour, reverse);
               setPixel(++x, y, colour, reverse);
               // pattern 1 backward.
               setPixel(x1L, y1L, colour, reverse);
               setPixel(--x1L, y1L, colour, reverse);
               D += incr1;
            }
            else
            {
               if (D < c)
               {
                  // pattern 2 forward.
                  setPixel(x, y, colour, reverse);
                  setPixel(++x, y += step, colour, reverse);
                  // pattern 2 backward.
                  setPixel(x1L, y1L, colour, reverse);
                  setPixel(--x1, y1L -= step, colour, reverse);
               }
               else
               {
                  // pattern 3 forwards.
                  setPixel(x, y += step, colour, reverse);
                  setPixel(++x, y, colour, reverse);
                  // pattern 3 backwards.
                  setPixel(x1L, y1L -= step, colour, reverse);
                  setPixel(--x1L, y1L, colour, reverse);
               }
               D += incr2;
            }
         }         // end for.

         // plot last pattern.
         if (pixelsLeft)
         {
            if (D < 0)
            {
               setPixel(++x, y, colour, reverse);
               // pattern 1.
               if (pixelsLeft > 1)
                  setPixel(++x, y, colour, reverse);
               if (pixelsLeft > 2)
                  setPixel(--x1L, y1L, colour, reverse);
            }
            else
            {
               if (D < c)
               {
                  setPixel(++x, y, colour, reverse);
                  // pattern 2.
                  if (pixelsLeft > 1)
                     setPixel(++x, y += step, colour, reverse);
                  if (pixelsLeft > 2)
                     setPixel(--x1L, y1L, colour, reverse);
               }
               else
               {
                  // pattern 3.
                  setPixel(++x, y += step, colour, reverse);
                  if (pixelsLeft > 1)
                     setPixel(++x, y, colour, reverse);
                  if (pixelsLeft > 2)
                     setPixel(--x1L, y1L -= step, colour, reverse);
               }
            }
         } // end if pixelsLeft.
      } // end slope < 1/2.
      else
      {                            // slope greater than 1/2.
         c = 2 * (dy - dx);
         incr1 = 2 * c;
         D = incr1 + dx;
         for (i = 0; i < xend; i++)
         {
            ++x;
            --x1;
            if (D > 0)
            {
               // pattern 4 forwards.
               setPixel(x, y += step, colour, reverse);
               setPixel(++x, y += step, colour, reverse);
               // pattern 4 backwards.
               setPixel(x1L, y1L -= step, colour, reverse);
               setPixel(--x1L, y1L -= step, colour, reverse);
               D += incr1;
            }
            else
            {
               if (D < c)
               {
                  // pattern 2 forwards
                  setPixel(x, y, colour, reverse);
                  setPixel(++x, y += step, colour, reverse);

                  // pattern 2 backwards.
                  setPixel(x1L, y1L, colour, reverse);
                  setPixel(--x1L, y1L -= step, colour, reverse);
               }
               else
               {
                  // pattern 3 forwards.
                  setPixel(x, y += step, colour, reverse);
                  setPixel(++x, y, colour, reverse);
                  // pattern 3 backwards.
                  setPixel(x1L, y1L -= step, colour, reverse);
                  setPixel(--x1L, y1L, colour, reverse);
               }
               D += incr2;
            }
         } // end for
         // plot last pattern.
         if (pixelsLeft)
         {
            if (D > 0)
            {
               setPixel(++x, y += step, colour, reverse);
               // pattern 4.
               if (pixelsLeft > 1)
                  setPixel(++x, y += step, colour, reverse);
               if (pixelsLeft > 2)
                  setPixel(--x1L, y1L -= step, colour, reverse);
            }
            else
            {
               if (D < c)
               {
                  setPixel(++x, y, colour, reverse);
                  // pattern 2.
                  if (pixelsLeft > 1)
                     setPixel(++x, y += step, colour, reverse);
                  if (pixelsLeft > 2)
                    setPixel(--x1L, y1L, colour, reverse);
               }
               else
               {
                  // pattern 3.
                  setPixel(++x, y += step, colour, reverse);
                  if (pixelsLeft > 1)
                     setPixel(++x, y, colour, reverse);
                  if (pixelsLeft > 2)
                  {
                     if (D > c)    // step 3.
                        setPixel(--x1L, y1L -= step, colour, reverse);
                     else          // step 2.
                        setPixel(--x1L, y1L, colour, reverse);
                  }
               }
            }
         }
      }
#undef swap
#undef absolute
   }

   /*!
    * \brief Draw an axis aligned box.
    * \param[in] xPos
    * \param[in] yPos Position of the top left corner.
    * \param[in] width
    * \param[in] height Size of the box.
    * \param[in] colour The colour.
    */
   void drawAABox(unsigned short xPos, unsigned short yPos,
                  unsigned short width, unsigned short height,
                  const typeCol* colour)
   {
      drawLine(xPos, yPos, xPos + width, yPos, colour);
      drawLine(xPos + width, yPos, xPos + width, yPos + height, colour);
      drawLine(xPos + width, yPos + height, xPos, yPos + height, colour);
      drawLine(xPos, yPos + height, xPos, yPos, colour);
   }

   /*!
    * \brief Draw an axis aligned box.
    * 
    * It assumes the image is panoramic (circular in x).
    *
    * \param[in] xPos
    * \param[in] yPos Position of the top left corner.
    * \param[in] width
    * \param[in] height Size of the box.
    * \param[in] colour The colour.
    */
   void drawAABox(long xPos, unsigned short yPos,
                  unsigned short width, unsigned short height,
                  const typeCol* colour)
   {
      drawLine(xPos, yPos,
               static_cast<long>(xPos + width), yPos, colour);
      drawLine(xPos + width, yPos,
               static_cast<long>(xPos + width), yPos + height, colour);
      drawLine(static_cast<long>(xPos + width), yPos + height,
               xPos, yPos + height, colour);
      drawLine(xPos, yPos + height,
               xPos, yPos, colour);
   }

protected:

   /*!
    * \brief Scale and/or blur an image.
    *
    * Transform the image so that it has a different size and/or is
    * blurred.  The blurring is done by taking the mean of the values in
    * a rectangle of the given size (reduced along the edges of the
    * image).  A size of 1 corresponds to no blurring, so a nearest
    * neighbour resampling in case of a change in scale.
    *
    * The type typeCol must be castable into a double and back.
    *
    * This is called by the public version for images and is used to
    * produce the thumbnails (which also means it cannot be constant).
    *
    * \param blurWidth (IN) The width of the neighbourhood for the
    * blurring (changed to the above odd number if even).
    * \param blurHeight (IN) The height of the neighbourhood for the
    * blurring (changed to the above odd number if even).
    * \param cols
    * \param rows (IN) The size of the new image.
    * \param img (OUT) The result.  Must have been allocated beforehand.
    */
   void scaleBlur(unsigned short blurWidth,
                  unsigned short blurHeight,
                  unsigned short cols, unsigned short rows,
                  typeCol* img) const
   {
      // To convert from new coordinate to old.
      double xScale = static_cast<double>(nbColumns_)
         / static_cast<double>(cols);
      double yScale = static_cast<double>(nbRows_)
         / static_cast<double>(rows);
      double xShift = 0;
      if (cols < nbColumns_)
         xShift = xScale * 0.5 + 0.5;
      double yShift = 0;
      if (rows < nbRows_)
         yShift = yScale * 0.5 + 0.5;

      // Pixel operations.
      double* values = new double[nbColComp];
      for (unsigned short ind = 0; ind < nbColComp; ++ind)
      {
         values[ind] = 0.0;
      }
      const typeCol* pixCol;
      const typeCol* pixRow;

      if ((blurWidth > 1) || (blurHeight > 1))
      {                         // Do blur
         unsigned short posX, posXMin, posXMax;
         unsigned short posY, posYMin, posYMax;

         long blurWidth2 = blurWidth / 2;
         long blurHeight2 = blurHeight / 2;

         // All pixels of the new image.
         for (unsigned short indY = 0; indY < rows; ++indY)
         {
            for (unsigned short indX = 0; indX < cols; ++indX)
            {
               // Position in the old image.
               posX = static_cast<long>(static_cast<double>(indX)
                                        * xScale + xShift);
               posY = static_cast<long>(static_cast<double>(indY)
                                        * yScale + yShift);

               long tmpPos = posX - blurWidth2;
               if (tmpPos < 0)
                  posXMin = 0;
               else
                  posXMin = tmpPos;
               posXMax = posX + blurWidth2;
               if (posXMax >= nbColumns_)
                  posXMax = nbColumns_ - 1;
               tmpPos = posY - blurHeight2;
               if (tmpPos < 0)
                  posYMin = 0;
               else
                  posYMin = tmpPos;
               posYMax = posY + blurHeight2;
               if (posYMax >= nbRows_)
                  posYMax = nbRows_ - 1;

               pixRow = getActualPixel(posXMin, posYMin);
               for (unsigned short wY = posYMin;
                    wY <= posYMax;
                    ++wY)
               {
                  pixCol = pixRow;
                  for (unsigned short wX = posXMin;
                       wX <= posXMax;
                       ++wX)
                  {
                     for (unsigned short ind = 0; ind < nbColComp; ++ind)
                     {
                        values[ind] += static_cast<double>(*(pixCol++));
                     }
                  }
                  pixRow += (nbColumns_ * nbColComp);
               }

               double nbPixels = (posXMax - posXMin + 1)
                  * (posYMax - posYMin + 1);
               for (unsigned short ind = 0; ind < nbColComp; ++ind)
               {
                  *(img++) =
                     static_cast<typeCol>(values[ind] / nbPixels + 0.5);
                  values[ind] = 0.0;
               }
            }
         }
      }
      else
      {                      // Do not blur
         unsigned short posX, posY;
         // All pixels of the new image.
         for (unsigned short indY = 0; indY < rows; ++indY)
         {
            for (unsigned short indX = 0; indX < cols; ++indX)
            {
               // Position in the old image.
               posX = static_cast<unsigned short>(static_cast<double>(indX)
                  * xScale + xShift);
               posY = static_cast<unsigned short>(static_cast<double>(indY)
                  * yScale + yShift);

               pixRow = getActualPixel(posX, posY);
               for (unsigned short ind = 0; ind < nbColComp; ++ind)
               {
                  *(img++) = *(pixRow++);
               }
            }
         }
      }

      delete [] values;
   }

private:

   /*!
     \brief The pixels of the image.
   */
   typeCol* pixels_;

   /*!
     \brief The pixels of the thumbnail.
   */
   typeCol* thumbPixels_;

   /*!
     \brief The horizontal size of the image (number of columns).
   */
   unsigned short nbColumns_;

   /*!
     \brief The vertical size of the image (number of rows).
   */
   unsigned short nbRows_;

   /*!
     \brief The horizontal size of the thumbnail.
   */
   unsigned short thumbNbColumns_;

   /*!
     \brief The vertical size of the thumbnail.
   */
   unsigned short thumbNbRows_;
};

} // namespace imalib

#endif // IMALIB_IMAGE_H
