/*
 * imalib
 * 
 * Copyright 2017 Aberystwyth University
 * 
 * This library and program are free software: you can redistribute them
 * and/or modify them under the terms of the GNU General Public License as 
 * published by the Free Software Foundation, either version 3 of the License,
 * or (at your option) any later version.
 * 
 * The program and library are distributed in the hope that they will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*!
 * \file imageD3Raw.h
 * \brief A class to use as a raw (type-less) container for images of 3 doubles.
 *
 * In particular, this class is useful to make computations involving pixel
 * accumulations of some sort.
 *
 * \author Fred Labrosse <ffl@aber.ac.uk>
 * \copyright 2017 Aberystwyth University
 */

#ifndef IMALIB_IMAGED3RAW_H
#define IMALIB_IMAGED3RAW_H

#include <imalib/imageD3.h>
#include <imalib/imageRGB.h>

namespace imalib
{
/*!
 *\class imageD3Raw imageD3Raw.h imalib/imageD3Raw.h
 * \brief A class to use as a raw (type-less) container for images of 3 doubles.
 *
 * In particular, this class is useful to make computations involving pixel
 * accumulations of some sort.
 *
 * \author Fred Labrosse <ffl@aber.ac.uk>
 */
class imageD3Raw : public imageD3
{
public:

   /*!
    * \brief Default constructor.
    */
   imageD3Raw()
      : imageD3()
   {
   }

   /*!
    * \brief Constructor from an imageD3.
    *
    * \warning No pixel conversion is performed.
    * \param[in] image The image to copy.
    */
   imageD3Raw(const imageD3& image)
   {
      copy(image);
   }

   /*!
    * \brief Constructor from an imageRGB.
    */
   imageD3Raw(const imageRGB& image)
   {
      copyFromRGB(image);
   }

   /*!
    * \brief Copy function
    *
    * The size of the image is changed to fit the source image.
    *
    * \warning: this function does not do any conversion.  This means the values
    * will be between 0 and 255.  If you want a conversion to take place, then
    * use one of the typed image classes.
    *
    * \param[in] image The imageRGB to copy.
    */
   void copyFromRGB(const imageRGB& image)
   {
      setSize(image.nbColumns(), image.nbRows());

      const unsigned char* rgb = image.getActualPixel(0, 0);
      double* pixels = getActualPixel(0, 0);

      for (unsigned long ind = 0;
           ind < nbColumns() * nbRows() * 3U;
           ind++)
      {
         pixels[ind] = static_cast<double>(rgb[ind]);
      }
   }

   /*!
    * \brief Copy the imageD3Raw into an imageRGB.
    *
    * The size of the target image is set here in accordance to the source image.
    *
    * \warning This function does not do any conversion.  This means the values
    * need to be between 0 and 255.  If you want a conversion to take place,
    * then use one of the typed image classes, or normalise it first.
    *
    * \param[out] image The imageRGB.
    * \sa normalise().
    */
   virtual void toRGB(imageRGB* image) const
   {
      image->setSize(nbColumns(), nbRows());

      const double* pixels = getActualPixel(0, 0);
      unsigned char* rgb = image->getActualPixel(0, 0);

      for (unsigned long ind = 0;
           ind < nbColumns() * nbRows() * 3U;
           ind++)
      {
         rgb[ind] = static_cast<unsigned char>(pixels[ind]);
      }
   }

   /*!
    * \brief Add an imageD3 to the current imageD3Raw.
    *
    * \warning No conversion or limit checking is done.
    *
    * Sizes must be compatible.  This is only checked if compiled without -NDEBUG.
    *
    * \param[in] image2 The imageD3 to add.
    */
   void add(const imageD3& image2)
   {
      assert((nbColumns() == image2.nbColumns())
             && (nbRows() == image2.nbRows()));

      const double* pixels2 = image2.getActualPixel(0, 0);
      double* pixels = getActualPixel(0, 0);

      for (unsigned long ind = 0;
           ind < nbColumns() * nbRows() * 3U;
           ind++)
      {
         pixels[ind] += pixels2[ind];
      }
   }

   /*!
    * \brief Devide the imageD3Raw by a double.
    *
    * The components of all the pixels are divided by the same value.
    *
    * \param[in] nb The double to divide the imageD3Raw by.
    */
   void divide(double nb)
   {
      double* pixels = getActualPixel(0, 0);

      for (unsigned long ind = 0;
           ind < nbColumns() * nbRows() * 3U;
           ind++)
      {
         pixels[ind] /= nb;
      }
   }

   /*!
    * \brief Normalise the image so that the maximum value is the given one.
    * 
    * This is useful to save the image.  The new values will be in [0;maxVal].
    *
    * \param[in] maxVal The maximum value.
    */
   void normalise(double maxVal = 255.0)
   {
      double* pixels = getActualPixel(0, 0);
      double currentMaxVal = std::numeric_limits<double>::min();
      double currentMinVal = std::numeric_limits<double>::max();

      for (unsigned long ind = 0;
           ind < nbColumns() * nbRows() * 3U;
           ind++)
      {
         if (currentMaxVal < pixels[ind])
            currentMaxVal = pixels[ind];
         if (currentMinVal > pixels[ind])
            currentMinVal = pixels[ind];
      }

      const double coeff = maxVal / (currentMaxVal - currentMinVal);
      for (unsigned long ind = 0;
           ind < nbColumns() * nbRows() * 3U;
           ind++)
      {
         pixels[ind] = (pixels[ind] - currentMinVal) * coeff;
      }
   }

protected:

private:

};
} // namespace imalib

#endif // IMALIB_IMAGED3RAW_H
