/*
 * imalib
 * 
 * Copyright 2017 Aberystwyth University
 * 
 * This library and program are free software: you can redistribute them
 * and/or modify them under the terms of the GNU General Public License as 
 * published by the Free Software Foundation, either version 3 of the License,
 * or (at your option) any later version.
 * 
 * The program and library are distributed in the hope that they will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*!
 * \file imageGrabber.h
 * \brief A class to wrap the grabbing of images.
 *
 * The image is just a colourful image generated by the class.  As such, the
 * class should be inherited by a real frame grabber class.
 *
 * \author Fred Labrosse <ffl@aber.ac.uk>
 * \copyright 2017 Aberystwyth University
 */

#ifndef IMAGEGRABBER_H
#define IMAGEGRABBER_H

#include "imalib/imageMapping.h"

#include <iostream>
#include <sstream>
#include <string>

namespace imalib
{
   const int IMAGE_DEFAULT_INPUTWIDTH = 768;   //PAL
   const int IMAGE_DEFAULT_INPUTHEIGHT = 576;   //PAL

   class imageRGB;

   /*!
    * \class imageGrabber imageGrabber.h imalib/imageGrabber.h
    * \brief A class to interface a generic image grabber.
    *
    * It does nothing else than generating colourful images.  You certainly
    * want to create your own class that should inherit imageGrabber.
    */
   class imageGrabber
   {
   public:
      /*!
       * \brief Default constructor.
       */
      imageGrabber();

      /*!
       * \brief Constructor.
       * \param[in] imgWidth
       * \param[in] imgHeight The width and height of the images to grab.
       * \param[in] upsideDown Set to true if the camera is upside-down.
       * \param[in] flip Set to true to horizontally flip the images.
       */
      imageGrabber(unsigned short imgWidth, unsigned short imgHeight,
                   bool upsideDown, bool flip);

      /*!
       * \brief destructor.
       */
      virtual ~imageGrabber();

      /*!
       * \brief Set the size of the image to be captured.
       * \param[in] imgWidth
       * \param[in] imgHeight The width and height of the images to grab.
       */
      void setImageSize(unsigned short imgWidth, unsigned short imgHeight)
      {
         imageWidth = imgWidth;
         imageHeight = imgHeight;

         updateMapping();
      }

      /*!
       * \brief Tell if we are connected to the required frame grabber.
       *
       * In the case of this fake grabber, we are always connected.
       *
       * \return true if we are connected, false otherwise.
       */
      virtual bool isConnected()
      {
         return (connected);
      }

      /*!
       * \brief Display on the standard output some information about the
       * source of the images.
       *\sa printInfoAboutSource(std::ostringstream*).
       */
      void printInfoAboutSource()
      {
         std::ostringstream aStream;
         printInfoAboutSource(&aStream);
         std::cout << aStream.str();
      }

      /*!
       * \brief Outputs some information about the source of the images in the
       * given stream.
       *
       * You should really override this!
       *
       * \param[in,out] aStream The stream to which the info is added.
       */
      virtual void printInfoAboutSource(std::ostringstream* aStream);

      /*!
       * \brief Output the basic info about the grabber.
       *
       * This should be called by all sub-classes in their
       * printInfoAboutSource() function.
       *
       * \param[in,out] aStream The stream to which the info is added.
       */
      void printBasicInfoAboutSource(std::ostringstream* aStream);

      /*!
       * \brief Capture the image from a grabber.
       *
       * This is there because the capture can take a while.  We thus first
       * capture and then use (with getImage()).  This does not change the
       * image data.
       *
       * You will have to implement that for your specific frame grabber.
       */
      virtual void captureImage();

      /*!
       * \brief Get an image from whatever image source.
       *
       * The image might have been pre-captured (with captureImage()), but if
       * not wil be captured here.
       *
       * The default implementation generates a moving image.  You may want
       * to override this.
       *
       * The images are resized here.
       *
       * \param[out] theImage The image in which to save the grabbed image, if
       * not NULL.
       * \param[out] rawImage The image in which to save the raw image (before
       * any transformation is applied), if not NULL.
       */
      virtual void getImage(imageRGB* theImage, imageRGB* rawImage = NULL);

      /*!
       * \brief Unwrap or not.
       * \param[in] yes Whether to unwrap or not.
       * \param[in] outerRadius The outer radius of the doughnut.
       * \param[in] innerRadius The inner radius of the doughnut.
       * \param[in] centreX
       * \param[in] centreY The Centre of the doughnut.
       * \param[in] angularRes The angular resolution (number of pixels per
       * degree of angle).
       * \param[in] power The power of the function to convert values along a
       * radial axis in the omni-directional image to a column in the panoramic
       * image (1.0 is linear, 2.0 is squared, etc).
       * \param[in] direction 1 is turn counter-clockwise (positive
       * trigonometric), -1 is turn clockwise (negative trigonometric) from
       * left to right in the unwrapped image.
       * \sa createUnwrappingMapping(),
       * imageMapping::createUnwrappingMapping().
       */
      void doUnwrap(bool yes,
                    unsigned short outerRadius,
                    unsigned short innerRadius,
                    unsigned short centreX,
                    unsigned short centreY,
                    unsigned short angularRes = 1,
                    double power = 1.0,
                    short direction = 1);

      /*!
       * \brief Unwrap or not.
       *
       * This function actually reads the configuration file, despite the fact
       * that the imageMapping class can do it as well.  This is because we
       * need the info here.
       *
       * \param[in] yes Whether to unwrap or not.
       * \param[in] fileName Name of the configuration file of the unwrapping.
       */
      void doUnwrap(bool yes,
                    const std::string& fileName);

      /*!
       * \brief Blur or not.
       * \param[in] yes Whether to blur or not.
       * \param[in] blurWidth
       * \param[in] blurHeight The size of the window used to blur.
       * \sa scaleBlur().
       */
      void doBlur(bool yes,
                  unsigned short blurWidth, unsigned short blurHeight);

      /*!
       * \brief Is the camera upside-down or not.
       * \param[in] yes Set to true if the camera is upside-down.
       */
      void setUpsideDown(bool yes)
      {
         _upsideDown = yes;
         updateMapping();
      }

      /*!
       * \brief Do or not flip the images horizontally.
       * \param[in] yes Set to true to flip, false otherwise.
       */
      void setFlip(bool yes)
      {
         _flip = yes;
         updateMapping();
      }

      /*!
       * \brief Set some parameters for the grabber.
       * \param[in] imgWidth
       * \param[in] imgHeight The width and height of the images to grab.
       * \param[in] upsideDown Set to true if the camera is upside-down.
       * \param[in] flip Set to true to flip.
       */
      void setParameters(unsigned short imgWidth, unsigned short imgHeight,
                         bool upsideDown, bool flip)
      {
         imageWidth = imgWidth;
         imageHeight = imgHeight;
         _upsideDown = upsideDown;
         _flip = flip;
         updateMapping();
      }

      /*!
       * \brief Return the current imageMapping used to transform the grabbed
       * images.
       * \return A pointer to a constant imageMapping.
       */
      const imageMapping* getImageMapping()
      {
         return(imgMap);
      }

   protected:

      imageGrabber(const imageGrabber* imgTrans);
                                // We don't provide an implementation of
                                // that on purpose!
      imageGrabber& operator=(const imageGrabber& imgTrans);
                                // Same as above!

      /*!
       * \brief Update the mapping.
       */
      void updateMapping();

      /*!
       * \brief True if we are connected to a frame grabber, false otherwise.
       */
      bool connected;

      /*!
       * \brief Width of the image to grab.
       */
      unsigned short imageWidth;

      /*!
       * \brief Height of the image to grab.
       */
      unsigned short imageHeight;

      /*!
       * \brief Whether to unwrap.
       */
      bool _unwrap;
      
      /*!
       * \brief Outer radius of the doughnut.
       */
      unsigned short _outerRadius;

      /*!
       * \brief The inner radius of the doughnut.
       */
      unsigned short _innerRadius;

      /*!
       * \brief X position of the centre of the doughnut.
       */
      unsigned short _centreX;
      
      /*!
       * \brief Y position of the centre of the doughnut.
       */
      unsigned short _centreY;

      /*!
       * \brief Number of pixels per degree of angle for the unwrapping.
       */
      unsigned short _angularRes;
      
      /*!
       * \brief Power of the unwrapping.
       */
      double _unwrappingPower;

      bool _blur;               // Whether to blur or not.
      unsigned short _blurWidth, _blurHeight;
                                // Size of the rectangle used to blur.

      bool _upsideDown;         // If true, then the images must be rotated by
                                // 180 degrees (camera upside-down)

      bool _flip;               // To flip horizontally the images.

      short _direction;         // Direction of the unwrapping.

      imageMapping* imgMap;     // The current mapping.

      bool capture;             // If true, then a capture has been
                                // initiated.
   };
} // namespace imalib

#endif
