/*
 * imalib
 * 
 * Copyright 2017 Aberystwyth University
 * 
 * This library and program are free software: you can redistribute them
 * and/or modify them under the terms of the GNU General Public License as 
 * published by the Free Software Foundation, either version 3 of the License,
 * or (at your option) any later version.
 * 
 * The program and library are distributed in the hope that they will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*!
 * \file imageHSV.h
 * \brief Defines classes to play with HSV images.
 * \author Fred Labrosse <ffl@aber.ac.uk>
 * \copyright 2017 Aberystwyth University
 */

#ifndef IMALIB_IMAGEHSV_H
#define IMALIB_IMAGEHSV_H

#include <cassert>
#include <cstdlib>
#include <exception>
#include <fstream>
#include <iostream>
#include <map>
#include <sstream>
#include <string>

#ifdef WITH_PNG
#  include <png.h>
#endif // WITH_PNG

#include <imalib/imageD2.h>
#include <imalib/imageD3.h>
#include <imalib/imageRGB.h>

namespace imalib
{
   /*!
    * \class imageHSV imageHSV.h imalib/imageHSV.h
    * \brief A class to represent images in the HSV colour space.
    *
    * The conversion is as follows (see for example
    * http://en.wikipedia.org/wiki/HSL_and_HSV).
    *
    * Conversion from RGB is provided. The transformation from RGB to HSV is as
    * follows.  For \f$ 0 \leq R,G,B \leq 1 \f$:
    * \f[
    *    \begin{array}{l}
    *    \displaystyle
    *    M = \max(R,G,B),\\
    *    m = \min(R,G,B),\\
    *    C = M - m.
    *    \end{array}
    * \f]
    * With this, HSV is defined as
    * \f[
    * \begin{array}{l}
    *    \displaystyle
    *    V = M, \\
    *    H = 60^\circ \times \left\{
    *       \begin{array}{ll}
    *         \displaystyle
    *         \mathrm{undefined}        & \mathrm{if} \; C = 0, \\
    *         \frac{G - B}{C} \;\bmod 6 & \mathrm{if} \; M = R, \\
    *         \frac{B - R}{C} + 2       & \mathrm{if} \; M = G, \\
    *         \frac{R - G}{C} + 4       & \mathrm{if} \; M = B,
    *       \end{array}\right. \\
    *    S = \left\{
    *       \begin{array}{ll}
    *         \displaystyle
    *         0           &\mathrm{if} \; C = 0, \\
    *         \frac{C}{V} &\mathrm{otherwise}.
    *       \end{array}
    *       \right.
    *  \end{array}
    * \f]
    */
   class imageHSV : public imageD3
   {
   public:
      /*!
       * \brief Constructor from an imageRGB.
       * \param image (IN) The imageRGB to copy.
       */
      imageHSV(const imageRGB& image) 
      {
         copyFromRGB(image);
      }
      
      /*!
       * \brief Copy function
       *
       * The size of the image is changed to fit the source image.
       *
       * \param image (IN) The imageRGB to copy.
       */      
      void copyFromRGB(const imageRGB& image)
      {
         setSize(image.nbColumns(), image.nbRows());

         const unsigned char* rgb = image.getActualPixel(0, 0);
         double* hsv = getActualPixel(0, 0);
         for (unsigned long ind = 0;
              ind < nbColumns() * nbRows() * 3U;
              ind += 3)
            RGBToHSV(&rgb[ind], &hsv[ind]);        
      }

      /*!
       * \brief Copy the image into an imageRGB.
       *
       * The size of the target image is set here in accordance to the
       * source image.
       *
       * \param image (OUT) The imageRGB.
       * \todo This needs to be implemented.
       */
      virtual void toRGB(imageRGB* image) const
      {
         std::cerr << "FIXME: imageHSV::toRGB(imageRGB* image)\n"
                   << "FIXME: This function needs to be implemented!.\n";
      }

   private:
      /*!
        \brief Convert a pixel in the RGB colour space to a pixel in the
        HSV colour space.
        \param rgb (OUT) the pixel in RGB colour space.
        \param hsv (IN) The pixel in HSV colour space (must be
        allocated beforehand).
      */
      void RGBToHSV(const unsigned char* rgb, double* hsv);

      /*!
         \brief The value H should be set to if it is undefined for
         a particular colour. 
      */
      static const double HSV_H_UNDEFINED;

   };
   
   
   /*!
    * \class imageHS_ imageHSV.h imalib/imageHSV.h
    * \brief A class to represent images in the
    * HSV colour space where only the
    * H ans S components are kept.
    *
    * The class is there to allow the efficient computation of distances in image
    * space in the HSV colour space
    * while not considering the darkness information. This can be done by
    * instantiating a distance_base with the number of colour components set to 2.
    *
    *
    * <small>The '_' in the name is there to remind of the missing V, not as a
    * link between two words ;-).</small>
    *
    * \sa imageHSV.
    */
   class imageHS_ : public imageD2
   {
   public:

      /*!
        \brief Constructor from an imageRGB.
        \param image (IN) The imageRGB to copy.
      */
      imageHS_(const imageRGB& image) 
      {
         copyFromRGB(image);
      }
      
      /*!
        \brief Copy function

        The size of the image is changed to fit the source image.

        \param image (IN) The imageRGB to copy.
      */      
      void copyFromRGB(const imageRGB& image)
      {
         setSize(image.nbColumns(), image.nbRows());

         for (unsigned short col = 0; col < image.nbColumns(); ++col)
         {
            for (unsigned row = 0; row < image.nbRows(); ++row)
            {
               RGBToHS_(image.getActualPixel(col, row),getActualPixel(col, row));
            }
         }
      }

      /*!
        \brief Copy the image into an imageRGB.

        The size of the target image is set here in accordance to the
        source image.

        \param image (OUT) The imageRGB.
        \todo This needs to be implemented.
      */
      virtual void toRGB(imageRGB* image) const
      {
         std::cerr << "FIXME: imageHSV::toRGB(imageRGB* image)\n"
                   << "FIXME: This function needs to be implemented!.\n";
      }

   protected:
   private:
      /*!
        \brief Convert a pixel in the RGB colour space to a pixel in the
        HSV colour space.
        \param rgb (OUT) the pixel in RGB colour space.
        \param hsv (IN) The pixel in HSV colour space (must be
        allocated beforehand).
      */
      void RGBToHS_(const unsigned char* rgb, double* hsv);

      /*!
         \brief The value H should be set to if it is undefined for
         a particular colour. 
      */
      static const double HSV_H_UNDEFINED;

   };
   
   

} // namespace imalib

#endif // IMALIB_IMAGEHSV_H

//  LocalWords:  param cstdlib NDEBUG nbColComp imageRGB imalib imageMapping
//  LocalWords:  createUnwrappingMapping bestEuclideanDistance translationOne
//  LocalWords:  createVerticalShiftMapping createHorizontalShiftMapping
//  LocalWords:  translationTwo gradOne gradTwo euclideanDistance imageLab
//  LocalWords:  typeCol LocalWords imageHSV
