/*
 * imalib
 * 
 * Copyright 2017 Aberystwyth University
 * 
 * This library and program are free software: you can redistribute them
 * and/or modify them under the terms of the GNU General Public License as 
 * published by the Free Software Foundation, either version 3 of the License,
 * or (at your option) any later version.
 * 
 * The program and library are distributed in the hope that they will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*!
 * \file imageLCS.h
 * \brief Defines classes to play with log-chromaticity space (LCS) images.
 * \author Marek Ososinski <mro7@aber.ac.uk>
 * \copyright 2017 Aberystwyth University
 */

#ifndef IMALIB_IMAGELCS_H
#define IMALIB_IMAGELCS_H

#include <cassert>
#include <cstdlib>
#include <exception>
#include <fstream>
#include <iostream>
#include <map>
#include <sstream>
#include <string>

#ifdef WITH_PNG
#  include <png.h>
#endif // WITH_PNG

#include <imalib/imageD2.h>
#include <imalib/imageD3.h>
#include <imalib/imageRGB.h>

namespace imalib
{
   /*!
    * \class imageLCS imageLCS.h imalib/imageLCS.h
    * \brief A class to represent images in the log-chromaticity space.
    *
    * This is a two dimensional colour space, defined as:
    *\f[
    * lcs_1 = \left\{
    *   \begin{array}{ll}
    *      \displaystyle
    *      log(r/g) & \mathrm{if} \; g>0,\\
    *      log(r)   & \mathrm{otherwise},
    *   \end{array}
    *   \right.
    * \f]
    * \f[
    * lcs_2 = \left\{
    *   \begin{array}{ll}
    *      \displaystyle
    *      log(b/g) & \mathrm{if} \; g>0,\\
    *      log(b)   & \mathrm{otherwise}.
    *   \end{array}
    *   \right.
    * \f]
    *
    * if either \f$ r \f$ or \f$ b \f$ are equal to zero, the value 1 is used
    * instead in the log.
    *
    * Conversion from RGB is provided.
    */
   class imageLCS : public imageD2
   {
   public:

      /*!
       * \brief Constructor from an imageRGB.
       * \param image (IN) The imageRGB to copy.
       */
      imageLCS(const imageRGB& image) 
      {
         copyFromRGB(image);
      }
      
      /*!
       * \brief Copy function
       *
       * The size of the image is changed to fit the source image.
       *
       * \param[in] image The imageRGB to copy.
       */      
      void copyFromRGB(const imageRGB& image)
      {
         setSize(image.nbColumns(), image.nbRows());

         for (unsigned short col = 0; col < image.nbColumns(); ++col)
         {
            for (unsigned row = 0; row < image.nbRows(); ++row)
            {
               RGBToLCS(image.getActualPixel(col, row),getActualPixel(col, row));
            }
         }
      }

      /*!
       * \brief Copy the image into an imageRGB.
       *
       * The size of the target image is set here in accordance to the
       * source image.
       *
       * \param[out] image The imageRGB.
       * \todo This needs to be implemented.
       */
      virtual void toRGB(imageRGB* image) const
      {
         std::cerr << "FIXME: imageLCS::toRGB(imageRGB* image)\n"
                   << "FIXME: This function is not possible to implement.\n";
      }

   protected:
   private:
      /*!
       * \brief Convert a pixel in the RGB colour space to a pixel in the
       * LCS colour space.
       * \param[in] rgb The pixel in RGB colour space.
       * \param[out] lcs The pixel in HSV colour space (must be
       * allocated beforehand).
       */
      void RGBToLCS(const unsigned char* rgb, double* lcs);

   };
   
   

} // namespace imalib

#endif // IMALIB_IMAGELCS_H
