/*
 * imalib
 * 
 * Copyright 2017 Aberystwyth University
 * 
 * This library and program are free software: you can redistribute them
 * and/or modify them under the terms of the GNU General Public License as 
 * published by the Free Software Foundation, either version 3 of the License,
 * or (at your option) any later version.
 * 
 * The program and library are distributed in the hope that they will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*!
 * \file imageLab.h
 * \brief Defines classes to play with CIEL*a*b* images.
 * \author Fred Labrosse <ffl@aber.ac.uk>
 * \copyright 2017 Aberystwyth University
 */

#ifndef IMALIB_IMAGELAB_H
#define IMALIB_IMAGELAB_H

#include <cassert>
#include <cstdlib>
#include <exception>
#include <fstream>
#include <iostream>
#include <map>
#include <sstream>
#include <string>

#ifdef WITH_PNG
#  include <png.h>
#endif // WITH_PNG

#include <imalib/imageD2.h>
#include <imalib/imageD3.h>
#include <imalib/imageRGB.h>

namespace imalib
{
/*!
 * \class imageLab imageLab.h imalib/imageLab.h
 * \brief A class to represent images in the
 * CIE&nbsp;L<sup>*</sup>a<sup>*</sup>b<sup>*</sup> colour space.
 *
 * Conversion to and from RGB is provided.
 *
 * The RGB to CIE&nbsp;L<sup>*</sup>a<sup>*</sup>b<sup>*</sup> conversion
 * is as follows:
 * \f[
 * \left(\begin{array}{c}X\\Y\\Z\end{array}\right)
 * =
 * \left(\begin{array}{ccc}
 * 2.7690 & 1.7518 & 1.1300\\
 * 1.0000 & 4.5907 & 0.0601\\
 * 0.0000 & 0.0565 & 5.5943
 * \end{array}\right)
 * \left(\begin{array}{c}R\\G\\B\end{array}\right)
 * \f]
 * \f[
 * \begin{array}{l}
 * \displaystyle
 * L^* = 116 \left(\frac{Y}{Y_n}\right)^{1/3} - 16 \\ \\
 * \displaystyle
 * a^* = 500 \left[\left(\frac{X}{X_n}\right)^{1/3}
 * - \left(\frac{Y}{Y_n}\right)^{1/3}\right] \\ \\
 * \displaystyle
 * b^* = 200 \left[\left(\frac{Y}{Y_n}\right)^{1/3}
 * - \left(\frac{Z}{Z_n}\right)^{1/3}\right],
 * \end{array}
 * \f]
 * where \f$X_n\f$, \f$Y_n\f$, and \f$Z_n\f$ are the \f$X\f$, \f$Y\f$,
 * and \f$Z\f$ values of the reference white, \f$R = G = B = 255\f$.
 *
 * The CIE&nbsp;L<sup>*</sup>a<sup>*</sup>b<sup>*</sup> to RGB conversion
 * is given by:
 * \f[
 * \begin{array}{l}
 * \displaystyle
 * Y = \frac{1}{1560896} Y_n L^{*3} + \frac{3}{97556} Y_n L^{*2}
 *     + \frac{12}{24389} Y_n L^* + \frac{64}{24389} Yn \\ \\
 * \displaystyle
 * C = \left(\frac{Y}{Yn}\right)^{1/3} \\ \\
 * \displaystyle
 * X = \frac{1}{125} X_n a^{*3} + \frac{3}{25} X_n a^{*2} C
 * + \frac{3}{5} X_n a C^2 + X_n C^3 \\ \\
 * \displaystyle
 * Z = -\frac{1}{8} Z_n b^{*3} + \frac{3}{4} Z_n b^{*2} C
 * - \frac{3}{2} Z_n b C^2 + Z_n C^3
 * \end{array}
 * \f]
 * \f[
 * \left(\begin{array}{c}R\\G\\B\end{array}\right)
 * =
 * \left(\begin{array}{ccc}
 * 0.4184383528 & -0.1586557990 & -0.08281646054 \\
 * -0.09116119246 & 0.2524253419 & 0.01570194384 \\
 * 0.0009206884461 & -0.002549386307 & 0.1785947912
 * \end{array}\right)
 * \left(\begin{array}{c}Z\\Y\\Z\end{array}\right).
 * \f].
 *
 * Black is (-16, 0, 0).  \f$L^*\f$ is in \f$[-16; 100]\f$.
 * \f$a^*\f$ and \f$b^*\f$ have no limit.
 *
 * \sa image_ab.
 *
 * \todo The special case for low values
 * of \f$X/X_n\f$ and \f$Y/Y_n\f$ should be dealt with [\ref Fai98].
 *
 * \todo Need to verify the information about
 * CIE&nbsp;L<sup>*</sup>a<sup>*</sup>b<sup>*</sup>.
 */
class imageLab : public imageD3
{
public:

   /*!
    * \brief Constructor.
    */
   imageLab()
      : imageD3()
   {
   }

   /*!
    * \brief Constructor.
    */
   imageLab(short cols, short rows, short tCols = 0, short tRows = 0)
      : imageD3(cols, rows, tCols, tRows)
   {
   }

   /*!
    * \brief Constructor from an imageRGB.
    * \param[in] image The imageRGB to copy.
    */
   imageLab(const imageRGB& image)
   {
      copyFromRGB(image);
   }

   /*!
    * \brief Copy function
    *
    * The size of the image is changed to fit the source image.
    *
    * \param[in] image The imageRGB to copy.
    */
   virtual void copyFromRGB(const imageRGB& image)
   {
      setSize(image.nbColumns(), image.nbRows());

      const unsigned char* rgb = image.getActualPixel(0, 0);
      double* lab = getActualPixel(0, 0);
      for (unsigned long ind = 0;
           ind < nbColumns() * nbRows() * 3U;
           ind += 3)
         RGBToLab(&rgb[ind], &lab[ind]);
   }

   /*!
    * \brief Copy the image into an imageRGB.
    *
    * The size of the target image is set here in accordance to the source
    * image.
    *
    * \param[out] image The imageRGB.
    */
   virtual void toRGB(imageRGB* image) const
   {
      image->setSize(nbColumns(), nbRows());

      unsigned char* rgb = image->getActualPixel(0, 0);
      const double* lab = getActualPixel(0, 0);
      for (unsigned long ind = 0;
           ind < nbColumns() * nbRows() * 3U;
           ind += 3)
         LabToRGB(&lab[ind], &rgb[ind]);
   }

   /*!
    * \brief Set a pixel from a RGB pixel.
    *
    * Checks on the coordinates are only performed if compiled without NDEBUG.
    *
    * \param[in] col The column of the pixel.
    * \param[in] row The row of the pixel.
    * \param[in] RGB The RGB pixel.
    */
   void setPixelRGB(unsigned short col, unsigned short row,
                    const unsigned char* RGB)
   {
      assert((col <= nbColumns())
             || (row <= nbRows()));
      RGBToLab(RGB, getActualPixel(col, row));
   }

   /*!
    * \brief Get a pixel and convert it to an RGB pixel.
    *
    * Checks on the coordinates are only performed if compiled without NDEBUG.
    *
    * \param[in] col The column of the pixel.
    * \param[in] row The row of the pixel.
    * \param[out] RGB Where the RGB pixel will be stored (must be
    * allocated beforehand).
    */
   void getPixelRGB(unsigned short col, unsigned short row,
                    unsigned char* RGB)
   {
      assert((col <= nbColumns())
             || (row <= nbRows()));
      LabToRGB(getActualPixel(col, row), RGB);
   }

   /*!
    * \brief Change the L<sup>*</sup> component (luminance) of all pixels to
    * the given value.
    *
    * Is that useful?
    *
    * \param[in] lumin The new luminance.
    */
   void changeLuminance(double lumin)
   {
      double* pixel = getActualPixel(0, 0);
      for (unsigned long ind = 0;
           ind < nbColumns() * nbRows() * 3U;
           ind += 3)
         pixel[ind] = lumin;
   }

   /*!
    * \brief Scale the L<sup>*</sup> component (luminance) of all pixels.
    * \param[in] factor The scaling factor.
    */
   void scaleLuminance(double factor)
   {
      double* pixel = getActualPixel(0, 0);
      for (unsigned long ind = 0;
           ind < nbColumns() * nbRows() * 3U;
           ind += 3)
         pixel[ind] = pixel[ind] * factor;
   }

   /*!
    * \brief Copy the a<sup>*</sup> and b<sup>*</sup> components in an
    * RGB image using the R and G components only (B is set to 124).
    *
    * The a<sup>*</sup> and b<sup>*</sup> components are normalised to the
    * interval [0; 255] using the minimum and maximum values obtainable for
    * them by converting the RGB values (255, 0, 0) (red), (0, 255, 0) (green),
    * (255, 255, 0) (yellow) and (0, 0, 255) (blue) which roughly correspond
    * to the axes defined by the components a<sup>*</sup> and b<sup>*</sup>.
    * This range can be altered.
    *
    * \param[out] img The image where to save the a<sup>*</sup> and
    * b<sup>*</sup> components (its size is set here).
    * \param[in] coeff Divide the "theoretical" range of a<sup>*</sup> and
    * b<sup>*</sup> by this coefficient (to perform some kind of contrast
    * modification --- enhancement for coeff > 1.0, reduction for coeff < 1.0).
    */
   void copyAB(imageRGB* img, double coeff = 1.0) const;

   /*!
    * \brief Convert a pixel in the
    * CIE&nbsp;L<sup>*</sup>a<sup>*</sup>b<sup>*</sup> colour space to a
    * pixel in the RGB colour space.
    *
    * \param[in] lab The pixel in
    * CIE&nbsp;L<sup>*</sup>a<sup>*</sup>b<sup>*</sup> colour space.
    * \param[out] rgb The pixel in RGB colour space (must be allocated
    * beforehand).
    */
   static void LabToRGB(const double* lab, unsigned char* rgb);

   /*!
    * \brief Convert a pixel in the RGB colour space to a pixel in the
    * CIE&nbsp;L<sup>*</sup>a<sup>*</sup>b<sup>*</sup> colour space.
    *
    * \param[in] rgb The pixel in RGB colour space.
    * \param[out] lab The pixel in
    * CIE&nbsp;L<sup>*</sup>a<sup>*</sup>b<sup>*</sup> colour space (must
    * be allocated beforehand).
    */
   static void RGBToLab(const unsigned char* rgb, double* lab);

   /*!
    * \brief Convert a pixel in the RGB colour space to a pixel in the
    * CIE&nbsp;L<sup>*</sup>a<sup>*</sup>b<sup>*</sup> colour space.
    *
    * Note that the RGB values are in [0, 255].
    *
    * \param[in] rgb The pixel in RGB colour space.
    * \param[out] lab The pixel in
    * CIE&nbsp;L<sup>*</sup>a<sup>*</sup>b<sup>*</sup> colour space (must
    * be allocated beforehand).
    */
   static void RGBToLab(const double* rgb, double* lab);

protected:
private:
};



/*!
 * \class image_ab imageLab.h imalib/imageLab.h
 * \brief A class to represent images in the
 * CIE&nbsp;L<sup>*</sup>a<sup>*</sup>b<sup>*</sup> colour space where only the
 * a<sup>*</sup> and b<sup>*</sup> components are kept.
 *
 * The class is there to allow the efficient computation of distances in image
 * space in the CIE&nbsp;L<sup>*</sup>a<sup>*</sup>b<sup>*</sup> colour space
 * while not considering the luminance information.  This can be done by
 * instantiating a distance_base with the number of colour components set to 2.
 *
 * Conversions to and from RGB are provided, although the conversion to RGB
 * is not too meaningful since the luminance information is not present.  The
 * conversions are the same as for the imageLab class.
 *
 * <small>The '_' in the name is there to remind of the missing L, not as a
 * link between two words ;-).</small>
 *
 * \sa imageLab.
 */
class image_ab : public imageD2
{
public:

   /*!
    * \brief Constructor.
    */
   image_ab()
      : imageD2(), defaultL(60)
   {
   }

   /*!
    * \brief Constructor.
    */
   image_ab(short cols, short rows, short tCols = 0, short tRows = 0)
      : imageD2(cols, rows, tCols, tRows)
   {
   }

   /*!
    * \brief Constructor from an imageRGB.
    * \param[in] image The imageRGB to copy.
    */
   image_ab(const imageRGB& image)
   {
      copyFromRGB(image);
   }

   /*!
    * \brief Copy function
    *
    * The size of the image is changed to fit the source image.
    *
    * \param[in] image The imageRGB to copy.
    */
   virtual void copyFromRGB(const imageRGB& image)
   {
      setSize(image.nbColumns(), image.nbRows());

      for (unsigned short col = 0; col < image.nbColumns(); ++col)
      {
         for (unsigned row = 0; row < image.nbRows(); ++row)
         {
            RGBTo_ab(image.getActualPixel(col, row),
                     getActualPixel(col, row));
         }
      }
   }

   /*!
    * \brief Copy the image into an imageRGB.
    *
    * The size of the target image is set here in accordance to the source
    * image.
    *
    * \param[out] image The imageRGB.
    */
   virtual void toRGB(imageRGB* image) const
   {
      image->setSize(nbColumns(), nbRows());

      for (unsigned short col = 0; col < nbColumns(); ++col)
      {
         for (unsigned row = 0; row < nbRows(); ++row)
         {
            _abToRGB(getActualPixel(col, row),
                     image->getActualPixel(col, row));
         }
      }
   }

   /*!
    * \brief Convert a pixel in the
    * CIE&nbsp;L<sup>*</sup>a<sup>*</sup>b<sup>*</sup> colour space with only
    * the a<sup>*</sup> and b<sup>*</sup> components to a pixel in the RGB
    * colour space.
    *
    * The value for L<sup>*</sup> is taken to be the default L;
    *
    * \param[in] ab The pixel in
    * CIE&nbsp;_a<sup>*</sup>b<sup>*</sup> colour space.
    * \param[out] rgb The pixel in RGB colour space (must be allocated
    * beforehand).
    */
   void _abToRGB(const double* ab, unsigned char* rgb) const;

   /*!
    * \brief Convert a pixel in the RGB colour space to a pixel in the
    * CIE&nbsp;L<sup>*</sup>a<sup>*</sup>b<sup>*</sup> colour space with only
    * the a<sup>*</sup> and b<sup>*</sup> components.
    *
    * \param[out] rgb The pixel in RGB colour space.
    * \param[in] ab The pixel in
    * CIE&nbsp;_a<sup>*</sup>b<sup>*</sup> colour space (must
    * be allocated beforehand).
    */
   void RGBTo_ab(const unsigned char* rgb, double* ab) const;

   /*!
    * \brief Set the default value for L<sup>*</sup> when converting to other
    * colour spaces.
    * \param[in] L The new value.
    */
   void setDefaultL(double L)
   {
      defaultL = L;
   }

   /*!
    * \brief Get the default value for L<sup>*</sup> when converting to other
    * colour spaces.
    * \return The value.
    */
   double getDefaultL() const
   {
      return(defaultL);
   }

protected:
private:

   /*!
    * \brief value for L<sup>*</sup> when converting to other colour spaces.
    */
   double defaultL;
};

} // namespace imalib

#endif // IMALIB_IMAGELAB_H
