/*
 * imalib
 * 
 * Copyright 2017 Aberystwyth University
 * 
 * This library and program are free software: you can redistribute them
 * and/or modify them under the terms of the GNU General Public License as 
 * published by the Free Software Foundation, either version 3 of the License,
 * or (at your option) any later version.
 * 
 * The program and library are distributed in the hope that they will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*!
 * \file imageMCh.h
 * \brief Defines classes to play with mean chroma images.
 * \author Marek Ososinski <mro7@aber.ac.uk>
 * \copyright 2017 Aberystwyth University
 */

#ifndef IMALIB_IMAGEMCH_H
#define IMALIB_IMAGEMCH_H

#include <cassert>
#include <cstdlib>
#include <exception>
#include <fstream>
#include <iostream>
#include <map>
#include <sstream>
#include <string>

#ifdef WITH_PNG
#  include <png.h>
#endif // WITH_PNG

#include <imalib/imageRGB.h>
#include <imalib/imageD1.h>
#include <imalib/imageD3.h>

namespace imalib
{
   /*!
    * \class imageMCh imageMCh.h imalib/imageMCh.h
    * \brief A class to represent mean chroma.
    *
    * This is as defined in [\ref KCB09].  This is a one-dimentional colour
    * space:
    * \f[
    * MCh = \frac{Cb + Cr + 2a^*}{4},
    * \f]
    * where \f$ Cb \f$ and \f$ Cr \f$  are two of the YCbCr (\ref imageYCbCr)
    * components and\f$ a^* \f$ comes from the
    * CIE&nbsp;L<sup>*</sup>a<sup>*</sup>b<sup>*</sup> colour space
    * (\ref imageLab).
    *
    * The RGB conversion will only copy the value to all channels.
    */
   class imageMCh : public imageD1
   {
   public:
      /*!
       * \brief Constructor.
       */
      imageMCh()
         : imageD1()
      {
      }

      /*!
       * \brief Constructor from an imageRGB.
       * \param[in] image The imageRGB to copy.
       */
      imageMCh(const imageRGB& image)
      {
         copyFromRGB(image);
      }

      /*!
       * \brief Copy function
       *
       * The size of the image is changed to fit the source image.
       *
       * \param[in] image The imageRGB to copy.
       */
      void copyFromRGB(const imageRGB& image)
      {
         setSize(image.nbColumns(), image.nbRows());

         const unsigned char* rgb = image.getActualPixel(0, 0);
         double* mch = getActualPixel(0, 0);
         for (unsigned long ind = 0;
              ind < (static_cast<unsigned long>(nbColumns())
                     * static_cast<unsigned long>(nbRows()));
              ind += 1)
            RGBToMCh(&rgb[ind*3], &mch[ind]);
      }

      /*!
       * \brief Copy the image into an imageRGB.
       *
       * The size of the target image is set here in accordance to the
       * source image.
       *
       * \param[out] image The imageRGB.
       */
      virtual void toRGB(imageRGB* image) const
      {
         image->setSize(nbColumns(), nbRows());

         unsigned char* rgb = image->getActualPixel(0, 0);
         const double* mch = getActualPixel(0, 0);
         for (unsigned long ind = 0;
              ind < (static_cast<unsigned long>(nbColumns())
                     * static_cast<unsigned long>(nbRows()));
              ind += 1)
            MChToRGB(&mch[ind], &rgb[ind*3]);
      }

   private:

      /*!
       * \brief Convert a pixel in the RGB colour space to a mean chroma of the pixel.
       * \param[in] rgb The pixel in RGB colour space.
       * \param[out] mch The pixel in mch colour space (must be
       * allocated beforehand).
       */
      void RGBToMCh(const unsigned char* rgb, double* mch)
      {
         double R = static_cast<double>(rgb[0]) / 255.0;
         double G = static_cast<double>(rgb[1]) / 255.0;
         double B = static_cast<double>(rgb[2]) / 255.0;
         
	 
         static const double Xn = 2.7690 * 255.0 + 1.7518 * 255.0 + 1.1300 * 255.0;
         static const double Yn = 1.0000 * 255.0 + 4.5907 * 255.0 + 0.0601 * 255.0;
         double X;
         double Y;

         X = 2.7690 * rgb[0] + 1.7518 * rgb[1] + 1.1300 * rgb[2];
         Y = 1.0000 * rgb[0] + 4.5907 * rgb[1] + 0.0601 * rgb[2];

         X = cbrt(X / Xn);
         Y = cbrt(Y / Yn);
         
	 double a = ((500.0 * (X - Y)) + 128.0)/255.0;
         

         double Cb = 0.5 - 0.169 * R - 0.331 * G + 0.500 * B;
         double Cr = 0.5 + 0.500 * R - 0.419 * G - 0.081 * B;
         
         mch[0] = (Cb+Cr+2*a)/4;
	 /*mch[0] = Cb;
	 mch[1] = Cr;
	 mch[2] = a;*/
         
      }

      /*!
       * \brief Copy a mean chroma of a pixel to a pixel in the RGB
       * colour space.
       * \param[in] mch The mean chroma pixel.
       * \param[out] rgb The pixel in RGB colour space (must be allocated
       * beforehand).
       */
      void MChToRGB(const double* mch, unsigned char* rgb) const
      {
         double tmp = mch[0]*255;
         if (tmp > 255)
            tmp = 255;
         else
            tmp = tmp + 0.5;
         rgb[0] = static_cast<unsigned char>(tmp);
         rgb[1] = static_cast<unsigned char>(tmp);
         rgb[2] = static_cast<unsigned char>(tmp);
      }
   };




   class imageMC3 : public imageD3
   {
   public:
      /*!
       * \brief Constructor.
       */
      imageMC3()
         : imageD3()
      {
      }

      /*!
       * \brief Constructor from an imageRGB.
       * \param[in] image The imageRGB to copy.
       */
      imageMC3(const imageRGB& image)
      {
         copyFromRGB(image);
      }

      /*!
       * \brief Copy function
       *
       * The size of the image is changed to fit the source image.
       *
       * \param[in] image The imageRGB to copy.
       */
      void copyFromRGB(const imageRGB& image)
      {
         setSize(image.nbColumns(), image.nbRows());

         const unsigned char* rgb = image.getActualPixel(0, 0);
         double* mch = getActualPixel(0, 0);
         for (unsigned long ind = 0;
              ind < (static_cast<unsigned long>(nbColumns())
                     * static_cast<unsigned long>(nbRows()));
              ind += 1)
            RGBToMC3(&rgb[ind*3], &mch[ind*3]);
      }

      /*!
       * \brief Copy the image into an imageRGB.
       *
       * The size of the target image is set here in accordance to the
       * source image.
       *
       * \param[out] image The imageRGB.
       */
      virtual void toRGB(imageRGB* image) const
      {
         image->setSize(nbColumns(), nbRows());

         unsigned char* rgb = image->getActualPixel(0, 0);
         const double* mch = getActualPixel(0, 0);
         for (unsigned long ind = 0;
              ind < (static_cast<unsigned long>(nbColumns())
                     * static_cast<unsigned long>(nbRows()));
              ind += 1)
            MC3ToRGB(&mch[ind*3], &rgb[ind*3]);
      }

   private:

      /*!
       * \brief Convert a pixel in the RGB colour space to a mean chroma of the
       * pixel.
       * \param[in] rgb The pixel in RGB colour space.
       * \param[out] mc3 The pixel in mc3 colour space (must be
       * allocated beforehand).
       */
      void RGBToMC3(const unsigned char* rgb, double* mc3)
      {
         double R = static_cast<double>(rgb[0]) / 255.0;
         double G = static_cast<double>(rgb[1]) / 255.0;
         double B = static_cast<double>(rgb[2]) / 255.0;

         static const double Xn = 2.7690 * 255.0 + 1.7518 * 255.0 + 1.1300 * 255.0;
         static const double Yn = 1.0000 * 255.0 + 4.5907 * 255.0 + 0.0601 * 255.0;
         double X;
         double Y;

         X = 2.7690 * rgb[0] + 1.7518 * rgb[1] + 1.1300 * rgb[2];
         Y = 1.0000 * rgb[0] + 4.5907 * rgb[1] + 0.0601 * rgb[2];

         X = cbrt(X / Xn);
         Y = cbrt(Y / Yn);

         double a = ((500.0 * (X - Y)) + 128.0)/255.0;

         double Cb = 0.5 - 0.169 * R - 0.331 * G + 0.500 * B;
         double Cr = 0.5 + 0.500 * R - 0.419 * G - 0.081 * B;

         //mch[0] = (Cb+Cr+2*a)/4;
         mc3[0] = Cb;
         mc3[1] = Cr;
         mc3[2] = a;
      }

      /*!
       * \brief Copy a mean chroma of a pixel to a pixel in the RGB
       * colour space.
       * \param[in] mc3 The mean chroma pixel.
       * \param[out] rgb The pixel in RGB colour space (must be allocated
       * beforehand).
       */
      void MC3ToRGB(const double* mc3, unsigned char* rgb) const
      {
         double tmp = mc3[0]*255;
         if (tmp > 255)
            tmp = 255;
         else
            tmp = tmp + 0.5;
         rgb[0] = static_cast<unsigned char>(tmp);
         tmp = mc3[1]*255;
         if (tmp > 255)
            tmp = 255;
         else
            tmp = tmp + 0.5;
         rgb[1] = static_cast<unsigned char>(tmp);
         tmp = mc3[2]*255;
         if (tmp > 255)
            tmp = 255;
         else
            tmp = tmp + 0.5;
         rgb[2] = static_cast<unsigned char>(tmp);
      }
   };

   /*!
    * \class imageMCS imageMCh.h imalib/imageMCh.h
    * \brief A class to represent chroma information.
    *
    * This is a colour space similar to MCh but also using the b* component
    * from \em CIE \em L*a*b* and normalising all four components and the
    * result combination:
    * \f[
    * \begin{array}{l}
    *  \mathrm{Cb}_n = \mathrm{Cb} \times 255,\\                                                         *
    *  \mathrm{Cr}_n = \mathrm{Cr} \times 255,\\
    *  a_n = (a^* + 99.6749) \times 1.232539626,\\
    *  b_n = (b^* + 92.5584) \times 2.433977176,
    * \end{array}
    * \f]
    * followed by appropriate clamping.  Because of non-homogeneity of the coverage
    * of the 0 to 255 range, we define MCS as
    * \f[
    * \mathrm{MCS} = \left(\frac{\mathrm{Cb}_n + \mathrm{Cr}_n + a_n + b_n}{4} - 90\right) \times 2.65625,
    * \f]
    * again followed by appropriate clamping.
    *
    * The RGB conversion will only copy the values to all channels
    */   
   class imageMCS : public imageD1
   {
   public:
      /*!
       * \brief Constructor.
       */
      imageMCS()
         : imageD1()
      {
      }

      /*!
       * \brief Constructor from an imageRGB.
       * \param[in] image The imageRGB to copy.
       */
      imageMCS(const imageRGB& image)
      {
         copyFromRGB(image);
      }

      /*!
       * \brief Copy function
       *
       * The size of the image is changed to fit the source image.
       *
       * \param[in] image The imageRGB to copy.
       */
      void copyFromRGB(const imageRGB& image)
      {
         setSize(image.nbColumns(), image.nbRows());

         const unsigned char* rgb = image.getActualPixel(0, 0);
         double* mcs = getActualPixel(0, 0);
         for (unsigned long ind = 0;
              ind < (static_cast<unsigned long>(nbColumns())
                     * static_cast<unsigned long>(nbRows()));
              ind += 1)
            RGBToMCS(&rgb[ind*3], &mcs[ind]);
      }

      /*!
       * \brief Copy the image into an imageRGB.
       *
       * The size of the target image is set here in accordance to the
       * source image.
       *
       * \param[out] image The imageRGB.
       */
      virtual void toRGB(imageRGB* image) const
      {
         image->setSize(nbColumns(), nbRows());

         unsigned char* rgb = image->getActualPixel(0, 0);
         const double* mcs = getActualPixel(0, 0);
         for (unsigned long ind = 0;
              ind < (static_cast<unsigned long>(nbColumns())
                     * static_cast<unsigned long>(nbRows()));
              ind += 1)
            MCSToRGB(&mcs[ind], &rgb[ind*3]);
      }

   private:

      /*!
       * \brief Convert a pixel in the RGB colour space to a mean chroma of the pixel.
       * \param[in] rgb The pixel in RGB colour space.
       * \param[out] mcs The pixel in mcs colour space (must be
       * allocated beforehand).
       */
      void RGBToMCS(const unsigned char* rgb, double* mcs)
      {
         double R = static_cast<double>(rgb[0]) / 255.0;
         double G = static_cast<double>(rgb[1]) / 255.0;
         double B = static_cast<double>(rgb[2]) / 255.0;

         static const double Xn = 2.7690 * 255.0 + 1.7518 * 255.0 + 1.1300 * 255.0;
         static const double Yn = 1.0000 * 255.0 + 4.5907 * 255.0 + 0.0601 * 255.0;
         static const double Zn =                + 0.0565 * 255.0 + 5.5943 * 255.0;
         double X = 2.7690 * rgb[0] + 1.7518 * rgb[1] + 1.1300 * rgb[2];
         double Y = 1.0000 * rgb[0] + 4.5907 * rgb[1] + 0.0601 * rgb[2];
         double Z =                   0.0565 * rgb[1] + 5.5943 * rgb[2];

         X = cbrt(X / Xn);
         Y = cbrt(Y / Yn);
         Z = cbrt(Z / Zn);

         double a = 500.0 * (X - Y);
         double b = 200.0 * (Y - Z);

         double Cb = 0.5 - 0.169 * R - 0.331 * G + 0.500 * B;
         double Cr = 0.5 + 0.500 * R - 0.419 * G - 0.081 * B;

         /* Values acquired from 40k+ images of typical roads */
 
         a = (a+99.6749)*1.232539626;
         if(a<0) a=0; else if(a>255) a=255; 
         b = (b+92.5584)*2.433977176;
         if(b<0) b=0; else if(b>255) b=255;
         Cb *= 255; //(Cb-0.111079)*647.4694102;
         if(Cb<0) Cb=0; else if(Cb>255) Cb=255;
         Cr *= 255; //(Cr-0.185604)*415.2195046;
         if(Cr<0) Cr=0; else if(Cr>255) Cr=255; 

         /* All the values are now normalised to be between 0 and 255
          * The following equation will stretch the result of the equation
          * Cr+Cb+a+b to populate the 0-255 space */
 
         mcs[0] = (((Cb+Cr+a+b)/4)-90)*2.65625;
         if(mcs[0]<0) mcs[0]=0; else if(mcs[0]>255) mcs[0]=255;
      }

      /*!
       * \brief Copy a mean chroma of a pixel to a pixel in the RGB
       * colour space.
       * \param[in] mcs The mean chroma pixel.
       * \param[out] rgb The pixel in RGB colour space (must be allocated
       * beforehand).
       */
      void MCSToRGB(const double* mcs, unsigned char* rgb) const
      {
         double tmp = mcs[0]+0.5;
         rgb[0] = static_cast<unsigned char>(tmp);
         rgb[1] = static_cast<unsigned char>(tmp);
         rgb[2] = static_cast<unsigned char>(tmp);
      }
   };

   class imageMCC : public imageD3
   {
   public:
      /*!
       * \brief Constructor.
       */
      imageMCC()
         : imageD3()
      {
      }

      /*!
       * \brief Constructor from an imageRGB.
       * \param[in] image The imageRGB to copy.
       */
      imageMCC(const imageRGB& image)
      {
         copyFromRGB(image);
      }

      /*!
       * \brief Copy function
       *
       * The size of the image is changed to fit the source image.
       *
       * \param[in] image The imageRGB to copy.
       */
      void copyFromRGB(const imageRGB& image)
      {
         setSize(image.nbColumns(), image.nbRows());

         const unsigned char* rgb = image.getActualPixel(0, 0);
         double* mch = getActualPixel(0, 0);
         for (unsigned long ind = 0;
              ind < (static_cast<unsigned long>(nbColumns())
                     * static_cast<unsigned long>(nbRows()));
              ind += 1)
            RGBToMCC(&rgb[ind*3], &mch[ind*3]);
      }

      /*!
       * \brief Copy the image into an imageRGB.
       *
       * The size of the target image is set here in accordance to the
       * source image.
       *
       * \param[out] image The imageRGB.
       */
      virtual void toRGB(imageRGB* image) const
      {
         image->setSize(nbColumns(), nbRows());

         unsigned char* rgb = image->getActualPixel(0, 0);
         const double* mch = getActualPixel(0, 0);
         for (unsigned long ind = 0;
              ind < (static_cast<unsigned long>(nbColumns())
                     * static_cast<unsigned long>(nbRows()));
              ind += 1)
            MCCToRGB(&mch[ind*3], &rgb[ind*3]);
      }

   private:

      /*!
       * \brief Convert a pixel in the RGB colour space to a mean chroma of the pixel.
       * \param[in] rgb The pixel in RGB colour space.
       * \param[out] mcc The pixel in mcc colour space (must be
       * allocated beforehand).
       */
      void RGBToMCC(const unsigned char* rgb, double* mcc)
      {
         double R = static_cast<double>(rgb[0]) / 255.0;
         double G = static_cast<double>(rgb[1]) / 255.0;
         double B = static_cast<double>(rgb[2]) / 255.0;

         static const double Xn = 2.7690 * 255.0 + 1.7518 * 255.0 + 1.1300 * 255.0;
         static const double Yn = 1.0000 * 255.0 + 4.5907 * 255.0 + 0.0601 * 255.0;
         double X = 2.7690 * rgb[0] + 1.7518 * rgb[1] + 1.1300 * rgb[2];
         double Y = 1.0000 * rgb[0] + 4.5907 * rgb[1] + 0.0601 * rgb[2];

         X = cbrt(X / Xn);
         Y = cbrt(Y / Yn);

         double a = ((500.0 * (X - Y)) + 128.0)/255.0;         

         double Cb = 0.5 - 0.169 * R - 0.331 * G + 0.500 * B;
         double Cr = 0.5 + 0.500 * R - 0.419 * G - 0.081 * B;

         //mch[0] = (Cb+Cr+2*a)/4;
         mcc[0] = Cb;
         mcc[1] = Cr;
         mcc[2] = a*2;
      }

      /*!
       * \brief Copy a mean chroma of a pixel to a pixel in the RGB
       * colour space.
       * \param[in] mcc The mean chroma pixel.
       * \param[out] rgb The pixel in RGB colour space (must be allocated
       * beforehand).
       */
      void MCCToRGB(const double* mcc, unsigned char* rgb) const
      {
         double tmp = mcc[0]*255;
         if (tmp > 255)
            tmp = 255;
         else
            tmp = tmp + 0.5;
         rgb[0] = static_cast<unsigned char>(tmp);
         tmp = mcc[1]*255;
         if (tmp > 255)
            tmp = 255;
         else
            tmp = tmp + 0.5;
         rgb[1] = static_cast<unsigned char>(tmp);
         tmp = mcc[2]*255;
         if (tmp > 255)
            tmp = 255;
         else
            tmp = tmp + 0.5;
         rgb[2] = static_cast<unsigned char>(tmp);
      }
   };


   
} // namespace imalib

#endif // IMALIB_IMAGEMCH_H
