/*
 * imalib
 * 
 * Copyright 2017 Aberystwyth University
 * 
 * This library and program are free software: you can redistribute them
 * and/or modify them under the terms of the GNU General Public License as 
 * published by the Free Software Foundation, either version 3 of the License,
 * or (at your option) any later version.
 * 
 * The program and library are distributed in the hope that they will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*!
 * \file imageMapping.h
 * \brief A class to represent the transformation of an image into another.
 *
 * \author Fred Labrosse <ffl@aber.ac.uk>
 * \copyright 2017 Aberystwyth University
 */

#ifndef IMAGEMAPPING_H
#define IMAGEMAPPING_H

#include "imalib/pixelSelector.h"
#include "imalib/shape.h"

#include <map>

namespace imalib
{
/*!
 * \class imageMapping imageMapping.h imalib/imageMapping.h
 * \brief A class to represent the mapping of an image into another.
 *
 * A mapping is applied to an image using image::apply().
 *
 * The mapping is discrete, i.e. pixel by pixel.
 *
 * Source width and height and target width and height are the source and
 * target image sizes.  These should be guaranteed to contain all the pixels
 * of the mapping.  If it is not the case,
 * then it's because there is a bug in one of the functions that create the
 * mapping.
 *
 * The minimum and maximum columns are rows can also be stored (not all
 * mapping functions do it, in which case they will be 0).  These should be
 * the bounding boxes of the pixels
 * actually involved in the mapping.  This can be usefull if a further
 * transformation is applied to the mapping to chack that the pixels will
 * be contained in image boundaries.
 *
 * Also, some mappings are one-to-one, some are one-to-many (e.g. the one
 * created by createUnwrappingMapping()).
 *
 * Hash tables can also be created (and will be when the mapping is searched)
 * to make the search of sources and targets more efficient.  For this
 * mechanism to work, clear() is called when the mapping is created (and should
 * be in newly written creation functions).  If the mapping is modified but
 * not clear()ed, then hashTablesValid should be set to false so that the next
 * time the mapping is searched (using findRangeSource(), findRangeTarget(),
 * findSource() or findTarget()), the hash tables are updated
 * (updateHashTables() can also be used to that effect).
 */
class imageMapping
{
public:

   /*!
    * \brief Default constructor.
    */
   imageMapping()
      : sourceMinRow_(0),
      sourceMaxRow_(0),
      sourceMinCol_(0),
      sourceMaxCol_(0),
      targetMinRow_(0),
      targetMaxRow_(0),
      targetMinCol_(0),
      targetMaxCol_(0),
      hashTablesValid(false)
   {
   }

   /*!
    * \brief Copy constructor.
    * \param[in] mapp The mapping to copy.
    */
   imageMapping(const imageMapping& mapp)
      : sourceMinRow_(0),
      sourceMaxRow_(0),
      sourceMinCol_(0),
      sourceMaxCol_(0),
      targetMinRow_(0),
      targetMaxRow_(0),
      targetMinCol_(0),
      targetMaxCol_(0),
      hashTablesValid(false)
   {
      copy(mapp);
   }

   /*!
    * \class coordsPair imageMapping.h imalib/imageMapping.h
    * \brief A class to represent a pair of coordinates in the mapping.
    */
   class coordsPair
   {
   public:
      /*!
       * \brief Constructor.
       */
      coordsPair(unsigned short _sX, unsigned short _sY,
                 unsigned short _tX, unsigned short _tY)
         : sX(_sX), sY(_sY), tX(_tX), tY(_tY)
      {
      }

      /*!
       * \brief X coordinate of the source pixel.
       */
      unsigned short sX;

      /*!
       * \brief Y coordinate of the source pixel.
       */
      unsigned short sY;

      /*!
       * \brief X coordinate of the target pixel.
       */
      unsigned short tX;

      /*!
       * \brief Y coordinate of the target pixel.
       */
      unsigned short tY;

      /*!
       * \brief Are the two coordinates of the pair the same?
       * \return true if the coordinates are equal.
       */
      bool same() const
      {
         return ((sX == tX) && (sY == tY));
      }

      /*!
       * \brief Are the two coordinate pairs the same?
       * \param[in] coords2 The coordinates to compare.
       * \return true if the coordinates are equal.
       * \todo Passing coords2 by reference instead of values seems to be
       * impossible.  Is that true?
       */
      bool same(const coordsPair& coords2) const
      {
         return ((sX == coords2.sX) && (sY == coords2.sY)
                 && (tX == coords2.tX) && (tY == coords2.tY));
      }

      /*!
       * \brief Are the sources equal?
       * \param[in] coords2 The coordinates to compare.
       * \return true if the sources of the two sets of coordinates are equal.
       * \todo Passing coords2 by reference instead of values seems to be
       * impossible.  Is that true?
       */
      bool sameSources(coordsPair coords2) const
      {
         return ((sX == coords2.sX)
                 && (sY == coords2.sY));
      }

      /*!
       * \brief Are the sources approximatively equal?
       *
       * Approximatively means that coords2 will be compared with the current
       * coordsPair and its four nearest neighbours.
       *
       * \param[in] coords2 The coordinates to compare.
       * \return true if the sources of the two sets of coordinates are
       * approximatively equal.
       * \todo Passing coords2 by reference instead of values seems to be
       * impossible.  Is that true?
       */
      bool sameSourcesApprox(coordsPair coords2) const
      {
         return (((sX == coords2.sX) && (sY == coords2.sY))
                 || ((sX+1 == coords2.sX) && (sY == coords2.sY))
                 || ((sX-1 == coords2.sX) && (sY == coords2.sY))
                 || ((sX == coords2.sX) && (sY+1 == coords2.sY))
                 || ((sX == coords2.sX) && (sY-1 == coords2.sY))
                 || ((sX+1 == coords2.sX) && (sY+1 == coords2.sY))
                 || ((sX-1 == coords2.sX) && (sY-1 == coords2.sY))
                 || ((sX-1 == coords2.sX) && (sY+1 == coords2.sY))
                 || ((sX+1 == coords2.sX) && (sY-1 == coords2.sY)));
      }

      /*!
       * \brief Is the source of the current coordsPair the same as the
       * target of the given coordsPair?
       * \param[in] coords2 The coordinates to compare.
       * \return true if the source of this is equal to the target of coords2.
       * \todo Passing coords2 by reference instead of values seems to be
       * impossible.  Is that true?
       */
      bool sameSourceTarget(coordsPair coords2) const
      {
         return ((sX == coords2.tX) && (sY == coords2.tY));
      }

      /*!
       * \brief Is the source of the current coordsPair approximatively
       * the same as the target of the given coordsPair?
       *
       * Approximatively means that coords2 will be compared with the
       * current coordsPair and its four nearest neighbours.
       *
       * \param[in] coords2 The coordinates to compare.
       * \return true if the source of this is equal to the target of coords2.
       * \todo Passing coords2 by reference instead of values seems to be
       * impossible.  Is that true?
       */
      bool sameSourceTargetApprox(coordsPair coords2) const
      {
         return (((sX == coords2.tX) && (sY == coords2.tY))
                 || ((sX+1 == coords2.tX) && (sY == coords2.tY))
                 || ((sX-1 == coords2.tX) && (sY == coords2.tY))
                 || ((sX == coords2.tX) && (sY+1 == coords2.tY))
                 || ((sX == coords2.tX) && (sY-1 == coords2.tY))
                 || ((sX+1 == coords2.tX) && (sY+1 == coords2.tY))
                 || ((sX-1 == coords2.tX) && (sY-1 == coords2.tY))
                 || ((sX-1 == coords2.tX) && (sY+1 == coords2.tY))
                 || ((sX+1 == coords2.tX) && (sY-1 == coords2.tY)));
      }

      /*!
       * \brief Is the target of the current coordsPair the same as the
       * source of the given coordsPair?
       * \param[in] coords2 The coordinates to compare.
       * \return true if the target of this is equal to the source of coords2.
       * \todo Passing coords2 by reference instead of values seems to be
       * impossible.  Is that true?
       */
      bool sameTargetSource(coordsPair coords2) const
      {
         return ((tX == coords2.sX) && (tY == coords2.sY));
      }
   };

   /*!
    * \brief Get the source width.
    * \return The source width.
    */
   unsigned short sourceWidth() const
   {
      return(sourceWidth_);
   }

   /*!
    * \brief Get the source height.
    * \return The source height.
    */
   unsigned short sourceHeight() const
   {
      return(sourceHeight_);
   }

   /*!
    * \brief Get the target width.
    * \return The target width.
    */
   unsigned short targetWidth() const
   {
      return(targetWidth_);
   }

   /*!
    * \brief Get the target height.
    * \return The target height.
    */
   unsigned short targetHeight() const
   {
      return(targetHeight_);
   }

   /*!
    * \brief Get the source minimum row of the pixels actually involved in the
    * mapping (inclusive).
    */
   unsigned short sourceMinRow() const
   {
      return(sourceMinRow_);
   }

   /*!
    * \brief Get the source maximum row of the pixels actually involved in the
    * mapping (inclusive).
    */
   unsigned short sourceMaxRow() const
   {
      return(sourceMaxRow_);
   }

   /*!
    * \brief Get the source minimum column of the pixels actually involved in
    * the mapping (inclusive).
    */
   unsigned short sourceMinCol() const
   {
      return(sourceMinCol_);
   }

   /*!
    * \brief Get the source maximum column of the pixels actually involved in
    * the mapping (inclusive).
    */
   unsigned short sourceMaxCol() const
   {
      return(sourceMaxCol_);
   }

   /*!
    * \brief Get the target minimum row of the pixels actually involved in the
    * mapping (inclusive).
    */
   unsigned short targetMinRow() const
   {
      return(targetMinRow_);
   }

   /*!
    * \brief Get the target maximum row of the pixels actually involved in the
    * mapping (inclusive).
    */
   unsigned short targetMaxRow() const
   {
      return(targetMaxRow_);
   }

   /*!
    * \brief Get the target minimum column of the pixels actually involved in
    * the mapping (inclusive).
    */
   unsigned short targetMinCol() const
   {
      return(targetMinCol_);
   }

   /*!
    * \brief Get the targetmaximum column of the pixels actually involved in
    * the mapping (inclusive).
    */
   unsigned short targetMaxCol() const
   {
      return(targetMaxCol_);
   }

   /*!
    * \brief The vector that contains the pixel to pixel mapping.
    */
   std::vector<coordsPair> map;

   /*!
    * \brief Copy an imageMapping.
    * \param[in] mapp The imageMapping to copy.
    */
   void copy(const imageMapping& mapp)
   {
      sourceWidth_ = mapp.sourceWidth_;
      sourceHeight_ = mapp.sourceHeight_;
      targetWidth_ = mapp.targetWidth_;
      targetHeight_ = mapp.targetHeight_;
      map = mapp.map;
      hashSources = mapp.hashSources;
      hashTargets = mapp.hashTargets;
      hashTablesValid = mapp.hashTablesValid;
   }

   /*!
    * \brief Create the imageMapping to unwrap an omni-directional into a
    * panoramic image.
    *
    * So far, some of the parameters of the mapping are hard-coded:
    * - omni-directional image must be square,
    * - size of the target (panoramic image): (360 * angularRes) x
    * imgSize / 2 - innerRadius,
    *
    * Also, the mapping only applies to square images.
    *
    * This mapping is one-to-many.
    *
    * \param[in] imgSize The size of the (square) image this mapping
    * will be applied to.
    * \param[in] innerRadius The inner radius of the considered doughnut as a
    * percentage of the image size (the outer radius is 95% of the image
    * size).  When negative, 40% of the image size is used.
    * \param[in] angularRes The angular resolution (number of pixels per
    * degree of angle).
    * \param[in] power The power of the function to convert values along a
    * radial axis in the omni-directional image to a column in the panoramic
    * image (1.0 is linear, 2.0 is squared, etc).
    * \param[in] direction 1 is turn counter-clockwise (positive
    * trigonometric), -1 is turn clockwise (negative trigonometric) from left
    * to right in the unwrapped image.
    */
   void createUnwrappingMapping(unsigned short imgSize,
                                short innerRadius = -1,
                                unsigned short angularRes = 1,
                                double power = 1.0,
                                short direction = 1);

   /*!
    * \brief Create the imageMapping to unwrap an omni-directional into a
    * panoramic image.
    *
    * The position of the doughnut is verified only if not compiled with NDBUG.
    *
    * This mapping is one-to-many.
    *
    * \param[in] imgWidth
    * \param[in] imgHeight The size of the images this mapping
    * will be applied to.
    * \param[in] outerRadius The outer radius of the doughnut.
    * \param[in] innerRadius The inner radius of the doughnut.
    * \param[in] centreX
    * \param[in] centreY The coordinates of the centre of the doughnut
    * \param[in] angularRes The angular resolution (number of pixels per
    * degree of angle).
    * \param[in] power The power of the function to convert values along a
    * radial axis in the omni-directional image to a column in the panoramic
    * image (1.0 is linear, 2.0 is squared, etc).
    * \param[in] direction 1 is turn counter-clockwise (positive
    * trigonometric), -1 is turn clockwise (negative trigonometric) from left
    * to right in the unwrapped image.
    */
   void createUnwrappingMapping(unsigned short imgWidth,
                                unsigned short imgHeight,
                                unsigned short outerRadius,
                                unsigned short innerRadius,
                                unsigned short centreX,
                                unsigned short centreY,
                                unsigned short angularRes = 1,
                                double power = 1.0,
                                short direction = 1);

   /*!
    * \brief Create the imageMapping to unwrap  an omni-directional into a
    * panoramic image.
    *
    * The parameters of the mapping are taken from the configuration file
    * specified.
    *
    * \param[in] imgSize The size of the (square) image this mapping
    * will be applied to.
    * \param[in] fileName Name of the configuration file.
    */
   void createUnwrappingMapping(unsigned short imgSize,
                                const std::string& fileName);

   /*!
    * \brief Create the imageMapping to unwrap an omni-directional into a
    * panoramic image.
    *
    * The parameters of the mapping are taken from the configuration file
    * specified.
    *
    * \param[in] imgWidth
    * \param[in] imgHeight The size of the image this mapping will be
    * applied to.
    * \param[in] fileName Name of the configuration file.
    */
   void createUnwrappingMapping(unsigned short imgWidth,
                                unsigned short imgHeight,
                                const std::string& fileName);

   /*!
    * \brief Create the imageMapping to unwrap a transformed
    * omni-directional image into a panoramic image.
    *
    * So far, some of the parameters of the mapping are hard-coded:
    * - omni-directional image must be square,
    * - size of the target (panoramic image): 360 x 60,
    * - radius of the ignored centre: 40 * omni-directional image size / 200.
    *
    * This mapping is one-to-many.
    *
    * The original mapping must use and produce square images of the same size.
    * This is only verified if compiled without NDEBUG.
    *
    * \param[in] mapping The mapping to apply to the omni-directional
    * image before unwrapping.
    * \param[in] power The power of the function to convert values along a
    * radial axis in the omni-directional image to a column in the panoramic
    * image (1.0 is linear, 2.0 is squared, etc).
    */
   void createUnwrappingMapping(imageMapping& mapping,
                                double power = 1.0);

   /*!
    * \brief Create a mapping that does nothing!
    *
    * The mapping will map an image onto itself.  It is a one-to-one mapping.
    *
    * \param[in] width The width if the image.
    * \param[in] height The height of the image.
    */
   void createNoopMapping(unsigned short width,
                          unsigned short height);

   /*!
    * \brief Create a mapping that will vertically shift pixels.
    *
    * If a shift moves a pixel outside of the target image, then the
    * pixel is ignored.
    *
    * It is a one-to-one mapping.
    *
    * \param width
    * \param height (IN) The size of the images.
    * \param shift (IN) By how much to shift (positive shifts down).
    * \param minRow
    * \param maxRow (IN) Minimum and maximum (inclusive) rows that will
    * be shifted.  The other pixels will not be shifted.
    */
   void createVerticalShiftMapping(unsigned short width,
                                   unsigned short height,
                                   long shift,
                                   unsigned short minRow,
                                   unsigned short maxRow);

   /*!
    * \brief Create a mapping that will horizontally shift pixels.
    *
    * If a shift moves a pixel outside of the target image, then the
    * pixel is ignored.
    *
    * It is a one-to-one mapping.
    *
    * \param width
    * \param height (IN) The size of the images.
    * \param shift (IN) By how much to shift (positive shifts to the right).
    * \param minColumn
    * \param maxColumn (IN) Minimum and maximum (inclusive) columns that
    * will be shifted.  The other pixels will not be shifted.
    */
   void createHorizontalShiftMapping(unsigned short width,
                                     unsigned short height,
                                     long shift,
                                     unsigned short minColumn,
                                     unsigned short maxColumn);

   /*!
    * \brief Create a warping mapping.
    *
    * It uses the algorithm (and the code is based on the one) described
    * in [\ref Hec94].
    *
    * The boundaries of the target in the source define the size of the target.
    * Top and bottom should thus have the same size.  The same applies to
    * right and left.  This is only verified if compiled without NDEBUG.
    *
    * The boundaries must run clockwise from the top left corner of the source.
    *
    * Coordinates in the source image (implied by the warping boundaries) will
    * be truncated to the limits of the source image, unless wrapX is true, in
    * which case the x coordinates will be wrapped around (useful if the image
    * is a panoramic image).
    *
    * \param widthS
    * \param heightS (IN) The width and height of the source image.
    * \param top
    * \param right
    * \param bottom
    * \param left (IN) The boundaries of the target in the source.
    * \param wrapX (IN) If true, then wrap the X coordinates in the source
    * image.
    */
   void createWarpingMapping(unsigned short widthS,
                             unsigned short heightS,
                             const curve& top,
                             const curve& right,
                             const curve& bottom,
                             const curve& left,
                             bool wrapX = false);

   /*!
    * \brief Create a mapping that rotates images by 180deg, for when the
    * camera is upside-down.
    *
    * \param[in] width The width if the image.
    * \param[in] height The height of the image.
    */
   void createUpsideDownMapping(unsigned short width,
                                unsigned short height);

   /*!
    * \brief Create a mapping that does a horizontal flip.
    *
    * \param[in] width The width if the image.
    * \param[in] height The height of the image.
    */
   void createHorizontalFlipMapping(unsigned short width,
                                    unsigned short height);

   /*!
    * \brief Apply to the mapping the given one.
    *
    * The source and target pixels will be transformed according to the
    * given mapping.
    *
    * If a source or target pixel is not found in the given mapping, that pair
    * is removed from the original mapping and false is returned.
    *
    * The application of the mapping can be strict (the default) or
    * approximative.  The problem is that a strict application can result in
    * holes left between pixels from the current mapping if the second mapping
    * does not necessarily use contiguous image pixels (such as the mapping
    * created by createUnwrappingMapping()).  The approximative application
    * will look for neighbours of the source and target pixels of the current
    * mapping in the mapping to be applied if these pixels are not handled by
    * this mapping.  The considered neighbours are the four closest pixels.
    *
    * This should work for one-to-one and one-to-many mappings.
    *
    * \param[in] mapping The mapping to apply.
    * \param[in] approx If true, then allow an approximative application
    * of the mapping.
    * \return true if all points of the current mapping have been
    * transformed, false otherwise.
    * \todo Some of this function could use the hashtables for more efficient
    * searches.  Is it worth doing it?
    */
   bool apply(imageMapping& mapping, bool approx = false);

   /*!
    * \brief Apply a pixelSelector to the mapping.
    *
    * Only those pixels that are mapped by the mapping AND in the selector are
    * kept.
    *
    * The function is certainly not efficient and is therefore not meant to be
    * used repeatedly.  In particular, the current mapping is copied into a
    * temporary one and then filtered back into the original.
    *
    * \param[in] selector The pixel selector to use.
    * \todo This function does not seem to work (no pixels kept?).
    */
   void apply(const pixelSelector& selector);

   /*!
    * \brief Concatenate two image mappings.
    *
    * A new mapping is created that will have all the pairs of the second
    * mapping that have a source equal to a target in the first (original)
    * mapping.  Some pairs of the second or first mappings might not be
    * represented in the new mapping.  The new mapping replaces the first
    * (original) mapping.
    *
    * The source size of the new mapping is the source size of the original
    * mapping while the target size of the new mapping is the target size of the
    * second mapping.
    *
    * \param[in] mapping The second image mapping.
    */
   void concatenate(const imageMapping& mapping);

   /*!
    * \brief Is the mapping an identity mapping?
    * 
    * This only means that the source and target coordinates in the mapping
    * are all equal, not that an image to which the mapping is applied remains
    * unchanged.
    * \return true if the mapping is an identity, false otherwise.
    */
   bool isIdentity()
   {
      std::vector<imageMapping::coordsPair>::const_iterator
         iter = map.begin();
      while (iter < map.end())
      {
         if (iter->same())
            ++iter;
         else
            return (false);
      }
      return(true);
   }

   /*!
    * \brief The type of hash table.
    */
   typedef std::multimap<unsigned long, unsigned long> hashTableType;

   /*!
    * \brief Find a range of sources in the current mapping.
    *
    * A range is returned (a pair of iterators).  The two iterators are equal
    * if the source was not found.
    *
    * \param[in] x
    * \param[in] y The coordinates of the searched point.
    */
   std::pair<hashTableType::const_iterator,hashTableType::const_iterator>
      findRangeSource(unsigned short x, unsigned short y)
   {
      if (!hashTablesValid)
         updateHashTables();
      return(hashSources.equal_range(y * sourceWidth_ + x));
   }

   /*!
    * \brief Find a range of targets in the current mapping.
    *
    * A range is returned (a pair of iterators).  The two iterators are equal
    * if the source was not found.
    *
    * \param[in] x
    * \param[in] y The coordinates of the searched point.
    */
   std::pair<hashTableType::const_iterator,hashTableType::const_iterator>
      findRangeTarget(unsigned short x, unsigned short y)
   {
      if (!hashTablesValid)
         updateHashTables();
      return(hashTargets.equal_range(y * targetWidth_ + x));
   }

   /*!
    * \brief Find the first occurence of the given position as a source
    * in the current mapping.
    *
    * The iterator is equal to endHashSources() if the position is not found.
    *
    * \param[in] x
    * \param[in] y The coordinates of the searched point.
    */
   hashTableType::const_iterator findSource(unsigned short x, unsigned short y)
   {
      if (!hashTablesValid)
         updateHashTables();
      return(hashSources.find(y * sourceWidth_ + x));
   }

   /*!
    * \brief Find the first occurence of the given position as a target
    * in the current mapping.
    *
    * The iterator is equal to endHashTargets() if the position is not found.
    *
    * \param[in] x
    * \param[in] y The coordinates of the searched point.
    */
   hashTableType::const_iterator findTarget(unsigned short x, unsigned short y)
   {
      if (!hashTablesValid)
         updateHashTables();
      return(hashTargets.find(y * targetWidth_ + x));
   }

   /*!
    * \brief Get the end of the source hashtable.
    */
   hashTableType::const_iterator endHashSources()
   {
      return(hashSources.end());
   }

   /*!
    * \brief Get the end of the target hashtable.
    */
   hashTableType::const_iterator endHashTargets()
   {
      return(hashTargets.end());
   }

protected:

   /*!
    * \brief Clear the mapping.
    */
   void clear()
   {
      map.clear();
      hashTablesValid = false;
   }

   /*!
    * \brief The width of the source image.
    */
   unsigned short sourceWidth_;

   /*!
    * \brief The height of the source image.
    */
   unsigned short sourceHeight_;

   /*!
    * \brief The width of the target image.
    */
   unsigned short targetWidth_;

   /*!
    * \brief The height of the target image.
    */
   unsigned short targetHeight_;

   /*!
    * \brief Source minimum row (inclusive).
    * 
    * In some cases, it is interesting to have the bounding box of the actual
    * pixels involved in the mapping.
    */
   unsigned short sourceMinRow_;

   /*!
    * \brief Source maximum row (inclusive).
    * 
    * In some cases, it is interesting to have the bounding box of the actual
    * pixels involved in the mapping.
    */
   unsigned short sourceMaxRow_;

   /*!
    * \brief Source minimum column (inclusive).
    * 
    * In some cases, it is interesting to have the bounding box of the actual
    * pixels involved in the mapping.
    */
   unsigned short sourceMinCol_;

   /*!
    * \brief Source maximum column (inclusive).
    * 
    * In some cases, it is interesting to have the bounding box of the actual
    * pixels involved in the mapping.
    */
   unsigned short sourceMaxCol_;

   /*!
    * \brief Target minimum row (inclusive).
    * 
    * In some cases, it is interesting to have the bounding box of the actual
    * pixels involved in the mapping.
    */
   unsigned short targetMinRow_;

   /*!
    * \brief Target maximum row (inclusive).
    * 
    * In some cases, it is interesting to have the bounding box of the actual
    * pixels involved in the mapping.
    */
   unsigned short targetMaxRow_;

   /*!
    * \brief Target minimum column (inclusive).
    * 
    * In some cases, it is interesting to have the bounding box of the actual
    * pixels involved in the mapping.
    */
   unsigned short targetMinCol_;

   /*!
    * \brief Target maximum column (inclusive).
    * 
    * In some cases, it is interesting to have the bounding box of the actual
    * pixels involved in the mapping.
    */
   unsigned short targetMaxCol_;

private:

   /*!
    * \brief Update the hash tables used for efficient searching.
    *
    * This function should be called at the end of each mapping modifications.
    */
   void updateHashTables();

   /*!
    * \brief The hash table to look for sources.
    *
    * The first value is the pixel index (y * sourceWidth + x) of the source.
    * The second value is the index in the map.
    */
   hashTableType hashSources;

   /*!
    * \brief The hash table to look for targets.
    *
    * The first value is the pixel index (y * targetWidth + x) of the target.
    * The second value is the index in the map.
    */
   hashTableType hashTargets;

   /*!
    * \brief Tells if the hash tables are valid or not.
    *
    * If the hash tables are not valid, then they will be updated when
    * findRangeSource() or findRangeTarget() are called.
    */
   bool hashTablesValid;
};

} // namespace imalib

#endif // IMAGEMAPPING_H
