/*
 * imalib
 * 
 * Copyright 2017 Aberystwyth University
 * 
 * This library and program are free software: you can redistribute them
 * and/or modify them under the terms of the GNU General Public License as 
 * published by the Free Software Foundation, either version 3 of the License,
 * or (at your option) any later version.
 * 
 * The program and library are distributed in the hope that they will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*!
 * \file imageRGB.h
 * \brief Defines classes to play with RGB and rgb images.
 * \author Fred Labrosse <ffl@aber.ac.uk>
 * \copyright 2017 Aberystwyth University
 */

#ifndef IMALIB_IMAGERGB_H
#define IMALIB_IMAGERGB_H

#include <imalib/imageD3.h>

#include <cassert>
#include <cstdlib>
#include <exception>
#include <fstream>
#include <iostream>
#include <map>
#include <sstream>
#include <string>

namespace imalib
{
   /*!
    * \class imageRGB imageRGB.h imalib/imageRGB.h
    * \brief A class to represent images in the RGB colour space (values
    * between 0 and 255).
    */
   class imageRGB : public image<unsigned char,3>
   {
   public:
      /*!
       * \brief Constructor.
       */
      imageRGB()
         : image<unsigned char,3>()
      {
      }

      /*!
       * \brief Copy constructor
       */
      imageRGB(const imageRGB& img)
         : image<unsigned char, 3>()
      {
         copy(img);
      }

      /*!
       * \brief Constructor.
       */
      imageRGB(short cols, short rows, short tCols = 0, short tRows = 0)
         : image<unsigned char,3>(cols, rows, tCols, tRows)
      {
      }

      /*!
       * \brief Set a pixel from a BGR pixel.
       *
       * Checks on the coordinates are only performed if compiled without
       * NDEBUG.
       *
       * \param col (IN) The column of the pixel.
       * \param row (IN) The row of the pixel.
       * \param BGR (IN) The BGR pixel.
       */
      void setPixelBGR(unsigned short col, unsigned short row,
                       const unsigned char* BGR)
      {
         assert((col <= nbColumns())
                || (row <= nbRows()));

         unsigned char* pixel = getActualPixel(col, row);
         pixel[0] = BGR[2];
         pixel[1] = BGR[1];
         pixel[2] = BGR[0];
      }

      /*!
       * \brief Output the supported image formats.
       * \param output (IN/OUT) The stream where the formats are written.
       * \sa getReadFormats(), getWriteFormats().
       */
      static void outputSupportedImageFormats(std::ostream& output);

      /*!
       * \brief Give the supported image formats while reading.
       * \return A list of char*, each containing a format.
       */
      static const std::string& getReadFormats() {return(readFormats);}

      /*!
       * \brief Give the supported image formats while writing.
       * \return A list of char*, each containing a format.
       */
      static const std::string& getWriteFormats() {return(writeFormats);}

      /*!
       * \brief Read an imageRGB from a file.
       * \param[in] fileName The name of the file containing the image.
       * \return true if the operation is successful.
       */
      bool read(const std::string& fileName);

      /*!
       * \brief Write the image into a file.
       * \param[in] fileName The name of the file.
       * \return true if successful, false otherwise.
       */
      bool write(const std::string& fileName) const;

      /*!
       * \brief Write the image into a file.
       * \param[in] fileName The name of the file.
       * \param[in] comment The comment to add to the file, new line has to start with a # so '\n#'
       * \return true if successful, false otherwise.
       */
      bool write(const std::string& fileName, const std::string& comment) const;


   protected:

#ifdef WITH_PNG
      /*!
       * \brief Reads an image in the PNG format.
       * \param[in] fileName  The name of the file containing the image.
       * \return true if the operation is successful.
       */
      bool readPNG(const std::string& fileName);

      /*!
       * \brief Writes an image in the PNG format.
       * \param[in] fileName The name of the file that will contain the
       * image.
       * \return true if the operation is successful.
       */
      bool writePNG(const std::string& fileName) const;
#endif // WITH_PNG

#ifdef WITH_JPEG
      /*!
       * \brief Reads an image in the JPEG format.
       * \param[in] fileName  The name of the file containing the image.
       * \return true if the operation is successful.
       */
      bool readJPEG(const std::string& fileName);
#endif // WITH_JPEG

      /*!
       * \brief Reads an image in the PPM format.
       * \param[in] fileName The name of the file containing the image.
       * \return true if the operation is successful.
       */
      bool readPPM(const std::string& fileName);

      /*!
       * \brief Write an image in the PPM format.
       * \param[in] fileName The name of the file containing the image.
       * \return true if the operation is successful.
       */
      bool writePPM(const std::string& fileName) const;

      /*!
       * \brief Write an image in the PPM format. multi-line comments have to include # after the new line so '\n#'
       * \param[in] fileName The name of the file containing the image.
       * \param[in] comment The string to be embedded in the image.
       * \return true if the operation is successful.
       */
      bool writePPM(const std::string& fileName ,const std::string& comment) const;

      /*!
       * \brief Write an image in the XPM format.
       * \param[in] fileName The name of the file that will contain the
       * image.
       * \return true if the operation is successful.
       */
      bool writeXPM(const std::string& fileName) const;

      /*!
       * \brief Write an image in txt format.
       * 
       * This write a desaturated image (grey) where each line of the txt file
       * corresponds to a row of the image and therefore has nbRows() lines.
       *  Values are between 0 and 255.
       * \param[in] fileName The name of the file that will contain the image.
       * \return true if the operation is successful.
       */
      bool writeTxt(const std::string& fileName) const;

   private:
      // Formats in read and write.
      static const std::string readFormats;
      static const std::string writeFormats;

   };


   /*!
     \class imageRgb imageRGB.h imalib/imageRGB.h
     \brief A class to represent images in the RGB colour space (values
     between 0.0 and 1.0 as doubles).
   */
   class imageRgb : public imageD3
   {
   public:
      /*!
        \brief Constructor.
      */
      imageRgb()
         : imageD3()
      {
      }

      /*!
        \brief Constructor.
      */
      imageRgb(short cols, short rows, short tCols = 0, short tRows = 0)
         : imageD3(cols, rows, tCols, tRows)
      {
      }

      /*!
        \brief Constructor from an imageRGB.
        \param image (IN) The imageRGB to copy.
      */
      imageRgb(const imageRGB& image)
      {
         copyFromRGB(image);
      }

      /*!
        \brief Copy function

        The size of the image is changed to fit the source image.

        \param image (IN) The imageRGB to copy.
      */
      void copyFromRGB(const imageRGB& image)
      {
         setSize(image.nbColumns(), image.nbRows());

         const unsigned char* RGB = image.getActualPixel(0, 0);
         double* rgb = getActualPixel(0, 0);
         for (unsigned long ind = 0;
              ind < nbColumns() * nbRows() * 3U;
              ind += 3)
            RGBToRgb(&RGB[ind], &rgb[ind]);
      }

      /*!
        \brief Copy the imageRgb into an imageRGB.

        The size of the target image is set here in accordance to the
        source image.

        \param image (OUT) The imageRGB.
      */
      void toRGB(imageRGB* image) const
      {
         image->setSize(nbColumns(), nbRows());

         unsigned char* RGB = image->getActualPixel(0, 0);
         const double* rgb = getActualPixel(0, 0);
         for (unsigned long ind = 0;
              ind < nbColumns() * nbRows() * 3U;
              ind += 3)
            rgbToRGB(&rgb[ind], &RGB[ind]);
      }

      /*!
        \brief Set a pixel from a RGB pixel (unsigned chars).

        Checks on the coordinates are only performed if compiled without
        NDEBUG.

        \param col (IN) The column of the pixel.
        \param row (IN) The row of the pixel.
        \param RGB (IN) The RGB pixel.
      */
      void setPixelRGB(unsigned short col, unsigned short row,
                       const unsigned char* RGB)
      {
         assert((col <= nbColumns())
                || (row <= nbRows()));
         RGBToRgb(RGB, getActualPixel(col, row));
      }

      /*!
        \brief Get a pixel and convert it to an RGB pixel (unsigned chars).

        Checks on the coordinates are only performed if compiled without
        NDEBUG.

        \param col (IN) The column of the pixel.
        \param row (IN) The row of the pixel.
        \param RGB (OUT) Where the RGB pixel will be stored (must be
        allocated beforehand).
      */
      void getPixelRGB(unsigned short col, unsigned short row,
                       unsigned char* RGB)
      {
         assert((col <= nbColumns())
                || (row <= nbRows()));
         rgbToRGB(getActualPixel(col, row), RGB);
      }

      /*!
        \brief Read an imageRGB from a file.
        \param fileName (IN) The name of the file containing the image.
        \return true if the operation is successful.
      */
      bool read(const std::string& fileName);

      /*!
        \brief Write the image into a file.
        \param fileName (IN) The name of the file.
        \return true if successful, false otherwise.
      */
      bool write(const std::string& fileName);

   protected:

#ifdef WITH_OPENEXR
      /*!
        \brief Reads an image in the OpenEXR format.
        \param fileName  (IN) The name of the file containing the image.
        \return true if the operation is successful.
      */
      bool readOpenEXR(const std::string& fileName);

      /*!
        \brief Writes an image in the OpenEXR format.
        \param fileName (IN) The name of the file that will contain the
        image.
        \return true if the operation is successful.
      */
      bool writeOpenEXR(const std::string& fileName);
#endif

   private:

      /*!
        \brief Convert a pixel in the rgb colour space (0.0 to 1.0) to a
        pixel in the RGB colour space (0 to 255).

        rgb values are checked only if compiled without NDEBUG.

        \param rgb (IN) The pixel in rgb colour space.
        \param RGB (OUT) The pixel in RGB colour space (must be allocated
        beforehand).
      */
      static void rgbToRGB(const double* rgb, unsigned char* RGB);

      /*!
        \brief Convert a pixel in the RGB colour space (0 to 255) to a
        pixel in the rgb colour space (0.0 to 1.0).
        \param RGB (OUT) The pixel in RGB colour space.
        \param rgb (IN) The pixel in rgb colour space (must
        be allocated beforehand).
      */
      static void RGBToRgb(const unsigned char* RGB, double* rgb);
   };

} // namespace imalib

#endif // IMALIB_IMAGERGB_H
