/*
 * imalib
 * 
 * Copyright 2017 Aberystwyth University
 * 
 * This library and program are free software: you can redistribute them
 * and/or modify them under the terms of the GNU General Public License as 
 * published by the Free Software Foundation, either version 3 of the License,
 * or (at your option) any later version.
 * 
 * The program and library are distributed in the hope that they will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*!
 * \file imageReceiver.h
 * \brief A class to wrap the reception of an image from a socket.
 *
 * \author Fred Labrosse <ffl@aber.ac.uk>
 * \copyright 2017 Aberystwyth University
 */

#ifndef IMAGERECEIVER_H
#define IMAGERECEIVER_H

#include <imalib/image.h>
#include <imalib/imageRGB.h>

// #include <clientSocket++.h>

#include <string>

#include <boost/asio.hpp>
using boost::asio::ip::tcp;

namespace imalib
{
/*!
 * \class imageReceiver imageReceiver.h imalib/imageReceiver.h
 * \brief A class to encapsulate the reception on an imageRGB from a socket.
 *
 * The class blocks on read of an image.
 *
 * Most functions will catch "normal" boost::asio exceptions and throw
 * std::runtime_error exceptions.
 *
 * \sa imageTransmitter.
 */
class imageReceiver
{
public:
   /*!
    * \brief Constructor.
    * \param[in] ioService The boost::asio IO service.
    * \param[in] server The name of the computer where the server is running.
    * \param[in] port The port number the server is listening to.
    */
   imageReceiver(boost::asio::io_service& ioService,
                 const std::string& server,
                 const std::string& port);

   /*!
    * \brief Destructor.
    */
   ~imageReceiver()
   {
   }

   /*!
    * \brief Are we connected to a server?
    */
   bool isConnected() {return(connected);}

   /*!
    * \brief To receive an image.
    *
    * Should follow a successful call to askForImage().
    *
    * \param[out] img The received image.
    * \sa askForImage().
    */
   void receiveImage(imageRGB* img) throw(std::runtime_error);

   /*!
    * \brief Tell the server we want an image.
    *
    * Use askForImage() first, then use receiveImage().
    *
    * \sa receiveImage().
    */
   void askForImage() throw(std::runtime_error);

   /*!
    * \brief Tell the server we are done.
    */
   void tellToFinish() throw(std::runtime_error);

private:

   /*!
    * \brief Tells if we are connected or not.
    */
   bool connected;

   /*!
    * \brief The TCP socket used.
    */
   tcp::socket sock;
};
} // namespace imalib

#endif
