/*
 * imalib
 * 
 * Copyright 2017 Aberystwyth University
 * 
 * This library and program are free software: you can redistribute them
 * and/or modify them under the terms of the GNU General Public License as 
 * published by the Free Software Foundation, either version 3 of the License,
 * or (at your option) any later version.
 * 
 * The program and library are distributed in the hope that they will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*!
 * \file imageTransmitter.h
 * \brief A class to wrap the serving of images on a socket.
 *
 * \author Fred Labrosse <ffl@aber.ac.uk>
 * \copyright 2017 Aberystwyth University
 */

#ifndef IMAGETRANSMITTER_H
#define IMAGETRANSMITTER_H

#include "imalib/imageGrabber.h"
#include "imalib/image.h"
#include "imalib/imageRGB.h"

#include <boost/bind.hpp>
#include <boost/asio.hpp>
using boost::asio::ip::tcp;

namespace imalib
{

/*!
 * \class imageTransmitter imageTransmitter.h imalib/imageTransmitter.h
 * \brief A class to encapsulate the transmission of an imageRGB to a socket.
 *
 * The server can accept several simultaneous connections.
 *
 * When an image is sent (following the reception of a get image command (see
 * commProtocol.h), 2 buffers are sent.
 *
 * The first buffer is 4 bytes long and contains the size of the image, the
 * first 2 bytes being the width, the second 2 bytes being the height.
 * Encoding is unsigned, which gives widths and heights of up to 65536
 * (unsigned short under linux).
 *
 * The second buffer is the data.  There is three bytes
 * per pixel (RGB) and the number of pixels depends on the size of the image.
 *
 * All the functions can throw exceptions.  They will be the "normal"
 * boost::asio exceptions.
 *
 * \sa imageReceiver.
 */
class imageTransmitter
{

protected:
   // Forward decleration.
   class session;

public:
   /*!
    * \brief Constructor.
    * \param[in] ioService The boost::asio IO service.
    * \param[in] sockPort The port to use for the server to wait on.
    * \param[in,out] imgGrabber Where the images come from.
    */
   imageTransmitter(boost::asio::io_service& ioService,
                    int sockPort,
                    imageGrabber* imgGrabber);

   /*!
    * \brief Destructor.
    */
   ~imageTransmitter()
   {
   }

   /*!
    * \brief Handle acception of connections.
    * \param[in,out] newSession The accepted session.
    * \param[in] error The error code.
    */
   void handleAccept(session* newSession,
                     const boost::system::error_code& error)
   {
      if (!error)
      {
         std::cout << "Connection accepted.\n";
         newSession->start();
         newSession = new session(ioService_, theGrabber);
         acceptor_.async_accept(newSession->socket(),
                                boost::bind(&imageTransmitter::handleAccept,
                                            this,
                                            newSession,
                                            boost::asio::placeholders::error));
      }
      else
      {
         delete newSession;
      }
   }

protected:

   /*!
    * \class session imageTransmitter.h imalib/imageTransmitter.h
    * \brief A class to handle a server session.
    */
   class session
   {
   public:
      /*!
       * \brief Constructor.
       * \param[in,out] ioService The boost:asio IO service.
       * \param[in,out] imgGrabber Where the images come from.
       */
      session(boost::asio::io_service& ioService,
              imageGrabber* imgGrabber)
         : socket_(ioService),
         grabber_(imgGrabber)
      {
      }

      /*!
       * \brief Destructor.
       */
      ~session()
      {
         std::cout << "Connection closed.\n";
      }

      /*!
       * \brief Get the socket used.
       * \return A reference to the socket.
       */
      tcp::socket& socket()
      {
         return socket_;
      }

      /*!
       * \brief Start the reading of a command.
       */
      void start()
      {
         socket_.async_read_some(
                   boost::asio::buffer(&command_, 1),
                   boost::bind(&session::handleRead, this,
                               boost::asio::placeholders::error));
      }

      /*!
       * \brief Handle the reading of a command.
       *
       * The commands are defined in commProtocol.h.
       *
       * \param[in] error The error code.
       */
      void handleRead(const boost::system::error_code& error);

      /*!
       * \brief Handle the writing of data.
       *
       * When data has been sent successfully, a new reading is initiated.
       * \param[in] error The error code.
       */
      void handleWrite(const boost::system::error_code& error)
      {
         if (!error)
         {
            start();
         }
         else
         {
            delete this;
         }
      }

   private:

      /*!
       * \brief The tcp socket used.
       */
      tcp::socket socket_;

      /*!
       * \brief The buffer to receive read commands.
       */
      char command_;

      /*!
       * \brief The imageGrabber to use.
       */
      imageGrabber* grabber_;

      /*!
       * \brief A grabbed image.
       */
      imageRGB image_;

   };

private:

   /*!
    * \brief Copy constructor.
    *
    * We don't provide an implementation of that on purpose!
    */
   imageTransmitter(const imageTransmitter* imgTrans);

   /*!
    * \brief Assignment operator.
    *
    * We don't provide an implementation of that on purpose!
    */
   imageTransmitter& operator=(const imageTransmitter& imgTrans);

   /*!
    * \brief The imageGrabber to use.
    */
   imageGrabber* theGrabber;

   /*!
    * \brief The boost::asio IO service
    */
   boost::asio::io_service& ioService_;

   /*!
    * \brief The asio tcp acceptor.
    */
   tcp::acceptor acceptor_;

};

} // namespace imalib

#endif
