/*
 * imalib
 * 
 * Copyright 2017 Aberystwyth University
 * 
 * This library and program are free software: you can redistribute them
 * and/or modify them under the terms of the GNU General Public License as 
 * published by the Free Software Foundation, either version 3 of the License,
 * or (at your option) any later version.
 * 
 * The program and library are distributed in the hope that they will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*!
 * \file imageYCbCr.h
 * \brief Defines classes to play with YCbCr images.
 * \author Fred Labrosse <ffl@aber.ac.uk>
 * \copyright 2017 Aberystwyth University
 */

#ifndef IMALIB_IMAGEYCBCR_H
#define IMALIB_IMAGEYCBCR_H

#include <cassert>
#include <cstdlib>
#include <exception>
#include <fstream>
#include <iostream>
#include <map>
#include <sstream>
#include <string>

#ifdef WITH_PNG
#  include <png.h>
#endif // WITH_PNG

#include <imalib/imageRGB.h>
#include <imalib/imageD3.h>
#include <imalib/imageD2.h>

namespace imalib
{
   /*!
    * \class imageYCbCr imageYCbCr.h imalib/imageYCbCr.h
    * \brief A class to represent images in the YCbCr colour space.
    *
    * This is the YCbCr as defined at http://www.equasys.de/colorconversion.html#YCbCr-RGBColorFormatConversion:
    * \f[
    * \begin{array}{l}
    * \displaystyle
    * Y = 0.299 R + 0.587 G + 0.114 B,\\
    * Cb = 0.5 - 0.169 R - 0.331 G + 0.500 B,\\
    * Cr = 0.5 + 0.500 R - 0.419 G - 0.081 B,
    * \end{array}
    * \f]
    * \f[
    * \begin{array}{l}
    * \displaystyle
    * R = Y + 1.140 (Cr - 0.5),\\
    * G = Y - 0.343 (Cb - 0.5) - 0.711 (Cr - 0.5),\\
    * B = Y + 1.765 ( Cb - 0.5),
    * \end{array}
    * \f]
    * where \f$ 0 \leq R,G,B \leq 1 \f$, \f$ 0 \leq Y,Cb,Cr \leq 1 \f$.
    *
    * Conversion from and to RGB is provided.
    */
   class imageYCbCr : public imageD3
   {
   public:
      /*!
       * \brief Constructor.
       */
      imageYCbCr()
         : imageD3()
      {
      }

      /*!
       * \brief Constructor from an imageRGB.
       * \param[in] image The imageRGB to copy.
       */
      imageYCbCr(const imageRGB& image)
      {
         copyFromRGB(image);
      }

      /*!
       * \brief Copy function
       *
       * The size of the image is changed to fit the source image.
       *
       * \param[in] image The imageRGB to copy.
       */
      void copyFromRGB(const imageRGB& image)
      {
         setSize(image.nbColumns(), image.nbRows());

         const unsigned char* rgb = image.getActualPixel(0, 0);
         double* ycbcr = getActualPixel(0, 0);
         for (unsigned long ind = 0;
              ind < nbColumns() * nbRows() * 3U;
              ind += 3)
            RGBToYCbCr(&rgb[ind], &ycbcr[ind]);
      }

      /*!
       * \brief Copy the image into an imageRGB.
       *
       * The size of the target image is set here in accordance to the
       * source image.
       *
       * \param[out] image The imageRGB.
       */
      virtual void toRGB(imageRGB* image) const
      {
         image->setSize(nbColumns(), nbRows());

         unsigned char* rgb = image->getActualPixel(0, 0);
         const double* ycbcr = getActualPixel(0, 0);
         for (unsigned long ind = 0;
              ind < nbColumns() * nbRows() * 3U;
              ind += 3)
            YCbCrToRGB(&ycbcr[ind], &rgb[ind]);
      }

   private:

      /*!
       * \brief Convert a pixel in the RGB colour space to a pixel in the
       * YCbCr colour space.
       * \param[in] rgb The pixel in RGB colour space.
       * \param[out] ycbcr The pixel in YCbCr colour space (must be
       * allocated beforehand).
       */
      void RGBToYCbCr(const unsigned char* rgb, double* ycbcr)
      {
         double R = static_cast<double>(rgb[0]) / 255.0;
         double G = static_cast<double>(rgb[1]) / 255.0;
         double B = static_cast<double>(rgb[2]) / 255.0;
         ycbcr[0] = 0.299 * R + 0.587 * G + 0.114 * B;
         ycbcr[1] = 0.5 - 0.169 * R - 0.331 * G + 0.500 * B;
         ycbcr[2] = 0.5 + 0.500 * R - 0.419 * G - 0.081 * B;
      }

      /*!
       * \brief Convert a pixel in the YCbCr colour space to a pixel in the RGB
       * colour space.
       * \param[in] ycbcr The pixel in YCbCr colour space.
       * \param[out] rgb The pixel in RGB colour space (must be allocated
       * beforehand).
       */
      void YCbCrToRGB(const double* ycbcr, unsigned char* rgb) const
      {
         double tmp = (ycbcr[0] + 1.140 * (ycbcr[2] - 0.5)) * 255.0;
         if (tmp > 255)
            tmp = 255;
         else
            tmp = tmp + 0.5;
         rgb[0] = static_cast<unsigned char>(tmp);
         tmp = (ycbcr[0]
                - 0.343 * (ycbcr[1] - 0.5)
                - 0.711 * (ycbcr[2] - 0.5)) * 255.0;
         if (tmp > 255)
            tmp = 255;
         else
            tmp = tmp + 0.5;
         rgb[1] = static_cast<unsigned char>(tmp);
         tmp = (ycbcr[0] + 1.765 * (ycbcr[1] - 0.5)) * 255.0;
         if (tmp > 255)
            tmp = 255;
         else
            tmp = tmp + 0.5;
         rgb[2] = static_cast<unsigned char>(tmp);
      }
   };
   
   
      /*!
    * \class image_CbCr imageYCbCr.h imalib/imageYCbCr.h
    * \brief A class to represent images in the _CbCr colour space (without the Y).
    *
    * This is the YCbCr as defined at http://www.equasys.de/colorconversion.html#YCbCr-RGBColorFormatConversion:
    * \f[
    * \begin{array}{l}
    * \displaystyle
    * Y = 0.299 R + 0.587 G + 0.114 B,\\
    * Cb = 0.5 - 0.169 R - 0.331 G + 0.500 B,\\
    * Cr = 0.5 + 0.500 R - 0.419 G - 0.081 B,
    * \end{array}
    * \f]
    * \f[
    * \begin{array}{l}
    * \displaystyle
    * R = Y + 1.140 (Cr - 0.5),\\
    * G = Y - 0.343 (Cb - 0.5) - 0.711 (Cr - 0.5),\\
    * B = Y + 1.765 ( Cb - 0.5),
    * \end{array}
    * \f]
    * where \f$ 0 \leq R,G,B \leq 1 \f$, \f$ 0 \leq Y,Cb,Cr \leq 1 \f$.
    *
    * Conversion from and to RGB is provided.
    */
   class image_CbCr : public imageD2
   {
   public:
      /*!
       * \brief Constructor.
       */
      image_CbCr()
         : imageD2()
      {
      }

      /*!
       * \brief Constructor from an imageRGB.
       * \param[in] image The imageRGB to copy.
       */
      image_CbCr(const imageRGB& image)
      {
         copyFromRGB(image);
      }

      /*!
       * \brief Copy function
       *
       * The size of the image is changed to fit the source image.
       *
       * \param[in] image The imageRGB to copy.
       */
      void copyFromRGB(const imageRGB& image)
      {
         setSize(image.nbColumns(), image.nbRows());

         const unsigned char* rgb = image.getActualPixel(0, 0);
         double* _cbcr = getActualPixel(0, 0);
         for (unsigned long ind = 0;
              ind < (static_cast<unsigned long>(nbColumns())
                     * static_cast<unsigned long>(nbRows()));
              ind ++)
            RGBTo_CbCr(&rgb[ind*3], &_cbcr[ind*2]);
      }

      /*!
       * \brief Copy the image into an imageRGB.
       *
       * The size of the target image is set here in accordance to the
       * source image.
       *
       * \param[out] image The imageRGB.
       */
      virtual void toRGB(imageRGB* image) const
      {
         image->setSize(nbColumns(), nbRows());

         unsigned char* rgb = image->getActualPixel(0, 0);
         const double* _cbcr = getActualPixel(0, 0);
         for (unsigned long ind = 0;
              ind < (static_cast<unsigned long>(nbColumns())
                     * static_cast<unsigned long>(nbRows()));
              ind ++)
            _CbCrToRGB(&_cbcr[ind*2], &rgb[ind*3]);
      }

   private:

      /*!
       * \brief Convert a pixel in the RGB colour space to a pixel in the
       * YCbCr colour space.
       * \param[in] rgb The pixel in RGB colour space.
       * \param[out] _cbcr The pixel in _CbCr colour space (must be
       * allocated beforehand).
       */
      void RGBTo_CbCr(const unsigned char* rgb, double* _cbcr)
      {
         double R = static_cast<double>(rgb[0]) / 255.0;
         double G = static_cast<double>(rgb[1]) / 255.0;
         double B = static_cast<double>(rgb[2]) / 255.0;
         //_cbcr[0] = 0.299 * R + 0.587 * G + 0.114 * B;
         _cbcr[0] = 0.5 - 0.169 * R - 0.331 * G + 0.500 * B;
         _cbcr[1] = 0.5 + 0.500 * R - 0.419 * G - 0.081 * B;
      }

      /*!
       * \brief save a pixel in the RGB ( just a copy ).
       * \param[in] _cbcr The pixel in _CbCr colour space.
       * \param[out] rgb The pixel in RGB colour space (must be allocated
       * beforehand).
       */
      void _CbCrToRGB(const double* _cbcr, unsigned char* rgb) const
      {
         double tmp = _cbcr[0] * 255.0;
         if (tmp > 255)
            tmp = 255;
         else
            tmp = tmp + 0.5;
         rgb[0] = static_cast<unsigned char>(tmp);
         tmp = _cbcr[1] * 255.0;
         if (tmp > 255)
            tmp = 255;
         else
            tmp = tmp + 0.5;
         rgb[1] = static_cast<unsigned char>(tmp);
         rgb[2] = 0;
      }
   };
   
} // namespace imalib

#endif // IMALIB_IMAGEYCBCR_H
