/*
 * imalib
 * 
 * Copyright 2017 Aberystwyth University
 * 
 * This library and program are free software: you can redistribute them
 * and/or modify them under the terms of the GNU General Public License as 
 * published by the Free Software Foundation, either version 3 of the License,
 * or (at your option) any later version.
 * 
 * The program and library are distributed in the hope that they will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*!
 * \file imageYUV.h
 * \brief Defines classes to play with YUV images.
 * \author Fred Labrosse <ffl@aber.ac.uk>
 * \copyright 2017 Aberystwyth University
 */

#ifndef IMALIB_IMAGEYUV_H
#define IMALIB_IMAGEYUV_H

#include <cassert>
#include <cstdlib>
#include <exception>
#include <fstream>
#include <iostream>
#include <map>
#include <sstream>
#include <string>

#ifdef WITH_PNG
#  include <png.h>
#endif // WITH_PNG

#include <imalib/imageRGB.h>
#include <imalib/imageD2.h>
#include <imalib/imageD3.h>

namespace imalib
{
   /*!
     \class imageYUV imageYUV.h imalib/imageYUV.h
     \brief A class to represent images in the YUV colour space.

     This is the YUV as used in PAL:
     \f[
     \begin{array}{l}
     \displaystyle
     Y = 0.299 R + 0.587 G + 0.114 B,\\
     U = 0.492 (B - Y),\\
     V = 0.877 (R - Y),
     \end{array}
     \f]
     \f[
     \begin{array}{l}
     \displaystyle
     R = Y + 1.140 V,\\
     G = Y - 0.395 U - 0.581 V,\\
     B = Y + 2.032 U,
     \end{array}
     \f]
     where \f$ 0 \leq R,G,B \leq 1 \f$, \f$ 0 \leq Y \leq 1 \f$, \f$ -0.492
     \leq U \leq 0.492 \f$ and \f$ -0.877 \leq V \leq 0.877 \f$.

     Conversion from and to RGB is provided.
   */
   class imageYUV : public imageD3
   {
   public:
      /*!
        \brief Constructor.
      */
      imageYUV()
         : imageD3()
      {
      }

      /*!
        \brief Constructor from an imageRGB.
        \param image (IN) The imageRGB to copy.
      */
      imageYUV(const imageRGB& image) 
      {
         copyFromRGB(image);
      }

      /*!
        \brief Copy function

        The size of the image is changed to fit the source image.

        \param image (IN) The imageRGB to copy.
      */
      void copyFromRGB(const imageRGB& image)
      {
         setSize(image.nbColumns(), image.nbRows());

         const unsigned char* rgb = image.getActualPixel(0, 0);
         double* yuv = getActualPixel(0, 0);
         for (unsigned long ind = 0;
              ind < nbColumns() * nbRows() * 3U;
              ind += 3)
            RGBToYUV(&rgb[ind], &yuv[ind]);        
      }

      /*!
        \brief Change the Y component (luminance) of all
        pixels to the given value.

        Is that useful?

        \param y (IN) The new luminance.
      */
      void changeY(double y)
      {
         double* pixel = getActualPixel(0, 0);
         for (unsigned long ind = 0;
              ind < nbColumns() * nbRows() * 3U;
              ind += 3)
            pixel[ind] = y;
      }

      /*!
        \brief Copy the image into an imageRGB.

        The size of the target image is set here in accordance to the
        source image.

        \param image (OUT) The imageRGB.
      */
      virtual void toRGB(imageRGB* image) const
      {
         image->setSize(nbColumns(), nbRows());

         unsigned char* rgb = image->getActualPixel(0, 0);
         const double* yuv = getActualPixel(0, 0);
         for (unsigned long ind = 0;
              ind < nbColumns() * nbRows() * 3U;
              ind += 3)
            YUVToRGB(&yuv[ind], &rgb[ind]);
      }

   private:
      /*!
       * \brief Convert a pixel in the RGB colour space to a pixel in the
       * YUV colour space.
       * \param[in] rgb The pixel in RGB colour space.
       * \param[out] yuv The pixel in YUV colour space (must be
       * allocated beforehand).
       */
      void RGBToYUV(const unsigned char* rgb, double* yuv)
      {
         yuv[0] = 0.299 * static_cast<double>(rgb[0]) / 255.0
            + 0.587 * static_cast<double>(rgb[1]) / 255.0
            + 0.114 * static_cast<double>(rgb[2]) / 255.0;
         yuv[1] = 0.492 * (static_cast<double>(rgb[2]) / 255.0 - yuv[0]);
         yuv[2] = 0.877 * (static_cast<double>(rgb[0]) / 255.0 - yuv[0]);
      }

      /*!
      \brief Convert a pixel in the YUV colour space to a pixel in the RGB
      colour space.
      \param[in] yuv The pixel in YUV colour space.
      \param[out] rgb The pixel in RGB colour space (must be allocated
      beforehand).
      */
      void YUVToRGB(const double* yuv, unsigned char* rgb) const
      {
         double tmp = (yuv[0] + 1.140 * yuv[2]) * 255.0;
         if (tmp > 255)
            tmp = 255;
         else
            tmp = tmp + 0.5;
         rgb[0] = static_cast<unsigned char>(tmp);
         tmp = (yuv[0] - 0.395 * yuv[1] - 0.581 * yuv[2]) * 255.0;
         if (tmp > 255)
            tmp = 255;
         else
            tmp = tmp + 0.5;
         rgb[1] = static_cast<unsigned char>(tmp);
         tmp = (yuv[0] + 2.032 * yuv[1]) * 255.0;
         if (tmp > 255)
            tmp = 255;
         else
            tmp = tmp + 0.5;
         rgb[2] = static_cast<unsigned char>(tmp);
      }

   };

      /*!
     \class image_UV imageYUV.h imalib/imageYUV.h
     \brief A class to represent images in the YUV colour space with only 
     the U and V components.

     Conversion from and to RGB is provided ( assuming a default V = 0.5).
   */
   class image_UV : public imageD2
   {
   public:
      /*!
        \brief Constructor.
      */
      image_UV()
         : imageD2(), defaultY(0.5)
      {
      }

      /*!
        \brief Constructor from an imageRGB.
        \param image (IN) The imageRGB to copy.
      */
      image_UV(const imageRGB& image)
         : defaultY(0.5)
      {
         copyFromRGB(image);
      }

      /*!
        \brief Copy function

        The size of the image is changed to fit the source image.

        \param image (IN) The imageRGB to copy.
      */
      void copyFromRGB(const imageRGB& image)
      {
         setSize(image.nbColumns(), image.nbRows());

         for (unsigned short col = 0; col < image.nbColumns(); ++col)
         {
            for (unsigned row = 0; row < image.nbRows(); ++row)
            {
               RGBTo_UV(image.getActualPixel(col, row),getActualPixel(col, row));
            }
         }
      }

      /*!
        \brief Change the Y component (luminance) of all
        pixels to the given value.

        Is that useful?

        \param y (IN) The new luminance.
      */
      void changeY(double y)
      {
         defaultY = y;
      }

      /*!
        \brief Copy the image into an imageRGB.

        The size of the target image is set here in accordance to the
        source image.

        \param image (OUT) The imageRGB.
        \todo This needs to be implemented.
      */
      virtual void toRGB(imageRGB* image) const
      {
         image->setSize(nbColumns(), nbRows());

         for (unsigned short col = 0; col < nbColumns(); ++col)
         {
            for (unsigned row = 0; row < nbRows(); ++row)
            {
               _UVToRGB(getActualPixel(col, row),image->getActualPixel(col, row));
            }
         }
      }

   private:
     
      /*!
       *\brief The default Y value .
       */   
      double defaultY;
      
      /*!
        \brief Convert a pixel in the RGB colour space to a pixel in the
        YUV colour space.
        \param[in] rgb The pixel in RGB colour space.
        \param[out] yuv The pixel in HSV colour space (must be
        allocated beforehand).
      */
      void RGBTo_UV(const unsigned char* rgb, double* yuv)
      {
         double y = 0.299 * static_cast<double>(rgb[0]) / 255.0
            + 0.587 * static_cast<double>(rgb[1]) / 255.0
            + 0.114 * static_cast<double>(rgb[2]) / 255.0;
         yuv[0] = 0.492 * (static_cast<double>(rgb[2]) / 255.0 - y);
         yuv[1] = 0.877 * (static_cast<double>(rgb[0]) / 255.0 - y);
      }

      /*!
      \brief Convert a pixel in the YUV colour space to a pixel in the RGB
      colour space.
      \param[in] yuv The pixel in YUV colour space.
      \param[out] rgb The pixel in RGB colour space (must be allocated
      beforehand).
      */
      void _UVToRGB(const double* yuv, unsigned char* rgb) const
      {
         double tmp = (defaultY + 1.140 * yuv[1]) * 255.0;
         if (tmp > 255)
            tmp = 255;
         else
            tmp = tmp + 0.5;
         rgb[0] = static_cast<unsigned char>(tmp);
         tmp = (defaultY - 0.395 * yuv[0] - 0.581 * yuv[1]) * 255.0;
         if (tmp > 255)
            tmp = 255;
         else
            tmp = tmp + 0.5;
         rgb[1] = static_cast<unsigned char>(tmp);
         tmp = (defaultY + 2.032 * yuv[0]) * 255.0;
         if (tmp > 255)
            tmp = 255;
         else
            tmp = tmp + 0.5;
         rgb[2] = static_cast<unsigned char>(tmp);
      }

   };
   
} // namespace imalib

#endif // IMALIB_IMAGEYUV_H
