/*
 * imalib
 * 
 * Copyright 2017 Aberystwyth University
 * 
 * This library and program are free software: you can redistribute them
 * and/or modify them under the terms of the GNU General Public License as 
 * published by the Free Software Foundation, either version 3 of the License,
 * or (at your option) any later version.
 * 
 * The program and library are distributed in the hope that they will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*!
 * \file picassoImageGrabber.h
 * \brief Image grabber classes for a picasso frame grabber.
 *
 * \author Fred Labrosse <ffl@aber.ac.uk>
 * \copyright 2017 Aberystwyth University
 */

#ifndef PICASSOIMAGEGRABBER_H
#define PICASSOIMAGEGRABBER_H

#include "imageGrabber.h"

#ifdef WITH_PICASSO
#  include <picasso.h>
#endif // WITH_PICASSO

#include <unistd.h>

namespace imalib
{

/*!
 * \class picassoImageGrabber picassoImageGrabber.h imalib/picassoImageGrabber.h
 * \brief A class to interface a picasso image grabber.
 *
 * If the class is available, WITH_PICASSO is defined (if you use
 * imalib-config --cflags when compiling, see \ref autotools).
 */
class picassoImageGrabber : public imageGrabber
{
public:
   /*!
    * \brief Constructor.
    *
    * \param[in] imgWidth
    * \param[in] imgHeight The width and height of the image.
    * \param[in] upsideDown Set to true if the camera is upside-down.
    * \param[in] flip Set to true to horizontally flip the images.
    * \param[in] grabIndex The index of the Picasso grabber to use.
    * \param[in] portNumber The number of the port on the device.
    * \param[in] posX
    * \param[in] posY The position of the viewport of the grabbed part.
    * \param[in] width
    * \param[in] height The width and height of the grabbed area.
    */
   picassoImageGrabber(unsigned short imgWidth, unsigned short imgHeight,
                       bool upsideDown,
                       bool flip,
                       int grabIndex, int portNumber,
                       unsigned short posX = 0,
                       unsigned short posY = 0,
                       unsigned short width = IMAGE_DEFAULT_INPUTWIDTH,
                       unsigned short height = IMAGE_DEFAULT_INPUTHEIGHT);

   /*!
    * \brief Constructor.
    * \param[in] fileName Name of the configuration file.
    */
   picassoImageGrabber(const std::string& fileName);

   /*!
    * \brief Destructor.
    */
   virtual ~picassoImageGrabber();

   /*!
    * \brief Display on the standard output some information about the
    * source of the images.
    *
    * \param[in,out] aStream The stream to which the info is added.
    */
   virtual void printInfoAboutSource(std::ostringstream* aStream);

   /*!
    * \brief Capture the image from a grabber.
    *
    * This is there because the capture can take a while.  We thus first
    * capture and then use (with getImage()).  This does not change the
    * image data.
    */
   virtual void captureImage();

   /*!
    * \brief Get an image from the frame grabber.
    *
    * The image might have been pre-captured (with captureImage()), but
    * if not will be captured here.
    *
    * The images are resized here.
    *
    * \param[out] theImage The image in which to save the grabbed image, if
    * not NULL.
    * \param[out] rawImage The image in which to save the raw image (before any
    * transformation is applied), if not NULL.
    */
   virtual void getImage(imageRGB* theImage, imageRGB* rawImage = NULL);

   /*!
    * \brief Specify the input rectangle (viewport on the grabbed image).
    * \param[in] x
    * \param[in] y Coordinates of the top-left corner of the rectangle.
    * \param[in] w
    * \param[in] h Size of the rectangle.
    * \return true is successful, false otherwise.
    * \sa useInputRectangle().
    */
   bool setInputRectangle(unsigned short x, unsigned short y,
                          unsigned short w, unsigned short h);

   /*!
    * \brief Use the input rectangle currently set.
    * \return true is successful, false otherwise.
    * \sa setInputRectangle().
    */
   bool useInputRectangle();

   /*!
    * \brief Set the properties of the picture.
    * \param[in] b The brightness.
    * \param[in] sa The saturation.
    * \param[in] cn The contrast.
    * \param[in] h The hue.
    * \return true is successful, false otherwise.
    */
   bool setPictureProperties(unsigned short b, unsigned short sa,
                             unsigned short cn, unsigned short h);

   /*!
    * \brief Use the picture properties currently set.
    * \return true is successful, false otherwise.
    * \sa setPictureProperties().
    */
   bool usePictureProperties();

   /*!
    * \brief Move the viewport to the right.
    * \sa setInputRectangle().
    */
   void moveRight();

   /*!
    * \brief Move the viewport to the left.
    * \sa setInputRectangle().
    */
   void moveLeft();

   /*!
    * \brief Move the viewport up.
    * \sa setInputRectangle().
    */
   void moveUp();

   /*!
    * \brief Move the viewport down.
    * \sa setInputRectangle().
    */
   void moveDown();

   /*!
    * \brief Increase the width of the viewport.
    * \sa setInputRectangle().
    */
   void increaseWidth();

   /*!
    * \brief Decrease the width of the viewport.
    * \sa setInputRectangle().
    */
   void decreaseWidth();

   /*!
    * \brief Increase the height of the viewport.
    * \sa setInputRectangle().
    */
   void increaseHeight();

   /*!
    * \brief Decrease the height of the viewport.
    * \sa setInputRectangle().
    */
   void decreaseHeight();

   /*!
    * \brief Set the input port.
    *
    * Don't do that too often as there is a delay (300ms) after doing so to let
    * the frame grabber stabilise.
    * \param[in] nb The new input port number.
    * \return true if all went well, false otherwise.
    */
   bool setInputPort(long nb)
   {
      long param = nb;
      PcssRetType ret = PcssFunc(grabberHandle, PCSSF_SET_INPUTCONNPORT, &param);
      if (ret != PCSS_NO_ERROR)
      {
         std::cerr << "Could not connect to input port " << nb << ".\n";
         return(false);
      }
      usleep(300000);

      return(true);
   }

protected:

   /*!
    * \brief Initialise the grabber.
    * \param[in] imgWidth
    * \param[in] imgHeight The width and height of the image.
    * \param[in] upsideDown Set to true if the camera is upside-down.
    * \param[in] flip Set to true to horizontally flip the images.
    * \param[in] grabIndex The index of the Picasso grabber to use.
    * \param[in] portNumber The number of the port on the device.
    * \param[in] posX
    * \param[in] posY The position of the viewport of the grabbed part.
    * \param[in] width
    * \param[in] height The width and height of the grabbed area.
    */
   void init(unsigned short imgWidth, unsigned short imgHeight,
             bool upsideDown,
             bool flip,
             int grabIndex, int portNumber,
             unsigned short posX = 0,
             unsigned short posY = 0,
             unsigned short width = IMAGE_DEFAULT_INPUTWIDTH,
             unsigned short height = IMAGE_DEFAULT_INPUTHEIGHT);

private:

   /*!
    * \brief A static variable that tells us how many instances have been
    * created.
    *
    * This is because some of the hardware is common to all ports and so the
    * input port must be reset before each grab.
    */
   static short numberOfInstances;

   /*!
    * \brief A static variable to make sure that only one board can be opened.
    */
   static int boardIndex;

   /*!
    * \brief Low level initialisation of the frame grabber.
    *
    * This is to be called after all changes in the acquisition parameters
    * and is called in useInputRectangle().
    * \param[in] vfld Fields selected by the grabber (see the picasso
    * documentation).  This should be EVENODD_FIELD if a full height is
    * captured (interlaced image) or EVENODD_FIELD_SINGLE_PITCH if half of
    * the height is captured (only even or odd fields).
    * \param[in] cfld Fields captured. This should be CAPT_ODDEVEN_FRAME if a
    * full height is captured (interlaced image) or CAPT_ALL_FIELDS if half of
    * the height is captured (only even or odd fields).
    * \return true if all went well, false otherwise.
    */
   bool initFrameGrabberCapture(PcssVFields vfld, PcssCaptFields cfld);

   /*!
    * \brief the handle to the frame grabber.
    */
   PCSS_HANDLE grabberHandle;

   /*
    * \brief The grabbed image.
    */
//    unsigned char* grabbedImage;

   /*!
    * \brief Input rectangle X position.
    */
   long inputX;

   /*!
    * \brief Input rectangle Y position.
    */
   long inputY;

   /*!
    * \brief Input rectangle width.
    */
   long inputWidth;

   /*!
    * \brief Input rectangle height.
    */
   long inputHeight;

}; // class picassoImageGrabber

} // namespace imalib

#endif // PICASSOIMAGEGRABBER_H
