/*
 * imalib
 * 
 * Copyright 2017 Aberystwyth University
 * 
 * This library and program are free software: you can redistribute them
 * and/or modify them under the terms of the GNU General Public License as 
 * published by the Free Software Foundation, either version 3 of the License,
 * or (at your option) any later version.
 * 
 * The program and library are distributed in the hope that they will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*!
 * \file pixelCoord.h
 * \brief Class to represent pixel coordinates.
 * \author Fred Labrosse <ffl@aber.ac.uk>
 * \copyright 2017 Aberystwyth University
 */

#ifndef PIXELCOORD_H
#define PIXELCOORD_H

namespace imalib
{

/*!
 * \class pixelCoord pixelCoord.h imalib/pixelCoord.h
 * \brief A class to represent pixel coordinates.
 *
 * The coordinates are doubles because we do some sub-pixel stuff.  There are
 * getters that convert to integral values.
 *
 * An mutable attribute records whether an instance is used in distance
 * calculations.  Any newly created instance, whichever the constructor used,
 * is marked as not having been used.
 */
class pixelCoord
{
public:

   /*!
    * \brief Default constructor.
    */
   pixelCoord()
   : X_(0), Y_(0), usedInDistance_(false)
   {
   }

   /*!
    * \brief Copy constructor.
    * \param[in] coord Coordinates to copy.
    */
   pixelCoord(const pixelCoord& coord)
   : X_(coord.X_),
   Y_(coord.Y_),
   usedInDistance_(coord.usedInDistance_)
   {
   }

   /*!
    * \brief Constructor from two values.
    * \param[in] X
    * \param[in] Y The X and Y coordinates.
    */
   pixelCoord(double X, double Y)
   : X_(X), Y_(Y), usedInDistance_(false)
   {
   }

   /*!
    * \brief Get the X coordinate.
    * \return The X coordinate.
    * \sa getXround().
    */
   double getX() const
   {
      return(X_);
   }

   /*!
    * \brief Get the Y coordinate.
    * \return The Y coordinate.
    * \sa getYround().
    */
   double getY() const
   {
      return(Y_);
   }

   /*!
    * \brief Get the X coordinate as an integral value.
    *
    * The value is rounded to the nearest integer value.
    *
    * \return The X coordinate.
    * \sa getX().
    */
   unsigned short getXround() const
   {
      return(static_cast<unsigned short>(X_ + 0.5));
   }

   /*!
    * \brief Get the Y coordinate as an integral value.
    *
    * The value is rounded to the nearest integer value.
    *
    * \return The Y coordinate.
    * \sa getY().
    */
   unsigned short getYround() const
   {
      return(static_cast<unsigned short>(Y_ + 0.5));
   }

   /*!
    * \brief Set the X coordinate.
    * \param[in] X The new value.
    */
   void setX(double X)
   {
      X_= X;
   }

   /*!
    * \brief Set the Y coordinate.
    * \param[in] Y The new value.
    */
   void setY(double Y)
   {
      Y_= Y;
   }

   /*!
    * \brief Set the coordinates
    * \param[in] X
    * \param[in] Y The new coordinates.
    */
   void set(double X, double Y)
   {
      X_ = X;
      Y_ = Y;
   }

   /*!
    * \brief Assignment operator.
    * \param coord The new value.
    * \return A reference to the calling object.
    */
   pixelCoord& operator=(const pixelCoord& coord)
   {
      X_ = coord.X_;
      Y_ = coord.Y_;
      usedInDistance_ = coord.usedInDistance_;
      return(*this);
   }

   /*!
    * \brief Less than operator.
    *
    * A position is less than another if both its coordinates are less than that
    * of the other.
    *
    * \param[in] r The right operand.
    * \return true if *this < r.
    */
   bool operator<(const pixelCoord& r) const
   {
      return((Y_ < r.Y_)
             || ((Y_ == r.Y_) && (X_ < r.X_)));
   }

   /*!
    * \brief Was this used in the last distance calculations.
    * \return true if this instance has been used.
    */
   bool wasUsedInDistance() const
   {
      return(usedInDistance_);
   }

   /*!
    * \brief Set the instance to have been used or not.
    * 
    * NOTE: this is const, so that instances can be used as keys in maps,
    * but the underlying attribute is mutable.
    *
    * \param[in] yes Set this to this value.
    */
   void setUsedInDistance(bool yes) const
   {
      usedInDistance_ = yes;
   }

private:

   /*!
    * \brief the X coordinate of the pixel.
    */
   double X_;

   /*!
    * \brief The Y coordinate of the pixel.
    */
   double Y_;

   /*!
    * \brief If true, this means that that position has been used in
    * calculating distances.
    */
   mutable bool usedInDistance_;
};

} // namespace imalib

#endif // PIXELCOORD_H
