/*
 * imalib
 * 
 * Copyright 2017 Aberystwyth University
 * 
 * This library and program are free software: you can redistribute them
 * and/or modify them under the terms of the GNU General Public License as 
 * published by the Free Software Foundation, either version 3 of the License,
 * or (at your option) any later version.
 * 
 * The program and library are distributed in the hope that they will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*!
 * \file pixelSelector.h
 * \brief A class to select a subset of all pixels in an image.
 * 
 * \author Fred Labrosse <ffl@aber.ac.uk>
 * \copyright 2017 Aberystwyth University
 */

#ifndef PIXELSELECTOR_H
#define PIXELSELECTOR_H

#include <assert.h>

#include <vector>
#include <map>

namespace imalib
{
/*!
 * \class pixelSelector pixelSelector.h imalib/pixelSelector.h
 * \brief A class to represent a subset of pixels of an image.
 *
 * This is just a vector of positions in an image that is meant to be
 * iterated over whenever some pixels of an image should be accessed.
 * 
 * A hash table is maintained for quick searches.
 */
class pixelSelector
{
public:

   /*!
    * \brief Constructor.
    */
   pixelSelector()
      : hashTableValid(false)
   {
   }

   /*!
    * \brief Get the X position of the top left corner, inclusive.
    * \return The X position of the top left corner.
    */
   unsigned short topLeftX() const
   {
      return(topLeftX_);
   }

   /*!
    * \brief Get the Y position of the top left corner, inclusive.
    * \return The Y position of the top left corner.
    */
   unsigned short topLeftY() const
   {
      return(topLeftY_);
   }

   /*!
    * \brief Get the X position of the bottom right corner, inclusive.
    * \return The X position of the bottom left corner.
    */
   unsigned short bottomRightX() const
   {
      return(bottomRightX_);
   }

   /*!
    * \brief Get the Y position of the bottom right corner, inclusive.
    * \return The Y position of the bottom left corner.
    */
   unsigned short bottomRightY() const
   {
      return(bottomRightY_);
   }

   /*!
   \class coords pixelSelector.h imalib/pixelSelector.h
   \brief A class to represent pixel coordinates.
   */
   class coords
   {
   public:
      /*!
       * \brief Constructor.
       */
      coords(unsigned short x, unsigned short y)
         : X(x), Y(y)
      {
      }

      /*!
       * \brief X coordinate of the source pixel.
       */
      unsigned short X;

      /*!
       * \brief Y coordinate of the source pixel.
       */
      unsigned short Y;
   };

   /*!
    * \brief Get the number of pixels in the selector.
    * \return The number of pixels.
    */
   std::size_t size()
   {
      return(pixels_.size());
   }

   /*!
    * \brief Get the vector of pixels that make up the selector.
    * \return The vector of pixels making up the selector.
    * \warning This returns a constant reference to the vector of pixels.
    *  Don't destroy it!
    */
   const std::vector<coords>& pixels() const
   {
      return(pixels_);
   }

   /*!
    * \brief Create a selector that keeps all the pixels.
    *
    * One usually does not want to do that (less efficient than just looking
    * at all pixels of the image) but this is useful for debugging.
    *
    * \param[in] width The width if the image.
    * \param[in] height The height of the image.
    */
   void createNoopPixelSelector(unsigned short width,
                                unsigned short height);

   /*!
    * \brief Create an area selector.
    * \param[in] pos1X
    * \param[in] pos1Y The position of the top left corner of the area,
    * inclusive.
    * \param[in] pos2X
    * \param[in] pos2Y The position of the bottom right corner of the area,
    * inclusive.
    */
   void createAreaSelector(unsigned short pos1X,
                           unsigned short pos1Y,
                           unsigned short pos2X,
                           unsigned short pos2Y);

   /*!
    * \brief Create a random selector.
    *
    * A random set of pixels is kept, based on the given probability, in a
    * rectangular area of specified size.
    *
    * \param[in] width The width of the rectangular area.
    * \param[in] height The height of the rectangular area.
    * \param[in] probability The probability of keeping pixels.
    */
   void createRandomPixelSelector(unsigned short width,
                                  unsigned short height,
                                  double probability);

   /*!
    * \brief Create a random selector.
    *
    * A random set of pixels is kept, based on the given probability, in a
    * rectangular area of specified size and position.
    *
    * \param[in] pos1X
    * \param[in] pos1Y The position of the top left corner of the area,
    * inclusive.
    * \param[in] pos2X
    * \param[in] pos2Y The position of the bottom right corner of the area,
    * inclusive.
    * \param[in] probability The probability of keeping pixels.
    */
   void createRandomPixelSelector(unsigned short pos1X,
                                  unsigned short pos1Y,
                                  unsigned short pos2X,
                                  unsigned short pos2Y,
                                  double probability);

   /*!
    * \brief Create a random selector from the front/back areas of a panoramic
    * image.
    *
    * This assumes that the front of the camera is centred at 3/4 of the width
    * of the panoramic image from the left while the back of the camera is
    * centred at 1/4 of the width of the image from the left.
    *
    * \param[in] width The width if the image.
    * \param[in] height The height of the image.
    * \param[in] probability The probability of keeping pixels.
    * \param[in] frontBack The width in pixels of the front and back areas.
    * \param[in] horizon Sets the maximum, exclusive, index of rows to consider,
    * 0 being top and height being bottom (the value is clamped to height).
    * \param[in] frontOnly If true, then only consider the front part of the
    * images.
    */
   void createRandomFrontBackHorizonPixelSelector(unsigned short width,
                                                  unsigned short height,
                                                  double probability,
                                                  unsigned short frontBack = 60,
                                                  unsigned short horizon = -1,
                                                  bool frontOnly = false);

   /*!
    * \brief Create the union of two pixelSelector's.
    * 
    * Any position already there is not added again (this remains a set).
    *
    * \param[in] selector The pixelSelector to add to the current.
    */
   const pixelSelector& operator+(const pixelSelector& selector);

   /*!
    * \brief The type of hash table.
    */
   typedef std::multimap<unsigned long, unsigned long> hashTableType;

   /*!
    * \brief Find the given position in the selector.
    *
    * The iterator is equal to endHash() if the position is not found.
    * 
    * The returned iterator (<tt>iter</tt>) can be used to find the index of
    * the searched position in the vector of pixels by doing 
    * <tt>iter->second</tt>.
    *
    * \param[in] x
    * \param[in] y The coordinates of the searched pixel.
    * \return The iterator found (or not) in the hash table.
    */
   hashTableType::const_iterator find(unsigned short x, unsigned short y) const
   {
      assert(hashTableValid);
      return(hash.find(y * (bottomRightX_ - topLeftX_) + x));
   }

   /*!
    * \brief Get the end of the hash table.
    */
   hashTableType::const_iterator endHash() const
   {
      return(hash.end());
   }

protected:

private:

   /*!
    * \brief The X position of the top left corner, inclusive.
    */
   unsigned short topLeftX_;

   /*!
    * \brief The Y position of the top left corner, inclusive.
    */
   unsigned short topLeftY_;

   /*!
    * \brief The X position of the bottom right corner, inclusive.
    */
   unsigned short bottomRightX_;

   /*!
    * \brief The Y position of the bottom right corner, inclusive.
    */
   unsigned short bottomRightY_;

   /*!
    * \brief The set of pixels that make up the subset.
    */
   std::vector<coords> pixels_;

   /*!
    * \brief Update the hash table used for efficient searching.
    *
    * This function needs to be called at the end of each selector
    * modification.
    */
   void createHashTable();

   /*!
    * \brief Add a position to the hash table.
    * \param[in] coord The position to add.
    * \param[in] ind The index of the position in the vector of positions.
    */
   void addToHashTable(coords coord, unsigned int ind)
   {
      hash.insert(std::pair<unsigned long, unsigned long>(
         coord.Y * (bottomRightX_ - topLeftX_) + coord.X, ind));
   }

   /*!
    * \brief Hash table for the coordinates of the selector.
    */
   hashTableType hash;

   /*!
    * \brief Tells if the hash table is valid or not.
    *
    * If the hash table is not valid, then it will be updated when
    * find() is called.
    */
   bool hashTableValid;
};

} // namespace imalib

#endif // PIXELSELECTOR_H
