/*
 * imalib
 * 
 * Copyright 2017 Aberystwyth University
 * 
 * This library and program are free software: you can redistribute them
 * and/or modify them under the terms of the GNU General Public License as 
 * published by the Free Software Foundation, either version 3 of the License,
 * or (at your option) any later version.
 * 
 * The program and library are distributed in the hope that they will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*!
 * \file pixelValue.h
 * \brief Class to represent the value of a pixel.
 * \author Fred Labrosse <ffl@aber.ac.uk>
 * \copyright 2017 Aberystwyth University
 */

#ifndef PIXELVALUE_H
#define PIXELVALUE_H

#include <cstring>

namespace imalib
{

/*!
 * \class pixelValue pixelValue.h imalib/pixelValue.h
 * \brief A class to represent the value of a pixel.
 *
 * The value of a pixel is made of nbColComp components of type typeCol.
 *
 * The data is statically allocated at compilation time.
 */
template  <typename typeCol, int nbColComp>
class pixelValue
{
public:

   /*!
    * \brief Default constructor.
    */
   pixelValue()
   {
      for (unsigned short ind = 0; ind < nbColComp; ind++)
      {
         data_[ind] = typeCol();
      }
   }

   /*!
    * \brief Copy constructor.
    * \param[in] newVal The new value to copy.
    */
   pixelValue(const pixelValue& newVal)
   {
      memcpy(data_, newVal.data_, sizeof(typeCol) * nbColComp);
   }

   /*!
    * \brief Constructor from an array of values.
    *
    * It is assumed that the array contains enough values to copy from.
    * \param[in] newVal The array of new values.
    */
   pixelValue(const typeCol* newVal)
   {
      set(newVal);
   }

   /*!
    * \brief Set the value from an array of values.
    *
    * It is assumed that the array contains enough values to copy from.
    * \param[in] newVal The array of new values.
    */
   void set(const typeCol* newVal)
   {
      memcpy(data_, newVal, sizeof(typeCol) * nbColComp);
   }

   /*!
    * \brief Assignment operator.
    * \param[in] newVal The new value to use.
    */
   void operator=(const pixelValue& newVal)
   {
      memcpy(data_, newVal.data_, sizeof(typeCol) * nbColComp);
   }

   /*!
    * \brief Get the value of the pixel.
    * \return The value.
    */
   const typeCol* get() const
   {
      return(data_);
   }

private:

   /*!
    * \brief The pixel data.
    */
   typeCol data_[nbColComp];
};

} // namespace imalib

#endif // PIXELCOORD_H
