/*
 * imalib
 * 
 * Copyright 2017 Aberystwyth University
 * 
 * This library and program are free software: you can redistribute them
 * and/or modify them under the terms of the GNU General Public License as 
 * published by the Free Software Foundation, either version 3 of the License,
 * or (at your option) any later version.
 * 
 * The program and library are distributed in the hope that they will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*!
 * \file shape.h
 * \brief Classes to represent objects having some sort of geometric shape.
 *
 * \author Fred Labrosse
 * \copyright 2017 Aberystwyth University
 */

#ifndef SHAPE_H
#define SHAPE_H

#include <math.h>
#include <string>
#include <vector>

namespace imalib
{
/*!
 * \class point shape.h imalib/shape.h
 * \brief A class to represent coordinates, for example of an image
 * point.
 *
 * Yes, I know, it's all public...
 */
class point
{
public:

   /*!
    * \brief Default constructor.
    */
   point()
    : x(0.0), y(0.0)
   {
   }

   /*!
    * \brief Constructor.
    */
   point(double X, double Y)
      : x(X), y(Y)
   {
   }

    /*!
     * \brief X coordinate.
     */
   double x;

   /*!
    * \brief Y coordinate.
    */
   double y;

   /*!
    * \brief Distance between two points.
    */
   double distance(const point& point2)
   {
      return(sqrt((point2.x - x) * (point2.x - x)
                  + (point2.y - y) * (point2.y - y)));
   }
};

/*!
 * \class segment shape.h imalib/shape.h
 * \brief A class to represent a segment.
 *
 * It is four values: coordinates of one of the points and incremental
 * displacement to reach the other point.
 */
class segment
{
public:
   /*!
    * \brief Constructor.
    */
   segment(long X, long Y, long dX, long dY)
      : x(X), y(Y), dx(dX), dy(dY)
   {
   }

   /*!
    * \brief X coordinate of the first point.
    */
   long x;

   /*!
    * \brief Y coordinate of the first point.
    */
   long y;

   /*!
    * \brief X displacement to reach the second point.
    */
   long dx;

   /*!
    * \brief Y displacement to reach the second point.
    */
   long dy;
};

/*!
 * \class curve shape.h imalib/shape.h
 * \brief A class to represent a curve.
 *
 * It is just a vector of imalib::point's.
 */
class curve
{
public:
   /*!
    * \brief Constructor.
    */
   curve()
   {
   }

   /*!
    * \brief Re-sample a curve.
    *
    * The new curve can be uniformly sampled or not.
    *
    * It is based on code from [\ref Hec94].
    *
    * \param[out] newCurve The re-sampled curve.
    * \param[in] n The number of points on the new curve (must be 2 or more,
    * only verified if compiled without NDEBUG).
    * \param[in] nonUniform If true, then re-sample non-uniformly.
    */
   void resample(curve* newCurve,
                 unsigned short n,
                 bool nonUniform = false);

   /*!
    * \brief outputs the curve in a string, one point per line.
    * \return The string.
    */
   std::string toString();

   /*!
    * \brief The coordinates of the points making up the curve.
    */
   std::vector<point> points;
};

}

#endif // SHAPE_H
