/*
 * imalib
 * 
 * Copyright 2017 Aberystwyth University
 * 
 * This library and program are free software: you can redistribute them
 * and/or modify them under the terms of the GNU General Public License as 
 * published by the Free Software Foundation, either version 3 of the License,
 * or (at your option) any later version.
 * 
 * The program and library are distributed in the hope that they will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*!
 * \file transformation.h
 * \brief Defines classes to geometrically transform appearances.
 * \author Fred Labrosse <ffl@aber.ac.uk>
 * \copyright 2017 Aberystwyth University
 */

#ifndef IMALIB_TRANSFORMATION_H
#define IMALIB_TRANSFORMATION_H

#include <math.h>
#include <string.h>

#include <iostream>
#include <iomanip>
#include <limits>

#include "imalib/appearancePixel.h"
#include "imalib/pixelCoord.h"

namespace imalib
{

/*!
 * \class transformation transformation.h imalib/transformation.h
 * \brief Base class of all transformation classes.
 *
 * It is an abstract class and only defines the interface as well as common
 * data needed for transformations.
 *
 * The transformation is expressed as a 3x3 generalised affine 2D
 * transformation matrix.  This matrix can be set in a number of ways but must
 * be settable from the parameters given to setParams() and gettable using
 * getParams().
 *
 * By default, the transformation is the identity (noop).
 *
 * Limits on the parameters can be set (see setParamsLimits()).  These limits
 * may or may not be used by setLimits() but are used by the optimisation
 * functions in the classes inheriting from distance_base.
 *
 * The matrix is as follows:
 * \f[
 * \left(\begin{array}{ccc}
 * t11 & t21 & t31 \\
 * t12 & t22 & t32 \\
 * t13 & t23 & t33
 * \end{array}\right)
 * \f]
 */
class transformation
{
public:

   /*!
    * \brief Constructor.
    *
    * Sets the transformation to the identity.
    *
    * \param[in] dof The degrees of freedom.
    */
   transformation(unsigned short dof)
      : t11_ (1.0), t21_ (0.0), t31_ (0.0),
          t12_ (0.0), t22_ (1.0), t32_ (0.0),
          t13_ (0.0), t23_(0.0), t33_(1.0),
        dof_(dof),
        paramsMin_(NULL), paramsMax_(NULL), paramsStep_(NULL)
   {
      allocateMemory();
   }

   /*!
    * \brief Copy constructor.
    * \param[in] transf The transformation to copy.
    */
   transformation(const transformation& transf)
      : t11_ (transf.t11_ ), t21_ (transf.t21_ ), t31_ (transf.t31_ ),
          t12_ (transf.t12_ ), t22_ (transf.t22_ ), t32_ (transf.t32_ ),
          t13_ (transf.t13_ ), t23_(transf.t23_), t33_(transf.t33_),
      dof_(transf.dof_),
      paramsMin_(NULL), paramsMax_(NULL), paramsStep_(NULL)
   {
      allocateMemory();
      memcpy(paramsMin_, transf.paramsMin_, dof_ * sizeof(double));
      memcpy(paramsMax_, transf.paramsMax_, dof_ * sizeof(double));
      memcpy(paramsStep_, transf.paramsStep_, dof_ * sizeof(double));
   }

   /*!
    * \brief Destructor.
    */
   virtual ~transformation()
   {
      freeMemory();
   }

   /*!
    * \brief Assignment operator.
    * \param[in] transf The transformation to copy.
    */
   transformation& operator=(const transformation& transf)
   {
      t11_ = transf.t11_;
      t21_ = transf.t21_;
      t31_ = transf.t31_;
      t12_ = transf.t12_;
      t22_ = transf.t22_;
      t32_ = transf.t32_;
      t13_ = transf.t13_;
      t23_ = transf.t23_;
      t33_ = transf.t33_;
      setDoF(transf.dof_);
      memcpy(paramsMin_, transf.paramsMin_, dof_ * sizeof(double));
      memcpy(paramsMax_, transf.paramsMax_, dof_ * sizeof(double));
      memcpy(paramsStep_, transf.paramsStep_, dof_ * sizeof(double));

      return(*this);
   }

   /*!
    * \brief Set the degrees of freedom of the transformation.
    *
    * This is the number of parameters given to setParams() or received from
    * getParams().
    *
    * \param[in] dof The new degrees of freedom.
    */
   virtual void setDoF(unsigned short dof)
   {
      if (dof != dof_)
      {
         freeMemory();
         dof_ = dof;
         allocateMemory();
      }
   }

   /*!
    * \brief Get the degrees of freedom of the transformation.
    *
    * This is the number of parameters given to setParams() or received from
    * getParams().
    */
   unsigned short getDoF()
   {
      return(dof_);
   }

   /*!
    * \brief Set the parameters of the transformation.
    *
    * This function is virtual pure and must be implemented by sub-classes.
    * The format of the parameters is up to the sub-class.
    *
    * If the transformation is to be optimised in a distance function (see
    * distance_base), it is a good idea to have them such that their ranges
    * are similar.
    *
    * \param[in] params The new parameters of the transformation.
    * \sa setDoF().
    */
   virtual void setParams(const double* params) = 0;

   /*!
    * \brief Get the parameters of the transformation.
    *
    * This function is virtual pure and must be implemented by sub-classes.
    * The format of the parameters is up to the sub-class.
    *
    * \param[out] params The parameters of the transformation.  This must be
    * allocated elsewhere.
    * \sa getDoF().
    */
   virtual void getParams(double* params) const = 0;

   /*!
    * \brief Set the limits of the parameters of the transformation.
    *
    * These are <b>not</b> absolute limits on the values and it is up to
    * the setParams() function to check the validity of given parameters.  The
    * limits are however used in optimisation functions in the distance_base
    * classes by forcing the optimisation inside the limits set here.
    *
    * \param[in] paramsMin The new minimum values of the parameters.
    * \param[in] paramsMax The new maximum values of the parameters.
    * \sa setParams().
    */
   virtual void setParamsLimits(const double* paramsMin,
                                const double* paramsMax) = 0;

   /*!
    * \brief Get the limits of the parameters of the transformation.
    *
    * \param[out] paramsMin The minimum values of the parameters.  This must be
    * allocated elsewhere.
    * \param[out] paramsMax The maximum values of the parameters.  This must be
    * allocated elsewhere.
    * \sa setParamsLimits().
    */
   virtual void getParamsLimits(double* paramsMin,
                                double* paramsMax) const = 0;

   /*!
    * \brief Set the step for the parameters of the transformation.
    *
    * This should be values such that a transformation with parameters
    * "values" and one with parameters "values + step" should be significantly
    * different but not enough to be in different valley of a distance_base
    * optimisation.  The steps are used in optimisations of the transformation
    * in the distance_base classes.
    *
    * \param[in] steps The new step for the parameters of the transformation.
    * \sa getParamsSteps().
    */
   virtual void setParamsSteps(const double* steps) = 0;

   /*!
    * \brief Get the steps for the parameters of the transformation.
    *
    * This should be values such that a transformation with parameters
    * "values" and one with parameters "values + step" should be significantly
    * different but not enough to be in different valley of a distance_base
    * optimisation.  The steps are used in optimisations of the transformation
    * in the distance_base classes.
    *
    * \param[out] steps The step for the parameters of the transformation.
    * This must be allocated elsewhere.
    * \sa setParamsSteps().
    */
   virtual void getParamsSteps(double* steps) const = 0;

   /*
    * \brief Apply the transformation to an appearancePixel.
    * \param[in,out] appPix The appearancePixel to transform.
    */
//    template <typename typeCol, int nbColComp>
//    void apply(appearancePixel<typeCol, nbColComp>* appPix) const
//    {
//       appPix->setX(t11_ * appPix->getX() + t21_ * appPix->getY() + t31_);
//       appPix->setY(t12_ * appPix->getX() + t22_ * appPix->getY() + t32_);
//    }

   /*!
    * \brief Apply the transformation to a pixelCoord.
    * \param[in] coordIn The position to transform.
    * \param[out] coordOut The transformed position.
    */
   void apply(const pixelCoord& coordIn, pixelCoord* coordOut) const
   {
      coordOut->set(t11_ * coordIn.getX() + t21_ * coordIn.getY() + t31_,
                    t12_ * coordIn.getX() + t22_ * coordIn.getY() + t32_);
   }

   /*!
    * \brief Apply the transformation to a 2D position.
    * \param[in] xIn
    * \param[in] yIn The position of the point to transform.
    * \param[out] xOut
    * \param[out] yOut The position of the transformed point.
    */
   void apply(double xIn, double yIn, double* xOut, double* yOut) const
   {
      *xOut = t11_ * xIn + t21_ * yIn + t31_;
      *yOut = t12_ * xIn + t22_ * yIn + t32_;
   }

   /*!
    * \brief Apply the transformation to a 2D position.
    * \param[in] xIn
    * \param[in] yIn The position of the point to transform.
    * \param[out] xOut
    * \param[out] yOut The position of the transformed point.
    */
   void apply(unsigned short xIn, unsigned short yIn,
              unsigned short* xOut, unsigned short* yOut) const
   {
      *xOut = static_cast<unsigned short>(round( t11_ * static_cast<double>(xIn)
                                               + t21_ * static_cast<double>(yIn)
                                               + static_cast<double>( t31_ )));
      *yOut = static_cast<unsigned short>(round( t12_ * static_cast<double>(xIn)
                                               + t22_ * static_cast<double>(yIn)
                                               + static_cast<double>( t32_ )));
   }

   /*!
    * \brief Insertion operator.
    */
   friend std::ostream& operator<<(std::ostream& out,
                                   const transformation& transf)
   {
      out << std::setw(8) << transf.t11_ << " "
          << std::setw(8) << transf.t21_ << " "
          << std::setw(8) << transf.t31_ << std::endl
          << std::setw(8) << transf.t12_ << " "
          << std::setw(8) << transf.t22_ << " "
          << std::setw(8) << transf.t32_ << std::endl
          << std::setw(8) << transf.t13_ << " "
          << std::setw(8) << transf.t23_ << " "
          << std::setw(8) << transf.t33_ << std::endl
          << "     Min      Max      Step\n";
      for (unsigned short ind = 0; ind < transf.dof_; ind++)
      {
         out << std::setw(8) << transf.paramsMin_[ind]
         << " " << std::setw(8) << transf.paramsMax_[ind]
         << " " << std::setw(8) << transf.paramsStep_[ind] << std::endl;
      }
      return(out);
   }

protected:

   /*!
    * \brief Parameter t11 of the matrix.
    */
   double t11_;

   /*!
    * \brief Parameter t21 of the matrix.
    */
   double t21_;

   /*!
    * \brief Parameter t31 of the matrix.
    */
   double t31_;

   /*!
    * \brief Parameter t12 of the matrix.
    */
   double t12_;

   /*!
    * \brief Parameter t22 of the matrix.
    */
   double t22_;

   /*!
    * \brief Parameter t32 of the matrix.
    */
   double t32_;

   /*!
    * \brief Parameter t13 of the matrix.
    */
   double t13_;

   /*!
    * \brief Parameter t23 of the matrix.
    */
   double t23_;

   /*!
    * \brief Parameter t33 of the matrix.
    */
   double t33_;

   /*!
    * \brief Degrees of freedom of the transformation.
    *
    * This is how many parameters are necessary to specify the transformation.
    * \sa getDoF(), setDoF(), getParams(), setParams().
    */
   unsigned short dof_;

   /*!
    * \brief The minimum values "allowed" for the parameters when searching
    * for an optimum transformation in the distance_base classes.
    */
   double* paramsMin_;

   /*!
    * \brief The maximum values "allowed" for the parameters when searching
    * for an optimum transformation in the distance_base classes.
    */
   double* paramsMax_;

   /*!
    * \brief The steps for the parameters of the transformation.
    */
   double* paramsStep_;

private:

   /*!
    * \brief Allocate memory.
    */
   void allocateMemory()
   {
      if (paramsMin_ != NULL)
         freeMemory();

      paramsMin_ = new double[dof_];
      paramsMax_ = new double[dof_];
      paramsStep_ = new double[dof_];
      for (unsigned short ind = 0; ind < dof_; ++ind)
      {
         paramsMin_[ind] = std::numeric_limits<double>::min();
         paramsMax_[ind] = std::numeric_limits<double>::max();
         paramsStep_[ind] = 2.0; // Arbitrary values.
      }
   }

   /*!
    * \brief Free memory.
    */
   void freeMemory()
   {
      delete [] paramsStep_;
      paramsStep_ = NULL;
      delete [] paramsMax_;
      paramsMax_ = NULL;
      delete [] paramsMin_;
      paramsMin_ = NULL;
   }

};

/*!
 * \class translation transformation.h imalib/transformation.h
 * \brief A class to represent a 2D (in the x,y plane) translation.
 */
class translation : public transformation
{
public:
   /*!
    * \brief Default constructor.
    *
    * Sets the transformation to the identity.
    */
   translation()
      : transformation(2)
   {
   }

   /*!
    * \brief Destructor.
    */
   virtual ~translation()
   {
   }

   /*!
    * \brief Set the parameters of the translation.
    *
    * \param[in] params The new parameters of the translation:
    * - params[0] along the x direction,
    * - params[1] along the y direction.
    * .
    * \sa setDoF().
    */
   virtual void setParams(const double* params)
   {
        t31_ = params[0];
        t32_ = params[1];
   }

   /*!
    * \brief Get the parameters of the translation.
    *
    * \param[out] params The parameters of the translation:
    * - params[0] along the x direction,
    * - params[1] along the y direction.
    * .
    * This must be allocated elsewhere.
    * \sa getDoF().
    */
   virtual void getParams(double* params) const
   {
      params[0] = t31_;
      params[1] = t32_;
   }

   /*!
    * \brief Set the limits of the parameters of the transformation.
    *
    * These are <b>not</b> absolute limits on the values and it is up to
    * the setParams() function to check the validity of given parameters.  The
    * limits are however used in optimisation functions in the distance_base
    * classes by forcing the optimisation inside the limits set here.
    *
    * \param[in] paramsMin The new minimum values of the parameters.
    * \param[in] paramsMax The new maximum values of the parameters.
    * \sa setParams().
    */
   virtual void setParamsLimits(const double* paramsMin,
                                const double* paramsMax)
   {
      memcpy(paramsMin_, paramsMin, dof_ * sizeof(double));
      memcpy(paramsMax_, paramsMax, dof_ * sizeof(double));
   }

   /*!
    * \brief Get the limits of the parameters of the transformation.
    *
    * \param[out] paramsMin The minimum values of the parameters.  This must be
    * allocated elsewhere.
    * \param[out] paramsMax The maximum values of the parameters.  This must be
    * allocated elsewhere.
    * \sa setParamsLimits().
    */
   virtual void getParamsLimits(double* paramsMin,
                                double* paramsMax) const
   {
      memcpy(paramsMin, paramsMin_, dof_ * sizeof(double));
      memcpy(paramsMax, paramsMax_, dof_ * sizeof(double));
   }

   /*!
    * \brief Set the step for the parameters of the transformation.
    *
    * This should be values such that a transformation with parameters
    * "values" and one with parameters "values + step" should be significantly
    * different but not enough to be in different valley of a distance_base
    * optimisation.  The steps are used in optimisations of the transformation
    * in the distance_base classes.
    *
    * \param[in] steps The new step for the parameters of the transformation.
    * \sa getParamsSteps().
    */
   virtual void setParamsSteps(const double* steps)
   {
      memcpy(paramsStep_, steps, dof_ * sizeof(double));
   }

   /*!
    * \brief Get the steps for the parameters of the transformation.
    *
    * This should be values such that a transformation with parameters
    * "values" and one with parameters "values + step" should be significantly
    * different but not enough to be in different valley of a distance_base
    * optimisation.  The steps are used in optimisations of the transformation
    * in the distance_base classes.
    *
    * \param[out] steps The step for the parameters of the transformation.
    * This must be allocated elsewhere.
    * \sa setParamsSteps().
    */
   virtual void getParamsSteps(double* steps) const
   {
      memcpy(steps, paramsStep_, dof_ * sizeof(double));
   }

protected:
private:

};

/*!
 * \class rotation transformation.h imalib/transformation.h
 * \brief A class to represent a rotation around the axis perpendicular to
 * the image plane.
 */
class rotation : public transformation
{
public:
   /*!
    * \brief Default constructor.
    *
    * Sets the transformation to the identity.
    */
   rotation()
      : transformation(1)
   {
   }

   /*!
    * \brief Destructor.
    */
   virtual ~rotation()
   {
   }

   /*!
    * \brief Set the parameter of the rotation.
    *
    * \param[in] params The new parameter of the rotation:
    * - params[0] angle in radians.
    * .
    * Because the y axis is positive downwards, a positive rotation is
    * clock-wise.
    * \sa setDoF().
    */
   virtual void setParams(const double* params)
   {
        t11_ = t22_ = cos(params[0]);
        t12_ = sin(params[0]);
        t21_ = -t12_;
   }

   /*!
    * \brief Get the parameter of the rotation.
    *
    * \param[out] params The parameter of the rotation:
    * - params[0] angle in radians.
    * .
    * This must be allocated elsewhere.  Because the y axis is positive
    * downwards, a positive rotation is clock-wise.
    * \sa getDoF().
    */
   virtual void getParams(double* params) const
   {
      params[0] = acos( t11_ );
   }

   /*!
    * \brief Set the limits of the parameters of the transformation.
    *
    * These are <b>not</b> absolute limits on the values and it is up to
    * the setParams() function to check the validity of given parameters.  The
    * limits are however used in optimisation functions in the distance_base
    * classes by forcing the optimisation inside the limits set here.
    *
    * \param[in] paramsMin The new minimum values of the parameters.
    * \param[in] paramsMax The new maximum values of the parameters.
    * \sa setParams().
    */
   virtual void setParamsLimits(const double* paramsMin,
                                const double* paramsMax)
   {
      memcpy(paramsMin_, paramsMin, dof_ * sizeof(double));
      memcpy(paramsMax_, paramsMax, dof_ * sizeof(double));
   }

   /*!
    * \brief Get the limits of the parameters of the transformation.
    *
    * \param[out] paramsMin The minimum values of the parameters.  This must be
    * allocated elsewhere.
    * \param[out] paramsMax The maximum values of the parameters.  This must be
    * allocated elsewhere.
    * \sa setParamsLimits().
    */
   virtual void getParamsLimits(double* paramsMin,
                                double* paramsMax) const
   {
      memcpy(paramsMin, paramsMin_, dof_ * sizeof(double));
      memcpy(paramsMax, paramsMax_, dof_ * sizeof(double));
   }

   /*!
    * \brief Set the step for the parameters of the transformation.
    *
    * This should be values such that a transformation with parameters
    * "values" and one with parameters "values + step" should be significantly
    * different but not enough to be in different valley of a distance_base
    * optimisation.  The steps are used in optimisations of the transformation
    * in the distance_base classes.
    *
    * \param[in] steps The new step for the parameters of the transformation.
    * \sa getParamsSteps().
    */
   virtual void setParamsSteps(const double* steps)
   {
      memcpy(paramsStep_, steps, dof_ * sizeof(double));
   }

   /*!
    * \brief Get the steps for the parameters of the transformation.
    *
    * This should be values such that a transformation with parameters
    * "values" and one with parameters "values + step" should be significantly
    * different but not enough to be in different valley of a distance_base
    * optimisation.  The steps are used in optimisations of the transformation
    * in the distance_base classes.
    *
    * \param[out] steps The step for the parameters of the transformation.
    * This must be allocated elsewhere.
    * \sa setParamsSteps().
    */
   virtual void getParamsSteps(double* steps) const
   {
      memcpy(steps, paramsStep_, dof_ * sizeof(double));
   }

protected:
private:

};

}

#endif // IMALIB_TRANSFORMATION_H
