/*
 * imalib
 * 
 * Copyright 2017 Aberystwyth University
 * 
 * This library and program are free software: you can redistribute them
 * and/or modify them under the terms of the GNU General Public License as 
 * published by the Free Software Foundation, either version 3 of the License,
 * or (at your option) any later version.
 * 
 * The program and library are distributed in the hope that they will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*!
 * \file ueyeImageGrabber.h
 * \brief An image grabber for a uEye USB camera.
 *
 * \author Fred Labrosse <ffl@aber.ac.uk>, with contributions from Suzana
 * Barreto <szb@aber.ac.uk> and Mark Neal <mjn@aber.ac.uk>
 * \copyright 2017 Aberystwyth University
 */

#ifndef UEYEIMAGEGRABBER_H
#define UEYEIMAGEGRABBER_H

#define UEYE_IMAGE_DEFAULT_INPUTWIDTH 1600
#define UEYE_IMAGE_DEFAULT_INPUTHEIGHT 1200

#include "imageGrabber.h"

#include <ueye.h>

namespace imalib
{

/*!
 * \class ueyeImageGrabber ueyeImageGrabber.h imalib/ueyeImageGrabber.h
 * \brief A class to interface a uEye USB camera.
 *
 * The camera is run in Freerun mode (i.e. continuously captures images)
 * and snap mode (i.e. images are only downloaded when requested).
 *
 * If the class is available, WITH_UEYE is defined (if you use
 * imalib-config --cflags when compiling, see \ref autotools).
 */
class ueyeImageGrabber : public imageGrabber
{
public:

   /*!
    * \brief Constructor.
    *
    * \param[in] upsideDown Set to true if the camera is upside-down.
    * \param[in] flip Set to true to horizontally flip the images.
    * \param[in] camNumber The number of the camera (0 is the first available).
    * \param[in] posX
    * \param[in] posY The position of the viewport of the grabbed part.
    * \param[in] width
    * \param[in] height The width and height of the grabbed area and of the
    * images.
    */
   ueyeImageGrabber(bool upsideDown,
                   bool flip,
                   int camNumber,
                   unsigned short posX = 0,
                   unsigned short posY = 0,
                   unsigned short width = UEYE_IMAGE_DEFAULT_INPUTWIDTH,
                   unsigned short height = UEYE_IMAGE_DEFAULT_INPUTHEIGHT);

   /*!
    * \brief Constructor.
    * \param[in] fileName Name of the configuration file.
    */
   ueyeImageGrabber(const std::string& fileName);

   /*!
    * \brief Destructor.
    */
   virtual ~ueyeImageGrabber();

   /*!
    * \brief Tell if we are connected to the required frame grabber.
    * \return true if we are connected, false otherwise.
    */
   virtual bool isConnected()
   {
      return (connected);
   }

   /*!
    * \brief Display on the standard output some information about the
    * source of the images.
    *
    * \param[in,out] aStream The stream to which the info is added.
    */
   virtual void printInfoAboutSource(std::ostringstream* aStream);

   /*!
    * \brief Capture the image from a grabber.
    *
    * This is there because the capture can take a while.  We thus first
    * capture and then use (with getImage()).
    */
   virtual void captureImage();

   /*!
    * \brief Get an image from the frame grabber.
    *
    * The image might have been pre-captured (with captureImage()), but
    * if not will be captured here.
    *
    * The images are resized here.
    *
    * \param[out] theImage The image in which to save the grabbed image, if
    * not NULL.
    * \param[out] rawImage The image in which to save the raw image (before
    * any transformation is applied), if not NULL.
    */
   virtual void getImage(imageRGB* theImage, imageRGB* rawImage = NULL);

   /*!
    * \brief Specify the input rectangle (viewport on the grabbed image).
    * \param[in] x
    * \param[in] y Coordinates of the top-left corner of the rectangle.
    * \param[in] w
    * \param[in] h Size of the rectangle.
    * \return true is successful, false otherwise.
    * \sa useInputRectangle().
    */
   bool setInputRectangle(int x, int y, int w, int h);

   /*!
    * \brief Use the input rectangle currently set.
    * \return true is successful, false otherwise.
    * \sa setInputRectangle().
    */
   bool useInputRectangle();
   /*!
    * \brief set the camera mode, default is freerun continuous capture
    * \param[in] mode the mode - always continuous capture, default = 0,
    * \ software trigger = 1, hardware trigger = 2
    * \return true if mode set, false otherwise.
    */
   bool setMode(int mode);

   /*!
    * \brief Get a list of discrete pixel clocks for the camera.
    * \param[out] pixelClocks The array of pixel cloacks.  This must have been
    * allocated elsewhere.
    * \sa getNumberOfPixelClocks().
    */
   void getPixelClockList(UINT* pixelClocks);

   /*!
    * \brief Get the number of supported pixel clocks.
    * \return The number of pixel clocks.
    */
   UINT getNumberOfPixelClocks();

   /*!
    * \brief Get the supported pixel clock range.
    * \param[out] range The minimum, maximum and mid values in range.
    *  Must have been allocated elsewhere (array of 3).
    * \return min max and mid values in range.
    */
   void getPixelClockRange(UINT* range);

   /*!
    * \brief get the current pixel clock value 
    * \return the pixel clock value.
    */
   UINT getCurrentPixelClockValue();
   /*!
    * \brief get the default pixel clock value
    * \return default pixel clock value.
    */
   UINT getDefaultPixelClockValue();
   /*!
    * \brief set the pixel clock
    * \param[in] newValue the new pixel clock value 
    * \return success or failure
    */
   INT setPixelClockVal(UINT newValue);
   
   /*
    * \brief Set the properties of the picture.
    *
    * If a parameter is negative, then it is not used.
    *
    * Values are given as a percentage.
    *
    * \param[in] bri Brightness
    * \param[in] col Colour
    * \param[in] con Contrast
    * \param[in] hue Hue
    * \param[in] whi Whiteness
    * \param[in] dep Depth
    * \return true if the function was successful.
    */
//    bool setPictureProperties(long bri, long col,
//                              long con, long hue,
//                              long whi, long dep);

   /*
    * \brief Use the picture properties currently set.
    * \return true is successful, false otherwise.
    * \sa setPictureProperties().
    */
//    bool usePictureProperties();

   /*!
    * \brief Move the viewport to the right.
    * \sa setInputRectangle().
    */
   void moveRight();

   /*!
    * \brief Move the viewport to the left.
    * \sa setInputRectangle().
    */
   void moveLeft();

   /*!
    * \brief Move the viewport up.
    * \sa setInputRectangle().
    */
   void moveUp();

   /*!
    * \brief Move the viewport down.
    * \sa setInputRectangle().
    */
   void moveDown();

   /*!
    * \brief Increase the width of the viewport.
    * \sa setInputRectangle().
    */
   void increaseWidth();

   /*!
    * \brief Decrease the width of the viewport.
    * \sa setInputRectangle().
    */
   void decreaseWidth();

   /*!
    * \brief Increase the height of the viewport.
    * \sa setInputRectangle().
    */
   void increaseHeight();

   /*!
    * \brief Decrease the height of the viewport.
    * \sa setInputRectangle().
    */
   void decreaseHeight();

   /*!
    * \brief Set the gain.
    *
    * \param[in] value The new gain value, between 0 and 100.  A negative value
    * sets automatic gain control.
    */
   void setGain(long value);

   /*!
    * \brief Get the current gain value.
    * \return The value.
    */
   long getGain();

   /*!
    * \brief Get the minimum value for the gain.
    * \return The minimum value.
    */
   long getMinGain()
   {
      return(static_cast<long>(minGain));
   }

   /*!
    * \brief Get the maximum value for the gain.
    * \return The maximum value.
    */
   long getMaxGain()
   {
      return(static_cast<long>(maxGain));
   }

   /*!
    * \brief Set the exposure.
    *
    * \param[in] value The new exposure value.  A negative value means automatic
    * exposure control.
    */
   void setExposure(long value);

   /*!
    * \brief Get the current exposure value.
    * \return The value.
    */
   long getExposure();

   /*!
    * \brief Get the minimum value for the exposure.
    * \return The minimum value.
    */
   long getMinExposure()
   {
      return(static_cast<long>(minExposure));
   }

   /*!
    * \brief Get the maximum value for the exposure.
    * \return The maximum value.
    */
   long getMaxExposure()
   {
      return(static_cast<long>(maxExposure));
   }

   /*!
    * \brief Get the exposure interval.
    * \return The exposure interval value.
    */
   long getExposureIntervall()
   {
      return(static_cast<long>(exposureInterval));
   }

   /*!
    * \brief Set the framerate.
    *
    * \param[in] value The new framerate value.
    */
   void setFramerate(double value);

   /*!
    * \brief Get the current framerate value.
    * \return The value.
    */
   double getFramerate();

   /*!
    * \brief Get the minimum value for the framerate.
    * \return The minimum value.
    */
   double getMinFramerate()
   {
      return(minFramerate);
   }

   /*!
    * \brief Get the maximum value for the framerate.
    * \return The maximum value.
    */
   double getMaxFramerate()
   {
      return(maxFramerate);
   }
   /*!
    * \brief Get the current auto exposure setting.
    * \return The auto exposure setting.
    */
   long getAutoExposureSetting();
   /*!
    * \brief Set the current auto exposure setting.
    * \param[in] val The auto exposure setting, 1 enables control, 0 disables
    * control.
    */
   void setAutoExposureControl(int val);

protected:

   /*!
    * \brief Initialise the grabber.
    * \param[in] upsideDown Set to true if the camera is upside-down.
    * \param[in] flip Set to true to horizontally flip the images.
    * \param[in] camNumber The number of the camera (0 is the first available).
    * \param[in] posX
    * \param[in] posY The position of the viewport of the grabbed part.
    * \param[in] width
    * \param[in] height The width and height of the grabbed area and fo the
    * images.
    */
   void init(bool upsideDown,
             bool flip,
             int camNumber,
             unsigned short posX = 0,
             unsigned short posY = 0,
             unsigned short width = UEYE_IMAGE_DEFAULT_INPUTWIDTH,
             unsigned short height = UEYE_IMAGE_DEFAULT_INPUTHEIGHT);
    

private:

   /*!
    * \brief The grabbed image.
    */
   char* grabbedImage;
   char* ringbuffer[5];
   int sequenceBufIDs[5];
   /*!
    * \brief Major version of the library this was created for.
    */
   const static unsigned char majorVersion = 3;

   /*!
    * \brief Minor version of the library this was created for.
    */
   const static unsigned char minorVersion = 50;

   /*!
    * \brief Camera handle.
    */
   HIDS cameraHandle;

   /*!
    * \brief ID of the memory allocated for the grabbed image.
    * \sa grabbedImage->
    */
   INT memID;

   /*!
    * \brief Input rectangle X position.
    */
   INT inputX;

   /*!
    * \brief Input rectangle Y position.
    */
   INT inputY;

   /*!
    * \brief Information about the camera's sensor.
    */
   SENSORINFO sensorInfo;

   /*!
    * \brief Min gain.
    */
   double minGain;

   /*!
    * \brief Max gain.
    */
   double maxGain;

   /*!
    * \brief Min exposure.
    */
   double minExposure;

   /*!
    * \brief Max exposure.
    */
   double maxExposure;

   /*!
    * \brief Exposure interval.
    */
   double exposureInterval;

   /*!
    * \brief Min framerate.
    */
   double minFramerate;

   /*!
    * \brief Max framerate.
    */
   double maxFramerate;

   /*!
    * \brief Framerate interval.
    */
   double framerateInterval;
};

} // namespace imalib

#endif // UEYEIMAGEGRABBER_H
