/*
 * imalib
 * 
 * Copyright 2017 Aberystwyth University
 * 
 * This library and program are free software: you can redistribute them
 * and/or modify them under the terms of the GNU General Public License as 
 * published by the Free Software Foundation, either version 3 of the License,
 * or (at your option) any later version.
 * 
 * The program and library are distributed in the hope that they will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*!
 * \file utilities.h
 * \brief A file where several utility functions are defined.
 *
 * \author Fred Labrosse <ffl@aber.ac.uk>
 * \copyright 2017 Aberystwyth University
 */

#ifndef UTILITIES_H
#define UTILITIES_H

#include <stdlib.h>
#include <cmath>
#include <vector>

#include <gsl/gsl_multimin.h>

#include <boost/program_options.hpp>
namespace po = boost::program_options;

namespace imalib
{

/*!
 * \class imalibOptions imalib/utilities.h imalib/utilities.h
 * \brief A class to store global options for imalib.
 */
class imalibOptions
{
public:

   /*!
    * \brief Default constructor.
    */
   imalibOptions();

   /*!
    * \brief Get the options so that they can be incorporated into programs.
    * \return The options.
    */
   po::options_description* getOptions()
   {
      return (&imalibOptions_);
   }

   /*!
    * \brief Process the command line.
    * \param[in] vm The boost program_options variable map.
    */
   void commandLine(po::variables_map& vm);

private:

   /*!
    * \brief The imalib global program_options options.
    */
   static po::options_description imalibOptions_;
};

/*!
 * \brief Computes the sign of the parameter.
 * \param[in] val The value of which we compute the sign.
 * \return -1 if val < 0, 0 if val == 0 and +1 if val > 0.
 */
double sign(double val);

/*!
 * \brief Computes the modulo of a number, even when its value is negative.
 *
 * We assume that the devider is positive (only tested if compiled with NDEBUG).
 *
 * \param[in] number The number.
 * \param[in] mod The divider.
 * \return The new value.
 */
template <class T>
T modulo(T number, T mod)
{
   assert(mod >= 0);
   if (number < 0)
   {
      do
      {
         number += mod;
      } while (number < 0);
   }
   // In some cases, rounding of the previous do-while makes number == mod...
   if (number >= mod)
   {
      do
      {
         number -= mod;
      } while (number >= mod);
   }

   return(number);
}


/*!
 * \brief Computes a random number uniformly distributed between min and max.
 *
 * The type T must be castable to and from double, must have operator>
 * and operator-.
 *
 * \param[in] min The min value.
 * \param[in] max The max value.
 * \return The random number.
 */
template <class T>
T random(T min, T max)
{
   if (min > max)
   {
      T tmp = max;
      max = min;
      min = tmp;
   }
   return(min + static_cast<T>(static_cast<double>(max - min)
          * static_cast<double>(rand())
          / static_cast<double>(RAND_MAX + 1.0)));
}

/*!
 * \brief Round a double to nearest long.
 *
 * Also works for negative numbers.
 *
 * \param num (IN) The number to round.
 * \return The rounded number.
 */
inline long round(double num)
{
   if (num < 0)
      return(static_cast<long>(num - 0.5));
   else
      return(static_cast<long>(num + 0.5));
}

/*
 * \brief Compute the maximum of two doubles.
 * \param nb1 (IN) The first number.
 * \param nb2 (IN) The second number.
 * \return The maximum of the two values.
 */
// inline double max(double nb1, double nb2)
// {
//    return((nb1 > nb2) ? nb1 : nb2);
// }

/*
\brief Compute the minimum of two doubles.
\param nb1 (IN) The first number.
\param nb2 (IN) The second number.
\return The minimum of the two values.
*/
// inline double min(double nb1, double nb2)
// {
// return((nb1 > nb2) ? nb2 : nb1);
// }

/*!
 * \brief Compute the median value from the values in the given vector.
 *
 * The type T must be castable into a double and back.
 *
 * \param[in] values The values from which to extract the median.  The vector is
 * partially ordered at the end of the process.
 * \return The median value.
 */
template <class T>
T median(std::vector<T>& values)
{
   double size2d = static_cast<double>(values.size()) / 2.0;
   unsigned short size2 = static_cast<unsigned short>(size2d);
   if (size2d != 0)
   {
      // Odd case
      nth_element(values.begin(),
                  values.begin() + size2,
                  values.end());
      return(values[size2]);
   }
   else
   {
      // Even case
      nth_element(values.begin(),
                  values.begin() + size2 - 1,
                  values.end());
      nth_element(values.begin() + size2,
                  values.begin() + size2,
                  values.end());
      return(static_cast<T>(static_cast<double>(values[size2-1] +
                                                values[size2]) / 2.0));
   }
}

/*!
 * \brief Round to the nearest multiple of the given parameter.
 * \param[in] num The number to round.
 * \param[in] multiple The multiple.
 * \return The number rounded.
 */
inline long roundNearestMultiple(long num, long multiple)
{
   return(lround(static_cast<double>(num) / static_cast<double>(multiple))
          * multiple);
}

/*!
 * \brief Gives a global verbosity level.
 * 
 * Hides a global variable (see [\ref Mey98, Item 47]).
 * 
 * The function can be manipulated as if a number.
 * 
 * \return A global verbosity level.
 */
inline unsigned short& verboseLevel()
{
   static unsigned short level = 0;
   return(level);
}

/*!
 * \class gslMultiminFunction imalib/utilities.h imalib/utilities.h
 * \brief This class is used as a wrapper for gsl_multimin_function's so that
 * class members can be used in GSL code ("objectiveFunction()" below).
 * 
 * This is inspired by http://stackoverflow.com/questions/13074756/how-to-avoid-static-member-function-when-using-gsl-with-c
 *
 * This is then used as follows in the class that needs the optimisation:
 * \verbatim
 auto fctptr = [this](const gsl_vector* x)->double{return objectiveFunction(x);};
 gslMultiminFunction<decltype(fctptr)> Fp(fctptr, n); // n is the size of the gsl_vector.
 gsl_multimin_function *F = static_cast<gsl_multimin_function*>(&Fp);
 \endverbatim
 * The lambda function could obviously capture other local variables if needed
 * as well as all members of the class.
 */
template< typename F >
class gslMultiminFunction : public gsl_multimin_function
{
public:

   /*!
    * \brief Constructor.
    * \param[in] func The member function to optimise.  It must return a double
    * and take as argument a const gsl_vector* of size nDim.  This is typically a 
    * \param[in] nDim The number of dimensions of the problem.
    */
   gslMultiminFunction(const F& func, size_t nDim)
   : _func(func)
   {
      // Both function and params are from gsl_function.
      f = &gslMultiminFunction::invoke;
      n = nDim;
      params = this;  // This is actually not used.
   }

private:

   const F& _func;
   static double invoke(const gsl_vector* x, void *params)
   {
      return static_cast<gslMultiminFunction*>(params)->_func(x);
   }
};

} // namespace imalib

#endif // UTILITIES_H
// Don't write anything important below.
