/*
 * imalib
 * 
 * Copyright 2017 Aberystwyth University
 * 
 * This library and program are free software: you can redistribute them
 * and/or modify them under the terms of the GNU General Public License as 
 * published by the Free Software Foundation, either version 3 of the License,
 * or (at your option) any later version.
 * 
 * The program and library are distributed in the hope that they will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*!
 * \file v4l2ImageGrabber.h
 * \brief An image grabber for a v4l2 frame grabber.
 *
 * \author Fred Labrosse <ffl@aber.ac.uk>
 * \copyright 2017 Aberystwyth University
 */

#ifndef V4L2IMAGEGRABBER_H
#define V4L2IMAGEGRABBER_H

#include "imageGrabber.h"

#include <asm/types.h>          /* for videodev2.h */
#include <linux/videodev2.h>
#include <sys/ioctl.h>
#include <errno.h>

#include <vector>

namespace imalib
{

/*!
 * \class v4l2ImageGrabber v4l2ImageGrabber.h imalib/v4l2ImageGrabber.h
 * \brief A class to interface a v4l2 image grabber.
 *
 * If the class is available, WITH_V4L2 is defined (if you use
 * imalib-config --cflags when compiling, see \ref autotools that page).
 *
 * At the moment, only the READ and MMAP (preferred) io methods are supported.
 *
 * This is largelly inspired from the example at
 * <http://v4l2spec.bytesex.org/spec/capture-example.html>.
 */
class v4l2ImageGrabber : public imageGrabber
{
public:

   /*!
    * \brief Possible io methods to get images.
    */
   typedef enum
   {
      IO_METHOD_READ,    //! Direct read method.
      IO_METHOD_MMAP,    //! Memory mapping method.
      IO_METHOD_USERPTR  //! User pointer method.
   } possibleIoMethods;

   /*!
    * \brief Constructor.
    *
    * \param[in] imgWidth
    * \param[in] imgHeight The width and height of the image.
    * \param[in] upsideDown Set to true if the camera is upside-down.
    * \param[in] flip Set to true to horizontally flip the images.
    * \param[in] devNumber The number of the video device (/dev/video?).
    * \param[in] inputNumber The video input number.
    * \param[in] videoStandard The video standard number.
    * \param[in] ioMethod The io method to use.
    * \param[in] posX
    * \param[in] posY The position of the viewport of the grabbed part.
    * \param[in] width
    * \param[in] height The width and height of the grabbed area.
    */
   v4l2ImageGrabber(unsigned short imgWidth, unsigned short imgHeight,
                    bool upsideDown,
                    bool flip,
                    int devNumber,
                    int inputNumber,
                    int videoStandard,
                    possibleIoMethods ioMethod = IO_METHOD_MMAP,
                    unsigned short posX = 0,
                    unsigned short posY = 0,
                    unsigned short width = IMAGE_DEFAULT_INPUTWIDTH,
                    unsigned short height = IMAGE_DEFAULT_INPUTHEIGHT);

   /*!
    * \brief Constructor.
    * \param[in] fileName Name of the configuration file.
    */
   v4l2ImageGrabber(const std::string& fileName);

   /*!
    * \brief Destructor.
    */
   virtual ~v4l2ImageGrabber();

   /*!
    * \brief Tell if we are connected to the required frame grabber.
    * \return true if we are connected, false otherwise.
    */
   virtual bool isConnected()
   {
      return (connected);
   }

   /*!
    * \brief Display on the standard output some information about the
    * source of the images.
    *
    * \param[in,out] aStream The stream to which the info is added.
    */
   virtual void printInfoAboutSource(std::ostringstream* aStream);

   /*!
    * \brief Capture the image from a grabber.
    *
    * This blocks until data is available to be read.  This should be called
    * before calling getImage().
    */
   virtual void captureImage();

   /*!
    * \brief Get an image from the frame grabber.
    *
    * The images are resized here.
    *
    * captureImage() should be called before, but will be called if it wasn't.
    *
    * \param[out] theImage The image in which to save the grabbed image, if
    * not NULL.
    * \param[out] rawImage The image in which to save the raw image (before
    * any transformation is applied), if not NULL.
    */
   virtual void getImage(imageRGB* theImage, imageRGB* rawImage = NULL);

   /*!
    * \brief Specify the input rectangle (viewport on the grabbed image).
    * \param[in] x
    * \param[in] y Coordinates of the top-left corner of the rectangle.
    * \param[in] w
    * \param[in] h Size of the rectangle.
    * \return true is successful, false otherwise.
    * \sa useInputRectangle().
    */
   bool setInputRectangle(int x, int y, int w, int h);

   /*!
    * \brief Use the input rectangle currently set.
    * \return true is successful, false otherwise.
    * \sa setInputRectangle().
    */
   bool useInputRectangle();

   /*!
    * \brief Set the properties of the picture.
    *
    * If a parameter is negative, then it is not used.
    *
    * Values are given as a percentage.
    *
    * \param[in] bri Brightness
    * \param[in] col Colour
    * \param[in] con Contrast
    * \param[in] hue Hue
    * \param[in] whi Whiteness
    * \param[in] dep Depth
    * \return true if the function was successful.
    */
   bool setPictureProperties(long bri, long col,
                             long con, long hue,
                             long whi, long dep);

   /*!
    * \brief Use the picture properties currently set.
    * \return true is successful, false otherwise.
    * \sa setPictureProperties().
    */
   bool usePictureProperties();

   /*!
    * \brief Move the viewport to the right.
    * \sa setInputRectangle().
    */
   void moveRight();

   /*!
    * \brief Move the viewport to the left.
    * \sa setInputRectangle().
    */
   void moveLeft();

   /*!
    * \brief Move the viewport up.
    * \sa setInputRectangle().
    */
   void moveUp();

   /*!
    * \brief Move the viewport down.
    * \sa setInputRectangle().
    */
   void moveDown();

   /*!
    * \brief Increase the width of the viewport.
    * \sa setInputRectangle().
    */
   void increaseWidth();

   /*!
    * \brief Decrease the width of the viewport.
    * \sa setInputRectangle().
    */
   void decreaseWidth();

   /*!
    * \brief Increase the height of the viewport.
    * \sa setInputRectangle().
    */
   void increaseHeight();

   /*!
    * \brief Decrease the height of the viewport.
    * \sa setInputRectangle().
    */
   void decreaseHeight();

protected:

   /*!
    * \brief Initialise the grabber.
    * \param[in] imgWidth
    * \param[in] imgHeight The width and height of the image.
    * \param[in] upsideDown Set to true if the camera is upside-down.
    * \param[in] flip Set to true to horizontally flip the images.
    * \param[in] devNumber The number of the video device (/dev/video?).
    * \param[in] inputNumber The video input number.
    * \param[in] videoStandard The video standard number.
    * \param[in] ioMethod The io method to use.
    * \param[in] posX
    * \param[in] posY The position of the viewport of the grabbed part.
    * \param[in] width
    * \param[in] height The width and height of the grabbed area.
    */
   void init(unsigned short imgWidth, unsigned short imgHeight,
             bool upsideDown,
             bool flip,
             int devNumber,
             int inputNumber,
             int videoStandard,
             possibleIoMethods ioMethod = IO_METHOD_MMAP,
             unsigned short posX = 0,
             unsigned short posY = 0,
             unsigned short width = IMAGE_DEFAULT_INPUTWIDTH,
             unsigned short height = IMAGE_DEFAULT_INPUTHEIGHT);

private:

   /*!
    * \brief A structure to store image buffers (start and length).
    */
   struct dataBuffer
   {
      unsigned char *start; //! Start in memory of the buffer.
      size_t length; //! Length in memory of the buffer.
   };

   /*!
    * \brief Image buffers.
    */
   dataBuffer* imageBuffers;

   /*!
    * \brief Number of buffers used.
    */
   unsigned int nbBuffers;

   /*!
    * \brief The file descriptor for the video device.
    */
   int fd;

   /*!
    * \brief The device number (on the frame grabber).
    */
   int deviceNumber;

   /*!
    * \brief Device name.
    */
   std::string deviceName;

   /*
    * \brief Video input.
    */
//    int videoInput_;

   /*!
    * \brief The io method used.
    */
   possibleIoMethods ioMethod_;

   /*!
    * \brief Hardware capability.
    */
   struct v4l2_capability hdwCap_;

   /*!
    * \brief Cropping capability of the hardware.
    */
   struct v4l2_cropcap cropcap;

   /*!
    * \brief Current hardware crop.
    */
   struct v4l2_crop crop;

   /*!
    * \brief Image format.
    */
   struct v4l2_format fmt;

   /*!
    * \brief Supported pixel formats.
    *
    * They must be in order of preference, preferred first.
    */
   static const std::vector<__u32> supportedPixelFormats;

   /*!
    * \brief Pixel format currently used.
    */
    __u32 pixelFormat_;

    /*!
     * True if we are streaming.
     */
    bool streaming_;

   /*!
    * \brief Repeatedly calls ioctl() until something happens.
    * \param[in] fd File descriptor.
    * \param[in] request Request.
    * \param[in] arg Arugments of the request.
    * \return Return value of ioctl().
    */
   int xioctl(int fd, int request, void *arg)
   {
      int r;

      do
      {
         r = ioctl(fd, request, arg);
      }
      while (-1 == r && EINTR == errno);

      return(r);
   }

   /*!
    * \brief Initialise memory for the IO_METHOD_READ io method.
    * \param[in] bufferSize Needed buffer size.
    * \return true is all went well, false otherwise.
    */
   bool initRead(unsigned int bufferSize);

   /*!
    * \brief Initialise memory for the IO_METHOD_MMAP io method.
    * \return true is all went well, false otherwise.
    */
   bool initMmap();

   /*!
    * \brief Initialise memory for the IO_METHOD_USERPTR io method.
    *
    * This method is not yet supported.
    *
    * \param[in] bufferSize Needed buffer size.
    * \return true is all went well, false otherwise.
    */
   bool initUserPtr(unsigned int bufferSize);

   /*!
    * \brief Process the new image.
    *
    * This is taking the raw data from the grabber and putting it into the
    * required image.
    *
    * \param[in] buffer The buffer to take the data from.
    * \param[out] theImage The image in which to save the grabbed image, if
    * not NULL.
    * \param[out] rawImage The image in which to save the raw image (before any
    * transformation is applied), if not NULL.
    */
   void processNewImage(unsigned char* buffer,
                        imageRGB* theImage, imageRGB* rawImage);

   /*!
    * \brief Convert and clamp a double RGB colour component to unsigned char.
    */
   unsigned char clamp(double comp)
   {
      if (comp <= 0)
         return(0);
      else if (comp >= 255)
         return(255);
      else
         return(static_cast<unsigned char>(lround(comp)));
   }

   /*!
    * \brief Convert a Y Cr Cb colour to RGB.
    */
   void YCbCrToRGB(unsigned char* YCbCr, unsigned char* RGB)
   {
      double r, g, b;
      double y1, pb, pr;

      y1 = (255.0 / 219.0) * (YCbCr[0] - 16);
      pb = (255.0 / 224.0) * (YCbCr[1] - 128);
      pr = (255.0 / 224.0) * (YCbCr[2] - 128);

      r = 1.0 * y1 + 0.0   * pb + 1.402 * pr;
      g = 1.0 * y1 - 0.344 * pb - 0.714 * pr;
      b = 1.0 * y1 + 1.772 * pb + 0.0   * pr;

      RGB[0] = clamp(r); /* [ok? one should prob. limit y1,pb,pr] */
      RGB[1] = clamp(g);
      RGB[2] = clamp(b);
   }

   /*!
    * \brief Convert a Bayer GBRG to RGB image.
    *
    * \param[in] gbrg The array of GBRG data.
    * \param[out] rgb The resulting RGB image.
    */
    void GBRGtoRGB(unsigned char* gbrg, unsigned char* rgb);
};

} // namespace imalib

#endif
