/*
 * imalib
 * 
 * Copyright 2017 Aberystwyth University
 * 
 * This library and program are free software: you can redistribute them
 * and/or modify them under the terms of the GNU General Public License as 
 * published by the Free Software Foundation, either version 3 of the License,
 * or (at your option) any later version.
 * 
 * The program and library are distributed in the hope that they will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*!
 * \file v4lImageGrabber.h
 * \brief An image grabber for a v4l frame grabber.
 * 
 * \author Fred Labrosse
 * \copyright 2017 Aberystwyth University
 */

#ifndef V4LIMAGEGRABBER_H
#define V4LIMAGEGRABBER_H

#include "imalib/imageGrabber.h"

#include <linux/videodev.h>

namespace imalib
{

/*!
 * \class v4lImageGrabber v4lImageGrabber.h imalib/v4lImageGrabber.h
 * \brief A class to interface a v4l image grabber.
 *
 * If the class is available, WITH_V4L is defined (if you use
 * imalib-config --cflags when compiling, see \ref autotools that page).
 */
class v4lImageGrabber : public imageGrabber
{
public:

   /*!
    * \brief Constructor.
    *
    * \param[in] imgWidth
    * \param[in] imgHeight The width and height of the image.
    * \param[in] upsideDown Set to true if the camera is upside-down.
    * \param[in] flip Set to true to horizontally flip the images.
    * \param[in] devNumber The number of the video device (/dev/video?).
    * \param[in] portNumber The number of the port on the device.
    * \param[in] posX
    * \param[in] posY The position of the viewport of the grabbed part.
    * \param[in] width
    * \param[in] height The width and height of the grabbed area.
    */
   v4lImageGrabber(unsigned short imgWidth, unsigned short imgHeight,
                   bool upsideDown,
                   bool flip,
                   int devNumber, int portNumber,
                   unsigned short posX = 0,
                   unsigned short posY = 0,
                   unsigned short width = IMAGE_DEFAULT_INPUTWIDTH,
                   unsigned short height = IMAGE_DEFAULT_INPUTHEIGHT);

   /*!
    * \brief Constructor.
    * \param fileName (IN) Name of the configuration file.
    */
   v4lImageGrabber(const std::string& fileName);

   /*!
    * \brief Destructor.
    */
   virtual ~v4lImageGrabber();

   /*!
    * \brief Tell if we are connected to the required frame grabber.
    * \return true if we are connected, false otherwise.
    */
   virtual bool isConnected()
   {
      return (connected);
   }

   /*!
    * \brief Display on the standard output some information about the
    * source of the images.
    */
   virtual void printInfoAboutSource();

   /*!
    * \brief Capture the image from a grabber.
    *
    *  This is there because the capture can take a while.  We thus first
    * capture and then use (with getImage()).  This does not change the
    * image data.
    */
   virtual void captureImage();

   /*!
    * \brief Get an image from the frame grabber.
    *
    * The image might have been pre-captured (with captureImage()), but
    * if not will be captured here.
    *
    * The images are resized here.
    *
    * \param[out] theImage The image in which to save the grabbed image, if
    * not NULL.
    * \param[out] rawImage The image in which to save the raw image (before
    * any transformation is applied), if not NULL.
    */
   virtual void getImage(imageRGB* theImage, imageRGB* rawImage = NULL);

   /*!
    * \brief Specify the input rectangle (viewport on the grabbed image).
    * \param[in] x
    * \param[in] y Coordinates of the top-left corner of the rectangle.
    * \param[in] w
    * \param[in] h Size of the rectangle.
    * \return true is successful, false otherwise.
    * \sa useInputRectangle().
    */
   bool setInputRectangle(int x, int y, int w, int h);

   /*!
    * \brief Use the input rectangle currently set.
    * \return true is successful, false otherwise.
    * \sa setInputRectangle().
    */
   bool useInputRectangle();

   /*!
    * \brief Set the properties of the picture.
    *
    * If a parameter is negative, then it is not used.
    *
    * Values are given as a percentage.
    *
    * \param[in] bri Brightness
    * \param[in] col Colour
    * \param[in] con Contrast
    * \param[in] hue Hue
    * \param[in] whi Whiteness
    * \param[in] dep Depth
    * \return true if the function was successful.
    */
   bool setPictureProperties(long bri, long col,
                             long con, long hue,
                             long whi, long dep);

   /*!
    * \brief Use the picture properties currently set.
    * \return true is successful, false otherwise.
    * \sa setPictureProperties().
    */
   bool usePictureProperties();

   /*!
    * \brief Move the viewport to the right.
    * \sa setInputRectangle().
    */
   void moveRight();

   /*!
    * \brief Move the viewport to the left.
    * \sa setInputRectangle().
    */
   void moveLeft();

   /*!
    * \brief Move the viewport up.
    * \sa setInputRectangle().
    */
   void moveUp();

   /*!
    * \brief Move the viewport down.
    * \sa setInputRectangle().
    */
   void moveDown();

   /*!
    * \brief Increase the width of the viewport.
    * \sa setInputRectangle().
    */
   void increaseWidth();

   /*!
    * \brief Decrease the width of the viewport.
    * \sa setInputRectangle().
    */
   void decreaseWidth();

   /*!
    * \brief Increase the height of the viewport.
    * \sa setInputRectangle().
    */
   void increaseHeight();

   /*!
    * \brief Decrease the height of the viewport.
    * \sa setInputRectangle().
    */
   void decreaseHeight();

protected:

   /*!
    * \brief Initialise the grabber.
    * \param[in] imgWidth
    * \param[in] imgHeight The width and height of the image.
    * \param[in] upsideDown Set to true if the camera is upside-down.
    * \param[in] flip Set to true to horizontally flip the images.
    * \param[in] devNumber The number of the video device (/dev/video?).
    * \param[in] portNumber The number of the port on the device.
    * \param[in] posX
    * \param[in] posY The position of the viewport of the grabbed part.
    * \param[in] width
    * \param[in] height The width and height of the grabbed area.
    */
   void init(unsigned short imgWidth, unsigned short imgHeight,
             bool upsideDown,
             bool flip,
             int devNumber, int portNumber,
             unsigned short posX = 0,
             unsigned short posY = 0,
             unsigned short width = IMAGE_DEFAULT_INPUTWIDTH,
             unsigned short height = IMAGE_DEFAULT_INPUTHEIGHT);

private:

   /*!
    * \brief The grabbed image.
    */
   unsigned char* grabbedImage;

   /*!
    * \brief The file descriptor for the video device.
    */
   int fd;

   /*!
    * \brief The video_picture structure,
    */
   video_picture vPicture;

   /*!
    * \brief the video_channel structure.
    */
   video_channel vChannel;

   /*!
    * \brief The video_channel structure.
    */
   video_capture vCapture;

   /*!
    * \brief The video_window structure.
    */
   video_window vWindow;

   /*!
    * \brief The video_mmap structure.
    */
   video_mmap vMmap;

   /*!
    * \brief The device number (on the frame grabber).
    */
   int deviceNumber;
};

} // namespace imalib

#endif

