/*
 * imalib
 * 
 * Copyright 2017 Aberystwyth University
 * 
 * This library and program are free software: you can redistribute them
 * and/or modify them under the terms of the GNU General Public License as 
 * published by the Free Software Foundation, either version 3 of the License,
 * or (at your option) any later version.
 * 
 * The program and library are distributed in the hope that they will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*!
 * \file picassoImageGrabber.cpp
 * \brief An image grabber for a picasso frame grabber.  Implementation.
 *
 * \author Fred Labrosse
 * \copyright 2017 Aberystwyth University
 */

#include "imalib/imageRGB.h"
#include "imalib/picassoImageGrabber.h"

#include <sys/mman.h>
#include <math.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <stdio.h>
#include <unistd.h>

#include <fstream>
#include <iostream>

#include <boost/program_options.hpp>
namespace po = boost::program_options;

namespace imalib
{

/*----------------------------------------------------------------*\
Class picassoImageGrabber
\*----------------------------------------------------------------*/

short picassoImageGrabber::numberOfInstances = 0;
int picassoImageGrabber::boardIndex = -1;

/*!
 * \brief Constructor.
 *
 * \param[in] imgWidth
 * \param[in] imgHeight The width and height of the image.
 * \param[in] upsideDown Set to true if the camera is upside-down.
 * \param[in] flip Set to true to horizontally flip the images.
 * \param[in] grabNum The number of the Picasso grabber to use.
 * \param[in] devPort The number of the port on the device.
 * \param[in] posX
 * \param[in] posY The position of the viewport of the grabbed part.
 * \param[in] width
 * \param[in] height The width and height of the grabbed area.
 */
picassoImageGrabber::picassoImageGrabber(unsigned short imgWidth,
                                         unsigned short imgHeight,
                                         bool upsideDown,
                                         bool flip,
                                         int grabNum, int devPort,
                                         unsigned short posX,
                                         unsigned short posY,
                                         unsigned short width,
                                         unsigned short height)
   : grabberHandle(0)
{
   init(imgWidth, imgHeight,
        upsideDown,
        flip,
        grabNum, devPort,
        posX, posY,
        width, height);
}

/*!
 * \brief Constructor.
 * \param[in] fileName Name of the configuration file.
 */
picassoImageGrabber::picassoImageGrabber(const std::string& fileName)
{
   // Let's extract the options from the configuration file.
   po::options_description optionsDesc("picassoImageGrabber options");
   int grabNum = 0;
   int devPo = 0;
   int poX = 0;
   int poY = 0;
   int soWidth = IMAGE_DEFAULT_INPUTWIDTH;
   int soHeight = IMAGE_DEFAULT_INPUTHEIGHT;
   int imWidth = 200;
   int imHeight = 200;
   bool upDown = false;
   bool fl = false;
   optionsDesc.add_options()
      ("grabberNumber", po::value<int>(&grabNum)->default_value(grabNum), "")
      ("devicePort", po::value<int>(&devPo)->default_value(devPo), "")
      ("posX", po::value<int>(&poX)->default_value(poX), "")
      ("posY", po::value<int>(&poY)->default_value(poY), "")
      ("sourceWidth",
       po::value<int>(&soWidth)->default_value(soWidth), "")
      ("sourceHeight",
       po::value<int>(&soHeight)->default_value(soHeight), "")
      ("imageWidth", po::value<int>(&imWidth)->default_value(imWidth), "")
      ("imageHeight", po::value<int>(&imHeight)->default_value(imHeight), "")
      ("upsideDown", po::value<bool>(&upDown)->default_value(upDown), "")
      ("flip", po::value<bool>(&fl)->default_value(fl), "")
      ;

   po::variables_map vm;
   std::ifstream configFile(fileName.c_str());
   if (configFile.is_open())
   {
      po::store(po::parse_config_file(configFile, optionsDesc), vm);
      po::notify(vm);
      configFile.close();
   }
   else
   {
      std::cerr << "picassoImageGrabber::picassoImageGrabber():\n"
         << "WARNING: could not open configuration file " << fileName
         << "!\n"
         << "\tUsing defaults.\n";
   }

   init(imWidth, imHeight,
        upDown,
        fl,
        grabNum, devPo,
        poX, poY,
        soWidth, soHeight);
}


/*!
 * \brief Destructor.
 */
picassoImageGrabber::~picassoImageGrabber()
{
   numberOfInstances--;
   if (numberOfInstances == 0)
   {
      PcssRetType ret = PcssClose(grabberHandle);
      if (ret != PCSS_NO_ERROR)
      {
         std::cerr << "Error while closing the Picasso board.\n";
      }
      else
      {
         boardIndex = -1;
      }

      ret = PcssCloseLibrary();
      if (ret != PCSS_NO_ERROR)
      {
         std::cerr << "Error while closing the Picasso library.\n";
      }
   }
}

/*!
 * \brief Display on the standard output some information about the
 * source of the images.
 *
 * \param[in,out] aStream The stream to which the info is added.
 */
void picassoImageGrabber::printInfoAboutSource(std::ostringstream* aStream)
{
   *aStream << "----------------------------------------------\n"
      << "Image source:\n";

   SPcssCardInfo cardInfo;
   PCSSINITVAR(cardInfo);
   PcssRetType ret = PcssFunc(grabberHandle, PCSSF_GET_CARDINFO, &cardInfo);
   if (ret != PCSS_NO_ERROR)
   {
      std::cerr << "Unable to get the Picasso board info.\n";
   }
   else
   {
      char buff[128];
      PcssEnumToStr(PCSSBOARDTYPE, cardInfo.btype, buff, 128);
      *aStream << "\tgrabber: " << buff << std::endl;
   }

   long param;
   ret = PcssFunc(grabberHandle, PCSSF_GET_INPUTCONNPORT, &param);
   if (ret != PCSS_NO_ERROR)
   {
      std::cerr << "Unable to get the video port of the frame grabber.\n";
   }
   else
   {
      char buff[128];
      PcssEnumToStr(PCSSINPUTCONN, param, buff, 128);
      *aStream << "\tvideo input port: " << buff << std::endl;
   }

   SPcssStatus stat;
   PCSSINITVAR(stat);
   ret = PcssFunc(grabberHandle, PCSSF_GET_STATUS, &stat);
   if (ret != PCSS_NO_ERROR)
   {
      std::cerr << "Could not get status of board.\n";
   }

   if (stat.inpstat & SQ_HLCK) /* check HLCK bit */
   {
      *aStream << "\t==> no video detected on input port.\n";
      //          return(false);
   }
   else
   {
      if (stat.inpstat & SQ_FIDT)
      {       /* 60 Hz */
         if (stat.inpstat & SQ_CODE)
            *aStream << "\tNTSC video detected (60 Hz, color).\n";
         else
            *aStream << "\tRS170 video detected (60 Hz, b/w).\n";
      }
      else
      { /* 50 Hz */
         if (stat.inpstat & SQ_CODE)
            *aStream << "\tPAL/SECAM video detected (50 Hz, color).\n";
         else
            *aStream << "\tCCIR video detected (50 Hz, b/w).\n";
      }
   }

   printBasicInfoAboutSource(aStream);
   *aStream << "----------------------------------------------\n";
}

/*!
 * \brief Capture the image from a grabber.
 *
 * This is there because the capture can take a while.  We thus first
 * capture and then use (with getImage()).
 */
void picassoImageGrabber::captureImage()
{
   if (connected)
   {
      long doStart = 1;
      PcssRetType ret = PcssFunc(grabberHandle, PCSSF_START_CAPTURE, &doStart);
      if (ret != PCSS_NO_ERROR)
      {
         std::cerr << "Could not start grabbing.\n";
         connected = false;
         capture = false;
      }
      else
         capture = true;
   }
   else
      capture = false;
// }
}

/*!
 * \brief Get an image from the frame grabber.
 *
 * The image might have been pre-captured (with captureImage()), but
 * if not will be captured here.
 *
 * The images are resized here.
 *
 * \param[out] theImage The image in which to save the grabbed image, if
 * not NULL.
 * \param[out] rawImage The image in which to save the raw image (before any
 * transformation is applied), if not NULL.
 */
void picassoImageGrabber::getImage(imageRGB* theImage, imageRGB* rawImage)
{
   unsigned char* grabbedImage;

   if (connected)
   {
      if (!capture)
         captureImage();

      // Wait for the capture to be finished.
      ArvDWORD isReady;
      PcssRetType ret;
      do
      {
         ret = PcssFunc(grabberHandle, PCSSF_ISCAPT_READY, &isReady);
         if (ret != PCSS_NO_ERROR)
         {
            std::cerr << "Could not talk to the grabber.\n";
            connected = false;
         }
      } while (!isReady);
   }

   // Get the data.
   SPcssCaptureInfo captureInfo;
   PCSSINITVAR(captureInfo);
   if (connected)
   {
      PcssRetType ret = PcssFunc(grabberHandle, PCSSF_GET_CAPTUREINFO,
                                 &captureInfo);
      if (ret != PCSS_NO_ERROR)
      {
         std::cerr << "Could not talk to the grabber.\n";
         connected = false;
      }
      else
      {
         grabbedImage = static_cast<unsigned char*>(captureInfo.lin_mem);
      }
   }

   // Play with the data.
   if (connected)
   {
//       std::cout << "Capture info: " << captureInfo.w << "x" << captureInfo.h
//          << std::endl;
      assert((captureInfo.w == imgMap->sourceWidth())
             && (captureInfo.h == imgMap->sourceHeight()));

      unsigned long index;
      if (theImage)
      {
         theImage->setSize(imgMap->targetWidth(),
                           imgMap->targetHeight());
         std::vector<imageMapping::coordsPair>::iterator iter;
         for (iter = imgMap->map.begin();
              iter < imgMap->map.end();
              ++iter)
         {
            index = iter->sY * captureInfo.pitch + iter->sX * 3;
            theImage->setPixelBGR(iter->tX, iter->tY,
                                  &grabbedImage[index]);
         }
      }
      if (rawImage)
      {
         rawImage->setSize(imgMap->sourceWidth(),
                           imgMap->sourceHeight());
         for (unsigned short indY = 0;
              indY < imgMap->sourceHeight();
              ++indY)
         {
            index = indY * captureInfo.pitch;
            for (unsigned short indX = 0;
                 indX < imgMap->sourceWidth();
                 ++indX)
            {
               rawImage->setPixelBGR(indX, indY,
                                     &grabbedImage[index]);
               index += 3;
            }
         }
      }

      capture = false;    // We have used the captured image.
   }
}

/*!
 * \brief Specify the input rectangle (viewport on the grabbed image).
 * \param[in] x
 * \param[in] y Coordinates of the top-left corner of the rectangle.
 * \param[in] w
 * \param[in] h Size of the rectangle.
 * \return true is successful, false otherwise.
 * \sa useInputRectangle().
 */
bool picassoImageGrabber::setInputRectangle(unsigned short x,
                                            unsigned short y,
                                            unsigned short w,
                                            unsigned short h)
{
   inputX = x;
   inputY = y;
   inputWidth = w;
   inputHeight = h;
   return(useInputRectangle());
}

/*!
 * \brief Use the input rectangle currently set.
 * \return true if successful, false otherwise.
 * \sa setInputRectangle().
 */
bool picassoImageGrabber::useInputRectangle()
{
   PcssVFields vfld;
   PcssCaptFields cfld;
   long inputH = inputHeight;
   if (inputH >= 2 * imgMap->sourceHeight())
   {
      inputH /= 2;
//       vfld = EVENODD_FIELD_SINGLE_PITCH;
//       cfld = CAPT_ALL_FIELDS;
      vfld = ODD_FIELD;
      cfld = CAPT_ODD_FIELD;

   }
   else
   {
      vfld = EVENODD_FIELD;
      cfld = CAPT_ODDEVEN_FRAME;
   }
   PcssRetType ret = PcssFunc(grabberHandle, PCSSF_SET_SOURCE_W, &inputWidth);
   if (ret != PCSS_NO_ERROR)
   {
      std::cerr << "Could not set source width to " << inputWidth << std::endl;
      return(false);
   }
   else
      ret = PcssFunc(grabberHandle, PCSSF_SET_SOURCE_H, &inputH);
   if (ret != PCSS_NO_ERROR)
   {
      std::cerr << "Could not set source height to " << inputH << std::endl;
      return(false);
   }
   else
      ret = PcssFunc(grabberHandle, PCSSF_SET_SOURCE_X, &inputX);
   if (ret != PCSS_NO_ERROR)
   {
      std::cerr << "Could not set source x position to " << inputX << std::endl;
      return(false);
   }
   else
      ret = PcssFunc(grabberHandle, PCSSF_SET_SOURCE_Y, &inputY);
   long param = 0;
   if (ret != PCSS_NO_ERROR)
   {
      std::cerr << "Could not set source y position to " << inputY << std::endl;
      return(false);
   }
   else
      ret = PcssFunc(grabberHandle, PCSSF_SET_DEST_X, &param);
   if (ret != PCSS_NO_ERROR)
   {
      std::cerr << "Could not set destination x position to " << param
         << std::endl;
      return(false);
   }
   else
      ret = PcssFunc(grabberHandle, PCSSF_SET_DEST_Y, &param);
   if (ret != PCSS_NO_ERROR)
   {
      std::cerr << "Could not set destination y position to " << param
         << std::endl;
      return(false);
   }
   else
   {
      param = imgMap->sourceWidth();
      ret = PcssFunc(grabberHandle, PCSSF_SET_DEST_W, &param);
   }
   if (ret != PCSS_NO_ERROR)
   {
      std::cerr << "Could not set destination width to " << param << std::endl;
      return(false);
   }
   else
   {
      param = imgMap->sourceHeight();
      ret = PcssFunc(grabberHandle, PCSSF_SET_DEST_H, &param);
   }
   if (ret != PCSS_NO_ERROR)
   {
      std::cerr << "Could not set destination height to " << param << std::endl;
      return(false);
   }
   else
   {
      if (initFrameGrabberCapture(vfld, cfld))
      {
         std::cout << "Capture rectangle: " << inputWidth << "x" << inputHeight
            << " at (" << inputX << "," << inputY << ")\n";
         return(true);
      }
      else
         return(false);
   }
}

void picassoImageGrabber::moveRight()
{
   static unsigned int delta = 2;
   if (inputX >= delta)
   {
      inputX -= delta;
      useInputRectangle();
   }
}

void picassoImageGrabber::moveLeft()
{
   static unsigned int delta = 2;
   inputX += delta;
   useInputRectangle();
}

void picassoImageGrabber::moveUp()
{
   inputY += 1;
   useInputRectangle();
}

void picassoImageGrabber::moveDown()
{
   if (inputY >= 1)
   {
      inputY -= 1;
      useInputRectangle();
   }
}

void picassoImageGrabber::increaseWidth()
{
   if (inputWidth >= 1)
   {
      inputWidth -= 1;
      useInputRectangle();
   }
}

void picassoImageGrabber::decreaseWidth()
{
   inputWidth += 1;
   useInputRectangle();
}

void picassoImageGrabber::increaseHeight()
{
   if (inputHeight >= 1)
   {
      inputHeight -= 1;
      useInputRectangle();
   }
}

void picassoImageGrabber::decreaseHeight()
{
   inputHeight += 1;
   useInputRectangle();
}

/*!
 * \brief Initialise the grabber.
 * \param[in] imgWidth
 * \param[in] imgHeight The width and height of the image.
 * \param[in] upsideDown Set to true if the camera is upside-down.
 * \param[in] flip Set to true to horizontally flip the images.
 * \param[in] grabIndex The index of the Picasso grabber to use.
 * \param[in] portNumber The number of the port on the device.
 * \param[in] posX
 * \param[in] posY The position of the viewport of the grabbed part.
 * \param[in] width
 * \param[in] height The width and height of the grabbed area.
 */
void picassoImageGrabber::init(unsigned short imgWidth,
                               unsigned short imgHeight,
                               bool upsideDown,
                               bool flip,
                               int grabIndex,
                               int portNumber,
                               unsigned short posX,
                               unsigned short posY,
                               unsigned short width,
                               unsigned short height)
{
   bool cont = true;

   if (boardIndex == -1)
   {
      boardIndex = grabIndex;
      numberOfInstances++;
      if (numberOfInstances > 1)
      {
         std::cerr
            << "-------------------- WARNING ------------------------"
            << "More than one instance (" << numberOfInstances
            << ") of the class imalib::picassoImageGrabber created!\n"
            << "You will have to call imalib::picassoImageGrabber::setInputPort()\n"
            << "on each of them before each image grabbing (otherwise you will\n"
            << "only grab from the same input port).\n"
            << "-------------------- WARNING ------------------------";
      }
      
   }
   else
   {
      cont = false;
      std::cerr << "Only one board can be opened at a time.\n"
         << "Board " << boardIndex << " is already opened.\n";
   }

   connected = false;

   // We can't have the image larger than the grabbed area.
   if (imgWidth > width)
      imgWidth = width;
   if (imgHeight > height)
      imgHeight = height;

   setParameters(imgWidth, imgHeight, upsideDown, flip);

   PcssRetType ret = PcssOpenLibrary();
   if (ret != PCSS_NO_ERROR)
   {
      std::cerr << "Error while opening the Picasso library.\n";
      cont = false;
   }

   if (cont)
   {
      ArvSDWORD version;
      ret = PcssGetVersion(&version);
      if (ret != PCSS_NO_ERROR)
      {
         std::cerr << "Unable to get the Picasso version.\n";
         cont = false;
      }
      else if (version < PCSSA_VERSION)
      {
         std::cerr << "Error: incorrect Picasso library version: " << version
            << ".\n"
            << "Application compiled with version: " << PCSSA_VERSION << ".\n";
         cont = false;
      }
   }

   if (cont)
   {
      short numCards;
      ret = GetNumPcssCards(&numCards);
      if (ret != PCSS_NO_ERROR)
      {
         std::cerr << "Unable to get the number of Picasso boards.\n";
         cont = false;
      }
      else if (numCards < boardIndex)
      {
         std::cerr << "The specified grabber index (" << boardIndex
            << ") does not exist.\n";
         cont = false;
      }
   }

   PcssBoardType boardType;
   PcssBusSystem busSystem;
   if (cont)
   {
      ret = GetPcssBoardType(boardIndex, &boardType, &busSystem);
      if (ret != PCSS_NO_ERROR)
      {
         std::cerr << "Unable to get the Picasso board type.\n";
         cont = false;
      }
      else if (boardType == BOARDT_UNKNOWN)
      {
         std::cerr << "Unknown Picasso board type.\n";
         cont = false;
      }
   }

   if (cont)
   {
      ret = PcssOpen(boardIndex, boardType, &grabberHandle);
      if (ret != PCSS_NO_ERROR)
      {
         std::cerr << "Unable to open the Picasso board.\n";
         cont = false;
         boardIndex = -1;
      }
      else
      {
         connected = true;
      }
   }

   //
   // Setting parameters.
   //
   // Timing.
   long param = PCSS_VTIME_PAL;
   if (cont)
   {
      ret = PcssFunc(grabberHandle, PCSSF_SET_VIDEOTIMETYPE, &param);
      if (ret != PCSS_NO_ERROR)
      {
         std::cerr << "Could not set timing to PAL.\n";
         cont = false;
      }
   }

   // Port.
   if (cont)
   {
      cont = setInputPort(portNumber);
   }

   // Colour
   if (cont)
   {
      param = PCSS_RGB24_888; // RGB, 24 bit per pixel.
      ret = PcssFunc(grabberHandle, PCSSF_SET_VIDEOFORMAT, &param);
      if (ret != PCSS_NO_ERROR)
      {
         std::cerr << "Could not set the colour format to RGB 24.\n";
         cont = false;
      }
   }

   // Interlaced video
   if (cont)
   {
      param = EVENODD_FIELD;
      //       param = EVENODD_FIELD_SINGLE_PITCH;
      ret = PcssFunc(grabberHandle, PCSSF_SET_FIELDS, &param);
      if (ret != PCSS_NO_ERROR)
      {
         std::cerr << "Could not set the fields to interlaced mode.\n";
         cont = false;
      }
   }

   // Image sizes.
   if (cont)
   {
      cont = setInputRectangle(posX, posY, width, height);
      // This also initialises the frame grabber.
   }

   if (!cont && connected)
   {
      // Something wrong happened after connection. We need to close.
      ret = PcssClose(grabberHandle);
      if (ret != PCSS_NO_ERROR)
      {
         std::cerr << "Could not close the Picasso board.\n";
      }
      else
      {
         connected = false;
      }
   }
}


/*!
 * \brief Low level initialisation of the frame grabber.
 *
 * This is to be called after all changes in the acquisition parameters
 * and is called in useInputRectangle().
 * \param[in] vfld Fields selected by the grabber (see the picasso
 * documentation).  This should be EVENODD_FIELD if a full height is
 * captured (interlaced image) or EVENODD_FIELD_SINGLE_PITCH if half of
 * the height is captured (only even or odd fields).
 * \param[in] cfld Fields captured. This should be CAPT_ODDEVEN_FRAME if a
 * full height is captured (interlaced image) or CAPT_ALL_FIELDS if half of
 * the height is captured (only even or odd fields).
 * \return true if all went well, false otherwise.
 */
bool picassoImageGrabber::initFrameGrabberCapture(PcssVFields vfld,
                                                  PcssCaptFields cfld)
{
   bool cont = true;
   PcssRetType ret;

   ret = PcssFunc(grabberHandle, PCSSF_SET_FIELDS, &vfld);
   if (ret != PCSS_NO_ERROR)
   {
      std::cerr
         << "Could not initialise the capture with the Picasso grabber:\n";
      char buff[128];
      PcssEnumToStr(PCSSRETTYPE, ret, buff, 128);
      std::cerr << buff << std::endl;
      cont = false;
   }
   else
   {
      SPcssInitCapture capinit;
      PCSSINITVAR(capinit);
      capinit.num_capt = 1;
      capinit.flds_before = 0; // 0 fields before starting.
      capinit.trig = CAPTRIG_ON_CALL; // Start immediatly.
      capinit.capt_field = cfld; // Capture both fields.
      capinit.capt_overlap = PCSS_CAPT_OPT_OVERLAP; // Non-blocking
      capinit.use_cb = 0; // Don't use a callback.
      capinit.cbfunc = NULL;
      capinit.cbparm = NULL;

      ret = PcssFunc(grabberHandle, PCSSF_INIT_CAPTURE, &capinit);
      if (ret != PCSS_NO_ERROR)
      {
         std::cerr
            << "Could not initialise the capture with the Picasso grabber:\n";
         char buff[128];
         PcssEnumToStr(PCSSRETTYPE, ret, buff, 128);
         std::cerr << buff << std::endl;
         cont = false;
      }
   }

   return(cont);
}

} // namespace imalib.
