/*
 * imalib
 * 
 * Copyright 2017 Aberystwyth University
 * 
 * This library and program are free software: you can redistribute them
 * and/or modify them under the terms of the GNU General Public License as 
 * published by the Free Software Foundation, either version 3 of the License,
 * or (at your option) any later version.
 * 
 * The program and library are distributed in the hope that they will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*!
 * \file picassoImageGrabber.cpp
 * \brief An image grabber for a picasso frame grabber using the V4L driver.
 * Implementation.
 *
 * \author Fred Labrosse <ffl@aber.ac.uk>
 * \copyright 2017 Aberystwyth University
 */

#include "imalib/imageRGB.h"
#include "imalib/picassoV4LImageGrabber.h"

#include <sys/mman.h>
#include <math.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <stdio.h>
#include <unistd.h>

#include <fstream>
#include <iostream>

#include <boost/program_options.hpp>
namespace po = boost::program_options;

namespace imalib
{

/*----------------------------------------------------------------*\
Class picassoImageGrabberV4L
\*----------------------------------------------------------------*/

const unsigned int DEFAULT_FORMAT = VIDEO_PALETTE_RGB24;
// #define DEFAULT_SYNC SYNC_EXTERN
const unsigned short DEFAULT_NORM = PAL;

const unsigned short DEFAULT_BRIGHTNESS = 32768;
const unsigned short DEFAULT_CONTRAST = 32768;
const unsigned short DEFAULT_SATURATION = 32768;
const unsigned short DEFAULT_HUE = 32768;

const unsigned short MAXINPUTS = 100;

/*!
 * \brief Constructor.
 *
 * \param[in] imgWidth
 * \param[in] imgHeight The width and height of the image.
 * \param[in] upsideDown Set to true if the camera is upside-down.
 * \param[in] flip Set to true to horizontally flip the images.
 * \param[in] devNumber The number of the video device (/dev/video?).
 * \param[in] devPort The number of the port on the device.
 * \param[in] posX
 * \param[in] posY The position of the viewport of the grabbed part.
 * \param[in] width
 * \param[in] height The width and height of the grabbed area.
 */
picassoV4LImageGrabber::picassoV4LImageGrabber(unsigned short imgWidth,
                                               unsigned short imgHeight,
                                               bool upsideDown,
                                               bool flip,
                                               int devNumber, int devPort,
                                               unsigned short posX,
                                               unsigned short posY,
                                               unsigned short width,
                                               unsigned short height)
{
   init(imgWidth, imgHeight,
        upsideDown,
        flip,
        devNumber, devPort,
        posX, posY,
        width, height);
}

/*!
 * \brief Constructor.
 * \param[in] fileName Name of the configuration file.
 */
picassoV4LImageGrabber::picassoV4LImageGrabber(const std::string& fileName)
{
   // Let's extract the options from the configuration file.
   po::options_description optionsDesc("picassoV4LImageGrabber options");
   int vidDev = 0;
   int devPo = 0;
   int poX = 0;
   int poY = 0;
   int soWidth = IMAGE_DEFAULT_INPUTWIDTH;
   int soHeight = IMAGE_DEFAULT_INPUTHEIGHT;
   int imWidth = 200;
   int imHeight = 200;
   bool upDown = false;
   bool fl = false;
   optionsDesc.add_options()
      ("videoDevice", po::value<int>(&vidDev)->default_value(vidDev), "")
      ("devicePort", po::value<int>(&devPo)->default_value(devPo), "")
      ("posX", po::value<int>(&poX)->default_value(poX), "")
      ("posY", po::value<int>(&poY)->default_value(poY), "")
      ("sourceWidth",
       po::value<int>(&soWidth)->default_value(soWidth), "")
      ("sourceHeight",
       po::value<int>(&soHeight)->default_value(soHeight), "")
      ("imageWidth", po::value<int>(&imWidth)->default_value(imWidth), "")
      ("imageHeight", po::value<int>(&imHeight)->default_value(imHeight), "")
      ("upsideDown", po::value<bool>(&upDown)->default_value(upDown), "")
      ("flip", po::value<bool>(&fl)->default_value(fl), "")
      ;

   po::variables_map vm;
   std::ifstream configFile(fileName.c_str());
   if (configFile.is_open())
   {
      po::store(po::parse_config_file(configFile, optionsDesc), vm);
      po::notify(vm);
      configFile.close();
   }
   else
   {
      std::cerr << "picassoV4LImageGrabber::picassoV4LImageGrabber():\n"
         << "WARNING: could not open configuration file " << fileName
         << "!\n"
         << "\tUsing defaults.\n";
   }

   init(imWidth, imHeight,
        upDown,
        fl,
        vidDev, devPo,
        poX, poY,
        soWidth, soHeight);
}


/*!
 * \brief Destructor.
 */
picassoV4LImageGrabber::~picassoV4LImageGrabber()
{
   if (fd)
      close(fd);
}

/*!
 * \brief Display on the standard output some information about the
 * source of the images.
 */
void picassoV4LImageGrabber::printInfoAboutSource()
{
   char deviceName[255];
   video_capability vcap;
   video_channel vchan;
   bool wasOpen = false;

   std::cout << "----------------------------------------------\n"
      << "Image source:\n";

   // Open the video device
   sprintf(deviceName, "/dev/video%d", deviceNumber);
   if (fd <= 0)
   {
      wasOpen = false;
      fd = open(deviceName, O_RDWR);
      if (fd <= 0)
      {
         perror("Open");
         connected = false;
      }
   }
   else
      wasOpen = true;
   if (fd > 0)
   {
      // Retrieve the number of input ports
      if(ioctl(fd, VIDIOCGCAP, &vcap) < 0)
      {
         perror("VIDIOCGCAP");
         connected = false;
      }
      else
      {
         std::cout << "\tgrabber: " << vcap.name << std::endl;

         int numInputs = vcap.channels;
         char inputNames[MAXINPUTS][25];

         //Retrieve the name of each input
         for (int i = 0 ; i < numInputs ; i++)
         {
            vchan.channel = i;
            if(ioctl(fd, VIDIOCGCHAN, &vchan) < 0)
            {
               perror("VIDIOCGCHAN");
               connected = false;
            }
            else
            {
               strcpy(inputNames[i], vchan.name);
            }
         }

         std::cout << "\tvideo port: "
            << inputNames[vChannel.channel]
            << std::endl;
      }
      if (!wasOpen)
         close(fd);
   }
//    std::cout << "\timage size: " << imageWidth << "x" << imageHeight
//       << std::endl;
//    if (_unwrap)
//    {
//       std::cout << "Images are unwrapped:\n"
//          << "\tsize:               " << imgMap->targetWidth << "x"
//          << imgMap->targetHeight << std::endl
//          << "\tinner radius:       " << _innerRadius << std::endl
//          << "\tangular resolution: " << _angularRes << std::endl
//          << "\tpower:              " << _unwrappingPower << std::endl
//          << "\tdirection:          " << _direction << std::endl;
//    }
//    if (_blur)
//    {
//       std::cout << "Images are blurred:\n"
//          << "\tblur size: " << _blurWidth << "x" << _blurHeight
//          << std::endl;
//    }
//    if (_upsideDown)
//    {
//       std::cout << "Images are upside down\n";
//    }
   printBasicInfoAboutSource();
   std::cout << "----------------------------------------------\n";
}


/*!
 * \brief Capture the image from a grabber.
 *
 * This is there because the capture can take a while.  We thus first
 * capture and then use (with getImage()).
 */
void picassoV4LImageGrabber::captureImage()
{
   if (connected)
   {
      if (ioctl(fd, VIDIOCMCAPTURE, &vMmap) < 0)
      {
         perror("VIDIOCMCAPTURE");
         connected = false;
         capture = false;
      }
      else
         capture = true;
   }
}

/*!
 * \brief Get an image from the frame grabber.
 *
 * The image might have been pre-captured (with captureImage()), but
 * if not will be captured here.
 *
 * The images are resized here.
 *
 * \param[out] theImage The image in which to save the grabbed image, if
 * not NULL.
 * \param[out] rawImage The image in which to save the raw image (before any
 * transformation is applied), if not NULL.
 */
void picassoV4LImageGrabber::getImage(imageRGB* theImage, imageRGB* rawImage)
{
   if (connected)
   {
      if (!capture)
         captureImage();

      // Wait for the capture to be finished.
      if (ioctl(fd, VIDIOCSYNC, &vMmap.frame) < 0)
      {
         perror("VIDIOCSYNC");
         connected = false;
      }
      else
      {
         unsigned long index;
         if (theImage)
         {
            theImage->setSize(imgMap->targetWidth,
                              imgMap->targetHeight);
            std::vector<imageMapping::coordsPair>::iterator iter;
            for (iter = imgMap->map.begin();
                 iter < imgMap->map.end();
                 ++iter)
            {
               index = (iter->sY * imgMap->sourceWidth + iter->sX) * 3;
               theImage->setPixelBGR(iter->tX, iter->tY,
                                     &grabbedImage[index]);
            }
         }
         if (rawImage)
         {
            rawImage->setSize(imgMap->sourceWidth,
                              imgMap->sourceHeight);
            for (unsigned short indX = 0;
                 indX < imgMap->sourceWidth;
                 ++indX)
               for (unsigned short indY = 0;
                    indY < imgMap->sourceHeight;
                    ++indY)
               {
                  index = (indY * imgMap->sourceWidth + indX) * 3;
                  rawImage->setPixelBGR(indX, indY,
                                        &grabbedImage[index]);
               }
         }
         capture = false;    // We have used the captured image.
      }
   }
}


void picassoV4LImageGrabber::moveRight()
{
   static unsigned int delta = 2; // Why 2?
   if (vCapture.x >= delta)
   {
      vCapture.x = vCapture.x - delta;
      useInputRectangle();
   }
}

void picassoV4LImageGrabber::moveLeft()
{
   static unsigned int delta = 2; // Why 2?
   vCapture.x = vCapture.x + delta;
   useInputRectangle();
}

void picassoV4LImageGrabber::moveUp()
{
   vCapture.y = vCapture.y + 1;
   useInputRectangle();
}

void picassoV4LImageGrabber::moveDown()
{
   if (vCapture.y >= 1)
   {
      vCapture.y = vCapture.y - 1;
      useInputRectangle();
   }
}

void picassoV4LImageGrabber::increaseWidth()
{
   if (vCapture.width >= 1)
   {
      vCapture.width = vCapture.width - 1;
      useInputRectangle();
   }
}

void picassoV4LImageGrabber::decreaseWidth()
{
   vCapture.width = vCapture.width + 1;
   useInputRectangle();
}

void picassoV4LImageGrabber::increaseHeight()
{
   if (vCapture.height >= 1)
   {
      vCapture.height = vCapture.height - 1;
      useInputRectangle();
   }
}

void picassoV4LImageGrabber::decreaseHeight()
{
   vCapture.height = vCapture.height + 1;
   useInputRectangle();
}

bool picassoV4LImageGrabber::setInputRectangle(int x, int y, int w, int h)
{
   vCapture.x = x;
   vCapture.y = y;
   vCapture.width = w;
   vCapture.height = h / 2;     // See "capture" example from ARVOO.
   return(useInputRectangle());
}

bool picassoV4LImageGrabber::useInputRectangle()
{
   if (ioctl(fd, VIDIOCSCAPTURE, &vCapture) < 0)
   {
      perror("VIDIOCSCAPTURE");
      return(false);
   }
   else
   {
      std::cout << vCapture.width << "x" << vCapture.height * 2
         << " at (" << vCapture.x << "," << vCapture.y << ")\n";
      return(true);
   }
}

bool picassoV4LImageGrabber::setPictureProperties(unsigned short b,
   unsigned short sa,
   unsigned short cn,
   unsigned short h)
{
   //Set brightness, contrast, saturation and hue
   vPicture.brightness =  b;
   vPicture.colour = sa;
   vPicture.contrast = cn;
   vPicture.hue = h;
   return(usePictureProperties());
}

bool picassoV4LImageGrabber::usePictureProperties()
{
   if (ioctl(fd, VIDIOCSPICT, &vPicture) < 0 )
   {
      perror("VIDIOCSPICT");
      return(false);
   }
   return(true);
}

/*!
 * \brief Initialise the grabber.
 * \param[in] imgWidth
 * \param[in] imgHeight The width and height of the image.
 * \param[in] upsideDown Set to true if the camera is upside-down.
 * \param[in] flip Set to true to horizontally flip the images.
 * \param[in] videoDevice The number of the video device (/dev/video?).
 * \param[in] devicePort The number of the port on the device.
 * \param[in] posX
 * \param[in] posY The position of the viewport of the grabbed part.
 * \param[in] width
 * \param[in] height The width and height of the grabbed area.
 */
void picassoV4LImageGrabber::init(unsigned short imgWidth,
                                  unsigned short imgHeight,
                                  bool upsideDown,
                                  bool flip,
                                  int videoDevice,
                                  int devicePort,
                                  unsigned short posX,
                                  unsigned short posY,
                                  unsigned short width,
                                  unsigned short height)
{
   std::cout << "Warning: the class picassoV4LImageGrabber is deprecated.\n"
      << "consider using the class picassoImageGrabber instead.\n";

   deviceNumber = videoDevice;
   setParameters(imgWidth, imgHeight, upsideDown, flip);

   //set input channel number and norm (PAL/NTSC etc)
   vChannel.channel = devicePort;
   vChannel.norm = DEFAULT_NORM;

   // Set output properties
   vMmap.frame = 0;
   vMmap.width = imgMap->sourceWidth;
   vMmap.height = imgMap->sourceHeight;
   vMmap.format = DEFAULT_FORMAT;

   // Open the video device
   char deviceName[255];
   std::cout << "Opening /dev/video" << deviceNumber << std::endl;
   sprintf(deviceName, "/dev/video%d", deviceNumber);
   fd = open(deviceName, O_RDWR);
   if (fd <= 0)
   {
      perror("Open");
      connected = false;
   }
   else
   {
      //Set brightness, contrast, saturation and hue
      if (setPictureProperties(DEFAULT_BRIGHTNESS,
                               DEFAULT_SATURATION,
                               DEFAULT_CONTRAST,
                               DEFAULT_HUE))
      {
         if (setInputRectangle(posX, posY, width, height))
         {
            if (ioctl(fd, VIDIOCSCHAN, &vChannel) < 0)
            {
               perror("VIDIOCSCHAN");
               connected = false;
            }
            else
            {
               //Get the buffer props
               video_mbuf mbuf;

               if(ioctl(fd, VIDIOCGMBUF, &mbuf) < 0)
               {
                  perror("VIDIOCGMBUF");
                  connected = false;
               }
               else
               {
                  //map the buffer
                  grabbedImage = (unsigned char*)mmap(0, mbuf.size,
                     PROT_READ|PROT_WRITE,
                     MAP_SHARED,
                     fd, 0);
                  if (!grabbedImage)
                  {
                     perror("mmap");
                     connected = false;
                  }
                  else
                  {
                     grabbedImage = grabbedImage + mbuf.offsets[0];
                  }
               }
            }
         }
      }
   }
}

} // namespace imalib.
