/*
 * imalib
 * 
 * Copyright 2017 Aberystwyth University
 * 
 * This library and program are free software: you can redistribute them
 * and/or modify them under the terms of the GNU General Public License as 
 * published by the Free Software Foundation, either version 3 of the License,
 * or (at your option) any later version.
 * 
 * The program and library are distributed in the hope that they will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*!
 * \file pixelSelector.cpp
 * \brief A class to select a subset of all pixels in an image (implementation).
 * 
 * \author Fred Labrosse <ffl@aber.ac.uk>
 * \copyright 2017 Aberystwyth University
 */

#include "imalib/pixelSelector.h"

#include <stdlib.h>
#include <iostream>

namespace imalib
{
/*!
 * \brief Create a selector that keeps all the pixels.
 *
 * One usually does not want to do that (less efficient than just looking at all
 * pixels of the image) but this is useful for debugging.
 *
 * \param[in] width The width if the image.
 * \param[in] height The height of the image.
 */
void pixelSelector::createNoopPixelSelector(unsigned short width,
                                            unsigned short height)
{
   hashTableValid = false;
   topLeftX_ = 0;
   topLeftY_ = 0;
   bottomRightX_ = width - 1;
   bottomRightY_ = height - 1;

   pixels_.clear();

   for (unsigned short col = 0; col <= width; ++col)
   {
      for (unsigned short row =0; row <= height; ++row)
      {
         pixels_.push_back(pixelSelector::coords(col, row));
      }
   }

   createHashTable();
}

/*!
 * \brief Create an area selector.
 * \param[in] pos1X
 * \param[in] pos1Y The position of the top left corner of the area,
 * inclusive.
 * \param[in] pos2X
 * \param[in] pos2Y The position of the bottom right corner of the area,
 * inclusive.
 */
void pixelSelector::createAreaSelector(unsigned short pos1X,
                                       unsigned short pos1Y,
                                       unsigned short pos2X,
                                       unsigned short pos2Y)
{
   hashTableValid = false;
   topLeftX_ = pos1X;
   topLeftY_ = pos1Y;
   bottomRightX_ = pos2X;
   bottomRightY_ = pos2Y;

   pixels_.clear();

   for (unsigned short col = topLeftX_; col <= bottomRightX_; ++col)
   {
      for (unsigned short row = topLeftY_; row <= bottomRightY_; ++row)
      {
         pixels_.push_back(pixelSelector::coords(col, row));
      }
   }

   createHashTable();
}

/*!
 * \brief Create a random selector.
 *
 * A random set of pixels is kept, based on the given probability, in a
 * rectangular area of specified size.
 *
 * \param[in] width The width of the rectangular area.
 * \param[in] height The height of the rectangular area.
 * \param[in] probability The probability of keeping pixels.
 */
void pixelSelector::createRandomPixelSelector(unsigned short width,
                                              unsigned short height,
                                              double probability)
{
   hashTableValid = false;
   topLeftX_ = 0;
   topLeftY_ = 0;
   bottomRightX_ = width - 1;
   bottomRightY_ = height - 1;

   pixels_.clear();

   for (unsigned short col = 0; col <= width; ++col)
   {
      for (unsigned short row = 0; row <= height; ++row)
      {
         if (((rand() / (RAND_MAX + 1.0))) <= probability)
            pixels_.push_back(pixelSelector::coords(col, row));
      }
   }

   createHashTable();
}

/*!
 * \brief Create a random selector.
 *
 * A random set of pixels is kept, based on the given probability, in a
 * rectangular area of specified size and position.
 *
 * \param[in] pos1X
 * \param[in] pos1Y The position of the top left corner of the area,
 * inclusive.
 * \param[in] pos2X
 * \param[in] pos2Y The position of the bottom right corner of the area,
 * inclusive.
 * \param[in] probability The probability of keeping pixels.
 */
void pixelSelector::createRandomPixelSelector(unsigned short pos1X,
                                              unsigned short pos1Y,
                                              unsigned short pos2X,
                                              unsigned short pos2Y,
                                              double probability)
{
   hashTableValid = false;
   topLeftX_ = pos1X;
   topLeftY_ = pos1Y;
   bottomRightX_ = pos2X;
   bottomRightY_ = pos2Y;

   pixels_.clear();

   for (unsigned short col = topLeftX_; col <= bottomRightX_; ++col)
   {
      for (unsigned short row = topLeftY_; row <= bottomRightY_; ++row)
      {
         if (((rand() / (RAND_MAX + 1.0))) <= probability)
            pixels_.push_back(pixelSelector::coords(col, row));
      }
   }

   createHashTable();
}

/*!
 * \brief Create a random selector from the front/back areas of a panoramic
 * image.
 *
 * This assumes that the front of the camera is centred at 3/4 of the width of
 * the panramic image from the left while the back of the camera is centred at
 * 1/4 of the width of the image from the left.
 *
 * \param[in] width The width if the image.
 * \param[in] height The height of the image.
 * \param[in] probability The probability of keeping pixels.
 * \param[in] frontBack The width in pixels of the front and back areas.
 * \param[in] horizon Sets the maximum, exclusive, index of rows to consider,
 * 0 being top and height being bottom (the value is clamped to height).
 * \param[in] frontOnly If true, then only consider the front part of the
 * images.
 */
void
   pixelSelector::
   createRandomFrontBackHorizonPixelSelector(unsigned short width,
                                             unsigned short height,
                                             double probability,
                                             unsigned short frontBack,
                                             unsigned short horizon,
                                             bool frontOnly)
{
   hashTableValid = false;
   if (horizon > height)
      horizon = height;

   topLeftX_ = 0;
   topLeftY_ = 0;
   bottomRightX_ = width - 1;
   bottomRightY_ = height - 1;

   pixels_.clear();

   unsigned short nbCols34 = width * 3/4;
   unsigned short nbCols4 = width / 4;
   unsigned short frontBackWidth2 = frontBack / 2;

   unsigned long start, end;

   for (unsigned short row = 0; row < horizon; ++row)
   {
      // Front
      start = (nbCols34 - frontBackWidth2);
      end = start + frontBack;
      for (unsigned long col = start; col < end; ++col)
      {
         if (((rand() / (RAND_MAX + 1.0))) <= probability)
         {
            pixels_.push_back(pixelSelector::coords(col, row));
         }
      }

      // Back
      if (!frontOnly)
      {
         start = (nbCols4 - frontBackWidth2);
         end = start + frontBack;
         for (unsigned long col = start; col < end; ++col)
         {
            if (((rand() / (RAND_MAX + 1.0))) <= probability)
            {
               pixels_.push_back(pixelSelector::coords(col, row));
            }
         }
      }
   }

   createHashTable();
}

/*!
 * \brief Update the hash table used for efficient searching.
 *
 * This function should to be called at the end of each selector
 * modification.
 */
void pixelSelector::createHashTable()
{
  // Empty the current hash tables.
   hash.clear();

   // Populate the hash tables.
   for (unsigned int ind = 0; ind < pixels_.size(); ++ind)
   {
      addToHashTable(pixels_[ind], ind);
   }
   hashTableValid = true;
}

} // namespace imalib
