/*
 * imalib
 * 
 * Copyright 2017 Aberystwyth University
 * 
 * This library and program are free software: you can redistribute them
 * and/or modify them under the terms of the GNU General Public License as 
 * published by the Free Software Foundation, either version 3 of the License,
 * or (at your option) any later version.
 * 
 * The program and library are distributed in the hope that they will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*!
 * \file receiveImages.cpp
 * \brief Receives images from a socket.
 *
 * It uses imlib and gdk for the visualisation.  This is just an example on how
 * to use the imalib classes.
 *
 * \author Fred Labrosse <ffl@aber.ac.uk>
 * \copyright 2017 Aberystwyth University
 */

#include "imalib/imageReceiver.h"

#include <stdio.h>
#include <stdlib.h>     /* getenv(), etc. */
#include <termios.h>
#include <unistd.h>
#include <sys/time.h>

#include <string>
#include <iostream>
#include <iomanip>

#ifdef WITH_GTK
#   include <gtk/gtk.h>
#   include <gdk-pixbuf/gdk-pixbuf.h>
GtkWidget* window;
int windowWidth = 384;
int windowHeight = 288;
#endif // WITH_GTK

#include <boost/program_options.hpp>
namespace po = boost::program_options;

std::string port;               // The port to use.
std::string host;               // The host from where to get the images.

imalib::imageReceiver* imgReceiv; // The image receiver.
imalib::imageRGB theImage;      // The image received from the server.

struct termios origTermios;

bool statistics;                // If true, then display some statistics
                                // while running.
double lastTime, currentTime;
double framesPerSecond;
bool flip;                      // If true, then horizontally flip the
                                // image.
bool save;                      // If true, then save the images.
bool saveOne = false;           // If true, save one image.
unsigned short savedCounter = 0;

double firstTime;               // Time when it all starts.
unsigned short counter = 0;     // How many frames have been processed.

// bool cont = true;

#ifdef WITH_GTK
bool gtkMainLoopRunning = false;
#endif // WITH_GTK

unsigned short delay;

/*
  \brief Retrieves a timestamp in seconds using gettimeofday() for portable
  timing
  \return The current time since the  Epoch  (see time(2)).
*/
inline double time()
{
        timeval tv;
        gettimeofday(&tv, 0);

        return(tv.tv_sec + static_cast<double>(tv.tv_usec) / 1000000.0);
}

void commandLine(int argc, char** argv)
{
   std::ostringstream helpKeys;
   helpKeys
      << "While running, 'q' will quit, 's' will save one image and\n"
      << "'S' will flip in and out of continuous saving"
#ifdef WITH_GTK
      << ", 'r' will reset the size"
#endif // WITH_GTK
      << ".";

   po::options_description optionsDesc("General options");
   optionsDesc.add_options()
      ("help", "Prints this help message and exits")
      ("host,h", po::value<std::string>(&host)->default_value("localhost"),
       "Name of the host the server is running on")
      ("port,p", po::value<std::string>(&port)->default_value("7777"),
       "Port the server is listening to")
      ("statistics,s", "To print statistics about what is done")
      ("flip,f", "To flip horizontally the images")
      ("delay,d", po::value<unsigned short>(&delay)->default_value(0),
       "Delay between frames (in seconds)");

   po::variables_map vm;
   try
   {
      po::store(po::parse_command_line(argc, argv, optionsDesc), vm);
      po::notify(vm);
   }
   catch (const std::exception& e)
   {
      std::cerr << "Error: " << e.what() << std::endl << std::endl
         << optionsDesc << std::endl;
      exit(0);
   }

   if (vm.count("help"))
   {
      std::cout << optionsDesc << std::endl
          << helpKeys.str() << std::endl;
      exit(1);
   }

   if (vm.count("statistics"))
   {
      statistics = true;
   }
   else
   {
      statistics = false;
   }

   if (vm.count("flip"))
   {
      flip = true;
   }
   else
   {
      flip = false;
   }
}

void saveImage()
{
   static std::string baseName = "grabbedImage_";
   static std::string extension = ".ppm";
   char fileName[25];

   sprintf(fileName, "%s%04d%s", baseName.c_str(), savedCounter,
           extension.c_str());
   theImage.write(fileName);
   savedCounter++;
}

inline void flipSave()
{
   save = !save;
}

inline void doSaveOne()
{
   saveOne = true;
}

void quit()
{
   tcsetattr(fileno(stdin), TCSANOW, &origTermios);

   double currentTime = time();
   double framesPerSecond = static_cast<double>(counter)
      / (currentTime - firstTime);
   if (statistics)
      std::cout << std::endl;
   std::cout << "Received " << counter << " frames ("
             << framesPerSecond << " f/s on average)\n";

   // Let's tell the server we are done
   try
   {
      imgReceiv->tellToFinish();
   }
   catch (const std::exception& e)
   {
      // We don't do anything because we are quiting anyway.
   }

#ifdef WITH_GTK
   if (gtkMainLoopRunning)
      gtk_main_quit();
#endif // WITH_GTK

   exit(0);
}

#ifdef WITH_GTK
void resetWindowSize()
{
   windowWidth = theImage.nbColumns();
   windowHeight = theImage.nbRows();
   gtk_window_resize(GTK_WINDOW(window), windowWidth, windowHeight);
}
#endif // WITH_GTK

bool processKeys()
{
   bool cont = true;

   int key = getchar();
   if (key != EOF)
   {
      if (key == 'q')
         cont = false;
      else if (key == 's')
         doSaveOne();
      else if (key == 'S')
         flipSave();
#ifdef WITH_GTK
      else if (key == 'r')
         resetWindowSize();
#endif // WITH_GTK
//       else
//          std::cout << static_cast<int>(key) << std::endl;
   }

   return(cont);
}

void initNonBlocking()
{
   struct termios newTermios;

   tcgetattr(fileno(stdin), &origTermios);

   tcgetattr(fileno(stdin), &newTermios);

   newTermios.c_cc[VTIME] = 0;
   newTermios.c_cc[VMIN] = 0;
   newTermios.c_lflag &= (~ECHO & ~ICANON);
   tcsetattr(fileno(stdin), TCSANOW, &newTermios);
}

#ifdef WITH_GTK
void resize(GtkWidget* widget, GtkAllocation* size, gpointer ptr)
{
   widget = widget; ptr = ptr;
   windowWidth = size->width;
   windowHeight = size->height;
}
#endif // WITH_GTK

int animate(void* impixmap)
{
#ifdef WITH_GTK
   static GdkPixbuf * pixbuf;
   static GdkPixbuf * prescaled;
   static GdkPixmap * pixmap;
#endif // WITH_GTK

   bool cont = processKeys();

   if (cont)
   {
      // Let's tell the server we want an image.
      try
      {
         imgReceiv->askForImage();
         imgReceiv->receiveImage(&theImage);
      }
      catch (const std::exception& e)
      {
         std::cerr << "Exception: " << e.what() << std::endl;
         cont = false;
      }

      if (cont)
      {
#ifdef WITH_GTK
         pixbuf = gdk_pixbuf_new_from_data(theImage.getActualPixel(0, 0),
                                           GDK_COLORSPACE_RGB, FALSE, 8,
                                           theImage.nbColumns(),
                                           theImage.nbRows(),
                                           theImage.nbColumns() * 3,
                                           NULL, NULL);
         if (!pixbuf)
         {
            std::cerr << "Could not show image\n";
            return(TRUE);
         }
         if (flip)
         {
            prescaled = pixbuf;
            pixbuf= gdk_pixbuf_flip(prescaled, TRUE);
         }
         if ((windowWidth != theImage.nbColumns())
             || (windowHeight != theImage.nbRows()))
         {
            prescaled = pixbuf;
            pixbuf = gdk_pixbuf_scale_simple(prescaled,
                                             windowWidth, windowHeight,
                                             GDK_INTERP_NEAREST);
//                                              GDK_INTERP_BILINEAR);
            g_object_unref(prescaled);
//             gdk_pixbuf_unref(prescaled);
            if (!pixbuf)
            {
               std::cerr << "Cannot rescale to size " << windowWidth
                  << "x" << windowHeight << ".\n";
               return(TRUE);
            }
         }
         gdk_pixbuf_render_pixmap_and_mask(pixbuf, &pixmap, NULL, 1);
         gdk_draw_pixmap(GTK_WIDGET(impixmap)->window,
                         GTK_WIDGET(impixmap)->style->fg_gc[GTK_WIDGET_STATE(GTK_WIDGET(impixmap))],
                         pixmap, 0, 0, 0, 0, windowWidth, windowHeight);
         g_object_unref(pixbuf);
//          gdk_pixbuf_unref(pixbuf);
         g_object_unref(pixmap);
//          gdk_pixmap_unref(pixmap);
#endif // WITH_GTK

         counter ++;
         if (statistics)
         {
            lastTime = currentTime;
            currentTime = time();
            framesPerSecond = 1.0 / (currentTime - lastTime);
            std::cout << "\rFrame: " << std::setw(5) << counter << " ("
               << std::setw(6) << std::setprecision(2) << std::fixed
               << framesPerSecond
               << " f/s)";
            if (save || saveOne)
            {
               std::cout << " (saved image " << savedCounter+1 << ")";
            }
            else
            {
               std::cout << "                              ";
            }
         }

         if (saveOne)
         {
            saveImage();
            saveOne = false;
         }
         else if (save)
            saveImage();
      }

      sleep(delay);
   }

   if (!cont)
   {
      quit();
   }

   return(true);
}


int main(int argc, char** argv)
{
   commandLine(argc, argv);

   initNonBlocking();

   boost::asio::io_service ioService;
   std::cout << "Attempting to connect to " << host << " on port " << port
      << "\n";
   try
   {
      imgReceiv = new imalib::imageReceiver(ioService, host, port);
   }
   catch (const std::exception& e)
   {
      std::cerr << "Exception: " << e.what() << "\n";
      quit();
   }
//    if (!imgReceiv->isConnected())
//    {
//       std::cout << "No server on " << host << " listening to port "
//                 << port << ": quitting.\n";
//       quit();
//    }

   // Let's get the first image (to have a size).
   try
   {
      imgReceiv->askForImage();
      imgReceiv->receiveImage(&theImage);
   }
   catch (const std::exception& e)
   {
      std::cerr << "Exception: " << e.what() << std::endl;
      quit();
   }

#ifdef WITH_GTK
   windowWidth = theImage.nbColumns();
   windowHeight = theImage.nbRows();
   GtkWidget* impixmap;
   gtk_init(&argc, &argv);

   window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
   char* windowTitleImage;
   windowTitleImage = new char[host.length()+1];
   sprintf(windowTitleImage, "%s", host.c_str());
   gtk_window_set_title(GTK_WINDOW(window),
                           windowTitleImage);
   gtk_container_set_border_width(GTK_CONTAINER(window), 2);
   gtk_signal_connect(GTK_OBJECT(window), "delete-event",
                      GTK_SIGNAL_FUNC(quit), NULL);
   impixmap = GTK_WIDGET(gtk_drawing_area_new());
   gtk_widget_set_usize(GTK_WIDGET(impixmap), windowWidth, windowHeight);
   gtk_signal_connect(GTK_OBJECT(impixmap), "size-allocate",
                      GTK_SIGNAL_FUNC(resize), NULL);
   gtk_container_add(GTK_CONTAINER(window), GTK_WIDGET(impixmap));
   gtk_widget_show_all(GTK_WIDGET(window));
   gtk_idle_add(animate, impixmap);
//    gtk_widget_set_usize(GTK_WIDGET(impixmap), 4, 3);
#else // WITH_GTK
   std::cout << "\nGTK+ is not available.\n"
             << "--> Images will not be displayed.\n";
#endif // WITH_GTK

   firstTime = currentTime = time();

#ifdef WITH_GTK
   gtkMainLoopRunning = true;
   gtk_main();
#else // WITH_GTK
   while(true)
   {
      animate(NULL);
   }
#endif // WITH_GTK


   return(0);                   // Even though this will never be reached.
}
