/*
 * imalib
 * 
 * Copyright 2017 Aberystwyth University
 * 
 * This library and program are free software: you can redistribute them
 * and/or modify them under the terms of the GNU General Public License as 
 * published by the Free Software Foundation, either version 3 of the License,
 * or (at your option) any later version.
 * 
 * The program and library are distributed in the hope that they will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*!
 * \file shape.cpp
 * \brief Classes to represent objects having some sort of geometric shape.
 * Implementation.
 * 
 * \author Fred Labrosse
 * \copyright 2017 Aberystwyth University
 */

#include "imalib/shape.h"

#include <cassert>
#include <iostream>
#include <sstream>

namespace imalib
{
//----------------
// Class curve
//----------------

/*!
 * \brief Re-sample a curve.
 *
 * The new curve can be uniformly sampled or not.
 *
 * It is based on code from [\ref Hec94].
 *
 * \param[out] newCurve The re-sampled curve.
 * \param[in] n The number of points on the new curve (must be 2 or
 * more, only verified if compiled without NDEBUG).
 * \param[in] nonUniform If true, then re-sample non-uniformly.
 */
void curve::resample(curve* newCurve,
                     unsigned short n,
                     bool nonUniform)
{
   assert(n >= 2);

   // a current curve, b new curve.
   if (nonUniform)
   {
      long ai, ai1, bi, n1 = n-1;
      double ax, af;
      double divisor = (static_cast<double>(points.size()) - 1.0)
         / (static_cast<double>(n) - 1.0);

      newCurve->points.clear();
      newCurve->points.reserve(n);

      // n-1 first points
      for (bi = 0; bi < n1; bi++)
      {
         ax = static_cast<double>(bi) * divisor;
         ai = static_cast<long>(ax);
         ai1 = ai + 1;
         af = ax - ai;
         newCurve->points.push_back(point(points[ai].x
                                          + af * (points[ai1].x
                                                  - points[ai].x),
                                          points[ai].y
                                          + af * (points[ai1].y
                                                  - points[ai].y)));
      }
      // Last point.  This is because the loop above will do an overflow on
      // points if the last point is done in the loop; I think this was a bug
      // in the original implementation.
      newCurve->points.push_back(point(points[points.size()-1].x,
                                       points[points.size()-1].y));

   }
   else
   {
      unsigned short i, i1;
      double step = 0.0;
      double l;
      double d = 0.0001;     // = 0 + tolerance for roundoff error

      for (i = 0; i < points.size() - 1; ++i)
         step += points[i].distance(points[i+1]);
      step /= n - 1;           // length of each output segment (ideally)

      newCurve->points.clear();
      newCurve->points.reserve(n);

      for (i = 0; i < points.size() - 1; ++i)
      {
         i1 = i + 1;
         l = points[i].distance(points[i1]);
         d += l;
         // d is the remaining length of the line segment from points[i]
         // to points[i+1] that needs to be subdivided into segments of
         // length step
         while (d > 0.0)
         {
            newCurve->points.push_back(point(points[i1].x
                                             - d / l * (points[i1].x
                                                - points[i].x),
                                             points[i1].y
                                             - d / l * (points[i1].y
                                                - points[i].y)));
            d -= step;
         }
      }
      if (newCurve->points.size() != n)
         std::cout << "WARNING: requested " << n << " points, created "
         << newCurve->points.size() << " d = " << d << "\n";
   }
}

/*!
\brief outputs the curve in a string, one point per line.
\return The string.
*/
std::string curve::toString()
{
   std::ostringstream aStream;
   std::vector<point>::iterator iter;
   for (iter = points.begin(); iter != points.end(); ++iter)
   {
      aStream << iter->x << " " << iter->y << std::endl;
   }
   return(aStream.str());
}

}
