/*
 * imalib
 * 
 * Copyright 2017 Aberystwyth University
 * 
 * This library and program are free software: you can redistribute them
 * and/or modify them under the terms of the GNU General Public License as 
 * published by the Free Software Foundation, either version 3 of the License,
 * or (at your option) any later version.
 * 
 * The program and library are distributed in the hope that they will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*!
 * \file test.cpp
 * \brief A program to test different things.  It is not installed by 'make
 * install'.
 *
 * \author Fred Labrosse <ffl@aber.ac.uk>
 * \copyright 2017 Aberystwyth University
 */

#include "imalib/appearance.h"
#include "imalib/distanceImgAppear.h"
#include "imalib/distancePanoramic.h"
// #include "imalib/distancePanoramicQuadtree.h"
#include "imalib/distanceMetric.h"
#include "imalib/imageLab.h"
#include "imalib/imageRGB.h"
#include "imalib/imageYUV.h"
#include "imalib/pixelSelector.h"
#include "imalib/transformation.h"
#include "imalib/distanceRectangular.h"
#include "imalib/shape.h"
#include "imalib/utilities.h"
#ifdef WITH_UEYE
#  include "imalib/ueyeImageGrabber.h"
#endif
#ifdef WITH_DC1394
#  include "imalib/dc1394ImageGrabber.h"
#endif

#include <stdio.h>
#include <stdlib.h>     /* getenv(), etc. */
#include <unistd.h>
#include <sys/time.h>

#include <iostream>
#include <string>
#include <vector>

#include <boost/program_options.hpp>
namespace po = boost::program_options;

std::vector<std::string> fileNames;

// bool testQuadtreeDistance = false;
bool testShift = false;
bool testUeyeFirewire = false;
bool testAppearTransform = false;
bool testFindAppearance = false;
bool testPartialMatch = false;

/*
  \brief Retrieves a timestamp in seconds using gettimeofday() for portable
  timing
  \return The current time since the  Epoch  (see time(2)).
*/
inline double time()
{
   timeval tv;
   gettimeofday(&tv, 0);

   return(tv.tv_sec + static_cast<double>(tv.tv_usec) / 1000000.0);
}

void commandLine(int argc, char** argv)
{
   std::ostringstream usage;
   usage << "Usage:\n" << "\ttest {options} {file names}";

   po::options_description testOptions("Options");

   imalib::imalibOptions generalOptions;
   testOptions.add(*(generalOptions.getOptions()));

   testOptions.add_options()
   ("help", "Prints this help message and exits")
//    ("quadtree", "Test the quadtree-based distance")
   ("shift","Test the shifting of an image")
   ("uEyeFirewire",
    "Test the grabbing of images from uEye and firewire cameras at the same time")
   ("appearTransf", "Test the transformation of an appearance")
   ("findAppearance", "Test the finding of an appearance in an image")
   ("partialMatch", "Test the partial match of an appearance on an image")

   ;

   po::options_description fileNamesOptions("Hidden options");
   fileNamesOptions.add_options()
   ("file", po::value< std::vector<std::string> >(), "List of filenames used")
   ;

   po::options_description cmdlineOptions;
   cmdlineOptions.add(testOptions).add(fileNamesOptions);

   po::positional_options_description p;
   p.add("file", -1);

   po::variables_map vm;
   try
   {
      po::store(po::command_line_parser(argc, argv).
                options(cmdlineOptions).positional(p).run(), vm);
      notify(vm);
   }
   catch (const std::exception& e)
   {
      std::cerr << "Error: " << e.what() << std::endl << std::endl
         << usage.str() << std::endl << std::endl
         << testOptions << std::endl;
      exit(0);
   }

   generalOptions.commandLine(vm);

   if (vm.count("help"))
   {
      std::cout << usage.str() << std::endl << std::endl
         << testOptions << std::endl;
      exit(1);
   }

//    if (vm.count("quadtree"))
//       testQuadtreeDistance = true;

   if (vm.count("shift"))
      testShift = true;

   if (vm.count("uEyeFirewire"))
      testUeyeFirewire = true;

   if (vm.count("appearTransf"))
      testAppearTransform = true;

   if (vm.count("findAppearance"))
      testFindAppearance = true;

   if (vm.count("partialMatch"))
      testPartialMatch = true;

   if (vm.count("file"))
      fileNames = vm["file"].as< std::vector<std::string> >();

}

int main(int argc, char** argv)
{
   commandLine(argc, argv);

//    if (testQuadtreeDistance)
//    {
//       std::cout << "Testing the quadtree-based distance...\n";
// 
//       if (fileNames.size() != 2)
//       {
//          std::cerr << "This test requires two images.\n";
//          exit(1);
//       }
// 
//       imalib::imageRGB imgOne, imgTwo;
//       if (!imgOne.read(fileNames[0]))
//       {
//          std::cerr << "Could not read " << fileNames[0] << std::endl;
//          exit(1);
//       }
//       if (!imgTwo.read(fileNames[1]))
//       {
//          std::cerr << "Could not read " << fileNames[1] << std::endl;
//          exit(1);
//       }
// 
// //       imalib::euclideanMetric<unsigned char, 3> theMetric;
// //       imalib::distancePanoramicQuadtree<unsigned char, 3,
// //                             imalib::euclideanMetric<unsigned char, 3> >
// //                             theDistanceFunction(theMetric, 10.0);
//       imalib::correlationMetric<unsigned char, 3> theMetric;
//       imalib::distancePanoramicQuadtree<unsigned char, 3,
//                             imalib::correlationMetric<unsigned char, 3> >
//                             theDistanceFunction(theMetric, 10.0);
// 
//       theDistanceFunction.setSimilarityThreshold(1000);
// 
//       double dist;
//       dist = theDistanceFunction.distance(imgOne, imgTwo, 0);
//       std::cout << "Distance: " << dist << std::endl;
//    }

   if (testShift)
   {
      std::cout << "Testing the shifting of an image...\n";

      if (fileNames.size() != 1)
      {
         std::cerr << "This test requires one image.\n";
         exit(1);
      }
      imalib::imageRGB anImage, imgTwo;
      if (!anImage.read(fileNames[0]))
      {
         std::cerr << "Could not read " << fileNames[0] << std::endl;
         exit(1);
      }

      anImage.shift(0, &imgTwo);
      imgTwo.write("shiftTest.ppm");
   }

   if (testUeyeFirewire)
   {
      std::cout
        << "Testing the grabbing of images from uEye and firewire cameras...\n";

#if defined(WITH_DC1394) && defined(WITH_UEYE)
      imalib::dc1394ImageGrabber dc1394Grabber(false, // upsidedown
                                               false, // flip
                                               0, // camera number
                                               0, 0, // position
                                               800,
                                               600);
      if (dc1394Grabber.isConnected())
      {
  //       dc1394Grabber.printInfoAboutSource();
         dc1394Grabber.setGain(6000);
      }
      else
      {
         return(0);
      }

      imalib::ueyeImageGrabber ueyeGrabber(false, // upsidedown
                                           false, // flip
                                           0, // camera number
                                           0, 0, // position
                                           800,
                                           600);
      if (ueyeGrabber.isConnected())
      {
//         ueyeGrabber.printInfoAboutSource();
      }
      else
      {
         return(0);
      }

      imalib::imageRGB ueyeImage;
      imalib::imageRGB dc1394Image;

      for (int ind = 0; ind < 100; ind++)
      {
         std::cout << "Ind: " << ind << std::endl;
         std::cout << "\tGrabbing uEye...\n";
         ueyeGrabber.getImage(&ueyeImage);
//       ueyeImage.write("ueyeImage.ppm");
         std::cout << "\tGrabbing dc1394...\n";
         dc1394Grabber.getImage(&dc1394Image);
//       dc1394Image.write("dc1394Image.ppm");
      }
#else
      std::cout << "Not compiled with uEye or dc1394 support.\n";
#endif
   }

   if (testAppearTransform)
   {
      std::cout << "Testing the transformation of an appearance...\n";

      if (fileNames.size() != 1)
      {
         std::cerr << "This test requires one image.\n";
         exit(1);
      }

      imalib::imageRGB anImage;
      if (!anImage.read(fileNames[0]))
      {
         std::cerr << "Could not read " << fileNames[0] << std::endl;
         exit(1);
      }
      imalib::appearance<unsigned char, 3> anAppear(anImage);
      anAppear.recentre();
      imalib::rotation rot;
      double rotParam[1];
      rotParam[0] = M_PI/4;
      rot.setParams(rotParam);
      anAppear.transform(rot);
      anAppear.toImage(&anImage);
      anImage.write("appearTrans.png");
   }

   if (testFindAppearance)
   {
      std::cout << "Testing the finding of an appearance in an image...\n";

      if (fileNames.size() != 2)
      {
         std::cerr << "This test requires two images (appearance and image).\n";
         exit(1);
      }

      imalib::imageRGB imgTarget;
      if (!imgTarget.read(fileNames[0]))
      {
         exit(1);
      }
      imalib::image_ab imgTarget_ab;
      imgTarget_ab.copyFromRGB(imgTarget);
      imalib::appearance<double, 2> anAppear_ab(imgTarget_ab);
//       anAppear_ab.recentre();

      imalib::imageRGB imgMain;
      if (!imgMain.read(fileNames[1]))
      {
         exit(1);
      }
      imalib::image_ab imgMain_ab;
      imgMain_ab.copyFromRGB(imgMain);

//       imalib::euclideanMetric<double, 2> theMetric;
      imalib::distanceImgAppearRectangular<double, 2,
                                      imalib::euclideanMetric<double, 2> >
                                      theDistanceFunction;
//                                       theDistanceFunction(theMetric);

      imalib::translation transf;
      double transParam[2];
      double transParamBest[2];
      double transParamMin[2];
      transParamMin[0] = 0.0;
      transParamMin[1] = 0.0;
      double transParamMax[2];
      transParamMax[0] = imgMain.nbColumns();
      transParamMax[1] = imgMain.nbRows();
      transf.setParamsLimits(transParamMin, transParamMax);
      double distMin = 1000000;
      double xMin = 0;
      double yMin = 0;
      double dist;
      imalib::translation transfMin;

      std::cout << "Exhaustive search...\n";
      for (double x = -anAppear_ab.getBBxMin();
           x < imgMain.nbColumns() - anAppear_ab.getBBxMax();
           x++)
      {
         for (double y = -anAppear_ab.getBByMin();
              y < imgMain.nbRows() - anAppear_ab.getBByMax(); y++)
         {
            std::cout.flush();
            transParam[0] = x;
            transParam[1] = y;
            transf.setParams(transParam);
            dist = theDistanceFunction.distanceNormalised(imgMain_ab,
                                                          anAppear_ab,
                                                          transf);
            if (dist < distMin)
            {
               distMin = dist;
               xMin = x;
               yMin = y;
               transfMin = transf;
               std::cout << "\rBest so far at x: " << x << " y: " << y
                         << " dist: " << dist << "         ";
            }
         }
      }

      std::cout << "\rExhaustive minimum at x: " << xMin << " y: " << yMin
         << " dist: " << distMin << std::endl;

      anAppear_ab.fromImage(imgTarget_ab);
//       anAppear_ab.recentre();
      distMin = 1000000;
      std::cout << "Semi-exhaustive search...\n";
      for (double x = 0; x < imgMain.nbColumns(); x += anAppear_ab.getBBxMax())
      {
         for (double y = 0; y < imgMain.nbRows(); y += anAppear_ab.getBByMax())
         {
            transParam[0] = x;
            transParam[1] = y;
            transf.setParams(transParam);
            dist = theDistanceFunction.distanceMin(imgMain_ab, anAppear_ab,
                                                   &transf, 0.5);
            if (fabs(dist) < distMin)
            {
               distMin = fabs(dist);
               transfMin = transf;
            }
         }
      }

      transfMin.getParams(transParamBest);
      std::cout << "Semi-exhaustive minimum at x: " << transParamBest[0]
         << " y: " << transParamBest[1]
         << " dist: " << distMin << std::endl;
   }

   if (testPartialMatch)
   {
      std::cout << "Testing the partial match of an appearance in an image...\n";

      if (fileNames.size() != 2)
      {
         std::cerr
            << "This test requires two images (appearance and image).\n";
         exit(1);
      }

      imalib::imageRGB imgTarget;
      if (!imgTarget.read(fileNames[0]))
      {
         exit(1);
      }
      imalib::appearance<unsigned char, 3> anAppear_RGB(imgTarget);
            // Whole image used, re-centred.

      imalib::image_ab imgTarget_ab;
      imgTarget_ab.copyFromRGB(imgTarget);
      imalib::appearance<double, 2> anAppear_ab(imgTarget_ab);
            // Whole image used, re-centred.

      imalib::imageRGB imgMain;
      if (!imgMain.read(fileNames[1]))
      {
         exit(1);
      }
      imalib::image_ab imgMain_ab;
      imgMain_ab.copyFromRGB(imgMain);

//       imalib::euclideanRobust<double, 2> robustMetric;
      imalib::distanceImgAppearRectangular<double, 2,
                               imalib::euclideanRobust<double, 2> >
                               robustDistance;
//                                robustDistance(robustMetric);

      imalib::translation transf;
      double transParam[2];
      std::cout << "X translation: ";
      std::cin >> transParam[0];
      std::cout << "Y translation: ";
      std::cin >> transParam[1];
      transf.setParams(transParam);

//       robustMetric.reset();

//       typename imalib::appearancePixels<double, 2>::const_iterator iter_ab;
//       double pixX, pixY;

//       for (iter_ab = anAppear_ab.constPixels().begin();
//            iter_ab != anAppear_ab.constPixels().end();
//            ++iter_ab)
//       {
//          transf.apply(iter_ab->first.getX(), iter_ab->first.getY(), &pixX, &pixY);
//          robustMetric.
//                  comparePixels(imgMain_ab.
//                      getActualPixel(static_cast<unsigned short>(round(pixX)),
//                                     static_cast<unsigned short>(round(pixY))),
//                                iter_ab->second.get(),
//                                &(iter_ab->first));
//       }

//       double dist = robustMetric.distanceNormalised();

      double dist = robustDistance.distance(imgMain_ab, anAppear_ab, transf);

      std::cout << "Robust Euclidean distance: " << dist << std::endl
         << "Number of pixels used: " << robustDistance.getNbPixels()
         << std::endl;

//       imalib::euclideanMetric<double, 2> theMetric;
      imalib::distanceImgAppearRectangular<double, 2,
                                 imalib::euclideanMetric<double, 2> >
                                 theDistance;
//                                  theDistance(theMetric);
      std::cout << "Euclidean distance: "
          << theDistance.distanceNormalised(imgMain_ab, anAppear_ab, transf)
          << std::endl;

      // Update the used pixels of the appearance from the main image and
      // save it to file.
      anAppear_ab.update(imgMain_ab, imgMain, &anAppear_RGB, transf, true);
      imalib::imageRGB updatedTarget;
      anAppear_RGB.toImage(&updatedTarget);
      std::cout << "Writing image updatedTarget.png\n";
      updatedTarget.write("updatedTarget.png");

      // Update all the pixels of the appearance from the main image and save
      // it to file.
      anAppear_ab.update(imgMain_ab, imgMain, &anAppear_RGB, transf);
      anAppear_RGB.toImage(&updatedTarget);
      std::cout << "Writing image updatedWholeTarget.png\n";
      updatedTarget.write("updatedWholeTarget.png");

      // Paint the used pixels in the target image and save it.
      unsigned char red[] = {255, 0, 0};

//       imalib::appearance<double, 2>::appearPixelsUsedIter
//          pixUsed(anAppear_ab);
//       for (; // pixUsed is alreading at begin().
//            pixUsed != anAppear_ab.end();
//            ++pixUsed)
//       {
//          imgTarget.setPixel(round(pixUsed->first.getX() - anAppear_RGB.getBBxMin()),
//                             round(pixUsed->first.getY() - anAppear_RGB.getBByMin()),
//                             red);
//       }
      anAppear_ab.saveUsedPixelsMask(red, &imgTarget);
      std::cout << "Writing image robustPixels.png\n";
      imgTarget.write("robustPixels.png");
   }

   return(0);
}
