/*
 * imalib
 * 
 * Copyright 2017 Aberystwyth University
 * 
 * This library and program are free software: you can redistribute them
 * and/or modify them under the terms of the GNU General Public License as 
 * published by the Free Software Foundation, either version 3 of the License,
 * or (at your option) any later version.
 * 
 * The program and library are distributed in the hope that they will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*!
 * \file ueyeImageGrabber.cpp
 * \brief An image grabber for a uEye USB camera.  Implementation.
 *
 * \author Fred Labrosse <ffl@aber.ac.uk>, with contributions from Suzana
 * Barreto <szb@aber.ac.uk> and Mark Neal <mjn@aber.ac.uk>
 * \copyright 2017 Aberystwyth University
 */

#include "imalib/imageRGB.h"
#include "imalib/ueyeImageGrabber.h"

#include <boost/program_options.hpp>

namespace po = boost::program_options;

namespace imalib
{

/*!
 * \brief Constructor.
 *
 * \param[in] upsideDown Set to true if the camera is upside-down.
 * \param[in] flip Set to true to horizontally flip the images.
 * \param[in] camNumber The number of the camera (0 is the first available).
 * \param[in] posX
 * \param[in] posY The position of the viewport of the grabbed part.
 * \param[in] width
 * \param[in] height The width and height of the grabbed area and of the
 * images.
 */
ueyeImageGrabber::ueyeImageGrabber(bool upsideDown,
                                   bool flip,
                                   int camNumber,
                                   unsigned short posX,
                                   unsigned short posY,
                                   unsigned short width,
                                   unsigned short height)
{
   init(upsideDown,
        flip,
        camNumber,
        posX, posY, width, height);
}

/*!
 * \brief Constructor.
 * \param[in] fileName Name of the configuration file.
 */
ueyeImageGrabber::ueyeImageGrabber(const std::string& fileName)
{
   // Let's extract the options from the configuration file.
   po::options_description optionsDesc("ueyeImageGrabber options");
   int camNum = 0;
   int poX = 0;
   int poY = 0;
   int width = IMAGE_DEFAULT_INPUTWIDTH;
   int height = IMAGE_DEFAULT_INPUTHEIGHT;
   bool upDown = false;
   bool fl = false;
   optionsDesc.add_options()
      ("cameraNumber", po::value<int>(&camNum)->default_value(camNum), "")
      ("posX", po::value<int>(&poX)->default_value(poX), "")
      ("posY", po::value<int>(&poY)->default_value(poY), "")
      ("width",
       po::value<int>(&width)->default_value(width), "")
      ("height",
       po::value<int>(&height)->default_value(height), "")
      ("upsideDown", po::value<bool>(&upDown)->default_value(upDown), "")
      ("flip", po::value<bool>(&fl)->default_value(fl), "")
      ;

   po::variables_map vm;
   std::ifstream configFile(fileName.c_str());
   if (configFile.is_open())
   {
      po::store(po::parse_config_file(configFile, optionsDesc), vm);
      po::notify(vm);
      configFile.close();
   }
   else
   {
      std::cerr << "ueyeImageGrabber::ueyeImageGrabber():\n"
         << "WARNING: could not open configuration file " << fileName
         << "!\n"
         << "\tUsing defaults.\n";
   }

   init(upDown,
        fl,
        camNum,
        poX, poY,
        width, height);
}


/*!
 * \brief Destructor.
 */
ueyeImageGrabber::~ueyeImageGrabber()
{
   // stop live capture
   if(capture){
       is_StopLiveVideo (cameraHandle, IS_WAIT);
   }
   //exit
   INT res = is_ExitCamera(cameraHandle);
   if (res != IS_SUCCESS)
   {
      std::cerr << "ueyeImageGrabber::~ueyeImageGrabber()" << std::endl
         << "Could not close the camera.\n";
   }
}

/*!
 * \brief Display on the standard output some information about the
 * source of the images.
 *
 * \param[in,out] aStream The stream to which the info is added.
 */
void ueyeImageGrabber::printInfoAboutSource(std::ostringstream* aStream)
{
   *aStream << "----------------------------------------------\n"
      << "Image source:\n";

   const std::string funcName = "void ueyeImageGrabber::printInfoAboutSource():";

   if (connected)
   {
      BOARDINFO boardInfo;
      INT resBoard = is_GetCameraInfo(cameraHandle, &boardInfo);
      if (resBoard != IS_SUCCESS)
      {
         std::cerr << funcName << std::endl
            << "Could not get the information about the camera.\n";
      }


      if (resBoard == IS_SUCCESS)
      {
         *aStream << "\tcamera Manufacturer: " << boardInfo.ID << std::endl
            << "\tcamera version: " << boardInfo.Version << std::endl
//             << "\tcamera ID: " << boardInfo.Select << std::endl
            << "\tcamera type: ";
         if (boardInfo.Type == IS_CAMERA_TYPE_UEYE_USB_SE)
            *aStream << "USB SE";
         else if (boardInfo.Type == IS_CAMERA_TYPE_UEYE_USB_ME)
            *aStream << "USB ME";
         else if (boardInfo.Type == IS_CAMERA_TYPE_UEYE_USB_RE)
            *aStream << "USB RE";
         else if (boardInfo.Type == IS_CAMERA_TYPE_UEYE_USB_LE)
            *aStream << "USB LE";
         else if (boardInfo.Type == IS_CAMERA_TYPE_UEYE_ETH_HE)
            *aStream << "GigE HE";
         else if (boardInfo.Type == IS_CAMERA_TYPE_UEYE_ETH_SE)
            *aStream << "GigE SE";
         *aStream << std::endl;
      }

      *aStream << "\tcolour mode: ";
      if (sensorInfo.nColorMode == IS_COLORMODE_BAYER)
         *aStream << "Bayer";
      else if (sensorInfo.nColorMode == IS_COLORMODE_MONOCHROME)
         *aStream << "Monochrome";
      *aStream << std::endl
         << "\tmaximum width: " << sensorInfo.nMaxWidth << std::endl
         << "\tmaximum height: " << sensorInfo.nMaxHeight << std::endl;
   }

   printBasicInfoAboutSource(aStream);
   *aStream << "----------------------------------------------\n";
}


/*!
 * \brief Starts video capture running and putting images in the ring buffer.
 *
 * We start video capture here and the use getImage() to extract a single frame
 *
 */
void ueyeImageGrabber::captureImage()
{
   if (connected)
   {
	//perhaps check which mode the camera is in and then either pass is_dont_wait or is_wait
      INT res = is_CaptureVideo(cameraHandle, IS_WAIT); // in freerun and trigger mode returns when 1st image in memory
      if (res != IS_SUCCESS)
      {
         std::cerr << "void ueyeImageGrabber::captureVideo()" << std::endl
            << "Could not start.\n";
         std::cerr << res << std::endl << "Could not start.\n";
         std::cerr << " bad structure size: " << IS_BAD_STRUCTURE_SIZE << " can't communicate with driver: " << IS_CANT_COMMUNICATE_WITH_DRIVER << " can't open device: " << IS_CANT_OPEN_DEVICE << " capture running: " << IS_CAPTURE_RUNNING << " invalid buffer size: " << IS_INVALID_BUFFER_SIZE << " invlaid camera type: " << IS_INVALID_CAMERA_TYPE << " invalid exposure time: " << IS_INVALID_EXPOSURE_TIME << " invalid camera handle: " << IS_INVALID_CAMERA_HANDLE << " invalid memory pointer: " << IS_INVALID_MEMORY_POINTER << " invalid param: " << IS_INVALID_PARAMETER << " io request failed: " << IS_IO_REQUEST_FAILED << " no active image mem: " << IS_NO_ACTIVE_IMG_MEM << " generic error:  " << IS_NO_SUCCESS << " not calibrated: " << IS_NOT_CALIBRATED << " not supported: " << IS_NOT_SUPPORTED << " out of memory: " << IS_OUT_OF_MEMORY << " success: " << IS_SUCCESS << " timed out: " << IS_TIMED_OUT << " trigger activated: " << IS_TRIGGER_ACTIVATED;
         connected = false;
         capture = false;
      }
      else
      {
         capture = true;
      }
   }
}

/*!
 * \brief Get an image from the frame grabber.
 *
 * The images are usually streaming continuously, but captureImage()
 * may need to be called to start video capture if not already running 
 *
 * The images are resized here.
 *
 * \param[out] theImage The image in which to save the grabbed image, if
 * not NULL.
 * \param[out] rawImage The image in which to save the raw image (before any
 * transformation is applied), if not NULL.
 */
void ueyeImageGrabber::getImage(imageRGB* theImage, imageRGB* rawImage)
{
   const std::string funcName =
      "void ueyeImageGrabber::getImage(imageRGB*, imageRGB*):";
   char* currmem;
   char* lastmem;
   INT pnNum;
   INT res;
//    BOOL captDone = false;
   if (connected)
   {
      if (!capture)
      {
         captureImage();   // Only start video capture if not already running
	 //capture = true; the capture image method already sets capture to true or false and if it failed, capture would be set from false to true here 
      }
    }

   pnNum = 0;
   // find the current point in the ring buffer that is currently active
   res = is_GetActSeqBuf(cameraHandle, &pnNum, &currmem, &lastmem);

   for (int i = 1; i < 5; i++)
   {
      if (lastmem == ringbuffer[i])
      {
         //lock sequence buffer
         res = is_LockSeqBuf (cameraHandle, sequenceBufIDs[i], ringbuffer[i]);
         if (res == IS_SUCCESS)
         {
            //get image
            grabbedImage = ringbuffer[i];
            unsigned long index;
            unsigned char* grabbedImageAlias =
               reinterpret_cast<unsigned char*>(grabbedImage);
            if (theImage)
            {
               theImage->setSize(imgMap->targetWidth(),imgMap->targetHeight()); // setSize is an inherited method from image
               std::vector<imageMapping::coordsPair>::iterator iter;
               for (iter = imgMap->map.begin();iter < imgMap->map.end();++iter)
               {
                  index = (iter->sY * imgMap->sourceWidth() + iter->sX) * 3;
                  theImage->setPixel(iter->tX, iter->tY,&grabbedImageAlias[index]);
               }
//                captDone = true;
            }
            if (rawImage)
            {
               rawImage->setSize(imgMap->sourceWidth(), imgMap->sourceHeight());
               for (unsigned short indX = 0;
                    indX < imgMap->sourceWidth();
                    ++indX)
               {
                  for (unsigned short indY = 0;
                       indY < imgMap->sourceHeight();
                       ++indY)
                  {
                     index = (indY * imgMap->sourceWidth() + indX) * 3;

                     rawImage->setPixel(indX, indY,
                                        &grabbedImageAlias[index]);
                  }
               }
//                captDone = true;
            }
            //unlock sequence buffer
            res = is_UnlockSeqBuf (cameraHandle, sequenceBufIDs[i], ringbuffer[i]);
         }
         else
         {
            std::cout<<"could not lock sequence buffer - did not capture image"<<std::endl;
         }
      }
   }
   
  /* // Take the previous ring buffer entry to avoid "jitter"
   if (pnNum > 1) {
      grabbedImage = ringbuffer[pnNum-1];
   std::cout<<"the grabbed image came from is \t"<<(pnNum -1)<<std::endl;
   }
   else{
      grabbedImage = ringbuffer[4];
      std::cout<<"the grabbed image came from ringbuffer 4"<<std::endl;
   }

   if (res != IS_SUCCESS)
   {
      std::cerr << funcName << std::endl
         << "Something bad happened during the transfer of the image.\n";
      connected = false;
   }

   if (connected)
   {
      unsigned long index;
      unsigned char* grabbedImageAlias =
            reinterpret_cast<unsigned char*>(grabbedImage);
      if (theImage)
      {
         theImage->setSize(imgMap->targetWidth(), imgMap->targetHeight());
         std::vector<imageMapping::coordsPair>::iterator iter;
         for (iter = imgMap->map.begin();
              iter < imgMap->map.end();
              ++iter)
         {
            index = (iter->sY * imgMap->sourceWidth + iter->sX) * 3;
            //index = (iter->sY * 1600 + iter->sX) * 3;

            theImage->setPixel(iter->tX, iter->tY,
                                  &grabbedImageAlias[index]);
         }
      }
      if (rawImage)
      {
         rawImage->setSize(imgMap->sourceWidth(), imgMap->sourceHeight());
         for (unsigned short indX = 0;
              indX < imgMap->sourceWidth();
              ++indX)
         {
            for (unsigned short indY = 0;
                 indY < imgMap->sourceHeight();
                 ++indY)
            {
               index = (indY * imgMap->sourceWidth() + indX) * 3;

               rawImage->setPixel(indX, indY,
                                     &grabbedImageAlias[index]);
            }
         }
      }
   }*/
}


// This function will break the ring buffer and stop the video capture from
// working. DO NOT USE. Needs rewriting
void ueyeImageGrabber::moveRight()
{
   if (inputX >= 2)
   {
      inputX -= 2;
      if (!useInputRectangle())
      {
         connected = false;
      }
   }
}

// This function will break the ring buffer and stop the video capture from
// working. DO NOT USE. Needs rewriting
void ueyeImageGrabber::moveLeft()
{
   inputX += 2;
   if (!useInputRectangle())
   {
      connected = false;
   }
}

// This function will break the ring buffer and stop the video capture from
// working. DO NOT USE. Needs rewriting
void ueyeImageGrabber::moveUp()
{
   inputY += 2;
   if (!useInputRectangle())
   {
      connected = false;
   }
}

// This function will break the ring buffer and stop the video capture from
// working. DO NOT USE. Needs rewriting
void ueyeImageGrabber::moveDown()
{
   if (inputY >= 2)
   {
      inputY -= 2;
      if (!useInputRectangle())
      {
         connected = false;
      }
   }
}

// This function will break the ring buffer and stop the video capture from
// working. DO NOT USE. Needs rewriting
void ueyeImageGrabber::increaseWidth()
{
   if (imageWidth >= 2)
   {
      imageWidth -= 2;
      if (!useInputRectangle())
      {
         connected = false;
      }
   }
}

// This function will break the ring buffer and stop the video capture from
// working. DO NOT USE. Needs rewriting
void ueyeImageGrabber::decreaseWidth()
{
   imageWidth += 2;
   if (!useInputRectangle())
   {
      connected = false;
   }
}

// This function will break the ring buffer and stop the video capture from
// working. DO NOT USE. Needs rewriting
void ueyeImageGrabber::increaseHeight()
{
   if (imageHeight >= 1)
   {
      imageHeight--;
      if (!useInputRectangle())
      {
         connected = false;
      }
   }
}

// This function will break the ring buffer and stop the video capture from
// working. DO NOT USE. Needs rewriting
void ueyeImageGrabber::decreaseHeight()
{
   imageHeight++;
   if (!useInputRectangle())
   {
      connected = false;
   }
}

// This function will break the ring buffer and stop the video capture from
// working. DO NOT USE. Needs rewriting
bool ueyeImageGrabber::setInputRectangle(int x, int y, int w, int h)
{
   inputX = x;
   inputY = y;
   imageWidth = w;
   imageHeight = h;

   return(useInputRectangle());
}

// This method is working but still needs to be fully tested to ensure that
// sizes and x y positions are correct TODO TESTING
bool ueyeImageGrabber::useInputRectangle()
{
   const std::string funcName = "bool ueyeImageGrabber::useInputRectangle():";
   bool cont = true;
   // to change the size of an image / select an area of interest it has to be done in increments of the
   // stated step size in height and width.
   int wstep = 4; //the step widths for the position and size of image - these are in pixels (4 pixels per step)
   int hstep = 2;// the step size for height for the position and size of image - in pixels, same as above
   int minXpos = 320; // the minimum x position an area of interest can start from
   int minYpos = 240; // the minimum y position an area of interest can start from

   //we check that the x and y positions are not smaller than the minimum stated x and y coordinates
   if (inputX < minXpos){
       //resize the width to make sure that we do not go out of the image bounds
       imageWidth -= (minXpos - inputX);
       //now set the x position to the min allowed
       inputX = minXpos;

   }
   else{
       // We check that the position is a multiple of 2.
       if (inputX % 2 != 0){
           inputX--;
       }
       if (inputX < 0){
           inputX = 0;
       }
       //check that we do not exceed the image size
       if (static_cast<unsigned long>(inputX) + imageWidth > sensorInfo.nMaxWidth){
           //reduce the width because we have already checked the x position is outside of the minimum
           //allowable x
          //inputX = sensorInfo.nMaxWidth - imageWidth;
           imageWidth = sensorInfo.nMaxWidth - inputX;
       }
   }
   //check y position
   if(inputY < minYpos){
       //resize the height to make sure that we do not go out of the image bounds
       imageHeight -= (minYpos - inputY);
       inputY = minYpos;
   }
    else{
       //check the position is a multiple of 2
       if (inputY % 2 != 0){
           inputY--;
       }
       if (inputY < 0){
           inputY = 0;
       }
       // check that the image height is not exceeded
       if (static_cast<unsigned long>(inputY) + imageHeight > sensorInfo.nMaxHeight){
           //reduce the height because we have already checked that the y position is outside of the
           //minimum allowable
           //inputY = sensorInfo.nMaxHeight - imageHeight;
           imageHeight = sensorInfo.nMaxHeight - inputY;
       }
   }
   // We check the width is a multiple of 4.
   if (imageWidth % 4 != 0){
       imageWidth += wstep - (imageWidth % 4);
       //imageWidth--;
   }
   //we check the height is a multiple of 2
   if (imageHeight % 2 != 0){
       imageHeight += hstep - (imageHeight % 2);
   }
   //We also check the sizes are within the max width and height but by this time this should not be the case
   if (imageWidth > sensorInfo.nMaxWidth){
       imageWidth = sensorInfo.nMaxWidth;
   }
   if (imageHeight > sensorInfo.nMaxHeight){
       imageHeight = sensorInfo.nMaxHeight;
   }
   //prepare the rectangle
   IS_RECT rectAOI;

   rectAOI.s32X     = inputX;
   rectAOI.s32Y     = inputY;
   rectAOI.s32Width = imageWidth;
   rectAOI.s32Height = imageHeight;
   // stop the live capture
   if(capture){
       is_StopLiveVideo(cameraHandle,IS_WAIT);
       capture = false;
   }
   //set the area of interest
   INT res = is_AOI( cameraHandle, IS_AOI_IMAGE_SET_AOI, (void*)&rectAOI, sizeof(rectAOI));
   if (res != IS_SUCCESS)
   {
      std::cerr << funcName << std::endl
         << "Could not set the input rectangle.\n";
      return(false);
   }
   else
   {
      if (grabbedImage)
      {
         INT res = is_FreeImageMem(cameraHandle, grabbedImage, memID);
         if (res != IS_SUCCESS)
         {
            std::cerr << "Could not free the image memory.\n";
            cont = false;
         }
      }
      if (cont)
      {
         res = is_AllocImageMem(cameraHandle,imageWidth, imageHeight,24, // RGB8
                                &grabbedImage, &memID);
         if (res != IS_SUCCESS)
         {
            std::cerr << funcName << std::endl
               << "Could not allocate memory for the grabbed images.\n";
            cont = false;
         }
      }
      if (cont)//make image memory active
      {
         res = is_SetImageMem(cameraHandle, grabbedImage, memID);
         if (res != IS_SUCCESS)
         {
            std::cerr << funcName << std::endl
               << "Could not set the image memory.\n";
            cont = false;
         }
      }
      if (cont)
      {
         imageGrabber::setImageSize(imageWidth, imageHeight);
         //start live capture again
         captureImage();
         return(true);
      }
      else
      {
         return(false);
      }
   }
}

/*!
 * \brief Initialise the grabber.
 * \param[in] upsideDown Set to true if the camera is upside-down.
 * \param[in] flip Set to true to horizontally flip the images.
 * \param[in] camNumber The number of the camera (0 is the first available).
 * \param[in] portNumber The number of the port on the device.
 * \param[in] posX
 * \param[in] posY The position of the viewport of the grabbed part.
 * \param[in] width
 * \param[in] height The width and height of the grabbed area and of the
 * images.
 */
void ueyeImageGrabber::init(bool upsideDown,
                            bool flip,
                            int camNumber,
                            unsigned short posX,
                            unsigned short posY,
                            unsigned short width,
                            unsigned short height)
{
   std::string funcName = "void ueyeImageGrabber::init():";
   bool cont = true;

   setParameters(width, height, upsideDown, flip);
   //find out how many cameras are available, and which if any are busy
   INT numCams;
   INT res = is_GetNumberOfCameras (&numCams);
   // At least one camera must be available
   if( res == IS_SUCCESS) {
       if( numCams >= 1 ) {

          // Create new list to hold camera details with suitable size
          UEYE_CAMERA_LIST* pucamlist;
          pucamlist = (UEYE_CAMERA_LIST*) new BYTE [sizeof (DWORD) + numCams * sizeof (UEYE_CAMERA_INFO)];
          pucamlist->dwCount = numCams;
          //Retrieve the camera info
          if (is_GetCameraList(pucamlist) == IS_SUCCESS) {
              //there are two ways of connecting - select a specific camera (camNumber >0) or select the first available
              int counter;
              if(camNumber > 0){
                  UINT camnum = camNumber; //unsigned int comparison
                  for (counter = 0; counter < (int)pucamlist->dwCount; counter++) {
                      //get a handle on the specific camera and check that it is available
                      if( camnum == pucamlist->uci[counter].dwCameraID && pucamlist->uci[counter].dwInUse == 0){
                          //camera id matches and camera is not being used - get camera id
                           cameraHandle = pucamlist->uci[counter].dwCameraID;
                           cont = true;
                           break;
                      }
                      else{
                          cont = false; //ensures that cannot carry on if no camera handle is allocated
                      }

                 }
             }
             else{
                  for (counter = 0; counter < (int)pucamlist->dwCount; counter++) {
                      //get a handle on the first usable camera
                      if(pucamlist->uci[counter].dwInUse == 0){
                          //camera is not being used - get camera id
                           cameraHandle = pucamlist->uci[counter].dwCameraID;
                           cont = true;
                           break;
                      }
                      else{
                          cont = false; //ensures that cannot carry on if no camera handle is allocated
                      }

                 }
              }
             delete [] pucamlist;
          }
          else{
              std::cerr << funcName << std::endl << "Could not get the camera list.\n";
              cont = false;
          }

      }
   }
   else{
       std::cerr << funcName << std::endl << "Could not get the number of connected cameras.\n";
       cont = false;
   }


   if (cont)
   {
      //cameraHandle = camNumber;
      res = is_InitCamera(&cameraHandle, NULL);
      if (res != IS_SUCCESS)
      {
         std::cerr << funcName << std::endl
            << "Could not initialise the camera.\n";
         cont = false;
      }
   }

/* Start of ring buffer memory allocation */

   capture = false;

   if (cont)
   {
      for (int bufnum = 1; bufnum < 5; bufnum++) // Just 4 buffer slots
      {
         //the 24 passed to the is_allocImageMem function is 24 bit RGB8
         res = is_AllocImageMem(cameraHandle,width, height,24,&ringbuffer[bufnum], &sequenceBufIDs[bufnum]);
         if (res != IS_SUCCESS)
         {
            std::cerr << funcName << std::endl
               << "Could not allocate memory for the ring buffer.\n";
            cont = false;
            break;
         }
         //activate the image memory
         res = is_SetImageMem (cameraHandle, ringbuffer[bufnum], sequenceBufIDs[bufnum]);
         if (res != IS_SUCCESS)
         {
            std::cerr << funcName << std::endl
               << "Could not make the memory active.\n";
            //cont = false;
            //break;
         }
         res = is_AddToSequence(cameraHandle, ringbuffer[bufnum], sequenceBufIDs[bufnum]);
         if (res != IS_SUCCESS)
         {
            std::cerr << funcName << std::endl
               << "Could not add memory to ring buffer.\n";
            cont = false;
            break;
         }
      }
   }

/* End of ring buffer set up */

/* Removed memory allocation for single buffer: replaced with ring buffer */

   if (cont)
   {
      res = is_SetColorMode (cameraHandle, IS_CM_RGB8_PACKED);
      if (res != IS_SUCCESS)
      {
         std::cerr << funcName << std::endl
            << "Could not set the colour mode to RGB8.\n";
         cont = false;
      }
   }

   if (cont)
   {
      /*res = is_SetHardwareGain(cameraHandle,
                               IS_SET_ENABLE_AUTO_GAIN, // Auto gain
                               IS_IGNORE_PARAMETER, // Don't touch red
                               IS_IGNORE_PARAMETER, // Don't touch green
                               IS_IGNORE_PARAMETER); // Don't touch blue*/
       double pval1 = 0;
       //0 disables autogain
       res = is_SetAutoParameter(cameraHandle,IS_SET_ENABLE_AUTO_GAIN,&pval1,NULL);
       if (res != IS_SUCCESS)
      {
         std::cerr << funcName << std::endl
            << "Could not set the hardware gain.\n";
      }

      pval1 = 1.0;
      res = is_SetAutoParameter(cameraHandle, IS_SET_ENABLE_AUTO_SHUTTER, &pval1, NULL);
     // res = is_SetExposureTime(cameraHandle,
        //                       IS_SET_ENABLE_AUTO_SHUTTER, // Auto
         //                      &expTime);
      if (res != IS_SUCCESS)
      {
         std::cerr << funcName << std::endl
            << "Could not set the exposure time.\n";
      }
   }

   res = is_GetSensorInfo(cameraHandle, &sensorInfo);
   if (res != IS_SUCCESS)
   {
      std::cerr << funcName << std::endl
      << "Could not get the information about the sensor.\n";
   }

// setInputRectangle() REMOVED AS BREAKS THE RING BUFFER DUE TO MEMORY ALLOCATION
// This functionality could be replaced with calls to the UEye API functions
// that provide the same functionality directly

// Let's get some control values.
   if (cont)
   {
      minGain = 0;
      maxGain = 100;

      //res = is_GetExposureRange(cameraHandle, &minExposure, &maxExposure, &exposureInterval);  now obsolete
      double exparr[3];
      res = is_Exposure(cameraHandle, IS_EXPOSURE_CMD_GET_EXPOSURE_RANGE, &exparr, 24);
      if(res != IS_SUCCESS)
      {
         std::cerr << "Error while getting the Exposure range of the camera\n";
	 cont = false;
      }
      minExposure = exparr[0];
      maxExposure = exparr[1];
      exposureInterval = exparr[2];

      res = is_GetFrameTimeRange(cameraHandle, &minFramerate, &maxFramerate, &framerateInterval);
      if(res != IS_SUCCESS)
      {
         std::cerr << "Error while getting the Framerate range of the camera\n";
	 cont = false;
      }
   }

   if (!cont)
   {
      connected = false;
   }
   else
   {
      connected = true;
   }
//captureImage();
}

/*!
 * \brief Set the gain.
 *
 * \param[in] value The new gain value, between 0 and 100.  A negative value sets automatic gain
 * control.
 */
void ueyeImageGrabber::setGain(long value)
{
   assert(cameraHandle);
   int err;
   if (value < 0)
   {
      err = is_SetHardwareGain(cameraHandle, IS_SET_ENABLE_AUTO_GAIN, IS_IGNORE_PARAMETER, IS_IGNORE_PARAMETER, IS_IGNORE_PARAMETER);
      if (err != IS_SUCCESS)
      {
          std::cerr << "Error while setting the gain of the camera.\n";
      }
   }
   else
   {
      if (value < minGain)
         value = minGain;
      if (value > maxGain)
         value = maxGain;

      err = is_SetHardwareGain(cameraHandle, value, IS_IGNORE_PARAMETER, IS_IGNORE_PARAMETER, IS_IGNORE_PARAMETER);
      if (err != IS_SUCCESS)
      {
          std::cerr << "Error while setting the gain of the camera.\n";
      }
   }
}

/*!
 * \brief Get the current gain value.
 * \return The value.
 */
long ueyeImageGrabber::getGain()
{
   assert(cameraHandle);

   int value = is_SetHardwareGain(cameraHandle, IS_GET_MASTER_GAIN, IS_IGNORE_PARAMETER, IS_IGNORE_PARAMETER, IS_IGNORE_PARAMETER);
   if (value < 0)
      std::cerr << "Error while getting the current gain\n";

   return(static_cast<long>(value));
}

/*!
 * \brief Set the exposure.
 *
 * \param[in] value The new exposure value.  A negative value means automatic
 * exposure control.
 */
void ueyeImageGrabber::setExposure(long value)
{
   assert(cameraHandle);
   int err;
   double newExp;
   double pval1 = 1.0;

   if (value >= 0)
   {
       if (value < minExposure){
          value = minExposure;
          std::cout<<"exposure set to min exposure of:"<<minExposure<<std::endl;
       }
       else if (value > maxExposure){
          value = maxExposure;
          std::cout<<"exposure set to max exposure of:"<<maxExposure<<std::endl;
       }
       newExp = value;
       //last parameter is the size of 3rd parameter
       err = is_Exposure(cameraHandle, IS_EXPOSURE_CMD_SET_EXPOSURE, &newExp, 8);
       if (err != IS_SUCCESS)
       {
           std::cerr << "Error while setting the Exposure time.\n";
       }
   }
   else
   {
        err = is_SetAutoParameter(cameraHandle, IS_SET_ENABLE_AUTO_SHUTTER, &pval1, NULL);
        std::cout<<"exposure set to autoshutter on"<<std::endl;
        if (err != IS_SUCCESS)
        {
            std::cerr << "Error while setting the auto Exposure mode.\n";
        }
   }
}

/*!
 * \brief Get the current exposure value.
 * \return The value.
 */
long ueyeImageGrabber::getExposure()
{
   assert(cameraHandle);
   int err;
   //double exp;
   double pval1;

   err = is_Exposure(cameraHandle, IS_EXPOSURE_CMD_GET_EXPOSURE, &pval1, 8);
   //err = is_SetExposureTime(cameraHandle, IS_GET_EXPOSURE_TIME, &exp);
   if (err != IS_SUCCESS)
      std::cerr << "Error while getting the current Exposure time.\n";

   return(static_cast<long>(pval1));
}

/*!
 * \brief Get the current auto exposure setting.
 * \return The setting.
 */
long ueyeImageGrabber::getAutoExposureSetting()
{
   assert(cameraHandle);
   int err;
   double pval1;
   double dblDummy = 0;

   err = is_SetAutoParameter(cameraHandle, IS_GET_ENABLE_AUTO_SHUTTER, &pval1, &dblDummy);
   if (err != IS_SUCCESS){
      std::cerr << "Error while getting the current Auto Exposure setting.\n";
    }
   return(static_cast<long>(pval1));
}


/*!
 * \brief Set the current auto exposure setting.
 * \param[in] The auto exposure setting, 1 enables control, 0 disables control.
 */
void ueyeImageGrabber::setAutoExposureControl(int val){
    assert(cameraHandle);
    int err;
    double pval1 = val;
    double dblDummy = 0;

    err = is_SetAutoParameter(cameraHandle, IS_SET_ENABLE_AUTO_SHUTTER, &pval1, &dblDummy);
    if (err != IS_SUCCESS){
       std::cerr << "Error while getting the current Auto Exposure setting.\n";
     }
}

/*!
 * \brief Set the framerate.
 *
 * \param[in] value The new framerate value.
 */
void ueyeImageGrabber::setFramerate(double value)
{
   assert(cameraHandle);
   int err;
   double newFPS, fps;

   if (value < 0)
   {
      //find out what the default is and set it to that newFPS = default value
      err = is_SetFrameRate(cameraHandle, IS_GET_DEFAULT_FRAMERATE, &newFPS);
      if (err != IS_SUCCESS)
         std::cerr << "Error while getting the default framerate.\n";
      else
      {
         //now set it to default value and fps returns what actual value was set
         err = is_SetFrameRate(cameraHandle, newFPS, &fps);
         if (err != IS_SUCCESS)
	    std::cerr << "Error while setting the default framerate.\n";
      }
   }
   else
   {
      if (value < minFramerate)
            value = minFramerate;
      if (value > maxFramerate)
            value = maxFramerate;

      err = is_SetFrameRate(cameraHandle, value, &newFPS);
      if (err != IS_SUCCESS)
         std::cerr << "Error while setting the new framerate.\n";
   }
}

/*!
 * \brief Get the current framerate value.
 * \return The value.
 */
double ueyeImageGrabber::getFramerate()
{
   assert(cameraHandle);
   int err;
   double fps;
   err = is_SetFrameRate(cameraHandle, IS_GET_FRAMERATE, &fps);
   //err = is_GetFramesPerSecond(cameraHandle, &fps); <- this method is recommended by IDS but was found not to work
   if (err != IS_SUCCESS)
      std::cerr << "Error while getting the current framerate.\n";
   
   return(fps);
}
/*!
 * \brief set the camera mode, default is freerun continuous capture
 * \param[in] mode the mode - always continuous capture, default = 0,
 * \ software trigger = 1, hardware trigger = 2
 * \return true if mode set, false otherwise.
 */
bool ueyeImageGrabber::setMode(int mode)
{
        assert(cameraHandle);
	int res = 0;
	//at this stage the live capture might have been started - if so stop
	if (capture)
	{
		is_StopLiveVideo(cameraHandle,IS_DONT_WAIT);
		capture = false;
	}
	//set the mode
	switch (mode)
	{
	case 0: 
		//default mode - no triggering (free run continuous)
		res = is_SetExternalTrigger(cameraHandle,IS_SET_TRIGGER_OFF);
		break;
	
	case 1:
		//software trigger mode
		res = is_SetExternalTrigger(cameraHandle,IS_SET_TRIGGER_SOFTWARE);
		break; 
	case 2:
		//hardware trigger mode
		res = is_SetExternalTrigger(cameraHandle,IS_SET_TRIGGER_LO_HI); 
		break;
	default:
		//anything else is incorrect and should be set back to default
		res = is_SetExternalTrigger(cameraHandle,IS_SET_TRIGGER_OFF);
		std::cerr << "void ueyeImageGrabber::setMode" << std::endl
            << "Mode value incorrect, setting to default value.\n";
		break;
	}// end switch
	if(res == IS_SUCCESS)
	{
		//call captureImage so camera is ready for continuous triggering again
		captureImage();
		return true;
	}
	else
	{
		std::cerr << "void ueyeImageGrabber::setMode" << std::endl
            << "An error occurred, could not set mode.\n";
        return false;
	}
   }

   /*!
    * \brief Get a list of discrete pixel clocks for the camera.
    * \param[out] pixelClocks The array of pixel cloacks.  This must have been
    * allocated elsewhere.
    * \sa getNumberOfPixelClocks().
    */
   void ueyeImageGrabber::getPixelClockList(UINT* pixelClocks)
   {
      assert(cameraHandle);
//       // No camera has more than 150 different pixel clocks.
//       UINT pixClockList[150];
//       UINT* pixPointer = pixClockList;
      UINT numOfSupportedPixClocks = 0;
      INT retVal = is_PixelClock(cameraHandle,
                                 IS_PIXELCLOCK_CMD_GET_NUMBER,
                                 (void*)&numOfSupportedPixClocks,
                                 sizeof(numOfSupportedPixClocks));
      //now get the list safely
      if ((retVal == IS_SUCCESS) && (numOfSupportedPixClocks > 0))
      {
         ZeroMemory(&pixelClocks, sizeof(pixelClocks)); // Really?
         retVal = is_PixelClock(cameraHandle,
                                IS_PIXELCLOCK_CMD_GET_LIST,
                                (void*)pixelClocks,
                                numOfSupportedPixClocks * sizeof(UINT));
      }
   }

   /*!
    * \brief Get the number of supported pixel clocks.
    * \return The number of pixel clocks.
    */
   UINT ueyeImageGrabber::getNumberOfPixelClocks()
   {
      assert(cameraHandle);
      UINT numOfSupportedPixClocks = 0;
      INT retVal = is_PixelClock(cameraHandle,
                                 IS_PIXELCLOCK_CMD_GET_NUMBER,
                                 (void*)&numOfSupportedPixClocks,
                                 sizeof(numOfSupportedPixClocks));
      return retVal;
   }

   /*!
    * \brief Get the supported pixel clock range.
    * \param[out] range The minimum, maximum and mid values in range.
    *  Must have been allocated elsewhere (array of 3).
    * \return min max and mid values in range.
    */
   void ueyeImageGrabber::getPixelClockRange(UINT* range)
   {
      assert(cameraHandle);
//       UINT range[3];
//       UINT* rangePointer = range;
      ZeroMemory(range, sizeof(UINT));
      // Get pixel clock range
      INT nRet = is_PixelClock(cameraHandle,
                               IS_PIXELCLOCK_CMD_GET_RANGE,
                               (void*)range,
                               sizeof(UINT));
      if(nRet != IS_SUCCESS)
      {
         std::cerr<<"could not get pixelclock range"<<std::endl;
      }
//       return rangePointer;
   }

   /*!
    * \brief get the current pixel clock value
    * \return the pixel clock value.
    */
   UINT ueyeImageGrabber::getCurrentPixelClockValue(){
       assert(cameraHandle);
       UINT currentVal;
       // Get current pixel clock setting value
       INT retVal = is_PixelClock(cameraHandle, IS_PIXELCLOCK_CMD_GET, (void*)&currentVal, sizeof(currentVal));
       if(retVal == IS_SUCCESS){
           return currentVal;
       }
       return 0;
   }
   /*!
    * \brief get the default pixel clock value
    * \param[in] cameraHandle the ueye camera handle 
    * \return default pixel clock value.
    */
   UINT ueyeImageGrabber::getDefaultPixelClockValue(){
       assert(cameraHandle);
       UINT defaultVal;
       // Get current pixel clock setting value
       INT retVal = is_PixelClock(cameraHandle, IS_PIXELCLOCK_CMD_GET_DEFAULT, (void*)&defaultVal, sizeof(defaultVal));
       if(retVal == IS_SUCCESS){
           return defaultVal;
       }
       return 0;
   }
   /*!
    * \brief set the pixel clock
    * \param[in] cameraHandle the ueye camera handle 
    * \param[in] newValue the new pixel clock value 
    * \return success or failure
    */
   INT ueyeImageGrabber::setPixelClockVal(UINT newValue){
       assert(cameraHandle);
        // Set this pixel clock
       INT retVal = is_PixelClock(cameraHandle, IS_PIXELCLOCK_CMD_SET,(void*)&newValue, sizeof(newValue));
       return retVal;
   }

} // namespace imalib.
