/*
 * imalib
 * 
 * Copyright 2017 Aberystwyth University
 * 
 * This library and program are free software: you can redistribute them
 * and/or modify them under the terms of the GNU General Public License as 
 * published by the Free Software Foundation, either version 3 of the License,
 * or (at your option) any later version.
 * 
 * The program and library are distributed in the hope that they will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*!
 * \file v4l2ImageGrabber.cpp
 * \brief An image grabber for a V4L2 frame grabber.  Implementation.
 *
 * \author Fred Labrosse <ffl@aber.ac.uk>
 * \copyright 2017 Aberystwyth University
 */

#include "imalib/imageRGB.h"
#include "imalib/v4l2ImageGrabber.h"

#include <sys/types.h>
#include <sys/stat.h> // for struct stat.
#include <fcntl.h>
#include <sys/mman.h>

#include <iostream>
#include <fstream>

#include <boost/program_options.hpp>
namespace po = boost::program_options;

namespace imalib
{
//    namespace
//    {
      static __u32 formats[] = {v4l2_fourcc('R', 'G', 'B', '3'),
                         v4l2_fourcc('Y', 'U', 'Y', 'V'),
                         v4l2_fourcc('U', 'Y', 'V', 'Y'),
                         v4l2_fourcc('G', 'B', 'R', 'G')};
//       static __u32 formats[] = {v4l2_fourcc('R', 'G', 'B', '3')};

      const std::vector<__u32> v4l2ImageGrabber::supportedPixelFormats(
         formats,
         formats + (sizeof(formats) / sizeof(formats[0])));
//    }

// const unsigned short DEFAULT_BRIGHTNESS = 50;
// const unsigned short DEFAULT_CONTRAST = 50;
// const unsigned short DEFAULT_SATURATION = 50;
// const unsigned short DEFAULT_HUE = 50;
// const unsigned short DEFAULT_WHITENESS = 50;
// const unsigned short DEFAULT_DEPTH = 50;

/*!
 * \brief Constructor.
 *
 * \param imgWidth
 * \param[in] imgHeight The width and height of the image.
 * \param[in] upsideDown Set to true if the camera is upside-down.
 * \param[in] flip Set to true to horizontally flip the images.
 * \param[in] devNumber The number of the video device (/dev/video?).
 * \param[in] inputNumber The video input number.
 * \param[in] videoStandard The video standard number.
 * \param[in] ioMethod The io method to use.
 * \param[in] posX
 * \param[in] posY The position of the viewport of the grabbed part.
 * \param[in] width
 * \param[in] height The width and height of the grabbed area.
 */
v4l2ImageGrabber::v4l2ImageGrabber(unsigned short imgWidth,
                                   unsigned short imgHeight,
                                   bool upsideDown,
                                   bool flip,
                                   int devNumber,
                                   int inputNumber,
                                   int videoStandard,
                                   possibleIoMethods ioMethod,
                                   unsigned short posX,
                                   unsigned short posY,
                                   unsigned short width,
                                   unsigned short height)
: imageBuffers(NULL), streaming_(false)
{
   init(imgWidth, imgHeight,
        upsideDown,
        flip,
        devNumber,
        inputNumber,
        videoStandard,
        ioMethod,
        posX, posY, width, height);
}

/*!
 * \brief Constructor.
 * \param[in] fileName Name of the configuration file.
 */
v4l2ImageGrabber::v4l2ImageGrabber(const std::string& fileName)
: imageBuffers(NULL), streaming_(false)
{
   // Let's extract the options from the configuration file.
   po::options_description optionsDesc("v4l2ImageGrabber options");
   int vidDev = 0;
   int vidNum = 0;
   int vidStd = 0;
   int poX = 0;
   int poY = 0;
   int soWidth = IMAGE_DEFAULT_INPUTWIDTH;
   int soHeight = IMAGE_DEFAULT_INPUTHEIGHT;
   int imWidth = 200;
   int imHeight = 200;
   bool upDown = false;
   bool fl = false;
   unsigned short ioMet = static_cast<unsigned short>(IO_METHOD_MMAP);
   optionsDesc.add_options()
      ("videoDevice", po::value<int>(&vidDev)->default_value(vidDev), "")
      ("videoInput", po::value<int>(&vidNum)->default_value(vidNum), "")
      ("videoStandard", po::value<int>(&vidStd)->default_value(vidStd), "")
      ("ioMethod", po::value<unsigned short>(&ioMet)->default_value(ioMet), "")
      ("posX", po::value<int>(&poX)->default_value(poX), "")
      ("posY", po::value<int>(&poY)->default_value(poY), "")
      ("sourceWidth",
       po::value<int>(&soWidth)->default_value(soWidth), "")
      ("sourceHeight",
       po::value<int>(&soHeight)->default_value(soHeight), "")
      ("imageWidth", po::value<int>(&imWidth)->default_value(imWidth), "")
      ("imageHeight", po::value<int>(&imHeight)->default_value(imHeight), "")
      ("upsideDown", po::value<bool>(&upDown)->default_value(upDown), "")
      ("flip", po::value<bool>(&fl)->default_value(fl), "")
      ;

   po::variables_map vm;
   std::ifstream configFile(fileName.c_str());
   if (configFile.is_open())
   {
      po::store(po::parse_config_file(configFile, optionsDesc), vm);
      po::notify(vm);
      configFile.close();
   }
   else
   {
      std::cerr << "v4lImageGrabber::v4lImageGrabber():\n"
         << "WARNING: could not open configuration file " << fileName
         << "!\n"
         << "\tUsing defaults.\n";
   }

   init(imWidth, imHeight,
        upDown,
        fl,
        vidDev,
        vidNum,
        vidStd,
        static_cast<possibleIoMethods>(ioMet),
        poX, poY,
        soWidth, soHeight);
}


/*!
\brief Constructor.
*/
v4l2ImageGrabber::~v4l2ImageGrabber()
{
   std::string funcName("v4l2ImageGrabber::~v4l2ImageGrabber()");

   switch (ioMethod_)
   {
      case IO_METHOD_READ:
         /* Nothing to do. */
         break;

      case IO_METHOD_MMAP:
      case IO_METHOD_USERPTR:
         if (streaming_)
         {
            enum v4l2_buf_type type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
            if (-1 == xioctl(fd, VIDIOC_STREAMOFF, &type))
            {
               std::cerr << std::endl << funcName << std::endl
                  << "Could not stop streaming (VIDIOC_STREAMOFF): "
                  << strerror(errno) << std::endl;
            }
         }
         break;
   }

   if (imageBuffers)
   {
      switch (ioMethod_)
      {
         case IO_METHOD_READ:
            delete [] imageBuffers[0].start;
            break;

         case IO_METHOD_MMAP:
            for (unsigned int i = 0; i < nbBuffers; ++i)
               if (-1 == munmap(imageBuffers[i].start, imageBuffers[i].length))
               {
                  std::cerr << std::endl << funcName << std::endl
                     << "Could not unmap memory: "
                     << strerror(errno) << std::endl;
               }
            break;

         case IO_METHOD_USERPTR:
//          for (i = 0; i < n_buffers; ++i)
//             free (buffers[i].start);
            break;
      }

      delete [] imageBuffers;
   }

   if (fd)
      close(fd);
}

/*!
 * \brief Display on the standard output some information about the
 * source of the images.
 *
 * \param[in,out] aStream The stream to which the info is added.
 */
void v4l2ImageGrabber::printInfoAboutSource(std::ostringstream* aStream)
{
   *aStream << "----------------------------------------------\n"
      << "Image source:\n";

   if (connected)
   {
      *aStream << "\tgrabber: " << hdwCap_.card << std::endl
         << "\t(driver: " << hdwCap_.driver << ")\n";

      if (IO_METHOD_MMAP == ioMethod_)
         *aStream << "\tusing memory mapping.\n";
      else if (IO_METHOD_READ == ioMethod_)
         *aStream << "\tusing 'read' io method.\n";
      else if (IO_METHOD_USERPTR == ioMethod_)
         *aStream << "\tusing 'user pointer' io method.\n";

      xioctl(fd, VIDIOC_CROPCAP, &cropcap);
      *aStream << "\timage bounds are:\n"
         << "\t\tcolumn min: " << cropcap.bounds.left << std::endl
         << "\t\trow min: " << cropcap.bounds.top << std::endl
         << "\t\twidth: " << cropcap.bounds.width << std::endl
         << "\t\theight: " << cropcap.bounds.height << std::endl;

      *aStream << "\tgrabbed size: " << crop.c.width
         << "x" << crop.c.height << " at ("
         << crop.c.left << "," << crop.c.top << ")" << std::endl;
   }

   printBasicInfoAboutSource(aStream);
   *aStream << "----------------------------------------------\n";
}

/*!
 * \brief Capture the image from a grabber.
 *
 * This blocks until data is available to be read.  This should be called
 * before calling getImage().
 */
void v4l2ImageGrabber::captureImage()
{
   std::string funcName("void v4l2ImageGrabber::captureImage()");

   if (connected)
   {
      capture = false;

      fd_set fds;
      struct timeval tv;
      int r;

      FD_ZERO (&fds);
      FD_SET (fd, &fds);

      /* Timeout. */
      tv.tv_sec = 60;
      tv.tv_usec = 0;

      r = select (fd + 1, &fds, NULL, NULL, &tv);

      if (-1 == r)
      {
         std::cerr << std::endl << "-->" << funcName << std::endl
            << "\tError while waiting for data on device '" << deviceName << "': "
            << strerror(errno) << std::endl;

         connected = false;
      }
      else if (0 == r)
      {
         std::cerr << std::endl << "-->" << funcName << std::endl
            << "\tTimeout while waiting for data on device '" << deviceName
            << "'.\n";

         connected = false;
      }
      else
      {
         capture = true;
      }
   }
}


/*!
 * \brief Get an image from the frame grabber.
 *
 * The images are resized here.
 *
 * captureImage() should be called before, but will be called if it wasn't.
 *
 * \param[out] theImage The image in which to save the grabbed image, if
 * not NULL.
 * \param[out] rawImage The image in which to save the raw image (before any
 * transformation is applied), if not NULL.
 */
void v4l2ImageGrabber::getImage(imageRGB* theImage, imageRGB* rawImage)
{
   std::string funcName("void v4l2ImageGrabber::getImage(...)");
   struct v4l2_buffer buf;

   if (!capture)
   {
      captureImage();
   }

   if (connected)
   {
      switch (ioMethod_)
      {
         case IO_METHOD_READ:
            if (-1 == read(fd, imageBuffers[0].start, imageBuffers[0].length))
            {
               std::cerr << std::endl << "-->" << funcName << std::endl
                  << "\tError while getting an image on device '"
                  << deviceName << "': "
                  << strerror(errno) << std::endl;
               connected = false;
            }
            else
            {
               processNewImage(imageBuffers[0].start,
                               theImage, rawImage);
            }

            break;

         case IO_METHOD_MMAP:
            memset(&buf, 0, sizeof(buf));

            buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
            buf.memory = V4L2_MEMORY_MMAP;

            if (-1 == xioctl(fd, VIDIOC_DQBUF, &buf))
            {
               std::cerr << std::endl << "-->" << funcName << std::endl
                  << "\tError in VIDIOC_DQBUF:" << std::endl
                  << strerror(errno) << std::endl;
               connected = false;
            }
            else
            {
               assert(buf.index < nbBuffers);

               processNewImage(imageBuffers[buf.index].start,
                               theImage, rawImage);

               if (-1 == xioctl(fd, VIDIOC_QBUF, &buf))
               {
                  std::cerr << std::endl << "-->" << funcName << std::endl
                     << "\tError in VIDIOC_QBUF:" << std::endl
                     << strerror(errno) << std::endl;
                  connected = false;
               }
            }

            break;

         case IO_METHOD_USERPTR:
            memset(&buf, 0, sizeof(buf));

            buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
            buf.memory = V4L2_MEMORY_USERPTR;

            if (-1 == xioctl(fd, VIDIOC_DQBUF, &buf))
            {
               std::cerr << std::endl << "-->" << funcName << std::endl
                  << "\tError in VIDIOC_DQBUF:" << std::endl
                  << strerror(errno) << std::endl;
               connected = false;
            }
            else
            {
               unsigned int  i;
               for (i = 0; i < nbBuffers; ++i)
               {
                  if ((buf.m.userptr == (unsigned long)imageBuffers[i].start)
                      && (buf.length == imageBuffers[i].length))
                  {
                     break;
                  }
               }

               assert(i < nbBuffers);

               processNewImage((unsigned char *) buf.m.userptr,
                               theImage, rawImage);

               if (-1 == xioctl(fd, VIDIOC_QBUF, &buf))
               {
                  std::cerr << std::endl << "-->" << funcName << std::endl
                     << "\tError in VIDIOC_QBUF:" << std::endl
                     << strerror(errno) << std::endl;
                  connected = false;
               }

               break;
            }
      }
   }
}



/*!
 * \brief Process the new image.
 *
 * This is taking the raw data from the grabber and putting it into the
 * required image.
 *
 * \param[in] buffer The buffer to take the data from.
 * \param[out] theImage The image in which to save the grabbed image, if
 * not NULL.
 * \param[out] rawImage The image in which to save the raw image (before any
 * transformation is applied), if not NULL.
 */
void v4l2ImageGrabber::processNewImage(unsigned char* buffer,
                                       imageRGB* theImage,
                                       imageRGB* rawImage)
{
   if (connected)
   {
      unsigned char RGB[3];
      unsigned char YCbCr[3];
      unsigned char* thePixels = NULL;
      unsigned long ind = 0;
      unsigned long indPix = 0;

      if (theImage)
      {
         theImage->setSize(imgMap->targetWidth(),
                           imgMap->targetHeight());
         std::vector<imageMapping::coordsPair>::iterator iter;

         imageRGB tmpImage;

         switch (pixelFormat_)
         {
            case v4l2_fourcc('R', 'G', 'B', '3'):
               for (iter = imgMap->map.begin();
                    iter < imgMap->map.end();
                    ++iter)
               {
                  ind = (iter->sY * imgMap->sourceWidth() + iter->sX) * 3;
                  theImage->setPixel(iter->tX, iter->tY,
                                     &buffer[ind]);
               }
               break;

            case v4l2_fourcc('Y', 'U', 'Y', 'V'):
               tmpImage.setSize(imgMap->sourceWidth(),
                                imgMap->sourceHeight());
               thePixels = tmpImage.getActualPixel(0, 0);
               for (ind = 0, indPix = 0;
                    ind < static_cast<unsigned long>(imgMap->sourceWidth())
                          * static_cast<unsigned long>(imgMap->sourceHeight())
                          / 2;
                    ++ind, indPix += 2)
               {
                  unsigned long ind2 = ind * 4;
                  unsigned long indPix2 = indPix * 3;
                  YCbCr[0] = buffer[ind2];
                  YCbCr[1] = buffer[ind2+1];
                  YCbCr[2] = buffer[ind2+3];
                  YCbCrToRGB(YCbCr, RGB);
                  memcpy(&(thePixels[indPix2]), RGB, 3);
                  YCbCr[0] = buffer[ind2+2];
                  YCbCrToRGB(YCbCr, RGB);
                  memcpy(&(thePixels[indPix2 + 3]), RGB, 3);
               }
               tmpImage.apply(*imgMap, theImage);
               break;

            case v4l2_fourcc('U', 'Y', 'V', 'Y'):
               tmpImage.setSize(imgMap->sourceWidth(),
                                imgMap->sourceHeight());
               thePixels = tmpImage.getActualPixel(0, 0);
               for (ind = 0, indPix = 0;
                    ind < static_cast<unsigned long>(imgMap->sourceWidth())
                        * static_cast<unsigned long>(imgMap->sourceHeight())
                        / 2;
                    ++ind, indPix += 2)
                    {
                       unsigned long ind2 = ind * 4;
                       unsigned long indPix2 = indPix * 3;
                       YCbCr[0] = buffer[ind2+1];
                       YCbCr[1] = buffer[ind2+0];
                       YCbCr[2] = buffer[ind2+2];
                       YCbCrToRGB(YCbCr, RGB);
                       memcpy(&(thePixels[indPix2]), RGB, 3);
                       YCbCr[0] = buffer[ind2+3];
                       YCbCrToRGB(YCbCr, RGB);
                       memcpy(&(thePixels[indPix2 + 3]), RGB, 3);
                    }
                    tmpImage.apply(*imgMap, theImage);
                    break;

            case v4l2_fourcc('G', 'B', 'R', 'G'):
               tmpImage.setSize(imgMap->sourceWidth(),
                                imgMap->sourceHeight());
               thePixels = tmpImage.getActualPixel(0, 0);
               GBRGtoRGB(buffer, thePixels);
               tmpImage.apply(*imgMap, theImage);
               break;

            default:
               std::cerr << "Pixel format not supported!\n";
         }
      }
      if (rawImage)
      {
         rawImage->setSize(imgMap->sourceWidth(),
                           imgMap->sourceHeight());
         switch (pixelFormat_)
         {
            case v4l2_fourcc('R', 'G', 'B', '3'):
               for (unsigned short indX = 0;
                    indX < imgMap->sourceWidth();
                    ++indX)
               {
                  for (unsigned short indY = 0;
                       indY < imgMap->sourceHeight();
                       ++indY)
                  {
                     ind = (indY * imgMap->sourceWidth() + indX) * 3;
                     rawImage->setPixel(indX, indY, &buffer[ind]);
                  }
               }
               break;
            case v4l2_fourcc('Y', 'U', 'Y', 'V'):
               thePixels = rawImage->getActualPixel(0, 0);
               for (ind = 0, indPix = 0;
                    ind < static_cast<unsigned long>(imgMap->sourceWidth())
                       * static_cast<unsigned long>(imgMap->sourceHeight())
                       / 2;
                    ++ind, indPix += 2)
               {
                  unsigned long ind2 = ind * 4;
                  unsigned long indPix2 = indPix * 3;
                  YCbCr[0] = buffer[ind2];
                  YCbCr[1] = buffer[ind2+1];
                  YCbCr[2] = buffer[ind2+3];
                  YCbCrToRGB(YCbCr, RGB);
                  memcpy(&(thePixels[indPix2]), RGB, 3);
                  YCbCr[0] = buffer[ind2+2];
                  YCbCrToRGB(YCbCr, RGB);
                  memcpy(&(thePixels[indPix2 + 3]), RGB, 3);
               }
               break;

            case v4l2_fourcc('U', 'Y', 'V', 'Y'):
               thePixels = rawImage->getActualPixel(0, 0);
               for (ind = 0, indPix = 0;
                    ind < static_cast<unsigned long>(imgMap->sourceWidth())
                       * static_cast<unsigned long>(imgMap->sourceHeight())
                       / 2;
                    ++ind, indPix += 2)
               {
                  unsigned long ind2 = ind * 4;
                  unsigned long indPix2 = indPix * 3;
                  YCbCr[0] = buffer[ind2+1];
                  YCbCr[1] = buffer[ind2+0];
                  YCbCr[2] = buffer[ind2+2];
                  YCbCrToRGB(YCbCr, RGB);
                  memcpy(&(thePixels[indPix2]), RGB, 3);
                  YCbCr[0] = buffer[ind2+3];
                  YCbCrToRGB(YCbCr, RGB);
                  memcpy(&(thePixels[indPix2 + 3]), RGB, 3);
               }
               break;

            case v4l2_fourcc('G', 'B', 'R', 'G'):
               thePixels = rawImage->getActualPixel(0, 0);
               GBRGtoRGB(buffer, thePixels);
               break;

            default:
               std::cerr << "Pixel format not supported!\n";
         }
      }

      capture = false;    // We have used the captured image.
   }
}


void v4l2ImageGrabber::moveRight()
{
   std::cerr << "v4l2ImageGrabber::moveRight() not implemented...\n";
//    static unsigned int delta = 1;
//    if (vWindow.x >= delta)
//    {
//       vWindow.x = vWindow.x - delta;
//       useInputRectangle();
//    }
}

void v4l2ImageGrabber::moveLeft()
{
   std::cerr << "v4l2ImageGrabber::moveLeft() not implemented...\n";
//    static unsigned int delta = 1;
//    vWindow.x = vWindow.x + delta;
//    useInputRectangle();
}

void v4l2ImageGrabber::moveUp()
{
   std::cerr << "v4l2ImageGrabber::moveUp() not implemented...\n";
//    vWindow.y = vWindow.y + 1;
//    useInputRectangle();
}

void v4l2ImageGrabber::moveDown()
{
   std::cerr << "v4l2ImageGrabber::moveDown() not implemented...\n";
//    if (vWindow.y >= 1)
//    {
//       vWindow.y = vWindow.y - 1;
//       useInputRectangle();
//    }
}

void v4l2ImageGrabber::increaseWidth()
{
   std::cerr << "v4l2ImageGrabber::increaseWidth() not implemented...\n";
//    if (vWindow.width >= 1)
//    {
//       vWindow.width = vWindow.width - 1;
//       useInputRectangle();
//    }
}

void v4l2ImageGrabber::decreaseWidth()
{
   std::cerr << "v4l2ImageGrabber::decreaseWidth() not implemented...\n";
//    vWindow.width = vWindow.width + 1;
//    useInputRectangle();
}

void v4l2ImageGrabber::increaseHeight()
{
   std::cerr << "v4l2ImageGrabber::increaseHeight() not implemented...\n";
//    if (vWindow.height >= 1)
//    {
//       vWindow.height = vWindow.height - 1;
//       useInputRectangle();
//    }
}

void v4l2ImageGrabber::decreaseHeight()
{
   std::cerr << "v4l2ImageGrabber::decreaseHeight() not implemented...\n";
//    vWindow.height = vWindow.height + 1;
//    useInputRectangle();
}

bool v4l2ImageGrabber::setInputRectangle(int x, int y, int w, int h)
{
   std::cerr << "v4l2ImageGrabber::setInputRectangle() not implemented...\n";
   return(true);
   // Read default fields value
//    if (ioctl(fd, VIDIOCGWIN, &vWindow ) < 0 )
//    {
//       perror( "v4l2ImageGrabber::setInputRectangle() (VIDIOCGWIN)" );
//       return(false);
//    }
//    vWindow.x = x;
//    vWindow.y = y;
//    vWindow.width = w;
//    vWindow.height = h;
//    return(useInputRectangle());
}

bool v4l2ImageGrabber::useInputRectangle()
{
   std::cerr << "v4l2ImageGrabber::useInputRectangle() not implemented...\n";
   return(true);
//    if (ioctl(fd, VIDIOCSWIN, &vWindow) < 0)
//    {
//       perror("v4lImageGrabber::useInputRectangle() (VIDIOCSWIN)");
//       return(false);
//    }
//    else
//    {
//       return(true);
//    }
}

/*!
 * \brief Set the properties of the picture.
 *
 * If a parameter is negative, then it is not used.
 *
 * Values are given as a percentage.
 *
 * \param[in] bri Brightness
 * \param[in] col Colour
 * \param[in] con Contrast
 * \param[in] hue Hue
 * \param[in] whi Whiteness
 * \param[in] dep Depth
 * \return true if the function was successful.
 */
bool v4l2ImageGrabber::setPictureProperties(long bri,
                                            long col,
                                            long con,
                                            long hue,
                                            long whi,
                                            long dep)
{
   std::cerr << "v4l2ImageGrabber::setPictureProperties() not implemented...\n";
   return(true);
//    //Set brightness, contrast, saturation and hue
//    if (bri > 0)
//       vPicture.brightness = FG_PERCENT(static_cast<unsigned short>(bri));
//    if (col > 0)
//       vPicture.colour = FG_PERCENT(static_cast<unsigned short>(col));
//    if (con > 0)
//       vPicture.contrast = FG_PERCENT(static_cast<unsigned short>(con));
//    if (hue > 0)
//       vPicture.hue = FG_PERCENT(static_cast<unsigned short>(hue));
//    if (whi > 0)
//       vPicture.whiteness = FG_PERCENT(static_cast<unsigned short>(whi));
//    if (dep > 0)
//       vPicture.depth = FG_PERCENT(static_cast<unsigned short>(dep));
//    return(usePictureProperties());
}

bool v4l2ImageGrabber::usePictureProperties()
{
   std::cerr << "v4l2ImageGrabber::usePictureProperties() not implemented...\n";
   return(true);
//    if (ioctl(fd, VIDIOCSPICT, &vPicture) < 0 )
//    {
//       perror("v4l2ImageGrabber::usePictureProperties() (VIDIOCSPICT)");
//       return(false);
//    }
//    return(true);
}

/*!
 * \brief Initialise the grabber.
 * \param[in] imgWidth
 * \param[in] imgHeight The width and height of the image.
 * \param[in] upsideDown Set to true if the camera is upside-down.
 * \param[in] flip Set to true to horizontally flip the images.
 * \param[in] devNumber The number of the video device (/dev/video?).
 * \param[in] inputNumber The video input number.
 * \param[in] videoStandard The video standard number.
 * \param[in] ioMethod The io method to use.
 * \param[in] posX
 * \param[in] posY The position of the viewport of the grabbed part.
 * \param[in] width
 * \param[in] height The width and height of the grabbed area.
 */
void v4l2ImageGrabber::init(unsigned short imgWidth,
                            unsigned short imgHeight,
                            bool upsideDown,
                            bool flip,
                            int devNumber,
                            int inputNumber,
                            int videoStandard,
                            possibleIoMethods ioMethod,
                            unsigned short posX,
                            unsigned short posY,
                            unsigned short width,
                            unsigned short height)
{
   std::string funcName = "void v4l2ImageGrabber::init(...)";
   bool error = false;

   unsigned int inNum = inputNumber;
   unsigned int vidStd = videoStandard;

   // Open the video device
   std::cout << "Opening /dev/video" << devNumber << std::endl;
   std::ostringstream aStream;
   aStream << "/dev/video" << devNumber;
   deviceName = aStream.str();

   struct stat st;

   if (-1 == stat(deviceName.c_str(), &st))
   {
      std::cerr << std::endl << "-->" << funcName << std::endl
         << "\tCannot identify '" << deviceName << "': "
         << strerror(errno) << std::endl;
      error = true;
   }

   if (!error && (!S_ISCHR(st.st_mode)))
   {
      std::cerr << std::endl << "-->" << funcName << std::endl
        << "\t" << deviceName << " is no device\n";
      error = true;
   }

   if (!error)
   {
      fd = open(deviceName.c_str(), O_RDWR, 0);

      if (-1 == fd)
      {
         std::cerr << std::endl << "-->" << funcName << std::endl
            << "\tCannot open '" << deviceName << "': "<< strerror(errno)
            << std::endl;
         error = true;
      }
   }

   // Verify that we have a v4l2 device.
   if (!error && (-1 == xioctl(fd, VIDIOC_QUERYCAP, &hdwCap_)))
   {
      if (EINVAL == errno)
      {
         std::cerr << std::endl << "-->" << funcName << std::endl
            << "\t" << deviceName << " is no V4L2 device\n";
         error = true;
      }
      else
      {
         std::cerr << std::endl << "-->" << funcName << std::endl
            << "\tError in VIDIOC_QUERYCAP:" << std::endl
            << strerror(errno) << std::endl;
         error = true;
      }
   }

   // Verify this is a capture device
   if (!error && (!(hdwCap_.capabilities & V4L2_CAP_VIDEO_CAPTURE)))
   {
      std::cerr << std::endl << "-->" << funcName << std::endl
         << "\t" << deviceName << " is not a video capture device\n";
      error = true;
   }

   // Deal with the io method to use.
   ioMethod_ = ioMethod;
   if (!error && (IO_METHOD_READ == ioMethod_))
   {
      if (!(hdwCap_.capabilities & V4L2_CAP_READWRITE))
      {
         std::cerr << std::endl << "-->" << funcName << std::endl
            << "\tThe V4L2 device does not support the READ io method.\n";
         error = true;;
      }
   }
   if (!error && (IO_METHOD_MMAP == ioMethod_))
   {
      if (!(hdwCap_.capabilities & V4L2_CAP_STREAMING))
      {
         std::cout << std::endl << "-->" << funcName << std::endl
            << "\tThe V4L2 device does not support the MMAP io method.\n";
         error = true;;
      }
   }
   if (!error && (IO_METHOD_USERPTR == ioMethod_))
   {
      std::cerr << std::endl << "-->" << funcName << std::endl
         << "\tUser pointer io method not supported yet.\n";
      error = true;
   }

   // Let's see what the sources can be and choose.
   v4l2_input inputDesc;
   memset(&inputDesc, 0, sizeof(inputDesc));
   inputDesc.index = 0;
   if (!error)
   {
      std::cout << "Available inputs:\n";
      while (true)
      {
         if (-1 != xioctl(fd, VIDIOC_ENUMINPUT, &inputDesc))
         {
            std::cout << "\t" << inputDesc.index << ": "
               << inputDesc.name;
            if (inputDesc.index == inNum)
            {
               std::cout << " (used)";
            }
            std::cout << "\n";
         }
         else
         {
            break; // To get out of the while.
         }
         inputDesc.index++;
      }
      if (inNum < inputDesc.index)
      {
         if (-1 == ioctl(fd, VIDIOC_S_INPUT, &inNum))
         {
            perror ("VIDIOC_S_INPUT");
            error = true;
         }
      }
      else
      {
         std::cerr
            << "Input number can be at most " << inputDesc.index - 1
            << " (" << inNum << " given).\n";
         error = true;
      }
   }

   // Let's look at what video formats are available and chose.
   if (!error)
   {
      if (-1 == ioctl(fd, VIDIOC_G_INPUT, &inputDesc.index))
      {
         perror("VIDIOC_G_INPUT");
         error = true;
      }

      if (!error)
      {
         if (-1 == ioctl(fd, VIDIOC_ENUMINPUT, &inputDesc))
         {
            perror ("VIDIOC_ENUMINPUT");
            error = true;
         }
      }
      
//       if (inputDesc.std != 0) // USB exception...
//       {
      v4l2_standard standard;
      memset (&standard, 0, sizeof (standard));
      standard.index = 0;
      v4l2_std_id stdId;
      std::cout << "Current input supported standards:\n";
      while (0 == ioctl (fd, VIDIOC_ENUMSTD, &standard))
      {
         if (standard.id & inputDesc.std)
            std::cout << "\t" << standard.index << ": " << standard.name;
            if (standard.index == vidStd)
            {
               std::cout << " (used)";
               stdId = standard.id;
            }
            std::cout << "\n";
         standard.index++;
      }
      if (standard.index == 0)
      {
         std::cout << "\tNone reported, using default.\n";
      }
      else if (vidStd < standard.index)
      {
         if (-1 == ioctl (fd, VIDIOC_S_STD, &stdId))
         {
            perror ("VIDIOC_S_STD");
            error = true;
         }
      }
      else
      {
         std::cerr
            << "Video standard number can be at most "
            << static_cast<int>(standard.index) - 1
            << " (" << vidStd << " given).\n";
         error = true;
      }
   }

   // Let's look at what pixel formats are available.
   if (!error)
   {
      v4l2_fmtdesc fmtdesc;
      memset(&fmtdesc, 0, sizeof(fmtdesc));
      fmtdesc.index = 0;
      fmtdesc.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
      pixelFormat_ = 0;
      std::cout << "Available formats:\n";
      while (true)
      {
         if (-1 != xioctl(fd, VIDIOC_ENUM_FMT, &fmtdesc))
         {
            std::cout << "\t" << fmtdesc.description;
            std::vector<__u32>::const_iterator pos =
               find(supportedPixelFormats.begin(),
                    supportedPixelFormats.end(),
                    fmtdesc.pixelformat);
            if (pos != supportedPixelFormats.end())
            {
               std::cout << " (supported by imalib)";
               if (0 == pixelFormat_)
                  pixelFormat_ = *pos;
            }
            std::cout << std::endl;
         }
         else
         {
            break; // To get out of the while.
         }
         fmtdesc.index++;
      }
      if (0 == pixelFormat_)
      {
         std::cerr
            << "None of the available formats are supported by imalib.\n";
         error = true;
      }
   }

   if (!error)
   {
      // Video input, standard and some tuning.
      memset(&cropcap, 0, sizeof(cropcap));
      cropcap.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;

      // Some bits below need to be moved to the setRectangle() function.
      if (0 == ioctl(fd, VIDIOC_CROPCAP, &cropcap))
      {
//          std::cout << "Cropcap bounds " << cropcap.bounds.width
//            << "x" << cropcap.bounds.height
//            << " at (" << cropcap.bounds.left << "," << cropcap.bounds.top
//            << ")\n";
//          std::cout << "Cropcap defrect " << cropcap.defrect.width
//            << "x" << cropcap.defrect.height
//            << " at (" << cropcap.defrect.left << "," << cropcap.defrect.top
//            << ")\n";
         memset(&crop, 0, sizeof(crop));
         crop.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
//          crop.c = cropcap.defrect;
         crop.c.width = width;
         crop.c.height = height;
         crop.c.top = posY;
         crop.c.left = posX;
         std::cout << "Setting cropping to " << crop.c.width
           << "x" << crop.c.height
           << " at (" << crop.c.left << "," << crop.c.top << ")\n";
      }
      else
      {
         // This probably means that cropping is not supported so we just
         // use a default rectangle, which is probably not defined given that
         // the call to VIDIOC_CROPCAP failed...!!!.
//          std::cerr << std::endl << "-->" << funcName << std::endl
//             << "\tError in VIDIOC_CROPCAP:" << std::endl
//             << "\t" << strerror(errno) << std::endl;
//          std::cerr << "\tThe error above should not happen and probably\n"
//             << "\tmeans that cropping is not supported!!!\n";
         std::cout << std::endl << "-->" << funcName << std::endl
            << "\tCropping is probably not supported!\n";
         crop.c = cropcap.defrect;
      }

      if (-1 == ioctl(fd, VIDIOC_S_CROP, &crop))
      {
         switch (errno)
         {
            case EINVAL:
               /* Cropping not supported. */
               std::cout << std::endl << "-->" << funcName << std::endl
                  << "\tCropping is probably not supported (VIDIOC_S_CROP)!\n";
               break;
            default:
               // Seems to be harmless to ignore these...
//                std::cerr << std::endl << "-->" << funcName << std::endl
//                   << "\tError in VIDIOC_S_CROP:" << std::endl
//                   << "\t" << strerror(errno) << std::endl;
//                error = true;
               break;
         }
      }
      if (-1 == ioctl(fd, VIDIOC_G_CROP, &crop))
      {
         switch (errno)
         {
            case EINVAL:
               /* Cropping not supported. */
               std::cout << std::endl << "-->" << funcName << std::endl
                  << "\tCropping is probably not supported (VIDIOC_G_CROP)!\n";
               break;
            default:
               // Seems to be harmless to ignore these...
//                std::cerr << std::endl << "-->" << funcName << std::endl
//                   << "\tError in VIDIOC_G_CROP:" << std::endl
//                   << "\t" << strerror(errno) << std::endl;
//                error = true;
               break;
         }
      }
      else
      {
        std::cout << "Cropping set to " << crop.c.width
          << "x" << crop.c.height
          << " at (" << crop.c.left << "," << crop.c.top << ")\n";
      }
   }

   if (!error)
   {
      memset(&fmt, 0, sizeof(fmt));

      fmt.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
      fmt.fmt.pix.width = imgWidth;
      fmt.fmt.pix.height = imgHeight;
      fmt.fmt.pix.pixelformat = pixelFormat_;
      fmt.fmt.pix.field = V4L2_FIELD_INTERLACED;

      if (-1 == xioctl(fd, VIDIOC_S_FMT, &fmt))
      {
         std::cerr << std::endl << "-->" << funcName << std::endl
            << "\tError in VIDIOC_S_FMT" << std::endl
            << strerror(errno) << std::endl;
         error = true;
      }

// std::cout << "bytesperline " << fmt.fmt.pix.bytesperline << std::endl
// << "sizeimage " << fmt.fmt.pix.sizeimage << std::endl;

      // Buggy driver paranoia.
      // I don't see the point of that so I removed it!
//       unsigned int min = fmt.fmt.pix.width * 2;
//       if (fmt.fmt.pix.bytesperline < min)
//          fmt.fmt.pix.bytesperline = min;
//       min = fmt.fmt.pix.bytesperline * fmt.fmt.pix.height;
//       if (fmt.fmt.pix.sizeimage < min)
//          fmt.fmt.pix.sizeimage = min;

   }
   // Note VIDIOC_S_FMT may change width and height.
   // Some of the above should be in setRectangle().

// std::cout << "bytesperline " << fmt.fmt.pix.bytesperline << std::endl
// << "sizeimage " << fmt.fmt.pix.sizeimage << std::endl;

   if (!error)
   {
      if ((fmt.fmt.pix.width != imgWidth)
         || (fmt.fmt.pix.height != imgHeight))
      {
         std::cout << "-->" << funcName << std::endl
            << "\tImage size changed to " << fmt.fmt.pix.width << "x"
            << fmt.fmt.pix.height << " (from "
            << imgWidth << "x" << imgHeight
            << ")\n";
         imgWidth = fmt.fmt.pix.width;
         imgHeight = fmt.fmt.pix.height;
      }

      setParameters(imgWidth, imgHeight, upsideDown, flip);
   }

   // Do the memory initialisation.
   if (!error)
   {
      switch (ioMethod_)
      {
         case IO_METHOD_READ:
            error = !initRead(fmt.fmt.pix.sizeimage);
            break;
         case IO_METHOD_MMAP:
            error = !initMmap();
            break;
         case IO_METHOD_USERPTR:
            error = !initUserPtr(fmt.fmt.pix.sizeimage);
            break;
      }
   }

   // We now start the capture (if appropriate).
   if (!error)
   {
      enum v4l2_buf_type type;

      switch (ioMethod_)
      {
         case IO_METHOD_READ:
            // Nothing to do.
            break;

         case IO_METHOD_MMAP:
            for (unsigned int i = 0; i < nbBuffers; ++i)
            {
               struct v4l2_buffer buf;
               memset(&buf, 0, sizeof(buf));

               buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
               buf.memory = V4L2_MEMORY_MMAP;
               buf.index = i;

               if (-1 == xioctl(fd, VIDIOC_QBUF, &buf))
               {
                  std::cerr << "-->" << funcName << std::endl
                     << "\tError while queueing a buffer (VIDIOC_QBUF):\n"
                     << strerror(errno) << std::endl;
                  error = true;
                  break;
               }
            }

            if (!error)
            {
               type = V4L2_BUF_TYPE_VIDEO_CAPTURE;

               if (-1 == xioctl(fd, VIDIOC_STREAMON, &type))
               {
                  std::cerr << "-->" << funcName << std::endl
                     << "\tError when starting streaming (VIDIOC_STREAMON):\n"
                     << "\t\t" << strerror(errno) << std::endl;
                  error = true;
               }
               else
                  streaming_ = true;
            }
            break;

         case IO_METHOD_USERPTR:
//             for (i = 0; i < n_buffers; ++i) {
//                struct v4l2_buffer buf;
//
//                CLEAR (buf);
//
//                buf.type        = V4L2_BUF_TYPE_VIDEO_CAPTURE;
//                buf.memory      = V4L2_MEMORY_USERPTR;
//                buf.index       = i;
//                buf.m.userptr  = (unsigned long) buffers[i].start;
//                buf.length      = buffers[i].length;
//
//                if (-1 == xioctl (fd, VIDIOC_QBUF, &buf))
//                   errno_exit ("VIDIOC_QBUF");
//             }
//
//             type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
//
//             if (-1 == xioctl (fd, VIDIOC_STREAMON, &type))
//                errno_exit ("VIDIOC_STREAMON");
//
         break;
      }
   }

//    if (!error)
//    {
//       // We wait a bit, just to make sure everything is settled.
//       sleep(1);
//    }

   if (error)
   {
      std::cerr
         << "Something horribly wrong happened during initialisation of the frame"
         << " grabber.\n"
         << "\t--> connection aborted.\n";
      connected = false;
   }
}

/*!
 * \brief Initialise memory for the IO_METHOD_READ io method.
 * \param[in] bufferSize Needed buffer size.
 * \return true is all went well, false otherwise.
 */
bool v4l2ImageGrabber::initRead(unsigned int bufferSize)
{
   imageBuffers = new dataBuffer[1];

//    if (!buffers) {
//       fprintf (stderr, "Out of memory\n");
//       exit (EXIT_FAILURE);
//    }

   imageBuffers[0].length = bufferSize;
   imageBuffers[0].start = new unsigned char[bufferSize];

//    if (!buffers[0].start) {
//       fprintf (stderr, "Out of memory\n");
//       exit (EXIT_FAILURE);
//    }

   return(true);
}

/*!
 * \brief Initialise memory for the IO_METHOD_MMAP io method.
 * \return true is all went well, false otherwise.
 */
bool v4l2ImageGrabber::initMmap()
{
   std::string funcName = "void v4l2ImageGrabber::initMmap()";
   bool cont = true;

   struct v4l2_requestbuffers req;
   memset(&req, 0, sizeof(req));

   req.count = 4; // We want 4 buffers.
   req.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
   req.memory = V4L2_MEMORY_MMAP;

   if (-1 == xioctl(fd, VIDIOC_REQBUFS, &req))
   {
      std::cerr << std::endl << "-->" << funcName << std::endl;

      if (EINVAL == errno)
      {
         std::cerr << "\t" << deviceName
            << " does not support memory mapping\n";
      }
      else
      {
         std::cerr << "\tError while requesting buffers (VIDIOC_REQBUFS):\n"
            << strerror(errno) << std::endl;
      }
      cont = false;
   }
   // Do we really care if we have fewer than 2 buffers?
//    else
//    {
//       if (req.count < 2)
//       {
//       fprintf (stderr, "Insufficient buffer memory on %s\n",
//                dev_name);
//                exit (EXIT_FAILURE);
//       }
//    }

   if (cont)
   {
      imageBuffers = new dataBuffer[req.count];

      //    if (!buffers) {
      //       fprintf (stderr, "Out of memory\n");
      //       exit (EXIT_FAILURE);
      //    }

      for (nbBuffers = 0; nbBuffers < req.count; ++nbBuffers)
      {
         struct v4l2_buffer buf;
         memset(&buf, 0, sizeof(buf));

         buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
         buf.memory = V4L2_MEMORY_MMAP;
         buf.index = nbBuffers;

         if (-1 == xioctl(fd, VIDIOC_QUERYBUF, &buf))
         {
            std::cerr << "-->" << funcName << std::endl
               << "\tError while querying a buffer (VIDIOC_QUERYBUF):\n"
               << strerror(errno) << std::endl;
            cont = false;
            break;
         }

         imageBuffers[nbBuffers].length = buf.length;
         imageBuffers[nbBuffers].start =
            (unsigned char*)mmap(NULL /* start anywhere */,
                                 buf.length,
                                 PROT_READ | PROT_WRITE /* required */,
                                 MAP_SHARED /* recommended */,
                                 fd,
                                 buf.m.offset);

         if (MAP_FAILED == imageBuffers[nbBuffers].start)
         {
            std::cerr << std::endl << "-->" << funcName << std::endl
               << "\tMemory mapping error.\n";
            cont = false;
         }
      }
   }

   return(cont);
}

/*!
 * \brief Initialise memory for the IO_METHOD_USERPTR io method.
 *
 * This method is not yet supported.
 *
 * \param[in] bufferSize Needed buffer size.
 * \return true is all went well, false otherwise.
 */
bool v4l2ImageGrabber::initUserPtr(unsigned int bufferSize)
{
   std::cerr << "The IO_METHOD_USERPTR is not yet supported.\n";

//    struct v4l2_requestbuffers req;
//    unsigned int page_size;
//
//    page_size = getpagesize ();
//    buffer_size = (buffer_size + page_size - 1) & ~(page_size - 1);
//
//    CLEAR (req);
//
//    req.count               = 4;
//    req.type                = V4L2_BUF_TYPE_VIDEO_CAPTURE;
//    req.memory              = V4L2_MEMORY_USERPTR;
//
//    if (-1 == xioctl (fd, VIDIOC_REQBUFS, &req)) {
//       if (EINVAL == errno) {
//          fprintf (stderr, "%s does not support "
//          "user pointer i/o\n", dev_name);
//          exit (EXIT_FAILURE);
//       } else {
//          errno_exit ("VIDIOC_REQBUFS");
//       }
//    }
//
//    buffers = calloc (4, sizeof (*buffers));
//
//    if (!buffers) {
//       fprintf (stderr, "Out of memory\n");
//       exit (EXIT_FAILURE);
//    }
//
//    for (n_buffers = 0; n_buffers < 4; ++n_buffers) {
//       buffers[n_buffers].length = buffer_size;
//       buffers[n_buffers].start = memalign (/* boundary */ page_size,
//                                            buffer_size);
//
//                                            if (!buffers[n_buffers].start) {
//                                               fprintf (stderr, "Out of memory\n");
//                                               exit (EXIT_FAILURE);
//                                            }
//    }

   return(false);
}

/*!
 * \brief Convert a Bayer GBRG to RGB image.
 *
 * \param[in] gbrg The array of GBRG data.
 * \param[out] rgb The resulting RGB image.
 */
void v4l2ImageGrabber::GBRGtoRGB(unsigned char* gbrg, unsigned char* rgb)
{
   assert((0 == (imgMap->sourceWidth() % 2))
          && (0 == (imgMap->sourceHeight() % 2)));

   unsigned short indX, indY;
   unsigned long indRGB, indGBRG;

   // Top left corner.
   rgb[0] = gbrg[imgMap->sourceWidth()];
   rgb[1] = gbrg[0];
   rgb[2] = gbrg[1];

   // Top row minus corners, by pairs of pixels.
   indGBRG = 0;
   indRGB = 0;
   for (indX = 1; indX < imgMap->sourceWidth()-1; indX += 2)
   {
      // Left pixel
      indGBRG++;
      indRGB += 3;
      rgb[indRGB] = (gbrg[indGBRG+imgMap->sourceWidth()-1]
                     + gbrg[indGBRG+imgMap->sourceWidth()+1]) / 2;
      rgb[indRGB+1] = (gbrg[indGBRG-1] + gbrg[indGBRG+1]
                       + gbrg[indGBRG+imgMap->sourceWidth()]) / 3;
      rgb[indRGB+2] = gbrg[indGBRG];
      // Right pixel.
      indGBRG++;
      indRGB += 3;
      rgb[indRGB] = gbrg[indGBRG+imgMap->sourceWidth()];
      rgb[indRGB+1] = gbrg[indGBRG];
      rgb[indRGB+2] = (gbrg[indGBRG-1] + gbrg[indGBRG+1]) / 2;
   }

   // Top right corner.
   indGBRG = imgMap->sourceWidth() - 1;
   indRGB = indGBRG * 3;
   rgb[indRGB] = gbrg[indGBRG+imgMap->sourceWidth()-1];
   rgb[indRGB+1] = (gbrg[indGBRG-1] + gbrg[indGBRG+imgMap->sourceWidth()]) / 2;
   rgb[indRGB+2] = gbrg[indGBRG];

   // Left column minus corners, by pairs of pixels.
   indGBRG = 0;
   for (indY = 1; indY < imgMap->sourceHeight()-1; indY += 2)
   {
      // Top pixel
      indGBRG += imgMap->sourceWidth();
      indRGB = indGBRG * 3;
      rgb[indRGB] = gbrg[indGBRG];
      rgb[indRGB+1] = (gbrg[indGBRG-imgMap->sourceWidth()]
                       + gbrg[indGBRG+imgMap->sourceWidth()]
                       + gbrg[indGBRG+1]) / 3;
      rgb[indRGB+2] = (gbrg[indGBRG-imgMap->sourceWidth()+1]
                       + gbrg[indGBRG+imgMap->sourceWidth()+1]) / 2;
      // Bottom pixel
      indGBRG += imgMap->sourceWidth();
      indRGB = indGBRG * 3;
      rgb[indRGB] = (gbrg[indGBRG-imgMap->sourceWidth()]
                     + gbrg[indGBRG+imgMap->sourceWidth()]) / 2;
      rgb[indRGB+1] = gbrg[indGBRG];
      rgb[indRGB+2] = gbrg[indGBRG+1];
   }

   // Bottom left corner
   indGBRG = (imgMap->sourceHeight() - 1) * imgMap->sourceWidth();
   indRGB = indGBRG * 3;
   rgb[indRGB] = gbrg[indGBRG];
   rgb[indRGB+1] = (gbrg[indGBRG-imgMap->sourceWidth()] + gbrg[indGBRG+1]) / 2;
   rgb[indRGB+2] = gbrg[indGBRG-imgMap->sourceWidth()+1];

   // Bottom row minus corners, by pairs of pixels
   indGBRG = (imgMap->sourceHeight() - 1) * imgMap->sourceWidth();
   for (indX = 1; indX < imgMap->sourceWidth()-1; indX += 2)
   {
      // Left pixel.
      indGBRG++;
      indRGB = indGBRG * 3;
      rgb[indRGB] = (gbrg[indGBRG-1] + gbrg[indGBRG+1]) / 2;
      rgb[indRGB+1] = gbrg[indGBRG];
      rgb[indRGB+2] = gbrg[indGBRG-imgMap->sourceWidth()];
      // Right pixel.
      indGBRG++;
      indRGB += 3;
      rgb[indRGB] = gbrg[indGBRG];
      rgb[indRGB+1] = (gbrg[indGBRG-1]
                       + gbrg[indGBRG-imgMap->sourceWidth()]
                       + gbrg[indGBRG+1]) / 3;
      rgb[indRGB+2] = (gbrg[indGBRG-imgMap->sourceWidth()-1]
                       + gbrg[indGBRG-imgMap->sourceWidth()+1]) / 2;
   }

   // Bottom right corner
   indGBRG = (imgMap->sourceHeight() * imgMap->sourceWidth()) - 1;
   indRGB = indGBRG * 3;
   rgb[indRGB] = gbrg[indGBRG-1];
   rgb[indRGB+1] = gbrg[indGBRG];
   rgb[indRGB+2] = gbrg[indGBRG-imgMap->sourceWidth()];

   // Right column minus corners, by pairs of pixels
   indGBRG = imgMap->sourceWidth() - 1;
   for (indY = 1; indY < imgMap->sourceHeight()-1; indY += 2)
   {
      // Top pixel
      indGBRG += imgMap->sourceWidth();
      indRGB = indGBRG * 3;
      rgb[indRGB] = gbrg[indGBRG-1];
      rgb[indRGB+1] = gbrg[indGBRG];
      rgb[indRGB+2] = (gbrg[indGBRG-imgMap->sourceWidth()]
                       + gbrg[indGBRG+imgMap->sourceWidth()]) / 2;
      // Bottom pixel.
      indGBRG += imgMap->sourceWidth();
      indRGB = indGBRG * 3;;
      rgb[indRGB] = (gbrg[indGBRG-imgMap->sourceWidth()-1]
                       + gbrg[indGBRG+imgMap->sourceWidth()-1]) / 2;
      rgb[indRGB+1] = (gbrg[indGBRG-imgMap->sourceWidth()]
                       + gbrg[indGBRG-1]
                       + gbrg[indGBRG+imgMap->sourceWidth()]) / 3;
      rgb[indRGB+2] = gbrg[indGBRG];
   }

   // Centre groups of 4 pixels.
   for (indY = 1; indY < imgMap->sourceHeight()-1; indY += 2)
   {
      indGBRG = indY * imgMap->sourceWidth();
      indRGB = indGBRG * 3;
      for (indX = 1; indX < imgMap->sourceWidth()-1; indX += 2)
      {
         // Top left pixel.
         indGBRG++;
         indRGB += 3;
         rgb[indRGB] = (gbrg[indGBRG-1] + gbrg[indGBRG+1]) / 2;
         rgb[indRGB+1] = gbrg[indGBRG];
         rgb[indRGB+2] = (gbrg[indGBRG-imgMap->sourceWidth()]
                          + gbrg[indGBRG+imgMap->sourceWidth()]) / 2;

         // Bottom left pixel.
         indGBRG += imgMap->sourceWidth();
         indRGB = indGBRG * 3;
         rgb[indRGB] = (gbrg[indGBRG-imgMap->sourceWidth()-1]
                        + gbrg[indGBRG-imgMap->sourceWidth()+1]
                        + gbrg[indGBRG+imgMap->sourceWidth()-1]
                        + gbrg[indGBRG+imgMap->sourceWidth()+1]) / 4;
         rgb[indRGB+1] = (gbrg[indGBRG-imgMap->sourceWidth()]
                        + gbrg[indGBRG+1]
                        + gbrg[indGBRG-1]
                        + gbrg[indGBRG+imgMap->sourceWidth()]) / 4;
         rgb[indRGB+2] = gbrg[indGBRG];

         // Bottom right pixel.
         indGBRG++;
         indRGB += 3;
         rgb[indRGB] = (gbrg[indGBRG-imgMap->sourceWidth()]
                        + gbrg[indGBRG+imgMap->sourceWidth()]) / 2;
         rgb[indRGB+1] = gbrg[indGBRG];
         rgb[indRGB+2] = (gbrg[indGBRG-1] + gbrg[indGBRG+1]) / 2;

         // Top right pixel.
         indGBRG -= imgMap->sourceWidth();
         indRGB = indGBRG * 3;
         rgb[indRGB] = gbrg[indGBRG];
         rgb[indRGB+1] = (gbrg[indGBRG-imgMap->sourceWidth()]
                          + gbrg[indGBRG+1]
                          + gbrg[indGBRG-1]
                          + gbrg[indGBRG+imgMap->sourceWidth()]) / 4;
         rgb[indRGB+2] = (gbrg[indGBRG-imgMap->sourceWidth()-1]
                          + gbrg[indGBRG-imgMap->sourceWidth()+1]
                          + gbrg[indGBRG+imgMap->sourceWidth()-1]
                          + gbrg[indGBRG+imgMap->sourceWidth()+1]) / 4;
      }
   }
}

} // namespace imalib.
