/*
 * imalib
 * 
 * Copyright 2017 Aberystwyth University
 * 
 * This library and program are free software: you can redistribute them
 * and/or modify them under the terms of the GNU General Public License as 
 * published by the Free Software Foundation, either version 3 of the License,
 * or (at your option) any later version.
 * 
 * The program and library are distributed in the hope that they will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*!
 * \file v4lImageGrabber.cpp
 * \brief An image grabber for a V4L frame grabber.  Implementation.
 *
 * \author Fred Labrosse <ffl@aber.ac.uk>
 * \copyright 2017 Aberystwyth University
 */

#include "imalib/imageRGB.h"
#include "imalib/v4lImageGrabber.h"

#include <sys/mman.h>
#include <math.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <stdio.h>
#include <unistd.h>

#include <iostream>
#include <fstream>

#include <boost/program_options.hpp>
namespace po = boost::program_options;

namespace imalib
{
const unsigned int DEFAULT_PALETTE = VIDEO_PALETTE_RGB24;
// #define DEFAULT_SYNC SYNC_EXTERN
const unsigned short DEFAULT_NORM = VIDEO_MODE_PAL; // PAL

#define MAXINPUTS 100

// Percentage of a unsigned short
#define FG_PERCENT(n)           ((n)*65535/100)
#define FG_50PC                 FG_PERCENT(50)

const unsigned short DEFAULT_BRIGHTNESS = 50;
const unsigned short DEFAULT_CONTRAST = 50;
const unsigned short DEFAULT_SATURATION = 50;
const unsigned short DEFAULT_HUE = 50;
const unsigned short DEFAULT_WHITENESS = 50;
const unsigned short DEFAULT_DEPTH = 50;

/*!
 * \brief Constructor.
 *
 * \param imgWidth
 * \param imgHeight (IN) The width and height of the image.
 * \param[in] upsideDown Set to true if the camera is upside-down.
 * \param[in] flip Set to true to horizontally flip the images.
 * \param devNumber (IN) The number of the video device (/dev/video?).
 * \param devPort (IN) The number of the port on the device.
 * \param posX
 * \param posY (IN) The position of the viewport of the grabbed part.
 * \param width
 * \param height (IN) The width and height of the grabbed area.
 */
v4lImageGrabber::v4lImageGrabber(unsigned short imgWidth,
                                 unsigned short imgHeight,
                                 bool upsideDown,
                                 bool flip,
                                 int devNumber, int devPort,
                                 unsigned short posX,
                                 unsigned short posY,
                                 unsigned short width,
                                 unsigned short height)
{
   init(imgWidth, imgHeight,
        upsideDown,
        flip,
        devNumber, devPort,
        posX, posY, width, height);
}

/*!
 * \brief Constructor.
 * \param[in] fileName Name of the configuration file.
 */
v4lImageGrabber::v4lImageGrabber(const std::string& fileName)
{
   // Let's extract the options from the configuration file.
   po::options_description optionsDesc("v4lImageGrabber options");
   int vidDev = 0;
   int vidPo = 0;
   int poX = 0;
   int poY = 0;
   int soWidth = IMAGE_DEFAULT_INPUTWIDTH;
   int soHeight = IMAGE_DEFAULT_INPUTHEIGHT;
   int imWidth = 200;
   int imHeight = 200;
   bool upDown = false;
   bool fl = false;
   optionsDesc.add_options()
      ("videoDevice", po::value<int>(&vidDev)->default_value(vidDev), "")
      ("videoPort", po::value<int>(&vidPo)->default_value(vidPo), "")
      ("posX", po::value<int>(&poX)->default_value(poX), "")
      ("posY", po::value<int>(&poY)->default_value(poY), "")
      ("sourceWidth",
       po::value<int>(&soWidth)->default_value(soWidth), "")
      ("sourceHeight",
       po::value<int>(&soHeight)->default_value(soHeight), "")
      ("imageWidth", po::value<int>(&imWidth)->default_value(imWidth), "")
      ("imageHeight", po::value<int>(&imHeight)->default_value(imHeight), "")
      ("upsideDown", po::value<bool>(&upDown)->default_value(upDown), "")
      ("flip", po::value<bool>(&fl)->default_value(fl), "")
      ;

   po::variables_map vm;
   std::ifstream configFile(fileName.c_str());
   if (configFile.is_open())
   {
      po::store(po::parse_config_file(configFile, optionsDesc), vm);
      po::notify(vm);
      configFile.close();
   }
   else
   {
      std::cerr << "v4lImageGrabber::v4lImageGrabber():\n"
         << "WARNING: could not open configuration file " << fileName
         << "!\n"
         << "\tUsing defaults.\n";
   }

   init(imWidth, imHeight,
        upDown,
        fl,
        vidDev, vidPo,
        poX, poY,
        soWidth, soHeight);
}


/*!
\brief Constructor.
*/
v4lImageGrabber::~v4lImageGrabber()
{
   if (fd)
      close(fd);
}

/*!
  \brief Display on the standard output some information about the
  source of the images.
*/
void v4lImageGrabber::printInfoAboutSource()
{
   char deviceName[255];
   video_capability vcap;
   video_channel vchan;
   bool wasOpen = false;

   std::cout << "----------------------------------------------\n"
      << "Image source:\n";

   // Open the video device
   sprintf(deviceName, "/dev/video%d", deviceNumber);
   if (fd <= 0)
   {
      wasOpen = false;
      fd = open(deviceName, O_RDWR);
      if (fd <= 0)
      {
         perror("Open");
         connected = false;
      }
   }
   else
      wasOpen = true;
   if (fd > 0)
   {
      // Retrieve the number of input ports
      if(ioctl(fd, VIDIOCGCAP, &vcap) < 0)
      {
         perror("VIDIOCGCAP");
         connected = false;
      }
      else
      {
         std::cout << "\tgrabber: " << vcap.name << std::endl;

         int numInputs = vcap.channels;
         char inputNames[MAXINPUTS][25];

         //Retrieve the name of each input
         for (int i = 0 ; i < numInputs ; i++)
         {
            vchan.channel = i;
            if(ioctl(fd, VIDIOCGCHAN, &vchan) < 0)
            {
               perror("VIDIOCGCHAN");
               connected = false;
               i = numInputs;
            }
            else
            {
               strcpy(inputNames[i], vchan.name);
            }
         }

         std::cout << "\tvideo port: "
            << inputNames[vChannel.channel]
            << std::endl;

         // Window
         video_window aWindow;
         if (ioctl(fd, VIDIOCGWIN, &aWindow ) < 0 )
         {
            perror( "v4lImageGrabber::printInfoAboutSource() (VIDIOCGWIN)" );
            connected = false;
         }
         else
         {
            std::cout << "\tgrabbed image size: " << aWindow.width
               << "x" << aWindow.height
               << " at (" << aWindow.x << "," << aWindow.y << ")"
               << std::endl;
         }

      }
      if (!wasOpen)
         close(fd);
   }
//    std::cout << "\tprovided image size: " << imageWidth << "x" << imageHeight
//       << std::endl;
//    if (_unwrap)
//    {
//       std::cout << "Images are unwrapped:\n"
//          << "\tsize:               " << imgMap->targetWidth << "x"
//          << imgMap->targetHeight << std::endl
//          << "\tinner radius:       " << _innerRadius << std::endl
//          << "\tangular resolution: " << _angularRes << std::endl
//          << "\tpower:              " << _unwrappingPower << std::endl
//          << "\tdirection:          " << _direction << std::endl;
//    }
//    if (_blur)
//    {
//       std::cout << "Images are blurred:\n"
//          << "\tblur size: " << _blurWidth << "x" << _blurHeight
//          << std::endl;
//    }
//    if (_upsideDown)
//    {
//       std::cout << "Images are upside down\n";
//    }
   printBasicInfoAboutSource();
   std::cout << "----------------------------------------------\n";
}


/*!
\brief Capture the image from a grabber.

This is there because the capture can take a while.  We thus first
capture and then use (with getImage()).
*/
void v4lImageGrabber::captureImage()
{
   if (connected)
   {
      if (ioctl(fd, VIDIOCMCAPTURE, &vMmap) < 0)
      {
         perror("v4lImageGrabber::captureImage() (VIDIOCMCAPTURE)");
         connected = false;
         capture = false;
      }
      else
         capture = true;
   }
}

/*!
 * \brief Get an image from the frame grabber.
 *
 * The image might have been pre-captured (with captureImage()), but
 * if not will be captured here.
 *
 * The images are resized here.
 *
 * \param[out] theImage The image in which to save the grabbed image, if
 * not NULL.
 * \param[out] rawImage The image in which to save the raw image (before any
 * transformation is applied), if not NULL.
 */
void v4lImageGrabber::getImage(imageRGB* theImage, imageRGB* rawImage)
{
   if (connected)
   {
      if (!capture)
         captureImage();

      // Wait for the capture to be finished.
      if (ioctl(fd, VIDIOCSYNC, &vMmap.frame) < 0)
      {
         perror("VIDIOCSYNC");
         connected = false;
      }
      else
      {
         unsigned long index;
         if (theImage)
         {
            theImage->setSize(imgMap->targetWidth(),
                              imgMap->targetHeight());
            std::vector<imageMapping::coordsPair>::iterator iter;
            for (iter = imgMap->map.begin();
                 iter < imgMap->map.end();
                 ++iter)
            {
               index = (iter->sY * imgMap->sourceWidth() + iter->sX) * 3;
               theImage->setPixelBGR(iter->tX, iter->tY,
                                     &grabbedImage[index]);
            }
         }
         if (rawImage)
         {
            rawImage->setSize(imgMap->sourceWidth(),
                              imgMap->sourceHeight());
            for (unsigned short indX = 0;
                 indX < imgMap->sourceWidth();
                 ++indX)
               for (unsigned short indY = 0;
                    indY < imgMap->sourceHeight();
                    ++indY)
               {
                  index = (indY * imgMap->sourceWidth() + indX) * 3;
                  rawImage->setPixelBGR(indX, indY,
                                        &grabbedImage[index]);
               }
         }
         capture = false;    // We have used the captured image.
      }
   }
}


void v4lImageGrabber::moveRight()
{
   static unsigned int delta = 1;
   if (vWindow.x >= delta)
   {
      vWindow.x = vWindow.x - delta;
      useInputRectangle();
   }
}

void v4lImageGrabber::moveLeft()
{
   static unsigned int delta = 1;
   vWindow.x = vWindow.x + delta;
   useInputRectangle();
}

void v4lImageGrabber::moveUp()
{
   vWindow.y = vWindow.y + 1;
   useInputRectangle();
}

void v4lImageGrabber::moveDown()
{
   if (vWindow.y >= 1)
   {
      vWindow.y = vWindow.y - 1;
      useInputRectangle();
   }
}

void v4lImageGrabber::increaseWidth()
{
   if (vWindow.width >= 1)
   {
      vWindow.width = vWindow.width - 1;
      useInputRectangle();
   }
}

void v4lImageGrabber::decreaseWidth()
{
   vWindow.width = vWindow.width + 1;
   useInputRectangle();
}

void v4lImageGrabber::increaseHeight()
{
   if (vWindow.height >= 1)
   {
      vWindow.height = vWindow.height - 1;
      useInputRectangle();
   }
}

void v4lImageGrabber::decreaseHeight()
{
   vWindow.height = vWindow.height + 1;
   useInputRectangle();
}

bool v4lImageGrabber::setInputRectangle(int x, int y, int w, int h)
{
   // Read default fields value
   if (ioctl(fd, VIDIOCGWIN, &vWindow ) < 0 )
   {
      perror( "v4lImageGrabber::setInputRectangle() (VIDIOCGWIN)" );
      return(false);
   }
   vWindow.x = x;
   vWindow.y = y;
   vWindow.width = w;
   vWindow.height = h;
   return(useInputRectangle());
}

bool v4lImageGrabber::useInputRectangle()
{
   if (ioctl(fd, VIDIOCSWIN, &vWindow) < 0)
   {
      perror("v4lImageGrabber::useInputRectangle() (VIDIOCSWIN)");
      return(false);
   }
   else
   {
      return(true);
   }
}

/*!
\brief Set the properties of the picture.

   If a parameter is negative, then it is not used.

Values are given as a percentage.

\param[in] bri Brightness
\param[in] col Colour
\param[in] con Contrast
\param[in] hue Hue
\param[in] whi Whiteness
\param[in] dep Depth
\return true if the function was successful.
*/
bool v4lImageGrabber::setPictureProperties(long bri,
                                           long col,
                                           long con,
                                           long hue,
                                           long whi,
                                           long dep)
{
   //Set brightness, contrast, saturation and hue
   if (bri > 0)
      vPicture.brightness =  FG_PERCENT(static_cast<unsigned short>(bri));
   if (col > 0)
      vPicture.colour = FG_PERCENT(static_cast<unsigned short>(col));
   if (con > 0)
      vPicture.contrast = FG_PERCENT(static_cast<unsigned short>(con));
   if (hue > 0)
      vPicture.hue = FG_PERCENT(static_cast<unsigned short>(hue));
   if (whi > 0)
      vPicture.whiteness = FG_PERCENT(static_cast<unsigned short>(whi));
   if (dep > 0)
      vPicture.depth = FG_PERCENT(static_cast<unsigned short>(dep));
   return(usePictureProperties());
}

bool v4lImageGrabber::usePictureProperties()
{
   if (ioctl(fd, VIDIOCSPICT, &vPicture) < 0 )
   {
      perror("v4lImageGrabber::usePictureProperties() (VIDIOCSPICT)");
      return(false);
   }
   return(true);
}

/*!
 * \brief Initialise the grabber.
 * \param imgWidth
 * \param imgHeight (IN) The width and height of the image.
 * \param[in] upsideDown Set to true if the camera is upside-down.
 * \param[in] flip Set to true to horizontally flip the images.
 * \param devNumber (IN) The number of the video device (/dev/video?).
 * \param portNumber (IN) The number of the port on the device.
 * \param posX
 * \param posY (IN) The position of the viewport of the grabbed part.
 * \param width
 * \param height (IN) The width and height of the grabbed area.
 */
void v4lImageGrabber::init(unsigned short imgWidth,
                           unsigned short imgHeight,
                           bool upsideDown,
                           bool flip,
                           int devNumber,
                           int portNumber,
                           unsigned short posX,
                           unsigned short posY,
                           unsigned short width,
                           unsigned short height)
{
   bool error = true;

   deviceNumber = devNumber;
   setImageSize(imgWidth, imgHeight);
   setUpsideDown(upsideDown);

   //set input channel number and norm (PAL/NTSC etc)
   vChannel.channel = portNumber;
   vChannel.norm = DEFAULT_NORM;

   // Set output properties
   vMmap.frame = 0;
   vMmap.width = imgMap->sourceWidth();
   vMmap.height = imgMap->sourceHeight();
   vMmap.format = DEFAULT_PALETTE;

   // Open the video device
   char deviceName[255];
   std::cout << "Opening /dev/video" << devNumber << std::endl;
   sprintf(deviceName, "/dev/video%d", devNumber);
   fd = open(deviceName, O_RDWR);
   if (fd <= 0)
   {
      perror("Open");
   }
   else
   {
      //Set brightness, contrast, saturation and hue
      vPicture.palette = DEFAULT_PALETTE;
      if (!setPictureProperties(DEFAULT_BRIGHTNESS,
                                DEFAULT_SATURATION,
                                DEFAULT_CONTRAST,
                                DEFAULT_HUE,
                                DEFAULT_WHITENESS,
                                DEFAULT_DEPTH))
      {
         std::cerr
            << "v4lImageGrabber::init(): Warning:\n"
            << "\tcould not set the pictures properties.\n";
      }
      if (!setInputRectangle(posX, posY,
                             width, height))
      {
         std::cerr
            << "v4lImageGrabber::init(): Warning:\n"
            << "\tcould not set the input rectangle.\n";
      }
      if (ioctl(fd, VIDIOCSCHAN, &vChannel) < 0)
      {
         perror("v4lImageGrabber::init() (VIDIOCSCHAN)");
      }
      else
      {
         //Get the buffer props
         video_mbuf mbuf;

         if(ioctl(fd, VIDIOCGMBUF, &mbuf) < 0)
         {
            perror("v4lImageGrabber::init() (VIDIOCGMBUF)");
         }
         else
         {
            //map the buffer
            grabbedImage = (unsigned char*)mmap(0, mbuf.size,
                                                PROT_READ|PROT_WRITE,
                                                MAP_SHARED,
                                                fd, 0);
            if (!grabbedImage)
            {
               perror("v4lImageGrabber::init() (mmap)");
            }
            else
            {
               grabbedImage = grabbedImage + mbuf.offsets[0];
               error = false;
            }
         }
      }
   }
   if (error)
   {
      std::cerr
         << "Something horribly wrong happened during initialisation of the frame"
         << " grabber.\n"
         << "\t--> connection aborted.\n";
      connected = false;
   }
}

} // namespace imalib.
